/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include <json-c/json.h>
#include <stdnoreturn.h>

#include <doca_log.h>
#include <doca_version.h>

#include "doca_argp.h"

DOCA_LOG_REGISTER(ARGP);

#define MAX_SERVER_ADDRESS 24			/* Maximum length for gRPC server address string */
#define MAX_PROGRAM_NAME 32			/* Maximal length for program name */
#define MAX_STRING_LEN 255			/* Maximal length for string in json file */
#define MAX_PARAMS_LEN 16			/* Maximal number for DOCA params */
#define MAX_DPDK_MISC_FLAGS_LEN 128		/* Maximal length for DPDK misc flags string */
#define DPDK_PARAM_LEN 128			/* Maximal length for DPDK flag format */
#define DPDK_PARAM_SIZE 20			/* Maximal number for DPDK flags on argv buffer */
#define DPDK_DEVICE_JSON_KEYS_SIZE 4		/* Maximal number for DPDK device keys on json file */
/* Command line length in json mode is equal to 3 {bin file, -j/--json, json file path} */
#define JSON_MODE_CLI_LEN 3
/* Command line length for general flags {-h,-v} is equal to 3 {bin file, --, -h/-v} */
#define GENERAL_FLAGS_CLI_LEN 3

/* DOCA general flags values as provided to the program */
struct doca_argp_program_general_config {
	int log_level;					/* The log level as provided by the user */
	char grpc_address[MAX_SERVER_ADDRESS];		/* The gRPC server address as provided by the user */
};

/* @brief Program flag information */
struct doca_argp_param {
	const char *short_flag;				/*Flag long name */
	const char *long_flag;				/* Flag short name */
	const char *arguments;				/* Flag expected arguments */
	const char *description;			/* Flag description */
	callback_func callback;				/* Flag program callback */
	enum doca_argp_type arg_type;			/* Flag argument type */
	bool is_mandatory;				/* Is flag mandatory for the program */
	bool is_cli_only;				/* Is flag supported only in cli mode */
	/* Internal fields */
	int nb_appears;					/* Will be used for input validation */
};

/* Information about program configuration */
struct doca_argp_program_type_config {
	bool is_dpdk;					/* Is the program based on DPDK API */
	bool is_grpc;					/* Is the program based on gRPC API */
};

struct argp_dpdk_ctx {
	int index;					/* Index of next available place in the buffer */
	dpdk_callback cb;				/* Program callback to forward DPDK flags */
	char buffer[DPDK_PARAM_SIZE][DPDK_PARAM_LEN];	/* DPDK arguments buffer for rte_eal_init call */
};
struct argp_params_array {
	int size;					/* Array size */
	struct doca_argp_param array[MAX_PARAMS_LEN];	/* Array of params */
};
struct argp_ctx {
	struct argp_dpdk_ctx dpdk_ctx;				/* DPDK context for json mode */
	void *program_config;					/* Pointer to program configuration struct */
	validation_callback program_validation_callback;	/* Program validation function */
	char program_name[MAX_PROGRAM_NAME];			/* Program name, for usage uses */
	struct doca_argp_program_type_config type_config;	/* Program configuration type */
	struct doca_argp_program_general_config general_config;	/* Store values of general flags as user provided */
	struct argp_params_array dpdk_params;			/* DPDK params array */
	struct argp_params_array general_params;		/* DOCA general params array */
	struct argp_params_array program_params;		/* DOCA program params array */
};

static struct argp_ctx parser_ctx;

/*
 * Checks if current object contains valid keys, function accepts list of valid keys should be in the object.
 * json_obj: json object, checking its keys.
 * params: set of valid keys.
 * params_size: set size.
 */
static doca_error_t
json_keys_validation(struct json_object *json_obj, const char **params, int params_size)
{
	bool match;
	int param_index;
	const char *key_name;
	struct json_object_iterator it = json_object_iter_begin(json_obj);
	struct json_object_iterator itEnd = json_object_iter_end(json_obj);

	/* Pass over current object keys, and check its validation */
	while (!json_object_iter_equal(&it, &itEnd)) {
		match = false;
		key_name = json_object_iter_peek_name(&it);

		/* Pass over params and checking if there is a match */
		for (param_index = 0; param_index < params_size; param_index++) {
			if (strcmp(key_name, params[param_index]) == 0) {
				match = true;
				break;
			}
		}
		if (!match) {
			DOCA_LOG_ERR("The param \"%s\" is not supported.", key_name);
			return DOCA_ERROR_NOT_SUPPORTED;
		}
		json_object_iter_next(&it);
	}
	return DOCA_SUCCESS;
}

/*
 * Print usage depends on params type.
 * params: array of params.
 */
static void
usage_print(struct argp_params_array *params)
{
	int i, num_written = 0;
	bool has_short_flag = false;
	int cli_alignment_length = parser_ctx.type_config.is_grpc ? 42 : 36;

	for (i = 0; i < params->size; i++) {
		has_short_flag = false;
		num_written = printf("  ");
		if (params->array[i].short_flag != NULL) {
			num_written += printf("-%s", params->array[i].short_flag);
			has_short_flag = true;
		}
		if (params->array[i].long_flag != NULL) {
			if (has_short_flag)
				num_written += printf(", ");
			num_written += printf("--%s ", params->array[i].long_flag);
		}
		if (params->array[i].arguments != NULL)
			num_written += printf("%s ", params->array[i].arguments);
		if (num_written < cli_alignment_length)
			printf("%*s", cli_alignment_length - num_written, " ");
		printf("%s\n", params->array[i].description);
	}
}

/*
 * Print program arguments and exit.
 */
static void
argp_usage()
{
	doca_argp_usage();
	exit(0);
}

/*
 * Print program arguments.
 */
void
doca_argp_usage()
{
	printf("\n\nUsage: doca_%s", parser_ctx.program_name);
	if (parser_ctx.type_config.is_dpdk)
		printf(" [DPDK Flags] --");
	if (parser_ctx.general_params.size != 0)
		printf(" [DOCA Flags]");
	if (parser_ctx.program_params.size != 0)
		printf(" [Program Flags]");
	printf("\n");

	if (parser_ctx.general_params.size != 0) {
		printf("\nDOCA Flags:\n");
		usage_print(&parser_ctx.general_params);
	}
	if (parser_ctx.program_params.size != 0) {
		printf("\nProgram Flags:\n");
		usage_print(&parser_ctx.program_params);
	}
	printf("\n\n");
}

static void
print_version(void)
{
	printf("DOCA Version: %s\n", doca_version());
}

/*
 * Remove comments from json file.
 * buffer: function output, will contain Json file content without comments.
 * size: size of the buffer.
 * json_fp: pointer to the json file.
 */
static void
delete_comments_from_json_file(char *buffer, size_t size, FILE *json_fp)
{
	char *ptr = buffer;
	int curr = fgetc(json_fp);
	int prev = ' ';

	while (curr != EOF && ptr < buffer + size) {
		if (prev == '/' && curr == '/') {
			while (curr != EOF && curr != '\n')
				curr = fgetc(json_fp);
			/* Delete the first '/' in the buffer */
			ptr--;
		} else {
			*ptr = curr;
			ptr++;
			prev = curr;
			curr = fgetc(json_fp);
		}
	}
}

/*
 * Create the correct format for dpdk device type (flag -a).
 * buf: output param.
 * device: device type (sf, regex ...).
 * id: device id (if sf, id is sf number, ...)
 * sft_value: sft value (enabled, disable).
 * sft_en: indicate if user insert sft value for current device.
 * hws_en: indicate if hw steering is enabled for current device.
 * return: DOCA_SUCCESS in success, DOCA_ERROR_XXX otherwise.
 */
static doca_error_t
create_dpdk_device_type(char *buf, const char *device, const char *id, bool sft_value, bool sft_en, bool hws_en)
{
	int buf_len;

	if (strcmp(device, "sf") == 0)
		buf_len = snprintf(buf, DPDK_PARAM_LEN, "auxiliary:mlx5_core.sf.%s", id);
	else if (strcmp(device, "vf") == 0 || strcmp(device, "pf") == 0 || strcmp(device, "gpu") == 0)
		buf_len = snprintf(buf, DPDK_PARAM_LEN, "%s", id);
	else if (strcmp(device, "regex") == 0)
		buf_len = snprintf(buf, DPDK_PARAM_LEN, "%s,class=regex", id);
	else {
		DOCA_LOG_ERR("Device \"%s\" is currently unsupported, please use the \"flags\" field for now.", device);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	if ((buf_len + sizeof(",sft_en=") + 1) < DPDK_PARAM_LEN && sft_en) {
		if (sft_value)
			buf_len += snprintf(buf + buf_len, DPDK_PARAM_LEN - buf_len, ",sft_en=1");
		else
			buf_len += snprintf(buf + buf_len, DPDK_PARAM_LEN - buf_len, ",sft_en=0");
	}

	if ((buf_len + sizeof(",dv_flow_en=") + 1) < DPDK_PARAM_LEN && hws_en)
		buf_len += snprintf(buf + buf_len, DPDK_PARAM_LEN - buf_len, ",dv_flow_en=2");

	if (buf_len >= DPDK_PARAM_LEN) {
		DOCA_LOG_ERR("The value of \"%s\" refers to a %s ID that is longer than the limit (%d)", id, device,
			DPDK_PARAM_LEN - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	return DOCA_SUCCESS;
}

/*
 * Checks if DPDK device keys are valid.
 * devices_obj: json object for dpdk devices.
 */
static doca_error_t
json_dpdk_device_validation(struct json_object *devices_obj)
{
	const char *device_params[DPDK_DEVICE_JSON_KEYS_SIZE] = {"device", "id", "sft", "hws"};

	return json_keys_validation(devices_obj, device_params, DPDK_DEVICE_JSON_KEYS_SIZE);
}

/*
 * Allocate memory according to the json file size.
 * fp: pointer to json file.
 * file_length: output param, length of the json data buffer.
 * json_data: allocation place.
 */
static doca_error_t
allocate_json_buffer_dynamic(FILE *fp, size_t *file_length, char **json_data)
{
	ssize_t buf_len = 0;

	/* Use fseek to put file counter to the end, and calculate file length */
	if (fseek(fp, 0L, SEEK_END) == 0) {
		buf_len = ftell(fp);
		if (buf_len < 0) {
			DOCA_LOG_ERR("ftell() function failed");
			return DOCA_ERROR_IO_FAILED;
		}
		/* Check if json file is empty */
		if (buf_len == 0) {
			DOCA_LOG_ERR("Json file is empty");
			return DOCA_ERROR_INVALID_VALUE;
		}

		/* Dynamic allocation */
		*json_data = (char *)calloc((buf_len + 1), sizeof(char));
		if (*json_data == NULL) {
			DOCA_LOG_ERR("calloc() function failed");
			return DOCA_ERROR_NO_MEMORY;
		}

		/* Return file counter to the beginning */
		if (fseek(fp, 0L, SEEK_SET) != 0) {
			free(*json_data);
			*json_data = NULL;
			DOCA_LOG_ERR("fseek() function failed");
			return DOCA_ERROR_IO_FAILED;
		}
	}
	*file_length = buf_len;
	return DOCA_SUCCESS;
}

/*
 * Checks if Current params  object keys are valid.
 * params_obj: json object for params.
 * params: array of valid params.
 */
static doca_error_t
json_args_keys_validation(struct json_object *params_obj, struct argp_params_array *params)
{
	int param_index;
	const char *program_params[MAX_PARAMS_LEN] = {0};

	for (param_index = 0; param_index < params->size; param_index++)
		program_params[param_index] = params->array[param_index].long_flag;

	return json_keys_validation(params_obj, program_params, params->size);
}

/*
 * Extract DOCA flags.
 * parsed_json: json object for parsed data.
 * flags_name: general or Program flags.
 * params: array of params.
 * parsed_config: configuration struct.
 */
static doca_error_t
handle_json_args(struct json_object *parsed_json, char *flags_name, struct argp_params_array *params,
	void *parsed_config)
{
	doca_error_t result = DOCA_SUCCESS;
	int param_index, num = 0;
	struct json_object *doca_flags = NULL;
	struct json_object *doca_flag = NULL;
	bool enable = false;
	const char *temp = NULL;
	char str[MAX_STRING_LEN] = {0};

	/* Missing flags name in json file is not an error (doca_flags will be equal to NULL),
	 * and we still need to check if the mandatory params are missing.
	 */
	if (json_object_object_get_ex(parsed_json, flags_name, &doca_flags)) {
		/* params validation in json mode */
		result = json_args_keys_validation(doca_flags, params);
		if (result != DOCA_SUCCESS)
			return result;
	}

	for (param_index = 0; param_index < params->size; param_index++) {
		/* In case doca_flags == NULL we still need to check if the mandatory flags are missing */
		if (doca_flags != NULL && json_object_object_get_ex(doca_flags, params->array[param_index].long_flag,
			&doca_flag)) {
			if (params->array[param_index].is_cli_only) {
				DOCA_LOG_WARN("Flag \"%s\" is only supported in the CLI",
					params->array[param_index].long_flag);
				continue;
			}
			switch (params->array[param_index].arg_type) {

			/* In case flag accepts string argument */
			case DOCA_ARGP_TYPE_STRING:
				if (json_object_get_type(doca_flag) != json_type_string) {
					DOCA_LOG_ERR("Expecting a string value for \"%s\"",
						params->array[param_index].long_flag);
					return DOCA_ERROR_INVALID_VALUE;
				}
				temp = json_object_get_string(doca_flag);
				if (strnlen(temp, MAX_STRING_LEN) >= MAX_STRING_LEN) {
					DOCA_LOG_ERR("The value of \"%s\" is longer than the limit %d",
						params->array[param_index].long_flag, MAX_STRING_LEN - 1);
					return DOCA_ERROR_INVALID_VALUE;
				}
				strcpy(str, temp);
				result = params->array[param_index].callback((void *)str, parsed_config);
				break;

			/* In case flag accepts integer argument */
			case DOCA_ARGP_TYPE_INT:
				if (json_object_get_type(doca_flag) != json_type_int) {
					DOCA_LOG_ERR("Expecting a int value for \"%s\"",
						params->array[param_index].long_flag);
					return DOCA_ERROR_INVALID_VALUE;
				}
				num = json_object_get_int(doca_flag);
				result = params->array[param_index].callback((void *)&num, parsed_config);
				break;

			/* In case flag accepts boolean argument */
			case DOCA_ARGP_TYPE_BOOLEAN:
				if (json_object_get_type(doca_flag) != json_type_boolean) {
					DOCA_LOG_ERR("Expecting a boolean value for \"%s\"",
						params->array[param_index].long_flag);
					return DOCA_ERROR_INVALID_VALUE;
				}
				enable = json_object_get_boolean(doca_flag);
				if (enable)
					result = params->array[param_index].callback((void *)&enable, parsed_config);
				break;

			/* In case param is dpdk flag */
			case DOCA_ARGP_TYPE_JSON_OBJ:
				result = params->array[param_index].callback(doca_flag, parsed_config);
				break;

			default:
				DOCA_LOG_ERR("Invalid arg_type: %s", params->array[param_index].long_flag);
				return DOCA_ERROR_NOT_SUPPORTED;
			}
			if (result != DOCA_SUCCESS)
				return result;
		} else {
			if (params->array[param_index].is_mandatory) {
				DOCA_LOG_ERR("Parameter \"%s\" is mandatory", params->array[param_index].long_flag);
				return DOCA_ERROR_INVALID_VALUE;
			}
		}
	}
	return DOCA_SUCCESS;
}

static doca_error_t
send_json_dpdk_flags(char *program_path)
{
	doca_error_t result = DOCA_SUCCESS;
	char *argv[DPDK_PARAM_SIZE + 1] = {0};
	int i, argc, nb_params = parser_ctx.dpdk_ctx.index;

	argv[0] = program_path;
	for (i = 1; i <= nb_params; i++)
		argv[i] = &parser_ctx.dpdk_ctx.buffer[i-1][0];

	argv[i] = "--";
	argc = nb_params + 2; /* plus 2 for program path and "--" */
	if (parser_ctx.dpdk_ctx.cb != NULL) {
		result = parser_ctx.dpdk_ctx.cb(argc, argv);
		if (result != DOCA_SUCCESS)
			return result;
	}

	return result;
}

/*
 * Main function for json mode.
 * Creating dpdk, program and DOCA general params with correct format by extracting the data from the json file.
 * json_file_path: path to json file.
 */
static doca_error_t
handle_json_arguments(char *program_path, char *json_file_path)
{
	FILE *json_fp;
	size_t file_length = 0;
	char *json_data = NULL;
	doca_error_t result = DOCA_SUCCESS;
	struct json_object *parsed_json;

	/* Open and read JSON file */
	json_fp = fopen(json_file_path, "r");
	if (json_fp == NULL) {
		DOCA_LOG_ERR("Failed to open json file: \"%s\"", json_file_path);
		return DOCA_ERROR_IO_FAILED;
	}

	/* Dynamic allocate for json data buffer */
	result = allocate_json_buffer_dynamic(json_fp, &file_length, &json_data);
	if (result != DOCA_SUCCESS) {
		fclose(json_fp);
		DOCA_LOG_ERR("Failed to allocate data buffer for the json file");
		return result;
	}

	/* Remove comments from json data */
	delete_comments_from_json_file(json_data, file_length, json_fp);
	fclose(json_fp);

	parsed_json = json_tokener_parse(json_data);
	free(json_data);
	if (parsed_json == NULL) {
		DOCA_LOG_ERR("Invalid Json file format");
		return DOCA_ERROR_IO_FAILED;
	}

	/* Create dpdk params and send them to the program */
	if (parser_ctx.type_config.is_dpdk) {
		result = handle_json_args(parsed_json, "doca_dpdk_flags", &parser_ctx.dpdk_params, NULL);
		if (result != DOCA_SUCCESS)
			goto free_parsed_json;

		result = send_json_dpdk_flags(program_path);
		if (result != DOCA_SUCCESS)
			goto free_parsed_json;
	}

	/* Create general and program params */
	result = handle_json_args(parsed_json, "doca_general_flags", &parser_ctx.general_params,
		&parser_ctx.general_config);
	if (result != DOCA_SUCCESS)
		goto free_parsed_json;

	result = handle_json_args(parsed_json, "doca_program_flags", &parser_ctx.program_params,
		parser_ctx.program_config);
	if (result != DOCA_SUCCESS)
		goto free_parsed_json;

free_parsed_json:
	json_object_put(parsed_json);
	return result;
}

/*
 * Check if there is a match, and return the matched value.
 * param_name: a pionter to the matched value.
 * param: current param.
 * flag: Current flag on argv.
 */
static bool
check_param_match(const char **param_name, struct doca_argp_param *param, char *flag)
{
	int len = strlen(flag);

	*param_name = NULL;
	/* Check short name match, short flag on cli appears with '-' so we have to ignore it */
	if (param->short_flag != NULL && len > 1 && strcmp(param->short_flag, flag + 1) == 0) {
		*param_name = param->short_flag;
		return true;
	}

	/* Check long name match, long flag on cli appears with "--" so we have to ignore them */
	if (param->long_flag != NULL && len > 2 && strcmp(param->long_flag, flag + 2) == 0) {
		*param_name = param->long_flag;
		return true;
	}
	return false;
}

/*
 * Helper function, checks if there is a match for current param
 * params: array of valid params.
 * flag: input flag, check if it's valid.
 * match_index: in case of match return the index in params.
 */
static bool
cli_arg_validation(struct argp_params_array *params, char *flag, int *match_index)
{
	int param_i;
	const char *param_name = NULL;

	for (param_i = 0; param_i < params->size; param_i++) {
		if (check_param_match(&param_name, &params->array[param_i], flag)) {
			params->array[param_i].nb_appears++;
			*match_index = param_i;
			return true;
		}
	}
	return false;
}

/*
 * Program params validation in cli mode.
 */
static doca_error_t
cli_args_validation(int argc, char **argv)
{
	bool match;
	int flag_i, param_i;

	for (flag_i = 1; flag_i < argc; flag_i++) {
		/* Check if flag is general param */
		match = cli_arg_validation(&parser_ctx.general_params, argv[flag_i], &param_i);
		if (match) {
			if (parser_ctx.general_params.array[param_i].nb_appears > 1) {
				DOCA_LOG_ERR("Param \"%s\" should not appear more than once", argv[flag_i]);
				return DOCA_ERROR_INVALID_VALUE;
			}
			/* Skip 1 if current flag input is string or int */
			if (parser_ctx.general_params.array[param_i].arg_type != DOCA_ARGP_TYPE_BOOLEAN)
				flag_i++;
			continue;
		}

		/* Check if flag is program param */
		match = cli_arg_validation(&parser_ctx.program_params, argv[flag_i], &param_i);
		if (match) {
			if (parser_ctx.program_params.array[param_i].nb_appears > 1) {
				DOCA_LOG_ERR("Param \"%s\" should not appear more than once", argv[flag_i]);
				return DOCA_ERROR_INVALID_VALUE;
			}
			/* Skip 1 if current flag input is string or int */
			if (parser_ctx.program_params.array[param_i].arg_type != DOCA_ARGP_TYPE_BOOLEAN)
				flag_i++;
			continue;
		}

		if (!match) {
			DOCA_LOG_ERR("The param \"%s\" is not supported.", argv[flag_i]);
			return DOCA_ERROR_NOT_SUPPORTED;
		}
	}
	return DOCA_SUCCESS;
}

static doca_error_t
convert_string_to_int(char *buff, int *val)
{
	char *buff_end;
	long sl = strtol(buff, &buff_end, 0);
	doca_error_t result = DOCA_SUCCESS;

	errno = 0;
	if (buff_end == buff) {
		DOCA_LOG_ERR("Integer argument \"%s\" is not a number", buff);
		result = DOCA_ERROR_INVALID_VALUE;
	} else if ('\0' != *buff_end) {
		DOCA_LOG_ERR("Integer argument \"%s\" has extra characters at end of input: %s", buff, buff_end);
		result = DOCA_ERROR_INVALID_VALUE;
	} else if ((LONG_MIN == sl || LONG_MAX == sl) && ERANGE == errno) {
		DOCA_LOG_ERR("The argument \"%s\" is out of range of type long", buff);
		result = DOCA_ERROR_INVALID_VALUE;
	} else if (sl > INT_MAX) {
		DOCA_LOG_ERR("Failed to convert value of \"%s\" to \"%ld\" - greater than INT_MAX", buff, sl);
		result = DOCA_ERROR_INVALID_VALUE;
	} else if (sl < INT_MIN) {
		DOCA_LOG_ERR("Failed to convert value of \"%s\" to \"%ld\" - smaller than INT_MIN", buff, sl);
		result = DOCA_ERROR_INVALID_VALUE;
	}
	*val = (int) sl;
	return result;

}

/*
 * Extract general or program params from command line interface.
 * params: parameteres array for general or program params.
 * parsed_config: Configuration struct to store the parsed flags.
 */
static doca_error_t
parse_cli_args(int argc, char **argv, struct argp_params_array *params, void *parsed_config)
{
	doca_error_t result;
	int flag_i, param_i, value;
	const char *p_name, *param_name = NULL;
	bool flag, match = false;

	for (param_i = 0; param_i < params->size; param_i++) {
		match = false;
		for (flag_i = 1; flag_i < argc && !match; flag_i++) {
			if (check_param_match(&param_name, &params->array[param_i], argv[flag_i])) {
				match = true;
				switch (params->array[param_i].arg_type) {
				case DOCA_ARGP_TYPE_STRING:
					if (flag_i + 1 == argc) {
						DOCA_LOG_ERR("Missing value for parameter \"%s\"", param_name);
						return DOCA_ERROR_INVALID_VALUE;
					}
					result = params->array[param_i].callback(argv[flag_i + 1], parsed_config);
					break;
				case DOCA_ARGP_TYPE_INT:
					if (flag_i + 1 == argc) {
						DOCA_LOG_ERR("Missing value for parameter \"%s\"", param_name);
						return DOCA_ERROR_INVALID_VALUE;
					}
					result = convert_string_to_int(argv[flag_i + 1], &value);
					if (result != DOCA_SUCCESS)
						return result;
					result = params->array[param_i].callback(&value, parsed_config);
					break;
				case DOCA_ARGP_TYPE_BOOLEAN:
					flag = true;
					result = params->array[param_i].callback(&flag, parsed_config);
					break;
				default:
					DOCA_LOG_ERR("Invalid arg_type for parameter: %s", param_name);
					return DOCA_ERROR_NOT_SUPPORTED;
				}
				if (result !=  DOCA_SUCCESS)
					return result;
			}
		}
		p_name = (params->array[param_i].long_flag != NULL) ? params->array[param_i].long_flag
							     : params->array[param_i].short_flag;
		if (params->array[param_i].is_mandatory && !match) {
			DOCA_LOG_ERR("Parameter \"%s\" is mandatory", p_name);
			return DOCA_ERROR_INVALID_VALUE;
		}
	}
	return DOCA_SUCCESS;
}

static doca_error_t
send_cli_dpdk_flags(int *argc, char **argv[])
{
	doca_error_t result;
	int i = 0;
	char **tmp_p = *argv;
	char *params[DPDK_PARAM_SIZE + 1] = {0};

	while (i < *argc && strcmp(tmp_p[i], "--") != 0) {
		if (i > DPDK_PARAM_SIZE) {
			DOCA_LOG_ERR("DPDK flags size is bigger than the maximum - %d", DPDK_PARAM_SIZE);
			return DOCA_ERROR_INVALID_VALUE;
		}
		params[i] = &tmp_p[i][0];
		i++;
	}

	if (parser_ctx.dpdk_ctx.cb != NULL) {
		result = parser_ctx.dpdk_ctx.cb(i, params);
		if (result != DOCA_SUCCESS)
			return result;
	}

	/* Skip "--" if exist */
	if (i != *argc)
		i++;

	/* Skip DPDK flags */
	*argc -= (i-1);
	*argv += (i-1);

	/* Expected to have program path in the first element */
	*argv[0] = params[0];
	return DOCA_SUCCESS;
}

/*
 * Main function for regular command line mode.
 * argc: argv size.
 * argv: command line params.
 */
static doca_error_t
handle_cli_arguments(int argc, char **argv)
{
	doca_error_t result = DOCA_SUCCESS;

	if (parser_ctx.type_config.is_dpdk) {
		result = send_cli_dpdk_flags(&argc, &argv);
		if (result != DOCA_SUCCESS)
			return result;
	}

	/* Check input args validation */
	result = cli_args_validation(argc, argv);
	if (result != DOCA_SUCCESS)
		return result;

	/* Function send_cli_dpdk_flags updates the argc and argv according to the number of dpdk flags had read */
	if (argc >= 1) {
		/* Parse DOCA general flags */
		result = parse_cli_args(argc, argv, &parser_ctx.general_params, &parser_ctx.general_config);
		if (result != DOCA_SUCCESS)
			return result;

		/* Parse DOCA program flags */
		result = parse_cli_args(argc, argv, &parser_ctx.program_params, parser_ctx.program_config);
		if (result != DOCA_SUCCESS)
			return result;
	}
	return result;
}

/*
 * Register a supported flag.
 * input_param: param details, add it to params array.
 * params: params array.
 */
static doca_error_t
register_param(struct doca_argp_param *input_param, struct argp_params_array *params)
{
	int current_size = params->size;

	if (current_size >= MAX_PARAMS_LEN) {
		DOCA_LOG_ERR("Too many general params were registered, limit is %d", MAX_PARAMS_LEN);
		return DOCA_ERROR_INVALID_VALUE;
	}
	memcpy(&params->array[current_size], input_param, sizeof(*input_param));
	params->size = current_size + 1;
	return DOCA_SUCCESS;
}

static doca_error_t
log_level_callback(void *param, void *doca_config)
{
	doca_error_t result = DOCA_SUCCESS;
	struct doca_argp_program_general_config *program_config = (struct doca_argp_program_general_config *)doca_config;

	program_config->log_level = *(int *)param;

	result = doca_log_global_level_set(program_config->log_level);
	if (result != DOCA_SUCCESS)
		DOCA_LOG_ERR("The log-level \"%d\" is not supported.", program_config->log_level);
	return result;
}

static doca_error_t
grpc_callback(void *param, void *doca_config)
{
	struct doca_argp_program_general_config *program_config = (struct doca_argp_program_general_config *)doca_config;
	char *address = (char *)param;

	if (strnlen(address, MAX_SERVER_ADDRESS) >= MAX_SERVER_ADDRESS) {
		DOCA_LOG_ERR("gRPC Server address length greater than limit (%d)",
			     MAX_SERVER_ADDRESS - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(program_config->grpc_address, address);
	return DOCA_SUCCESS;
}

noreturn static doca_error_t
version_callback(void *param, void *doca_config)
{
	print_version();
	exit(0);
}

noreturn static doca_error_t
help_callback(void *param, void *doca_config)
{
	doca_argp_usage();
	exit(0);
}

static doca_error_t
register_general_params()
{
	doca_error_t result;
	struct doca_argp_param help_param = {
		.short_flag = "h",
		.long_flag = "help",
		.arguments = NULL,
		.description = "Print a help synopsis",
		.callback = help_callback,
		.arg_type = DOCA_ARGP_TYPE_BOOLEAN,
		.is_mandatory = false,
		.is_cli_only = true
	};
	struct doca_argp_param version_param = {
		.short_flag = "v",
		.long_flag = "version",
		.arguments = NULL,
		.description = "Print program version information",
		.callback = version_callback,
		.arg_type = DOCA_ARGP_TYPE_BOOLEAN,
		.is_mandatory = false,
		.is_cli_only = true
	};
	struct doca_argp_param log_level_param = {
		.short_flag = "l",
		.long_flag = "log-level",
		.arguments = NULL,
		/* Changing these values (from the Log's side) is a breaking change, hence it should be
		 * relatively safe to use the hard coded values here (instead of building the string in runtime)
		 */
		.description = "Set the log level for the program <CRITICAL=20, ERROR=30, WARNING=40, INFO=50, DEBUG=60>",
		.callback = log_level_callback,
		.arg_type = DOCA_ARGP_TYPE_INT,
		.is_mandatory = false,
		.is_cli_only = false
	};

	result = register_param(&help_param, &parser_ctx.general_params);
	if (result != DOCA_SUCCESS)
		return result;
	result = register_param(&version_param, &parser_ctx.general_params);
	if (result != DOCA_SUCCESS)
		return result;
	result = register_param(&log_level_param, &parser_ctx.general_params);
	if (result != DOCA_SUCCESS)
		return result;

	return DOCA_SUCCESS;
}

static doca_error_t
register_grpc_param()
{
	struct doca_argp_param grpc_param = {
		.short_flag = "g",
		.long_flag = "grpc-address",
		.arguments = "ip_address[:port]",
		.description = "Set the IP address for the grpc server",
		.callback = grpc_callback,
		.arg_type = DOCA_ARGP_TYPE_STRING,
		.is_mandatory = true,
		.is_cli_only = false
	};

	return register_param(&grpc_param, &parser_ctx.general_params);
}

static doca_error_t
dpdk_devices_callback(void *json_obj, void *config)
{
	doca_error_t result;
	struct json_object *devices = (struct json_object *)json_obj;
	bool sft_en, parsed_stf_en = false, parsed_hws_en = false;
	int i, index = parser_ctx.dpdk_ctx.index;
	const char *parsed_device_name = {0};
	const char *parsed_device_id = {0};
	struct json_object *device;
	struct json_object *device_name, *device_id, *device_sft, *device_hws;
	int n_devices = json_object_array_length(devices);

	for (i = 0; i < n_devices; i++) {
		/* Disable sft set by default */
		sft_en = false;
		device = json_object_array_get_idx(devices, i);
		if (device == NULL) {
			DOCA_LOG_ERR("Failed to get DPDK device object");
			return DOCA_ERROR_IO_FAILED;
		}

		/* device keys validation */
		result = json_dpdk_device_validation(device);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to validate device keys");
			return result;
		}

		/* Pull device name */
		if (!json_object_object_get_ex(device, "device", &device_name)) {
			DOCA_LOG_ERR("Failed to get DPDK device name");
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (json_object_get_type(device_name) != json_type_string) {
			DOCA_LOG_ERR("Expecting a string value for \"device\"");
			return DOCA_ERROR_INVALID_VALUE;
		}

		parsed_device_name = json_object_get_string(device_name);

		/* Pull device id */
		if (!json_object_object_get_ex(device, "id", &device_id)) {
			DOCA_LOG_ERR("Failed to get DPDK device id");
			return DOCA_ERROR_INVALID_VALUE;
		}

		if (json_object_get_type(device_id) != json_type_string) {
			DOCA_LOG_ERR("Expecting a string value for \"id\"");
			return DOCA_ERROR_INVALID_VALUE;
		}

		parsed_device_id = json_object_get_string(device_id);

		/* Pull sft enable */
		if (json_object_object_get_ex(device, "sft", &device_sft)) {
			if (json_object_get_type(device_sft) != json_type_boolean) {
				DOCA_LOG_ERR("Expecting a boolean value for \"sft\"");
				return DOCA_ERROR_INVALID_VALUE;
			}

			parsed_stf_en = json_object_get_boolean(device_sft);
			/* Notify SFT is enabled for current device */
			sft_en = true;
		}

		/* Pull hws enable */
		if (json_object_object_get_ex(device, "hws", &device_hws)) {
			if (json_object_get_type(device_hws) != json_type_boolean) {
				DOCA_LOG_ERR("Expecting a boolean value for \"hws\"");
				return DOCA_ERROR_INVALID_VALUE;
			}

			parsed_hws_en = json_object_get_boolean(device_hws);
		}

		/* Add device to buffer */
		if (index + 2 > DPDK_PARAM_SIZE) {
			DOCA_LOG_ERR("DPDK buffer is full");
			return DOCA_ERROR_INVALID_VALUE;
		}

		memcpy(parser_ctx.dpdk_ctx.buffer[index], "-a", sizeof("-a"));
		index++;
		result = create_dpdk_device_type(parser_ctx.dpdk_ctx.buffer[index], parsed_device_name,
					      parsed_device_id, parsed_stf_en, sft_en, parsed_hws_en);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Failed to create DPDK device");
			return result;
		}
		index++;
	}
	parser_ctx.dpdk_ctx.index = index;
	return DOCA_SUCCESS;
}

static doca_error_t
dpdk_core_list_callback(void *json_obj, void *config)
{
	int index = parser_ctx.dpdk_ctx.index;
	const char *core_param = NULL;
	struct json_object *core_flag = (struct json_object *)json_obj;

	if (index + 2 > DPDK_PARAM_SIZE) {
		DOCA_LOG_ERR("DPDK buffer is full");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (json_object_get_type(core_flag) != json_type_string) {
		DOCA_LOG_ERR("Expecting a string value for \"core-list\"");
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* Add core-list param to buffer */
	core_param = json_object_get_string(core_flag);
	memcpy(parser_ctx.dpdk_ctx.buffer[index], "-l", sizeof("-l"));
	index++;
	sprintf(parser_ctx.dpdk_ctx.buffer[index], "%s", core_param);
	index++;
	parser_ctx.dpdk_ctx.index = index;
	return DOCA_SUCCESS;
}

static doca_error_t
dpdk_core_mask_callback(void *json_obj, void *config)
{
	int index = parser_ctx.dpdk_ctx.index;
	const char *core_param = NULL;
	struct json_object *core_flag = (struct json_object *)json_obj;

	if (index + 2 > DPDK_PARAM_SIZE) {
		DOCA_LOG_ERR("DPDK buffer is full");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (json_object_get_type(core_flag) != json_type_string) {
		DOCA_LOG_ERR("Expecting a string value for \"core-mask\"");
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* Add core-mask param to buffer */
	core_param = json_object_get_string(core_flag);
	memcpy(parser_ctx.dpdk_ctx.buffer[index], "-c", sizeof("-c"));
	index++;
	sprintf(parser_ctx.dpdk_ctx.buffer[index], "%s", core_param);
	index++;
	parser_ctx.dpdk_ctx.index = index;
	return DOCA_SUCCESS;
}

static doca_error_t
dpdk_flags_callback(void *json_obj, void *config)
{
	char *ptr, delim[] = " ";
	int index = parser_ctx.dpdk_ctx.index;
	const char *parsed_misc_f = NULL;
	char misc_flags_str[MAX_DPDK_MISC_FLAGS_LEN] = {0};
	struct json_object *misc_flags = (struct json_object *)json_obj;

	if (json_object_get_type(misc_flags) != json_type_string) {
		DOCA_LOG_ERR("Expecting a string value for \"flags\"");
		return DOCA_ERROR_INVALID_VALUE;
	}

	parsed_misc_f = json_object_get_string(misc_flags);
	if (strnlen(parsed_misc_f, MAX_DPDK_MISC_FLAGS_LEN) >= MAX_DPDK_MISC_FLAGS_LEN) {
		DOCA_LOG_ERR("The value of \"flags\" is longer than the limit (%d)", MAX_DPDK_MISC_FLAGS_LEN - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}
	strcpy(misc_flags_str, parsed_misc_f);

	/* Add the additional dpdk flags to buffer */
	if (strlen(misc_flags_str) != 0) {
		ptr = strtok(misc_flags_str, delim);
		while (ptr != NULL) {
			if (strnlen(ptr, DPDK_PARAM_LEN) >= DPDK_PARAM_LEN) {
				DOCA_LOG_ERR("Misc flag is longer than the limit (%d)", DPDK_PARAM_LEN - 1);
				return DOCA_ERROR_INVALID_VALUE;
			}

			if (index >= DPDK_PARAM_SIZE) {
				DOCA_LOG_ERR("DPDK buffer is full");
				return DOCA_ERROR_INVALID_VALUE;
			}

			strcpy(parser_ctx.dpdk_ctx.buffer[index], ptr);
			index++;
			ptr = strtok(NULL, delim);
		}
	}
	parser_ctx.dpdk_ctx.index = index;
	return DOCA_SUCCESS;
}

static doca_error_t
register_dpdk_params()
{
	doca_error_t result;
	struct doca_argp_param devices_param = {
		.short_flag = "a",
		.long_flag = "devices",
		.arguments = NULL,
		.description = "Add a device to the allow list",
		.callback = dpdk_devices_callback,
		.arg_type = DOCA_ARGP_TYPE_JSON_OBJ,
		.is_mandatory = false,
		.is_cli_only = false
	};
	struct doca_argp_param core_list_param = {
		.short_flag = "l",
		.long_flag = "core-list",
		.arguments = "<CORELIST>",
		.description = "List of cores to run on",
		.callback = dpdk_core_list_callback,
		.arg_type = DOCA_ARGP_TYPE_JSON_OBJ,
		.is_mandatory = false,
		.is_cli_only = false
	};
	struct doca_argp_param core_mask_param = {
		.short_flag = "c",
		.long_flag = "core-mask",
		.arguments = "<COREMASK>",
		.description = "Hexadecimal bitmask of cores to run on",
		.callback = dpdk_core_mask_callback,
		.arg_type = DOCA_ARGP_TYPE_JSON_OBJ,
		.is_mandatory = false,
		.is_cli_only = false
	};
	struct doca_argp_param flags_param = {
		.short_flag = NULL,
		.long_flag = "flags",
		.arguments = NULL,
		.description = "Additional DPDK (EAL) flags",
		.callback = dpdk_flags_callback,
		.arg_type = DOCA_ARGP_TYPE_JSON_OBJ,
		.is_mandatory = false,
		.is_cli_only = false
	};

	result = register_param(&devices_param, &parser_ctx.dpdk_params);
	if (result != DOCA_SUCCESS)
		return result;
	result = register_param(&core_list_param, &parser_ctx.dpdk_params);
	if (result != DOCA_SUCCESS)
		return result;
	result = register_param(&core_mask_param, &parser_ctx.dpdk_params);
	if (result != DOCA_SUCCESS)
		return result;
	result = register_param(&flags_param, &parser_ctx.dpdk_params);
	if (result != DOCA_SUCCESS)
		return result;

	return DOCA_SUCCESS;
}

doca_error_t
doca_argp_init(const char *program_name, void *program_config)
{
	doca_error_t result;

	if (strnlen(program_name, MAX_PROGRAM_NAME) >= MAX_PROGRAM_NAME) {
		DOCA_LOG_ERR("Program name is longer than limit (%d)", MAX_PROGRAM_NAME - 1);
		return DOCA_ERROR_INVALID_VALUE;
	}

	memset(&parser_ctx, 0, sizeof(parser_ctx));
	parser_ctx.program_validation_callback = NULL;
	strcpy(parser_ctx.program_name, program_name);
	parser_ctx.program_config = program_config;
	parser_ctx.general_config.log_level = doca_log_global_level_get();

	/* DOCA general flags Registration */
	result = register_general_params();
	if (result != DOCA_SUCCESS)
		return result;

	return DOCA_SUCCESS;
}

doca_error_t
doca_argp_register_version_callback(callback_func callback)
{
	if (callback == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	int param_i, params_size = parser_ctx.general_params.size;

	for (param_i = 0; param_i < params_size; param_i++) {
		if (strcmp(parser_ctx.general_params.array[param_i].long_flag, "version") == 0) {
			parser_ctx.general_params.array[param_i].callback = callback;
			break;
		}
	}
	return DOCA_SUCCESS;
}

doca_error_t
doca_argp_register_validation_callback(validation_callback callback)
{
	if (callback == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	parser_ctx.program_validation_callback = callback;
	return DOCA_SUCCESS;
}

static void
print_version_using_callback()
{
	int param_i, params_size = parser_ctx.general_params.size;
	bool value = true;

	for (param_i = 0; param_i < params_size; param_i++) {
		if (strcmp(parser_ctx.general_params.array[param_i].long_flag, "version") == 0) {
			parser_ctx.general_params.array[param_i].callback(&value, parser_ctx.program_config);
			break;
		}
	}
	exit(0);
}

doca_error_t
doca_argp_start(int argc, char **argv)
{
	doca_error_t result;

	if (parser_ctx.type_config.is_dpdk) {
		/* Supported DPDK flags Registration */
		result = register_dpdk_params();
		if (result != DOCA_SUCCESS)
			goto failure;
	}

	/* Register gRPC address option */
	if (parser_ctx.type_config.is_grpc) {
		result = register_grpc_param();
		if (result != DOCA_SUCCESS)
			goto failure;
	}

	/* Start parsing program input */
	if (argc > 1 &&
	    ((strcmp(argv[1], "--json")) == 0 || (strcmp(argv[1], "-j")) == 0)) { /* Json mode */
		if (argc == JSON_MODE_CLI_LEN) {
			result = handle_json_arguments(argv[0], argv[2]);
			if (result != DOCA_SUCCESS)
				goto failure;
		} else {
			DOCA_LOG_ERR("Json mode was invoked with incorrect number of arguments");
			result = DOCA_ERROR_INVALID_VALUE;
			goto failure;
		}
	} else { /* Command line mode */
		if (parser_ctx.type_config.is_dpdk) {
			if (argc == GENERAL_FLAGS_CLI_LEN && strcmp(argv[1], "--") == 0) {
				if (strcmp(argv[2], "-h") == 0 || strcmp(argv[2], "--help") == 0)
					argp_usage();
				else if (strcmp(argv[2], "-v") == 0 || strcmp(argv[2], "--version") == 0)
					print_version_using_callback();
			}
		}
		result = handle_cli_arguments(argc, argv);
		if (result != DOCA_SUCCESS)
			goto failure;
	}

	/* calling program validation function */
	if (parser_ctx.program_validation_callback != NULL) {
		result = parser_ctx.program_validation_callback(parser_ctx.program_config);
		if (result != DOCA_SUCCESS)
			goto failure;
	}
	return DOCA_SUCCESS;

failure:
	doca_argp_usage();
	return result;
}

/* Note: we kept this function even though it's empty for future needs and to maintain backward compatibility */
doca_error_t
doca_argp_destroy()
{
	return DOCA_SUCCESS;
}

void
doca_argp_set_dpdk_program(dpdk_callback callback)
{
	parser_ctx.type_config.is_dpdk = true;
	parser_ctx.dpdk_ctx.cb = callback;
}

void
doca_argp_set_grpc_program()
{
	parser_ctx.type_config.is_grpc = true;
}

doca_error_t
doca_argp_get_grpc_addr(const char **address)
{
	if (address == NULL)
		return DOCA_ERROR_INVALID_VALUE;


	*address = &parser_ctx.general_config.grpc_address[0];
	return DOCA_SUCCESS;
}

doca_error_t
doca_argp_get_log_level(int *log_level)
{
	if (log_level == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*log_level = parser_ctx.general_config.log_level;
	return DOCA_SUCCESS;
}

doca_error_t
doca_argp_param_create(struct doca_argp_param **param)
{
	if (param == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	*param = (struct doca_argp_param *)calloc(1, sizeof(struct doca_argp_param));
	if (*param == NULL)
		return DOCA_ERROR_NO_MEMORY;

	return DOCA_SUCCESS;
}

void
doca_argp_param_set_short_name(struct doca_argp_param *param, const char *name)
{
	if (param == NULL)
		return;

	param->short_flag = name;
}

void
doca_argp_param_set_long_name(struct doca_argp_param *param, const char *name)
{
	if (param == NULL)
		return;

	param->long_flag = name;
}

void
doca_argp_param_set_arguments(struct doca_argp_param *param, const char *arguments)
{
	if (param == NULL)
		return;

	param->arguments = arguments;
}

void
doca_argp_param_set_description(struct doca_argp_param *param, const char *description)
{
	if (param == NULL)
		return;

	param->description = description;
}

void
doca_argp_param_set_callback(struct doca_argp_param *param, callback_func callback)
{
	if (param == NULL)
		return;

	param->callback = callback;
}

void
doca_argp_param_set_type(struct doca_argp_param *param, enum doca_argp_type type)
{
	if (param == NULL)
		return;

	param->arg_type = type;
}

void
doca_argp_param_set_mandatory(struct doca_argp_param *param)
{
	if (param == NULL)
		return;

	param->is_mandatory = true;
}

void
doca_argp_param_set_cli_only(struct doca_argp_param *param)
{
	if (param == NULL)
		return;

	param->is_cli_only = true;
}

doca_error_t
doca_argp_param_destroy(struct doca_argp_param *param)
{
	if (param == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	free(param);
	return DOCA_SUCCESS;
}

static doca_error_t
validate_param_values(struct doca_argp_param *param)
{
	doca_error_t result = DOCA_SUCCESS;

	if (param->short_flag == NULL && param->long_flag == NULL) {
		DOCA_LOG_ERR("Missing param name, at least one name should be set.");
		result = DOCA_ERROR_INITIALIZATION;
	} else if (param->description == NULL) {
		DOCA_LOG_ERR("Missing param description.");
		result = DOCA_ERROR_INITIALIZATION;
	} else if (param->callback == NULL) {
		DOCA_LOG_ERR("Missing param callback.");
		result = DOCA_ERROR_INITIALIZATION;
	} else if (param->arg_type == DOCA_ARGP_TYPE_UNKNOWN) {
		DOCA_LOG_ERR("Missing param arguments type.");
		result = DOCA_ERROR_INITIALIZATION;
	}

	return result;
}

doca_error_t
doca_argp_register_param(struct doca_argp_param *input_param)
{
	doca_error_t result;

	if (input_param == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	int index = parser_ctx.program_params.size;

	if (index >= MAX_PARAMS_LEN) {
		DOCA_LOG_ERR("Too many parameters registered, limit is %d", MAX_PARAMS_LEN);
		doca_argp_param_destroy(input_param);
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* Check if the user sets the mandatory fields */
	result = validate_param_values(input_param);
	if (result != DOCA_SUCCESS) {
		doca_argp_param_destroy(input_param);
		return result;
	}

	/* Param doesn't have a long flag name(JSON KEY), so supported on cli only */
	if (input_param->long_flag == NULL)
		input_param->is_cli_only = true;

	/* Add param to params array */
	memcpy(&parser_ctx.program_params.array[index], input_param, sizeof(*input_param));
	parser_ctx.program_params.size = index + 1;

	return doca_argp_param_destroy(input_param);
}
