/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the buffer, exposing implementation details.
 * This header should not be exposed to applications.
 */

#ifndef PRIV_DOCA_BUF_H_
#define PRIV_DOCA_BUF_H_

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <priv_doca_compat.h>
#include <doca_error.h>
#include <doca_mmap.h>
#include <doca_buf.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Mmap private API */
DOCA_PRIVATE
void priv_doca_mmap_increase_doca_buf_num(struct doca_mmap *mmap);

DOCA_PRIVATE
void priv_doca_mmap_decrease_doca_buf_num(struct doca_mmap *mmap);

DOCA_PRIVATE
bool priv_doca_mmap_is_started(struct doca_mmap *mmap);

DOCA_PRIVATE
struct priv_doca_mchunk *priv_doca_mmap_is_buf_in_chunk(struct doca_mmap *mmap, void *addr, size_t len);

DOCA_PRIVATE
uint32_t priv_doca_mmap_access_flags_to_device_access_flags(uint32_t doca_access);

/**
 * @brief Creates memory map object representing entire memory space for remote system.
 *
 * INTERNAL: Creates memory map object representing remote MR defined by
 * introspection MKey.
 *
 * @param [in] user_data
 * User_data of newly created DOCA memory map.
 * @param [in] rep_dev
 * Representor DOCA Device with introspection mkey capability.
 * @param [out] mmap
 * DOCA memory map granting access to remote memory.
 *
 * @note the created object is not backed by local memory.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure.
 */
DOCA_PRIVATE
doca_error_t priv_doca_mmap_create_from_rep(const union doca_data *user_data, struct doca_dev_rep *rep_dev,
						    struct doca_mmap **mmap);

DOCA_PRIVATE
doca_error_t priv_doca_mchunk_extract_mkey_by_device(const struct priv_doca_mchunk *mchunk, const struct doca_dev *dev,
						     uint32_t doca_access_flags, uint32_t *lkey);

/**
 * @brief This method returns the address of a doca buf
 * @param [in] buf
 * @return address
 */
DOCA_PRIVATE
void *priv_doca_buf_get_addr(const struct doca_buf *buf);

/**
 * @brief This method returns the length of a doca buf
 * @param [in] buf
 * @return address
 */
DOCA_PRIVATE
size_t priv_doca_buf_get_len(const struct doca_buf *buf);

/**
 * @brief This method returns the data address of doca buf
 * @param [in] buf
 * @return data address
 */
DOCA_PRIVATE
void *priv_doca_buf_get_data(const struct doca_buf *buf);

/**
 * @brief This method returns the data length of doca buf
 * @param [in] buf
 * @return data length
 */
DOCA_PRIVATE
size_t priv_doca_buf_get_data_len(const struct doca_buf *buf);

/**
 * @brief Get zero-based mkey for specific doca_buf.
 *
 * @details In order to extract zero-base mkey, the memory address use for the doca_buf must be align with the system
 *	    page-size.
 *
 * @param [in] buf
 * DOCA Buf element.
 *
 * @param [in] dev
 * The device that been registered for the stream context.
 *
 * @param [out] zba_mkey
 * The zba-mkey of the doca_buf.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * doca_error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received or if cannot find mkey by the given device.
 * - DOCA_ERROR_NOT_SUPPORTED - the ZBA mkey not supported for this doca_buf.
 * - DOCA_ERROR_NO_MEMORY - allocate zba_mkey object from pool for doca_buf failed.
 */
DOCA_PRIVATE
doca_error_t priv_doca_buf_zba_mkey_get(struct doca_buf *buf, const struct doca_dev *dev, uint32_t *zba_mkey);

/**
 * @brief Get the number of the elements from the current buffer.
 *
 * @param [in] buf
 * DOCA Buf element.
 *
 * @param [out] num_elements
 * If 'Linked List' extension is enabled: Number of elements until the end of the list
 * If 'Linked List' extension is disabled: 1
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
DOCA_PRIVATE
doca_error_t priv_doca_buf_num_elements_till_end_of_list(const struct doca_buf *buf, uint32_t *num_elements);

/**
 * @brief Get next DOCA Buf in linked list.
 *
 * @param [in] buf
 * DOCA Buf element. Buf must have the linked list extension.
 *
 * @param [out] next_buf
 * The next DOCA Buf or null if the 'Linked List' extension of the buffer is disabled.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
DOCA_PRIVATE
doca_error_t priv_doca_buf_list_next(const struct doca_buf *buf, const struct doca_buf **next_buf);

/**
 * @brief Get last DOCA Buf in linked list.
 *
 * @param [in] buf
 * DOCA Buf element. Buf must have the linked list extension.
 *
 * @param [out] last_buf
 * The last DOCA Buf in the linked list, which may be buf, or buf if the 'Linked List' extension of the buffer
 * is disabled.
 *
 * @return
 * DOCA_SUCCESS - in case of success.
 * Error code - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
 */
DOCA_PRIVATE
doca_error_t priv_doca_buf_list_last(const struct doca_buf *buf, const struct doca_buf **last_buf);

#ifdef __cplusplus
}
#endif

/** @} */

#endif /* PRIV_DOCA_BUF_H_ */
