/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the device, exposing implementation details.
 * This header should not be exposed to applications.
 */

#ifndef PRIV_DOCA_CQ_H_
#define PRIV_DOCA_CQ_H_

#include <stdint.h>

#include <doca_error.h>
#include <priv_doca_compat.h>
#include <priv_doca_types.h>
#include <priv_doca_uar.h>

#ifdef __cplusplus
extern "C" {
#endif

struct doca_dev;
struct priv_doca_cq;
struct priv_doca_comp_events_channel;


struct priv_doca_error_cqe {
	uint8_t		rsvd0[54];
	uint8_t		vendor_err_synd;
	uint8_t		syndrome;
	uint8_t		rsvd56[8];
};

struct priv_doca_64b_cqe {
	uint8_t data[64];
};

struct priv_doca_cq_attr {
	size_t nb_cqes;
	bool initiator_src_dct;

	/* Fields for event handling.*/
	union doca_data user_data;
	priv_doca_cqe_event_handler_cb_t handler;
};

static inline struct priv_doca_cq_attr
priv_doca_cq_attr_create(size_t nb_cqes)
{
	struct priv_doca_cq_attr attr = {0};
	attr.nb_cqes = nb_cqes;

	return attr;
}

/**
 * @brief This method creates a completion queue.
 *
 * @param [in] dev Device used to create the CQ.
 * @param [in] uar User access region.
 * @param [in] nb_cqes The size of the CQ, indicated by number of entries (CQEs) it can hold.
 * @param [in] user_data Opaque structure provided by user of completion queue. Will be provided to handler.
 * @param [in] handler Handler invoked by completion events channel, once CQE event is received. Can be NULL.
 *		       See priv_doca_completion_events_channel.
 * @param [out] cq_out The newly created CQ.
 *
 * @return doca_error_t
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_cq_create(const struct doca_dev *dev, const struct priv_doca_uar *uar,
				     const struct priv_doca_cq_attr *attr, struct priv_doca_cq **cq_out);

/**
 * @brief This method destroys a completion queue.
 *
 * @param [out] cq CQ created by priv_doca_dev_cq_create().
 *
 * @return doca_error_t
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_cq_destroy(struct priv_doca_cq *cq);

/**
 * @brief Get user data from CQ.
 *
 * @details retrieves user data that was provided on creation of the CQ. See priv_doca_cq_create().
 *
 * @param [in] cq The CQ. MUST NOT BE NULL.
 * @param [out] user_data Address that will hold the returned user data. MUST NOT BE NULL.
 *
 * @return doca_error_t
 */
DOCA_PRIVATE
doca_error_t priv_doca_cq_user_data_get(struct priv_doca_cq *cq, union doca_data *user_data);

/**
 * @brief Get user data from CQ.
 *
 * @details retrieves user data that was provided on creation of the CQ. See priv_doca_cq_create().
 *
 * @param [in] cq The CQ.
 * @param [out] cqn Address that will hold the returned CQ number.
 *
 * @return
 * DOCA_SUCCESS - cqn holds valid CQ number.
 * DOCA_ERROR_INVALID_VALUE - cq or cqn are NULL.
 */
DOCA_PRIVATE
doca_error_t priv_doca_cq_number_get(const struct priv_doca_cq *cq, uint32_t *cqn);

/**
 * @brief Poll CQ
 * @details This method is used for polling a number of CQEs.
 * The range [0, *nb_polled - 1] always holds valid CQEs.
 * Returned CQEs may include CQEs with error.
 * @note Returned CQEs are consumed, in the actual CQ, and the CI is updated. cqes_out will hold a copy of the entries.
 *                     -------------------------
 *               0 --> |       Valid CQE       |
 *                     -------------------------
 *                     -------------------------
 *               1 --> |       Valid CQE       |
 *                     -------------------------
 *                            ...
 *                     -------------------------
 *   nb_polled - 1 --> |       Valid CQE       |
 *                     -------------------------
 *                     -------------------------
 *       nb_polled --> |      Invalid CQE      |
 *                     -------------------------
 *                            ...
 *                     -------------------------
 * nb_max_cqes - 1 --> |      Invalid CQE      |
 *                     -------------------------
 * @param [in] cq The CQ to be polled.
 * @param [out] cqes_out Array to hold polled CQEs. MUST NOT BE NULL.
 * @param [in] nb_max_cqes Number of maximum CQEs to poll, and size of cqes_out array.
 * @return number of successfully polled CQEs (nb_polled). Where nb_polled <= nb_max_cqes.
 */
DOCA_PRIVATE
size_t priv_doca_cq_poll(struct priv_doca_cq *cq, struct priv_doca_64b_cqe *cqes_out, size_t nb_max_cqes);

/**
 * @brief Poll CQ for one completion
 * @details This method is used for polling one CQE.
 * In case no CQE was found DOCA_ERROR_AGAIN is returned.
 * In case a CQE or CQE with error was found, DOCA_SUCCESS is returned.
 * @note Returned CQE is consumed, from the actual CQ, and the CI is updated. cqe_out will hold a copy of the entry.
 * @param [in] cq The CQ to be polled. MUST NOT BE NULL.
 * @param [out] cqe_out The received CQE. MUST NOT BE NULL.
 * @return
 * DOCA_SUCCESS - CQE or CQE with error was found.
 * DOCA_ERROR_AGAIN - No CQE was found.
 */
DOCA_PRIVATE
doca_error_t priv_doca_cq_poll_one(struct priv_doca_cq *cq, struct priv_doca_64b_cqe *cqe_out);

/**
 * @brief Check if there are no CQEs available.
 * @details Check if there is at least 1 CQE present in the CQ.
 * @param [in] cq The CQ to be checked.
 * @return
 * true - no CQEs available.
 * false - at least 1 CQE available.
 */
DOCA_PRIVATE
bool priv_doca_cq_empty(const struct priv_doca_cq *cq);

/**
 * @brief Arm CQ
 * @details This method arms the CQ
 * @param [in] cq
 */
DOCA_PRIVATE
void priv_doca_cq_arm(struct priv_doca_cq *cq);

/**
 * @brief This method creates a completion events channel.
 *
 * @details The completion events channel can be used to receive completion events on one or more CQs.
 *
 * @param [in] dev Device used to create the channel.
 * @param [in] is blocking (blocks on get event, supported only in Linux)
 * @param [out] channel_out The newly created completion events channel.
 *
 * @return
 * DOCA_SUCCESS - channel was successfully created.
 * DOCA_ERROR_INVALID_VALUE - dev or channel_out are NULL.
 * DOCA_ERROR_NO_MEMORY - system ran out of memory.
 * DOCA_ERROR_OPERATING_SYSTEM - call to device driver failed.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_comp_events_channel_create(const struct doca_dev *dev, bool is_blocking,
						      struct priv_doca_comp_events_channel **channel_out);

/**
 * @brief This method destroys a completion events channel.
 *
 * @param [in] channel Channel created by priv_doca_dev_comp_events_channel_create().
 */
DOCA_PRIVATE
void priv_doca_dev_comp_events_channel_destroy(struct priv_doca_comp_events_channel *channel);

/**
 * @brief Arms all CQs that are no longer armed.
 *
 * @details The completion events channel needs to be armed after handling of events,
 *          such that CQs that previously received an event, will receive additional events.
 *          It is best to use this method after calling priv_doca_completion_events_channel_handle_events().
 *
 * @param [in] channel The completion events channel.
 * @param [out] handle Handle used for waiting for events. For Linux represents a file descriptor. For Windows a HANDLE.
 *
 * @return
 * DOCA_SUCCESS - handle holds valid value.
 * DOCA_ERROR_INVALID_VALUE - channel or handle are NULL.
 */
DOCA_PRIVATE
doca_error_t priv_doca_comp_events_channel_handle_get(struct priv_doca_comp_events_channel *channel,
						      doca_event_handle_t *handle);

/**
 * @brief Register a CQ to the completion events channel.
 *
 * @details Once a completion event has been received for this CQ, the CQE handler will be called.
 *          See priv_doca_dev_cq_create().
 *
 * @param [in] channel The channel where completion events will be received.
 * @param [in] cq The CQ to register.
 *
 * @return
 * DOCA_SUCCESS - CQ successfully registered.
 * DOCA_ERROR_INVALID_VALUE - channel or cq are NULL.
 * DOCA_ERROR_NO_MEMORY - system ran out of memory.
 * DOCA_ERROR_IN_USE - same CQ already registered.
 * DOCA_ERROR_OPERATING_SYSTEM - call to device driver failed.
 */
DOCA_PRIVATE
doca_error_t priv_doca_comp_events_channel_register(struct priv_doca_comp_events_channel *channel,
						    struct priv_doca_cq *cq);

/**
 * @brief Unregister a CQ from the completion events channel.
 *
 * @param [in] channel The channel where completion events will be received.
 * @param [in] cq a CQ previously registered using priv_doca_comp_events_channel_register().
 *
 * @return
 * DOCA_SUCCESS - CQ successfully unregistered.
 * DOCA_ERROR_INVALID_VALUE - channel or cq are NULL.
 * DOCA_ERROR_NOT_FOUND - cq is already unregistered.
 */
DOCA_PRIVATE
doca_error_t priv_doca_comp_events_channel_unregister(struct priv_doca_comp_events_channel *channel,
						      struct priv_doca_cq *cq);

/**
 * @brief Arms all CQs that are no longer armed.
 *
 * @details The completion events channel needs to be armed after handling of events,
 *          such that CQs that previously received an event, will receive additional events.
 *          It is best to use this method after calling priv_doca_completion_events_channel_handle_events().
 *
 * @param [in] channel The completion events channel. MUST NOT BE NULL.
 */
DOCA_PRIVATE
void priv_doca_comp_events_channel_arm(struct priv_doca_comp_events_channel *channel);

/**
 * @brief Arms all CQs regardless of their state
 *
 * @details The completion events channel needs to be armed after handling of events,
 *          such that CQs that previously received an event, will receive additional events.
 *          It is best to use this method after calling priv_doca_completion_events_channel_handle_events().
 *
 * @param [in] channel The completion events channel. MUST NOT BE NULL.
 */
DOCA_PRIVATE
void priv_doca_comp_events_channel_force_arm(struct priv_doca_comp_events_channel *channel);

/**
 * @brief Handles all completion events received.
 *
 * @details Each CQ that has received a CQE, will be notified using the CQE handler. see priv_doca_dev_cq_create().
 *
 * @param [in] channel The completion events channel. MUST NOT BE NULL.
 * @param [in] handle For Windows the OVERLAPPED* object retrieved as part of the event. For Linux ignored.
 *
 * @return
 * DOCA_SUCCESS - successfully handled completion events.
 * DOCA_ERROR_INVALID_INPUT - channel is NULL.
 * DOCA_ERROR_OPERATING_SYSTEM - call to device driver failed.
 */
DOCA_PRIVATE
doca_error_t priv_doca_comp_events_channel_handle_events(struct priv_doca_comp_events_channel *channel,
							 doca_event_handle_t handle);

#ifdef __cplusplus
}
#endif

#endif /* PRIV_DOCA_CQ_H_ */
