/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef PRIV_DOCA_CTX_H_
#define PRIV_DOCA_CTX_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include <priv_doca_compat.h>
#include <priv_doca_types.h>
#include <doca_error.h>

#include "doca_ctx.h"

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************************************
 * WorkQ - internal
 *********************************************************************************************************************/

enum priv_doca_workq_progress_state {
	PRIV_DOCA_WORKQ_PROGRESS_SUSPENDED,
	PRIV_DOCA_WORKQ_PROGRESS_RUNNING,
};

typedef uint16_t priv_doca_ctx_id_t;

/**
 * @brief Method for progressing of jobs and/or events.
 *
 * @details This method may be called whenever the user calls doca_workq_progress_retrieve().
 * The method should progress an inflight job or an event. And in case there is a result it should be writted in ev.
 *
 * @param [in] data_priv
 * The same data_priv provided in priv_doca_workq_progress_entry_add() along with this callback.
 * @param [out] ev
 * Where to write the result in case there is a result ready.
 *
 * @return
 * Should follow the same convention as doca_workq_progress_retrieve().
 */
typedef doca_error_t (*priv_doca_workq_progress_cb_t)(union doca_data data_priv, struct doca_event *ev);

struct priv_doca_workq_progress_entry {
	priv_doca_workq_progress_cb_t progress_cb;
	union doca_data data_priv;
	enum priv_doca_workq_progress_state state;
};

/**
 * @brief Add a progress entry to the WorkQ.
 *
 * @details The progress entry is responsible for progress of jobs and events submitted to the WorkQ.
 * When doca_workq_progress_retrieve() is called the progress entry is run.
 * In case there are multiple entries in the WorkQ then the entries will run in turn,
 * until the first that returns an event. see priv_doca_workq_progress_cb_t.
 *
 * @note The added progress entry will start in running state. see enum priv_doca_workq_progress_state.
 *
 * @param [in] workq
 * The WorkQ managing the progress entries.
 * @param [in] entry
 * The engine's initial configuration. The memory has to stay valid until rm is called.
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
DOCA_PRIVATE
doca_error_t priv_doca_workq_progress_entry_add(struct doca_workq *workq, struct priv_doca_workq_progress_entry *entry);

/**
 * @brief Update the state of a progress entry that is in the WorkQ.
 *
 * @param [in] workq
 * The WorkQ managing the progress entries.
 * @param [in] entry
 * Entry previously added using priv_doca_workq_progress_entry_add().
 * @param [in] state
 * New state of the progress entry.
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
DOCA_PRIVATE
doca_error_t priv_doca_workq_progress_entry_state_update(struct doca_workq *workq,
							 struct priv_doca_workq_progress_entry *entry,
							 enum priv_doca_workq_progress_state state);
/**
 * @brief Remove a progress entry from the WorkQ.
 *
 * @details Once a progress entry is removed it will no longer be run.
 * After this method it is possible to free the progress entry.
 *
 * @param [in] workq
 * The WorkQ managing the progress entries.
 * @param [in] entry
 * Entry previously added using priv_doca_workq_progress_entry_add(). Can be freed now.
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
DOCA_PRIVATE
doca_error_t priv_doca_workq_progress_entry_rm(struct doca_workq *workq, struct priv_doca_workq_progress_entry *entry);

/**
 * @brief Get the CTX private data previously added to the WorkQ.
 *
 * @param [in] workq
 * The WorkQ storing the data.
 * @param [in] ctx_id
 * ID of the CTX. Use ctx->id.
 * @param [out] data_priv
 * Private data belonging to CTX stored in this WorkQ. This is the same data_priv provided in ctx->ops.workq_add().
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
DOCA_PRIVATE
doca_error_t priv_doca_workq_data_priv_get(struct doca_workq *workq, priv_doca_ctx_id_t ctx_id,
					   union doca_data *data_priv);

/**********************************************************************************************************************
 * Event Driven WorkQ APIs
 **********************************************************************************************************************/

/**
 * @brief Add an event handle to the WorkQ.
 *
 * @details The WorkQ will manage multiple event handles, with behaviour provided by a ctx.
 * arm callback: Called once application calls doca_workq_event_handle_arm(). Application will call this to request
 *               next event before waiting.
 * clear callback: Called once application calls doca_workq_event_handle_clear(). Application will call this to clear
 *                 the received events, giving the ctx a chance to handle them.
 *
 * @param [in] workq
 * The WorkQ that will manage the event handle.
 * @param [in] event_handle
 * The event handle. For linux represents an fd.
 * @param [in] event_handle_ops
 * Callbacks provided by the ctx, to define custom behaviour of the event handle. See above
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
DOCA_PRIVATE
doca_error_t priv_doca_workq_event_handle_register(struct doca_workq *workq,
						   doca_event_handle_t event_handle,
						   const struct priv_doca_event_handle_ops *event_handle_ops,
						   union doca_data ops_data);
/**
 * @brief Remove an event handle that was previously added to the WorkQ.
 *
 * @details This method removes an event handle from the WorkQ.
 * The event handle must have been added using priv_doca_workq_event_handle_register().
 *
 * @param [in] workq
 * The WorkQ managing the event channel.
 * @param [in] event_handle
 * The event handle to unregister, previously provided to priv_doca_workq_event_handle_register().
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
DOCA_PRIVATE
doca_error_t priv_doca_workq_event_handle_unregister(struct doca_workq *workq,
						     doca_event_handle_t event_handle);

/*********************************************************************************************************************
 * CTX - internal
 *********************************************************************************************************************/

#define PRIV_DOCA_CTX_INVALID_ID 0

/**
 * @brief Method for starting a CTX.
 *
 * @details This method is provided by the entity deriving from doca_ctx.
 * This method will be called once the user calls doca_ctx_start() for the CTX.
 * Once the CTX has been started further configurations can't be provided, call stop in order to reconfigure.
 *
 * The following become possible only after start:
 * - Adding WorkQ to CTX using doca_ctx_workq_add()
 * - Removing WorkQ from CTX using doca_ctx_workq_rm()
 * - Submitting a job using doca_workq_submit()
 *
 * The following are NOT possible after start:
 * - Adding device to CTX using doca_ctx_dev_add()
 * - Removing device from CTX using doca_ctx_dev_rm()
 *
 * @param [in] ctx
 * The CTX to start, provided from doca_ctx_start().
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - In case of failure:
 * - DOCA_ERROR_INVALID_VALUE - either an invalid input was received or no devices were added to the CTX.
 * - DOCA_ERROR_NOT_SUPPORTED - one of the provided devices is not supported by CTX.
 * - DOCA_ERROR_INITIALIZATION - resource initialization failed (could be due to allocation failure),
 * 				 or the device is in a bad state or another reason caused initialization to fail.
 *
 * @note In case this method returns DOCA_SUCCESS the doca_ctx will be automatically marked as started.
 */
typedef doca_error_t (*priv_doca_ctx_start_t)(struct doca_ctx *ctx);

/**
 * @brief Method for stopping a CTX, allowing reconfiguration.
 *
 * @details This method is provided by the entity deriving from doca_ctx.
 * Once a CTX has started, it can't be configured any further.
 * This method should be called in case the CTX needs to be reconfigured after starting.
 * Please see priv_doca_ctx_start_t for allowed operations.
 *
 * @param [in] ctx
 * The CTX to stop, provided from doca_ctx_stop().
 * The implementation can assume this value is not NULL.
 *
 * @return
 * DOCA_SUCCESS - in case the device was added successfully.
 * Required error codes - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NOT_PERMITTED - either some jobs are still pending or not all WorkQs have been removed.
 *
 * @note In case this method returns DOCA_SUCCESS the doca_ctx will automatically be marked as stopped.
 */
typedef doca_error_t (*priv_doca_ctx_stop_t)(struct doca_ctx *ctx);

/**
 * @brief Method for adding a device to a CTX.
 *
 * @details This method is provided by the entity deriving from doca_ctx.
 * The method, checks the capabilities of the device and then decides to accept it or not.
 *
 * @param [in] ctx
 * The CTX to add the device to, provided from doca_ctx_dev_add().
 * @param [in] dev
 * The device to add, provided from doca_ctx_dev_add().
 *
 * @return
 * DOCA_SUCCESS - in case the device was added successfully.
 * Required error codes - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_IN_USE - the device was already added.
 * - DOCA_ERROR_NOT_SUPPORTED - the provided device is not supported by CTX,
 * 				I.e., the device is not useful for any job, or missing the capabilities.
 */
typedef doca_error_t (*priv_doca_ctx_dev_add_t)(struct doca_ctx *ctx, struct doca_dev *dev);

/**
 * @brief Method for removing a device from a CTX.
 *
 * @details This method is provided by the entity deriving from doca_ctx.
 * The method removes the device from a CTX.
 *
 * @param [in] ctx
 * The CTX to remove the device from. Provided from doca_ctx_dev_rm().
 * @param [in] dev
 * The device to remove. Provided from doca_ctx_dev_rm().
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - On failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NOT_FOUND - the provided device was never added to the ctx or was already removed.
 */
typedef doca_error_t (*priv_doca_ctx_dev_rm_t)(struct doca_ctx *ctx, struct doca_dev *dev);

/**
 * @brief Method for adding a workQ to a CTX.
 *
 * @details This method is provided by the entity deriving from doca_ctx.
 * Once a WorkQ has been added it will start accepting jobs defined by the CTX & retrieve events from the CTX.
 * The jobs can be progressed using doca_workq_progress_retrieve().
 *
 * @param [in] ctx
 * The CTX that will handle the jobs. Provided from doca_ctx_workq_add().
 * @param [in] workq
 * The WorkQ to be used for submit of jobs and progress of events. Provided from doca_ctx_workq_add().
 * @param [in, out] data_priv
 * Private data belonging to CTX, residing in WorkQ. Such that for each (CTX, WorkQ) pair there is unique private data.
 * The method must set data_priv to point to the private data of the CTX.
 * The same value will be provided back in doca_workq_submit().
 * This argument is provided by the WorkQ once doca_ctx_workq_add() is called.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - on failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - memory allocation failed.
 * - DOCA_ERROR_INITIALIZATION - initialization of WorkQ failed.
 */
typedef doca_error_t (*priv_doca_ctx_workq_add_t)(struct doca_ctx *ctx, struct doca_workq *workq,
						  union doca_data *data_priv);

/**
 * @brief Method for removing a workQ from a CTX.
 *
 * @details This method is provided by the entity deriving from doca_ctx.
 * Once a WorkQ has been removed it will stop accepting jobs & retrieval of events is no longer possible.
 * Calls to doca_workq_progress_retrieve(), will no longer progress the CTX.
 *
 * @param [in] ctx
 * The CTX that was handling the jobs. Provided from doca_ctx_workq_rm().
 * @param [in] workq
 * The WorkQ to be used for submit of jobs and progress of events. Provided from doca_ctx_workq_rm().
 * @param [in, out] data_priv
 * Private data belonging to CTX, residing in WorkQ. Such that for each (CTX, WorkQ) pair there is unique private data.
 * This is the same private data set in priv_doca_ctx_workq_add_t.
 *
 * @return
 * DOCA_SUCCESS - In case of success.
 * Error code - on failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_BAD_STATE - CTX is not started.
 * - DOCA_ERROR_NOT_PERMITTED - some jobs are still pending completion.
 * - DOCA_ERROR_NOT_FOUND - WorkQ does not exist within CTX
 */
typedef doca_error_t (*priv_doca_ctx_workq_rm_t)(struct doca_ctx *ctx, struct doca_workq *workq,
						 union doca_data data_priv);

/**
 * @brief Method for handling of a job by a CTX.
 *
 * @details This method is provided by the entity deriving from doca_ctx.
 * This method will be called by the WorkQ, whenever a user submits a job to the WorkQ.
 *
 * @param [in] data_priv
 * In case priv_doca_workq_submit_entry_add() was previously called using job->ctx->id,
 * then data_priv will be same as what was provided in add. Otherwise it will be NULL.
 * @param [in] job
 * The job that was submitted by user through doca_workq_submit().
 *
 * @return
 * DOCA_SUCCESS - in case the job was submitted successfully, doca_workq_progress_retrieve() can be called next.
 * Required error codes - in case of failure:
 * - DOCA_ERROR_INVALID_VALUE - received invalid input.
 * - DOCA_ERROR_NO_MEMORY - in case the queue is full. See WorkQ depth.
 */
typedef doca_error_t (*priv_doca_ctx_job_submit_t)(union doca_data data_priv, struct doca_job const *job);

/**
 * @brief Variant method for adding a workQ to a CTX. In case event driven mode requested in WorkQ.
 *
 * @details Same as priv_doca_ctx_workq_add_t, but should support event driven mode, by adding an event handle
 * to the workq using priv_doca_workq_event_handle_register().
 * In case this callback is kept as NULL in CTX, it is considered that the CTX does not support event driven mode.
 *
 * @param [in] ctx
 * The CTX that will handle the jobs. Provided from doca_ctx_workq_add().
 * @param [in] workq
 * The WorkQ to be used for submit of jobs and progress of events. Provided from doca_ctx_workq_add().
 * @param [in, out] data_priv
 * Private data belonging to CTX, residing in WorkQ. Such that for each (CTX, WorkQ) pair there is unique private data.
 * The method must set data_priv to point to the private data of the CTX.
 * The same value will be provided back in doca_workq_submit().
 * This argument is provided by the WorkQ once doca_ctx_workq_add() is called.
 */
typedef doca_error_t (*priv_doca_ctx_event_driven_workq_add_t)(struct doca_ctx *ctx, struct doca_workq *workq,
							       union doca_data *data_priv);

/**
 * @brief Variant method for removing a workQ from a CTX. In case event driven mode requested in WorkQ.
 *
 * @details Same as priv_doca_ctx_workq_rm_t, but cleans up after priv_doca_ctx_event_driven_workq_add_t instead.
 * This means in addition to removing all progress entries, it should remove all event handles as well, using
 * priv_doca_workq_event_handle_unregister().
 *
 * @param [in] ctx
 * The CTX that was handling the jobs. Provided from doca_ctx_workq_rm().
 * @param [in] workq
 * The WorkQ to be used for submit of jobs and progress of events. Provided from doca_ctx_workq_rm().
 * @param [in, out] data_priv
 * Private data belonging to CTX, residing in WorkQ. Such that for each (CTX, WorkQ) pair there is unique private data.
 * This is the same private data set in priv_doca_ctx_workq_add_t.
 */
typedef doca_error_t (*priv_doca_ctx_event_driven_workq_rm_t)(struct doca_ctx *ctx, struct doca_workq *workq,
							      union doca_data data_priv);

struct priv_doca_ctx_ops {
	priv_doca_ctx_start_t start;
	priv_doca_ctx_stop_t stop;
	priv_doca_ctx_dev_add_t dev_add;
	priv_doca_ctx_dev_rm_t dev_rm;
	priv_doca_ctx_workq_add_t workq_add;
	priv_doca_ctx_workq_rm_t workq_rm;
	priv_doca_ctx_job_submit_t job_submit;

	/* Required for event driven mode only. */
	priv_doca_ctx_event_driven_workq_add_t event_workq_add;
	priv_doca_ctx_event_driven_workq_rm_t event_workq_rm;
};

struct doca_ctx {
	priv_doca_ctx_id_t id; /**< Handle created automatically through CTX init API. Read only field. */
	bool is_started;       /**< Read only field.*/
	struct priv_doca_ctx_ops ops;
};

/**
 * @brief Initializes a newly allocated DOCA CTX.
 *
 * @details Initializes a new DOCA CTX.
 * This method should be called upon creation of a library instance that derives from doca_ctx.
 *
 * @param [in] ctx
 * pointer to the newly created CTX object. Usually a field inside the library instance.
 * @param [in] create_id
 * If set ctx->id will be initialized and can be used with the various objects that require it.
 */
DOCA_PRIVATE
doca_error_t priv_doca_ctx_init(struct doca_ctx *ctx, bool create_id);

/**
 * @brief Destroys a previously initialized DOCA CTX.
 *
 * @details Destroys a DOCA CTX.
 * This method should be called upon destroy of a library instance that derives from doca_ctx.
 *
 * @param [in] ctx
 * Pointer to the doca_ctx object. The ctx should have been previously initialized using priv_doca_ctx_init().
 */
DOCA_PRIVATE
void priv_doca_ctx_destroy(struct doca_ctx *ctx);

static inline struct priv_doca_ctx_ops *
priv_doca_ctx_workq_ops_get(struct doca_ctx *ctx)
{
	return &ctx->ops;
}

#ifdef __cplusplus
}
#endif

/** @} */

#endif /* PRIV_DOCA_CTX_H_ */
