/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the device, exposing implementation details.
 * This header should not be exposed to applications.
 */

#ifndef PRIV_DOCA_DEV_H_
#define PRIV_DOCA_DEV_H_

#ifdef __linux__

#include <infiniband/verbs.h>

#endif // __linux__

#include <stdbool.h>
#include <stdint.h>

#include <doca_error.h>
#include <doca_dev.h>
#include <priv_doca_compat.h>
#include <priv_doca_cq.h>
#include <priv_doca_uar.h>

#ifdef __cplusplus
extern "C" {
#endif

#define PRIV_DOCA_VUID_SIZE 128

/* Specifies the type of the represented device. */
enum priv_doca_rep_type {
	PRIV_DOCA_REP_UNKNOWN = 0,
	PRIV_DOCA_REP_NET,		/* Network Device */
	PRIV_DOCA_REP_VIRTIO_FS,	/* VIRTIO Filesystem Emulated Device */
	PRIV_DOCA_REP_VIRTIO_NET,	/* VIRTIO Network Emulated Device */
	PRIV_DOCA_REP_VIRTIO_BLK,	/* VIRTIO Block Emulated Device */
	PRIV_DOCA_REP_NVME,		/* NVMe Emulated Device */
};

struct priv_doca_vuid {
	char id[PRIV_DOCA_VUID_SIZE];
};

struct doca_devx_object;

struct doca_umem_object;

struct priv_doca_dev_port_info {
	bool vhca_id_valid;
	uint16_t vhca_id;
	uint16_t vport;
};

/**
 * @brief A DOCA device observer callback for device-close event.
 *
 * CB to be called before the DOCA device is being closed.
 * The closed device is being passed to the CB together with the cookie used for observer registration on the DOCA device.
 *
 * NOTE: Calling doca_dev API with 'dev' is not allowed.
 *
 * @param [in] dev
 * The observed DOCA device that is being closed.
 *
 * @param [in] cookie
 * User data provided on registration of the observer on the DOCA device.
 */
typedef void (*priv_doca_dev_observer_on_close_t)(const struct doca_dev *dev, union doca_data cookie);

struct priv_doca_dev_observer_ops {
	priv_doca_dev_observer_on_close_t on_dev_close; /*!< Callback to be called before the DOCA device is being closed. */
};

typedef uintptr_t priv_doca_dev_observer_handle_t;

/**
 * @brief Regiters a DOCA device observer to observe a subject DOCA device for special events.
 *
 * Multiple registration of the same observers ({priv_doca_dev_observer_ops observer_ops, union doca_data cookie} pair), is not allowed.
 * For multiple registration of the same event callbacks (struct priv_doca_dev_observer_ops), one should use different cookie (union doca_data).
 *
 * 'observer_ops' is copied on registration.
 *
 * @param [in] dev
 * The subject DOCA device to be observed.
 *
 * @param [in] observer_ops
 * Set of callbacks for supported events.
 *
 * @param [in] cookie
 * User data to pass to each assigned CB on event occurrence.
 *
 * @param [out] handle
 * A DOCA device observer handle to use for unregistration.
 *
 * @return
 * - DOCA_SUCCESS on success.
 * - DOCA_ERROR_INVALID_VALUE in case of invalid paramters
 * - DOCA_ERROR_IN_USE in case of registration of already-registered observer.
 * - DOCA_ERROR_NO_MEMORY in case of allocation failure.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_observer_register(struct doca_dev *dev, const struct priv_doca_dev_observer_ops *observer_ops,
					     union doca_data cookie, priv_doca_dev_observer_handle_t *handle);

/**
 * @brief Unregiters a DOCA device observer from a subject DOCA device.
 *
 * @param [in] dev
 * The subject DOCA device to unregister observer from.
 *
 * @param [in] handle
 * A DOCA device observer handle returned from previous priv_doca_dev_observer_register call.
 *
 * @return
 * - DOCA_SUCCESS on success.
 * - DOCA_ERROR_INVALID_VALUE in case of invalid paramters
 * - DOCA_ERROR_NOT_FOUND in case of invalid handle.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_observer_unregister(struct doca_dev *dev, priv_doca_dev_observer_handle_t handle);

#ifdef __linux__

/**
 * Retrieves an ibv_context from a devinfo.
 * This method is supported only for Linux as other OS don't support ibverbs.
 */
DOCA_PRIVATE
struct ibv_context *priv_doca_devinfo_to_ibv_ctx(const struct doca_devinfo *di);

/**
 * Retrieves an ibv_context from a dev.
 * This method is supported only for Linux as other OS don't support ibverbs.
 */
DOCA_PRIVATE
struct ibv_context *priv_doca_dev_to_ibv_ctx(const struct doca_dev *dev);

#endif // __linux__

/**
 * Retrieves a pd id from a dev
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_pd_id_get(const struct doca_dev *dev, uint32_t *pd_id);

/**
 * Retrieves the vhca ID of a doca_devinfo.
 */
DOCA_PRIVATE
doca_error_t priv_doca_devinfo_vhca_id_get(const struct doca_devinfo *di, uint16_t *vhca_id);

/**
 * Retrieves the vhca ID of a doca_devinfo_rep.
 */
DOCA_PRIVATE
doca_error_t priv_doca_devinfo_rep_vhca_id_get(const struct doca_devinfo_rep *di, uint16_t *vhca_id);

/**
 * Retrieves the Vendor Unique ID of a doca_devinfo.
 */
DOCA_PRIVATE
doca_error_t priv_doca_devinfo_vuid_get(const struct doca_devinfo *di, void *value, uint32_t size);

/**
 * @brief Get IP address for a given device info.
 * This allows to query information about device's IP address before opening
 * it.
 *
 * @param [in] devinfo
 * The devinfo structure of the requested device.
 *
 * @param [in] is_ipv4
 * Chooses between quering IPv4 or IPv6
 *
 * @param [out] value
 * Pointer to where the sockaddr_in data is placed.
 *
 * @param [in] size
 * Memory size allocated for value
 *
 * @return
 * DOCA_SUCCESS on success, if DOCA_ERROR_NOT_SUPPORTED is returned then this devinfo does not have the relevant
 * IP address.
 */
DOCA_PRIVATE
doca_error_t priv_doca_devinfo_sockaddr_get(const struct doca_devinfo *devinfo, bool is_ipv4, void *value, uint32_t size);

/**
 * @brief Get vhca id of a representor device
 * @param [in] dev representor
 * @return vhca id
 */
DOCA_PRIVATE
uint16_t priv_doca_dev_vhca_id_get(const struct doca_dev_rep *dev);

/**
 * @brief Query whether the representor device is an hotplugged device
 * @param [in] dev representor
 * @return Is the representor device is an hotplugged device
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_rep_is_hotplug_get(const struct doca_dev_rep *dev, bool *is_hotplug);

/**
 * @brief Issues general PRM command.
 *
 * @param [in] dev DOCA device to issue PRM command on.
 * @param [in] in input buffer, of specification structure, to send.
 * @param [in] in_size input buffer size.
 * @param [out] out command's output data.
 * @param [in] out_size output data size.
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_prm_cmd_send(const struct doca_dev *dev, const void *in, size_t in_size, void *out, size_t out_size);

/**
 * @brief Creates devx object.
 *
 * @param [in] dev DOCA device to create the devx on.
 * @param [in] in input buffer, of specification structure, to send.
 * @param [in] in_size input buffer size.
 * @param [out] out command's output data.
 * @param [in] out_size output data size.
 * @param [out] devx object
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_devx_object_create(const struct doca_dev *dev, const void *in, size_t in_size, void *out,
					      size_t out_size, struct doca_devx_object **devx_obj);

/**
 * @brief Destroys devx object
 *
 * @param [in] devx object to destroy
 *
  @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_devx_object_destroy(struct doca_devx_object *devx_obj);

/**
 * @brief Modifies devx object.
 *
 * @param [in] devx object to modify
 * @param [in] in input buffer, of specification structure, to send.
 * @param [in] in_size input buffer size.
 * @param [out] out command's output data.
 * @param [in] out_size output data size.
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_devx_object_modify(struct doca_devx_object *devx_obj, const void *in, size_t in_size,
					      void *out, size_t out_size);

/**
 * @brief retrieves the wrapped devx object
 *
 * @param [in] devx object
 * @param [out] wrapped devx object pointer
 *
  @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_devx_object_get(struct doca_devx_object *devx_obj, void **obj);

/**
 * @brief creates umem
 *
 * @param [in] dev DOCA device to create umem on
 * @param [in] address
 * @param [in] size
 * @param [in] access flags (see doca_access_flags in doca_types.h)
 * @param [out] umem object
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_umem_create(const struct doca_dev *dev, void *address, size_t size, uint32_t access_flags,
				   struct doca_umem_object **umem_obj);

/**
 * @brief creates umem
 *
 * @param [in] dev DOCA device to create umem on
 * @param [in] address
 * @param [in] size
 * @param [in] page size
 * @param [in] access flags (see doca_access_flags in doca_types.h)
 * @param [out] umem object
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_umem_create_pg_sz(const struct doca_dev *dev, void *address, size_t size, size_t page_size,
				   uint32_t access_flags, struct doca_umem_object **umem_obj);

/**
 * @brief destroys umem
 *
 * @param [in] umem object
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 *
 */
DOCA_PRIVATE
doca_error_t priv_doca_umem_destroy(struct doca_umem_object *umem_obj);

/**
 * @brief This method retrieves the umem id from the umem object
 *
 * @param [in] umem object
 * @param [out] umem id
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_umem_id_get(struct doca_umem_object *umem_obj, uint32_t *umem_id);

/**
 * @brief Issues QUERY_HCA_CAP PRM command with custom op_mod.
 *
 * @details Method for usage by libraries to query custom HCA capabilities.
 *          In case this method is used it is recommended to use multi_os_prm.h for parsing of output buffer.
 *
 * @param [in] devinfo DOCA device info to issue PRM command on.
 * @param [in] op_mod operation mode of the QUERY_HCA_CAP command.
 * @param [out] out QUER_HCA_CAP_OUT buffer.
 * @param [out] out_size output data size.
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_devinfo_hca_caps_by_opmod_query(const struct doca_devinfo *devinfo, uint16_t op_mod,
						       void *out, size_t out_size);

/**
 * @brief Get DOCA local device of DOCA representor device
 *
 * @param [in] rep_dev DOCA representor device to get parent of.
 * @param [out] local_dev DOCA local device parent.
 *
 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_rep_parent_get(const struct doca_dev_rep *rep_dev, struct doca_dev **local_dev);

/**
 * @brief Retrieves the VPORT of a doca device.
 *
 * @param [in] dev
 * doca representor device
 *
 * @param [out] vport
 * the vport value
 *
 * @return
 * DOCA_SUCCESS on success, DOCA_ERROR_INVALID_VALUE in case of invalid input
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_rep_vport_get(const struct doca_dev_rep * const dev, uint16_t *vport);

/**
 * @brief Set introspection mkey of a doca representor device.
 *
 * @param [in] dev
 * doca representor device
 *
 * @param [in] mkey_in
 * the mkey value
 *
 * @return
 * DOCA_SUCCESS on success, DOCA_ERROR_INVALID_VALUE in case of invalid input
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_rep_introspection_mkey_set(struct doca_dev_rep *dev, uint32_t mkey_in);

/**
 * @brief Retrieves the introspection mkey of a doca representor device.
 *
 * @param [in] dev
 * doca representor device
 *
 * @param [out] mkey_out
 * introspection mkey value
 *
 * @return
 * DOCA_SUCCESS on success, DOCA_ERROR_INVALID_VALUE in case of invalid input
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_rep_introspection_mkey_get(const struct doca_dev_rep *const dev, uint32_t *mkey_out);

#ifdef __cplusplus
}
#endif

#endif /* PRIV_DOCA_DEV_H_ */
