/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the doca types, exposing private types.
 * These types may be relevant only for libraries in doca.
 * This header should not be exposed to applications.
 */

#ifndef PRIV_DOCA_TYPES_H_
#define PRIV_DOCA_TYPES_H_

#include <doca_types.h>
#include <doca_error.h>

#ifdef __cplusplus
extern "C" {
#endif

struct priv_doca_cq;

/**
 * @brief Method for arming of event handle.
 *
 * @details This method is called whenever the user application requests an event before sleeping.
 *
 * @param [in] ops_data
 * Data provided upon adding the event channel.
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
typedef doca_error_t (*priv_doca_event_handle_arm_cb_t)(union doca_data ops_data);

/**
 * @brief Method for handling of events.
 *
 * @details This method is called whenever an event has been received on the associated event handle
 * (see priv_doca_workq_event_handle_register()), once the method is called it is guaranteed that some event had been
 * received on the event handle.
 *
 * @param [in] ops_data
 * Data provided upon adding the event handle in priv_doca_workq_event_handle_register().
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
typedef doca_error_t (*priv_doca_event_handle_clear_cb_t)(union doca_data ops_data);

struct priv_doca_event_handle_ops {
	priv_doca_event_handle_arm_cb_t arm_cb;
	priv_doca_event_handle_clear_cb_t clear_cb;
};

/**
 * @brief Method for handling of completion events.
 *
 * @details This method is called by the completion event channel whenever a CQE arrives on a registered CQ.
 * If this handler is called, it means that at least one CQE has been added to the registered CQ.
 *
 * @param [in] user_data
 * user_data provided on CQ creation. See priv_doca_cq_create().
 *
 * @return
 * 0 - In case of success.
 * Error code - On failure. See doca_error_t.
 */
typedef void (*priv_doca_cqe_event_handler_cb_t)(union doca_data user_data);

#ifdef __cplusplus
}
#endif

/** @} */

#endif /* PRIV_DOCA_TYPES_H_ */
