/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the device, exposing implementation details.
 * This header should not be exposed to applications.
 */

#pragma once

#include <doca_error.h>
#include <priv_doca_compat.h>

#ifdef __cplusplus
extern "C" {
#endif

struct doca_buf;
struct doca_dev;

struct priv_doca_umr;
struct priv_doca_umr_handle;

struct priv_doca_data_segment {
	uint32_t byte_count;	/* Big Endian */
	uint32_t lkey;		/* Big Endian */
	uint64_t address;	/* Big Endian */
};

/**
 * @brief Create UMR container
 *
 * @param [in] dev device for memory registration
 * @param [in] num_elements number of maximum infilght requests using UMR key.
 * Represents number of calls that can be made to buf_to_wqe before calling mkey_free.
 * @param [out] umr_out The UMR container
 * @return doca_error_t
 * DOCA_SUCCESS - successfully created UMR.
 * DOCA_ERROR_INVALID_VALUE - invalid input was provided.
 * DOCA_ERROR_NO_MEMORY - failed to allocate memory.
 * DOCA_ERROR_DRIVER - failed to allocate an mkey DevX command failed.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_umr_create(const struct doca_dev *dev, uint32_t num_elements,
				      struct priv_doca_umr **umr_out);

/**
 * @brief Create UMR WQE from doca buf. On WQE submit.
 * @details Fills a UMR WQE taking space of 3 WQEBBs. From a doca_buf representing a gather list.
 * Returns a single indirect key that can be used as the source buffer of the next WQE,
 * in case more than one buffer was provided, a small fence must be applied on the WQE using the indirect key.
 * Otherwise NOPS will be filled instead of the UMR and there is no need for a fence on the next WQE.
 *
 * @param [in] umr The UMR container created from priv_doca_dev_umr_create().
 * @param [in] buf Gather list of buffers. Can be single buffer, in that case 3 NOPs are filled.
 * @param [in] wqe Start address of the WQE where the UMR WQE will be written.
 * @param [in] pi Producer index of the QP. Neccessary for filling control segment of the UMR.
 * @param [in] qp_num Number of the QP. Neccessary for filling control segment of the UMR.
 * @param [out] indirect_key Pointer to the WQE source data segment where the produced UMR mkey will be used.
 * @param [out] need_fence Whether a fence needs to be applied on WQE using the indirect key.
 * @param [out] handle_out A unique handle for identifying the UMR mkey instance.
 * @return doca_error_t
 * DOCA_SUCCESS - UMR WQE successfully filled, taking space of 3 WQEBBs.
 * DOCA_ERROR_INVALID_VALUE - invalid input was provided.
 * DOCA_ERROR_NOT_PERMITTED - buf list elements exceed maximum allowed. Or buf is not head of list.
 * Or maximum umr operations exceeded, see num_elements in priv_doca_dev_umr_create.
 */
DOCA_PRIVATE
doca_error_t priv_doca_umr_buf_to_wqe(struct priv_doca_umr *umr, const struct doca_buf *buf, void *wqe, uint16_t pi,
				      uint32_t qp_num, struct priv_doca_data_segment *indirect_key,
				      bool *need_fence, struct priv_doca_umr_handle **handle_out);


/**
 * @brief Release the UMR mkey for reuse in other operations. On CQE receive.
 * @details Method must be called once the UMR mkey is no longer in use, to allow reuse of the same mkey
 * for coming operations. This can be known once CQE is received for the operation that used the mkey.
 *
 * @param [in] umr the UMR container instance
 * @param [in] handle Handle received from priv_doca_umr_buf_to_wqe().
 *
 * @return doca_error_t
 * DOCA_SUCCESS - successfully freed UMR mkey, for reuse.
 * DOCA_ERROR_INVALID_VALUE - invalid input was provided.
 */
DOCA_PRIVATE
void priv_doca_umr_mkey_free(struct priv_doca_umr *umr, struct priv_doca_umr_handle *handle);

/**
 * @brief Destroy the UMR container instance.
 *
 * @param [in] umr the UMR container instance
 *
 * @return doca_error_t
 * DOCA_SUCCESS - successfully destroyed UMR container.
 * DOCA_ERROR_INVALID_VALUE - invalid input was provided.
 */
DOCA_PRIVATE
doca_error_t priv_doca_dev_umr_destroy(struct priv_doca_umr *umr);

#ifdef __cplusplus
}
#endif
