/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <unordered_map>
#include <atomic>
#include <algorithm>

#include <doca_log.h>
#include "buf_extensions.hpp"
#include "internal_doca_buf.h"

DOCA_LOG_REGISTER(CORE::BUF_EXTENSIONS);

static const std::vector<doca_core::buf_extension_attributes>
extensions_attributes{
	{doca_core::buf_extension_type::LINKED_LIST, doca_buf_extension::DOCA_BUF_EXTENSION_LINKED_LIST,
	 sizeof(doca_buf_linked_list_extension),
	 "linked list"},
};

const doca_core::supported_extensions_attributes &
doca_core::get_supported_extensions() noexcept
{
	return extensions_attributes;
}

/**
 * @brief This method calculates the supported extension masks
 * The method calculates the mask only in the first time it is called because it relies on a static const map above.
*  @return supported extensions mask.
 */
uint32_t
supported_extensions_mask() noexcept
{
	// extensions_mask is set at the end of this function so if it is not 0 then it was already calculated.
	static std::atomic_ulong extensions_mask{0};
	if (extensions_mask != 0) {
		return extensions_mask;
	}

	// Calculating the mask directly on extensions_mask may result in invalid value if more than one thread runs
	// this method at the first time. Therefore it uses temp_mask and sets extensions_mask after calculation is done.
	uint32_t temp_mask{0};
	for (auto &entry : extensions_attributes) {
		temp_mask |= entry.extension_flag;
	}

	extensions_mask = temp_mask;
	return extensions_mask;
}

bool
doca_core::is_buf_extensions_valid(uint32_t extension_flags) noexcept
{
	auto supported_mask = supported_extensions_mask();

	if ((~supported_mask & extension_flags) != 0) {
		// Extension flags can be 0 or any combination of flags that are defined in doca_buf_extension.
		// masking with ~supported_mask will leaves only the unsupported flags and should be 0.
		return false;
	}

	return true;
}

size_t
doca_core::calculate_extensions_buffer_size(uint32_t extension_flags,
	const supported_extensions_attributes &attributes /* = get_supported_extensions()*/) noexcept
{
	size_t required_size{0};
	for (auto &attribute : attributes) {
		if (extension_flags & static_cast<uint32_t>(attribute.extension_flag)) {
			required_size += attribute.size;
		}
	}

	return required_size;
}

