/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <vector>
#include <string>

#include <priv_doca_compat.h>
#include <doca_buf.h>
#include "buf_extensions.hpp"

namespace doca_core
{

/**
 * This class implements a doca buffer extension map.
 * The map defines and maintains the extensions offsets.
 * doca buffers can uses this map to get the address or offset of their extensions
 */
class buf_extensions_map
{
public:
	/**
	 * Constructor
	 * @param extensions_flags [in] set of extensions that this map should maintain.
	 * @param supported_extensions_attributes [in] DOCA supported extensions attributes
	 */
	explicit buf_extensions_map(uint32_t extensions_flags,
				    const supported_extensions_attributes &extensions_attributes);

	/**
	 * Destructor
	 */
	~buf_extensions_map() = default;

	/**
	 * This method calculates an extension address for a doca buffer. The method trusts that the extension type
	 * is boundaries safe and that the extensions head is valid.
	 * @param [in] extensions head
	 * @param [in] extension type
	 * @return address or null (if the inventory doesn't support this extension)
	 */
	inline void *get_extension(void *extensions_head, buf_extension_type extension_type) const noexcept
	{
		auto offset_entry = &m_offsets[static_cast<uint32_t>(extension_type)];
		return reinterpret_cast<void *>(
			(reinterpret_cast<uintptr_t>(extensions_head) + offset_entry->offset) * offset_entry->valid);
	}

	/**
	 * This method calculates an extension address for a doca buffer. The method trusts that the extension type
	 * is boundaries safe and that the extensions head is valid.
	 * @param [in] extensions head
	 * @param [in] extension type
	 * @return address or null (if the inventory doesn't support this extension)
	 */
	inline const void *get_extension(const void *extensions_head, buf_extension_type extension_type) const noexcept
	{
		auto offset_entry = &m_offsets[static_cast<uint32_t>(extension_type)];
		return reinterpret_cast<const void *>((reinterpret_cast<uintptr_t>(extensions_head) + offset_entry->offset) *
						offset_entry->valid);
	}

	/**
	 * This method returns an extension offset.
	 * @param [in] extension type
	 * @param [out] offset
	 * @return DOCA_SUCCESS if extension is supported, DOCA_ERROR_NOT_SUPPORTED if not.
	 */
	inline std::pair<doca_error_t, size_t>
	get_extension_offset(buf_extension_type extension_type) const
	{
		static const doca_error_t ret_val[]{DOCA_ERROR_NOT_SUPPORTED, DOCA_SUCCESS};
		auto offset_entry = &m_offsets[static_cast<uint32_t>(extension_type)];

		// Assumption: valid field can only be 1 or 0.
		return {ret_val[(offset_entry->valid & 0x1)], offset_entry->offset};
	}

	/**
	 * This method returns an extension size.
	 * @param [in] extension type
	 * @param [out] offset
	 * @return DOCA_SUCCESS if extension is supported, DOCA_ERROR_NOT_SUPPORTED if not.
	 */
	inline std::pair<doca_error_t, size_t> get_extension_size(buf_extension_type extension_type) const
	{
		static const doca_error_t ret_val[]{DOCA_ERROR_NOT_SUPPORTED, DOCA_SUCCESS};
		auto offset_entry = &m_offsets[static_cast<uint32_t>(extension_type)];

		// Assumption: valid field can only be 1 or 0.
		return {ret_val[(offset_entry->valid & 0x1)], offset_entry->size};
	}

	/**
	 * @brief This method returns the extensions flags
	 * @return extensions flags
	 */
	inline uint32_t get_extensions() const
	{
		return m_extensions_flags;
	}

private:

	struct extension_entry {
		size_t offset{0};
		size_t size{0};
		uint8_t valid{0};
	};

	static const size_t sc_max_extensions{32}; /** 32 possible extension flags */
	extension_entry m_offsets[sc_max_extensions];

	const uint32_t m_extensions_flags;

	// Non copyable
	buf_extensions_map(buf_extensions_map const &) = delete;
	buf_extensions_map &operator=(buf_extensions_map const &) = delete;
	buf_extensions_map(buf_extensions_map &&) = delete;
	buf_extensions_map &operator=(buf_extensions_map &&) = delete;
};

} /* namespace doca_core */


/** @} */
