/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the buffer, exposing implementation details.
 * This header should not be exposed to applications.
 */

#pragma once

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <vector>

#include <priv_doca_compat.h>

struct doca_buf;

namespace doca_core
{

/**
 * This class implements a specialized pool of doca buffers
 * doca buffers implement a variable sized extensions buffer that requires special allocation of memory and use
 * of placement new that are not supported by other pools (typed or generic)
 */
class buf_pool
{
public:
	/**
	 * @brief Constructor
	 * @param num_buffers [in] num buffers that the pool should maintain
	 * @param extensions_buffer_size [in] extensions memory size for one doca_buf.
	 */
	buf_pool(uint32_t num_buffers, size_t extensions_buffer_size);

	/**
	 * Destructor
	 */
	~buf_pool();

	/**
	 * @brief This method returns an element from the pool
	 * @return element or null if pool is empty.
	 */
	inline struct doca_buf *
	acquire() noexcept
	{
		/* Using try to because m_curr is incremented inside the at() access  */
		try {
			// No reason to set m_elements_ptrs.at(m_curr++) to null because it will not be used until
			// release is called for this index (and it will be overwritten).
			return m_elements_ptrs.at(m_curr++);
		} catch (...) {
			/* Catching all exceptions to assure that the pool doesn't throw any exception. */
			m_curr = m_elements_ptrs.size();
			return nullptr;
		}
	}

	/**
	 * @brief This method returns an element to the pool
	 * @param [in] element. Do not release the same element twice, the pool will break, this is undefined behavior.
	 */
	inline void
	release(struct doca_buf *element) noexcept
	{
		/* Using try to because m_curr is decremented inside the at() access */
		try {
			m_elements_ptrs.at(--m_curr) = element;
		} catch (...) {
			/* Catching all exceptions to assure that the pool doesn't throw any exception. */
			m_curr = 0;
		}
	}

	/**
	 * This method returns the number of free elements in the pool
	 * @return uint32_t
	 */
	inline uint32_t
	num_free_elements() const noexcept
	{
		return static_cast<uint32_t>(m_elements_ptrs.size() - m_curr);
	}

	/**
	 * This method returns the pool size
	 * @return uint32_t
	 */
	inline uint32_t
	pool_size() const noexcept
	{
		return static_cast<uint32_t>(m_elements_ptrs.size());
	}

	/**
	 * This method returns the number of allocated elements in the pool
	 * @return uint32_t
	 */
	inline uint32_t
	num_allocated_elements() const noexcept
	{
		return static_cast<uint32_t>(m_curr);
	}

	/**
	 * @bfief This method returns the amount of memory used for the pool in bytes
	 * @return size_t
	 */
	inline size_t memory_usage () const noexcept
	{
		return m_elements_buffer.size();
	}

private:

	/** doca_buf size is variable (according to the extensions). This vector holds the memory for the buffers */
	std::vector<uint8_t> m_elements_buffer;
	/** This vector holds references to the elements. It is modified every time that release is called */
	std::vector<doca_buf *> m_elements_ptrs;
	size_t m_curr{0};

	// Non copyable
	buf_pool(buf_pool const &) = delete;
	buf_pool &operator=(buf_pool const &) = delete;
	buf_pool(buf_pool &&) = delete;
	buf_pool &operator=(buf_pool &&) = delete;
};

} /* namespace doca_core */


/** @} */
