/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <tuple>

#include <doca_log.h>

#include <memory_region.hpp>
#include <idevx_adapter.hpp>
#include "alias_memory_region.hpp"

DOCA_LOG_REGISTER(CORE::ALIAS_MR);

using namespace device_abstraction;

create_mr_ret alias_memory_region::create(std::shared_ptr<idevx_adapter> devx_adapter,
					  std::shared_ptr<iprotection_domain> protection_domain, uint32_t lkey,
					  uint16_t other_vhca_id,
					  const allow_other_vhca_access_key_t &access_key)noexcept
{
	try {
		auto mkey_obj = std::make_shared<alias_memory_region>(devx_adapter, protection_domain);
		auto status = mkey_obj->initialize(lkey, other_vhca_id, access_key);
		if (status != DOCA_SUCCESS)
			mkey_obj.reset();

		return {status, mkey_obj};
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate alias memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<alias_memory_region>(nullptr)};
	}
}

alias_memory_region::alias_memory_region(std::shared_ptr<idevx_adapter> devx_adapter,
					 std::shared_ptr<iprotection_domain> protection_domain) noexcept
	: memory_region_base(devx_adapter, protection_domain)
{
}

doca_error_t
alias_memory_region::initialize(uint32_t lkey, uint16_t other_vhca_id,
				const allow_other_vhca_access_key_t &access_key) noexcept
{
	create_alias_mkey_in in = { 0 };
	create_alias_mkey_in out = { 0 };

	auto status = fill_alias_memory_region_cmd(in, m_pd, lkey, other_vhca_id, access_key);
	if (status != DOCA_SUCCESS)
		return status;

	std::tie(status, m_mkey) = m_devx_adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create mkey object");
		return status;
	}

	status = get_alias_memory_region_key_index(out, m_key_index);
	if (status != DOCA_SUCCESS) {
		m_mkey->destroy();
		m_mkey.reset();
		return status;
	}

	return DOCA_SUCCESS;
}
