/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <tuple>

#include <doca_log.h>
#include <multi_os_prm.h>

#include "channel_connection.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using create_channel_connection_in = uint32_t[MLX5_ST_SZ_DW(create_channel_connection_obj_in)];
using create_channel_connection_out = uint32_t[MLX5_ST_SZ_DW(create_channel_connection_obj_out)];

static doca_error_t
fill_create_channel_connection_cmd(create_channel_connection_in &in,
	ichannel_connection::attributes const &attr) noexcept
{
	if (attr.service_name.length() >= MLX5_CHANNEL_SERVICE_NAME_LENGTH) {
		DOCA_DLOG_ERR("Invalid channel service name length");
		return DOCA_ERROR_INVALID_VALUE;
	}

	// header
	void *hdr = MLX5_ADDR_OF(create_channel_connection_obj_in, in, hdr);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJ_TYPE_CHANNEL_CONNECTION);

	// obj_context
	void *obj_context = MLX5_ADDR_OF(create_channel_connection_obj_in, in, ch_conn_obj);
	MLX5_SET(channel_connection_obj, obj_context, my_qpn_or_dci, attr.qpn_or_dci);

	if (attr.transport_type == ichannel_connection::transport_types::dc)
		MLX5_SET(channel_connection_obj, obj_context, transport_type, MLX5_CHANNEL_TRANSPORT_TYPE_DC);
	else
		MLX5_SET(channel_connection_obj, obj_context, transport_type, MLX5_CHANNEL_TRANSPORT_TYPE_UD);

	if (attr.connection_type == ichannel_connection::connection_types::vhca_id) {
		MLX5_SET(channel_connection_obj, obj_context, connection_type, MLX5_CHANNEL_CONNECT_BY_VHCA_ID);
		MLX5_SET(channel_connection_obj, obj_context, dst_vhca_id, attr.dst_vhca_id);
	} else {
		MLX5_SET(channel_connection_obj, obj_context, connection_type, MLX5_CHANNEL_CONNECT_BY_SERVICE_NAME);
		MLX5_SET64(channel_connection_obj, obj_context, match_token, attr.token);

		void *service_name = MLX5_ADDR_OF(channel_connection_obj, obj_context, service_name);
		try {
			attr.service_name.copy((char *)service_name, MLX5_CHANNEL_SERVICE_NAME_LENGTH);
		} catch (...) {
			DOCA_DLOG_ERR("Failed to copy service name");
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	return DOCA_SUCCESS;
}

static doca_error_t
get_channel_connection_id(create_channel_connection_out &out, uint32_t &id) noexcept
{
	void *hdr = MLX5_ADDR_OF(create_channel_connection_obj_out, out, hdr);

	auto status = MLX5_GET(general_obj_out_cmd_hdr, hdr, status);
	if (status != 0) {
		auto syndrome = MLX5_GET(general_obj_out_cmd_hdr, hdr, syndrome);
		(void)syndrome;

		DOCA_DLOG_ERR("Failed to create channel connection object, status = %u, syndrome = %u", status, syndrome);

		return DOCA_ERROR_DRIVER;
	}

	id = MLX5_GET(general_obj_out_cmd_hdr, hdr, obj_id);

	return DOCA_SUCCESS;
}

ichannel_connection::create_ret
channel_connection::create(std::shared_ptr<idevx_adapter> adapter, ichannel_connection::attributes const &attr) noexcept
{
	std::shared_ptr<channel_connection> channel_connection_obj{nullptr};
	try {
		channel_connection_obj = std::make_shared<channel_connection>();
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create channel connection object");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<channel_connection>(nullptr)};
	}

	auto status = channel_connection_obj->initialize(adapter, attr);
	if (status != DOCA_SUCCESS)
		channel_connection_obj.reset();

	return {status, channel_connection_obj};
}

channel_connection::~channel_connection() noexcept
{
	if (destroy() != DOCA_SUCCESS)
		DOCA_DLOG_ERR("Failed to destroy channel connection");
}

doca_error_t
channel_connection::initialize(std::shared_ptr<idevx_adapter> adapter, ichannel_connection::attributes const &attr) noexcept
{
	if (adapter == nullptr) {
		DOCA_DLOG_ERR("Invalid devx adapter handle");
		return DOCA_ERROR_INVALID_VALUE;
	}
	m_devx_adapter = adapter;

	create_channel_connection_in in{0};
	create_channel_connection_out out{0};

	auto status = fill_create_channel_connection_cmd(in, attr);
	if (status != DOCA_SUCCESS)
		return status;

	std::unique_ptr<idevx_object> devx_object;
	std::tie(status, devx_object) = m_devx_adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS)
		return status;

	status = get_channel_connection_id(out, m_id);
	if (status != DOCA_SUCCESS)
		return status;

	m_channel_connection = std::move(devx_object);

	return DOCA_SUCCESS;
}

doca_error_t
channel_connection::destroy() noexcept
{
	if (m_channel_connection != nullptr) {
		auto status = m_channel_connection->destroy();
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to destroy channel connection, status = %u", status);
			return status;
		}
		m_channel_connection = nullptr;
	}

	return DOCA_SUCCESS;
}

uint32_t
channel_connection::get_id() const noexcept
{
	return m_id;
}
