/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <memory>

#include <doca_error.h>
#include <idevx_adapter.hpp>

#include "ichannel_connection.hpp"

namespace device_abstraction {

class channel_connection : public ichannel_connection
{
public:

	/**
	 * @brief This method allocates and initializes a channel_connection
	 * @param [in] adapter devx adapter handle
	 * @param [in] attr channel connection attributes
	 * @return ichannel_connection::create_ret (see interface)
	 */
	static ichannel_connection::create_ret
	create(std::shared_ptr<idevx_adapter> adapter, ichannel_connection::attributes const &attr) noexcept;

	/**
	 * Constructor
	 */
	channel_connection() noexcept = default;

	/**
	 * Destructor
	 */
	virtual ~channel_connection() noexcept;

	/**
	 * @brief This method initializes the channel_connection
	 * @param [in] adapter devx adapter handle
	 * @param [in] attr channel connection attributes
	 * @return doca_error_t
	 */
	doca_error_t
	initialize(std::shared_ptr<idevx_adapter> adapter, ichannel_connection::attributes const &attr) noexcept;

	/**
	 * @brief This method explicitly destroys the channel connection.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept override;

	/**
	 * @brief This method returns the channel connection id
	 * @return uint32_t
	 */
	virtual uint32_t get_id() const noexcept override;

	channel_connection(channel_connection const &) = delete;
	channel_connection &operator=(channel_connection const &) = delete;
	channel_connection(channel_connection &&) = delete;
	channel_connection &operator=(channel_connection &&) = delete;

private:

	std::shared_ptr<idevx_adapter> m_devx_adapter{nullptr};
	std::unique_ptr<idevx_object> m_channel_connection{nullptr};
	uint32_t m_id{0};
};

} /* namespace device_abstraction */
