/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <tuple>

#include <doca_log.h>
#include <multi_os_prm.h>

#include "channel_service.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using create_channel_service_in = uint32_t[MLX5_ST_SZ_DW(create_channel_service_obj_in)];
using create_channel_service_out = uint32_t[MLX5_ST_SZ_DW(create_channel_service_obj_out)];

/**
 * @brief This method fills the create channel service command
 * @param [in][out] command buffer
 * @param [in] channel service attributes
 * @return doca_error_t
 */
static doca_error_t
fill_create_channel_service_cmd(create_channel_service_in &in, ichannel_service::attributes const &attr) noexcept
{
	if (attr.service_name.length() >= MLX5_CHANNEL_SERVICE_NAME_LENGTH) {
		DOCA_DLOG_ERR("Invalid channel service name length");
		return DOCA_ERROR_INVALID_VALUE;
	}

	// header
	void *hdr = MLX5_ADDR_OF(create_channel_service_obj_in, in, hdr);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJ_TYPE_CHANNEL_SERVICE);

	// obj_context
	void *obj_context = MLX5_ADDR_OF(create_channel_service_obj_in, in, ch_svc_obj);
	MLX5_SET(channel_service_obj, obj_context, allowed_vhca_id, attr.allowed_vhca_id);
	MLX5_SET(channel_service_obj, obj_context, enforce_token, attr.enforce_token);
	MLX5_SET(channel_service_obj, obj_context, service_qpn_or_dct, attr.qpn_or_dct);
	MLX5_SET(channel_service_obj, obj_context, max_num_connected_client, attr.max_num_connected_client);
	MLX5_SET64(channel_service_obj, obj_context, token, attr.token);

	if (attr.transport_type == ichannel_service::transport_types::dc) {
		MLX5_SET(channel_service_obj, obj_context, transport_type, MLX5_CHANNEL_TRANSPORT_TYPE_DC);
		MLX5_SET64(channel_service_obj, obj_context, dct_key, attr.dct_key);
	} else {
		MLX5_SET(channel_service_obj, obj_context, transport_type, MLX5_CHANNEL_TRANSPORT_TYPE_UD);
		MLX5_SET(channel_service_obj, obj_context, service_qkey, attr.service_qkey);
	}

	void *service_name = MLX5_ADDR_OF(channel_service_obj, obj_context, service_name);
	try {
		attr.service_name.copy((char *)service_name, MLX5_CHANNEL_SERVICE_NAME_LENGTH);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to copy service name");
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

/**
 * @brief This method returns the channel service id
 * @param [in] command buffer
 * @param [in][out] channel service id
 * @return doca_error_t
 */
static doca_error_t
get_channel_service_id(create_channel_service_out &out, uint32_t &id) noexcept
{
	void *hdr = MLX5_ADDR_OF(create_channel_service_obj_out, out, hdr);

	auto status = MLX5_GET(general_obj_out_cmd_hdr, hdr, status);
	if (status != 0) {
		auto syndrome = MLX5_GET(general_obj_out_cmd_hdr, hdr, syndrome);
		(void)syndrome;

		DOCA_DLOG_ERR("Failed to create channel service object, status = %u, syndrome = %u", status, syndrome);

		return DOCA_ERROR_DRIVER;
	}

	id = MLX5_GET(general_obj_out_cmd_hdr, hdr, obj_id);

	return DOCA_SUCCESS;
}

ichannel_service::create_ret
channel_service::create(std::shared_ptr<idevx_adapter> adapter, ichannel_service::attributes const &attr) noexcept
{
	std::shared_ptr<channel_service> channel_service_obj{nullptr};
	try {
		channel_service_obj = std::make_shared<channel_service>();
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create channel service object");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<channel_service>(nullptr)};
	}

	auto status = channel_service_obj->initialize(adapter, attr);
	if (status != DOCA_SUCCESS)
		channel_service_obj.reset();

	return {status, channel_service_obj};
}

channel_service::~channel_service() noexcept
{
	if (destroy() != DOCA_SUCCESS)
		DOCA_DLOG_ERR("Failed to destroy channel service");
}

doca_error_t
channel_service::initialize(std::shared_ptr<idevx_adapter> adapter, ichannel_service::attributes const &attr) noexcept
{
	if (adapter == nullptr) {
		DOCA_DLOG_ERR("Invalid devx adapter handle");
		return DOCA_ERROR_INVALID_VALUE;
	}
	m_devx_adapter = adapter;

	create_channel_service_in in{0};
	create_channel_service_out out{0};

	auto status = fill_create_channel_service_cmd(in, attr);
	if (status != DOCA_SUCCESS)
		return status;

	std::unique_ptr<idevx_object> devx_object;
	std::tie(status, devx_object) = m_devx_adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS)
		return status;

	status = get_channel_service_id(out, m_id);
	if (status != DOCA_SUCCESS)
		return status;

	m_channel_service = std::move(devx_object);

	return DOCA_SUCCESS;
}

doca_error_t
channel_service::destroy() noexcept
{
	if (m_channel_service != nullptr) {
		auto status = m_channel_service->destroy();
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to destroy channel service, status = %u", status);
			return status;
		}
		m_channel_service = nullptr;
	}

	return DOCA_SUCCESS;
}

uint32_t
channel_service::get_id() const noexcept
{
	return m_id;
}
