/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <memory>

#include <doca_error.h>
#include <idevx_adapter.hpp>

#include "ichannel_service.hpp"

namespace device_abstraction {

class channel_service : public ichannel_service
{
public:

	/**
	 * @brief This method allocates and initializes a channel_service
	 * @param [in] adapter devx adapter handle
	 * @param [in] attr channel service attributes
	 * @return ichannel_service::create_ret (see interface)
	 */
	static ichannel_service::create_ret
	create(std::shared_ptr<idevx_adapter> adapter, ichannel_service::attributes const &attr) noexcept;

	/**
	 * Constructor
	 */
	channel_service() noexcept = default;

	/**
	 * Destructor
	 */
	virtual ~channel_service() noexcept;

	/**
	 * @brief This method initializes the channel_service
	 * @param [in] adapter devx adapter handle
	 * @param [in] attr channel service attributes
	 * @return doca_error_t
	 */
	doca_error_t
	initialize(std::shared_ptr<idevx_adapter> adapter, ichannel_service::attributes const &attr) noexcept;

	/**
	 * @brief This method explicitly destroys the channel service.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept override;

	/**
	 * @brief This method returns the channel service id
	 * @return uint32_t
	 */
	virtual uint32_t get_id() const noexcept override;

	channel_service(channel_service const &) = delete;
	channel_service &operator=(channel_service const &) = delete;
	channel_service(channel_service &&) = delete;
	channel_service &operator=(channel_service &&) = delete;

private:

	std::shared_ptr<idevx_adapter> m_devx_adapter{nullptr};
	std::unique_ptr<idevx_object> m_channel_service{nullptr};
	uint32_t m_id{0};
};

} /* namespace device_abstraction */
