/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <cmath>
#include <tuple>

#include <multi_os_dv.h>
#include <multi_os_prm.h>
#include <doca_log.h>

#include <priv_doca_os_abstraction_pagesize.hpp>
#include <utils.hpp>
#include <idevx_adapter.hpp>
#include <imapped_user_memory.hpp>
#include <iuser_access_region.hpp>
#include "completion_queue.hpp"

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using namespace device_abstraction;

namespace {

static constexpr int core = 0;

static constexpr uint32_t cacheline_size = 64;

using create_cq_in = uint32_t[MLX5_ST_SZ_DW(create_cq_in)];
using create_cq_out = uint32_t[MLX5_ST_SZ_DW(create_cq_out)];

inline void
fill_create_cq_cmd(create_cq_in &in, std::shared_ptr<iuser_access_region> uar, uint32_t log_nb_cqes,
		   std::shared_ptr<imapped_user_memory> umem_obj, uint32_t db_umem_offset, uint32_t log_page_size,
		   uint32_t eqn, bool initiator_src_dct)
{

	MLX5_SET(create_cq_in, in, opcode, MLX5_CMD_OP_CREATE_CQ);
	MLX5_SET(create_cq_in, in, cq_context.c_eqn, eqn);
	MLX5_SET(create_cq_in, in, cq_context.cqe_sz, MLX5_CQE_SIZE_64B);
	MLX5_SET(create_cq_in, in, cq_context.log_cq_size, log_nb_cqes);
	MLX5_SET(create_cq_in, in, cq_context.initiator_src_dct, static_cast<uint8_t>(initiator_src_dct));
	MLX5_SET(create_cq_in, in, cq_context.uar_page, uar->get_id());
	MLX5_SET(create_cq_in, in, cq_umem_id, umem_obj->get_id());
	MLX5_SET(create_cq_in, in, cq_umem_valid, 1);
	MLX5_SET(create_cq_in, in, cq_context.dbr_umem_id, umem_obj->get_id());
	MLX5_SET(create_cq_in, in, cq_context.dbr_umem_valid, 1);
	MLX5_SET64(create_cq_in, in, cq_context.dbr_addr, db_umem_offset);
	/* Page size provided in granularity of 4K. */
	MLX5_SET(create_cq_in, in, cq_context.log_page_size, static_cast<uint32_t>(log_page_size - MLX5_ADAPTER_PAGE_SHIFT));
}

inline uint32_t
get_cq_cqn(create_cq_out &out)
{
	return MLX5_GET(create_cq_out, out, cqn);
}

inline void
init_cqes(struct mlx5_64b_cqe *cqes, uint32_t nb_cqes)
{
	for (uint32_t cqe_idx = 0; cqe_idx < nb_cqes; cqe_idx++) {
		auto& cqe = cqes[cqe_idx];
		cqe.op_own = (MLX5_CQE_INVALID << 4) | MLX5_CQE_OWNER_MASK;
	}
}

void
default_cqe_handler(union doca_data) {}

inline bool
is_cqe_valid(const struct mlx5_64b_cqe& cqe, uint32_t ci, uint32_t nb_cqes)
{
	const bool expected_op_owner = static_cast<bool>(ci & nb_cqes);
	const bool op_owner = static_cast<bool>(cqe.op_own & MLX5_CQE_OWNER_MASK);
	const uint8_t op_code = cqe.op_own >> MLX5_CQE_OPCODE_SHIFT;

	/* New CQE isn't found */
	if (likely((op_owner != expected_op_owner) || (op_code == MLX5_CQE_INVALID)))
		return false;

	return true;
}

}; /* namespace */

completion_queue::~completion_queue()
{
	static_cast<void>(destroy());
}

create_completion_queue_ret
completion_queue::create(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iuser_access_region> uar,
			 const icompletion_queue::attributes &attr, union doca_data user_data,
			 priv_doca_cqe_event_handler_cb_t handler) noexcept
{
	try {
		auto cq = std::make_shared<completion_queue>();

		auto ret = cq->initialize(adapter, uar, attr, user_data, handler);
		if (ret != DOCA_SUCCESS)
			cq.reset();
		return {ret, cq};
	} catch (...) {
		return {DOCA_ERROR_NO_MEMORY, nullptr};
	}
}

doca_error_t
completion_queue::initialize(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iuser_access_region> uar,
			     const icompletion_queue::attributes &attr,
			     union doca_data user_data, priv_doca_cqe_event_handler_cb_t handler) noexcept
{
	uint32_t eqn{0};
	auto ret = adapter->query_event_queue_number(core, eqn);
	if (ret != DOCA_SUCCESS)
		return ret;

	uint32_t log_nb_cqes = static_cast<uint32_t>(core_utils::log2_uint32_t(static_cast<uint32_t>(attr.nb_cqes)));
	/* Make number of CQEs a power of 2. */
	m_nb_cqes = 1U << log_nb_cqes;
	m_nb_cqes_mask = m_nb_cqes - 1;
	uint32_t cq_buffer_size = static_cast<uint32_t>(m_nb_cqes * sizeof(struct mlx5_64b_cqe));

	/* Make sure doorbell record fits inside a cacheline by itself. */
	uint32_t db_umem_offset = core_utils::align_uint32_t(cq_buffer_size, cacheline_size);
	uint32_t db_size = cacheline_size;

	uint32_t page_size = static_cast<uint32_t>(os_abstraction::priv_doca_get_page_size());
	uint32_t umem_size = core_utils::align_uint32_t(db_umem_offset + db_size, page_size);

	try {
		m_umem_buf = std::make_unique<os_abstraction::aligned_buffer<uint8_t>>(page_size, umem_size);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate umem aligned buffer. Out of memory.");
		return DOCA_ERROR_NO_MEMORY;
	}
	memset(m_umem_buf->begin(), 0, m_umem_buf->size());
	m_cqes = reinterpret_cast<struct mlx5_64b_cqe*>(m_umem_buf->begin());
	m_db_buffer = reinterpret_cast<uint32_t*>(m_umem_buf->begin() + db_umem_offset);

	std::tie(ret, m_umem_obj) = adapter->safe_create_mapped_user_memory(
		m_umem_buf->begin(), m_umem_buf->size(),
								       page_size, memory_access_flags::local_write);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create umem.");
		return ret;
	}

	create_cq_in in = {0};
	create_cq_out out = {0};
	fill_create_cq_cmd(in, uar, log_nb_cqes, m_umem_obj, db_umem_offset, core_utils::log2_uint32_t(page_size), eqn,
			   attr.initiator_src_dct);
	std::tie(ret, m_cq_obj) = adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create CQ DevX object. Fail syndrome=%u",
			      MLX5_GET(create_cq_out, out, syndrome));
		return ret;
	}
	m_uar = uar;
	m_uar_db_reg = reinterpret_cast<uint64_t*>(&static_cast<uint8_t*>(m_uar->get_page())[MLX5_CQ_DOORBELL]);

	m_cqn = get_cq_cqn(out);

	/* Mark all CQEs initially as invalid. */
	init_cqes(m_cqes, m_nb_cqes);

	m_user_data = user_data;

	m_cqe_handler = handler;
	if (m_cqe_handler == NULL)
		m_cqe_handler = default_cqe_handler;

	return DOCA_SUCCESS;
}

idevx_object*
completion_queue::get_devx_obj() const noexcept
{
	return m_cq_obj.get();
}

union doca_data
completion_queue::get_user_data() const noexcept
{
	return m_user_data;
}

uint32_t
completion_queue::get_cq_number() const noexcept
{
	return m_cqn;
}

void
completion_queue::arm() noexcept
{

	uint32_t sn = m_sni & 3;
	uint32_t ci = m_ci & 0xffffff;
	static constexpr uint32_t cmd = MLX5_CQ_DB_REQ_NOT;

	uint64_t doorbell = sn << 28 | cmd | ci;
	doorbell <<= 32;
	doorbell |= m_cqn;

	m_db_buffer[MLX5_CQ_ARM_DB] = htobe32(sn << 28 | cmd | ci);

	priv_doca_memory_barrier();

	*m_uar_db_reg = htobe64(doorbell);
	priv_doca_memory_barrier();
}

void
completion_queue::force_arm() noexcept
{
	if (m_armed_ci != m_ci)
		/*
		 * In this case we know for certain that an event was received.
		 * IMPORTANT: In case CQ was not polled after event was received, this logic will fail.
		 */
		m_sni++;
	arm();
	m_armed_ci = m_ci;
}

void
completion_queue::handle_event() noexcept
{
	++m_sni;
	m_cqe_handler(m_user_data);
}

doca_error_t
completion_queue::poll_one(struct mlx5_64b_cqe *cqe_out) noexcept
{
	const auto &cqe = m_cqes[m_ci & m_nb_cqes_mask];

	if (!is_cqe_valid(cqe, m_ci, m_nb_cqes))
		return DOCA_ERROR_AGAIN;

	memcpy(cqe_out, &cqe, sizeof(*cqe_out));
	++m_ci;

	update_ci();

	return DOCA_SUCCESS;
}

bool
completion_queue::empty() const noexcept
{
	return !is_cqe_valid(m_cqes[m_ci & m_nb_cqes_mask], m_ci, m_nb_cqes);
}

size_t
completion_queue::poll(struct mlx5_64b_cqe* cqes_out, size_t nb_max_cqes) noexcept
{
	/* Make sure returned CQEs are continuous. */
	size_t first_cqe_idx = m_ci & m_nb_cqes_mask;
	const size_t nb_cont_cqes = m_nb_cqes - first_cqe_idx;

	/* First copy CQEs to output buffer. */
	if (unlikely(nb_max_cqes > nb_cont_cqes)) {
		/* In case of CQEs wrap around. */
		memcpy(cqes_out, &m_cqes[first_cqe_idx], sizeof(*cqes_out) * nb_cont_cqes);
		memcpy(&cqes_out[nb_cont_cqes], m_cqes, sizeof(*cqes_out) * (nb_max_cqes - nb_cont_cqes));
	} else {
		memcpy(cqes_out, &m_cqes[first_cqe_idx], sizeof(*cqes_out) * nb_max_cqes);
	}

	size_t nb_cqes = 0;
	for (nb_cqes = 0; nb_cqes < nb_max_cqes; ++nb_cqes) {
		if (likely(!is_cqe_valid(cqes_out[nb_cqes], m_ci, m_nb_cqes)))
			break;
		++m_ci;
	}

	update_ci();

	return nb_cqes;
}

void
completion_queue::update_ci() noexcept
{
	/* Make sure CQE read is done before updating CI. So CQE being read does not get overwritten by HW. */
	priv_doca_memory_barrier();

	m_db_buffer[MLX5_CQ_SET_CI] = htobe32(m_ci & 0xffffff);
}

doca_error_t
completion_queue::destroy() noexcept
{
	if (m_cq_obj != nullptr) {
		auto ret = m_cq_obj->destroy();
		if (ret != DOCA_SUCCESS)
			return ret;
		m_cq_obj = nullptr;
	}

	if (m_umem_obj != nullptr) {
		auto ret = m_umem_obj->dereg();
		if (ret != DOCA_SUCCESS)
			return ret;
		m_umem_obj = nullptr;
	}

	return DOCA_SUCCESS;
}
