/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <priv_doca_types.h>

#include <priv_doca_os_abstraction_memory.hpp>
#include "icompletion_queue.hpp"

namespace device_abstraction {

class idevx_adapter;
class iuser_access_region;
class imapped_user_memory;

/**
 * Completion Queue, representing the device CQ.
 * Has unified implementation in DevX.
 */
class completion_queue : public icompletion_queue
{
public:

	/**
	 * @brief Create a completion queue
	 *
	 * @param [in] adapter The DevX adapter to use, representing a device.
	 * @param [in] uar User access region.
	 * @param [in] nb_cqes Number of maximum completion entries.
	 * @param [in] user_data An opaque structure stored within the CQ. Available later using get_user_data().
	 * @param [in] handler CQE event handler, may be called using icompletion_events_channel::handle_events(),
	 * 		       once CQE arrives on CQ. Can be NULL.
	 * @return create_completion_queue_ret. see icompletion_queue.
	 */
	static create_completion_queue_ret create(std::shared_ptr<idevx_adapter> adapter,
						  std::shared_ptr<iuser_access_region> uar,
						  const icompletion_queue::attributes &attr, union doca_data user_data,
						  priv_doca_cqe_event_handler_cb_t handler) noexcept;

	/**
	 * Constructor
	 */
	completion_queue() = default;

	/**
	 * Destructor
	 */
	virtual ~completion_queue();


	virtual doca_error_t destroy() noexcept override;

	virtual idevx_object* get_devx_obj() const noexcept override;

	virtual union doca_data get_user_data() const noexcept override;

	virtual uint32_t get_cq_number() const noexcept override;

	virtual void arm() noexcept override;

	virtual void force_arm() noexcept override;

	virtual void handle_event() noexcept override;

	virtual size_t poll(struct mlx5_64b_cqe* cqes_out, size_t nb_max_cqes) noexcept override;

	virtual doca_error_t poll_one(struct mlx5_64b_cqe *cqe_out) noexcept override;

	virtual bool empty() const noexcept override;

	completion_queue(completion_queue const &) = delete;
	completion_queue &operator=(completion_queue const &) = delete;
	completion_queue(completion_queue &&) = delete;
	completion_queue &operator=(completion_queue &&) = delete;
protected:

	virtual doca_error_t initialize(std::shared_ptr<idevx_adapter> adapter,
					std::shared_ptr<iuser_access_region> uar,
					const icompletion_queue::attributes &attr,
					union doca_data user_data,
					priv_doca_cqe_event_handler_cb_t handler) noexcept;

	virtual void update_ci() noexcept;

	struct mlx5_64b_cqe *m_cqes{nullptr};
	uint32_t m_nb_cqes{0};
	uint32_t m_nb_cqes_mask{0};
	uint32_t *m_db_buffer{nullptr};
	uint64_t *m_uar_db_reg{nullptr};
	uint32_t m_ci{0};
	uint32_t m_armed_ci{0};
	uint32_t m_sni{0};
	uint32_t m_cqn{0};
	priv_doca_cqe_event_handler_cb_t m_cqe_handler{nullptr};
	union doca_data m_user_data;

	/* Order of declaration should be maintained. */
	std::unique_ptr<os_abstraction::aligned_buffer<uint8_t>> m_umem_buf;
	std::shared_ptr<imapped_user_memory> m_umem_obj;
	std::shared_ptr<iuser_access_region> m_uar;
	std::unique_ptr<idevx_object> m_cq_obj;
};

} // namespace device_abstraction
