/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <tuple>

#include <doca_log.h>

#include <memory_region.hpp>
#include "crossing_memory_region.hpp"
#include <idevx_adapter.hpp>

DOCA_LOG_REGISTER(CORE::CROSSING_MR);

using namespace device_abstraction;

create_mr_ret
crossing_memory_region::create(std::shared_ptr<idevx_adapter> devx_adapter,
				std::shared_ptr<iprotection_domain> protection_domain,
				uint32_t introspection_mkey, uint16_t other_vhca_id) noexcept
{
	try {
		auto mkey_obj = std::make_shared<crossing_memory_region>(devx_adapter, protection_domain);
		auto status = mkey_obj->initialize(introspection_mkey, other_vhca_id);
		if (status != DOCA_SUCCESS)
			mkey_obj.reset();

		return {status, mkey_obj};
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate crossing memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<crossing_memory_region>(nullptr)};
	}
}

crossing_memory_region::crossing_memory_region(
	std::shared_ptr<idevx_adapter> devx_adapter, std::shared_ptr<iprotection_domain> protection_domain) noexcept
	: memory_region_base(devx_adapter, protection_domain)
{
}

doca_error_t
crossing_memory_region::initialize(uint32_t introspection_mkey, uint16_t other_vhca_id) noexcept
{
	create_crossing_mkey_in in = {0};
	create_crossing_mkey_out out = {0};

	auto status = fill_crossing_memory_region_cmd(in, m_pd, introspection_mkey, other_vhca_id);
	if (status != DOCA_SUCCESS)
		return status;

	std::tie(status, m_mkey) = m_devx_adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS) {
		//Most likely invalid introspection_mkey/other_vhca_id
		DOCA_DLOG_ERR("Failed to create crossing MKEY. Invalid input: introspection_mkey=%u, other_vhca_id=%u",
				introspection_mkey, other_vhca_id);
		DOCA_DLOG_ERR("Crossing mkey fail syndrome=%u", get_crossing_mkey_fail_syndrome(out));
		return DOCA_ERROR_INVALID_VALUE;
	}

	m_key_index = get_crossing_memory_region_key_index(out);

	return DOCA_SUCCESS;
}


