/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_error.h>
#include <doca_log.h>
#include <multi_os_dv.h>
#include <multi_os_prm.h>

#include "channel_connection.hpp"
#include "channel_service.hpp"
#include "completion_queue.hpp"
#include "device_adapter.hpp"

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using namespace device_abstraction;

query_hca_caps_ret
device_adapter::query_hca_caps() noexcept
{
	if (!m_hca_caps_cached) {
		const doca_error_t status = query_hca_caps_imp(m_hca_caps, false, 0);
		if (status != DOCA_SUCCESS)
			return {status, m_hca_caps};
		m_hca_caps_cached = true;
	}
	return {DOCA_SUCCESS, m_hca_caps};
};

doca_error_t
device_adapter::query_function_hca_caps(uint16_t function_id, hca_capabilities &caps) noexcept
{
	const doca_error_t ret = query_hca_caps_imp(caps, true, function_id);
	if (ret != DOCA_SUCCESS)
		return ret;
	return DOCA_SUCCESS;
}

doca_error_t
device_adapter::query_hca_caps_by_opmod(uint16_t op_mod, void *out, size_t out_size) noexcept
{
	uint32_t in[MLX5_ST_SZ_DW(query_hca_cap_in)]{0};

	/* Set operation code and operation mode of the command. */
	MLX5_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	MLX5_SET(query_hca_cap_in, in, op_mod, op_mod);

	if (send_prm_cmd(in, sizeof(in), out, out_size) != 0) {
		DOCA_DLOG_ERR("Query devx HCA capabilities with op_mod %u failed, syndrome = 0x%x",
			      op_mod, MLX5_GET(query_hca_cap_out, out, syndrome));
		/* Error is based on received syndrome.*/
		return DOCA_ERROR_DRIVER;
	}

	return DOCA_SUCCESS;
}

create_completion_queue_ret
device_adapter::create_completion_queue(std::shared_ptr<iuser_access_region> uar,
					const icompletion_queue::attributes &attr,
					union doca_data user_data, priv_doca_cqe_event_handler_cb_t handler) noexcept
{
	return completion_queue::create(get_devx_adapter(), uar, attr, user_data, handler);
}

ichannel_service::create_ret
device_adapter::create_channel_service(ichannel_service::attributes const &attr) noexcept
{
	query_hca_caps_ret ret = query_hca_caps();
	if (ret .first!= DOCA_SUCCESS)
		return {ret.first, std::shared_ptr<ichannel_service>(nullptr)};

	if ((ret.second.general_object_types_supported & MLX5_GENERAL_OBJ_TYPES_CAP_CHANNEL_SERVICE) == 0) {
		DOCA_DLOG_ERR("CHANNEL_SERVICE is not supported on this adapter");
		return {DOCA_ERROR_NOT_SUPPORTED, std::shared_ptr<ichannel_service>(nullptr)};
	}

	return channel_service::create(get_devx_adapter(), attr);
}

ichannel_connection::create_ret
device_adapter::create_channel_connection(ichannel_connection::attributes const &attr) noexcept
{
	query_hca_caps_ret ret = query_hca_caps();
	if (ret.first != DOCA_SUCCESS)
		return {ret.first, std::shared_ptr<ichannel_connection>(nullptr)};

	if ((ret.second.general_object_types_supported & MLX5_GENERAL_OBJ_TYPES_CAP_CHANNEL_CONNECTION) ==
	    0) {
		DOCA_DLOG_ERR("CHANNEL_CONNECTION is not supported on this adapter");
		return {DOCA_ERROR_NOT_SUPPORTED, std::shared_ptr<ichannel_connection>(nullptr)};
	}

	return channel_connection::create(get_devx_adapter(), attr);
}
