/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <tuple>

#include <doca_log.h>
#include <multi_os_prm.h>

#include "dynamically_connected_target.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using create_dct_in = uint32_t[MLX5_ST_SZ_DW(create_dct_in)];
using create_dct_out = uint32_t[MLX5_ST_SZ_DW(create_dct_out)];

static doca_error_t
fill_create_dct_cmd(create_dct_in &in, uint32_t pdn,
		    idynamically_connected_target::attributes const &attr) noexcept
{
	// mlx5_ifc_create_dct_in_bits
	MLX5_SET(create_dct_in, in, opcode, MLX5_CMD_OP_CREATE_DCT);

	// mlx5_ifc_dctc_bits
	void *dct_context = MLX5_ADDR_OF(create_dct_in, in, dct_context_entry);
	MLX5_SET(dctc, dct_context, cqn, attr.cq->get_cq_number());
	MLX5_SET(dctc, dct_context, port, attr.port);
	MLX5_SET(dctc, dct_context, mtu, attr.mtu);
	MLX5_SET(dctc, dct_context, srqn_xrqn, attr.rmp->get_id());
	MLX5_SET(dctc, dct_context, pd, pdn);
	MLX5_SET(dctc, dct_context, my_addr_index, attr.my_addr_index);
	MLX5_SET(dctc, dct_context, min_rnr_nak, attr.min_rnr_nak);
	MLX5_SET64(dctc, dct_context, dc_access_key, attr.dc_access_key);

	return DOCA_SUCCESS;
}

static doca_error_t
get_dynamically_connected_target_id(create_dct_out &out, uint32_t &id) noexcept
{
	auto status = MLX5_GET(create_dct_out, out, status);
	if (status != 0) {
		auto syndrome = MLX5_GET(create_dct_out, out, syndrome);
		(void)syndrome;

		DOCA_DLOG_ERR("Failed to create dct object, status = %u, syndrome = %u", status, syndrome);

		return DOCA_ERROR_DRIVER;
	}

	id = MLX5_GET(create_dct_out, out, dctn);

	return DOCA_SUCCESS;
}

idynamically_connected_target::create_ret
dynamically_connected_target::create(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iprotection_domain> pd,
				     idynamically_connected_target::attributes const &attr) noexcept
{
	std::shared_ptr<dynamically_connected_target> dct_obj{nullptr};
	try {
		dct_obj = std::make_shared<dynamically_connected_target>();
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create dynamically connected target object");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<dynamically_connected_target>(nullptr)};
	}

	auto status = dct_obj->initialize(adapter, pd, attr);
	if (status != DOCA_SUCCESS)
		dct_obj.reset();

	return {status, dct_obj};
}

dynamically_connected_target::~dynamically_connected_target()
{
	if (destroy() != DOCA_SUCCESS)
		DOCA_DLOG_ERR("Failed to destroy dynamically connected target");
}

doca_error_t
dynamically_connected_target::initialize(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iprotection_domain> pd,
					 idynamically_connected_target::attributes const &attr) noexcept
{
	if (adapter == nullptr) {
		DOCA_DLOG_ERR("Invalid devx adapter handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (pd == nullptr) {
		DOCA_DLOG_ERR("Invalid protection domain handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (attr.cq == nullptr) {
		DOCA_DLOG_ERR("Invalid completion queue handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (attr.rmp == nullptr) {
		DOCA_DLOG_ERR("Invalid receive memory pool handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	m_devx_adapter = adapter;
	m_pd = pd;
	m_cq = attr.cq;
	m_rmp = attr.rmp;

	create_dct_in in{0};
	create_dct_out out{0};

	auto status = fill_create_dct_cmd(in, m_pd->get_pd_id(), attr);
	if (status != DOCA_SUCCESS)
		return status;

	std::unique_ptr<idevx_object> devx_object;
	std::tie(status, devx_object) = m_devx_adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS)
		return status;

	status = get_dynamically_connected_target_id(out, m_id);
	if (status != DOCA_SUCCESS)
		return status;

	m_dct = std::move(devx_object);

	return DOCA_SUCCESS;
}

doca_error_t
dynamically_connected_target::destroy() noexcept
{
	if (m_dct != nullptr) {
		auto status = m_dct->destroy();
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to destroy dynamically connected target, status = %u", status);
			return status;
		}
		m_dct = nullptr;
	}

	return DOCA_SUCCESS;
}

uint32_t
dynamically_connected_target::get_id() const noexcept
{
	return m_id;
}
