/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <memory>

#include <doca_error.h>
#include <idevx_adapter.hpp>
#include <iprotection_domain.hpp>
#include <icompletion_queue.hpp>
#include <ireceive_memory_pool.hpp>

#include "idynamically_connected_target.hpp"

namespace device_abstraction {

class dynamically_connected_target : public idynamically_connected_target
{
public:

	/**
	 * @brief This method allocates and initializes a dct
	 * @param [in] adapter devx adapter handle
	 * @param [in] pd protection domain
	 * @param [in] attr dct attributes
	 * @return idynamically_connected_target::create_ret (see interface)
	 */
	static idynamically_connected_target::create_ret
	create(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iprotection_domain> pd,
	       idynamically_connected_target::attributes const &attr) noexcept;

	/**
	 * Constructor
	 */
	dynamically_connected_target() noexcept = default;

	/**
	 * Destructor
	 */
	virtual ~dynamically_connected_target();

	/**
	 * @brief This method initializes the dct
	 * @param [in] adapter devx adapter handle
	 * @param [in] pd protection domain
	 * @param [in] attr dct attributes
	 * @return doca_error_t
	 */
	doca_error_t
	initialize(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iprotection_domain> pd,
		   idynamically_connected_target::attributes const &attr) noexcept;

	/**
	 * @brief This method explicitly destroys the dct.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept override;

	/**
	 * @brief This method returns the dct id
	 * @return uint32_t
	 */
	virtual uint32_t get_id() const noexcept override;

	dynamically_connected_target(dynamically_connected_target const &) = delete;
	dynamically_connected_target &operator=(dynamically_connected_target const &) = delete;
	dynamically_connected_target(dynamically_connected_target &&) = delete;
	dynamically_connected_target &operator=(dynamically_connected_target &&) = delete;

private:

	std::shared_ptr<idevx_adapter> m_devx_adapter{nullptr};
	std::shared_ptr<iprotection_domain> m_pd{nullptr};
	std::shared_ptr<icompletion_queue> m_cq{nullptr};
	std::shared_ptr<ireceive_memory_pool> m_rmp{nullptr};
	std::unique_ptr<idevx_object> m_dct{nullptr};

	uint32_t m_id{0};
};

} /* namespace device_abstraction */
