/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <utility>
#include <memory>
#include <string>

#include <doca_error.h>

namespace device_abstraction {

class ichannel_connection
{
public:

	enum class connection_types {
		service_name = 0,
		vhca_id = 1,
	};

	enum class transport_types {
		ud = 0,
		dc = 1,
	};

	struct attributes {
		ichannel_connection::connection_types connection_type;
		ichannel_connection::transport_types transport_type;
		uint16_t dst_vhca_id;
		uint32_t qpn_or_dci;
		uint64_t token;
		std::string service_name;
	};

	/**
	 * @brief This pair is used as the return value for creating a channel connection
	 * @first: doca_error_t
	 *         DOCA_SUCCESS
	 *         DOCA_ERROR_NO_MEMORY
	 *         DOCA_ERROR_INVALID_VALUE
	 *         DOCA_ERROR_NOT_CONNECTED (devx failure)
	 *         DOCA_ERROR_DRIVER (devx failure)
	 * @second: shared pointer of the channel_connection.
	 */
	using create_ret = std::pair<doca_error_t, std::shared_ptr<ichannel_connection>>;

	/**
	 * Destructor
	 * Note: concrete destructor will attempt to destroy the channel connection.
	 */
	virtual ~ichannel_connection() noexcept = default;

	/**
	 * @brief Destroy the channel connection.
	 * @details Destroys the channel connection, and returns an error in case an error was encountered.
	 * The destructor can be used instead to destroy without returning an error.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/**
	 * @brief this method returns the channel connection id
	 * return uint32_t
	 */
	virtual uint32_t get_id() const noexcept = 0;

	ichannel_connection(ichannel_connection const &) = delete;
	ichannel_connection &operator=(ichannel_connection const &) = delete;
	ichannel_connection(ichannel_connection &&) = delete;
	ichannel_connection &operator=(ichannel_connection &&) = delete;

protected:
	/**
	 * Constructor
	 */
	ichannel_connection() noexcept = default;
};

} /* namespace device_abstraction */
