/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <utility>
#include <memory>
#include <string>

#include <doca_error.h>

namespace device_abstraction {

class ichannel_service
{
public:

	enum class transport_types {
		ud = 0,
		dc = 1,
	};

	struct attributes {
		ichannel_service::transport_types transport_type;
		uint16_t allowed_vhca_id;
		uint8_t enforce_token;
		uint32_t qpn_or_dct;
		uint32_t service_qkey;
		uint64_t token;
		uint64_t dct_key;
		uint16_t max_num_connected_client;
		std::string service_name;
	};

	/**
	 * @brief This pair is used as the return value for creating a channel service
	 * @first: doca_error_t
	 *         DOCA_SUCCESS
	 *         DOCA_ERROR_NO_MEMORY
	 *         DOCA_ERROR_INVALID_VALUE
	 *         DOCA_ERROR_NOT_SUPPORTED
	 *         DOCA_ERROR_NOT_CONNECTED (devx failure)
	 *         DOCA_ERROR_DRIVER (devx failure)
	 * @second: shared pointer of the channel_service.
	 */
	using create_ret = std::pair<doca_error_t, std::shared_ptr<ichannel_service>>;

	/**
	 * Destructor
	 * Note: concrete destructor will attempt to destroy the channel service.
	 */
	virtual ~ichannel_service() noexcept = default;

	/**
	 * @brief Destroy the channel service.
	 * @details Destroys the channel service, and returns an error in case an error was encountered.
	 * The destructor can be used instead to destroy without returning an error.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/**
	 * @brief this method returns the channel service id
	 * return uint32_t
	 */
	virtual uint32_t get_id() const noexcept = 0;

	ichannel_service(ichannel_service const &) = delete;
	ichannel_service &operator=(ichannel_service const &) = delete;
	ichannel_service(ichannel_service &&) = delete;
	ichannel_service &operator=(ichannel_service &&) = delete;

protected:
	/**
	 * Constructor
	 */
	ichannel_service() noexcept = default;
};

} /* namespace device_abstraction */
