/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <utility>
#include <memory>

#include <doca_error.h>
#include <doca_types.h>
#include <priv_doca_types.h>

namespace device_abstraction {

class icompletion_queue;

/**
 * Event channel used for reporting of completion events.
 * Multiple CQs can be registered to the channel, such that the channel can run a handler, once a CQE arrives at one
 * of the CQs.
 */
class icompletion_events_channel
{
public:

	/**
	 * Destructor
	 */
	virtual ~icompletion_events_channel() = default;

	/**
	 * @brief Register CQ to be monitored by the channel.
	 * @details Adds CQ to the channel, such that the channel starts receiving completion events,
	 * signaling that some CQE arrived.
	 * In case a CQE handler was provided on creation of CQ, it will be called each time a CQE event is received.
	 * @param [in] cq
	 * Represents the CQ that will receive CQEs.
	 * @return doca_error_t
	 */
	virtual doca_error_t register_cq(std::shared_ptr<icompletion_queue> cq) noexcept = 0;

	/**
	 * @brief Unregister CQ from the channel.
	 * @details Removes CQ from list of monitored CQs by the channel, such that the channel will no longer
	 * receive completion events related to this CQ.
	 * @param [in] cq
	 * A CQ previously provided in register_cq.
	 * @return doca_error_t
	 */
	virtual doca_error_t unregister_cq(std::shared_ptr<icompletion_queue> cq) noexcept = 0;

	/**
	 * @brief Arm all CQs that require arming.
	 * @details method for arming of all registered CQs in case they require arming.
	 * This method should be called in case user is interested to continue receiving events after calling
	 * handle_events().
	 */
	virtual void arm() noexcept = 0;

	/**
	 * @brief Arm all CQs regardless of their "armed" status.
	 * @details method for arming of all registered CQs
	 * This method should be called in case user is interested to continue receiving events after calling
	 * handle_events().
	 */
	virtual void force_arm() noexcept = 0;

	/**
	 * @brief Handle all completion events.
	 * @details Method for handling of all completion events, by calling handler of each CQ that has received a CQE.
	 * This method should only be called, in case an event has been triggered on the handle of this channel.
	 * @param [in] overlapped
	 * Relevant for Windows only, and represents the OVERLAPPED object received as part of the triggered event.
	 * For linux this argument will be ignored.
	 * @return doca_error_t
	 */
	virtual doca_error_t handle_events(doca_event_handle_t handle = 0) noexcept = 0;

	/**
	 * @brief Gets an event handle used for waiting for events.
	 * @details For Linux will return a file descriptor. For Windows will return an overlapped file.
	 * @return doca_error_t
	 */
	virtual doca_event_handle_t get_handle() const noexcept = 0;

	icompletion_events_channel(icompletion_events_channel const &) = delete;
	icompletion_events_channel &operator=(icompletion_events_channel const &) = delete;
	icompletion_events_channel(icompletion_events_channel &&) = delete;
	icompletion_events_channel &operator=(icompletion_events_channel &&) = delete;

protected:
	/**
	 * Constructor
	 */
	icompletion_events_channel() noexcept = default;
};

/**
 * Return value for creating a completion events channel.
 */
using create_comp_events_channel_ret = std::pair<doca_error_t, std::shared_ptr<icompletion_events_channel>>;

} // namespace device_abstraction
