/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <utility>
#include <memory>

#include <doca_types.h>

struct mlx5_64b_cqe;
struct mlx5_error_cqe;

namespace device_abstraction {

class idevx_object;

/**
 * Completion Queue, representing the device CQ.
 */
class icompletion_queue
{
public:

	struct attributes {
		size_t nb_cqes;
		bool initiator_src_dct;
	};

	/**
	 * Destructor
	 */
	virtual ~icompletion_queue() = default;

	/**
	 * @brief Destroy the CQ.
	 * @details Destroys the CQ, and returns an error in case an error was encountered.
	 * The destructor can be used instead to destroy without returning an error.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/** @brief Get devx object of the CQ. */
	virtual idevx_object* get_devx_obj() const noexcept = 0;

	/** @brief Get user data provided on CQ creation. */
	virtual union doca_data get_user_data() const noexcept = 0;

	/** @brief Get CQ number */
	virtual uint32_t get_cq_number() const noexcept = 0;

	/**
	 * @brief Arms the CQ to receive next completion event.
	 * It is recommend not to use this directly, but instead use the icompletion_events_channel.
	 */
	virtual void arm() noexcept = 0;

	/**
	 * @brief Attempts to verify that event was previously received before arming the CQ.
	 * It is recommend not to use this directly, but instead use the icompletion_events_channel.
	 * @note Prefer using arm() & handle_event() instead.
	 * @note Assumes that CQ is polled at least once, after event was received.
	 */
	virtual void force_arm() noexcept = 0;

	/**
	 * @brief Handle CQE event
	 * @details This method may be called from icompletion_events_channel::handle_events(), in case a CQE arrives.
	 * @return doca_error_t
	 */
	virtual void handle_event() noexcept = 0;

	/**
	 * @brief Poll CQ
	 * @details This method is used for polling a number of CQEs.
	 * The range [0, *nb_polled - 1] always holds valid CQEs.
	 * Returned CQEs may include CQEs with error.
	 * @note Returned CQEs are consumed, in the actual CQ, and the CI is updated.
	 * cqes_out will hold a copy of the entries.
	 *                     -------------------------
	 *               0 --> |       Valid CQE       |
	 *                     -------------------------
	 *                     -------------------------
	 *               1 --> |       Valid CQE       |
	 *                     -------------------------
	 *                            ...
	 *                     -------------------------
	 *   nb_polled - 1 --> |       Valid CQE       |
	 *                     -------------------------
	 *                     -------------------------
	 *       nb_polled --> |      Invalid CQE      |
	 *                     -------------------------
	 *                            ...
	 *                     -------------------------
	 * nb_max_cqes - 1 --> |      Invalid CQE      |
	 *                     -------------------------
	 * @param [out] cqes_out Array to hold polled CQEs. MUST NOT BE NULL.
	 * @param [in] nb_max_cqes Number of maximum CQEs to poll, and size of cqes_out array.
	 * @return number of successfully polled CQEs (nb_polled). Where nb_polled <= nb_max_cqes.
	 */
	virtual size_t poll(struct mlx5_64b_cqe* cqes_out, size_t nb_max_cqes) noexcept = 0;

	/**
	 * @brief Poll CQ for one completion
	 * @details This method is used for polling one CQE.
	 * In case no CQE was found DOCA_ERROR_AGAIN is returned.
	 * In case a CQE or CQE with error was found, DOCA_SUCCESS is returned.
	 * @note Returned CQE is consumed, from the actual CQ, and the CI is updated. cqe_out will hold a copy of the entry.
	 * @param [out] cqe_out The received CQE. MUST NOT BE NULL.
	 * @return
	 * DOCA_SUCCESS - CQE or CQE with error was found.
	 * DOCA_ERROR_AGAIN - No CQE was found.
	 */
	virtual doca_error_t poll_one(struct mlx5_64b_cqe *cqe_out) noexcept = 0;

	/**
	 * @brief Check if there are no CQEs available.
	 * @details Check if there is at least 1 CQE present in the CQ.
	 * @return
	 * true - no CQEs available.
	 * false - at least 1 CQE available.
	 */
	virtual bool empty() const noexcept = 0;

	icompletion_queue(icompletion_queue const &) = delete;
	icompletion_queue &operator=(icompletion_queue const &) = delete;
	icompletion_queue(icompletion_queue &&) = delete;
	icompletion_queue &operator=(icompletion_queue &&) = delete;

protected:
	/**
	 * Constructor
	 */
	icompletion_queue() noexcept = default;
};


/**
 * Return value for creating a completion events channel.
 */
using create_completion_queue_ret = std::pair<doca_error_t, std::shared_ptr<icompletion_queue>>;

} // namespace device_abstraction
