/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <memory>
#include <list>
#include <string>
#include <vector>
#include <unordered_map>

#include <priv_doca_types.h>
#include <doca_error.h>
#include <doca_dev.h>

#include "iprotection_domain.hpp"
#include "iuser_access_region.hpp"
#include "idevx_adapter.hpp"
#include "imapped_user_memory.hpp"
#include "icompletion_events_channel.hpp"
#include "icompletion_queue.hpp"
#include "ichannel_service.hpp"
#include "ichannel_connection.hpp"

#ifdef __linux__
struct ibv_context;
#endif /* __linux__ */

namespace device_abstraction {

enum atomic_capability {
	IB_ATOMIC_NONE,
	IB_ATOMIC_HCA,
	IB_ATOMIC_GLOB,
};

struct device_attributes {
	uint8_t phys_port_cnt{ 0 };
};

struct hca_capabilities {
	uint16_t vhca_id{0};
	bool vhca_resource_manager{false};
	bool hca_cap_2{false};
	bool cross_gvmi_mkey_enabled{false};
	bool emulation_manager_virtio_net{false};
	bool emulation_manager_virtio_blk{false};
	bool emulation_manager_nvme{false};
	bool emulation_manager_virtio_fs{false};
	bool hotplug_manager{false};
	bool eswitch_manager{false};
	bool introspection_mkey_access_allowed{false};
	uint32_t introspection_mkey{0};
	uint64_t general_object_types_supported{0ULL};
	bool ib_port_type{false};
	bool crossing_vhca_mkey_supported{false};
};

struct port_info {
	static const uint16_t sc_invalid_vhca_id{0xFFFF};
	bool vhca_id_valid{ false };
	uint16_t vhca_id{sc_invalid_vhca_id};
	uint16_t vport{sc_invalid_vhca_id};
};

struct vuid_t {
	static const uint32_t vuid_bytes{ 128 };
	uint8_t id[vuid_bytes];
};

enum class emulation_type {
	UNKNOWN = 0,
	VIRTIO_FS,
	VIRTIO_NET,
	VIRTIO_BLK,
	NVME,
};

struct function_info {
	doca_pci_func_type func_type { doca_pci_func_type::DOCA_PCI_FUNC_PF};
	uint16_t pf_vhca_id{0};
	uint16_t vhca_id{0};
	uint16_t bdf{0};
	uint16_t total_vf_cnt{0};
	uint16_t vf_cnt{0};
	uint16_t function_id{0};
};

struct emulated_devinfo {
	uint16_t vhca_id{0};
	uint16_t pci_bdf{0};
	uint16_t nb_total_vfs{0};
	bool hotplug {false};
	bool vf_exist {false};
};

/**
 * get_devices_ret
 * @ error:
 *   - DOCA_SUCCESS if got emulated devices list
 *   - DOCA_ERROR_NOT_FOUND if failed to get emulated devices list.
 *   - DOCA_ERROR_NO_MEMORY if failed to allocate memory.
 * @ devices
 */
using get_emulated_devices_ret = std::pair<doca_error_t, std::vector<emulated_devinfo>>;

/**
 * get_function_info_ret
 * @ error:
 *   - DOCA_SUCCESS if got function_info list
 *   - DOCA_ERROR_IO_FAILED if failed to get function_info list.
 *   - DOCA_ERROR_NO_MEMORY if failed to allocate memory.
 * @ functions info
 */
using get_function_info_ret = std::pair<doca_error_t, std::vector<function_info>>;

using query_hca_caps_ret = std::pair<doca_error_t, const hca_capabilities &>;

class idevice_adapter
{
public:

	/**
	 * Destructor
	 */
	virtual ~idevice_adapter() = default;

	/**************************************************************************
	 *
	 * Control
	 *
	 **************************************************************************/

	/**
	 * @brief This method opens the device
	 * @return doca_error_t
	 */
	virtual doca_error_t open() noexcept = 0;

	/**
	 * @brief This method closes the device
	 */
	virtual doca_error_t close() noexcept = 0;

	/**************************************************************************
	 *
	 * Queries
	 *
	 **************************************************************************/

	/**
	 * @brief This method compares two instances of the device adapter
	 * @param [in] other device adapter
	 * @return true if equal, false if not.
	 */
	virtual bool operator==(const idevice_adapter &other) const noexcept = 0;

	/**
	 * @brief This method returns the device name
	 * @return string
	 */
	virtual const std::string &get_name() const noexcept = 0;

	/**
	 * @brief This method returns the vendor id
	 * @return uint32_t
	 */
	virtual uint32_t get_vendor_id() const noexcept = 0;

	/**
	 * @brief This method returns the vendor part id
	 * @return uint32_t
	 */
	virtual uint32_t get_vendor_part_id() const noexcept = 0;

	/**
	 * @brief This method queries device HCA caps and caches the result
	 * @return query_hca_caps_ret
	 */
	virtual query_hca_caps_ret query_hca_caps() noexcept = 0;

	/**
	 * @brief This method queries function HCA caps
	 * @param [in] function_id
	 * @return query_hca_caps_ret
	 */
	virtual doca_error_t query_function_hca_caps(uint16_t function_id, hca_capabilities &caps) noexcept = 0;

	/**
	 * @brief This method queries vuids
	 * @param [in] vhca id
	 * @param [out] vuid
	 * @return doca_error_t
	 */
	virtual doca_error_t query_vuid(uint16_t vhca_id, vuid_t &vuid) noexcept = 0;

	/**
	 * @brief This method queries VUIDs of VFs through their PF
	 * @param [in] vhca_id The vHCA ID of the PF
	 * @param [in, out] vuids VUIDs of the VFs related to the PF
	 * @return doca_error_t
	 */
	virtual doca_error_t query_pf_vfs_vuids(uint16_t vhca_id, std::vector<vuid_t> &vuids) noexcept = 0;

	/**
	 * @brief this method queries the number of physical ports
	 * @param [out] number of physical ports
	 * @return doca_error_t
	 */
	virtual doca_error_t query_physical_ports_count(uint8_t &port_count) noexcept = 0;

	/**
	 * @brief this method queries port info
	 * @param [in] port id
	 * @param [out] port info
	 * @return doca_error_t
	 */
	virtual doca_error_t query_port_info(int port_id, port_info &info) noexcept = 0;

	/**
	 * @brief This method queries the device PCI bus device function
	 * @param [out] bus
	 * @param [out] device
	 * @param [out] function
	 * @return doca_error_t
	 */
	virtual doca_error_t query_pci_bdf(uint8_t &bus, uint8_t &device, uint8_t &function) const noexcept = 0;

	/**
	 * @brief This method queries the device ipv4 address.
	 * This API is used by the DOCA core and is OS specific, therefore the opaque data types.
	 * @param [in] opaque ipv4 address
	 * @param [in] buffer size
	 * @return doca_error_t
	 */
	virtual doca_error_t query_ipv4_address(void *value, size_t size) const noexcept = 0;

	/**
	 * @brief This method queries the device ipv6 address
	 * This API is used by the DOCA core and is OS specific, therefore the opaque data types.
	 * @param [in] opaque ipv6 address
	 * @param [in] buffer size
	 * @return doca_error_t
	 */
	virtual doca_error_t query_ipv6_address(void *value, size_t size) const noexcept = 0;

	/**
	 * @brief This method retrieves the device interface name
	 * @param [out] interface name
	 * @return doca_error_t
	 */
	virtual doca_error_t get_interface_name(std::string &interface_name) const noexcept = 0;

	/**
	 * @brief This method retrieves the device ibdev name
	 * @param [out] ibdev name
	 * @return doca_error_t
	 */
	virtual doca_error_t get_ibdev_name(std::string &ibdev_name) const noexcept = 0;

	/**
	 * @brief This method retrieves the information about the pci functions (Host PF, VFs and SFs) that are
	 * connected to the eswitch.
	 * Note - the SF information is only supported for Ethernet ports
	 * @return get_function_info_ret
	 */
	virtual get_function_info_ret get_net_pci_topology() noexcept = 0;

	/**
	 * @brief This method checks if the device supports representors.
	 * @return True if the device supports representors, False otherwise.
	 */
	virtual bool is_net_representors_supported() noexcept = 0;

	#ifdef __linux__
	/**
	 * @brief This method returns the ibv context of the device. The method is only supported for Linux
	 * Note using the ibv context out of the scope of the device adapter voids the ability of the device
	 * adapter to keep reference and may fail deallocating the ibv context at destruction.
	 * In addition to that, in case the ibv_context hasn't been initialized, this method initializes it.
	 * This is why this method is not declared as const.
	 * @return ibv context
	 */
	virtual ibv_context *get_ibv_context() noexcept = 0;
	#endif // __linux__

	/**************************************************************************
	 *
	 * PRM Objects Factory
	 *
	 **************************************************************************/

	/**
	 * @brief this method allocates a protection domain
	 * @return protection domain
	 */
	virtual std::shared_ptr<iprotection_domain> create_protection_domain() noexcept = 0;

	/**
	 * @brief this method allocates a user access region
	 * @param [in] uar allocation type
	 * @return iuser_access_region::create_ret (@see iuser_access_region.hpp)
	 */
	virtual iuser_access_region::create_ret
	create_user_access_region(iuser_access_region::allocation_type allocation_type) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a mapped user memory
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] access flags
	 * @return create_umem_ret (see interface)
	 */
	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size, size_t page_size,
							  uint32_t access_flags) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a mapped user memory with automatic page size
	 * @param [in] address
	 * @param [in] size
	 * @param [in] access flags
	 * @return create_umem_ret (see interface)
	 */
	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size,
							  uint32_t access_flags) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a mapped user memory using variadic args for type safety.
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] variadic access flags
	 * @return create_umem_ret (see interface)
	 */
	template <typename... Args>
	create_umem_ret safe_create_mapped_user_memory(void *address, size_t size, size_t page_size,
						       Args... flags) noexcept
	{
		int access_flags = memory_access_flags_to_uint(flags...);
		return create_mapped_user_memory(address, size, page_size, access_flags);
	}

	/**
	 * @brief This method creates a list of emulated PF devices per emulation type.
	 * @param [in] emu_type Type of the emulated physical functions.
	 * @return get_devices_ret (see above)
	 */
	virtual get_emulated_devices_ret create_emulated_pf_list(emulation_type emu_type) noexcept = 0;

	/**
	 * @brief This method creates a list of emulated VF devices per PF.
	 * @param [in] pf_vhca_id The emulated PF to query for its VFs on.
	 * @return get_devices_ret (see above)
	 */
	virtual get_emulated_devices_ret create_emulated_vf_list(uint16_t pf_vhca_id, uint16_t num_vfs) noexcept = 0;

	/**
	 * @brief Issues general PRM command.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual doca_error_t send_prm_cmd(const void *in, size_t in_size, void *out, size_t out_size) noexcept = 0;

	/**
	 * @brief Issues general PRM command.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual create_devx_object_ret create_devx_object(const void *in, size_t in_size, void *out,
							  size_t out_size) noexcept = 0;

	/**
	 * @brief This method queries device HCA caps with custom op_mod.
	 * @param [in] op_mod operation mode of the query hca capabilities command
	 * @param [out] out out QUER_HCA_CAP_OUT buffer
	 * @param [in] out_size size of allocated out buffer
	 * @return doca_error_t
	 */
	virtual doca_error_t query_hca_caps_by_opmod(uint16_t op_mod, void *out, size_t out_size) noexcept = 0;

	/**
	 * @brief This method creates a completion events channel
	 * @param [in] is blocking (blocks on get event, supported only in Linux)
	 * @return create_comp_events_channel_ret see icompletion_events_channel
	 */
	virtual create_comp_events_channel_ret create_completion_events_channel(bool is_blocking) noexcept = 0;

	/**
	 * @brief This method creates a completion queue
	 * @param [in] uar User access region.
	 * @param [in] nb_cqes The size of the CQ, indicated by number of entries (CQEs) it can hold.
	 * @param [in] user_data Opaque structure provided by user of completion queue.
	 * @param [in] handler Handler invoked by completion events channel, once CQE event is received. Can be NULL.
	 * 		       See icompletion_events_channel.
	 * @return create_completion_queue_ret see icompletion_queue
	 */
	virtual create_completion_queue_ret create_completion_queue(std::shared_ptr<iuser_access_region> uar,
								    const icompletion_queue::attributes &attr,
								    union doca_data user_data,
								    priv_doca_cqe_event_handler_cb_t handler) noexcept = 0;

	/**
	 * @brief This method creates a channel service
	 * @param [in] attr channel service attributes
	 * @return ichannel_service::create_ret (@see ichannel_service.hpp)
	 */
	virtual ichannel_service::create_ret
	create_channel_service(ichannel_service::attributes const &attr) noexcept = 0;

	/**
	 * @brief This method creates a channel connection
	 * @param [in] attr channel connection attributes
	 * @return ichannel_connection::create_ret (@see ichannel_connection.hpp)
	 */
	virtual ichannel_connection::create_ret
	create_channel_connection(ichannel_connection::attributes const &attr) noexcept = 0;

	idevice_adapter(idevice_adapter const &) = delete;
	idevice_adapter &operator=(idevice_adapter const &) = delete;
	idevice_adapter(idevice_adapter &&) = delete;
	idevice_adapter &operator=(idevice_adapter &&) = delete;

protected:

	/**
	 * Constructor
	 */
	idevice_adapter() noexcept = default;

	/**
	 * @brief This method queries device HCA caps
	 * @param [out] caps
	 * @return doca_error_t
	 */
	virtual doca_error_t query_hca_caps_imp(hca_capabilities &caps, bool other_function, uint16_t function_id) noexcept = 0;

	/**
	 * @brief This method returns the devx adapter used by the device.
	 * Since destroying the adapter may void device abilities, this was not declared as const.
	 * @return idevx_adapter
	 */
	virtual std::shared_ptr<idevx_adapter> get_devx_adapter() noexcept = 0;

	bool m_hca_caps_cached{false};
	struct hca_capabilities m_hca_caps;
};

} /* namespace device_abstraction */

