/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <doca_error.h>
#include <imapped_user_memory.hpp>

namespace device_abstraction {

class idevx_object;

using create_devx_object_ret = std::pair<doca_error_t, std::unique_ptr<idevx_object>>;

/**
 * @brief This interface defines a DEVX adapter that can be used by components in the device abstraction to send
 * commands and create devx objects (@see idevx_object interface below).
 */
class idevx_adapter
{
public:
	/**
	 * Destructor
	 */
	virtual ~idevx_adapter() = default;

	/**
	 * @brief Issues general PRM command.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual doca_error_t send_prm_cmd(const void *in, size_t in_size, void *out, size_t out_size) noexcept = 0;

	/**
	 * @brief Creates a devx object.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual create_devx_object_ret
	create_devx_object(const void *in, size_t in_size, void *out, size_t out_size) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a mapped user memory
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] access flags
	 * @return create_umem_ret (see interface)
	 */
	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size, size_t page_size,
							  uint32_t access_flags) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a mapped user memory with automatic page size
	 * @param [in] address
	 * @param [in] size
	 * @param [in] access flags (memory_access_flags see @memory_access_flags.hpp)
	 * @return create_umem_ret (see interface)
	 */
	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size,
							  uint32_t access_flags) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a mapped user memory using variadic args for type safety.
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] variadic access flags (memory_access_flags see @memory_access_flags.hpp)
	 * @return create_umem_ret (see interface)
	 */
	template <typename... Args>
	create_umem_ret safe_create_mapped_user_memory(void *address, size_t size, size_t page_size,
						       Args... flags) noexcept
	{
		int access_flags = memory_access_flags_to_uint(flags...);

		return create_mapped_user_memory(address, size, page_size, access_flags);
	}

	/**
	 * @brief This method finds recommended event queue number for a given core.
	 * @details The event queue number is used mainly for creation of a CQ. Use this method to find the event queue
	 * based on the core affinity.
	 * @param [in] core The core number where interrupts will be received.
	 * @param [out] eqn The recommended event queue number to use.
	 *
	 * @return doca_error_t
	 */
	virtual doca_error_t query_event_queue_number(uint32_t core, uint32_t &eqn) const noexcept = 0;

protected:
	/**
	 * Constructor
	 */
	idevx_adapter() noexcept = default;

private:
	idevx_adapter(idevx_adapter const &) = delete;
	idevx_adapter &operator=(idevx_adapter const &) = delete;
	idevx_adapter(idevx_adapter &&) = delete;
	idevx_adapter &operator=(idevx_adapter &&) = delete;
};

class idevx_object
{
public:
	/**
	 * Destructor
	 */
	virtual ~idevx_object() = default;

	/**
	 * @bfief This method explicitly destroys the devx object
	 * @return DOCA_SUCCESS or DOCA_ERROR_NOT_PERMITTED
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/**
	 * @brief Modifies the devx object.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual doca_error_t modify(const void *in, size_t in_size, void *out, size_t out_size) noexcept = 0;

	/**
	 * @brief This method retrieves the wrapped devx object
	 * @return wrapped devx object
	 */
	virtual void *get_object() const noexcept = 0;

protected:
	/**
	 * Constructor
	 */
	idevx_object() noexcept = default;
};

} // namespace device_abstraction

