/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <utility>
#include <memory>
#include <string>

#include <doca_error.h>
#include <iprotection_domain.hpp>
#include <icompletion_queue.hpp>
#include <ireceive_memory_pool.hpp>

namespace device_abstraction {

class idynamically_connected_target
{
public:

	struct attributes {
		std::shared_ptr<icompletion_queue> cq;
		std::shared_ptr<ireceive_memory_pool> rmp;
		uint8_t port;
		uint8_t mtu;
		uint8_t my_addr_index;
		uint8_t min_rnr_nak;
		uint64_t dc_access_key;
	};

	/**
	 * @brief This pair is used as the return value for creating a dct
	 * @first: doca_error_t
	 *         DOCA_SUCCESS
	 *         DOCA_ERROR_NO_MEMORY
	 *         DOCA_ERROR_INVALID_VALUE
	 *         DOCA_ERROR_NOT_CONNECTED (devx failure)
	 *         DOCA_ERROR_IO_FAILED (devx failure)
	 *         DOCA_ERROR_DRIVER (devx failure)
	 * @second: shared pointer of the dct.
	 */
	using create_ret = std::pair<doca_error_t, std::shared_ptr<idynamically_connected_target>>;

	/**
	 * Destructor
	 * Note: concrete destructor will attempt to destroy the dct.
	 */
	virtual ~idynamically_connected_target() = default;

	/**
	 * @brief Destroy the dct.
	 * @details Destroys the dct, and returns an error in case an error was encountered.
	 * The destructor can be used instead to destroy without returning an error.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/**
	 * @brief this method returns the dct id
	 * return uint32_t
	 */
	virtual uint32_t get_id() const noexcept = 0;

	idynamically_connected_target(idynamically_connected_target const &) = delete;
	idynamically_connected_target &operator=(idynamically_connected_target const &) = delete;
	idynamically_connected_target(idynamically_connected_target &&) = delete;
	idynamically_connected_target &operator=(idynamically_connected_target &&) = delete;

protected:

	/**
	 * Constructor
	 */
	idynamically_connected_target() noexcept = default;
};

} /* namespace device_abstraction */
