/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <doca_error.h>
#include <doca_types.h>

struct priv_doca_event_handle_ops;

namespace device_abstraction {

/**
 * Class used for monitoring of multiple event handles using the same thread.
 * The doca_event_handle_t is a handle created by the kernel that can be used to receive asynchronous events.
 */
class ievents_mux
{
public:

	/**
	 * Destructor
	 */
	virtual ~ievents_mux() = default;

	/**
	 * @brief Destroys the mux, operation may fail.
	 * Destructor will also call this method, ignoring the return value.
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/**
	 * @brief Register handle to be monitored by selector.
	 * @details Adds handle to list of monitored handles by the selector, such that selector can know when the
	 * handle becomes ready, based on an event that was received.
	 * @param [in] handle
	 * Represents a file descriptor in linux, or a file handle in Windows.
	 * @param [in] ops
	 * operations associated with the handle.
	 * @return doca_error_t
	 */
	virtual doca_error_t register_handle(doca_event_handle_t handle, const struct priv_doca_event_handle_ops& ops,
					     union doca_data ops_data) noexcept = 0;

	/**
	 * @brief Unregister handle from the selector.
	 * @details Removes handle from list of monitored handles by the selector, such that selector will no longer
	 * receive events related to this handle.
	 * @param [in] handle
	 * Represents a file descriptor in linux, or a file handle in Windows.
	 * @return doca_error_t
	 */
	virtual doca_error_t unregister_handle(doca_event_handle_t handle) noexcept = 0;

	virtual doca_error_t arm() noexcept = 0;

	virtual doca_error_t clear(doca_event_handle_t handle) noexcept = 0;

	/**
	 * @brief Manually trigger an event, so that sleeping thread immediately wakes up.
	 */
	virtual doca_error_t trigger_event() noexcept = 0;

	/**
	 * @brief Untrigger the maunually triggered event.
	 */
	virtual doca_error_t untrigger_event() noexcept = 0;

	/**
	 * @brief Gets an event handle used for waiting for events.
	 * @return doca_error_t
	 */
	virtual doca_event_handle_t get_handle() const noexcept = 0;

	ievents_mux(ievents_mux const &) = delete;
	ievents_mux &operator=(ievents_mux const &) = delete;
	ievents_mux(ievents_mux &&) = delete;
	ievents_mux &operator=(ievents_mux &&) = delete;

protected:
	/**
	 * Constructor
	 */
	ievents_mux() noexcept = default;
};

} // namespace device_abstraction
