/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <cstdlib>
#include <stdint.h>

#include <doca_error.h>

#include <memory_access_flags.hpp>

namespace device_abstraction {

class imapped_user_memory
{
public:

	/**
	 * Destructor
	 */
	virtual ~imapped_user_memory() = default;

	/**
	 * @brief This method deregisters the umem
	 * @return DOCA_SUCCESS or DOCA_ERROR_NOT_PERMITTED
	 */
	virtual doca_error_t dereg() noexcept = 0;

	/**
	 * brief This method returns the umem id
	 * @return uint32_t
	 */
	virtual uint32_t get_id() const noexcept = 0;

	/**
	 * brief This method returns the umem size
	 * return size_t
	 */
	virtual size_t get_size() const noexcept = 0;

	/**
	 * brief This method returns the umem buffer
	 * return void *
	 */
	virtual void *get_buffer() const noexcept = 0;

	/**
	 * brief This method returns the access flags
	 * @return umem_access
	 */
	virtual uint32_t get_access_flags() const noexcept= 0;

	/**
	 * @brief This method returns the page size set for this mapped memory
	 * @return size_t
	 */
	virtual size_t get_page_size() const noexcept = 0;

	imapped_user_memory(imapped_user_memory const &) = delete;
	imapped_user_memory &operator=(imapped_user_memory const &) = delete;
	imapped_user_memory(imapped_user_memory &&) = delete;
	imapped_user_memory &operator=(imapped_user_memory &&) = delete;

protected:
	/**
	 * Constructor
	 */
	imapped_user_memory() = default;
};

/**
 * @brief This pair is used as the return value for creating a user mapped memory
 * @first: doca_error_t
 *	   DOCA_SUCCESS
 *         DOCA_ERROR_NO_MEMORY
 *         DOCA_ERROR_DRIVER (devx failure)
 * @second shared pointer of the user mapped memory.
 */
using create_umem_ret = std::pair<doca_error_t, std::shared_ptr<imapped_user_memory>>;

}  /* namespace device_abstraction */

