/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>
#include <array>

#include <doca_error.h>

namespace device_abstraction {

// This file should not include prm, therefore it redefines the size (@see ALLOW_OTHER_VHCA_ACCESS_KEY_SIZE)
static const size_t allow_other_vhca_access_key_size{32};
using allow_other_vhca_access_key_t = std::array<uint8_t, allow_other_vhca_access_key_size>;

class imemory_region
{
public:

	/**
	 * Destructor
	 */
	virtual ~imemory_region() = default;

	/**
	 * This method returns the mkey id
	 * @return uint32_t
	 */
	virtual uint32_t get_id() const noexcept= 0;

	/**
	 * @brief This method explicitly destroys the key. Destruction may fail if the mkey is still used.
	 * @return doca_error
	 */
	virtual doca_error_t destroy_key() noexcept = 0;

	/**
	 * @brief This method allows other VHCA access
	 * @param [in] access key
	 * @return doca_error
	 */
	virtual doca_error_t allow_other_vhca_access(const allow_other_vhca_access_key_t &access_key) noexcept = 0;

	imemory_region(imemory_region const &) = delete;
	imemory_region &operator=(imemory_region const &) = delete;
	imemory_region(imemory_region &&) = delete;
	imemory_region &operator=(imemory_region &&) = delete;

protected:
	/**
	 * Constructor
	 */
	imemory_region() noexcept = default;
};

/**
 * @brief This pair is used as the return value for creating a memory region
 * @first: doca_error_t
 *	   DOCA_SUCCESS
 *         DOCA_ERROR_NO_MEMORY
 *         DOCA_ERROR_DRIVER (devx failure)
 * @second shared pointer of the memory region.
 */
using create_mr_ret = std::pair<doca_error_t, std::shared_ptr<imemory_region>>;

} // namespace device_abstraction

