/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>
#include <vector>

#include <doca_error.h>
#include <doca_log.h>
#include <multi_os_prm.h>

#include <iprotection_domain.hpp>
#include <idevx_adapter.hpp>

#include "indirect_memory_region.hpp"

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using namespace device_abstraction;

namespace {

void
fill_klm_mkey_create_in(uint32_t *in, uint32_t pdn, uint32_t num_klm)
{
	MLX5_SET(create_mkey_in, in, opcode, MLX5_CMD_OP_CREATE_MKEY);
	MLX5_SET(create_mkey_in, in, translations_octword_actual_size, num_klm);

	void *mkc = MLX5_ADDR_OF(create_mkey_in, in, memory_key_mkey_entry);
	MLX5_SET(mkc, mkc, access_mode_1_0, MLX5_MKC_ACCESS_MODE_KLMS);
	MLX5_SET(mkc, mkc, lw, 0x1);
	MLX5_SET(mkc, mkc, lr, 0x1);
	MLX5_SET(mkc, mkc, qpn, 0xffffff);
	MLX5_SET(mkc, mkc, pd, pdn);
	MLX5_SET(mkc, mkc, umr_en, 1);
	MLX5_SET(mkc, mkc, translations_octword_size_crossing_target_mkey, num_klm);
}

} /* namespace */

create_mr_ret
indirect_memory_region::create(std::shared_ptr<iprotection_domain> pd, std::shared_ptr<idevx_adapter> adapter,
			       uint32_t num_klm) noexcept
{

	try {
		auto mr_obj = std::make_shared<indirect_memory_region>();
		return {mr_obj->initialize(pd, adapter, num_klm), mr_obj};
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create indirect memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<imemory_region>(nullptr)};
	}
}

indirect_memory_region::~indirect_memory_region()
{
	static_cast<void>(destroy_key());
}

doca_error_t
indirect_memory_region::initialize(std::shared_ptr<iprotection_domain> pd, std::shared_ptr<idevx_adapter> adapter,
				   uint32_t num_klm) noexcept
{
	uint32_t out[MLX5_ST_SZ_DW(create_mkey_out)] = {0};
	const size_t in_size = MLX5_ST_SZ_DW(create_mkey_in) + (num_klm * MLX5_ST_SZ_DW(klm));

	try {
		std::vector<uint32_t> in(in_size, 0);
		fill_klm_mkey_create_in(&in[0], pd->get_pd_id(), num_klm);
		auto ret_obj = adapter->create_devx_object(&in[0], in_size * sizeof(uint32_t), out, sizeof(out));
		if (ret_obj.first != DOCA_SUCCESS)
			return ret_obj.first;

		m_mkey_obj = std::move(ret_obj.second);

		m_id = MLX5_GET(create_mkey_out, out, mkey_index);
		m_id <<= 8;
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate memory for DevX command input buffer. num_klm=%u", num_klm);
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

doca_error_t
indirect_memory_region::destroy_key() noexcept
{
	return m_mkey_obj->destroy();
}

doca_error_t
indirect_memory_region::allow_other_vhca_access(const allow_other_vhca_access_key_t &access_key) noexcept
{
	DOCA_DLOG_ERR("Allow other vhca access, not supported for indirect memory region (mkey).");
	return DOCA_ERROR_NOT_SUPPORTED;
}

uint32_t
indirect_memory_region::get_id() const noexcept
{
	return m_id;
}
