/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <doca_error.h>

#include <imemory_region.hpp>
#include <idevx_adapter.hpp>

namespace device_abstraction {

class iprotection_domain;

class indirect_memory_region : public imemory_region
{
public:

	/**
	 * @brief - This method allocates and initializes an empty indirect memory region.
	 * @param [in] pd Protection domain
	 * @param [in] adapter The DevX adapter
	 * @param [in] num_klm Maximum number of data segments making up the key
	 * @return create_mr_ret (see interface)
	 */
	static create_mr_ret
	create(std::shared_ptr<iprotection_domain> pd, std::shared_ptr<idevx_adapter> adapter, uint32_t num_klm) noexcept;

	/**
	 * Constructor
	 */
	indirect_memory_region() noexcept = default;

	/**
	 * Destructor
	 */
	virtual ~indirect_memory_region();

	/**
	 * @brief - this method initializes the region
	 * @param [in] pd protection domain
	 * @param [in] adapter The DevX adapter
	 * @param [in] num_klm Maximum number of data segments making up the key
	 * @return doca_error_t
	 */
	doca_error_t initialize(std::shared_ptr<iprotection_domain> pd, std::shared_ptr<idevx_adapter> adapter,
				uint32_t num_klm) noexcept;

	virtual doca_error_t destroy_key() noexcept override;

	virtual uint32_t get_id() const noexcept override;

	virtual doca_error_t allow_other_vhca_access(const allow_other_vhca_access_key_t &access_key) noexcept override;

private:

	std::unique_ptr<idevx_object> m_mkey_obj{nullptr};
	uint32_t m_id{0};

	indirect_memory_region(indirect_memory_region const &) = delete;
	indirect_memory_region &operator=(indirect_memory_region const &) = delete;
	indirect_memory_region(indirect_memory_region &&) = delete;
	indirect_memory_region &operator=(indirect_memory_region &&) = delete;
};

} /* namespace device_abstraction */
