/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include "imemory_region.hpp"
#include "ireceive_memory_pool.hpp"
#include "iuser_memory_registration.hpp"
#include "idynamically_connected_target.hpp"

struct ibv_pd;

namespace device_abstraction {

class iprotection_domain
{
public:

	/**
	 * Destructor
	 * Note: concrete destructor will attempt to destroy the protection domain.
	 */
	virtual ~iprotection_domain() = default;

	/**
	 * @brief This method explicitly destroys the protection domain. Protection domain destruction will fail if
	 *  the protection domain is still in use (e.g. a memory region was allocated but not released).
	 *  destroy method facilitates explicit destruction with a return value (that will let the user know if the PD
	 *  was destroyed or not).
	 * @return doca_error_t
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/**
	 * brief this method returns the pd id
	 * return uint32_t
	 */
	virtual uint32_t get_pd_id() const noexcept= 0;

	/**
	 * @brief This method returns the ibv pd. The method is only supported for Linux.
	 *  Note: deallocating the ibv pd will result in unexpected behavior.
	 * @return ibv_pd
	 */
	virtual struct ibv_pd *get_ibv_pd() const noexcept= 0;

	/**
	 * @brief - This method allocates and initializes a mapped memory region
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] access flags
	 * @return create_mr_ret (see imemory_region)
	 */
	virtual create_mr_ret
	create_mapped_memory_region(void *address, size_t size, size_t page_size, uint32_t access_flags) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a mapped memory region
	 *           using variadic args for type safety.
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] variadic access flags
	 * @return create_mr_ret (see imemory_region)
	 */
	template <typename... Args>
	create_mr_ret
	safe_create_mapped_memory_region(void *address, size_t size, size_t page_size, Args... flags) noexcept
	{
		uint32_t access_flags = memory_access_flags_to_uint(flags...);
		return create_mapped_memory_region(address, size, page_size, access_flags);
	}

	/**
	 * @brief - This method allocates and initializes a zero based mapped memory region
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] access flags
	 * @return create_mr_ret (see imemory_region)
	 */
	virtual create_mr_ret
	create_zero_based_mapped_memory_region(void *address, size_t size, size_t page_size,
					       uint32_t access_flags) noexcept = 0;

	/**
	 * @brief - This method allocates and initializes a zero based memory region
	 *          using variadic args for type safety.
	 * @param [in] ibv context
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] variadic access flags
	 * @param [out] doca_error_t (used in case of error)
	 * @return create_mr_ret (see imemory_region)
	 */
	template <typename... Args>
	create_mr_ret
	safe_create_zero_based_mapped_memory_region(void *address, size_t size, size_t page_size,
						    Args... flags) noexcept
	{
		uint32_t access_flags = memory_access_flags_to_uint(flags...);
		return create_zero_based_mapped_memory_region(address, size, page_size, access_flags);
	}

	/**
	 * @brief This method creates an alias memory key
	 * @param [in] lkey
	 * @param [in] vhca id
	 * @param [in] access key
	 * @return memory create_mr_ret (see imemory_region)
	 */
	virtual create_mr_ret create_alias_memory_region(uint32_t lkey, uint16_t vhca_id,
							 const allow_other_vhca_access_key_t &access_key) noexcept = 0;

	/**
	 * @brief This method creates an crossing memory key
	 * @param [in] introspection mkey
	 * @param [in] vhca id
	 * @return memory create_mr_ret (see imemory_region)
	 */
	virtual create_mr_ret
	create_crossing_memory_region(uint32_t introspection_mkey, uint16_t vhca_id) noexcept = 0;

	/**
	 * @brief This method creates a receive memory pool
	 * @param [in] wq_size work queue size in entries, must be a power of 2
	 * @return ireceive_memory_pool::create_ret (@see ireceive_memory_pool.hpp)
	 */
	virtual ireceive_memory_pool::create_ret
	create_receive_memory_pool(uint32_t wq_size) noexcept = 0;

	/**
	 * @brief This method creates an empty indirect memory region
	 * @param [in] num_regions Number of maximum regions that make up the key
	 * @return memory create_mr_ret (see imemory_region)
	 */
	virtual create_mr_ret
	create_indirect_memory_region(uint32_t num_regions) noexcept = 0;

	/**
	 * @brief This method creates a UMR instance
	 * @param [in] num_operations Maximum number of UMR operations to be done at same time.
	 * @return memory create_umr_ret (see iuser_memory_registration)
	 */
	virtual create_umr_ret
	create_user_memory_registration(uint32_t num_operations) noexcept = 0;

	/**
	 * @brief This method creates a dynamically connected target
	 * @param [in] attr dct attributes
	 * @return idynamically_connected_target::create_ret (@see idynamically_connected_target.hpp)
	 */
	virtual idynamically_connected_target::create_ret
	create_dynamically_connected_target(idynamically_connected_target::attributes const &attr) noexcept = 0;

	iprotection_domain(iprotection_domain const &) = delete;
	iprotection_domain &operator=(iprotection_domain const &) = delete;
	iprotection_domain(iprotection_domain &&) = delete;
	iprotection_domain &operator=(iprotection_domain &&) = delete;

protected:
	/**
	 * Constructor
	 */
	iprotection_domain() noexcept = default;
};

} /* namespace device_abstraction */

