/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <utility>
#include <memory>

#include <doca_error.h>

namespace device_abstraction {

class ireceive_memory_pool
{
public:

	/**
	 * @brief This pair is used as the return value for creating a receive memory pool
	 * @first: doca_error_t
	 *         DOCA_SUCCESS
	 *         DOCA_ERROR_NO_MEMORY
	 *         DOCA_ERROR_INVALID_VALUE
	 *         DOCA_ERROR_NOT_CONNECTED (devx failure)
	 *         DOCA_ERROR_DRIVER (devx failure)
	 * @second: shared pointer of the receive memory pool.
	 */
	using create_ret = std::pair<doca_error_t, std::shared_ptr<ireceive_memory_pool>>;

	/**
	 * Destructor
	 */
	virtual ~ireceive_memory_pool() noexcept = default;

	/**
	 * @brief This method explicitly destroys the rmp.
	 * Destruction may fail if the rmp is still in use.
	 * @return doca_error
	 */
	virtual doca_error_t destroy() noexcept = 0;

	/**
	 * @brief This method returns the rmp id
	 * @return uint32_t
	 */
	virtual uint32_t get_id() const noexcept = 0;

	/**
	 * @brief This method returns the rmp wq size in entries
	 * @return uint32_t
	 */
	virtual uint32_t get_wq_size() const noexcept = 0;

	/**
	 * @brief This method returns the rmp wq buffer
	 * @return void *
	 */
	virtual void *get_wq_buffer() const noexcept = 0;

	/**
	 * @brief This method returns the rmp wq doorbell
	 * @return uint32_t *
	 */
	virtual uint32_t *get_wq_doorbell() const noexcept = 0;

	ireceive_memory_pool(ireceive_memory_pool const &) = delete;
	ireceive_memory_pool &operator=(ireceive_memory_pool const &) = delete;
	ireceive_memory_pool(ireceive_memory_pool &&) = delete;
	ireceive_memory_pool &operator=(ireceive_memory_pool &&) = delete;

protected:
	/**
	 * Constructor
	 */
	ireceive_memory_pool() noexcept = default;
};

}  /* namespace device_abstraction */

