/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <doca_error.h>

namespace device_abstraction {

class iuser_memory_registration
{
public:

	struct mkey {
		uint32_t byte_count;	/* Big Endian */
		uint32_t id;		/* Big Endian */
		uint64_t address;	/* Big Endian */
	};

	using mkey_array = struct mkey*;

	/**
	 * Destructor
	 * Note: concrete destructor will attempt to destroy the user access region.
	 */
	virtual ~iuser_memory_registration() = default;

	/**
	 * @brief Start UMR wqe fill operation
	 * @details returns a pointer to the mkey array, user expected to fill the array and then call
	 * umr_done with same array and number of filled elements.
	 *
	 * @param [out] max_num_mkeys size of the mkey array.
	 *
	 * @return pointer to mkey_array user expected to fill the array and then call
	 * umr_done with same array and number of filled elements.
	 * Or nullptr in case no more arrays are free.
	 */
	virtual mkey_array *umr_start(uint32_t *max_num_mkeys) noexcept = 0;

	/**
	 * @brief This method fills the UMR WQE taking space of 3 WQEs.
	 * @details Used on submission the flow is as follows:
	 * On request submit (post to QP):
	 * 1. call umr_start.
	 * 2. fill mkey_array with desired buffers.
	 * 3. call umr_commit
	 * 4. use returned umr_mkey as source data segment of the wqe. Make sure to apply small fence.
	 * On request done (CQE received):
	 * 5. call umr_done with same mkey array.
	 *
	 * @param [in] mkeys array holding mkeys received from umr_start().
	 * @param [in] num_mkeys number of mkeys filled must be <= max_num_mkeys from umr_start().
	 * @param [in] total_len sum of all mkey lengths.
	 * @param [in] wqe start address of the WQE where the 3 UMR WQEs will be filled.
	 * @param [in] pi the current producer index of the QP. Needed for filling the control segment of the WQE.
	 * @param [in] qp_num the QP number. Needed for filling the control segment of the WQE.
	 * @param [out] umr_mkey single mkey representing a gather list of all provided mkeys in mkey_array.
	 * Can be used as source mkey in the next WQE, which must apply a small fence to insure validity.
	 * The provided pointer can be the address of the source data segment of the next wqe.
	 */
	virtual void umr_commit(mkey_array *mkeys, uint32_t num_mkeys, uint32_t total_len, void *wqe,
				uint16_t pi, uint32_t qp_num, struct mkey *umr_mkey) noexcept = 0;

	/**
	 * @brief This method returns the mkey array to be used for further operations.
	 * @details Method allows the mkey array to be reused in an upcoming UMR operation,
	 * this method is expected to be called, after a CQE has been received signifying that operation
	 * using the UMR mkey is done.
	 * @param [in] mkeys array received from umr_start().
	 */
	virtual void umr_done(mkey_array *mkeys) noexcept = 0;

	/**
	 * @brief Destroy the UMR instance.
	 * @return doca_error_t
	 * DOCA_SUCCESS - UMR successfully destroyed.
	 * DOCA_ERROR_IN_USE - umr_start called without calling umr_done.
	 * DOCA_ERROR_DRIVER - Failed to destroy mkey, this usually means it is still in use.
	 */
	virtual doca_error_t destroy() noexcept = 0;

	iuser_memory_registration(iuser_memory_registration const &) = delete;
	iuser_memory_registration &operator=(iuser_memory_registration const &) = delete;
	iuser_memory_registration(iuser_memory_registration &&) = delete;
	iuser_memory_registration &operator=(iuser_memory_registration &&) = delete;

protected:

	/**
	 * Constructor
	 */
	iuser_memory_registration() noexcept = default;
};

/**
 * @brief This pair is used as the return value for creating a user memory registration
 * @first: doca_error_t
 *	DOCA_SUCCESS
 *	DOCA_ERROR_NO_MEMORY
 *	DOCA_ERROR_DRIVER (devx failure)
 * @second: shared pointer of the user access region.
 */
using create_umr_ret = std::pair<doca_error_t, std::unique_ptr<iuser_memory_registration>>;

} /* namespace device_abstraction */

