/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <cerrno>
#include <functional>
#include <algorithm>
#include <fcntl.h>

#include <multi_os_prm.h>

#include <doca_log.h>

#include <idevx_adapter.hpp>
#include <icompletion_queue.hpp>
#include "linux_completion_events_channel.hpp"

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using namespace device_abstraction;

static_assert(sizeof(struct mlx5dv_devx_async_event_hdr) == sizeof(uint64_t),
	      "mlx5dv_devx_async_event_hdr struct size must be 64 bits aligned");

std::vector<linux_completion_events_channel::cookie*>::iterator
linux_completion_events_channel::find_cq(std::shared_ptr<icompletion_queue> cq)
{
	return std::find_if(m_registered_cqs.begin(), m_registered_cqs.end(),
		[cq](const linux_completion_events_channel::cookie *cookie) {
			return cq == cookie->cq;
		});
}

linux_completion_events_channel::~linux_completion_events_channel()
{
	for (auto cookie : m_registered_cqs)
		delete cookie;

	if (m_channel) {
		mlx5dv_devx_destroy_event_channel(m_channel);
		m_channel = nullptr;
	}
}

doca_error_t
linux_completion_events_channel::register_cq(std::shared_ptr<icompletion_queue> cq) noexcept
{
	if (find_cq(cq) != m_registered_cqs.end()) {
		DOCA_DLOG_ERR("Failed to register CQ. CQ already registered");
		return DOCA_ERROR_IN_USE;
	}

	auto cookie = new (std::nothrow) linux_completion_events_channel::cookie({cq.get(), false, cq});
	if (cookie == nullptr) {
		DOCA_DLOG_ERR("Failed to register CQ. Out of memory");
		return DOCA_ERROR_NO_MEMORY;
	}

	try {
		m_registered_cqs.emplace_back(cookie);
	} catch (...) {
		delete cookie;
		DOCA_DLOG_ERR("Failed to register CQ. Out of memory");
		return DOCA_ERROR_NO_MEMORY;
	}

	auto cq_obj = static_cast<struct mlx5dv_devx_obj *>(cq->get_devx_obj()->get_object());
	static uint16_t events_num[] = {MLX5_EVENT_TYPE_COMPLETION_EVENTS};
	auto ret = mlx5dv_devx_subscribe_devx_event(m_channel, cq_obj, sizeof(events_num), events_num,
						    reinterpret_cast<uint64_t>(cookie));
	if (ret != 0) {
		m_registered_cqs.pop_back();
		delete cookie;
		DOCA_DLOG_ERR("Failed to register CQ. mlx5dv_devx_subscribe_devx_event() failed with error %d", ret);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	return DOCA_SUCCESS;
}

doca_error_t
linux_completion_events_channel::unregister_cq(std::shared_ptr<icompletion_queue> cq) noexcept
{
	auto it = find_cq(cq);
	if (it == m_registered_cqs.end()) {
		DOCA_DLOG_ERR("Failed to unregister CQ. CQ already unregistered.");
		return DOCA_ERROR_NOT_FOUND;
	}

	auto cookie = *it;

	/* Only way to unsubscribe from mlx5dv_devx_event_channel is to destroy the CQ. */
	m_registered_cqs.erase(it);
	delete cookie;

	return DOCA_SUCCESS;
}

void
linux_completion_events_channel::arm() noexcept
{
	for (auto cq : m_registered_cqs) {
		if (cq->is_armed)
			continue;
		cq->is_armed = true;
		cq->fast_cq->arm();
	}
}

void
linux_completion_events_channel::force_arm() noexcept
{
	if(!m_is_blocking) {
		/*cleaning file descriptor before arming. */
		struct mlx5dv_devx_async_event_hdr event_header;
		mlx5dv_devx_get_event(m_channel, &event_header, sizeof(event_header));
	}

	for (auto cookie : m_registered_cqs) {
		cookie->is_armed = true;
		cookie->fast_cq->force_arm();
	}
}


doca_error_t
linux_completion_events_channel::handle_events(doca_event_handle_t) noexcept
{
	struct mlx5dv_devx_async_event_hdr event_header;
	if (mlx5dv_devx_get_event(m_channel, &event_header, sizeof(event_header)) != sizeof(event_header)) {

		if (m_is_blocking) {
			DOCA_DLOG_ERR(
				"Failed to handle events. failed to get event using mlx5dv_devx_get_event() errno=%d",
				errno);
			return DOCA_ERROR_OPERATING_SYSTEM;
		}
		else
			return DOCA_ERROR_AGAIN; /* Non blocking error implies that there is no event. */
	}

	linux_completion_events_channel::cookie &cq = *reinterpret_cast<linux_completion_events_channel::cookie*>(
								event_header.cookie);

	cq.is_armed = false;
	cq.fast_cq->handle_event();

	return DOCA_SUCCESS;
}

doca_event_handle_t
linux_completion_events_channel::get_handle() const noexcept
{
	return static_cast<doca_event_handle_t>(m_channel->fd);
}

doca_error_t
linux_completion_events_channel::initialize(struct ibv_context *ctx, bool is_blocking) noexcept
{
	if (ctx == nullptr) {
		DOCA_DLOG_ERR("Failed to initialize completion events channel. Received ibv_ctx=NULL");
		return DOCA_ERROR_UNEXPECTED;
	}
	m_channel = mlx5dv_devx_create_event_channel(ctx, MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA);
	if (m_channel == NULL) {
		DOCA_DLOG_ERR("Failed to initialize completion events channel.");
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	m_is_blocking = is_blocking;

	if (!is_blocking) {
		/* Marking the file descriptor as non block will facilitate non blocking handle_events. */
		int flags = fcntl(m_channel->fd, F_GETFL);
		if (flags == -1) {
			DOCA_DLOG_ERR("Failed to get flags from file descriptor\n");
			return DOCA_ERROR_OPERATING_SYSTEM;
		}

		int rc = fcntl(m_channel->fd, F_SETFL, flags | O_NONBLOCK);
		if (rc < 0) {
			DOCA_DLOG_ERR("Failed to change file descriptor of event channel to non block\n");
			return DOCA_ERROR_OPERATING_SYSTEM;
		}
	}

	return DOCA_SUCCESS;
}

create_comp_events_channel_ret
linux_completion_events_channel::create(struct ibv_context *ctx, bool is_blocking) noexcept
{
	try {
		auto channel = std::make_shared<linux_completion_events_channel>();
		auto ret = channel->initialize(ctx, is_blocking);
		if (ret != DOCA_SUCCESS)
			channel.reset();
		return {ret, channel};
	} catch (...) {
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<icompletion_events_channel>(nullptr)};
	}
}
