/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <dirent.h>
#include <ifaddrs.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <algorithm>
#include <unordered_map>
#include <vector>

#include <infiniband/verbs.h>
#include <infiniband/mlx5dv.h>

#include <mlx5_devx_cmds.h>
#include <doca_log.h>

#include "linux_device_adapter.hpp"
#include "linux_protection_domain.hpp"
#include "linux_user_access_region.hpp"
#include "linux_devx_adapter.hpp"
#include "linux_mapped_user_memory.hpp"
#include "linux_completion_events_channel.hpp"

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

#define LOCAL_MKEY_TO_REMOTE_MKEY_BIT 0x100
#define PRIV_DOCA_MAX_IFACE_NAME 256
#define SF_FIRST_VPORT_VALUE 0x8000

struct host_params_context {
	bool vf_info_valid{false};
	uint16_t total_vf_cnt{0};
	uint16_t vf_cnt{0};
	uint16_t pf_vhca_id{0};
	uint16_t pci_bus{0};
	uint16_t pci_device{0};
	uint16_t pci_function{0};
	uint16_t vf_stride{0};
	uint16_t first_vf_offset{0};
};

using namespace device_abstraction;

// Static free function helpers for extracting BDF elements from the BDF string
namespace {

static void
calc_vf_bdf(doca_pci_bdf pf_bdf, uint16_t first_vf_offset, uint16_t vf_index, doca_pci_bdf &vf_bdf)
{
	vf_bdf.bus = pf_bdf.bus;
	vf_bdf.device = (pf_bdf.device * PCI_FUNCTION_MAX_VALUE + pf_bdf.function +
		first_vf_offset + vf_index) / PCI_FUNCTION_MAX_VALUE;
	vf_bdf.function = (pf_bdf.function + first_vf_offset + vf_index) % PCI_FUNCTION_MAX_VALUE;
}

/* Converts an emulation type into its matching query emulated functions command opmod. */
static inline enum mlx5_emulated_functions_op_mod emu_type_to_opmod(emulation_type type)
{
	switch (type) {
	case emulation_type::VIRTIO_FS:
		return MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_VIRTIO_FS_PHYSICAL_FUNCTIONS;
	case emulation_type::VIRTIO_NET:
		return MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_VIRTIO_NET_PHYSICAL_FUNCTIONS;
	case emulation_type::VIRTIO_BLK:
		return MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_VIRTIO_BLK_PHYSICAL_FUNCTIONS;
	case emulation_type::NVME:
		return MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_NVME_PHYSICAL_FUNCTIONS;
	default:
		return MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_INVALID;
	}
}

static size_t
find_substring_pos_from_end(const std::string &source, size_t end_pos, const std::string &delimiter)
{
	auto end_it = source.begin() + end_pos;
	auto found_it = find_end(source.begin(), end_it, delimiter.begin(), delimiter.end());

	if (found_it == end_it) {
		throw std::logic_error("failed to find bdf element");
	}

	return (found_it - source.begin() + 1);
}

static uint8_t
extract_bdf_element(const std::string &pci_address_out, size_t &end_pos, const std::string &delimiter)
{
	auto element_pos = find_substring_pos_from_end(pci_address_out, end_pos, delimiter);
	auto element_str = pci_address_out.substr(element_pos, end_pos - element_pos);
	end_pos = (element_pos - 1);
	return static_cast<uint8_t>(stoul(element_str, nullptr, 16));
}

static inline doca_error_t
sf_pci_get(struct ibv_device *ibv_dev, std::string &pci_out)
{
	char pci_addr[IBV_SYSFS_PATH_MAX];
	int len;
	char *tmp_address, *tmp_save;
	char *pci_addr_found = NULL;
	const char *sf_name = pci_out.c_str();

	len = readlink(ibv_dev->ibdev_path, pci_addr, sizeof(pci_addr) - 1);
	if (len < 0) {
		DOCA_DLOG_ERR("readlink() failed with errno %d", errno);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}
	pci_addr[len] = '\0';

	/* We expect readlink output to look as follows '.../<PCI_ADDRESS>/<SF_NAME>/...' */
	for (tmp_address = strtok_r(pci_addr, "/", &tmp_save); tmp_address != NULL;
		 tmp_address = strtok_r(NULL, "/", &tmp_save)) {
		if (strcmp(tmp_address, sf_name) == 0)
			break;
		pci_addr_found = tmp_address;
	}

	if (pci_addr_found == NULL) {
		DOCA_DLOG_ERR("Unable to find pci address of SF device %s", ibv_get_device_name(ibv_dev));
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	pci_out = pci_addr_found;

	return DOCA_SUCCESS;
}

emulated_devinfo
emulated_fn_info_to_emulated_devinfo(struct emulated_fn_info &fn_info)
{
	return {
		.vhca_id = fn_info.vhca_id,
		.pci_bdf = fn_info.pci_bdf,
		.nb_total_vfs = fn_info.nb_total_vfs,
		.hotplug = fn_info.hotplug,
		.vf_exist = fn_info.vf_exist,
	};
}

} /* namespace */

linux_device_adapter::linux_device_adapter(struct ibv_device *handle) throw()
{
	m_ibv_dev = handle;

	m_name = ibv_get_device_name(m_ibv_dev);

	m_devx_adapter = std::make_shared<linux_devx_adapter>(handle);
}

linux_device_adapter::~linux_device_adapter()
{
	close();
}

doca_error_t
linux_device_adapter::open() noexcept
{
	return m_devx_adapter->open();
}

doca_error_t
linux_device_adapter::close() noexcept
{
	return m_devx_adapter->close();
}

bool
linux_device_adapter::operator==(const idevice_adapter &other) const noexcept
{
	// It is safe to assume that the interface is a linux_device_adapter.
	return m_ibv_dev == (static_cast<const linux_device_adapter &>(other)).m_ibv_dev;
}

const std::string &
linux_device_adapter::get_name() const noexcept
{
	return m_name;
}

uint32_t
linux_device_adapter::get_vendor_id() const noexcept
{
	return m_vendor_id;
}

uint32_t
linux_device_adapter::get_vendor_part_id() const noexcept
{
	return m_vendor_part_id;
}

struct ibv_context *
linux_device_adapter::get_ibv_context() noexcept
{
	if (open() != DOCA_SUCCESS)
		return nullptr;

	return m_devx_adapter->get_ibv_context();
}

doca_error_t
linux_device_adapter::query_hca_caps_imp(hca_capabilities &caps, bool other_function, uint16_t function_id) noexcept
{
	struct mlx5_hca_cap hca_caps{};

	const auto ret = open();
	if (ret != DOCA_SUCCESS)
		return ret;

	if (mlx5_devx_cmd_query_hca_caps(m_devx_adapter->get_ibv_context(), other_function, function_id, &hca_caps) !=
	    0) {
		DOCA_DLOG_ERR("Failed to query devx HCA capabilities");
		return DOCA_ERROR_DRIVER;
	}

	caps.vhca_id = hca_caps.vhca_id;
	caps.vhca_resource_manager = hca_caps.vhca_resource_manager;
	caps.hca_cap_2 = hca_caps.hca_cap_2;
	caps.emulation_manager_virtio_net = hca_caps.emulation_manager_virtio_net;
	caps.emulation_manager_virtio_blk = hca_caps.emulation_manager_virtio_blk;
	caps.emulation_manager_nvme = hca_caps.emulation_manager_nvme;
	caps.emulation_manager_virtio_fs = hca_caps.emulation_manager_virtio_fs;
	caps.hotplug_manager = hca_caps.hotplug_manager;
	caps.eswitch_manager = hca_caps.eswitch_manager;
	caps.general_object_types_supported = hca_caps.general_object_types_supported;
	caps.ib_port_type = hca_caps.ib_port_type;
	caps.crossing_vhca_mkey_supported = hca_caps.crossing_vhca_mkey_supported;

	if (caps.hca_cap_2) {
		struct mlx5_hca_cap_2 hca_caps2 {};
		if (mlx5_devx_cmd_query_hca_caps_2(m_devx_adapter->get_ibv_context(), other_function, function_id,
						   &hca_caps2) != 0) {
			DOCA_DLOG_ERR("Failed to query devx HCA capabilities 2");
			return DOCA_ERROR_DRIVER;
		}

		caps.cross_gvmi_mkey_enabled =
			(hca_caps2.cross_vhca_object_to_object_supported & LOCAL_MKEY_TO_REMOTE_MKEY_BIT) != 0;
		caps.introspection_mkey_access_allowed = hca_caps2.introspection_mkey_access_allowed;
		caps.introspection_mkey = hca_caps2.introspection_mkey;
	}
	return DOCA_SUCCESS;
}

doca_error_t
linux_device_adapter::query_pf_vfs_vuids(uint16_t vhca_id, std::vector<vuid_t> &vuids) noexcept
{
	auto status = open();
	if (status != DOCA_SUCCESS)
		return status;

	try {
		uint16_t nb_vuids = static_cast<uint16_t>(vuids.size());
		std::vector<mlx5_vuid> mlx_vuids(nb_vuids);

		status = mlx5_devx_cmd_query_vuid(m_devx_adapter->get_ibv_context(), vhca_id, mlx_vuids.data(),
						  &nb_vuids);

		if (status == DOCA_SUCCESS) {
			vuids.resize(nb_vuids);
			for (uint32_t i = 0; i < nb_vuids; i++) {
				memcpy(vuids[i].id, mlx_vuids[i].id, vuids[i].vuid_bytes);
			}
		}

		return status;

	} catch (...) {
		return DOCA_ERROR_NO_MEMORY;
	}
}

doca_error_t
linux_device_adapter::query_vuid(uint16_t vhca_id, vuid_t &vuid) noexcept
{
	mlx5_vuid mlx_vuid{};
	uint16_t nb_vuids{1};

	auto status = open();
	if (status != DOCA_SUCCESS)
		return status;

	status = mlx5_devx_cmd_query_vuid(m_devx_adapter->get_ibv_context(), vhca_id, &mlx_vuid, &nb_vuids);

	if (status == DOCA_SUCCESS) {
		memcpy(&vuid.id, mlx_vuid.id, MLX5_VUID_BYTES);
	}

	return status;
}

doca_error_t
linux_device_adapter::query_physical_ports_count(uint8_t &port_count) noexcept
{
	struct ibv_device_attr_ex device_attr{};

	const auto ret = open();
	if (ret != DOCA_SUCCESS)
		return ret;

	if (ibv_query_device_ex(m_devx_adapter->get_ibv_context(), nullptr, &device_attr) != 0) {
		DOCA_DLOG_ERR("Failed to ibv_query_device_ex(%s).", m_name.c_str());
		return DOCA_ERROR_DRIVER;
	}

	port_count = device_attr.phys_port_cnt_ex;

	return DOCA_SUCCESS;
}

doca_error_t
linux_device_adapter::query_port_info(int port_id, port_info &info) noexcept
{
	struct mlx5dv_port mlx_port_info{};

       const auto ret = open();
       if (ret != DOCA_SUCCESS)
               return ret;

       if (mlx5dv_query_port(m_devx_adapter->get_ibv_context(), port_id, &mlx_port_info) != 0) {
		return DOCA_ERROR_DRIVER;
	}

	//vport & vhca_id need to be valid
	if ((mlx_port_info.flags & (MLX5DV_QUERY_PORT_VPORT_VHCA_ID | MLX5DV_QUERY_PORT_VPORT)) ==
			(MLX5DV_QUERY_PORT_VPORT_VHCA_ID | MLX5DV_QUERY_PORT_VPORT)) {
		info.vhca_id_valid = true;
		info.vhca_id = mlx_port_info.vport_vhca_id;
		info.vport = mlx_port_info.vport;
	} else {
		info.vhca_id_valid = false;
	}

	return DOCA_SUCCESS;
}

doca_error_t
linux_device_adapter::query_pci_bdf(uint8_t &bus, uint8_t &device, uint8_t &function) const noexcept
{
	try {
		char pci_addr[IBV_SYSFS_PATH_MAX]{};
		char *pci_addr_out = nullptr;

		// check if interface is real HW
		std::string path = "/sys/class/infiniband/" + m_name + "/device";
		if (path.length() >= IBV_SYSFS_PATH_MAX) {
			DOCA_DLOG_ERR("link path length for pci address, which is %d, is too short. need %zu chars",
					  IBV_SYSFS_PATH_MAX, path.length() + 1);
			return DOCA_ERROR_NOT_SUPPORTED;
		}

		/* Open the command for reading. */
		int len = readlink(path.c_str(), pci_addr, sizeof(pci_addr) - 1);
		if (len < 0) {
			DOCA_DLOG_ERR("readlink() failed with errno %d", errno);
			return DOCA_ERROR_OPERATING_SYSTEM;
		}
		pci_addr[len] = '\0';

		pci_addr_out = strrchr(pci_addr, '/');
		if (pci_addr_out == NULL) {
			DOCA_DLOG_ERR("Unable to find pci address of device %s", m_name.c_str());
			return DOCA_ERROR_OPERATING_SYSTEM;
		}

		std::string pci_address_out = pci_addr_out + 1;
		if (strstr(pci_address_out.c_str(), "sf") != NULL)
		{
			auto status = sf_pci_get(m_ibv_dev, pci_address_out);
			if (status != DOCA_SUCCESS)
				return status;
		}

		size_t end_pos = pci_address_out.length();
		function = extract_bdf_element(pci_address_out, end_pos, ".");
		device = extract_bdf_element(pci_address_out, end_pos, ":");
		bus = extract_bdf_element(pci_address_out, end_pos, ":");

		return DOCA_SUCCESS;
	} catch (...) {
		DOCA_DLOG_ERR("query_pci_bdf failed with exception");
		return DOCA_ERROR_OPERATING_SYSTEM;
	}
}

static inline bool
priv_iface_dir_not_representor(const char *ib_path, const struct dirent *dir_entry)
{
	/* Net representors don't have the phy_stats directory, while PFs do.*/
	int n, ret;
	char phys_path[IBV_SYSFS_PATH_MAX] = {0};
	const char *path_to_check = "%s/%s/phy_stats";

	/* check if interface is real HW */
	n = snprintf(phys_path, sizeof(phys_path), path_to_check, ib_path, dir_entry->d_name);
	if ((n > 0) && (n < (int)sizeof(phys_path))) {
		ret = access(phys_path, F_OK);
		if (ret != 0) {
			DOCA_DLOG_DBG("couldn't access the path %s. error code = %d", phys_path, ret);
			return false;
		}
	}

	return true;
}

static inline bool
is_dot_dirs(const struct dirent *dir_entry)
{
	return (strcmp(dir_entry->d_name, ".") == 0) || (strcmp(dir_entry->d_name, "..") == 0);
}

doca_error_t
priv_doca_devinfo_iface_name_get(const char *device_name, char *ifname) noexcept
{
	int n;
	int dir_count;
	char ib_path[IBV_SYSFS_PATH_MAX] = {0};
	DIR *net_dir;
	struct dirent *dir_entry, *first_dir_found = NULL;
	doca_error_t ret;

	n = snprintf(ib_path, sizeof(ib_path), "/sys/class/infiniband/%s/device/net/", device_name);
	if ((n < 0) || (n >= (int)sizeof(ib_path)))
		return DOCA_ERROR_INVALID_VALUE;

	net_dir = opendir(ib_path);
	if (net_dir == NULL)
		return DOCA_ERROR_OPERATING_SYSTEM;

	/* See if there is more than a single interface in the net directory.*/
	dir_count = 0;
	while (((dir_entry = readdir(net_dir)) != NULL)) {
		if (dir_entry->d_type != DT_DIR || is_dot_dirs(dir_entry))
			continue;

		dir_count++;
		if (dir_count == 1) {
			/* Store first directory, until we find if there is more than 1 directory.*/
			first_dir_found = dir_entry;
			continue;
		}

		/* Only in case there is more than 1 directory we need to make this check.*/
		if (priv_iface_dir_not_representor(ib_path, dir_entry))
			break;
	}

	if (dir_count == 1) {
		/* If only a single entry exists then we take it.*/
		dir_entry = first_dir_found;
	} else if (dir_count > 1) {
		/* If more than 1 entry, then only 1 is a PF while others are representors.*/
		dir_entry = priv_iface_dir_not_representor(ib_path, first_dir_found) ? first_dir_found : dir_entry;
	}

	if (dir_entry == NULL) {
		DOCA_DLOG_ERR("Unable to find network interface name for IB dev %s", device_name);
		ret = DOCA_ERROR_NOT_FOUND;
		goto out;
	}

	memcpy(ifname, dir_entry->d_name, PRIV_DOCA_MAX_IFACE_NAME);

	ret = DOCA_SUCCESS;

out:
	if (closedir(net_dir))
		ret = DOCA_ERROR_OPERATING_SYSTEM;

	return ret;
}

doca_error_t
priv_doca_devinfo_sockaddr_get(const char *device_name, bool is_ipv4, void *value, uint32_t size)
{
	char iface_name[PRIV_DOCA_MAX_IFACE_NAME] = {0};
	struct ifaddrs *ifa = NULL, *ifa_start = NULL;
	doca_error_t ret;

	ret = priv_doca_devinfo_iface_name_get(device_name, iface_name);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to find network interface name for IB dev %s", device_name);
		return ret;
	}

	if (getifaddrs(&ifa)) {
		DOCA_DLOG_ERR("getifaddrs failed errno %d\n", errno);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	ret = DOCA_ERROR_NOT_FOUND;
	ifa_start = ifa;
	for (; ifa != NULL; ifa = ifa->ifa_next) {
		if (ifa->ifa_name == NULL || strcmp(ifa->ifa_name, iface_name) != 0 || ifa->ifa_addr == NULL)
			continue;

		/* Interface with matching name found and has an IP address.*/
		if (is_ipv4 && ifa->ifa_addr->sa_family == AF_INET) {
			memcpy(value, ifa->ifa_addr, sizeof(struct sockaddr_in));
			ret = DOCA_SUCCESS;
			break;
		}

		if (!is_ipv4 && ifa->ifa_addr->sa_family == AF_INET6) {
			memcpy(value, ifa->ifa_addr, sizeof(struct sockaddr_in6));
			ret = DOCA_SUCCESS;
			break;
		}
	}

	if (ifa_start)
		freeifaddrs(ifa_start);

	return ret;
}

doca_error_t
linux_device_adapter::query_ipv4_address(void *value, size_t size) const noexcept
{
	if (size != sizeof(struct in_addr)) {
		DOCA_DLOG_ERR("sockaddr struct size must be equal to buffer length supplied. need %zu value size",
			      sizeof(struct in_addr));
		return DOCA_ERROR_INVALID_VALUE;
	}

	struct sockaddr_in ipv4_addr{};

	auto status = priv_doca_devinfo_sockaddr_get(m_name.c_str(), true, &ipv4_addr, sizeof(ipv4_addr));
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("failed to extract sockaddr_in");
		return status;
	}

	memcpy(value, &(ipv4_addr.sin_addr), size);
	return DOCA_SUCCESS;
}

doca_error_t
linux_device_adapter::query_ipv6_address(void *value, size_t size) const noexcept
{
	if (size < sizeof(struct in6_addr)) {
		DOCA_DLOG_ERR("sockaddr struct size must be equal to buffer length supplied. need %zu value size",
			      sizeof(struct in6_addr));
		return DOCA_ERROR_INVALID_VALUE;
	}

	struct sockaddr_in6 ipv6_addr{};

	auto status = priv_doca_devinfo_sockaddr_get(m_name.c_str(), false, &ipv6_addr, sizeof(ipv6_addr));
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("failed to extract sockaddr_in6");
		return status;
	}

	memcpy(value, &(ipv6_addr.sin6_addr), size);
	return DOCA_SUCCESS;
}

doca_error_t
linux_device_adapter::get_interface_name(std::string &interface_name) const noexcept
{
	char iface_name[PRIV_DOCA_MAX_IFACE_NAME] = {0};
	auto status = priv_doca_devinfo_iface_name_get(m_name.c_str(), iface_name);
	if (status == DOCA_SUCCESS) {
		interface_name = iface_name;
	}

	return status;
}

doca_error_t
linux_device_adapter::get_ibdev_name(std::string &ibdev_name) const noexcept
{
	try {
		ibdev_name = get_name();
		return DOCA_SUCCESS;
	} catch (...) {
		DOCA_DLOG_ERR("Failed to assign ibdev name");
		return DOCA_ERROR_NO_MEMORY;
	}
}

std::shared_ptr<iprotection_domain>
linux_device_adapter::create_protection_domain() noexcept
{
	return linux_protection_domain::create(m_devx_adapter);
}

iuser_access_region::create_ret
linux_device_adapter::create_user_access_region(iuser_access_region::allocation_type allocation_type) noexcept
{
	return linux_user_access_region::create(m_devx_adapter->get_ibv_context(), allocation_type);
}

device_abstraction::create_umem_ret
linux_device_adapter::create_mapped_user_memory(void *address, size_t size, size_t page_size,
						uint32_t access_flags) noexcept
{
	return m_devx_adapter->create_mapped_user_memory(address, size, page_size, access_flags);
}

device_abstraction::create_umem_ret
linux_device_adapter::create_mapped_user_memory(void *address, size_t size, uint32_t access_flags) noexcept
{
	return m_devx_adapter->create_mapped_user_memory(address, size, access_flags);
}

get_emulated_devices_ret
linux_device_adapter::create_emulated_pf_list(emulation_type emu_type) noexcept
{
	/* Value based on maximum number that can be set in NVCONFIG.*/
	static constexpr unsigned int nb_max_emulated_pfs = 256;

	struct emulated_fn_info pfs[nb_max_emulated_pfs]{};
	uint16_t nb_pfs{nb_max_emulated_pfs};
	std::vector<emulated_devinfo> emulated_pf_list{};
	enum mlx5_emulated_functions_op_mod op_mod = emu_type_to_opmod(emu_type);

	if (op_mod == MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_INVALID) {
		DOCA_DLOG_ERR("Received invalid device emulation type %d.", (int)emu_type);
		return get_emulated_devices_ret{DOCA_ERROR_INVALID_VALUE, emulated_pf_list};
	}

	auto ret = open();
	if (ret != DOCA_SUCCESS)
		return get_emulated_devices_ret{ret, emulated_pf_list};

	ret = mlx5_devx_cmd_query_emulated_fns_info(m_devx_adapter->get_ibv_context(), op_mod,
						    /*ignored*/0, pfs, &nb_pfs);
	if (ret != DOCA_SUCCESS)
		return get_emulated_devices_ret{ret, emulated_pf_list};
	try {
		emulated_pf_list.reserve(nb_pfs);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to reserve space for additional representor devices."
			      "Attempted to increase capacity from %zu to %u.",
			      emulated_pf_list.capacity(), nb_pfs);
		return get_emulated_devices_ret{DOCA_ERROR_NO_MEMORY, emulated_pf_list};
	}

	for (int pf_idx = 0; pf_idx < nb_pfs; pf_idx++)
		emulated_pf_list.push_back(emulated_fn_info_to_emulated_devinfo(pfs[pf_idx]));

	return get_emulated_devices_ret{DOCA_SUCCESS, emulated_pf_list};
}

get_emulated_devices_ret
linux_device_adapter::create_emulated_vf_list(uint16_t pf_vhca_id, uint16_t num_vfs) noexcept
{
	struct emulated_fn_info vfs[num_vfs];
	std::vector<emulated_devinfo> emulated_vf_list{};

	memset(vfs, 0, sizeof(struct emulated_fn_info) * num_vfs);

	try {
		emulated_vf_list.reserve(num_vfs);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate space for virtual emulated representor devices");
		return get_emulated_devices_ret{DOCA_ERROR_NO_MEMORY, emulated_vf_list};
	}

	auto ret = open();
	if (ret != DOCA_SUCCESS)
		return get_emulated_devices_ret{ret, emulated_vf_list};

	ret = mlx5_devx_cmd_query_emulated_fns_info(m_devx_adapter->get_ibv_context(),
						    MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_VIRTUAL_FUNCTIONS, pf_vhca_id,
												vfs, &num_vfs);
	if (ret != DOCA_SUCCESS)
		return get_emulated_devices_ret{ret, emulated_vf_list};

	for (int vf_idx = 0; vf_idx < num_vfs; vf_idx++)
		emulated_vf_list.push_back(emulated_fn_info_to_emulated_devinfo(vfs[vf_idx]));

	return get_emulated_devices_ret{DOCA_SUCCESS, emulated_vf_list};
}

doca_error_t
linux_device_adapter::send_prm_cmd(const void *in, size_t in_size, void *out, size_t out_size) noexcept
{
	auto open_ret = open();
	if (open_ret != DOCA_SUCCESS)
		return open_ret;

	return m_devx_adapter->send_prm_cmd(in, in_size, out, out_size);
}

device_abstraction::create_devx_object_ret
linux_device_adapter::create_devx_object(const void *in, size_t in_size, void *out, size_t out_size) noexcept
{
	auto open_ret = open();
	if (open_ret != DOCA_SUCCESS)
		return {open_ret, std::unique_ptr<idevx_object>(nullptr)};

	return m_devx_adapter->create_devx_object(in, in_size, out, out_size);
}

create_comp_events_channel_ret
linux_device_adapter::create_completion_events_channel(bool is_blocking) noexcept
{
	return linux_completion_events_channel::create(m_devx_adapter->get_ibv_context(), is_blocking);
}

std::shared_ptr<idevx_adapter>
linux_device_adapter::get_devx_adapter() noexcept
{
	return m_devx_adapter;
}

static function_info
create_function_info(doca_pci_func_type func_type, uint16_t pf_vhca_id, uint16_t vhca_id, uint16_t bdf,
		uint16_t total_vf_cnt, uint16_t vf_cnt, uint16_t function_id)
{
	return {
		.func_type = func_type,
		.pf_vhca_id = pf_vhca_id,
		.vhca_id = vhca_id,
		.bdf = bdf,
		.total_vf_cnt = total_vf_cnt,
		.vf_cnt = vf_cnt,
		.function_id = function_id,
	};
}

static doca_error_t
query_esw_functions(struct ibv_context *context, host_params_context &host_params)
{
	uint32_t in[MLX5_ST_SZ_DW(query_esw_functions_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(query_esw_functions_out)] = {0};

	/* Set operation code and operation mode of the command. */
	MLX5_SET(query_esw_functions_in, in, opcode, MLX5_CMD_OP_QUERY_ESW_FUNCTIONS);

	/* Send the command to FW using DevX. */
	auto ret = mlx5dv_devx_general_cmd(context, in, sizeof(in), out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to query ESW functions");
		return DOCA_ERROR_IO_FAILED;
	}

	/* Inspect the returned status of the command and print the syndrome in case of error. */
	auto status = MLX5_GET(query_esw_functions_out, out, status);
	if (status) {
		DOCA_DLOG_ERR("Query ESW functions returned with status 0x%x, syndrome = 0x%x",
			       status, MLX5_GET(query_esw_functions_out, out, syndrome));
		return DOCA_ERROR_IO_FAILED;
	}

	/* Extract the host params field from the command output. */
	void *out_host_params = MLX5_ADDR_OF(query_esw_functions_out, out, host_params);

	if (MLX5_GET(host_params_context, out_host_params, host_pf_disabled) == 1) {
		DOCA_DLOG_ERR("Host PF is disabled");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	if (MLX5_GET(host_params_context, out_host_params, host_pf_not_exist) == 1) {
		DOCA_DLOG_ERR("Host PF doesn't exist for this context");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	if (MLX5_GET(host_params_context, out_host_params, host_pf_vhca_id_valid) == 0) {
		DOCA_DLOG_ERR("Invalid PF vhca id");
		return DOCA_ERROR_BAD_STATE;
	}

	/* Translate the host params from PRM format to esw_host_params structure. */
	host_params = (struct host_params_context) {
		.vf_info_valid = (bool)MLX5_GET(host_params_context, out_host_params, host_vf_info_valid),
		.total_vf_cnt = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_total_vfs),
		.vf_cnt = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_num_vfs),
		.pf_vhca_id = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_pf_vhca_id),
		.pci_bus = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_pci_bus),
		.pci_device = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_pci_device),
		.pci_function = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_pci_function),
		.vf_stride = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_pci_vf_stride),
		.first_vf_offset = (uint16_t)MLX5_GET(host_params_context, out_host_params, host_pci_first_vf_offset),
	};

	return DOCA_SUCCESS;
}

get_function_info_ret
linux_device_adapter::get_net_pci_topology() noexcept
{
	host_params_context host_params;
	auto ret = query_esw_functions(m_devx_adapter->get_ibv_context(), host_params);
	if (ret != DOCA_SUCCESS) {
		return get_function_info_ret{ret, {}};
	}

	query_hca_caps_ret hca_caps_ret = query_hca_caps();
	if (hca_caps_ret.first != doca_error::DOCA_SUCCESS)
		return get_function_info_ret{hca_caps_ret.first, {}};

	const auto &hca_caps = hca_caps_ret.second;

	uint8_t total_port_cnt{0};
	if (hca_caps.ib_port_type) {
		/* on IB, the pci functions are the PF and its VFs */
		total_port_cnt = host_params.vf_cnt + 1;
	}
	else {
		 /* on Ethernet, the pci functions are the PF, its VFs and its SFs.
		  * query_physical_ports_count retrieves the number of the accessible ports for Ethernet ports */
		ret = query_physical_ports_count(total_port_cnt);
			if (ret != DOCA_SUCCESS)
				return get_function_info_ret{ret, {}};
	}

	std::vector<function_info> functions_info_list{};
	try {
		functions_info_list.reserve(total_port_cnt);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate space for pci functions info");
		return get_function_info_ret{DOCA_ERROR_NO_MEMORY, {}};
	}

	/* save PF bdf & vhca_id for later use */
	doca_pci_bdf pf_bdf{0};
	pf_bdf.bus = host_params.pci_bus;
	pf_bdf.device = host_params.pci_device;
	pf_bdf.function = host_params.pci_function;
	auto pf_vhca_id = host_params.pf_vhca_id;

	/* add PF information to the list */
	functions_info_list.emplace_back(
		create_function_info(
			doca_pci_func_type::DOCA_PCI_FUNC_PF,  /* func_type */
			host_params.pf_vhca_id,  /* pf_vhca_id */
			host_params.pf_vhca_id,  /* vhca_id */
			pf_bdf.raw,  /* bdf */
			host_params.total_vf_cnt,  /* total_vf_cnt */
			host_params.vf_cnt,  /* vf_cnt */
			0 /* function_id */
		)
	);

	if (hca_caps.ib_port_type) {
		/* for IB ports, handle VFs representors using 'query function hca caps' PRM command
		 * since direct-verbs is not supported on IB */
		if (host_params.vf_info_valid == true && host_params.vf_cnt > 0) {

			for (auto vf_info_index = 0; vf_info_index < host_params.vf_cnt; vf_info_index++) {
				doca_pci_bdf vf_bdf{0};
				vf_bdf.bus = host_params.pci_bus;
				vf_bdf.device = (pf_bdf.device * PCI_FUNCTION_MAX_VALUE + pf_bdf.function +
					host_params.first_vf_offset + vf_info_index) / PCI_FUNCTION_MAX_VALUE;
				vf_bdf.function = (pf_bdf.device * PCI_FUNCTION_MAX_VALUE + pf_bdf.function
					+ host_params.first_vf_offset + vf_info_index) % PCI_FUNCTION_MAX_VALUE;

				hca_capabilities vf_caps;
				auto ret = query_function_hca_caps(vf_info_index + 1, vf_caps);
				if (ret != DOCA_SUCCESS) {
					DOCA_DLOG_ERR("Failed to query vhca id for other function");
					return get_function_info_ret{ret, functions_info_list};
				}

				/* add VF information to the list */
				functions_info_list.emplace_back(
					create_function_info(
						doca_pci_func_type::DOCA_PCI_FUNC_VF,  /* func_type */
						pf_vhca_id,  /* pf_vhca_id */
						vf_caps.vhca_id,  /* vhca_id */
						vf_bdf.raw,  /* bdf */
						0,  /* total_vf_cnt */
						0,  /* vf_cnt */
						vf_info_index + 1  /* function_id */
					)
				);
			}
		}
	} else {
		/* handle SFs & VFs representors for Ethernet ports, using direct-verbs */
		for (int port_idx = 0; port_idx < total_port_cnt; port_idx++) {
			port_info port_info{};

			if (query_port_info(port_idx, port_info) != DOCA_SUCCESS ||
				!port_info.vhca_id_valid ||
				port_info.vport == 0) // vport 0 represent a PF
				continue;

			doca_pci_bdf function_bdf;
			doca_pci_func_type function_type;
			if (port_info.vport >= SF_FIRST_VPORT_VALUE) { // case of SF
				function_type = doca_pci_func_type::DOCA_PCI_FUNC_SF;
				function_bdf.raw = pf_bdf.raw;
			}
			else { // case of VF
				function_type = doca_pci_func_type::DOCA_PCI_FUNC_VF;
				/* the VF index is vport - 1 */
				calc_vf_bdf(pf_bdf, host_params.first_vf_offset, port_info.vport - 1, function_bdf);
			}

			functions_info_list.emplace_back( // add VF/SF information to the list
				create_function_info(
					function_type, // func_type
					pf_vhca_id, // pf_vhca_id
					port_info.vhca_id, // vhca_id
					function_bdf.raw, // pci_bdf
					0, // total_vf_cnt
					0, // vf_cnt
					port_info.vport // function_id
				)
			);
		}
	}

	return get_function_info_ret{DOCA_SUCCESS, functions_info_list};
}

bool
linux_device_adapter::is_net_representors_supported() noexcept
{
	auto ret = open();
	if (ret != DOCA_SUCCESS)
		return false;

	host_params_context host_params{0};
	auto ibv_ctx = m_devx_adapter->get_ibv_context();
	if (ibv_ctx == nullptr) {
		DOCA_DLOG_ERR("Failed to get ibv context");
		return false;
	}

	/* send query_esw_functions prm command to check if the device supports representors */
	ret = query_esw_functions(m_devx_adapter->get_ibv_context(), host_params);
	if (ret != DOCA_SUCCESS) {
		return false;
	}

	return true;
}
