/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <string>
#include <mutex>

#include <device_adapter.hpp>
#include <doca_error.h>


struct ibv_device;
struct ibv_context;

namespace device_abstraction {

class linux_devx_adapter;

class linux_device_adapter : public device_adapter
{
public:

	/**
	 * Constructor
	 * @param [in] ibv device handle
	 */
	linux_device_adapter(struct ibv_device *handle) throw();

	/**
	 * Destructor
	 */
	virtual ~linux_device_adapter();

	virtual doca_error_t open() noexcept override;

	virtual doca_error_t close() noexcept override;

	/**************************************************************************
	 *
	 * Queries
	 *
	 **************************************************************************/

	virtual bool operator==(const idevice_adapter &other) const noexcept override;

	virtual const std::string &get_name() const noexcept override;

	virtual uint32_t get_vendor_id() const noexcept override;

	virtual uint32_t get_vendor_part_id() const noexcept override;

	virtual doca_error_t query_vuid(uint16_t vhca_id, vuid_t &vuid) noexcept override;

	virtual doca_error_t query_pf_vfs_vuids(uint16_t vhca_id, std::vector<vuid_t> &vuids) noexcept override;

	virtual doca_error_t query_physical_ports_count(uint8_t &port_count) noexcept override;

	virtual doca_error_t query_port_info(int port_id, port_info &info) noexcept override;

	virtual doca_error_t query_pci_bdf(uint8_t &bus, uint8_t &device, uint8_t &function) const noexcept override;

	virtual doca_error_t query_ipv4_address(void *value, size_t size) const noexcept override;

	virtual doca_error_t query_ipv6_address(void *value, size_t size) const noexcept override;

	virtual doca_error_t get_interface_name(std::string &interface_name) const noexcept override;

	virtual doca_error_t get_ibdev_name(std::string &ibdev_name) const noexcept override;

	virtual ibv_context *get_ibv_context() noexcept override;

	virtual get_function_info_ret get_net_pci_topology() noexcept override;

	virtual bool is_net_representors_supported() noexcept override;

	/**************************************************************************
	 *
	 * PRM Objects Factory
	 *
	 **************************************************************************/

	virtual std::shared_ptr<iprotection_domain> create_protection_domain() noexcept override;

	virtual iuser_access_region::create_ret
	create_user_access_region(iuser_access_region::allocation_type allocation_type) noexcept override;

	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size, size_t page_size,
							  uint32_t access_flags) noexcept override;

	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size,
							  uint32_t access_flags) noexcept override;

	virtual get_emulated_devices_ret create_emulated_pf_list(emulation_type emu_type) noexcept override;

	virtual get_emulated_devices_ret create_emulated_vf_list(uint16_t pf_vhca_id, uint16_t num_vfs) noexcept override;

	virtual doca_error_t send_prm_cmd(const void *in, size_t in_size, void *out, size_t out_size) noexcept override;

	virtual create_devx_object_ret create_devx_object(const void *in, size_t in_size, void *out,
							  size_t out_size) noexcept override;

	virtual create_comp_events_channel_ret create_completion_events_channel(bool is_blocking) noexcept override;


	linux_device_adapter(linux_device_adapter const &) = delete;
	linux_device_adapter &operator=(linux_device_adapter const &) = delete;
	linux_device_adapter(linux_device_adapter &&) = delete;
	linux_device_adapter &operator=(linux_device_adapter &&) = delete;

protected:

	virtual doca_error_t query_hca_caps_imp(hca_capabilities &caps, bool other_function, uint16_t function_id) noexcept override;

	virtual std::shared_ptr<idevx_adapter> get_devx_adapter() noexcept override;

private:

	std::mutex m_control_lock;
	struct ibv_device *m_ibv_dev{ nullptr };
	uint32_t m_vendor_id{ 0 };
	uint32_t m_vendor_part_id{ 0 };
	std::string m_name;
	std::shared_ptr<linux_devx_adapter> m_devx_adapter{nullptr};
};

} /* namespace device_abstraction */


