/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <infiniband/verbs.h>
#include <infiniband/mlx5dv.h>

#include <doca_log.h>

#include <devices_factory.hpp>

#include "linux_device_adapter.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

class ibv_list_deleter
{
public:
	void operator()(struct ibv_device **ibdev_list) const
	{
		ibv_free_device_list(ibdev_list);
	}
};

get_devices_ret
device_abstraction::get_devices() throw()
{
	std::vector<std::shared_ptr<idevice_adapter>> devices;

	int nb_ibdevs{0};
	auto ibdev_list = ibv_get_device_list(&nb_ibdevs);
	if(ibdev_list == nullptr) {
		DOCA_DLOG_ERR("Failed to get RDMA devices list, ibdev_list null");
		return get_devices_ret{DOCA_ERROR_NOT_FOUND, devices};
	}

	// The container and deleter are declared after checking for null because if ibdev_list are null then
	// ibv_free_device_list should not be called.
	ibv_list_deleter deleter;
	std::unique_ptr<struct ibv_device *[], ibv_list_deleter> ibdev_list_container(ibdev_list, deleter);

	if (nb_ibdevs == 0) {
		// Num devices = 0 is not an error.
		DOCA_DLOG_INFO("nb_ibdevs = 0");
		return get_devices_ret{DOCA_SUCCESS, devices};
	}

	try {
		devices.reserve(nb_ibdevs);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create devices vector (%d devices)", nb_ibdevs);

		// No point to continue if failed to reserve the vector.
		return get_devices_ret{DOCA_ERROR_NO_MEMORY, devices};
	}

	for (int i = 0; i < nb_ibdevs; ++i) {
		try {
			devices.emplace_back(std::make_shared<linux_device_adapter>(ibdev_list[i]));
		}
		catch(...) {
			DOCA_DLOG_ERR("Failed to create device adapter (index %d) or push it to the devices list", i);
		}
	}

	return get_devices_ret{DOCA_SUCCESS, devices};
}
