/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <unistd.h>

#include <doca_log.h>
#include "linux_devx_adapter.hpp"
#include "linux_mapped_user_memory.hpp"
#include <multi_os_prm.h>

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::linux_DEVX_ADAPTER);

#ifdef DOCA_LOGGING_ALLOW_DLOG

/**
 * @brief This method retrieves the syndrom and is used only for logging.
 * Therefore it is wrapped with DOCA_LOGGING_ALLOW_DLOG.
 */
static uint32_t
get_syndrome(const void *out, size_t out_size) noexcept
{
	// Assumption: syndrom is always in the same location for all create devx object commands.
	// Find syndrom: grep -i e5300 /mswg/release/fw/fw-4119/last_release/../etc/syndrome_list*
	if (out_size >= MLX5_ST_SZ_BYTES(general_obj_out_cmd_hdr))
		return MLX5_GET(general_obj_out_cmd_hdr, out, syndrome);

	DOCA_DLOG_ERR("out size (%zu) too small, can't extract syndrom", out_size);
	return 0;
}

#endif /* DOCA_LOGGING_ALLOW_DLOG */

linux_devx_adapter::linux_devx_adapter(struct ibv_device *handle) noexcept
	:m_ibv_device(handle)
{
}

doca_error_t
linux_devx_adapter::open() noexcept
{
	if (!m_ibv_ctx) {
		m_ibv_ctx = ibv_open_device(m_ibv_device);
		if (!m_ibv_ctx) {
			DOCA_DLOG_ERR("Failed to open device");
			return DOCA_ERROR_NOT_CONNECTED;
		}
	}

	return DOCA_SUCCESS;
}

doca_error_t
linux_devx_adapter::close() noexcept
{
	if (m_ibv_ctx) {
		auto status = ibv_close_device(m_ibv_ctx);
		if(status != 0) {
			DOCA_DLOG_ERR("Failed to close device with status %d", status);
			return DOCA_ERROR_NOT_PERMITTED;
		}

		m_ibv_ctx = nullptr;
	}

	return DOCA_SUCCESS;
}

struct ibv_context *
linux_devx_adapter::get_ibv_context() const noexcept
{
	return m_ibv_ctx;
}

doca_error_t
linux_devx_adapter::send_prm_cmd(const void *in, size_t in_size, void *out, size_t out_size) noexcept
{
	if (mlx5dv_devx_general_cmd(m_ibv_ctx, const_cast<void *>(in), in_size, out, out_size) != 0) {
		DOCA_DLOG_ERR("Failed to issue general PRM command with syndrome = 0x%x", get_syndrome(out, out_size));
		return DOCA_ERROR_DRIVER;
	}

	return DOCA_SUCCESS;
}

device_abstraction::create_devx_object_ret
linux_devx_adapter::create_devx_object(const void *in, size_t in_size, void *out, size_t out_size) noexcept
{
	auto devx_object = mlx5dv_devx_obj_create(m_ibv_ctx, const_cast<void *>(in), in_size, out, out_size);
	if (!devx_object) {
		DOCA_DLOG_ERR("Failed to create devx object with syndrome = 0x%x", get_syndrome(out, out_size));
		return {DOCA_ERROR_DRIVER, std::unique_ptr<idevx_object>(nullptr)};
	}

	try {
		return {DOCA_SUCCESS, std::make_unique<linux_devx_object>(devx_object)};
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create linux_devx_object");
		return {DOCA_ERROR_NO_MEMORY, std::unique_ptr<idevx_object>(nullptr)};
	}
}

device_abstraction::create_umem_ret
linux_devx_adapter::create_mapped_user_memory(void *address, size_t size, size_t page_size,
					      uint32_t access_flags) noexcept
{
	return linux_mapped_user_memory::create(m_ibv_ctx, address, size, page_size, access_flags);
}

device_abstraction::create_umem_ret
linux_devx_adapter::create_mapped_user_memory(void *address, size_t size, uint32_t access_flags) noexcept
{
	return linux_mapped_user_memory::create(m_ibv_ctx, address, size, getpagesize(), access_flags);
}

doca_error_t
linux_devx_adapter::query_event_queue_number(uint32_t core, uint32_t &eqn) const noexcept
{
	uint32_t eqn_out{0};
	auto ret = mlx5dv_devx_query_eqn(m_ibv_ctx, core, &eqn_out);
	if (ret != 0) {
		DOCA_DLOG_ERR("Failed to query eqn. mlx5dv_devx_query_eqn failed with errno=%d.", ret);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	eqn = eqn_out;
	return DOCA_SUCCESS;
}

linux_devx_object::linux_devx_object(struct mlx5dv_devx_obj *devx_object) noexcept
	:m_devx_object(devx_object)
{
}

doca_error_t
linux_devx_object::destroy() noexcept
{
	if (m_devx_object) {
		if (mlx5dv_devx_obj_destroy(m_devx_object) != 0) {
			DOCA_DLOG_ERR("Failed to destroy devx object");
			return DOCA_ERROR_NOT_PERMITTED;
		}

		m_devx_object = nullptr;
	}

	return DOCA_SUCCESS;
}

linux_devx_object::~linux_devx_object() noexcept
{
	static_cast<void>(destroy());
}

doca_error_t
linux_devx_object::modify(const void *in, size_t in_size, void *out, size_t out_size) noexcept
{
	if (mlx5dv_devx_obj_modify(m_devx_object, in, in_size, out, out_size) != 0) {
		DOCA_DLOG_ERR("Failed to modify devx object");
		return DOCA_ERROR_DRIVER;
	}

	return DOCA_SUCCESS;
}

void *
linux_devx_object::get_object() const noexcept
{
	return m_devx_object;
}
