/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <doca_error.h>

#include <idevx_adapter.hpp>

struct ibv_device;
struct ibv_context;
struct mlx5dv_devx_obj;

namespace device_abstraction {

class linux_devx_adapter : public idevx_adapter
{
public:
	/**
	 * Constructor
	 * @param [in] ibv device handle
	 */
	linux_devx_adapter(struct ibv_device *handle) noexcept;

	/**
	 * Destructor
	 */
	virtual ~linux_devx_adapter() = default;

	/**
	 * @brief This method opens the devx adapter.
	 * @return DOCA_SUCCESS or DOCA_ERROR_NOT_CONNECTED
	 */
	doca_error_t open() noexcept;

	/**
	 * @brief This method closes the devx adapter.
	 * @return DOCA_SUCCESS or DOCA_ERROR_NOT_PERMITTED
	 */
	doca_error_t close() noexcept;

	/**
	 * @brief This method returns the current ibv context
	 * @return ibv_context or null if closed.
	 */
	struct ibv_context *get_ibv_context() const noexcept;

	/**
	 * @brief Issues general PRM command.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual doca_error_t send_prm_cmd(const void *in, size_t in_size, void *out, size_t out_size) noexcept override;

	/**
	 * @brief Issues general PRM command.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual create_devx_object_ret
	create_devx_object(const void *in, size_t in_size, void *out, size_t out_size) noexcept override;

	/**
	 * @brief - This method allocates and initializes a mapped user memory
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] access flags
	 * @return create_umem_ret (see interface)
	 */
	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size, size_t page_size,
							  uint32_t access_flags) noexcept override;

	/**
	 * @brief - This method allocates and initializes a mapped user memory with automatic page size
	 * @param [in] address
	 * @param [in] size
	 * @param [in] access flags
	 * @return create_umem_ret (see interface)
	 */
	virtual create_umem_ret create_mapped_user_memory(void *address, size_t size,
							  uint32_t access_flags) noexcept override;

	virtual doca_error_t query_event_queue_number(uint32_t core, uint32_t &eqn) const noexcept override;

private:

	struct ibv_device *m_ibv_device{nullptr};
	struct ibv_context *m_ibv_ctx{nullptr};

	linux_devx_adapter(linux_devx_adapter const &) = delete;
	linux_devx_adapter &operator=(linux_devx_adapter const &) = delete;
	linux_devx_adapter(linux_devx_adapter &&) = delete;
	linux_devx_adapter &operator=(linux_devx_adapter &&) = delete;
};

class linux_devx_object : public idevx_object
{
public:
	/**
	 * Constructor
	 */
	linux_devx_object(struct mlx5dv_devx_obj *devx_object) noexcept;

	/**
	 * Constructor
	 */
	virtual ~linux_devx_object() noexcept;

	/**
	 * @bfief This method explicitly destroys the devx object
	 * @return DOCA_SUCCESS or DOCA_ERROR_NOT_PERMITTED
	 */
	virtual doca_error_t destroy() noexcept override;

	/**
	 * @brief Modifies the devx object.
	 * @param [in] in input buffer, of PRM specification structure, to send.
	 * @param [in] in_size input buffer size.
	 * @param [out] out command's output data of PRM specification size.
	 * @param [out] out_size output data size.
	 *
	 * @return DOCA_SUCCESS on success. On failure, an appropriate error.
	 */
	virtual doca_error_t modify(const void *in, size_t in_size, void *out, size_t out_size) noexcept override;

	/**
	 * @brief This method retrieves the wrapped devx object
	 * @return wrapped devx object
	 */
	virtual void *get_object() const noexcept override;

private:

	struct mlx5dv_devx_obj *m_devx_object{nullptr};
};

} /* namespace device_abstraction */
