/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <cerrno>
#include <stdexcept>

#include <unistd.h>

#include <doca_error.h>
#include <doca_types.h>
#include <doca_log.h>
#include <priv_doca_types.h>

#include "linux_events_mux.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

namespace {

doca_error_t manual_fd_arm(union doca_data)
{
	return DOCA_SUCCESS;
}

doca_error_t manual_fd_clear(union doca_data)
{
	return DOCA_SUCCESS;
}

static const struct priv_doca_event_handle_ops manual_fd_ops = {
	.arm_cb = manual_fd_arm,
	.clear_cb = manual_fd_clear
};

} /* anonymous namespace */

linux_events_mux::~linux_events_mux()
{
	static_cast<void>(destroy());
}

doca_error_t
linux_events_mux::initialize() noexcept
{
	static constexpr int initial_size = 16;
	try {
		events_buffer.resize(initial_size);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to resize events buffer vector.");
		return DOCA_ERROR_NO_MEMORY;
	}

	epoll_fd = epoll_create1(0);
	if (epoll_fd == -1) {
		DOCA_DLOG_ERR("Failed to create epoll file descriptor. errno: %d", errno);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	manual_fd = eventfd(0, 0);
	if (manual_fd == -1) {
		static_cast<void>(destroy());
		DOCA_DLOG_ERR("Failed to create event file descriptor, for manual triggering of events. errno: %d",
			      errno);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	union doca_data ops_data;
	ops_data.u64 = static_cast<uint64_t>(manual_fd);
	auto ret = register_handle(reinterpret_cast<doca_event_handle_t>(manual_fd), manual_fd_ops, ops_data);
	if (ret != DOCA_SUCCESS) {
		static_cast<void>(destroy());
		DOCA_DLOG_ERR("Failed to add event file descriptor for manual triggering of events.");
		return ret;
	}

	return DOCA_SUCCESS;
}

doca_error_t
linux_events_mux::destroy() noexcept
{
	if (manual_fd != -1) {
		auto ret = unregister_handle(reinterpret_cast<doca_event_handle_t>(manual_fd));
		/* It is possible that manual_fd was created but was not registered. */
		if (ret != DOCA_SUCCESS && ret != DOCA_ERROR_NOT_FOUND)
			return ret;

		if (close(manual_fd) != 0) {
			DOCA_DLOG_ERR("Failed to close event file descriptor. errno: %d", errno);
			return DOCA_ERROR_OPERATING_SYSTEM;
		}

		manual_fd = -1;
	}
	if (registered_handles.size() > 0) {
		DOCA_DLOG_ERR("Attempting to destroy epoll mux without removing all fds from it.");
		return DOCA_ERROR_IN_USE;
	}

	if (epoll_fd != -1) {
		if (close(epoll_fd) != 0) {
			DOCA_DLOG_ERR("Failed to close epoll file descriptor. errno: %d", errno);
			return DOCA_ERROR_OPERATING_SYSTEM;
		}

		epoll_fd = -1;
	}

	return DOCA_SUCCESS;
}

doca_error_t
linux_events_mux::register_handle(doca_event_handle_t handle, const struct priv_doca_event_handle_ops& ops,
				  union doca_data ops_data) noexcept
{
	auto it = registered_handles.end();
	try {
		auto it_inserted = registered_handles.emplace(handle, event_handle_cookie{ops, ops_data});
		it = it_inserted.first;
		if (!it_inserted.second) {
			DOCA_DLOG_ERR("Attempting to register handle that is already registered.");
			return DOCA_ERROR_IN_USE;
		}
	} catch (...) {
		DOCA_DLOG_ERR("Failed to register handle, failed to insert handle to map. Out of memory.");
		return DOCA_ERROR_NO_MEMORY;
	}

	if (registered_handles.size() > events_buffer.size()) {
		try {
			events_buffer.resize(registered_handles.size());
		} catch (...) {
			registered_handles.erase(it);
			DOCA_DLOG_ERR("Failed to register handle, failed to resize events buffer. Out of memory.");
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	struct epoll_event events_in = {
		.events = EPOLLIN,
		.data = {
			/* Address of data in map is guaranteed not to change. Unless erased.*/
			.ptr = static_cast<void*>(&it->second),
		},
	};
	if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, static_cast<int>(handle), &events_in) != 0) {
		DOCA_DLOG_ERR("Failed to add file descriptor to epoll fd using epoll_ctl(). errno: %d, fd: %d",
			      errno, static_cast<int>(handle));
		registered_handles.erase(it);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	return DOCA_SUCCESS;
}

doca_error_t
linux_events_mux::unregister_handle(doca_event_handle_t handle) noexcept
{
	auto it = registered_handles.find(handle);
	if (it == registered_handles.end()) {
		DOCA_DLOG_ERR("Attempting to unregister a handle that is not registered.");
		return DOCA_ERROR_NOT_FOUND;
	}
	if (epoll_ctl(epoll_fd, EPOLL_CTL_DEL, static_cast<int>(handle), nullptr) != 0) {
		DOCA_DLOG_ERR("Failed to remove fd from epoll using epoll_ctl. errno: %d", errno);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}
	registered_handles.erase(it);

	return DOCA_SUCCESS;
}

doca_error_t
linux_events_mux::arm() noexcept
{
	for (const auto& handle : registered_handles) {
		const auto& cookie = handle.second;
		auto ret = cookie.first.arm_cb(cookie.second);
		if (ret != DOCA_SUCCESS)
			return ret;
	}

	return DOCA_SUCCESS;
}

doca_error_t
linux_events_mux::clear(doca_event_handle_t) noexcept
{
	int nb_max_events = static_cast<int>(events_buffer.size());

	auto nb_events = epoll_wait(epoll_fd, &events_buffer[0], nb_max_events, /*timeout=*/0);
	if (nb_events == -1) {
		DOCA_DLOG_ERR("Failed to clear events, epoll_wait() with timeout 0 failed. errno: %d", errno);
		return DOCA_ERROR_OPERATING_SYSTEM;
	}

	for (int event_idx = 0; event_idx < nb_events; ++event_idx) {
		auto cookie = static_cast<event_handle_cookie*>(events_buffer[event_idx].data.ptr);
		auto ret = cookie->first.clear_cb(cookie->second);
		if (ret != DOCA_SUCCESS)
			return ret;
	}

	return DOCA_SUCCESS;
}

doca_event_handle_t
linux_events_mux::get_handle() const noexcept
{
	return static_cast<doca_event_handle_t>(epoll_fd);
}

doca_error_t
linux_events_mux::trigger_event() noexcept
{
	/* fd is considered ready for read, only in case the write is greater than 0. */
	static constexpr eventfd_t greater_than_zero = 1;

	if (eventfd_write(manual_fd, greater_than_zero) == 0)
		return DOCA_SUCCESS;

	DOCA_DLOG_ERR("Failed to manually trigger an event, eventfd_write failed. fd=%d, write_value=%lu",
		      manual_fd, greater_than_zero);
	return DOCA_ERROR_OPERATING_SYSTEM;
}

doca_error_t
linux_events_mux::untrigger_event() noexcept
{
	eventfd_t ignored{0};

	if (eventfd_read(manual_fd, &ignored) == 0)
		return DOCA_SUCCESS;

	DOCA_DLOG_ERR("Failed to untrigger manually triggered event, eventfd_read failed. fd=%d", manual_fd);
	return DOCA_ERROR_OPERATING_SYSTEM;
}
