/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_log.h>
#include "linux_mapped_user_memory.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

create_umem_ret
linux_mapped_user_memory::create(struct ibv_context *ibv_ctx, void *address, size_t size, size_t page_size,
				 uint32_t access_flags) noexcept
{
	std::shared_ptr<linux_mapped_user_memory> umem_obj{nullptr};

	try {
		umem_obj = std::make_shared<linux_mapped_user_memory>(address, size, page_size, access_flags);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create mapped user memory object");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<linux_mapped_user_memory>(nullptr)};
	}

	auto status = umem_obj->initialize(ibv_ctx);
	if (status != DOCA_SUCCESS)
		umem_obj.reset();

	return {status, umem_obj};
}

linux_mapped_user_memory::linux_mapped_user_memory(void *address, size_t size, size_t page_size,
						   uint32_t access_flags) noexcept
	:mapped_user_memory(address, size, page_size, access_flags)
{
}

linux_mapped_user_memory::~linux_mapped_user_memory()
{
	static_cast<void>(dereg());
}

doca_error_t
linux_mapped_user_memory::initialize(struct ibv_context *ibv_ctx) noexcept
{

	struct mlx5dv_devx_umem_in umem_in;

	umem_in.addr = m_address;
	umem_in.size = m_size;
	umem_in.access = m_access_flags;
	umem_in.pgsz_bitmap = m_page_size;
	umem_in.comp_mask = 0;

	m_handle = mlx5dv_devx_umem_reg_ex(ibv_ctx, &umem_in);
	if (!m_handle) {
		DOCA_DLOG_ERR("Failed to register user memory");
		return DOCA_ERROR_NO_MEMORY;
	}

	m_id = m_handle->umem_id;

	return DOCA_SUCCESS;
}

doca_error_t
linux_mapped_user_memory::dereg() noexcept
{
	if (m_handle) {
		int ret = mlx5dv_devx_umem_dereg(m_handle);
		if (ret != 0) {
			DOCA_DLOG_ERR("Failed to dereg user memory with ret %d", ret);
			return DOCA_ERROR_NOT_PERMITTED;
		}

		m_id = 0;
		m_handle = nullptr;
	}

	return DOCA_SUCCESS;
}

uint32_t
linux_mapped_user_memory::get_id() const noexcept
{
	return m_id;
}

