/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <infiniband/mlx5dv.h>

#include <doca_error.h>
#include <mapped_user_memory.hpp>

namespace device_abstraction {

class linux_mapped_user_memory : public mapped_user_memory
{
public:

	/**
	 * @brief - This method allocates and initializes a linux_user_memory
	 * @param [in] ibv context
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] access flags
	 * @return create_umem_ret (see interface)
	 */
	static create_umem_ret
	create(struct ibv_context *ibv_ctx, void* address, size_t size, size_t page_size,
		uint32_t access_flags) noexcept;

	/**
	 * @brief - This method allocates and initializes a linux_user_memory using variadic args for type safety.
	 * @param [in] ibv context
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] variadic access flags
	 * @return create_umem_ret (see interface)
	 */
	template <typename... Args>
	create_umem_ret
	safe_create(struct ibv_context *ibv_ctx, void* address, size_t size, size_t page_size,
		     Args... flags) noexcept
	{
		int access_flags = memory_access_flags_to_uint(flags...);
		return create(ibv_ctx, address, size, page_size, access_flags);
	}

	/**
	 * Constructor
	 * @param [in] address
	 * @param [in] size
	 * @param [in] page size
	 * @param [in] access flags
	 */
	linux_mapped_user_memory(void *address, size_t size, size_t page_size, uint32_t access_flags) noexcept;

	/**
	 * Destructor
	 */
	virtual ~linux_mapped_user_memory();

	/**
	 * brief - this method initializes the linux_mapped_user_memory
	 * @param [in] ibv context
	 * @return doca_error_t
	 */
	doca_error_t initialize(struct ibv_context *ibv_ctx) noexcept;

	/**
	 * @brief This method deregisters the umem
	 * @return DOCA_SUCCESS or DOCA_ERROR_NOT_PERMITTED
	 */
	virtual doca_error_t dereg() noexcept override;

	/**
	 * brief This method returns the linux_zero_based_mapped_user_memory id
	 */
	virtual uint32_t get_id() const noexcept override;

	linux_mapped_user_memory(linux_mapped_user_memory const &) = delete;
	linux_mapped_user_memory &operator=(linux_mapped_user_memory const &) = delete;
	linux_mapped_user_memory(linux_mapped_user_memory &&) = delete;
	linux_mapped_user_memory &operator=(linux_mapped_user_memory &&) = delete;

private:

	mlx5dv_devx_umem *m_handle{ nullptr };
	uint32_t m_id{ 0 };
};

} /* namespace device_abstraction */
