/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>
#include <infiniband/mlx5dv.h>

#include <doca_log.h>
#include "linux_devx_adapter.hpp"
#include "linux_protection_domain.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

std::shared_ptr<linux_protection_domain>
linux_protection_domain::create(std::shared_ptr<linux_devx_adapter> devx_adapter) noexcept
{
	std::shared_ptr<linux_protection_domain> pd_obj{nullptr};
	try {
		pd_obj = std::make_shared<linux_protection_domain>(devx_adapter);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create protection domain object");
		return std::shared_ptr<linux_protection_domain>(nullptr);
	}

	if (pd_obj) {
		if (pd_obj->initialize() != DOCA_SUCCESS)
			pd_obj.reset();
	}
	return pd_obj;
}

linux_protection_domain::linux_protection_domain(std::shared_ptr<linux_devx_adapter> devx_adapter) noexcept
	:protection_domain(devx_adapter)
{

}

linux_protection_domain::~linux_protection_domain()
{
	if (destroy() != DOCA_SUCCESS)
		DOCA_DLOG_ERR("Failed to destroy protection domain");
}

doca_error_t
linux_protection_domain::initialize() noexcept
{
	auto devx_adapter = std::static_pointer_cast<linux_devx_adapter>(m_devx_adapter);
	m_pd_obj = ibv_alloc_pd(devx_adapter->get_ibv_context());

	if (m_pd_obj == nullptr) {
		DOCA_DLOG_ERR("mlx5dv_devx_obj_create returned null");
		return DOCA_ERROR_NO_MEMORY;
	}

	/* Cast pd to devx object */

	struct mlx5dv_pd dv_pd{};
	struct mlx5dv_obj devx_obj{};

	devx_obj.pd.in = m_pd_obj;
	devx_obj.pd.out = &dv_pd;

	if (mlx5dv_init_obj(&devx_obj, MLX5DV_OBJ_PD) != 0) {
		DOCA_DLOG_ERR("Unable to expose pdn from pd.");
		return DOCA_ERROR_DRIVER;
	}

	m_pd = devx_obj.pd.out->pdn;

	return DOCA_SUCCESS;
}

doca_error_t
linux_protection_domain::destroy() noexcept
{
	if (m_pd_obj) {
		auto ret = ibv_dealloc_pd(m_pd_obj);

		if (ret != 0) {
			DOCA_DLOG_ERR("devx_obj_destroy failed for pd_obj with errno = %d\n", ret);
			return DOCA_ERROR_DRIVER;
		}

		m_pd_obj = nullptr;
	}

	return DOCA_SUCCESS;
}

uint32_t
linux_protection_domain::get_pd_id() const noexcept
{
	return m_pd;
}

struct ibv_pd *
linux_protection_domain::get_ibv_pd() const noexcept
{
	return m_pd_obj;
}

