/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <doca_log.h>
#include "linux_user_access_region.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

iuser_access_region::create_ret
linux_user_access_region::create(struct ibv_context *ibv_ctx,
				 iuser_access_region::allocation_type allocation_type) noexcept
{
	std::shared_ptr<linux_user_access_region> uar_obj{nullptr};
	try {
		uar_obj = std::make_shared<linux_user_access_region>();
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create user access region object");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<linux_user_access_region>(nullptr)};
	}

	auto status = uar_obj->initialize(ibv_ctx, allocation_type);
	if (status != DOCA_SUCCESS)
		uar_obj.reset();

	return {status, uar_obj};
}

linux_user_access_region::~linux_user_access_region()
{
	if (m_uar_obj != nullptr) {
		mlx5dv_devx_free_uar(m_uar_obj);
		m_uar_obj = nullptr;
	}
}

doca_error
linux_user_access_region::initialize(struct ibv_context *ibv_ctx,
				     iuser_access_region::allocation_type allocation_type) noexcept
{
	m_uar_obj = mlx5dv_devx_alloc_uar(ibv_ctx, static_cast<uint32_t>(allocation_type));
	if (!m_uar_obj) {
		DOCA_DLOG_ERR("mlx5dv_devx_obj_create returned null");
		return DOCA_ERROR_DRIVER;
	}

	return DOCA_SUCCESS;
}

uint32_t
linux_user_access_region::get_id() const noexcept
{
	return m_uar_obj->page_id;
}

void *
linux_user_access_region::get_page() const noexcept
{
	return m_uar_obj->base_addr;
}
