/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <tuple>

#include <doca_log.h>
#include <memory_region.hpp>
#include "mapped_memory_region.hpp"
#include "mapped_user_memory.hpp"
#include <idevx_adapter.hpp>

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::MAPPED_MR);


create_mr_ret
mapped_memory_region::create(std::shared_ptr<idevx_adapter> devx_adapter, std::shared_ptr<iprotection_domain> pd,
			      std::shared_ptr<imapped_user_memory> umem, bool is_zero_based) noexcept
{
	std::shared_ptr<mapped_memory_region> mr_obj{nullptr};
	try {
		mr_obj = std::make_shared<mapped_memory_region>(devx_adapter, pd, umem);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate mapped memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<mapped_memory_region>()};
	}

	auto status = mr_obj->initialize(is_zero_based);
	if (status != DOCA_SUCCESS)
		mr_obj.reset();

	return {status, mr_obj};
}

mapped_memory_region::mapped_memory_region(std::shared_ptr<idevx_adapter> devx_adapter,
					   std::shared_ptr<iprotection_domain> pd,
					   std::shared_ptr<imapped_user_memory> umem) noexcept
	: memory_region_base(devx_adapter, pd), m_umem(umem)
{
}


doca_error_t
mapped_memory_region::initialize(bool is_zero_based) noexcept
{
	create_mapped_mkey_in in{0};
	create_mapped_mkey_out out{0};

	auto status = fill_mapped_memory_region_cmd(in, m_pd, m_umem, is_zero_based);
	if (status != DOCA_SUCCESS)
		return status;

	std::tie(status, m_mkey) = m_devx_adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create mkey object");
		return status;
	}

	m_key_index = get_mapped_memory_region_key_index(out);

	return DOCA_SUCCESS;
}

mapped_memory_region::~mapped_memory_region()
{
	// mkey is destroyed in the base class it is too late because it has to be destroyed before the umem.
	// Therefore destroying the key explicitly here.
	destroy_key();
}

doca_error_t
mapped_memory_region::allow_other_vhca_access(const allow_other_vhca_access_key_t &access_key) noexcept
{
	allow_other_vhca_access_cmd_in in = {0};
	allow_other_vhca_access_cmd_out out = {0};

	auto status = fill_allow_other_vhca_access_cmd(in, access_key, get_id());
	if (status != DOCA_SUCCESS)
		return status;

	status = m_devx_adapter->send_prm_cmd(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to allow other vhca access syndrome = 0x%x",
			      get_allow_other_vhca_access_fail_syndrome(out));
		return status;
	}

	return DOCA_SUCCESS;
}


