/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <doca_error.h>

#include <multi_os_dv.h>
#include <multi_os_prm.h>
#include <imemory_region.hpp>
#include <imapped_user_memory.hpp>
#include <iprotection_domain.hpp>
#include "memory_region_base.hpp"

namespace device_abstraction {

class mapped_memory_region : public memory_region_base
{
public:

	/**
	 * @brief - This method allocates and initializes a mapped_memory_region
	 * @param [in] devx adapter
	 * @param [in] protection domain
	 * @param [in] umem
	 * @param [in] is zero based
	 * @return create_mr_ret (see interface)
	 */
	static create_mr_ret
	create(std::shared_ptr<idevx_adapter> devx_adapter, std::shared_ptr<iprotection_domain> pd,
		std::shared_ptr<imapped_user_memory> umem, bool is_zero_based) noexcept;

	/**
	 * Constructor
	 * @param [in] devx adapter
	 * @param [in] protection domain
	 * @param [in] umem
	 */
	mapped_memory_region(std::shared_ptr<idevx_adapter> devx_adapter,
			     std::shared_ptr<iprotection_domain> pd,
			     std::shared_ptr<imapped_user_memory> umem) noexcept;

	/**
	 * Destructor
	 */
	virtual ~mapped_memory_region();

	/**
	 * brief - this method initializes the memory region
	 * @param [in] is zero based
	 * @return doca_error_t
	 */
	doca_error_t initialize(bool is_zero_based) noexcept;

	/**
	 * @brief This method allows other VHCA access
	 * @param [in] access key
	 * @return doca_error_t
	 */
	virtual doca_error_t allow_other_vhca_access(const allow_other_vhca_access_key_t &access_key) noexcept override;

private:

	std::shared_ptr<imapped_user_memory> m_umem{ nullptr };

	mapped_memory_region(mapped_memory_region const &) = delete;
	mapped_memory_region &operator=(mapped_memory_region const &) = delete;
	mapped_memory_region(mapped_memory_region &&) = delete;
	mapped_memory_region &operator=(mapped_memory_region &&) = delete;
};

} /* namespace device_abstraction */

