/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>

namespace device_abstraction {

enum class memory_access_flags : uint32_t{
	local_write = 1 << 0,
	remote_write = 1 << 1,
	remote_read = 1 << 2,
	remote_atomic = 1 << 3,
	zero_based = 1 << 5,
};

/**
 * The methods below uses variadic templates to safely convert memory access flags to int
 * The methods allow only memory_access_flags and will fail in compile time if any other
 * parameter is used.
 * Variadic templates are extracted during compile time (hence no recursion during run time).
 */

/**
 * @brief This method converts empty memory access flags to int
 * @return 0
 */
inline uint32_t
memory_access_flags_to_uint() noexcept
{
	return 0;
}

/**
 * @brief This method converts memory access flags to int
 * @param [in] flag
 * @return int
 */
inline uint32_t
memory_access_flags_to_uint(memory_access_flags flag) noexcept
{
	return static_cast<uint32_t>(flag);
}

/**
 * @brief This method converts memory access flags to int
 * @param [in] flag
 * @param [in] variadic access flags
 * @return int
 */
template <typename... Args>
uint32_t
memory_access_flags_to_uint(memory_access_flags first_flag, Args... flags) noexcept
{
	return static_cast<uint32_t>(first_flag) | memory_access_flags_to_uint(flags...);
}

/**
 * @brief This method checks if a mask contains a memory access flag
 * @param mask
 * @param flag
 * @return true if flag enabled, false if not.
 */
inline bool
is_mem_access_flag_on(uint32_t mask, memory_access_flags flag) noexcept
{
	return mask & static_cast<uint32_t>(flag);
}

} /* namespace device_abstraction */

