/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <cmath>

#include "memory_region.hpp"
#include <doca_log.h>
#include <multi_os_dv.h>
#include <multi_os_prm.h>

#define MLX5_GENERAL_OBJ_TYPE_MKEY 0xff01
#define MLX5_GENERAL_OBJ_TYPE_MKEY 0xff01

using namespace device_abstraction;
using namespace std;

DOCA_LOG_REGISTER(CORE::MEMORY_REGION);

static const int variant_part_of_mkey = 0x42;

doca_error_t
device_abstraction::fill_mapped_memory_region_cmd(create_mapped_mkey_in &in, shared_ptr<iprotection_domain> pd,
						  shared_ptr<imapped_user_memory> umem, bool is_zero_based) noexcept
{
	int log_of_umem_page_size{0};
	auto page_size = umem->get_page_size();
	try {
		log_of_umem_page_size = static_cast<int>(std::log2(page_size));
	} catch (...) {
		DOCA_DLOG_ERR("Invalid page size %zu", umem->get_page_size());
		return DOCA_ERROR_INVALID_VALUE;
	}

	uintptr_t mkey_offset{0};

	if (is_zero_based) {
		auto addr = reinterpret_cast<uintptr_t>(umem->get_buffer());

		if ((addr % page_size) != 0) {
			DOCA_DLOG_ERR(
				"creation of zero-base mkey allowed only when buffer address align with memory page size");
			return DOCA_ERROR_NOT_SUPPORTED;
		}
	} else {
		mkey_offset = reinterpret_cast<uintptr_t>(umem->get_buffer());
	}

	MLX5_SET(create_mkey_in, in, opcode, MLX5_CMD_OP_CREATE_MKEY);
	MLX5_SET(create_mkey_in, in, translations_octword_actual_size, 1);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.access_mode_1_0, static_cast<int>(MLX5_MKC_ACCESS_MODE_MTT));
	auto access_flags = umem->get_access_flags();
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.a,
		 is_mem_access_flag_on(access_flags,memory_access_flags::remote_atomic));
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.rw,
		 is_mem_access_flag_on(access_flags,memory_access_flags::remote_write));
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.rr,
		 is_mem_access_flag_on(access_flags, memory_access_flags::remote_read));
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.lw,
		 is_mem_access_flag_on(access_flags, memory_access_flags::local_write));
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.lr, 1);

	MLX5_SET64(create_mkey_in, in, memory_key_mkey_entry.start_addr, mkey_offset);

	MLX5_SET64(create_mkey_in, in, memory_key_mkey_entry.len, umem->get_size());
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.pd, pd->get_pd_id());
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.qpn, 0xffffff);
	MLX5_SET(create_mkey_in, in, memory_key_mkey_entry.mkey_7_0, variant_part_of_mkey);
	MLX5_SET(create_mkey_in, in, pg_access, 1);
	MLX5_SET(create_mkey_in, in, mkey_umem_id, umem->get_id());
	void *mkc = MLX5_ADDR_OF(create_mkey_in, in, memory_key_mkey_entry);
	MLX5_SET(mkc, mkc, translations_octword_size_crossing_target_mkey, 1);
	MLX5_SET(mkc, mkc, log_entity_size, log_of_umem_page_size);

	return DOCA_SUCCESS;
}

uint32_t
device_abstraction::get_mapped_memory_region_key_index(const create_mapped_mkey_out &out) noexcept
{
	return MLX5_GET(create_mkey_out, out, mkey_index) << 8 | variant_part_of_mkey;
}

doca_error_t
device_abstraction::fill_alias_memory_region_cmd(create_alias_mkey_in &in,
						  std::shared_ptr<iprotection_domain> protection_domain, uint32_t lkey,
						 uint16_t vhca_id,
						 const allow_other_vhca_access_key_t &access_key) noexcept
{
	void *hdr, *alias_ctx, *access_key_addr, *metadata;

	static_assert(allow_other_vhca_access_key_size == MLX5_FLD_SZ_BYTES(alias_context, access_key),
		      "Access key size mismatch");

	/* Create alias */
	hdr = MLX5_ADDR_OF(create_alias_obj_in, in, hdr);
	alias_ctx = MLX5_ADDR_OF(create_alias_obj_in, in, alias_ctx);

	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJ_TYPE_MKEY);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, alias_object, 1);
	MLX5_SET(alias_context, alias_ctx, vhca_id_to_be_accessed, vhca_id);
	MLX5_SET(alias_context, alias_ctx, object_id_to_be_accessed, lkey >> 8);
	access_key_addr = MLX5_ADDR_OF(alias_context, alias_ctx, access_key);
	memcpy(access_key_addr, access_key.data(), access_key.size());
	metadata = MLX5_ADDR_OF(alias_context, alias_ctx, metadata);
	int pdn = htobe32(protection_domain->get_pd_id());
	memcpy(metadata, &pdn, sizeof(pdn));

	return DOCA_SUCCESS;
}

doca_error_t
device_abstraction::get_alias_memory_region_key_index(create_alias_mkey_out &out, uint32_t &key_index) noexcept
{
	auto ret = MLX5_GET(create_alias_obj_out, out, alias_ctx.status);
	if (ret) {
		DOCA_DLOG_ERR("Created mr alias object in bad state, ret = %d", ret);
		key_index = 0;
		return DOCA_ERROR_BAD_STATE;
	}

	key_index = MLX5_GET(create_alias_obj_out, out, hdr.obj_id) << 8;

	return DOCA_SUCCESS;
}

doca_error_t
device_abstraction::fill_allow_other_vhca_access_cmd(allow_other_vhca_access_cmd_in &in,
						     const allow_other_vhca_access_key_t &access_key,
						     uint32_t lkey) noexcept
{
	void *access_key_dest{nullptr};

	/* Allow crossing mkey */
	MLX5_SET(allow_other_vhca_access_in, in, opcode, MLX5_CMD_OP_ALLOW_OTHER_VHCA_ACCESS);
	MLX5_SET(allow_other_vhca_access_in, in, object_type_to_be_accessed, MLX5_GENERAL_OBJ_TYPE_MKEY);
	MLX5_SET(allow_other_vhca_access_in, in, object_id_to_be_accessed, lkey >> 8);
	access_key_dest = MLX5_ADDR_OF(allow_other_vhca_access_in, in, access_key);
	memcpy(access_key_dest, access_key.data(), access_key.size());

	return DOCA_SUCCESS;
}

uint32_t
device_abstraction::get_allow_other_vhca_access_fail_syndrome(allow_other_vhca_access_cmd_out &out) noexcept
{
	return MLX5_GET(allow_other_vhca_access_out, out, syndrome);
}

doca_error_t
device_abstraction::fill_crossing_memory_region_cmd(create_crossing_mkey_in &in,
				const std::shared_ptr<iprotection_domain> protection_domain,
				uint32_t introspection_mkey, uint16_t vhca_id) noexcept
{
	void *mkc = MLX5_ADDR_OF(create_mkey_in, in, memory_key_mkey_entry);

	MLX5_SET(create_mkey_in, in, opcode, MLX5_CMD_OP_CREATE_MKEY);
	MLX5_SET(mkc, mkc, access_mode_1_0, MLX5_MKC_ACCESS_MODE_CROSSING_VHCA_MKEY & 0x3);
	MLX5_SET(mkc, mkc, access_mode_4_2, (MLX5_MKC_ACCESS_MODE_CROSSING_VHCA_MKEY >> 2) & 0x7);
	MLX5_SET(mkc, mkc, a, 1);
	MLX5_SET(mkc, mkc, rw, 1);
	MLX5_SET(mkc, mkc, rr, 1);
	MLX5_SET(mkc, mkc, lw, 1);
	MLX5_SET(mkc, mkc, lr, 1);
	MLX5_SET(mkc, mkc, pd, protection_domain->get_pd_id());
	MLX5_SET(mkc, mkc, qpn, 0xffffff);
	MLX5_SET(mkc, mkc, length64, 1);
	MLX5_SET(mkc, mkc, mkey_7_0, 0x42);
	MLX5_SET(mkc, mkc, crossing_target_vhca_id, vhca_id);
	MLX5_SET(mkc, mkc, translations_octword_size_crossing_target_mkey, introspection_mkey);

	return DOCA_SUCCESS;
}

uint32_t
device_abstraction::get_crossing_memory_region_key_index(const create_crossing_mkey_out &out) noexcept
{
	return (MLX5_GET(create_mkey_out, out, mkey_index) << 8 | 0x42);
}

uint32_t
device_abstraction::get_crossing_mkey_fail_syndrome(create_crossing_mkey_out &out) noexcept
{
	return MLX5_GET(create_mkey_out, out, syndrome);
}
