/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <doca_error.h>
#include <iprotection_domain.hpp>
#include <imapped_user_memory.hpp>
#include <imemory_region.hpp>
#include <multi_os_dv.h>
#include <multi_os_prm.h>

namespace device_abstraction {

using create_mapped_mkey_in = uint32_t[MLX5_ST_SZ_DW(create_mkey_in)];
using create_mapped_mkey_out = uint32_t[MLX5_ST_SZ_DW(create_mkey_out)];

/**
 * @brief - this method fills the mapped memory region command
 * @param [in][out] command buffer
 * @param [in] protection domain
 * @param [in] user mapped memory
 * @param [in] is zero based
 * @return doca_error_t
 */
doca_error_t fill_mapped_memory_region_cmd(create_mapped_mkey_in &in, std::shared_ptr<iprotection_domain> pd,
					   std::shared_ptr<imapped_user_memory> umem, bool is_zero_based) noexcept;

/**
 * @brief This method retrieves the mapped memory region key index
 * @param [in] command out buffer
 * @return key index
 */
uint32_t get_mapped_memory_region_key_index(const create_mapped_mkey_out &out) noexcept;

using create_alias_mkey_in = uint32_t[MLX5_ST_SZ_DW(create_alias_obj_in)];
using create_alias_mkey_out = uint32_t[MLX5_ST_SZ_DW(create_alias_obj_out)];

/**
 * @brief This method fills the alias memory region command
 * @param [in] command buffer
 * @param [in] protection domain
 * @param [in] lkey
 * @param [in] vhca id
 * @param [in] access key
 * @return doca_error_t
 */
doca_error_t fill_alias_memory_region_cmd(create_alias_mkey_in &in,
					  std::shared_ptr<iprotection_domain> protection_domain, uint32_t lkey,
					  uint16_t vhca_id, const allow_other_vhca_access_key_t &access_key) noexcept;

/**
 * @brief This method retrieves teh key index out of the command
 * @param [in] command out buffer
 * @param [out] key index
 * @return doca_error_t
 */
doca_error_t get_alias_memory_region_key_index(create_alias_mkey_out &out, uint32_t &key_index) noexcept;

using allow_other_vhca_access_cmd_in = uint32_t[MLX5_ST_SZ_DW(allow_other_vhca_access_in)];
using allow_other_vhca_access_cmd_out = uint32_t [MLX5_ST_SZ_DW(allow_other_vhca_access_out)];

/**
 * @brief This method fills the allow other vhca access command
 * @param [in] command buffer
 * @param [in] access key

 * @param [in] lkey
 * @return doca_error_t
 */
doca_error_t fill_allow_other_vhca_access_cmd(allow_other_vhca_access_cmd_in &in,
					      const allow_other_vhca_access_key_t &access_key, uint32_t lkey) noexcept;

/**
 * @brief This method returns the allow vhca access failure syndrome.
 * @param [in] command out buffer
 * @return uint32_t
 */
uint32_t get_allow_other_vhca_access_fail_syndrome(allow_other_vhca_access_cmd_out &out) noexcept;

using create_crossing_mkey_in = uint8_t[MLX5_ST_SZ_BYTES(create_mkey_in)];
using create_crossing_mkey_out = uint8_t[MLX5_ST_SZ_BYTES(create_mkey_out)];

/**
 * @brief This method fills the crossing memory region command
 * @param [in] command buffer
 * @param [in] protection domain
 * @param [in] introspection mkey
 * @param [in] vhca id
 * @return doca_error_t
 */
doca_error_t fill_crossing_memory_region_cmd(create_crossing_mkey_in &in,
				const std::shared_ptr<iprotection_domain> protection_domain,
				uint32_t introspection_mkey, uint16_t vhca_id) noexcept;

/**
 * @brief This method retrieves teh key index out of the command
 * @param [in] command out buffer
 * @param [out] key index
 * @return uint32_t
 */
uint32_t get_crossing_memory_region_key_index(const create_crossing_mkey_out &out) noexcept;

/**
 * @brief This method returns the crossing_mkey failure syndrome.
 * @param [in] command out buffer
 * @return uint32_t
 */
uint32_t get_crossing_mkey_fail_syndrome(create_crossing_mkey_out &out) noexcept;

} /* namespace device_abstraction */
