/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>

#include <doca_log.h>
#include "idevx_adapter.hpp"
#include "protection_domain.hpp"
#include "mapped_memory_region.hpp"
#include "alias_memory_region.hpp"
#include "mapped_user_memory.hpp"
#include "crossing_memory_region.hpp"
#include "receive_memory_pool.hpp"
#include <indirect_memory_region.hpp>
#include <user_memory_registration.hpp>
#include <dynamically_connected_target.hpp>

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::PROTECTION_DOMAIN);

protection_domain::protection_domain(std::shared_ptr<idevx_adapter> devx_adapter) noexcept
	:m_devx_adapter(devx_adapter)
{
}

create_mr_ret
protection_domain::create_mapped_memory_region(void *address, size_t size, size_t page_size,
					       uint32_t access_flags) noexcept
{
	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<imapped_user_memory> umem_obj;
	std::tie(status, umem_obj) = m_devx_adapter->create_mapped_user_memory(address, size, page_size, access_flags);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create mapped user memory");
		return {status, std::shared_ptr<imemory_region>(nullptr)};
	}

	try {
		return mapped_memory_region::create(m_devx_adapter, shared_from_this(), umem_obj, false);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create mapped memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<imemory_region>(nullptr)};
	}
}

create_mr_ret
protection_domain::create_zero_based_mapped_memory_region(void *address, size_t size, size_t page_size,
							  uint32_t access_flags) noexcept
{
	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<imapped_user_memory> umem_obj;
	std::tie(status, umem_obj) = m_devx_adapter->create_mapped_user_memory(address, size, page_size, access_flags);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create mapped user memory");
		return {status, std::shared_ptr<imemory_region>(nullptr)};
	}

	try {
		return mapped_memory_region::create(m_devx_adapter, shared_from_this(), umem_obj, true);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create zero based mapped memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<imemory_region>(nullptr)};
	}
}

create_mr_ret
protection_domain::create_alias_memory_region(uint32_t lkey, uint16_t vhca_id,
					      const allow_other_vhca_access_key_t &access_key) noexcept
{
	try {
		return alias_memory_region::create(m_devx_adapter, shared_from_this(), lkey,
							 vhca_id, access_key);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create alias memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<imemory_region>(nullptr)};
	}
}

create_mr_ret
protection_domain::create_crossing_memory_region(uint32_t introspection_mkey, uint16_t vhca_id) noexcept
{
	try {
		return crossing_memory_region::create(m_devx_adapter, shared_from_this(), introspection_mkey, vhca_id);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create crossing memory region");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<imemory_region>(nullptr)};
	}
}

ireceive_memory_pool::create_ret
protection_domain::create_receive_memory_pool(uint32_t wq_size) noexcept
{
	try {
		return receive_memory_pool::create(m_devx_adapter, shared_from_this(), wq_size);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create receive memory pool");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<ireceive_memory_pool>(nullptr)};
	}
}

create_mr_ret
protection_domain::create_indirect_memory_region(uint32_t num_regions) noexcept
{
	try {
		return indirect_memory_region::create(shared_from_this(), m_devx_adapter, num_regions);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create indirect memory region.");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<imemory_region>(nullptr)};
	}
}

create_umr_ret
protection_domain::create_user_memory_registration(uint32_t num_operations) noexcept
{
	try {
		return user_memory_registration::create(shared_from_this(), num_operations);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create umr.");
		return {DOCA_ERROR_NO_MEMORY, std::unique_ptr<iuser_memory_registration>(nullptr)};
	}
}

idynamically_connected_target::create_ret
protection_domain::create_dynamically_connected_target(
	idynamically_connected_target::attributes const &attr) noexcept
{
	try {
		return dynamically_connected_target::create(m_devx_adapter, shared_from_this(), attr);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create dynamically connected target");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<idynamically_connected_target>(nullptr)};
	}
}