/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <tuple>

#include <doca_log.h>
#include <multi_os_prm.h>
#include <priv_doca_os_abstraction.hpp>
#include <utils.hpp>

#include "receive_memory_pool.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

using create_rmp_in = uint32_t[MLX5_ST_SZ_DW(create_rmp_in)];
using create_rmp_out = uint32_t[MLX5_ST_SZ_DW(create_rmp_out)];

struct receive_memory_pool_attributes {
	uint32_t pdn;
	uint32_t log_wq_size;
	uint32_t wq_umem_id;
	uint64_t wq_umem_offset;
	uint32_t db_umem_id;
	uint64_t db_umem_offset;
};

static doca_error_t
fill_create_rmp_cmd(create_rmp_in &in, receive_memory_pool_attributes const &attr) noexcept
{
	// mlx5_ifc_create_rmp_in_bits
	MLX5_SET(create_rmp_in, in, opcode, MLX5_CMD_OP_CREATE_RMP);

	// mlx5_ifc_rmpc_bits
	void *rmp_context = MLX5_ADDR_OF(create_rmp_in, in, ctx);
	MLX5_SET(rmpc, rmp_context, state, MLX5_RMPC_STATE_RDY);
	MLX5_SET(rmpc, rmp_context, basic_cyclic_rcv_wqe, 1);

	// mlx5_ifc_wq_bits
	void *wq_context = MLX5_ADDR_OF(rmpc, rmp_context, wq);
	MLX5_SET(wq, wq_context, wq_type, MLX5_WQ_TYPE_CYCLIC);
	MLX5_SET(wq, wq_context, pd, attr.pdn);

	MLX5_SET(wq, wq_context, log_wq_sz, attr.log_wq_size);
	MLX5_SET(wq, wq_context, log_wq_stride, CB_DEVX_RQE_BB_LOG);
	MLX5_SET(wq, wq_context, end_padding_mode, MLX5_WQ_END_PAD_MODE_ALIGN);

	MLX5_SET(wq, wq_context, wq_umem_id, attr.wq_umem_id);
	MLX5_SET64(wq, wq_context, wq_umem_offset, attr.wq_umem_offset);
	MLX5_SET(wq, wq_context, wq_umem_valid, 1);

	MLX5_SET(wq, wq_context, dbr_umem_id, attr.wq_umem_id);
	MLX5_SET64(wq, wq_context, dbr_addr, attr.db_umem_offset);
	MLX5_SET(wq, wq_context, dbr_umem_valid, 1);

	return DOCA_SUCCESS;
}

static doca_error_t
get_receive_memory_pool_id(create_rmp_out &out, uint32_t &id) noexcept
{
	auto status = MLX5_GET(create_rmp_out, out, status);
	if (status != 0) {
		auto syndrome = MLX5_GET(create_rmp_out, out, syndrome);
		(void)syndrome;

		DOCA_DLOG_ERR("Failed to create rmp object, status = %u, syndrome = %u", status, syndrome);

		return DOCA_ERROR_DRIVER;
	}

	id = MLX5_GET(create_rmp_out, out, rmpn);

	return DOCA_SUCCESS;
}

ireceive_memory_pool::create_ret
receive_memory_pool::create(std::shared_ptr<idevx_adapter> adapter,
	std::shared_ptr<iprotection_domain> pd, uint32_t wq_size) noexcept
{
	std::shared_ptr<receive_memory_pool> rmp_obj{nullptr};
	try {
		rmp_obj = std::make_shared<receive_memory_pool>();
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create receive memory pool object");
		return {DOCA_ERROR_NO_MEMORY, std::shared_ptr<receive_memory_pool>(nullptr)};
	}

	auto status = rmp_obj->initialize(adapter, pd, wq_size);
	if (status != DOCA_SUCCESS)
		rmp_obj.reset();

	return {status, rmp_obj};
}

receive_memory_pool::~receive_memory_pool()
{
	if (destroy() != DOCA_SUCCESS)
		DOCA_DLOG_ERR("Failed to destroy receive memory pool");
}

doca_error_t
receive_memory_pool::initialize(std::shared_ptr<idevx_adapter> adapter,
	std::shared_ptr<iprotection_domain> pd, uint32_t wq_size) noexcept
{
	if (adapter == nullptr) {
		DOCA_DLOG_ERR("Invalid devx adapter handle");
		return DOCA_ERROR_INVALID_VALUE;
	}
	m_devx_adapter = adapter;

	if (pd == nullptr) {
		DOCA_DLOG_ERR("Invalid protection domain handle");
		return DOCA_ERROR_INVALID_VALUE;
	}
	m_pd = pd;

	if (core_utils::is_power_of_two(wq_size) == false) {
		DOCA_DLOG_ERR("Invalid wq_size = %u", wq_size);
		return DOCA_ERROR_INVALID_VALUE;
	}
	m_wq_size = wq_size;

	auto status = create_umem();
	if (status != DOCA_SUCCESS)
		return status;

	receive_memory_pool_attributes attr = {0};

	attr.pdn = m_pd->get_pd_id();
	attr.wq_umem_id = m_umem->get_id();
	attr.wq_umem_offset = m_wq_umem_offset;
	attr.db_umem_id = m_umem->get_id();
	attr.db_umem_offset = m_db_umem_offset;
	attr.log_wq_size = os_abstraction::priv_doca_bit_scan_reverse(m_wq_size - 1);

	create_rmp_in in{0};
	create_rmp_out out{0};

	status = fill_create_rmp_cmd(in, attr);
	if (status != DOCA_SUCCESS)
		return status;

	std::unique_ptr<idevx_object> devx_object;
	std::tie(status, devx_object) = m_devx_adapter->create_devx_object(in, sizeof(in), out, sizeof(out));
	if (status != DOCA_SUCCESS)
		return status;

	status = get_receive_memory_pool_id(out, m_id);
	if (status != DOCA_SUCCESS)
		return status;

	m_rmp = std::move(devx_object);

	return DOCA_SUCCESS;
}

doca_error_t
receive_memory_pool::destroy() noexcept
{
	if (m_rmp != nullptr) {
		auto status = m_rmp->destroy();
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to destroy receive memory pool, status = %u", status);
			return status;
		}
		m_rmp = nullptr;
	}

	return DOCA_SUCCESS;
}

doca_error_t
receive_memory_pool::create_umem()
{
	/**
	 * Allocate memory buffer for WQEs and doorbell record.
	 * RQ must be aligned to 64B even if RQ stride is smaller than 64B
	 */
	size_t umem_size = MLX5_ALIGN_UP_BY(m_wq_size * CB_DEVX_RQE_BB, CB_DEVX_RQ_ALIGN);
	m_db_umem_offset = MLX5_ALIGN_UP_BY(umem_size, MLX5_DBR_SIZE);

	size_t alignment = os_abstraction::priv_doca_get_page_size();
	umem_size = MLX5_ALIGN_UP_BY(m_db_umem_offset + MLX5_DBR_SIZE, alignment);

	std::unique_ptr<os_abstraction::aligned_buffer<uint8_t>> umem_buf{nullptr};
	try {
		umem_buf = std::make_unique<os_abstraction::aligned_buffer<uint8_t>>(alignment, umem_size);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create aligned memory buffer");
		return DOCA_ERROR_NO_MEMORY;
	}
	m_umem_buf = std::move(umem_buf);

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<imapped_user_memory> umem_obj;
	std::tie(status, umem_obj) = m_devx_adapter->safe_create_mapped_user_memory(m_umem_buf->data(), m_umem_buf->size(), alignment);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to register umem for rmp");
		return status;
	}
	m_umem = umem_obj;

	uint32_t *db_address = get_wq_doorbell();
	*db_address = 0;

	return DOCA_SUCCESS;
}

uint32_t
receive_memory_pool::get_id() const noexcept
{
	return m_id;
}

uint32_t
receive_memory_pool::get_wq_size() const noexcept
{
	return m_wq_size;
}

void *
receive_memory_pool::get_wq_buffer() const noexcept
{
	uintptr_t umem_buffer = reinterpret_cast<uintptr_t>(m_umem->get_buffer());
	return reinterpret_cast<void *>(umem_buffer + m_wq_umem_offset);
}

uint32_t *
receive_memory_pool::get_wq_doorbell() const noexcept
{
	uintptr_t umem_buffer = reinterpret_cast<uintptr_t>(m_umem->get_buffer());
	return reinterpret_cast<uint32_t *>(umem_buffer + m_db_umem_offset);
}
