/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <memory>

#include <doca_error.h>
#include <idevx_adapter.hpp>
#include <iprotection_domain.hpp>
#include <imapped_user_memory.hpp>
#include <priv_doca_os_abstraction.hpp>

#include "ireceive_memory_pool.hpp"

namespace device_abstraction {

class receive_memory_pool : public ireceive_memory_pool
{
public:

	/**
	 * @brief This method allocates and initializes a receive memory pool
	 * @param [in] adapter devx adapter handle
	 * @param [in] pd protection domain
	 * @param [in] wq_size work queue size in entries, must be a power of 2
	 * @return ireceive_memory_pool::create_ret (see interface)
	 */
	static ireceive_memory_pool::create_ret
	create(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iprotection_domain> pd, uint32_t wq_size) noexcept;

	/**
	 * Constructor
	 */
	receive_memory_pool() noexcept = default;

	/**
	 * Destructor
	 */
	virtual ~receive_memory_pool();

	/**
	 * @brief This method initializes the receive memory pool
	 * @param [in] adapter devx adapter handle
	 * @param [in] pd protection domain
	 * @param [in] wq_size work queue size in entries, must be a power of 2
	 * @return doca_error_t
	 */
	doca_error_t
	initialize(std::shared_ptr<idevx_adapter> adapter, std::shared_ptr<iprotection_domain> pd, uint32_t wq_size) noexcept;

	/**
	 * @brief This method explicitly destroys the rmp.
	 * Destruction may fail if the rmp is still in use.
	 * @return doca_error
	 */
	virtual doca_error_t destroy() noexcept override;

	/**
	 * @brief This method returns the rmp id
	 * @return uint32_t
	 */
	virtual uint32_t get_id() const noexcept override;

	/**
	 * @brief This method returns the rmp wq size in entries
	 * @return uint32_t
	 */
	virtual uint32_t get_wq_size() const noexcept override;

	/**
	 * @brief This method returns the rmp wq buffer
	 * @return void *
	 */
	virtual void *get_wq_buffer() const noexcept override;

	/**
	 * @brief This method returns the rmp wq doorbell
	 * @return uint32_t *
	 */
	virtual uint32_t *get_wq_doorbell() const noexcept override;

	receive_memory_pool(receive_memory_pool const &) = delete;
	receive_memory_pool &operator=(receive_memory_pool const &) = delete;
	receive_memory_pool(receive_memory_pool &&) = delete;
	receive_memory_pool &operator=(receive_memory_pool &&) = delete;

private:

	/**
	 * @brief This method allocates buffer and creates umem for wq and dbr.
	 * @return doca_error
	 */
	doca_error_t create_umem();

	std::shared_ptr<idevx_adapter> m_devx_adapter{nullptr};
	std::shared_ptr<iprotection_domain> m_pd{nullptr};
	std::shared_ptr<imapped_user_memory> m_umem{nullptr};

	std::unique_ptr<os_abstraction::aligned_buffer<uint8_t>> m_umem_buf{nullptr};
	uint64_t m_wq_umem_offset{0};
	uint64_t m_db_umem_offset{0};

	std::unique_ptr<idevx_object> m_rmp{nullptr};
	uint32_t m_id{0};
	uint32_t m_wq_size{0};
};

} /* namespace device_abstraction */
