/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <priv_doca_os_abstraction_pagesize.hpp>
#include <priv_doca_os_abstraction_memory.hpp>
#include <indirect_memory_region.hpp>
#include <iprotection_domain.hpp>
#include <memory_access_flags.hpp>

#include <doca_log.h>
#include <multi_os_prm.h>
#include <priv_doca_os_abstraction_endian.h>

#include "user_memory_registration.hpp"

using namespace device_abstraction;

DOCA_LOG_REGISTER(CORE::DEVICE_ABSTRACTION);

namespace {

static constexpr uint32_t max_klm_segments = 128;

static constexpr size_t max_umr_inline_segments =  (MLX5_SEND_WQE_BB / sizeof(struct mlx5_wqe_umr_klm_seg));

struct umr_wqe {
	/* 1st WQEBB */
	struct mlx5_wqe_ctrl_seg general_ctrl_seg;
	struct mlx5_wqe_umr_ctrl_seg umr_ctrl_seg;
	/* 2nd WQEBB */
	struct mlx5_wqe_mkey_context_seg_member mkey_ctx_seg;
	/* 3rd WQEBB */
	struct mlx5_wqe_umr_klm_seg klm_arr[max_umr_inline_segments];
};

static constexpr size_t umr_wqe_size_bytes = sizeof(struct umr_wqe);
static constexpr size_t wqe_ds_bytes = 16;
static constexpr uint16_t klm_segments_alignment = 4;

} /* namespace */


create_umr_ret
user_memory_registration::create(std::shared_ptr<iprotection_domain> pd, uint32_t num_operations) noexcept
{
	try {
		auto umr_obj = std::make_unique<user_memory_registration>();
		return {umr_obj->initialize(pd, num_operations), std::move(umr_obj)};
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create indirect memory region");
		return {DOCA_ERROR_NO_MEMORY, std::unique_ptr<iuser_memory_registration>(nullptr)};
	}
}

user_memory_registration::~user_memory_registration()
{
	static_cast<void>(destroy());
}

doca_error_t
user_memory_registration::initialize(std::shared_ptr<iprotection_domain> pd, uint32_t num_operations) noexcept
{
	try {
		m_umr_pool = std::make_unique<typed_generic_pool<struct umr_mkey>>(num_operations);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to create umr instance. num_operations=%u", num_operations);
		return DOCA_ERROR_NO_MEMORY;
	}

	auto ret = m_umr_pool->iterate_over_elements([&](struct umr_mkey &umr){
		return umr_mkey_init(umr, pd);
	});
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to initialize all UMR mkeys.");
		destroy();
		return ret;
	}

	return DOCA_SUCCESS;
}

doca_error_t
user_memory_registration::destroy() noexcept
{
	if (m_umr_pool == nullptr)
		return DOCA_SUCCESS;

	const auto num_allocated = m_umr_pool->num_allocated_elements();
	if (num_allocated != 0) {
		DOCA_DLOG_ERR("Attempting to destroy UMR without freeing all allocated UMR keys. num_allocated=%u",
			      num_allocated);
		return DOCA_ERROR_IN_USE;
	}

	auto ret = m_umr_pool->iterate_over_elements([&](struct umr_mkey &umr){
		return umr_mkey_destroy(umr);
	});
	if (ret != DOCA_SUCCESS)
		return ret;

	m_umr_pool = nullptr;
	return DOCA_SUCCESS;
}

iuser_memory_registration::mkey_array*
user_memory_registration::umr_start(uint32_t *max_num_mkeys) noexcept
{
	*max_num_mkeys = max_klm_segments;
	/**
	 * Return pointer to first element in the umr_mkey structure. This will be used as a handle
	 * In commit and in done
	 */
	return static_cast<mkey_array *>(&m_umr_pool->acquire()->klm_array.segments);
}

void
user_memory_registration::umr_commit(mkey_array *mkeys, uint32_t num_mkeys, uint32_t total_len, void *wqe,
				     uint16_t pi, uint32_t qp_num, struct mkey *umr_mkey) noexcept
{
	auto umr = reinterpret_cast<struct umr_mkey*>(mkeys);

	auto umr_wqe = reinterpret_cast<struct umr_wqe*>(wqe);

	memset(umr_wqe, 0, umr_wqe_size_bytes);

	auto ctrl = &umr_wqe->general_ctrl_seg;

	if (num_mkeys == 1) {
		/**
		 * In case no UMR is required fill 3 WQEBB as NOP.
		 * This is to ensure each Job consumes exactly 4 WQEBB, to make calculation of the Job ID on
		 * received CQE, possible without additional metadata.
		 */
		mlx5dv_set_ctrl_seg(
			ctrl,					/* struct mlx5_wqe_ctrl_seg *seg */
			pi,					/* uint16_t pi */
			MLX5_OPCODE_NOP,			/* uint8_t opcode */
			0,					/* uint8_t opmod */
			qp_num,					/* uint32_t qp_num */
			0,					/* uint8_t fm_ce_se */
			umr_wqe_size_bytes / wqe_ds_bytes,	/* 12 ds: umr_wqe */
			0,					/* uint8_t signature */
			0);					/* uint32_t imm */

		/* Return the only mkey provided by user no need for UMR.*/
		*umr_mkey = umr->klm_array.segments[0];
		return;
	}

	/* Gather list provided fill UMR WQE. */

	const uint16_t num_klm_aligned = MLX5_ALIGN_UP_BY(num_mkeys, klm_segments_alignment);
	memset(&umr->klm_array.segments[num_mkeys], 0, (num_klm_aligned - num_mkeys) * sizeof(*umr->klm_array.segments));
	/* UMR WQE without the klm array, since array is provided as a pointer rather than inline. 9 DS*/
	static const uint8_t umr_ctrl_ds = 9;

	mlx5dv_set_ctrl_seg(
		ctrl,				 /* struct mlx5_wqe_ctrl_seg *seg */
		pi,				 /* uint16_t pi */
		MLX5_OPCODE_UMR,		 /* uint8_t opcode */
		0,				 /* uint8_t opmod */
		qp_num,				 /* uint32_t qp_num */
		0,				 /* uint8_t fm_ce_se */
		umr_ctrl_ds,			 /* 9 ds: umr_wqe - klm_array */
		0,				 /* uint8_t signature */
		htobe32(umr->indirect_mkey.id)); /* uint32_t imm */


	/* Set UMR WQE control seg. */
	auto umr_ctrl = &umr_wqe->umr_ctrl_seg;

	umr_ctrl->mkey_mask = htobe64(MLX5_WQE_UMR_CTRL_MKEY_MASK_LEN |
				      MLX5_WQE_UMR_CTRL_FLAG_TRNSLATION_OFFSET |
				      MLX5_WQE_UMR_CTRL_MKEY_MASK_ACCESS_LOCAL_WRITE);
	umr_ctrl->klm_octowords = htobe16(num_klm_aligned);

	/* Set mkey context seg. */
	auto mkey_ctx = reinterpret_cast<struct mlx5_wqe_mkey_context_seg*>(&umr_wqe->mkey_ctx_seg);

	mkey_ctx->len = htobe64(total_len);
	mkey_ctx->qpn_mkey = htobe32(0xffffff00 | (umr->indirect_mkey.id & 0xff));

	/* Set klm pointer to address of SW klm array. */
	auto klm = umr_wqe->klm_arr;

	klm->address = htobe64((uintptr_t)((uint8_t *)umr->klm_array.segments));
	klm->mkey = htobe32(umr->klm_array.id);
	klm->byte_count = htobe32(num_klm_aligned);

	/* Return the indirect mkey for user to use on next WQE.*/
	umr_mkey->id = htobe32(umr->indirect_mkey.id);
	umr_mkey->address = 0;
	umr_mkey->byte_count = htobe32(total_len);
}

void
user_memory_registration::umr_done(mkey_array *mkeys) noexcept
{
	auto umr = reinterpret_cast<struct umr_mkey*>(mkeys);

	m_umr_pool->release(umr);
}

doca_error_t
user_memory_registration::umr_mkey_init(struct umr_mkey &umr, std::shared_ptr<iprotection_domain> pd) noexcept
{
	auto indirec_ret_obj = pd->create_indirect_memory_region(max_klm_segments);
	if (indirec_ret_obj.first != DOCA_SUCCESS)
		return indirec_ret_obj.first;
	umr.indirect_mkey.obj = indirec_ret_obj.second;
	umr.indirect_mkey.id = indirec_ret_obj.second->get_id();

	const size_t klm_array_size_bytes = max_klm_segments * sizeof(struct mkey);
	umr.klm_array.segments = static_cast<struct mkey*>(priv_doca_allocate_aligned_buffer(klm_array_size_bytes,
											     klm_array_size_bytes));
	if (umr.klm_array.segments == nullptr) {
		DOCA_DLOG_ERR("Failed to allocate klm array for UMR mkey.");
		return DOCA_ERROR_NO_MEMORY;
	}

	auto mapped_ret_obj = pd->safe_create_mapped_memory_region(umr.klm_array.segments,
								   klm_array_size_bytes,
								   os_abstraction::priv_doca_get_page_size(),
								   memory_access_flags::local_write);
	if (mapped_ret_obj.first != DOCA_SUCCESS) {
		priv_doca_free_aligned_buffer(umr.klm_array.segments);
		umr.klm_array.segments = nullptr;
		return mapped_ret_obj.first;
	}
	umr.klm_array.mkey_obj = mapped_ret_obj.second;
	umr.klm_array.id = mapped_ret_obj.second->get_id();

	return DOCA_SUCCESS;
}

doca_error_t
user_memory_registration::umr_mkey_destroy(struct umr_mkey &umr) noexcept
{
	if (umr.klm_array.mkey_obj != nullptr) {
		auto ret = umr.klm_array.mkey_obj->destroy_key();
		if (ret != DOCA_SUCCESS)
			return ret;
	}

	priv_doca_free_aligned_buffer(umr.klm_array.segments);

	if (umr.indirect_mkey.obj != nullptr) {
		auto ret = umr.indirect_mkey.obj->destroy_key();
		if (ret != DOCA_SUCCESS)
			return ret;
	}
	return DOCA_SUCCESS;
}
