/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <memory>

#include <typed_generic_pool.hpp>

#include <iuser_memory_registration.hpp>

namespace device_abstraction {

class iprotection_domain;
class imemory_region;

class user_memory_registration : public iuser_memory_registration
{
public:

	/**
	 * @brief allocate and initialize a UMR instance.
	 *
	 * @param [in] pd Protection domain, same as QP using the UMR.
	 * @param [in] num_operations Maximum number of UMR keys to be used at same time.
	 *
	 * @return doca_error_t
	 * DOCA_SUCCESS - successfully created UMR.
	 * DOCA_ERROR_NO_MEMORY - failed to allocate memory.
	 * DOCA_ERROR_DRIVER - failed to allocate an mkey DevX command failed.
	 */
	static create_umr_ret create(std::shared_ptr<iprotection_domain> pd, uint32_t num_operations) noexcept;

	user_memory_registration() noexcept = default;

	virtual ~user_memory_registration();

	/**
	 * @brief initialize a UMR instance.
	 *
	 * @param [in] pd Protection domain, same as QP using the UMR.
	 * @param [in] num_operations Maximum number of UMR keys to be used at same time.
	 *
	 * @return doca_error_t
	 * DOCA_SUCCESS - successfully created UMR.
	 * DOCA_ERROR_NO_MEMORY - failed to allocate memory.
	 * DOCA_ERROR_DRIVER - failed to allocate an mkey DevX command failed.
	 */
	virtual doca_error_t initialize(std::shared_ptr<iprotection_domain> pd, uint32_t num_operations) noexcept;

	virtual doca_error_t destroy() noexcept override;

	virtual mkey_array *umr_start(uint32_t *max_num_mkeys) noexcept override;

	virtual void umr_commit(mkey_array *mkeys, uint32_t num_mkeys, uint32_t total_len, void *wqe,
				uint16_t pi, uint32_t qp_num, struct mkey *umr_mkey) noexcept override;

	virtual void umr_done(mkey_array *mkeys) noexcept override;

	user_memory_registration(user_memory_registration const &) = delete;
	user_memory_registration &operator=(user_memory_registration const &) = delete;
	user_memory_registration(user_memory_registration &&) = delete;
	user_memory_registration &operator=(user_memory_registration &&) = delete;

protected:

	struct umr_mkey {
		umr_mkey() = default;
		umr_mkey(const umr_mkey&) = delete;
		umr_mkey(umr_mkey&& other) noexcept = default;

		struct {
			struct mkey *segments{nullptr};
			std::shared_ptr<imemory_region> mkey_obj{nullptr};
			uint32_t id{0};
		} klm_array;
		struct {
			std::shared_ptr<imemory_region> obj{nullptr};
			uint32_t id{0};
		} indirect_mkey;
	};

	doca_error_t umr_mkey_init(struct umr_mkey &umr, std::shared_ptr<iprotection_domain> pd) noexcept;

	doca_error_t umr_mkey_destroy(struct umr_mkey &umr) noexcept;

	std::unique_ptr<typed_generic_pool<struct umr_mkey>> m_umr_pool{nullptr};
};

} /* namespace device_abstraction */

