/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>

#include <doca_log.h>
#include <priv_doca_buf.h>
#include <private/common_defines.h>
#include <imemory_region.hpp>
#include "internal_doca_buf.h"
#include "doca_buf_inventory.hpp"
#include "internal_doca_dev.h"
#include "memory_access_flags.hpp"

DOCA_LOG_REGISTER(CORE::DOCA_BUF);

doca_error_t
doca_buf_refcount_add(struct doca_buf * /*buf*/, uint16_t * /*refcount*/)
{
	return DOCA_ERROR_NOT_SUPPORTED;
}

static void
buf_refcount_rm_in_list(struct doca_buf *buf)
{
	struct doca_buf *current_buf = buf;
	struct doca_buf *next_buf;

	while (current_buf != NULL) {
		next_buf = get_linked_list_extension(current_buf)->next;

		/*
		 * Right now doca_buf can have only refcnt == 1 when adding the function doca_buf_refcount_add
		 * will need to adjust the function accordingly.
		 */
		if (current_buf->refcnt == 1)
			current_buf->resources->inventory->release_doca_buffer(current_buf);

		current_buf = next_buf;
	}
}

doca_error_t
doca_buf_refcount_rm(struct doca_buf *buf, uint16_t *refcount)
{
	if (buf == NULL) {
		DOCA_DLOG_ERR("Unable to remove reference from buffer: Received invalid input - buf=%p, refcount=%p",
			       buf, refcount);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto linked_list_extension = get_linked_list_extension(buf);
	if ((linked_list_extension != nullptr) && (buf->refcnt == 1) && (!linked_list_extension->is_head)) {
		DOCA_DLOG_ERR("Cannot delete a buffer that isn't at the beginning of a linked list");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	if (refcount != NULL)
		*refcount = buf->refcnt;

	--(buf->refcnt);

	if (buf->refcnt == 0) {
		if (linked_list_extension && (linked_list_extension->next != nullptr))
			buf_refcount_rm_in_list(linked_list_extension->next);
		buf->resources->inventory->release_doca_buffer(buf);
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_get_refcount(struct doca_buf *buf, uint16_t *refcount)
{
	if (buf == NULL || refcount == NULL) {
		DOCA_DLOG_ERR("Unable to get reference from buffer: Received invalid input - buf=%p, refcount=%p",
			       buf, refcount);
		return DOCA_ERROR_INVALID_VALUE;
	}
	*refcount = buf->refcnt;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_get_len(struct doca_buf *buf, size_t *len)
{
	if (buf == nullptr || len == nullptr) {
		DOCA_DLOG_ERR("Unable to get buffer length: Received invalid input - buf=%p, len=%p",
			       buf, len);
		return DOCA_ERROR_INVALID_VALUE;
	}
	*len = buf->len;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_get_head(struct doca_buf *buf, void **head)
{
	if (buf == nullptr || head == nullptr) {
		DOCA_DLOG_ERR("Unable to get buffer address: Received invalid input - buf=%p, head=%p",
			       buf, head);
		return DOCA_ERROR_INVALID_VALUE;
	}
	*head = buf->head;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_get_data_len(struct doca_buf *buf, size_t *data_len)
{
	if (buf == nullptr || data_len == nullptr) {
		DOCA_DLOG_ERR("Unable to get buffer data length: Received invalid input - buf=%p, data_len=%p",
			       buf, data_len);
		return DOCA_ERROR_INVALID_VALUE;
	}
	*data_len = buf->data_len;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_get_data(struct doca_buf *buf, void **data)
{
	if (buf == nullptr || data == nullptr) {
		DOCA_DLOG_ERR("Unable to get buffer data: Received invalid input - buf=%p, data=%p",
			       buf, data);
		return DOCA_ERROR_INVALID_VALUE;
	}
	*data = buf->data;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_set_data(struct doca_buf *buf, void *data, size_t data_len)
{
	if (buf == nullptr || data == nullptr) {
		DOCA_DLOG_ERR("Unable to set buffer data: Received invalid input - buf=%p, data=%p",
			       buf, data);
		return DOCA_ERROR_INVALID_VALUE;
	}
	uint8_t *cast_data = reinterpret_cast<uint8_t *>(data);

	if ((cast_data < buf->head) || (cast_data > (buf->head + buf->len)) ||
	     data_len > (buf->len - (cast_data - buf->head))) {
		DOCA_DLOG_ERR("Unable to set buffer data: Data address and length are outside buffer's limits");
		return DOCA_ERROR_INVALID_VALUE;
	}
	buf->data = cast_data;
	buf->data_len = data_len;
	return DOCA_SUCCESS;
}

void *
priv_doca_buf_get_addr(const struct doca_buf *buf)
{
	// It is likely that buf is not null, therefore the optimistic condition.
	if (likely(buf != NULL))
		return buf->head;
	return NULL;
}

size_t
priv_doca_buf_get_len(const struct doca_buf *buf)
{
	// It is likely that buf is not null, therefore the optimistic condition.
	if (likely(buf != NULL))
		return buf->len;
	return 0;
}

void *
priv_doca_buf_get_data(const struct doca_buf *buf)
{
	// It is likely that buf is not null, therefore the optimistic condition.
	if (likely(buf != NULL))
		return buf->data;
	return NULL;
}

size_t
priv_doca_buf_get_data_len(const struct doca_buf *buf)
{
	// It is likely that buf is not null, therefore the optimistic condition.
	if (likely(buf != NULL))
		return buf->data_len;
	return 0;
}

static std::pair<doca_error_t, std::shared_ptr<device_abstraction::imemory_region>>
allocate_zero_based_mkey(struct doca_buf *buf, const struct doca_dev *dev)
{
	// Once Rivermax change their library to use doca_buf data, we need to change here to data & data_len
	return dev->pd->safe_create_zero_based_mapped_memory_region(
		buf->head, buf->len, buf->mchunk->pg_sz, device_abstraction::memory_access_flags::local_write,
		device_abstraction::memory_access_flags::remote_write,
		device_abstraction::memory_access_flags::remote_read);
}

static std::pair<doca_error_t, std::shared_ptr<device_abstraction::imemory_region>>
get_zero_based_mkey(struct doca_buf *buf, const struct doca_dev *dev)
{
	auto mmap = buf->mchunk->mmap;
	auto buf_it = mmap->zba_mkey_map.find(buf);
	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<device_abstraction::imemory_region> mr;

	if (buf_it != mmap->zba_mkey_map.end()) {
		auto zba_key_it = buf_it->second.find(dev);
		if (zba_key_it != buf_it->second.end()) {
			return {DOCA_SUCCESS, zba_key_it->second};
		} else {
			std::tie(status, mr) = allocate_zero_based_mkey(buf, dev);

			if (status == DOCA_SUCCESS) {
				try {
					buf_it->second.emplace(dev, mr);
				} catch (...) {
					DOCA_DLOG_ERR("Failed to add zero based mkey to map");
					return {DOCA_ERROR_NO_MEMORY,
						std::shared_ptr<device_abstraction::imemory_region>(nullptr)};
				}
			}

			return {status, mr};
		}
	} else {
		std::tie(status, mr) = allocate_zero_based_mkey(buf, dev);
		if (status == DOCA_SUCCESS) {
			try {
				mmap->zba_mkey_map.emplace(buf, zba_mkey_map_entry{{dev, mr}});
			} catch (...) {
				DOCA_DLOG_ERR("Failed to add zero based mkey to map");
				return {DOCA_ERROR_NO_MEMORY,
					std::shared_ptr<device_abstraction::imemory_region>(nullptr)};
			}
		}

		return {status, mr};
	}
}

doca_error_t
priv_doca_buf_zba_mkey_get(struct doca_buf *buf, const struct doca_dev *dev, uint32_t *zba_mkey)
{
	if (buf == NULL || dev == NULL || zba_mkey == NULL) {
		DOCA_DLOG_ERR("Unable to get zero-based mkey: Received invalid input - buf=%p, dev=%p, zba_mkey=%p",
			      buf, dev, zba_mkey);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto mmap = buf->mchunk->mmap;
	auto device_it = std::find(mmap->device_arr.begin(), mmap->device_arr.end(), dev);
	if (device_it == mmap->device_arr.end()) {
		DOCA_DLOG_ERR("Unable to get zero-based mkey: Device not registered with buffer");
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<device_abstraction::imemory_region> mr;
	std::tie(status, mr) = get_zero_based_mkey(buf, dev);
	if (status == DOCA_SUCCESS) {
		*zba_mkey = mr->get_id();
		return status;
	} else {
		*zba_mkey = 0;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_list_next(struct doca_buf *buf, struct doca_buf **next_buf)
{
	if (buf == NULL || next_buf == NULL) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto linked_list_extension = get_linked_list_extension(buf);
	if (!linked_list_extension) {
		*next_buf = nullptr;
		return DOCA_SUCCESS;
	}

	*next_buf = linked_list_extension->next;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_buf_list_next(const struct doca_buf *buf, const struct doca_buf **next_buf)
{
	return doca_buf_list_next(const_cast<doca_buf *>(buf), const_cast<doca_buf **>(next_buf));
}

doca_error_t
doca_buf_list_last(struct doca_buf *buf, struct doca_buf **last_buf)
{
	struct doca_buf *current = buf;

	if (buf == NULL || last_buf == NULL) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto linked_list_extension = get_linked_list_extension(buf);
	if (!linked_list_extension) {
		*last_buf = buf;
		return DOCA_SUCCESS;
	}

	while (get_linked_list_extension(current)->next != NULL)
		current = get_linked_list_extension(current)->next;

	*last_buf = current;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_buf_list_last(const struct doca_buf *buf, const struct doca_buf **last_buf)
{
	return doca_buf_list_last(const_cast<doca_buf *>(buf), const_cast<doca_buf **>(last_buf));
}

doca_error_t
doca_buf_list_is_last(const struct doca_buf *buf, bool *is_last)
{
	if (buf == NULL || is_last == NULL) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto linked_list_extension = get_linked_list_extension(buf);
	if (!linked_list_extension) {
		*is_last = true;
		return DOCA_SUCCESS;
	}

	*is_last = (linked_list_extension->next == NULL);
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_list_is_first(const struct doca_buf *buf, bool *is_first)
{
	if (buf == NULL || is_first == NULL) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto linked_list_extension = get_linked_list_extension(buf);
	if (!linked_list_extension) {
		*is_first = true;
		return DOCA_SUCCESS;
	}

	*is_first = linked_list_extension->is_head;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_list_num_elements(const struct doca_buf *buf, uint32_t *num_elements)
{
	const struct doca_buf *current = buf;
	uint32_t count = 1;

	if (buf == NULL || num_elements == NULL) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto linked_list_extension = get_linked_list_extension(buf);
	if (!linked_list_extension) {
		*num_elements = 1;
		return DOCA_SUCCESS;
	}

	if (!linked_list_extension->is_head) {
		DOCA_DLOG_ERR("The buffer must be the head of a list");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	while (get_linked_list_extension(current)->next != NULL) {
		count++;
		current = get_linked_list_extension(current)->next;
	}

	*num_elements = count;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_buf_num_elements_till_end_of_list(const struct doca_buf *buf, uint32_t *num_elements)
{
	const struct doca_buf *current = buf;
	uint32_t count = 1;

	if (buf == NULL || num_elements == NULL) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto linked_list_extension = get_linked_list_extension(buf);
	if (!linked_list_extension) {
		*num_elements = 1;
		return DOCA_SUCCESS;
	}

	while (get_linked_list_extension(current)->next != NULL) {
		count++;
		current = get_linked_list_extension(current)->next;
	}

	*num_elements = count;
	return DOCA_SUCCESS;
}


doca_error_t
doca_buf_list_chain(struct doca_buf *list1, struct doca_buf *list2)
{
	struct doca_buf *list_last_of_list1;
	doca_error_t ret;

	if (list1 == NULL || list2 == NULL || list1 == list2) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto list1_extension = get_linked_list_extension(list1);
	auto list2_extension = get_linked_list_extension(list2);
	if ((!list1_extension) || (!list2_extension)) {
		DOCA_DLOG_ERR("Linked list extension for doca_buf is disabled");
		return DOCA_ERROR_NOT_PERMITTED;
	}
	if ((!list1_extension->is_head) || (!list2_extension->is_head)) {
		DOCA_DLOG_ERR("The buffers to chain must be the head of a list");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	ret = doca_buf_list_last(list1, &list_last_of_list1);
	if (ret != DOCA_SUCCESS)
		return ret;

	get_linked_list_extension(list_last_of_list1)->next = list2;
	get_linked_list_extension(list2)->is_head = false;

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_list_unchain(struct doca_buf *list1, struct doca_buf *list2)
{
	if (list1 == NULL || list2 == NULL || list1 == list2) {
		DOCA_DLOG_ERR("Received invalid input");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto list1_extension = get_linked_list_extension(list1);
	auto list2_extension = get_linked_list_extension(list2);
	if ((!list1_extension) || (!list2_extension)) {
		DOCA_DLOG_ERR("Linked list extension for doca_buf is disabled");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	if (list2_extension->is_head) {
		DOCA_DLOG_ERR("List 2 is head, can't unchain");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto *current = list1;
	while (get_linked_list_extension(current)->next != NULL) {
		if (get_linked_list_extension(current)->next == list2) {
			get_linked_list_extension(current)->next = NULL;
			get_linked_list_extension(list2)->is_head = true;
			return DOCA_SUCCESS;
		}

		current = get_linked_list_extension(current)->next;
	}

	DOCA_DLOG_ERR("List 2 is not a part of list 1, can't unchain");
	return DOCA_ERROR_INVALID_VALUE;
}

doca_error_t
doca_buf_get_lkey(const struct doca_buf *buf, struct doca_dev *dev, uint32_t doca_access_flags, uint32_t *lkey)
{
	if (unlikely(buf == NULL || dev == NULL || lkey == NULL)) {
		DOCA_DLOG_ERR("Invalid input: buf=%p, dev=%p, lkey=%p", buf, dev, lkey);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return priv_doca_mchunk_extract_mkey_by_device(buf->mchunk, dev, doca_access_flags, lkey);
}
