/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string>
#include <memory>

#include <doca_log.h>
#include <private/common_defines.h>

#include "doca_buf_inventory.hpp"
#include "buf_extensions_map.hpp"
#include "buf_extensions.hpp"

using namespace doca_core;

DOCA_LOG_REGISTER(CORE::DOCA_BUF_INVENTORY);

static void
init_linked_list_extension(struct doca_buf_linked_list_extension &extension, bool is_head)
{
	extension.is_head = is_head;
	extension.next = nullptr;
}

static void
init_linked_list_extension(struct doca_buf *buf, size_t extension_offset, bool is_head)
{
	init_linked_list_extension(*get_linked_list_extension(buf, extension_offset), is_head);
}


doca_buf_inventory::doca_buf_inventory(const union doca_data *user_data, uint32_t num_elements, uint32_t extensions)
{
	// Assumption: Extensions are valid.
	auto required_extensions_size = calculate_extensions_buffer_size(extensions);
	m_buf_pool = std::make_unique<buf_pool>(num_elements, required_extensions_size);
	m_extensions_map = std::make_unique<doca_core::buf_extensions_map>(extensions, get_supported_extensions());

	if (user_data != NULL)
		m_user_data = *user_data;
	else
		memset(&m_user_data, 0, sizeof(m_user_data));

	m_buffer_resources.extensions_map = m_extensions_map.get();
	m_buffer_resources.inventory = this;
}

bool
doca_buf_inventory::can_destroy() const noexcept
{
	if (m_buf_pool->num_allocated_elements() > 0) {
		DOCA_DLOG_ERR("Unable to destroy inventory: %d Buffers from the inventory are still in use",
			      m_buf_pool->num_allocated_elements());
		return false;
	}

	return true;
}

void
doca_buf_inventory::release_doca_buffer(struct doca_buf *buf) noexcept
{
	priv_doca_mmap_decrease_doca_buf_num(buf->mchunk->mmap);
	buf->mchunk->doca_buf_cnt--;
	buf->mchunk->mmap->zba_mkey_map.erase(buf);
	m_buf_pool->release(buf);
}

doca_error_t
doca_buf_inventory_create(const union doca_data *user_data, size_t num_elements, uint32_t extensions,
			  struct doca_buf_inventory **buf_inventory)
{
	if (num_elements == 0 || !is_buf_extensions_valid(extensions) || buf_inventory == nullptr) {
		DOCA_DLOG_ERR(
		     "Unable to create inventory: Received invalid input - num_elements=%zu, extensions=%u, buf_inventory=%p",
		     num_elements, extensions, buf_inventory);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (num_elements > UINT32_MAX) {
		DOCA_DLOG_ERR(
		   "Unable to create inventory: Received invalid num elements - Number of elements is greater than %u",
		   UINT32_MAX);
		return DOCA_ERROR_INVALID_VALUE;
	}

	// The section below allocates the inventory and since it is wrapped with try catch there is no need to
	// check for null (failure to allocate will throw an exception).
	try {
		*buf_inventory = new doca_buf_inventory(user_data, static_cast<uint32_t>(num_elements), extensions);
	} catch (...) {
		DOCA_DLOG_ERR("Unable to create inventory: Failed to allocate memory for buffer inventory");
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_inventory_destroy(struct doca_buf_inventory *inventory)
{
	if (unlikely(inventory == nullptr)) {
		DOCA_DLOG_ERR("Unable to destroy inventory: Received invalid input - buf_inventory is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (!inventory->can_destroy())
		return DOCA_ERROR_NOT_PERMITTED;

	delete inventory;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_inventory::get_num_elements(uint32_t *num_of_elements) const
{
	if (unlikely(num_of_elements == NULL)) {
		DOCA_DLOG_ERR(": Unable to get inventory number of elements: Received invalid input - value=%p",
			      num_of_elements);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*(num_of_elements) = m_buf_pool->pool_size();

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_inventory::get_num_free_elements(uint32_t *num_of_free_elements) const
{
	if (unlikely(num_of_free_elements == nullptr)) {
		DOCA_DLOG_ERR("Unable to get inventory number of free elements: Received invalid input - value=%p",
			      num_of_free_elements);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*(num_of_free_elements) = m_buf_pool->num_free_elements();

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_inventory::get_user_data(union doca_data *user_data) const
{
	if (unlikely(user_data == nullptr)) {
		DOCA_DLOG_ERR("%p: Unable to get inventory user_data: Received invalid input - value=%p", this,
			      user_data);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*user_data = m_user_data;

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_inventory::get_extensions(uint32_t *extensions) const
{
	if (unlikely(extensions == nullptr)) {
		DOCA_DLOG_ERR("Unable to get inventory extensions: Received invalid input - value=%p",
			      extensions);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*extensions = m_extensions_map->get_extensions();

	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_inventory_get_num_elements(struct doca_buf_inventory *inventory, uint32_t *num_of_elements)
{
	return inventory->get_num_elements(num_of_elements);
}

doca_error_t
doca_buf_inventory_get_num_free_elements(struct doca_buf_inventory *inventory, uint32_t *num_of_free_elements)
{
	return inventory->get_num_free_elements(num_of_free_elements);
}

doca_error_t
doca_buf_inventory_get_user_data(struct doca_buf_inventory *inventory, union doca_data *user_data)
{
	return inventory->get_user_data(user_data);
}

doca_error_t
doca_buf_inventory_get_list_supported(struct doca_buf_inventory *inventory, uint8_t *list_supported)
{
	if ((inventory == nullptr) || (list_supported == nullptr))
		return DOCA_ERROR_INVALID_VALUE;

	uint32_t extensions = 0;
	auto status = inventory->get_extensions(&extensions);
	if (status != DOCA_SUCCESS)
		return status;

	*list_supported = (extensions & DOCA_BUF_EXTENSION_LINKED_LIST) == DOCA_BUF_EXTENSION_LINKED_LIST;
	return DOCA_SUCCESS;
}

doca_error_t
doca_buf_inventory_start(struct doca_buf_inventory *inventory)
{
	if (unlikely(inventory == nullptr)) {
		DOCA_DLOG_ERR("Unable to start inventory: Received invalid input - inventory is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	inventory->start();

	return DOCA_SUCCESS;
}

void
doca_buf_inventory::start() noexcept
{
	m_state = state::started;
}

doca_error_t
doca_buf_inventory_stop(struct doca_buf_inventory *inventory)
{
	if (unlikely(inventory == nullptr)) {
		DOCA_DLOG_ERR("Unable to stop inventory: Received invalid input - inventory is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	inventory->stop();

	return DOCA_SUCCESS;
}

void
doca_buf_inventory::stop() noexcept
{
	// The inventory can't transition from not_started to stopped because stopped implies that the inventory was
	// previously started.
	if (likely(m_state != state::not_started)) {
		m_state = state::stopped;
	}
}

doca_error_t
doca_buf_inventory_buf_by_args(struct doca_buf_inventory *inventory, struct doca_mmap *mmap, void *addr, size_t len,
			       void *data, size_t data_len, struct doca_buf **buf)
{
	if (unlikely(buf == nullptr)) {
		DOCA_DLOG_ERR("Unable to retrieve buffer from inventory: Received invalid input - buf is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	// Setting buf to null here guarantees that the method will not output an uninitialized buf.
	*buf = nullptr;

	if (unlikely(inventory == nullptr)) {
		DOCA_DLOG_ERR(
			"Unable to retrieve buffer from inventory: Received invalid input - inventory=%p",
			inventory);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::tie(status, *buf) = inventory->allocate_buf_by_args(mmap, addr, len, data, data_len);
	return status;
}

std::pair<doca_error_t, struct doca_buf *>
doca_buf_inventory::allocate_buf_by_args(struct doca_mmap *mmap, void *addr, size_t len,
					 void *data, size_t data_len) noexcept
{
	if (unlikely(mmap == nullptr || addr == nullptr || data == nullptr)) {
		DOCA_DLOG_ERR(
			"Unable to retrieve buffer from inventory: Received invalid input - mmap=%p, addr=%p",
			mmap, addr);
		return {DOCA_ERROR_INVALID_VALUE, nullptr};
	}

	if (!priv_doca_mmap_is_started(mmap)) {
		DOCA_DLOG_ERR("Unable to retrieve buffer from inventory: Mmap is stopped");
		return {DOCA_ERROR_NOT_PERMITTED, nullptr};
	}

	if (m_state != state::started) {
		DOCA_DLOG_ERR("Unable to retrieve buffer from inventory: Inventory is stopped");
		return {DOCA_ERROR_NOT_PERMITTED, nullptr};
	}

	auto found_mchunk = priv_doca_mmap_is_buf_in_chunk(mmap, addr, len);
	if (found_mchunk == nullptr) {
		size_t ii;

		DOCA_DLOG_ERR(
			"Unable to retrieve buffer from inventory: There is no suitable memory range for the given address and len(%p, %zu)",
			addr, len);
		for (ii = 0; ii != mmap->curr_mchunk_num; ++ii) {
			DOCA_DLOG_DBG("Available range: [%p-%p) : length: %zu",
				       mmap->mchunk_arr[ii].addr,
				       (uint8_t *)mmap->mchunk_arr[ii].addr + mmap->mchunk_arr[ii].len,
				       mmap->mchunk_arr[ii].len);
		}
		return {DOCA_ERROR_INVALID_VALUE, nullptr};
	}

	uint8_t *cast_addr = reinterpret_cast<uint8_t *>(addr);
	uint8_t *cast_data = reinterpret_cast<uint8_t *>(data);

	if ((cast_data < cast_addr) || (cast_data > (cast_addr + len)) ||
	     data_len > (len - (cast_data - cast_addr))) {
		DOCA_DLOG_ERR("Unable to retrieve buffer from inventory: Data address and length are outside buffer's limits");
		return {DOCA_ERROR_INVALID_VALUE, nullptr};
	}

	auto new_buf = m_buf_pool->acquire();
	if (new_buf == nullptr) {
		DOCA_DLOG_ERR("Unable to retrieve buffer from inventory: Inventory is empty");
		return {DOCA_ERROR_NO_MEMORY, nullptr};
	}
	new_buf->head = cast_addr;
	new_buf->len = len;
	new_buf->data = cast_data;
	new_buf->data_len = data_len;
	new_buf->refcnt = 1;
	new_buf->mchunk = found_mchunk;
	new_buf->resources = &m_buffer_resources;

	auto linked_list_extension = get_linked_list_extension(new_buf);
	if (linked_list_extension != nullptr) {
		init_linked_list_extension(*linked_list_extension, true);
	}

	found_mchunk->doca_buf_cnt++;
	priv_doca_mmap_increase_doca_buf_num(mmap);

	return {DOCA_SUCCESS, new_buf};
}

doca_error_t
doca_buf_inventory_buf_dup(struct doca_buf_inventory *inventory, const struct doca_buf *src_buf,
			   struct doca_buf **dst_buf)
{
	if (unlikely(dst_buf == nullptr)) {
		DOCA_DLOG_ERR(
			"Unable to duplicate doca_buf: Received invalid input - dst_buf=%p", dst_buf);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*dst_buf = nullptr;

	if (unlikely(inventory == nullptr)) {
		DOCA_DLOG_ERR(
			"Unable to duplicate doca_buf: Received invalid input - inventory=%p", inventory);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::tie(status, *dst_buf) = inventory->duplicate_buffer(src_buf);
	return status;
}

std::pair<doca_error_t, struct doca_buf *>
doca_buf_inventory::duplicate_buffer(const struct doca_buf *src_buf) noexcept
{
	if (unlikely(src_buf == nullptr)) {
		DOCA_DLOG_ERR(
			"Unable to duplicate doca_buf: Received invalid input - src_buf=%p",
			src_buf);
		return {DOCA_ERROR_INVALID_VALUE, nullptr};
	}

	if (!priv_doca_mmap_is_started(src_buf->mchunk->mmap)) {
		DOCA_DLOG_ERR("Unable to duplicate doca_buf: Mmap is stopped");
		return {DOCA_ERROR_NOT_PERMITTED, nullptr};
	}

	if (m_state != state::started) {
		DOCA_DLOG_ERR("Unable to duplicate doca_buf: Inventory is stopped");
		return {DOCA_ERROR_NOT_PERMITTED, nullptr};
	}

	if (m_extensions_map->get_extensions() != src_buf->resources->extensions_map->get_extensions()) {
		DOCA_DLOG_ERR("Unable to duplicate doca_buf: Incompatible inventories' extensions");
		return {DOCA_ERROR_NOT_PERMITTED, nullptr};
	}

	auto linked_list_extension = get_linked_list_extension(src_buf);

	if (linked_list_extension != nullptr) {
		if (!linked_list_extension->is_head) {
			DOCA_DLOG_ERR("Cannot duplicate a buffer that is in the middle or the end of a linked list");
			return {DOCA_ERROR_NOT_PERMITTED, nullptr};
		}

		return duplicate_sg_list_doca_buf(src_buf);
	}

	return duplicate_single_doca_buf(src_buf);
}

std::pair<doca_error_t, struct doca_buf *>
doca_buf_inventory::duplicate_single_doca_buf(const struct doca_buf *src_buf)
{
	auto new_buf = m_buf_pool->acquire();
	if (new_buf == nullptr) {
		DOCA_DLOG_ERR("Unable to duplicate doca_buf: Inventory is empty");
		return {DOCA_ERROR_NO_MEMORY, nullptr};
	}

	priv_doca_mmap_increase_doca_buf_num(src_buf->mchunk->mmap);
	src_buf->mchunk->doca_buf_cnt++;

	*new_buf = *src_buf;
	new_buf->resources = &m_buffer_resources;

	// A doca buf may "point" to zero based memory keys in the mmap.
	// This section duplicates this pointing by creating a new entry in the zba mkey map and copying the content
	// of the source buffer entry to the new entry.
	auto zba_mkey_map = &src_buf->mchunk->mmap->zba_mkey_map;
	auto zba_entry = zba_mkey_map->find(src_buf);
	if (zba_entry != zba_mkey_map->end()) {
		try {
			zba_mkey_map->emplace(new_buf, zba_entry->second);
		} catch (...) {
			DOCA_DLOG_ERR("Unable to duplicate doca_buf: Failed to update zero based address map");
			return {DOCA_ERROR_NO_MEMORY, nullptr};
		}
	}
	new_buf->refcnt = 1;

	return {DOCA_SUCCESS, new_buf};
}

std::pair<doca_error_t, struct doca_buf *>
doca_buf_inventory::duplicate_sg_list_doca_buf(const struct doca_buf *src_buf)
{
	doca_error_t status{DOCA_SUCCESS};
	size_t offset{0};
	std::tie(status, offset) = m_extensions_map->get_extension_offset(doca_core::buf_extension_type::LINKED_LIST);

	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to get linked list extension offset with status %s",
			      doca_get_error_name(status));
		return {status, nullptr};
	}

	// First buffer can be duplicated as is.
	struct doca_buf *new_head{nullptr};
	std::tie(status, new_head) = duplicate_single_doca_buf(src_buf);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to duplicate doca_buf list");
		return {status, nullptr};
	}

	init_linked_list_extension(new_head, offset, true);

	auto current_dest = new_head;
	auto next_src = get_linked_list_extension(src_buf, offset)->next;
	while (next_src) {
		struct doca_buf *new_buf{nullptr};
		std::tie(status, new_buf) = duplicate_single_doca_buf(next_src);
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to duplicate the SG list, releasing allocated buffers");

			auto fold = new_head;
			while (fold) {
				auto fold_next = get_linked_list_extension(fold, offset)->next;
				release_doca_buffer(fold);
				fold = fold_next;
			}

			return {status, nullptr};
		}

		init_linked_list_extension(new_buf, offset, false);
		get_linked_list_extension(current_dest, offset)->next = new_buf;
		current_dest = new_buf;

		next_src = get_linked_list_extension(next_src, offset)->next;
	}

	return {DOCA_SUCCESS, new_head};
}
