/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string>
#include <memory>

#include <private/common_defines.h>
#include <doca_buf_inventory.h>
#include "internal_doca_buf.h"
#include "buf_extensions_map.hpp"
#include "buf_pool.hpp"

/**
 * doca_buf_inventory is declared in the API so it can't be declared as a class, but it is a class.
 */
struct doca_buf_inventory
{
public:
	/**
	 * @brief Constructor
	 * @param user_data [in] (optional) Inventory identifier provided by user
	 * @param num_elements [in] Number of elements that the inventory holds
	 * @param extensions_flags [in] flags that define which extensions should be supported by the inventory.
	 */
	doca_buf_inventory(const union doca_data *user_data, uint32_t num_elements, uint32_t extensions_flags);

	/**
	 * @briefDestructor
	 */
	~doca_buf_inventory() = default;

	/**
	 * @brief This method checks if the inventory can be destroyed
	 * @return true if can, false if can't.
	 */
	bool can_destroy() const noexcept;

	/**
	 * @brief This method starts the inventory
	 */
	void start() noexcept;

	/**
	 * @brief This method stops the inventory
	 */
	void stop() noexcept;

	/**
	 * @brief Allocate single element from buffer inventory and point it to the buffer
	 * defined by `addr`, `len`, `data` and `data_len` arguments.
	 *
	 * @param mmap [in] mmap that contains the memory that address points to
	 * @param addr [in] address of the memory to assign to the doca buffer
	 * @param len [in] buffer length
	 * @param data [in] the data placement inside the buffer.
	 * @param data_len [in] data length.
	 *
	 * @return pair of doca_error_t and doca_buf (upon success)
	 * DOCA_SUCCESS - in case of success.
	 * doca_error code - in case of failure:
	 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received
	 *                              or if there is no suitable memory range for the given address and length
	 *                              or if data address and length are outside buffer's limits.
	 * - DOCA_ERROR_NOT_PERMITTED - if doca_mmap or doca_buf_inventory is un-started/stopped.
	 * - DOCA_ERROR_NO_MEMORY - if doca_buf_inventory is empty.
	 */
	std::pair<doca_error_t, struct doca_buf *>
	allocate_buf_by_args(struct doca_mmap *mmap, void *addr, size_t len, void *data, size_t data_len) noexcept;

	/**
	 * @brief Duplicates content of the `buf` argument into element allocated from buffer inventory. (I.e., deep
	 * copy).
	 *
	 * @param src_buf [in] source buffer to duplicate
	 *
	 * @return doca_error_t and duplicated buffer if successful
	 * DOCA_SUCCESS - in case of success.
	 * doca_error code - in case of failure:
	 * - DOCA_ERROR_INVALID_VALUE - if an invalid input had been received.
	 * - DOCA_ERROR_NOT_PERMITTED - if src_buf mmap or input inventory unstarted/stopped
	 *                              or src_buf inventory extensions and the input inventory extensions are
	 * incompatible.
	 * - DOCA_ERROR_NO_MEMORY - if cannot alloc new doca_buf from the given inventory.
	 */
	std::pair<doca_error_t, struct doca_buf *>
	duplicate_buffer(const struct doca_buf *src_buf) noexcept;

	/**
	 * @brief Static callback for releasing a doca buf
	 * @param buf [in] doca_buf to release
	 */
	void release_doca_buffer(struct doca_buf *buf) noexcept;
	doca_error_t
	get_num_elements(uint32_t *num_of_elements) const;
	doca_error_t
	get_num_free_elements(uint32_t *num_of_free_elements) const;
	doca_error_t
	get_user_data(union doca_data *user_data) const;
	doca_error_t
	get_extensions(uint32_t *extensions) const;
private:

	/**
	 * @brief This method duplicates a single doca buffer
	 * @param [in] source buffer
	 * @return doca_error_t and duplicated buffer if successful
	 * DOCA_ERROR_NO_MEMORY
	 * DOCA_SUCCESS
	 */
	std::pair<doca_error_t, struct doca_buf *>
	duplicate_single_doca_buf(const struct doca_buf *src_buf);

	/**
	 * @brief This method duplicates an SG list doca buffer
	 * @param [in] source buffer
	 * @param [in] destination buffer
	 * @return doca_error_t and duplicated buffer if successful
	 * DOCA_ERROR_NO_MEMORY
	 * DOCA_SUCCESS
	 */
	std::pair<doca_error_t, struct doca_buf *>
	duplicate_sg_list_doca_buf(const struct doca_buf *src_buf);

	union doca_data m_user_data;						/**< Inventory's user data*/
	std::unique_ptr<doca_core::buf_pool> m_buf_pool;			/**< generic_pool of doca_bufs*/
	std::unique_ptr<doca_core::buf_extensions_map> m_extensions_map;

	// Inventory State
	enum class state {
		not_started,	/**< The inventory was never started */
		started,	/**< The inventory is started */
		stopped,	/**< The inventory is stopped, but was previously started */
	};
	state m_state{state::not_started};

	doca_buf_resources m_buffer_resources;

	// Non copyable
	doca_buf_inventory(doca_buf_inventory const &) = delete;
	doca_buf_inventory &operator=(doca_buf_inventory const &) = delete;
	doca_buf_inventory(doca_buf_inventory &&) = delete;
	doca_buf_inventory &operator=(doca_buf_inventory &&) = delete;
};
