/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>

#include <private/common_defines.h>
#include <doca_log.h>
#include <doca_dev.h>
#include "internal_doca_dev.h"
#include <idevice_adapter.hpp>
#include <ichannel_connection.hpp>
#include <priv_doca_channel_connection.h>

DOCA_LOG_REGISTER(CORE::DOCA_CHANNEL_CONNECTION);

using namespace device_abstraction;

doca_error_t
priv_doca_dev_channel_connection_create(const struct doca_dev *dev, const struct priv_doca_channel_connection_attributes *attr,
					struct priv_doca_channel_connection **channel_connection)
{
	if ((dev == nullptr) || (attr == nullptr) || (channel_connection == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, attr=%p, channel_connection=%p", dev, attr, channel_connection);
		return DOCA_ERROR_INVALID_VALUE;
	}

	ichannel_connection::attributes channel_connection_attr = {};

	channel_connection_attr.connection_type = static_cast<ichannel_connection::connection_types>(attr->connection_type);
	channel_connection_attr.transport_type = static_cast<ichannel_connection::transport_types>(attr->transport_type);
	channel_connection_attr.qpn_or_dci = attr->qpn_or_dci;
	channel_connection_attr.dst_vhca_id = attr->dst_vhca_id;
	channel_connection_attr.token = attr->token;

	if (attr->connection_type == PRIV_DOCA_CHANNEL_CONNECTION_CONNECT_BY_SERVICE_NAME) {
		if (attr->service_name == nullptr) {
			DOCA_DLOG_ERR("priv_doca_channel_connection_attributes::service_name cannot be NULL");
			return DOCA_ERROR_INVALID_VALUE;
		} else
			channel_connection_attr.service_name = attr->service_name;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<ichannel_connection> channel_connection_obj;
	std::tie(status, channel_connection_obj) = dev->device_adapter->create_channel_connection(channel_connection_attr);
	if (status != DOCA_SUCCESS)
		return status;

	*channel_connection = new (std::nothrow) priv_doca_channel_connection;
	if (*channel_connection == nullptr) {
		channel_connection_obj.reset();
		DOCA_DLOG_ERR("Failed to allocate doca_channel_connection");
		return DOCA_ERROR_NO_MEMORY;
	}

	(*channel_connection)->channel_connection_obj = channel_connection_obj;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_channel_connection_destroy(struct priv_doca_channel_connection *channel_connection)
{
	if (channel_connection == nullptr) {
		DOCA_DLOG_ERR("Invalid channel connection handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (channel_connection->channel_connection_obj == nullptr) {
		DOCA_DLOG_ERR("channel_connection object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	auto status = channel_connection->channel_connection_obj->destroy();
	if (status != DOCA_SUCCESS)
		return status;

	delete channel_connection;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_channel_connection_id_get(const struct priv_doca_channel_connection *channel_connection, uint32_t *id)
{
	if ((channel_connection == nullptr) || (id == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input channel_connection=%p, id=%p", channel_connection, id);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (channel_connection->channel_connection_obj == nullptr) {
		DOCA_DLOG_ERR("channel_connection object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	*id = channel_connection->channel_connection_obj->get_id();
	return DOCA_SUCCESS;
}
