/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>

#include <private/common_defines.h>
#include <doca_log.h>
#include <doca_dev.h>
#include "internal_doca_dev.h"
#include <idevice_adapter.hpp>
#include <ichannel_service.hpp>
#include <priv_doca_channel_service.h>

DOCA_LOG_REGISTER(CORE::DOCA_CHANNEL_SERVICE);

using namespace device_abstraction;

doca_error_t
priv_doca_dev_channel_service_create(const struct doca_dev *dev, const struct priv_doca_channel_service_attributes *attr,
				     struct priv_doca_channel_service **channel_service)
{
	if ((dev == nullptr) || (attr == nullptr) || (channel_service == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, attr=%p, channel_service=%p", dev, attr, channel_service);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (attr->service_name == nullptr) {
		DOCA_DLOG_ERR("priv_doca_channel_service_attributes::service_name cannot be NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	ichannel_service::attributes channel_service_attr = {};

	channel_service_attr.transport_type = static_cast<ichannel_service::transport_types>(attr->transport_type);
	channel_service_attr.allowed_vhca_id = attr->allowed_vhca_id;
	channel_service_attr.enforce_token = attr->enforce_token;
	channel_service_attr.qpn_or_dct = attr->qpn_or_dct;
	channel_service_attr.service_qkey = attr->service_qkey;
	channel_service_attr.token = attr->token;
	channel_service_attr.dct_key = attr->dct_key;
	channel_service_attr.service_name = attr->service_name;

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<ichannel_service> channel_service_obj;
	std::tie(status, channel_service_obj) = dev->device_adapter->create_channel_service(channel_service_attr);
	if (status != DOCA_SUCCESS)
		return status;

	*channel_service = new (std::nothrow) priv_doca_channel_service;
	if (*channel_service == nullptr) {
		channel_service_obj.reset();
		DOCA_DLOG_ERR("Failed to allocate doca_channel_service");
		return DOCA_ERROR_NO_MEMORY;
	}

	(*channel_service)->channel_service_obj = channel_service_obj;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_channel_service_destroy(struct priv_doca_channel_service *channel_service)
{
	if (channel_service == nullptr) {
		DOCA_DLOG_ERR("Invalid channel service handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (channel_service->channel_service_obj == nullptr) {
		DOCA_DLOG_ERR("channel_service object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	auto status = channel_service->channel_service_obj->destroy();
	if (status != DOCA_SUCCESS)
		return status;

	delete channel_service;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_channel_service_id_get(const struct priv_doca_channel_service *channel_service, uint32_t *id)
{
	if ((channel_service == nullptr) || (id == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input channel_service=%p, id=%p", channel_service, id);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (channel_service->channel_service_obj == nullptr) {
		DOCA_DLOG_ERR("channel_service object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	*id = channel_service->channel_service_obj->get_id();
	return DOCA_SUCCESS;
}
