/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>

#include <idevice_adapter.hpp>

#include <doca_log.h>
#include "internal_doca_dev.h"

#include <priv_doca_cq.h>

DOCA_LOG_REGISTER(CORE::DOCA_CQ);

using namespace device_abstraction;

doca_error_t
priv_doca_dev_cq_create(const struct doca_dev *dev, const struct priv_doca_uar *uar,
			const struct priv_doca_cq_attr *attr, struct priv_doca_cq **cq_out)
{
	if (dev == NULL || uar == NULL || cq_out == NULL || attr == NULL) {
		DOCA_DLOG_ERR("Unable to create CQ: received invalid arguments. dev=%p, uar=%p, attr=%p, cq_out=%p.",
			      dev, uar, attr, cq_out);
		return DOCA_ERROR_INVALID_VALUE;
	}

	icompletion_queue::attributes q_attr{};
	q_attr.nb_cqes = attr->nb_cqes;
	q_attr.initiator_src_dct = attr->initiator_src_dct;
	doca_error_t ret{DOCA_SUCCESS};
	std::shared_ptr<icompletion_queue> cq_obj{nullptr};
	std::tie(ret, cq_obj) = dev->device_adapter->create_completion_queue(uar->uar_obj, q_attr, attr->user_data,
									  attr->handler);
	if (ret != DOCA_SUCCESS)
		return ret;

	*cq_out = new (std::nothrow) struct priv_doca_cq;
	if (*cq_out == nullptr) {
		DOCA_DLOG_ERR("Unable to create CQ: out of memory.");
		return DOCA_ERROR_NO_MEMORY;
	}

	(*cq_out)->cq_obj = cq_obj;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_cq_destroy(struct priv_doca_cq *cq)
{
	if (cq == NULL) {
		DOCA_DLOG_ERR("Unable to destroy CQ: received invalid input. cq=%p.", cq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto ret = cq->cq_obj->destroy();
	if (ret != DOCA_SUCCESS)
		return ret;

	delete cq;
	return ret;
}

doca_error_t
priv_doca_cq_user_data_get(struct priv_doca_cq *cq, union doca_data *user_data)
{
	*user_data = cq->cq_obj->get_user_data();
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_cq_number_get(const struct priv_doca_cq *cq, uint32_t *cqn)
{
	if (cq == NULL || cqn == NULL) {
		DOCA_DLOG_ERR("Unable to get CQ number: reveiced invalid input. cq=%p, cqn=%p.", cq, cqn);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*cqn = cq->cq_obj->get_cq_number();

	return DOCA_SUCCESS;
}

size_t
priv_doca_cq_poll(struct priv_doca_cq *cq, struct priv_doca_64b_cqe *cqes_out, size_t nb_max_cqes)
{
	return cq->cq_obj->poll(reinterpret_cast<struct mlx5_64b_cqe*>(cqes_out), nb_max_cqes);
}

doca_error_t
priv_doca_cq_poll_one(struct priv_doca_cq *cq, struct priv_doca_64b_cqe *cqe_out)
{
	return cq->cq_obj->poll_one(reinterpret_cast<struct mlx5_64b_cqe*>(cqe_out));
}

bool
priv_doca_cq_empty(const struct priv_doca_cq *cq)
{
	return cq->cq_obj->empty();
}

void
priv_doca_cq_arm(struct priv_doca_cq *cq)
{
	cq->cq_obj->arm();
}

doca_error_t
priv_doca_dev_comp_events_channel_create(const struct doca_dev *dev, bool is_blocking,
					  struct priv_doca_comp_events_channel **channel_out)
{
	if (dev == NULL || channel_out == NULL) {
		DOCA_DLOG_ERR("Unable to create completion events channel: received invalid input. dev=%p, channel_out=%p.",
			      dev, channel_out);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t ret{DOCA_SUCCESS};
	std::shared_ptr<icompletion_events_channel> channel_obj;
	std::tie(ret, channel_obj) = dev->device_adapter->create_completion_events_channel(is_blocking);
	if (ret != DOCA_SUCCESS)
		return ret;

	*channel_out = new (std::nothrow) struct priv_doca_comp_events_channel;
	if (*channel_out == nullptr) {
		DOCA_DLOG_ERR("Unable to create completion events channel: out of memory.");
		return DOCA_ERROR_NO_MEMORY;
	}

	(*channel_out)->channel_obj = channel_obj;
	return DOCA_SUCCESS;
}

void
priv_doca_dev_comp_events_channel_destroy(struct priv_doca_comp_events_channel *channel)
{
	delete channel;
}

doca_error_t
priv_doca_comp_events_channel_handle_get(struct priv_doca_comp_events_channel *channel, doca_event_handle_t *handle)
{
	if (channel == NULL || handle == NULL) {
		DOCA_DLOG_ERR("Unable to get completion events channel handle: received invalid input. channel=%p, handle=%p.",
			      channel, handle);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*handle = channel->channel_obj->get_handle();
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_comp_events_channel_register(struct priv_doca_comp_events_channel *channel, struct priv_doca_cq *cq)
{
	if (channel == NULL || cq == NULL) {
		DOCA_DLOG_ERR("Unable to register CQ to completion events channel: received invalid input. channel=%p, cq=%p.",
			      channel, cq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return channel->channel_obj->register_cq(cq->cq_obj);
}

doca_error_t
priv_doca_comp_events_channel_unregister(struct priv_doca_comp_events_channel *channel, struct priv_doca_cq *cq)
{
	if (channel == NULL || cq == NULL) {
		DOCA_DLOG_ERR("Unable to unregister CQ from completion events channel: received invalid input. channel=%p, cq=%p.",
			      channel, cq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return channel->channel_obj->unregister_cq(cq->cq_obj);
}

void
priv_doca_comp_events_channel_arm(struct priv_doca_comp_events_channel *channel)
{
	channel->channel_obj->arm();
}

void
priv_doca_comp_events_channel_force_arm(struct priv_doca_comp_events_channel *channel)
{
	channel->channel_obj->force_arm();
}

doca_error_t
priv_doca_comp_events_channel_handle_events(struct priv_doca_comp_events_channel *channel, doca_event_handle_t handle)
{
	return channel->channel_obj->handle_events(handle);
}

