/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>
#include <stdbool.h>
#include <list>
#include <algorithm>
#include <array>
#include <unordered_set>
#include <cassert>

#include <ievents_mux.hpp>
#include <events_mux_factory.hpp>

#include <doca_log.h>
#include <doca_error.h>
#include <id_allocator.h>
#include <private/common_defines.h>
#include <priv_doca_ctx.h>

#define MAX_NUM_CTX 1024 /**< Maximum number of doca_ctx allowed within an application.*/

DOCA_LOG_REGISTER(CORE::DOCA_CTX);

ID_ALLOCATOR_REGISTER(ctx_id_aloc, MAX_NUM_CTX); /**< Allocator of ctx->id */

using namespace device_abstraction;

struct doca_workq_properties {
	/** Fields that are properties.*/
	uint32_t depth{0};
};

struct doca_workq {
	std::list<priv_doca_workq_progress_entry *> running_entry_head;
	std::list<priv_doca_workq_progress_entry *> suspend_entry_head;
	std::list<priv_doca_workq_progress_entry *>::iterator current_entry;
	std::list<priv_doca_workq_progress_entry *>::iterator last_entry;
	std::unordered_set<const priv_doca_workq_progress_entry *> entry_set;

	/** Index 0 is reserved for invalid ID default data. */
	std::array<union doca_data, MAX_NUM_CTX + 1> ctx_privs{};

	/** Bitmap used for marking which CTXs were added. */
	std::array <uint64_t, MAX_NUM_CTX / 64> ctx_map{};

	/** Event mux used for multiplexing of multiple event handles*/
	ievents_mux *mux{nullptr};
	bool event_triggered{false};

	doca_workq_properties properties;
};

#define BITMAP_IDX_GET(bitmap, idx) ((bitmap)[(idx) / 64] & (1ULL << ((idx) - 1)))
#define BITMAP_IDX_TOGGLE(bitmap, idx) ((bitmap)[(idx) / 64] ^= (1ULL << ((idx) - 1)))

static inline bool
priv_doca_data_priv_exists(const struct doca_workq *workq, priv_doca_ctx_id_t ctx_id)
{
	return BITMAP_IDX_GET(workq->ctx_map, ctx_id);
}

static inline void
priv_doca_data_priv_add(struct doca_workq *workq, priv_doca_ctx_id_t ctx_id, union doca_data data_priv)
{
	workq->ctx_privs[ctx_id] = data_priv;
	BITMAP_IDX_TOGGLE(workq->ctx_map, ctx_id);
}

static inline union doca_data
priv_doca_data_priv_get(const struct doca_workq *workq, priv_doca_ctx_id_t ctx_id)
{
	return workq->ctx_privs[ctx_id];
}

static inline void
priv_doca_data_priv_rm(struct doca_workq *workq, priv_doca_ctx_id_t ctx_id)
{
	workq->ctx_privs[ctx_id].ptr = NULL;
	BITMAP_IDX_TOGGLE(workq->ctx_map, ctx_id);
}

static inline bool
priv_doca_is_workq_in_any_ctx(struct doca_workq *workq)
{
	for (auto bitmap : workq->ctx_map)
		if (bitmap != 0)
			return true;
	return false;
}

doca_error_t
doca_workq_create(uint32_t depth, struct doca_workq **workq)
{
	if (depth == 0 || workq == NULL) {
		DOCA_DLOG_ERR("Unable to create workQ: Received invalid input - depth=%u, workq=%p", depth, workq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	try {
		*workq = new doca_workq();
		(*workq)->properties.depth = depth;
		(*workq)->current_entry = (*workq)->running_entry_head.end();
		(*workq)->last_entry = (*workq)->running_entry_head.end();
	} catch (...) {
		DOCA_DLOG_ERR("Unable to create workQ: Failed to allocate workq - Out of memory");
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_destroy(struct doca_workq *workq)
{
	if (workq == NULL) {
		DOCA_DLOG_ERR("Unable to destroy workQ: Received invalid input - workq is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (!workq->entry_set.empty()) {
		DOCA_DLOG_ERR(
		  "Unable to destroy workQ: Some progress entries still exist - remove workq from all ctxs before destroying it");
		return DOCA_ERROR_IN_USE;
	}
	auto ret = doca_workq_set_event_driven_enable(workq, 0);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to destroy workQ: Failed to disable event driven mode.");
		return ret;
	}
	delete workq;
	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_get_depth(const struct doca_workq *workq, uint32_t *depth)
{
	if (workq == NULL || depth == NULL) {
		DOCA_DLOG_ERR("Unable to get workQ depth property: Received invalid input - depth=%p, workq=%p",
			depth, workq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*depth = workq->properties.depth;

	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_set_depth(struct doca_workq *workq, uint32_t depth)
{
	if (depth == 0 || workq == NULL) {
		DOCA_DLOG_ERR("Unable to set workQ depth: Received invalid input - depth=%u, workq=%p",
			depth, workq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	workq->properties.depth = depth;

	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_progress_retrieve(struct doca_workq *workq, struct doca_event *ev, int /*flags*/)
{
	doca_error_t ret{DOCA_ERROR_AGAIN};

	if (unlikely(workq == NULL || ev == NULL)) {
		DOCA_DLOG_ERR("Unable to retrieve workQ progress: Received invalid input - workq=%p, ev=%p", workq, ev);
		return DOCA_ERROR_INVALID_VALUE;
	}

	workq->last_entry = workq->current_entry;

	if (unlikely(workq->last_entry == workq->running_entry_head.end())) {
		DOCA_DLOG_DBG(
			"Unable to retrieve workQ progress: Attempting to progress a workQ with no progress entries");
		return DOCA_ERROR_AGAIN;
	}

	do {
		auto current_entry = *workq->current_entry;
		ret = current_entry->progress_cb(current_entry->data_priv, ev);

		/* Increment curent_entry allowing the next iteration/call to continue from the next entry in list.*/

		workq->current_entry = (std::next(workq->current_entry) == workq->running_entry_head.end() ?
			workq->running_entry_head.begin() : std::next(workq->current_entry));
	} while ((ret == DOCA_ERROR_AGAIN) && (workq->current_entry != workq->last_entry));

	return ret;
}

static inline bool
priv_doca_progress_entry_exists(const struct doca_workq *workq, const struct priv_doca_workq_progress_entry *entry)
{
	return workq->entry_set.find(entry) != workq->entry_set.end();
}

doca_error_t
priv_doca_workq_progress_entry_add(struct doca_workq *workq, struct priv_doca_workq_progress_entry *entry)
{
	if (priv_doca_progress_entry_exists(workq, entry)) {
		DOCA_DLOG_ERR(
		  "Unable to add progress entry to workQ: Same entry already added to workq. Entry has address %p",
		  entry);
		return DOCA_ERROR_IN_USE;
	}

	try {
		workq->entry_set.insert(entry);
		if (workq->running_entry_head.empty()) {
			workq->running_entry_head.push_front(entry);
			workq->current_entry = workq->running_entry_head.begin();
			workq->last_entry = workq->current_entry;
		} else {
			workq->running_entry_head.insert(workq->current_entry, entry);
		}
	} catch (...) {
		DOCA_DLOG_ERR("Unable to add progress entry to workQ: Failed to add entry to list or set. Out of memory");
		static_cast<void>(workq->entry_set.erase(entry));
		return DOCA_ERROR_NO_MEMORY;
	}

	entry->state = PRIV_DOCA_WORKQ_PROGRESS_RUNNING;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_workq_progress_entry_rm(struct doca_workq *workq, struct priv_doca_workq_progress_entry *entry)
{
	if (!priv_doca_progress_entry_exists(workq, entry)) {
		DOCA_DLOG_ERR(
		"Unable to remove progress entry from workQ: Given progress entry, with address %p, doesn't exist in workq",
		entry);
		return DOCA_ERROR_NOT_FOUND;
	}

	auto &current_list = (entry->state == PRIV_DOCA_WORKQ_PROGRESS_RUNNING) ?
				workq->running_entry_head : workq->suspend_entry_head;
	auto it = std::find(current_list.begin(), current_list.end(), entry);
	if (it == current_list.end()) {
		DOCA_DLOG_ERR("Unable to remove progress entry from workQ: Given progress entry state was externally modified.");
		return DOCA_ERROR_BAD_STATE;
	}

	if (it == workq->current_entry)
		workq->current_entry = std::next(it);

	if (it == workq->last_entry)
		workq->last_entry = std::next(it);

	current_list.erase(it);
	workq->entry_set.erase(entry);

	workq->current_entry = workq->current_entry == workq->running_entry_head.end() ?
				workq->running_entry_head.begin() : workq->current_entry;
	workq->last_entry = workq->last_entry == workq->running_entry_head.end() ?
				workq->running_entry_head.begin() : workq->last_entry;

	return DOCA_SUCCESS;
}

doca_error_t priv_doca_workq_progress_entry_state_update(struct doca_workq *workq,
							 struct priv_doca_workq_progress_entry *entry,
							 enum priv_doca_workq_progress_state new_state)
{
	if (unlikely(workq == NULL || entry == NULL)) {
		DOCA_DLOG_ERR("Unable to update progress entry: Received invalid input - workq=%p, entry=%p", workq,
			       entry);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (unlikely(new_state != PRIV_DOCA_WORKQ_PROGRESS_RUNNING && new_state != PRIV_DOCA_WORKQ_PROGRESS_SUSPENDED)) {
		DOCA_DLOG_ERR("Unable to update progress entry: Received invalid input - Unknown state with value %d",
			       new_state);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (unlikely(!priv_doca_progress_entry_exists(workq, entry))) {
		DOCA_DLOG_ERR(
			"Unable to update progress entry: The progress entry being updated doesn't exist in workq. Entry has address %p",
			entry);
		return DOCA_ERROR_NOT_FOUND;
	}

	if (unlikely(entry->state == new_state)) {
		DOCA_DLOG_WARN("Unable to update progress entry: The progress entry already in requested state.");
		return DOCA_SUCCESS;
	}

	entry->state = new_state;

	auto &prev_list = (new_state == PRIV_DOCA_WORKQ_PROGRESS_RUNNING) ?
				workq->suspend_entry_head : workq->running_entry_head;
	auto &new_list = (new_state == PRIV_DOCA_WORKQ_PROGRESS_RUNNING) ?
				workq->running_entry_head : workq->suspend_entry_head;

	/* We want to move element before current entry in running list, or front of suspended list. */
	auto new_list_pos = (new_state == PRIV_DOCA_WORKQ_PROGRESS_RUNNING) ?
				workq->current_entry : new_list.begin();

	auto it = std::find(prev_list.begin(), prev_list.end(), entry);
	if (unlikely(it == prev_list.end())) {
		DOCA_DLOG_ERR("Unable to update progress entry: Entry state was externally modified.");
		return DOCA_ERROR_BAD_STATE;
	}

	auto next_current = std::next(workq->current_entry);
	auto next_last = std::next(workq->last_entry);

	new_list.splice(new_list_pos, prev_list, it);

	/* Find entry that should be after the last entry in a cyclic manner. Must be done after lists are updated.*/
	next_last = next_last == workq->running_entry_head.end() ? workq->running_entry_head.begin() : next_last;

	/* If the entry that was moved is the curent_entry, then update to be the next. */
	workq->current_entry = (it == workq->current_entry) ? next_current : workq->current_entry;

	/* In case current entry was moved and was last element, or running_entry_head was empty and new entry arrived. */
	workq->current_entry = workq->current_entry == workq->running_entry_head.end() ?
				workq->running_entry_head.begin() : workq->current_entry;

	/* If the entry that was moved is the last_entry, then update it as well. */
	workq->last_entry = (it == workq->last_entry) ? next_last : workq->last_entry;

	return DOCA_SUCCESS;
}

static inline bool
priv_doca_ctx_id_is_valid(priv_doca_ctx_id_t ctx_id)
{
	return (ctx_id <= MAX_NUM_CTX) && (ctx_id != PRIV_DOCA_CTX_INVALID_ID);
}

doca_error_t
priv_doca_workq_data_priv_get(struct doca_workq *workq, priv_doca_ctx_id_t ctx_id, union doca_data *data_priv)
{
	if (!priv_doca_ctx_id_is_valid(ctx_id)) {
		DOCA_DLOG_ERR("Unable to get data_priv: Received invalid context ID %u", ctx_id);
		return DOCA_ERROR_INVALID_VALUE;
	}
	*data_priv = workq->ctx_privs[ctx_id];
	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_submit(struct doca_workq *workq, struct doca_job const *job)
{
	if (unlikely(workq == NULL || job == NULL)) {
		DOCA_DLOG_ERR("Unable to submit workQ: Received invalid input - workq=%p, job=%p.", workq, job);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (unlikely(job->ctx == NULL)) {
		DOCA_DLOG_ERR("Unable to submit workQ: Received invalid input - job->ctx is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (unlikely(job->ctx->is_started == false)) {
		DOCA_DLOG_ERR("Unable to submit workQ: Attempting to submit job to a context that is stopped");
		return DOCA_ERROR_BAD_STATE;
	}

	if (unlikely(priv_doca_data_priv_exists(workq, job->ctx->id) == false)) {
		DOCA_DLOG_ERR("Unable to submit workQ: Attempting to submit job to a context that is not associated to the workQ.");
		return DOCA_ERROR_NOT_FOUND;
	}

	return job->ctx->ops.job_submit(workq->ctx_privs[job->ctx->id], job);
}

doca_error_t
priv_doca_ctx_init(struct doca_ctx *ctx, bool create_id)
{
	if (!create_id) {
		ctx->id = PRIV_DOCA_CTX_INVALID_ID;
		return DOCA_SUCCESS;
	}
	memset(ctx, 0, sizeof(*ctx));
	ctx->id = static_cast<priv_doca_ctx_id_t>(id_allocator_get(ctx_id_aloc));
	if (ctx->id == ID_ALLOCATOR_INVALID_ID) {
		DOCA_DLOG_ERR("Unable to initialize context: Failed to allocate ID for ctx");
		ctx->id = PRIV_DOCA_CTX_INVALID_ID;
		return DOCA_ERROR_INITIALIZATION;
	}
	return DOCA_SUCCESS;
}

void
priv_doca_ctx_destroy(struct doca_ctx *ctx)
{
	if (ctx->id == PRIV_DOCA_CTX_INVALID_ID)
		return;
	id_allocator_put(ctx_id_aloc, ctx->id);
	ctx->id = PRIV_DOCA_CTX_INVALID_ID;
}

doca_error_t
doca_ctx_dev_add(struct doca_ctx *ctx, struct doca_dev *dev)
{
	if (ctx == NULL || dev == NULL) {
		DOCA_DLOG_ERR("Unable to add device to context: Received invalid input - ctx=%p, dev=%p", ctx, dev);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (ctx->is_started) {
		DOCA_DLOG_ERR(
		     "Unable to add device to context: Can't add devices to a started context - Please stop ctx first");
		return DOCA_ERROR_BAD_STATE;
	}

	return ctx->ops.dev_add(ctx, dev);
}

doca_error_t
doca_ctx_dev_rm(struct doca_ctx *ctx, struct doca_dev *dev)
{
	if (ctx == NULL || dev == NULL) {
		DOCA_DLOG_ERR("Unable to remove device from context: Received invalid input - ctx=%p, dev=%p",
			       ctx, dev);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (ctx->is_started) {
		DOCA_DLOG_ERR(
		 "Unable to remove device from context: Can't remove devices from a started context - Please stop ctx first"
		 );
		return DOCA_ERROR_BAD_STATE;
	}

	return ctx->ops.dev_rm(ctx, dev);
}

doca_error_t
doca_ctx_start(struct doca_ctx *ctx)
{
	if (ctx == NULL) {
		DOCA_DLOG_ERR("Unable to start context: Received invalid input - ctx is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	const doca_error_t start_err = ctx->ops.start(ctx);

	if (start_err == DOCA_SUCCESS)
		ctx->is_started = true;

	return start_err;
}

doca_error_t
doca_ctx_stop(struct doca_ctx *ctx)
{
	if (ctx == NULL) {
		DOCA_DLOG_ERR("Unable to stop context: Received invalid input - ctx is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	const doca_error_t stop_err = ctx->ops.stop(ctx);

	if (stop_err == DOCA_SUCCESS)
		ctx->is_started = false;

	return stop_err;
}

doca_error_t
doca_ctx_workq_add(struct doca_ctx *ctx, struct doca_workq *workq)
{
	union doca_data data_priv;
	doca_error_t err;

	if (ctx == NULL || workq == NULL) {
		DOCA_DLOG_ERR("Unable to add workQ to context: Received invalid input - ctx=%p, workq=%p", ctx, workq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (!ctx->is_started) {
		DOCA_DLOG_ERR(
		     "Unable to add workQ to context: Can't add a workQ to a stopped context - Please start ctx first");
		return DOCA_ERROR_BAD_STATE;
	}

	if (ctx->id == PRIV_DOCA_CTX_INVALID_ID)
		return ctx->ops.workq_add(ctx, workq, NULL);

	if (!priv_doca_ctx_id_is_valid(ctx->id)) {
		DOCA_DLOG_ERR("Unable to add workQ to context: Received invalid input - ctx has invalid context ID: %u",
			       ctx->id);
		return DOCA_ERROR_INVALID_VALUE;
	}
	/* Check if this WorkQ was previously added to ctx.*/
	if (priv_doca_data_priv_exists(workq, ctx->id)) {
		DOCA_DLOG_ERR("Unable to add workQ to context: Same workq was already added to the context");
		return DOCA_ERROR_IN_USE;
	}

	/* Check if working in event-driven mode or polling mode.*/
	uint8_t workq_event_driven{0}, ctx_event_driven{0};
	static_cast<void>(doca_workq_get_event_driven_enable(workq, &workq_event_driven));
	static_cast<void>(doca_ctx_get_event_driven_supported(ctx, &ctx_event_driven));

	if (workq_event_driven && !ctx_event_driven) {
		DOCA_DLOG_ERR("Unable to add workQ to context: Attempting to add event driven workq to ctx that has no event driven support");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	/* Pass the WorkQ to CTX, and then register the returned data_priv.*/
	if (workq_event_driven) {
		err = ctx->ops.event_workq_add(ctx, workq, &data_priv);
	} else {
		err = ctx->ops.workq_add(ctx, workq, &data_priv);
	}
	if (err != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to add workQ to context: ctx failed to add workq");
		return err;
	}
	/* Add the CTX private data, representing it's per WorkQ state.*/
	priv_doca_data_priv_add(workq, ctx->id, data_priv);

	return DOCA_SUCCESS;
}

doca_error_t
doca_ctx_workq_rm(struct doca_ctx *ctx, struct doca_workq *workq)
{
	doca_error_t err;

	if (ctx == NULL || workq == NULL) {
		DOCA_DLOG_ERR("Unable to remove workQ from context: Received invalid input - ctx=%p, workq=%p",
			       ctx, workq);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (!ctx->is_started) {
		DOCA_DLOG_ERR(
		  "Unable to remove workQ from context: Can't remove workq from a stopped context - Please start ctx first");
		return DOCA_ERROR_BAD_STATE;
	}

	static const union doca_data data_priv_non{};

	if (ctx->id == PRIV_DOCA_CTX_INVALID_ID)
		return ctx->ops.workq_rm(ctx, workq, data_priv_non);

	if (!priv_doca_ctx_id_is_valid(ctx->id)) {
		DOCA_DLOG_ERR(
			"Unable to remove workQ from context: Received invalid input - ctx has invalid context ID: %u",
			ctx->id);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (!priv_doca_data_priv_exists(workq, ctx->id)) {
		DOCA_DLOG_ERR("Unable to remove workQ from context: workq was not previously added to ctx");
		return DOCA_ERROR_NOT_FOUND;
	}

	/* Check if working in event-driven mode or polling mode.*/
	uint8_t workq_event_driven{0};
	static_cast<void>(doca_workq_get_event_driven_enable(workq, &workq_event_driven));

	if (workq_event_driven) {
		err = ctx->ops.event_workq_rm(ctx, workq, priv_doca_data_priv_get(workq, ctx->id));
	} else {
		err = ctx->ops.workq_rm(ctx, workq, priv_doca_data_priv_get(workq, ctx->id));
	}
	if (err != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to remove workQ from context: ctx failed to remove workq");
		return err;
	}
	/* Reset CTX private data, mark it as unused with NULL.*/
	priv_doca_data_priv_rm(workq, ctx->id);
	return DOCA_SUCCESS;
}

/**********************************************************************************************************************
 * Event Driven Mode Methods
 **********************************************************************************************************************/

doca_error_t
doca_workq_event_handle_arm(struct doca_workq *workq)
{
	if (unlikely(workq == NULL)) {
		DOCA_DLOG_ERR("Unable to arm workq event handle: Received invalid input - workq=NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(workq->mux == nullptr)) {
		DOCA_DLOG_ERR("Unable to arm workq event handle: workq not in event driven mode.");
		return DOCA_ERROR_BAD_STATE;
	}

	auto ret = workq->mux->arm();
	if (ret != DOCA_SUCCESS)
		return ret;

	if (!workq->event_triggered && !workq->running_entry_head.empty()) {
		/* In case some progress entry is still added, application waiting on fd should wake up immediately.*/
		auto ret = workq->mux->trigger_event();
		if (ret != DOCA_SUCCESS)
			return ret;

		workq->event_triggered = true;
	} else if (workq->event_triggered && workq->running_entry_head.empty()) {
		/* In case no entry is running, application waiting on fd would wait for CTX to trigger event.*/
		auto ret = workq->mux->untrigger_event();
		if (ret != DOCA_SUCCESS)
			return ret;

		workq->event_triggered = false;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_event_handle_clear(struct doca_workq *workq, doca_event_handle_t handle)
{
	if (unlikely(workq == NULL)) {
		DOCA_DLOG_ERR("Unable to arm workq event handle: Received invalid input - workq=NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (unlikely(workq->mux == nullptr)) {
		DOCA_DLOG_ERR("Unable to arm workq event handle: workq not in event driven mode.");
		return DOCA_ERROR_BAD_STATE;
	}
	return workq->mux->clear(handle);
}

doca_error_t
doca_workq_get_event_handle(struct doca_workq *workq, doca_event_handle_t *handle)
{
	if (workq == nullptr || handle == nullptr) {
		DOCA_DLOG_ERR("Unable to get workq event handle: Received invalid value workq=%p, handle=%p",
			       workq, handle);
		return DOCA_ERROR_INVALID_VALUE;
	}

	uint8_t enabled{0};
	static_cast<void>(doca_workq_get_event_driven_enable(workq, &enabled));
	if (!enabled) {
		DOCA_DLOG_ERR("Unable to get workq event handle: Event-Driven mode not enabled.");
		return DOCA_ERROR_BAD_STATE;
	}

	*handle = workq->mux->get_handle();
	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_get_event_driven_enable(struct doca_workq *workq, uint8_t *enabled)
{
	if (workq == nullptr || enabled == nullptr) {
		DOCA_DLOG_ERR("Unable to get workq event driven enable: Received invalid value workq=%p, enabled=%p",
			       workq, enabled);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*enabled = (workq->mux != nullptr);

	return DOCA_SUCCESS;
}

doca_error_t
doca_workq_set_event_driven_enable(struct doca_workq *workq, uint8_t enable)
{
	if (workq == nullptr) {
		DOCA_DLOG_ERR("Unable to set workq event driven enable: Received invalid value workq=NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}


	/* See if requested state is same as existing state. */
	uint8_t enabled{0};
	static_cast<void>(doca_workq_get_event_driven_enable(workq, &enabled));
	if (!!enabled == !!enable)
		return DOCA_SUCCESS;

	if (priv_doca_is_workq_in_any_ctx(workq)) {
		DOCA_DLOG_ERR("Unable to set workq event driven enable: Please remove WorkQ from all CTXs first.");
		return DOCA_ERROR_BAD_STATE;
	}

	if (enable) {
		doca_error_t ret{DOCA_SUCCESS};
		ievents_mux *mux_tmp{nullptr};
		std::tie(ret, mux_tmp) = device_abstraction::create_events_mux();
		if (ret != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Unable to set workq event driven enable: Failed to create event mux.");
			return ret;
		}
		workq->mux = mux_tmp;
	} else {
		auto ret = workq->mux->destroy();
		if (ret != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Unable to set workq event driven enable: Failed to destroy event mux.");
			return ret;
		}
		delete workq->mux;
		workq->mux = nullptr;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_ctx_get_event_driven_supported(struct doca_ctx *ctx, uint8_t *event_supported)
{
	if (ctx == nullptr || event_supported == nullptr) {
		DOCA_DLOG_ERR("Unable to query event driven support: Received invalid input - ctx=%p, event_supported=%p",
				ctx, event_supported);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*event_supported = (ctx->ops.event_workq_add == nullptr || ctx->ops.event_workq_rm == nullptr) ? 0 : 1;

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_workq_event_handle_register(struct doca_workq *workq,
				      doca_event_handle_t event_handle,
				      const struct priv_doca_event_handle_ops *event_handle_ops,
				      union doca_data ops_data)
{
	if (workq == NULL || event_handle_ops == NULL) {
		DOCA_DLOG_ERR("Unable to register event handle to workq: Received invalid input - workq=%p, event_handle_ops=%p",
			       workq, event_handle_ops);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (event_handle_ops->arm_cb == nullptr || event_handle_ops->clear_cb == nullptr) {
		DOCA_DLOG_ERR("Unable to add event handle to workq: Received invalid input - arm_cb=%p, clear_cb=%p",
			       event_handle_ops->arm_cb, event_handle_ops->clear_cb);
		return DOCA_ERROR_INVALID_VALUE;
	}
	uint8_t enabled{0};
	static_cast<void>(doca_workq_get_event_driven_enable(workq, &enabled));
	if (!enabled) {
		DOCA_DLOG_ERR("Unable to add event handle to workq: workq event-driven mode is disabled.");
		return DOCA_ERROR_BAD_STATE;
	}

	return workq->mux->register_handle(event_handle, *event_handle_ops, ops_data);
}

doca_error_t
priv_doca_workq_event_handle_unregister(struct doca_workq *workq, doca_event_handle_t event_handle)
{
	if (workq == NULL) {
		DOCA_DLOG_ERR("Unable to register event handle to workq: Received invalid input - workq=NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}
	uint8_t enabled{0};
	static_cast<void>(doca_workq_get_event_driven_enable(workq, &enabled));
	if (!enabled) {
		DOCA_DLOG_ERR("Unable to remove event handle from workq: workq event-driven mode is disabled.");
		return DOCA_ERROR_BAD_STATE;
	}

	return workq->mux->unregister_handle(event_handle);
}

doca_error_t
doca_ctx_get_max_num_ctx(uint32_t *max_num_ctx)
{
	if (max_num_ctx == NULL) {
		DOCA_DLOG_ERR("Unable to get max num of contexts: Received invalid input - max_num_ctx=%p",
			max_num_ctx);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*max_num_ctx = MAX_NUM_CTX;

	return DOCA_SUCCESS;
}
