/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>

#include <private/common_defines.h>
#include <doca_log.h>
#include <doca_dev.h>
#include "internal_doca_dev.h"
#include <idevice_adapter.hpp>
#include <idynamically_connected_target.hpp>
#include <priv_doca_dct.h>

DOCA_LOG_REGISTER(CORE::DOCA_DCT);

using namespace device_abstraction;

doca_error_t
priv_doca_dev_dct_create(const struct doca_dev *dev, const struct priv_doca_dct_attributes *attr,
			 struct priv_doca_dct **dct)
{
	if ((dev == nullptr) || (attr == 0) || (dct == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, attr=%p, dct=%p", dev, attr, dct);
		return DOCA_ERROR_INVALID_VALUE;
	}

	idynamically_connected_target::attributes dct_attr = {};

	dct_attr.cq = attr->cq->cq_obj;
	dct_attr.rmp = attr->rmp->rmp_obj;
	dct_attr.port = attr->port;
	dct_attr.mtu = attr->mtu;
	dct_attr.my_addr_index = attr->my_addr_index;
	dct_attr.min_rnr_nak = attr->min_rnr_nak;
	dct_attr.dc_access_key = attr->dc_access_key;

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<idynamically_connected_target> dct_obj;
	std::tie(status, dct_obj) = dev->pd->create_dynamically_connected_target(dct_attr);
	if (status != DOCA_SUCCESS)
		return status;

	*dct = new (std::nothrow) struct priv_doca_dct;
	if (*dct == nullptr) {
		DOCA_DLOG_ERR("Failed to allocate priv_doca_dct");
		return DOCA_ERROR_NO_MEMORY;
	}

	(*dct)->dct_obj = dct_obj;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_dct_destroy(struct priv_doca_dct *dct)
{
	if (dct == nullptr) {
		DOCA_DLOG_ERR("Invalid dct handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (dct->dct_obj == nullptr) {
		DOCA_DLOG_ERR("dct object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	auto status = dct->dct_obj->destroy();
	if (status != DOCA_SUCCESS)
		return status;

	delete dct;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dct_id_get(const struct priv_doca_dct *dct, uint32_t *id)
{
	if ((dct == nullptr) || (id == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input dct=%p, id=%p", dct, id);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (dct->dct_obj == nullptr) {
		DOCA_DLOG_ERR("dct object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	*id = dct->dct_obj->get_id();
	return DOCA_SUCCESS;
}
