/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#include <stdbool.h>
#include <ctype.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include <vector>
#include <unordered_map>
#include <array>
#include <algorithm>
#include <list>
#include <memory>

#include <doca_log.h>
#include "internal_doca_dev.h"
#include <doca_dev.h>
#include <priv_doca_dev.h>
#include <devices_factory.hpp>
#include <idevice_adapter.hpp>
#include <priv_doca_os_abstraction.hpp>
#include <private/common_defines.h>
#include <doca_rdma.h>

DOCA_LOG_REGISTER(CORE::DOCA_DEVICE);

using namespace device_abstraction;

/* Anonymous namespace - used for declaring symbols that are visible from same file only. */
namespace {

#define ENABLE_EMULATED_VFS 0

/**
 * This struct contains a created devinfos list which is given to the user.
 * When the user calls destroy with the devinfos nb_devs is extracted by
 * decrementing the pointer.
 */
struct devinfo_list_priv {
	uint32_t nb_devs;
	uint32_t reserved; /* To keep 64-bit aligned */
	struct doca_devinfo* devinfos[1]; /** Some compilers warn against structs with size of 0, therefore the size is 1 */
};

static std::list<doca_dev *> global_dev_list;
std::mutex dev_list_lock;

static const int filters_all = DOCA_DEV_REP_FILTER_NET | DOCA_DEV_REP_FILTER_VIRTIO_FS |
			       DOCA_DEV_REP_FILTER_VIRTIO_NET | DOCA_DEV_REP_FILTER_VIRTIO_BLK |
			       DOCA_DEV_REP_FILTER_NVME;

struct doca_dev_rep_convert_entry {
	enum doca_dev_rep_filter rep_filter;
	enum priv_doca_rep_type rep_type;
	enum emulation_type emulation_type;
	const char *descr;
};

static const struct doca_dev_rep_convert_entry doca_dev_rep_convert_arr[] = {
	{ DOCA_DEV_REP_FILTER_VIRTIO_FS, PRIV_DOCA_REP_VIRTIO_FS, emulation_type::VIRTIO_FS, "VIRTIO_FS", },
	{ DOCA_DEV_REP_FILTER_VIRTIO_NET, PRIV_DOCA_REP_VIRTIO_NET, emulation_type::VIRTIO_NET, "VIRTIO_NET", },
	{ DOCA_DEV_REP_FILTER_VIRTIO_BLK, PRIV_DOCA_REP_VIRTIO_BLK, emulation_type::VIRTIO_BLK, "VIRTIO_BLK", },
	{ DOCA_DEV_REP_FILTER_NVME, PRIV_DOCA_REP_NVME, emulation_type::NVME, "NVME", },
};

/* Converts an emulation type into its matching representor type. */
static inline enum priv_doca_rep_type emu_type_to_rep_type(emulation_type type)
{
	for (unsigned int i = 0; i < ARRAY_SIZE(doca_dev_rep_convert_arr); i++) {
		if (doca_dev_rep_convert_arr[i].emulation_type == type)
			return doca_dev_rep_convert_arr[i].rep_type;

	}

	return PRIV_DOCA_REP_UNKNOWN;
}

void
dev_hold(struct doca_dev *dev)
{
	dev->refcount++;
}

doca_error_t
dev_put(struct doca_dev *dev)
{
	std::lock_guard<std::mutex> guard(dev_list_lock);

	if(dev->refcount.fetch_sub(1) <= 1) {
		for (auto &observer : dev->observers) { // dev->observers can't be changed at this point (acquired the device lock)
			if (observer.ops.on_dev_close)
				// no possible deadlock (DOCA device APIs are not allowed from within the observer CBs)
				observer.ops.on_dev_close(dev, observer.cookie);
		}

		if (dev->pd) {
			if (dev->pd->destroy() != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("Unable to close device: Failed to deallocate PD for dev %s",
				       dev->device_adapter->get_name().c_str());
				return DOCA_ERROR_IN_USE;
			}
			dev->pd.reset();
		}

		if (dev->device_adapter->close() != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Unable to close device: Failed to release context for %s",
				       dev->device_adapter->get_name().c_str());
			return DOCA_ERROR_IN_USE;
		}

		global_dev_list.remove(dev);
		delete dev;
	}

	return DOCA_SUCCESS;
}

/**
 * This method expects that dev_list_lock is locked
 * @param [in] device handles
 */
doca_error_t
global_dev_list_add_new_devices(std::vector<std::shared_ptr<device_abstraction::idevice_adapter>> &device_adapters)
{
	try {
		for (auto device_handle : device_adapters) {
			auto found_it = std::find_if(
				global_dev_list.begin(), global_dev_list.end(),
						[&](doca_dev *current_device) {
							return ((*current_device->device_adapter) == (*device_handle));
						});

			// Adding the device only if it is not in the global device list.
			if (found_it == global_dev_list.end())
			{
				auto new_device = new doca_dev();

				// doca_dev doesn't have a constructor at this point to keep changes minimal
				new_device->device_adapter = device_handle;
				new_device->info.dev = new_device;
				global_dev_list.push_back(new_device);
			}
		}
	}
	catch (...) {
		DOCA_DLOG_ERR("Unable to create global device list: Failed to allocate new doca device");
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

doca_error_t
reserve_extra_rep_devs(std::vector<struct doca_devinfo_rep> &rep_devs, size_t extra)
{
	const size_t min_size = rep_devs.size() + extra;

	try {
		rep_devs.reserve(min_size);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to reserve space for additional representor devices."
			      "Attempted to increase capacity from %zu to %zu.",
			      rep_devs.capacity(), min_size);
		return DOCA_ERROR_NO_MEMORY;
	}
	return DOCA_SUCCESS;
}

/**
 * This method expects that dev_list_lock is locked
 * @param [in] device handles
 */
void
global_dev_list_mark_zombie_devices(std::vector<std::shared_ptr<device_abstraction::idevice_adapter>> device_adapters)
{
	try {
		for (auto current_device : global_dev_list) {
			if (std::find_if(device_adapters.begin(), device_adapters.end(),
					 [&](std::shared_ptr<device_abstraction::idevice_adapter> device_adapter) {
					return ((*current_device->device_adapter) == (*device_adapter));
				}) == device_adapters.end()) {
				current_device->zombie = true;
			}
			else {
				current_device->zombie = false;
				// This means that every time that this method is called the ref count will be incremented. Possible bug.
				dev_hold(current_device);
			}
		}
	}
	catch (...) {
		return;
	}
}

struct devinfo_list_priv *
create_list_priv(size_t num_devices)
{
	auto length = sizeof(struct devinfo_list_priv) - sizeof(struct doca_devinfo *) + /* metadata only */
		      (num_devices * sizeof(struct doca_devinfo *));
	try {
		auto list_priv_buf = new uint8_t[length]();
		auto list_priv = reinterpret_cast<struct devinfo_list_priv *>(list_priv_buf);
		list_priv->nb_devs = static_cast<uint32_t>(num_devices);

		return list_priv;
	} catch (...) {
		return nullptr;
	}
}

inline struct devinfo_list_priv *
doca_devinfo_list_to_priv(struct doca_devinfo **devinfo_list)
{
	return (struct devinfo_list_priv *)PTR_SUB(devinfo_list,
			(sizeof(struct devinfo_list_priv) - sizeof(devinfo_list_priv::devinfos)));
}

void
destroy_list_priv(struct devinfo_list_priv *list_priv)
{
	auto list_priv_buf = reinterpret_cast<uint8_t *>(list_priv);

	delete[] list_priv_buf;
}

/**
 * Adds network representors to end of @rep_devs.
 */
doca_error_t
net_representors_add(struct doca_dev *dev, std::vector<struct doca_devinfo_rep> &rep_devs)
{
	/* First check if the capability exists for the device. */
	auto devinfo = doca_dev_as_devinfo(dev);
	uint8_t net_supported{0};
	doca_error_t ret = doca_devinfo_rep_get_is_list_net_supported(devinfo, &net_supported);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (net_supported == 0) {
		DOCA_DLOG_ERR("Device does not support network representor device discovery.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	auto device_adapter = dev->device_adapter;
	std::vector<function_info> function_info_list;
	std::tie(ret, function_info_list) = device_adapter->get_net_pci_topology();
	if (ret != DOCA_SUCCESS) {
		return ret;
	}

	/* reserve extra space for representors */
	ret = reserve_extra_rep_devs(rep_devs, function_info_list.size());
	if (ret != DOCA_SUCCESS) {
		return ret;
	}

	size_t nb_new_devs = rep_devs.size();
	/* Add representors */
	for (const auto& info : function_info_list) {

		rep_devs.emplace_back(
			dev,
			info.func_type,
			PRIV_DOCA_REP_NET,
			info.pf_vhca_id, /* pf_vhca_id */
			info.vhca_id, /* vhca_id */
			info.bdf, /* pci_bdf */
			info.function_id, /* vport */
			0, /* introspection_mkey*/
			info.total_vf_cnt, /* nb_total_vfs */
			info.vf_cnt, /* nb_vfs */
			false /* hotplug */
		);
	}

	const auto dev_name = device_adapter->get_name();

	nb_new_devs = rep_devs.size() - nb_new_devs;
	DOCA_DLOG_INFO("Found a total of %zu vhca IDs for dev %s", nb_new_devs, dev_name.c_str());

	if (nb_new_devs == 0) {
		DOCA_DLOG_ERR("No net representors were found for device %s", dev_name.c_str());
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	return DOCA_SUCCESS;
}

/**
 *  Creates a list of (PF, VF list) pairs that represent all emulated functions of type opmod.
 *  emulated_pf_info_list_destroy() must be called with the same variables. Even if the create fails.
 */
doca_error_t
emulated_fns_add(struct doca_dev *dev, emulation_type emu_type, std::vector<struct doca_devinfo_rep> &rep_devs)
{
	auto device_adapter = dev->device_adapter;

	/* First get list of emulated physical functions.*/
	doca_error_t ret{DOCA_SUCCESS};
	std::vector<emulated_devinfo> emulated_pf_list;
	std::tie(ret, emulated_pf_list) = device_adapter->create_emulated_pf_list(emu_type);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (emulated_pf_list.size() == 0) {
		DOCA_DLOG_INFO("No emulated devices of type %d were received.", (int)emu_type);
		return DOCA_SUCCESS;
	}

	/* Reserve enough space for all potential PFs and VFs. */
	size_t nb_new_devs = emulated_pf_list.size();

	// Can be tighter but with more allocations, by reserve space only for *valid* VFs after querying *each* PF
	for (const auto& pf_info : emulated_pf_list)
		nb_new_devs += pf_info.nb_total_vfs;

	ret = reserve_extra_rep_devs(rep_devs, nb_new_devs);
	if (ret != DOCA_SUCCESS)
		return ret;

	/* Then for each PF create its list of VFs.*/
	for (const auto& pf_info : emulated_pf_list) {
		/* Populate information about the PF first.*/
		rep_devs.emplace_back(pf_info, dev, DOCA_PCI_FUNC_PF, /*nb_vfs=*/0,
				      emu_type_to_rep_type(emu_type), /*vport=*/0 , /*intro_mkey=*/0);

		if (pf_info.nb_total_vfs == 0)
			continue;

		/* Populate the VFs in the list.*/
		std::vector<emulated_devinfo> emulated_vf_list{};
		std::tie(ret, emulated_vf_list) = device_adapter->create_emulated_vf_list(pf_info.vhca_id, pf_info.nb_total_vfs);
		if (ret != DOCA_SUCCESS)
			return ret;

		/* After VFs created from PF are successfully found, update number of VFs for the PF.*/
		rep_devs.back().nb_vfs = static_cast<uint16_t>(emulated_vf_list.size());

		/* add each VF as representor device. */
		for (const auto& vf_info : emulated_vf_list)
			rep_devs.emplace_back(vf_info, dev, DOCA_PCI_FUNC_VF, /*nb_vfs=*/0,
				      emu_type_to_rep_type(emu_type), /*vport=*/0 , /*intro_mkey=*/0);
	}

	return DOCA_SUCCESS;
}
} /* namespace */

/** Internal method for checking if a representor type is supported.
 *  Implementation is based on public API for checking capability, to ensure consistency.
 */
bool
is_rep_filter_supported(const struct doca_devinfo *devinfo, enum doca_dev_rep_filter rep_filter)
{
	doca_error_t ret;
	uint8_t is_supported{0};

	if (devinfo == nullptr) {
		DOCA_DLOG_ERR(
			"Unable to get is representor filter supported: Received invalid input - doca_devinfo=%p",
			devinfo);
		return false;
	}

	switch (rep_filter) {
	case DOCA_DEV_REP_FILTER_NET:
		ret = doca_devinfo_rep_get_is_list_net_supported(devinfo, &is_supported);
		break;
	case DOCA_DEV_REP_FILTER_VIRTIO_FS:
		ret = doca_devinfo_rep_get_is_list_virtio_fs_supported(devinfo, &is_supported);
		break;
	case DOCA_DEV_REP_FILTER_VIRTIO_NET:
		ret = doca_devinfo_rep_get_is_list_virtio_net_supported(devinfo, &is_supported);
		break;
	case DOCA_DEV_REP_FILTER_VIRTIO_BLK:
		ret = doca_devinfo_rep_get_is_list_virtio_blk_supported(devinfo, &is_supported);
		break;
	case DOCA_DEV_REP_FILTER_NVME:
		ret = doca_devinfo_rep_get_is_list_nvme_supported(devinfo, &is_supported);
		break;
	default:
		ret = DOCA_ERROR_INVALID_VALUE;
	}

	return ret == DOCA_SUCCESS && is_supported != 0;
}
doca_error_t
doca_devinfo_list_create(struct doca_devinfo ***devinfo_list, uint32_t *nb_devs)
{
	if (nb_devs == nullptr || devinfo_list == nullptr) {
		DOCA_DLOG_ERR(
			"Unable to create list of local devices: Received invalid input - devinfo_list=%p, nb_devs=%p",
			devinfo_list, nb_devs);
		return DOCA_ERROR_INVALID_VALUE;
	}
	// The scope of this try is large, but since all the methods below uses device adapters then this is the
	// minimal possible try scope size.
	doca_error_t status{DOCA_SUCCESS};
	std::vector<std::shared_ptr<idevice_adapter>> devices;
	std::tie(status, devices) = device_abstraction::get_devices();
	if (status != DOCA_SUCCESS) {
		// Empty devices list may occur and is not considered an error.
		// Device abstraction will set the status out param to an error value if it fails
		// to get the devices list or fails to allocate memory.
		DOCA_DLOG_ERR("Unable to create list of local devices: Failed to get RDMA devices list. Maybe missing RDMA kernel support?");
		return status;
	}
	/* Allocate first so we don't fail after dev list update.*/
	auto list_priv = create_list_priv(devices.size());
	if (list_priv == NULL) {
		DOCA_DLOG_ERR(
			       "Unable to create list of local devices: Failed to allocate memory for doca devinfo list"
			       );
		return DOCA_ERROR_NO_MEMORY;
	}
	std::lock_guard<std::mutex> guard(dev_list_lock);
	status = global_dev_list_add_new_devices(devices);
	if(status != DOCA_SUCCESS) {
		destroy_list_priv(list_priv);
		return status;
	}
	global_dev_list_mark_zombie_devices(devices);
	int curr_idx = 0;
	for (auto current_dev : global_dev_list) {
		if (!current_dev->zombie) {
			list_priv->devinfos[curr_idx++] = &current_dev->info;
		}
		if (current_dev->zombie)
			continue;
	}
	*nb_devs = static_cast<uint32_t>(global_dev_list.size());
	*devinfo_list = &list_priv->devinfos[0];
	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_list_destroy(struct doca_devinfo **devinfo_ptr_list)
{
	struct devinfo_list_priv *list_priv;
	uint32_t dev_idx, nb_devs;
	doca_error_t err;

	if (devinfo_ptr_list == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	list_priv = doca_devinfo_list_to_priv(devinfo_ptr_list);
	nb_devs = list_priv->nb_devs;
	for (dev_idx = 0; dev_idx < nb_devs; dev_idx++) {
		if (devinfo_ptr_list[dev_idx]) {
			err = dev_put(devinfo_ptr_list[dev_idx]->dev);
			if (err != DOCA_SUCCESS) {
				DOCA_DLOG_ERR(
					"Unable to destroy list of local devices: Failed to free device at index %u",
					dev_idx);
				return err;
			}
		}
	}

	destroy_list_priv(list_priv);

	return DOCA_SUCCESS;
}

// Assumption: This method is called under a thread safe flow (therefore di can not be deleted during the call
device_abstraction::idevice_adapter *priv_doca_devinfo_to_device_handle(const struct doca_devinfo* di)
{
	if (unlikely(di == NULL || di->dev == NULL))
		return nullptr;

	return di->dev->device_adapter.get();
}

doca_error_t
doca_devinfo_rep_list_create(struct doca_dev *dev, int filter, struct doca_devinfo_rep ***dev_list_rep,
			     uint32_t *nb_devs_rep)
{
	struct doca_devinfo_rep **dev_list{nullptr};
	struct doca_devinfo_rep *dev_list_flat{nullptr};
	/* Using vector for faster memory traversal, and since element insertion is done at back only.*/
	std::vector<struct doca_devinfo_rep> rep_devs;

	if (dev == nullptr || dev_list_rep == nullptr || nb_devs_rep == nullptr) {
		DOCA_DLOG_ERR("Received invalid input dev=%p dev_list_rep=%p nb_devs_rep=%p.",
			      dev, dev_list_rep, nb_devs_rep);

		return DOCA_ERROR_INVALID_VALUE;
	}
	if ((filter & filters_all) != filter) {
		DOCA_DLOG_ERR("Received invalid filter combination filters=%d, filters_all=%d", filter, filters_all);
		return DOCA_ERROR_INVALID_VALUE;
	}

	uint8_t all_supported{0};
	const struct doca_devinfo* devinfo = doca_dev_as_devinfo(dev);
	doca_error_t ret = doca_devinfo_rep_get_is_list_all_supported(devinfo, &all_supported);
	if (ret != DOCA_SUCCESS)
		return ret;

	if (all_supported == 0) {
		DOCA_DLOG_ERR("None of the representor device types are supported.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	/* Add all network representors to the array.*/
	if (filter == DOCA_DEV_REP_FILTER_ALL) {
		if (!is_rep_filter_supported(devinfo, DOCA_DEV_REP_FILTER_NET))
			DOCA_DLOG_WARN("Unsupported emulation device requested NET");
		else
			static_cast<void>(net_representors_add(dev, rep_devs));
	} else if ((static_cast<int>(DOCA_DEV_REP_FILTER_NET) & filter) != 0) {
		if (!is_rep_filter_supported(devinfo, DOCA_DEV_REP_FILTER_NET)) {
			DOCA_DLOG_ERR("Unsupported emulation device requested NET");
			return DOCA_ERROR_NOT_SUPPORTED;
		}

		ret = net_representors_add(dev, rep_devs);
		if (ret != DOCA_SUCCESS)
			return ret;
	}

	/* For each emulated function type, add all physical functions and their virtual functions.*/
	for (int i = 0; i < ARRAY_SIZE(doca_dev_rep_convert_arr); i++) {
		enum doca_dev_rep_filter filter_type = doca_dev_rep_convert_arr[i].rep_filter;
		emulation_type emulation_type = doca_dev_rep_convert_arr[i].emulation_type;
		const char *emulation_descr = doca_dev_rep_convert_arr[i].descr;
		(void)emulation_descr;

		if (filter == DOCA_DEV_REP_FILTER_ALL) {
			if (!is_rep_filter_supported(devinfo, filter_type)) {
				DOCA_DLOG_WARN("Unsupported emulation device requested %s",
					       emulation_descr);
				continue;
			}

			(void)emulated_fns_add(dev, emulation_type, rep_devs);
		} else if (static_cast<int>(filter_type) & filter) {
			if (!is_rep_filter_supported(devinfo, filter_type)) {
				DOCA_DLOG_ERR("Unsupported emulation device requested %s",
					      emulation_descr);
				return DOCA_ERROR_NOT_SUPPORTED;
			}

			ret = emulated_fns_add(dev, emulation_type, rep_devs);
			if (ret != DOCA_SUCCESS)
				return ret;
		}
	}

	const size_t nb_total_devs = rep_devs.size();

	try {
		dev_list = new struct doca_devinfo_rep* [nb_total_devs + 1]();
		/* In case no devices found but the filters are supported we return an empty list. */
		if (nb_total_devs > 0)
			dev_list_flat = new struct doca_devinfo_rep [nb_total_devs]();
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate list of devinfo_rep. nb_total_devs=%zu, list=%p, flat_list=%p",
				rep_devs.size(), dev_list, dev_list_flat);
		delete []dev_list_flat;
		delete []dev_list;

		DOCA_DLOG_ERR("Unable to create list of representor devices: Failed to allocate sufficient memory");
		return DOCA_ERROR_NO_MEMORY;
	}

	/* At this stage it is not possible to fail. Increment refcount so dev is not destroyed while list exists.*/
	if (nb_total_devs > 0)
		dev_hold(dev);

	/* Copy all the emulated functions (PF/VF) to a single devinfo flat array.*/
	/* Then connect each pointer to its matching devinfo in the flat array.*/
	size_t dev_idx = 0;
	for (dev_idx = 0; dev_idx < nb_total_devs; dev_idx++) {
		dev_list_flat[dev_idx] = rep_devs[dev_idx];
		dev_list[dev_idx] = &dev_list_flat[dev_idx];
	}

	/* Last element points to NULL. */
	dev_list[dev_idx] = nullptr;

	*dev_list_rep = dev_list;
	*nb_devs_rep = static_cast<uint32_t>(nb_total_devs);

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_list_destroy(struct doca_devinfo_rep **dev_list_rep)
{
	if (dev_list_rep == nullptr) {
		DOCA_DLOG_ERR("Received invalid input devinfo_list_rep=nullptr");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (*dev_list_rep != nullptr) {
		/* The list no longer needs to hold the parent device to exist.*/
		const doca_error_t dev_put_err = dev_put((*dev_list_rep)->dev_parent);

		if (dev_put_err != DOCA_SUCCESS) {
			DOCA_DLOG_ERR(
				"Unable to destroy list of representor devices: Failed to decrement refcount for parent device - dev=%p",
				(*dev_list_rep)->dev_parent);
			return dev_put_err;
		}

		/* The first element points to the start of the flat list array (see doca_devinfo_rep_list_create).*/
		delete []*dev_list_rep;
		*dev_list_rep = nullptr;
	}

	delete []dev_list_rep;

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_vuid(const struct doca_devinfo_rep *devinfo_rep, char *rep_vuid, uint32_t size)
{
	if ((devinfo_rep == NULL) || (rep_vuid == NULL)) {
		DOCA_DLOG_ERR(
			"Unable to get representor device vuid: Received invalid input - devinfo_rep=%p, rep_vuid=%p",
			devinfo_rep, rep_vuid);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (size != device_abstraction::vuid_t::vuid_bytes) {
		DOCA_DLOG_ERR("Unable to get representor device vuid: vuid is of size %dByte",
			DOCA_DEVINFO_REP_VUID_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}
	device_abstraction::vuid_t vuid{};
	auto status = devinfo_rep->dev_parent->device_adapter->query_vuid(devinfo_rep->vhca_id, vuid);
	if (status == doca_error_t::DOCA_SUCCESS) {
		memcpy(rep_vuid, vuid.id, DOCA_DEVINFO_REP_VUID_SIZE);
	}

	return status;
}

doca_error_t
doca_dev_open(struct doca_devinfo *devinfo, struct doca_dev **dev_out)
{
	struct doca_dev *dev;

	if (devinfo == NULL || dev_out == NULL) {
		DOCA_DLOG_ERR("Unable to open local device: Received invalid input - devinfo=%p, dev_out=%p",
			devinfo, dev_out);
		return DOCA_ERROR_INVALID_VALUE;
	}

	dev = devinfo->dev;
	if (dev == NULL) {
		DOCA_DLOG_ERR("Unable to open local device: Received invalid devinfo - dev is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	std::lock_guard<std::mutex> guard(dev_list_lock);
	doca_error status = dev->device_adapter->open();
	if(status != doca_error::DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to open local device: Failed to open following device %s",
			dev->device_adapter->get_name().c_str());
		return status;
	}

	if (!dev->pd) {
		dev->pd = dev->device_adapter->create_protection_domain();
		if (!dev->pd) {
			DOCA_DLOG_ERR("Unable to open local device: Failed to allocate PD for device %s",
				dev->device_adapter->get_name().c_str());
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	dev_hold(dev);
	*dev_out = dev;
	return DOCA_SUCCESS;
}

doca_error_t
doca_dev_close(struct doca_dev *dev)
{
	if (dev == NULL)
		return DOCA_SUCCESS;
	return dev_put(dev);
}

doca_error_t
doca_dev_rep_open(struct doca_devinfo_rep *devinfo, struct doca_dev_rep **dev_rep)
{
	struct doca_dev_rep *dev;

	if (devinfo == NULL || dev_rep == NULL) {
		DOCA_DLOG_ERR("Unable to open representor device: Received invalid input - devinfo=%p, dev_rep=%p",
			devinfo, dev_rep);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (devinfo->dev_parent == NULL) {
		DOCA_DLOG_ERR(
			"Unable to open representor device: Received representor device with invalid parent_dev=NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	try {
		dev = new doca_dev_rep();
	} catch (...) {
		DOCA_DLOG_ERR("Unable to open representor device: Failed to allocate memory for representor device size=%zu",
			sizeof(*dev));
		return DOCA_ERROR_NO_MEMORY;
	}

	dev_hold(devinfo->dev_parent);

	dev->info = *devinfo;

	*dev_rep = dev;

	return DOCA_SUCCESS;
}

doca_error_t
doca_dev_rep_close(struct doca_dev_rep *dev)
{
	doca_error_t err;

	if (dev == NULL)
		return DOCA_SUCCESS;

	if (dev->info.dev_parent == NULL) {
		DOCA_DLOG_ERR("Unable to close representor device: Received representor device with invalid dev_parent=NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	err = dev_put(dev->info.dev_parent);
	if (err != DOCA_SUCCESS)
		return err;

	delete dev;

	return err;
}

struct doca_devinfo *
doca_dev_as_devinfo(const struct doca_dev *dev)
{
	if (unlikely(dev == NULL))
		return NULL;
	return const_cast<struct doca_devinfo *>(&dev->info);
}

struct doca_devinfo_rep *
doca_dev_rep_as_devinfo(struct doca_dev_rep *dev_rep)
{
	if (unlikely(dev_rep == NULL))
		return NULL;
	return &dev_rep->info;
}

#ifdef __linux__
struct ibv_context *
priv_doca_devinfo_to_ibv_ctx(const struct doca_devinfo *di)
{
	if (unlikely(di == NULL))
		return NULL;
	return priv_doca_dev_to_ibv_ctx(di->dev);
}

struct ibv_context *
priv_doca_dev_to_ibv_ctx(const struct doca_dev *dev)
{
	if (unlikely(dev == NULL || dev->device_adapter == NULL))
		return NULL;

	return dev->device_adapter->get_ibv_context();
}

#endif // __linux__

doca_error_t
priv_doca_dev_pd_id_get(const struct doca_dev *dev, uint32_t *pd_id)
{
	if (unlikely(dev == nullptr || pd_id == nullptr || dev->pd == nullptr))
		return DOCA_ERROR_INVALID_VALUE;

	*pd_id = dev->pd->get_pd_id();
	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_get_pci_addr(const struct doca_devinfo *devinfo, struct doca_pci_bdf *pci_addr)
{
	if ((devinfo == NULL) || (pci_addr == NULL)) {
		DOCA_DLOG_ERR("Unable to get device pci address: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	uint8_t bus{0};
	uint8_t device{0};
	uint8_t function{0};
	auto status = devinfo->dev->device_adapter->query_pci_bdf(bus, device, function);
	if (status == DOCA_SUCCESS) {
		doca_pci_bdf *doca_bdf = pci_addr;
		doca_bdf->bus = bus;
		doca_bdf->device = device;
		doca_bdf->function = function;
	}

	return status;
}

doca_error_t
doca_devinfo_get_ipv4_addr(const struct doca_devinfo *devinfo,
			   uint8_t *ipv4_addr, uint32_t size)
{
	if ((devinfo == NULL) || (ipv4_addr == NULL)) {
		DOCA_DLOG_ERR("Unable to get device ipv4 address: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (size < DOCA_DEVINFO_IPV4_ADDR_SIZE) {
		DOCA_DLOG_ERR("Unable to get ipv4 address: Address is of size %dByte",
			DOCA_DEVINFO_IPV4_ADDR_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return devinfo->dev->device_adapter->query_ipv4_address(ipv4_addr, size);;
}

doca_error_t doca_devinfo_get_ipv6_addr(const struct doca_devinfo *devinfo,
					uint8_t *ipv6_addr, uint32_t size)
{
	if ((devinfo == NULL) || (ipv6_addr == NULL)) {
		DOCA_DLOG_ERR("Unable to get device ipv6 address: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (size < DOCA_DEVINFO_IPV6_ADDR_SIZE) {
		DOCA_DLOG_ERR("Unable to get ipv6 address: Address is of size %dByte",
			DOCA_DEVINFO_IPV6_ADDR_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return devinfo->dev->device_adapter->query_ipv6_address(ipv6_addr, size);
}

doca_error_t
doca_devinfo_get_iface_name(const struct doca_devinfo *devinfo,
			    char *iface_name, uint32_t size)
{
	if ((devinfo == NULL) || (iface_name == NULL)) {
		DOCA_DLOG_ERR("Unable to get device iface name: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (size < DOCA_DEVINFO_IFACE_NAME_SIZE) {
		DOCA_DLOG_ERR("Unable to get iface name: iface name is of size %dByte",
			DOCA_DEVINFO_IFACE_NAME_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}

	std::string value;
	auto status = devinfo->dev->device_adapter->get_interface_name(value);
	if (status == DOCA_SUCCESS) {
		try {
			auto length = value.copy(iface_name, value.length());
			iface_name[length] = '\0';
		} catch (...) {
			DOCA_DLOG_ERR("Failed to copy interface name");
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	return status;
}

doca_error_t
doca_devinfo_get_ibdev_name(const struct doca_devinfo *devinfo,
			    char *ibdev_name, uint32_t size)
{
	if ((devinfo == NULL) || (ibdev_name == NULL)) {
		DOCA_DLOG_ERR("Unable to get device ibdev name: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (size < DOCA_DEVINFO_IBDEV_NAME_SIZE) {
		DOCA_DLOG_ERR("Unable to get ibdev name: ibdev is of size %dByte",
			DOCA_DEVINFO_IBDEV_NAME_SIZE);
		return DOCA_ERROR_INVALID_VALUE;
	}

	std::string name;
	auto status = devinfo->dev->device_adapter->get_ibdev_name(name);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to get device ibdev name with status %s", doca_get_error_name(status));
		return status;
	}

	try {
		auto length = name.copy(ibdev_name, name.length());
		ibdev_name[length] = '\0';
	} catch (...) {
		DOCA_DLOG_ERR("Failed to copy device name");
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_get_is_hotplug_manager_supported(const struct doca_devinfo *devinfo,
				    	      uint8_t *is_hotplug_manager)
{
	if ((devinfo == NULL) || (is_hotplug_manager == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_hotplug_manager flag value: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto status_and_hca_caps = devinfo->dev->device_adapter->query_hca_caps();
	if (status_and_hca_caps.first != doca_error::DOCA_SUCCESS)
		return status_and_hca_caps.first;

	*is_hotplug_manager = static_cast<uint8_t>((status_and_hca_caps.second).hotplug_manager);

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_get_is_mmap_export_supported(const struct doca_devinfo *devinfo,
					  uint8_t *mmap_export)
{
	if ((devinfo == NULL) || (mmap_export == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_mmap_export: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto status_and_hca_caps = devinfo->dev->device_adapter->query_hca_caps();
	if (status_and_hca_caps.first != doca_error::DOCA_SUCCESS)
		return status_and_hca_caps.first;

	*mmap_export = static_cast<uint8_t>((status_and_hca_caps.second).cross_gvmi_mkey_enabled);

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_get_is_mmap_from_export_supported(const struct doca_devinfo *devinfo,
				     	       uint8_t *from_export)
{
	if ((devinfo == NULL) || (from_export == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_mmap_from_export flag value: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto status_and_hca_caps = devinfo->dev->device_adapter->query_hca_caps();
	if (status_and_hca_caps.first != doca_error::DOCA_SUCCESS)
		return status_and_hca_caps.first;

	*from_export = static_cast<uint8_t>((status_and_hca_caps.second).cross_gvmi_mkey_enabled);

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_is_list_all_supported(const struct doca_devinfo *devinfo, uint8_t *all_supported)
{
	if ((devinfo == NULL) || (all_supported == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_rep_list_all supported: Received invalid input - devinfo=%p, all_supported=%p",
			      devinfo, all_supported);
		return DOCA_ERROR_INVALID_VALUE;
	}

	/* Check if net representor type is supported. */
	if (is_rep_filter_supported(devinfo, DOCA_DEV_REP_FILTER_NET)) {
		*all_supported = 1;
		return DOCA_SUCCESS;
	}

	/* Check if all other filter types are supported. */
	for (int i = 0; i < ARRAY_SIZE(doca_dev_rep_convert_arr); i++) {
		enum doca_dev_rep_filter filter_type = doca_dev_rep_convert_arr[i].rep_filter;

		if (!is_rep_filter_supported(devinfo, filter_type))
			continue;

		/* In case any type is supported then we return true. */
		*all_supported = 1;
		return DOCA_SUCCESS;
	}

	*all_supported = 0;

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_is_list_net_supported(const struct doca_devinfo *devinfo,
				 	   uint8_t *net_supported)
{
	if ((devinfo == NULL) || (net_supported == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_rep_list_net flag value: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*net_supported = static_cast<uint8_t>(devinfo->dev->device_adapter->is_net_representors_supported());

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_is_list_virtio_fs_supported(const struct doca_devinfo *devinfo,
				       		 uint8_t *virtio_fs_supported)
{
	if ((devinfo == NULL) || (virtio_fs_supported == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_rep_list_virtio_fs flag value: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto status_and_hca_caps = devinfo->dev->device_adapter->query_hca_caps();
	if (status_and_hca_caps.first != doca_error::DOCA_SUCCESS)
		return status_and_hca_caps.first;

	*virtio_fs_supported = static_cast<uint8_t>((status_and_hca_caps.second).emulation_manager_virtio_fs);

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_is_list_virtio_net_supported(const struct doca_devinfo *devinfo,
						  uint8_t *virtio_net_supported)
{
	if ((devinfo == NULL) || (virtio_net_supported == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_rep_list_virtio_net flag value: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto status_and_hca_caps = devinfo->dev->device_adapter->query_hca_caps();
	if (status_and_hca_caps.first != doca_error::DOCA_SUCCESS)
		return status_and_hca_caps.first;

	*virtio_net_supported = static_cast<uint8_t>((status_and_hca_caps.second).emulation_manager_virtio_net);

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_is_list_virtio_blk_supported(const struct doca_devinfo *devinfo,
						  uint8_t *virtio_blk_supported)
{
	if ((devinfo == NULL) || (virtio_blk_supported == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_rep_list_virtio_blk flag value: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto status_and_hca_caps = devinfo->dev->device_adapter->query_hca_caps();
	if (status_and_hca_caps.first != doca_error::DOCA_SUCCESS)
		return status_and_hca_caps.first;

	*virtio_blk_supported = static_cast<uint8_t>((status_and_hca_caps.second).emulation_manager_virtio_blk);

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_is_list_nvme_supported(const struct doca_devinfo *devinfo,
				  	    uint8_t *nvme_supported)
{
	if ((devinfo == NULL) || (nvme_supported == NULL)) {
		DOCA_DLOG_ERR("Unable to get device is_rep_list_nvme flag value: Received invalid input - devinfo=%p",devinfo);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto status_and_hca_caps = devinfo->dev->device_adapter->query_hca_caps();
	if (status_and_hca_caps.first != doca_error::DOCA_SUCCESS)
		return status_and_hca_caps.first;

	*nvme_supported = static_cast<uint8_t>((status_and_hca_caps.second).emulation_manager_nvme);

	return DOCA_SUCCESS;
}

uint16_t
priv_doca_dev_vhca_id_get(const struct doca_dev_rep *dev)
{
	return dev->info.vhca_id;
}

doca_error_t
priv_doca_devinfo_rep_vhca_id_get(const struct doca_devinfo_rep *di, uint16_t *vhca_id)
{
	if ((di == NULL) || (vhca_id == NULL)) {
		DOCA_DLOG_ERR("Illegal NULL parameter (di=%p, vhca_id=%p)", di, vhca_id);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*vhca_id = di->vhca_id;

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_rep_is_hotplug_get(const struct doca_dev_rep *dev, bool *is_hotplug)
{
	if ((dev == NULL) || (is_hotplug == NULL)) {
		DOCA_DLOG_ERR("Illegal NULL parameter (dev=%p, is_hotplug=%p)", dev, is_hotplug);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*is_hotplug = dev->info.hotplug;

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_prm_cmd_send(const struct doca_dev *dev, const void *in, size_t in_size, void *out, size_t out_size)
{
	if ((dev == NULL) || (in == NULL) || (in_size == 0) || (out == NULL) || (out_size == 0)) {
		DOCA_DLOG_ERR("Illegal parameter (dev=%p, in=%p, in_size=%zu, out=%p, out_size=%zu)",
			dev, in, in_size, out, out_size);
		return DOCA_ERROR_INVALID_VALUE;
	}
	return dev->device_adapter->send_prm_cmd(in, in_size, out, out_size);
}

doca_error_t
priv_doca_dev_devx_object_create(const struct doca_dev *dev, const void *in, size_t in_size, void *out, size_t out_size,
				 struct doca_devx_object **devx_obj)
{
	if ((dev == nullptr) || (in == nullptr) || (in_size == 0) || (out == nullptr) || (out_size == 0) ||
	     (devx_obj == nullptr)) {
		DOCA_DLOG_ERR("Illegal parameter (dev=%p, in=%p, in_size=%zu, out=%p, out_size=%zu, devx_obj=%p)", dev,
			       in, in_size, out, out_size, devx_obj);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::unique_ptr<idevx_object> devx_object;
	std::tie(status, devx_object) = dev->device_adapter->create_devx_object(in, in_size, out, out_size);
	if (status == DOCA_SUCCESS) {
		(*devx_obj) = new (std::nothrow) struct doca_devx_object;
		if (*devx_obj) {
			(*devx_obj)->devx_object = std::move(devx_object);
			return DOCA_SUCCESS;
		}
		DOCA_DLOG_ERR("Failed to allocate devx object wrapper");
		return DOCA_ERROR_NO_MEMORY;
	}
	return status;
}

doca_error_t
priv_doca_devx_object_destroy(struct doca_devx_object *devx_obj)
{
	if (devx_obj == nullptr){
		DOCA_DLOG_ERR("Illegal parameter (devx_obj is null)");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (devx_obj->devx_object) {
		auto status = devx_obj->devx_object->destroy();
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to destroy devx object %p", devx_obj);
			return status;
		}
	}

	delete devx_obj;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_devx_object_modify(struct doca_devx_object *devx_obj, const void *in, size_t in_size, void *out,
				 size_t out_size)
{
	if (devx_obj == nullptr) {
		DOCA_DLOG_ERR("Illegal parameter (devx_obj wrapper is null)");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (!devx_obj->devx_object) {
		DOCA_DLOG_ERR("devx object is null");
		return DOCA_ERROR_BAD_STATE;
	}

	return devx_obj->devx_object->modify(in, in_size, out, out_size);
}

DOCA_PRIVATE
doca_error_t
priv_doca_devx_object_get(struct doca_devx_object *devx_obj, void **obj)
{
	if (devx_obj == nullptr) {
		DOCA_DLOG_ERR("Illegal parameter (devx_obj wrapper is null)");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (!devx_obj->devx_object) {
		DOCA_DLOG_ERR("devx object is null");
		return DOCA_ERROR_BAD_STATE;
	}

	*obj = devx_obj->devx_object->get_object();
	return DOCA_SUCCESS;
}


doca_error_t
priv_doca_umem_create(const struct doca_dev *dev, void *address, size_t size, uint32_t access_flags,
		      struct doca_umem_object **umem_obj)
{
	if ((dev == NULL) || (address == NULL) || (size == 0) || (umem_obj== NULL)) {
		DOCA_DLOG_ERR("Illegal parameter (dev=%p, address=%p, size=%zu, umem_obj=%p)", dev, address, size,
			      umem_obj);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<imapped_user_memory> umem_object;
	std::tie(status, umem_object) = dev->device_adapter->create_mapped_user_memory(address, size, access_flags);
	if (status == DOCA_SUCCESS) {
		(*umem_obj) = new (std::nothrow) struct doca_umem_object;
		if (*umem_obj) {
			(*umem_obj)->umem = umem_object;
			return DOCA_SUCCESS;
		}
		DOCA_DLOG_ERR("Failed to allocate umem wrapper");
		return DOCA_ERROR_NO_MEMORY;
	}
	return status;
}

doca_error_t
priv_doca_umem_create_pg_sz(const struct doca_dev *dev, void *address, size_t size, size_t page_size, uint32_t access_flags,
		      struct doca_umem_object **umem_obj)
{
	if ((dev == NULL) || (address == NULL) || (size == 0) || (umem_obj == NULL)) {
		DOCA_DLOG_ERR("Illegal parameter (dev=%p, address=%p, size=%zu, umem_obj=%p)", dev, address, size,
			      umem_obj);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<imapped_user_memory> umem_object;
	std::tie(status, umem_object) =
		dev->device_adapter->create_mapped_user_memory(address, size, page_size, access_flags);
	if (status == DOCA_SUCCESS) {
		(*umem_obj) = new (std::nothrow) struct doca_umem_object;
		if (*umem_obj) {
			(*umem_obj)->umem = umem_object;
			return DOCA_SUCCESS;
		}
		DOCA_DLOG_ERR("Failed to allocate umem wrapper");
		return DOCA_ERROR_NO_MEMORY;
	}
	return status;
}

doca_error_t
priv_doca_umem_destroy(struct doca_umem_object *umem_obj)
{
	if (umem_obj == nullptr) {
		DOCA_DLOG_ERR("Illegal parameter (umem_obj wrapper is null)");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (!umem_obj->umem) {
		DOCA_DLOG_ERR("unexpected, umem is null");
		return DOCA_ERROR_BAD_STATE;
	}

	auto status = umem_obj->umem->dereg();
	if (status != DOCA_SUCCESS) {
		return status;
	}

	delete umem_obj;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_umem_id_get(struct doca_umem_object *umem_obj, uint32_t *umem_id)
{
	if (umem_obj == nullptr) {
		DOCA_DLOG_ERR("Illegal parameter (umem_obj wrapper is null)");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (!umem_obj->umem) {
		DOCA_DLOG_ERR("Umem object is null");
		return DOCA_ERROR_BAD_STATE;
	}

	*umem_id = umem_obj->umem->get_id();
	return DOCA_SUCCESS;
}

doca_error_t priv_doca_devinfo_hca_caps_by_opmod_query(const struct doca_devinfo *devinfo, uint16_t op_mod,
						       void *out, size_t out_size)
{
	if ((devinfo == NULL) || (out == NULL) || (out_size == 0)) {
		DOCA_DLOG_ERR("Illegal parameter (devinfo=%p, opmod=%u, out=%p, out_size=%zu)",
			      devinfo, op_mod, out, out_size);
		return DOCA_ERROR_INVALID_VALUE;
	}
	return devinfo->dev->device_adapter->query_hca_caps_by_opmod(op_mod, out, out_size);
}

doca_error_t
priv_doca_dev_rep_parent_get(const struct doca_dev_rep *rep_dev, struct doca_dev **local_dev)
{
	if ((rep_dev == NULL) || (local_dev == NULL) ) {
		DOCA_DLOG_ERR("Illegal parameter (rep_dev=%p, local_dev=%p)", rep_dev, local_dev);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*local_dev = rep_dev->info.dev_parent;

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_rep_vport_get(const struct doca_dev_rep * const dev, uint16_t *vport)
{
	if ((dev == NULL) || (vport == NULL)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, vport=%p", dev, vport);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*vport = dev->info.vport;
	return DOCA_SUCCESS;
}

static doca_error_t
priv_doca_dev_rep_introspection_mkey_query(const struct doca_dev_rep *dev, uint32_t *mkey)
{
	if ((dev == NULL) || (mkey == NULL)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, mkey=%p", dev, mkey);
		return DOCA_ERROR_INVALID_VALUE;
	}

	struct doca_dev *parent;
	auto ret = priv_doca_dev_rep_parent_get(dev, &parent);
	if (ret != DOCA_SUCCESS)
		return ret;

	device_abstraction::hca_capabilities hca_caps{};
	auto device_adapter = parent->device_adapter;
	std::tie(ret, hca_caps) = device_adapter->query_hca_caps();
	if (ret != DOCA_SUCCESS)
		return ret;

	if (!hca_caps.introspection_mkey_access_allowed) {
		DOCA_DLOG_ERR("Introspection mkey access is not allowed");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	ret = device_adapter->query_function_hca_caps(dev->info.vport, hca_caps);
	if (ret != DOCA_SUCCESS)
		return ret;

	*mkey = hca_caps.introspection_mkey;

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_rep_introspection_mkey_get(const struct doca_dev_rep *const dev, uint32_t *mkey_out)
{
	if ((dev == NULL) || (mkey_out == NULL)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, mkey_out=%p", dev, mkey_out);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (dev->info.rep_type == PRIV_DOCA_REP_NET) {
		return priv_doca_dev_rep_introspection_mkey_query(dev, mkey_out);
	} else { // emulated device - due to bug, using the previously manually set (by the user) introspection mkey
		*mkey_out = dev->info.introspection_mkey;
		return DOCA_SUCCESS;
	}
}

doca_error_t
priv_doca_dev_rep_introspection_mkey_set(struct doca_dev_rep *dev, uint32_t mkey_in)
{
	if ((dev == NULL) || (mkey_in == 0)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, mkey_in=%u", dev, mkey_in);
		return DOCA_ERROR_INVALID_VALUE;
	}

	dev->info.introspection_mkey = mkey_in;
	return DOCA_SUCCESS;
}

doca_error_t
doca_dev_get_pd(const struct doca_dev *dev, struct ibv_pd **pd)
{
	if ((dev == NULL) || (pd == NULL)) {
		DOCA_DLOG_ERR("Invalid input: dev=%p, pd=%p", dev, pd);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (dev->pd == NULL) {
		DOCA_DLOG_ERR("pd handle is null");
		return DOCA_ERROR_BAD_STATE;
	}

	*pd = dev->pd->get_ibv_pd();

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_observer_register(struct doca_dev *dev, const struct priv_doca_dev_observer_ops *observer_ops,
				union doca_data cookie, priv_doca_dev_observer_handle_t *handle)
{
	if ((dev == nullptr) || (observer_ops == nullptr) || (handle == nullptr)) {
		DOCA_DLOG_ERR("Invalid input dev=%p, observer=%p, handle=%p", dev, observer_ops, handle);
		return DOCA_ERROR_INVALID_VALUE;
	}

	try {
		std::lock_guard<std::mutex> guard(dev_list_lock);
		auto iter = dev->observers.emplace(*observer_ops, cookie);
		if (!iter.second) {
			DOCA_DLOG_ERR("Failed to register an already-registered observer");
			return DOCA_ERROR_IN_USE;
		}
		*handle = reinterpret_cast<priv_doca_dev_observer_handle_t>(&*iter.first);
	}
	catch (...) {
		DOCA_DLOG_ERR("Failed to register a device observer");
		return DOCA_ERROR_NO_MEMORY;
	}

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_observer_unregister(struct doca_dev *dev, priv_doca_dev_observer_handle_t handle)
{
	if ((dev == nullptr) || (handle == (priv_doca_dev_observer_handle_t)nullptr)) {
		DOCA_DLOG_ERR("Invalid input dev=%p, observer=%lu", dev, (unsigned long)handle);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto observer = reinterpret_cast<struct priv_doca_dev_observer *>(handle);

	std::lock_guard<std::mutex> guard(dev_list_lock);
	if (dev->observers.erase(*observer) == 0)
		return DOCA_ERROR_NOT_FOUND;

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_pci_addr(const struct doca_devinfo_rep *devinfo_rep, struct doca_pci_bdf *pci_addr)
{
	if ((devinfo_rep == NULL) || (pci_addr == NULL)) {
		DOCA_DLOG_ERR(
		  "Unable to get representor device pci addr: Received invalid input - devinfo_rep=%p, pci_addr=%p",
		  	devinfo_rep, pci_addr);
		return DOCA_ERROR_INVALID_VALUE;
	}

	pci_addr->raw = devinfo_rep->pci_bdf;

	return DOCA_SUCCESS;
}

doca_error_t
doca_devinfo_rep_get_pci_func_type(const struct doca_devinfo_rep *devinfo_rep, enum doca_pci_func_type *pci_func_type)
{
	if ((devinfo_rep == NULL) || (pci_func_type == NULL)) {
		DOCA_DLOG_ERR(
		  "Unable to get representor device pci function type: Received invalid input - devinfo_rep=%p, pci_fn_type=%p",
		  	devinfo_rep, pci_func_type);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*pci_func_type = devinfo_rep->func_type;

	return DOCA_SUCCESS;
}
