/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <iomanip>
#include <unistd.h>
#include <stdint.h>
#include <cstring>
#include <sstream>
#include <unordered_map>
#include <rte_ethdev.h>
#include <doca_log.h>
#include <doca_rdma.h>
#include <doca_dev.h>
#include <priv_doca_dev.h>
#include "doca_dpdk.h"

DOCA_LOG_REGISTER(CORE::DOCA_DPDK);

static std::unordered_map<uint16_t, doca_dev *> port_to_dev;

static void
detach_port(uint16_t port_id)
{
	struct rte_eth_dev_info dev_info{0};

	if (rte_eth_dev_stop(port_id) != 0)
		DOCA_DLOG_ERR("Failed to stop DPDK port: %d", port_id);

	rte_eth_dev_close(port_id);

	if (rte_eth_dev_info_get(port_id, &dev_info) < 0) {
		DOCA_DLOG_ERR("Failed to get DPDK dev info of port: %d", port_id);
		return;
	}

	if (rte_dev_remove(dev_info.device) < 0)
		DOCA_DLOG_ERR("Failed to remove DPDK dev of port: %d", port_id);
}

static void
ports_cleanup_on_dev_close(const struct doca_dev *dev, union doca_data cookie)
{
	for (auto iter = port_to_dev.begin(); iter != port_to_dev.end();) {
		if (iter->second != dev) {
			iter++;
			continue;
		}

		detach_port(iter->first);

		iter = port_to_dev.erase(iter);
	}
}

static void
detach_ports(const char *identifier)
{
	uint16_t pi = RTE_MAX_ETHPORTS;
	struct rte_dev_iterator iter;

	RTE_ETH_FOREACH_MATCHING_DEV(pi, identifier, &iter) {
		detach_port(pi);
	}
}

static doca_error_t
extract_pci_bdf_identifier(const struct doca_dev *dev, std::ostringstream &pci_bdf_identifier)
{
	struct doca_pci_bdf pci_bdf{0};
	auto status = doca_devinfo_get_pci_addr(doca_dev_as_devinfo(dev), &pci_bdf);
	if (status != DOCA_SUCCESS)
		return status;

	pci_bdf_identifier << "0000:"
				<< std::setfill('0') << std::setw(2) << std::hex << pci_bdf.bus << ":"
				<< std::setfill('0') << std::setw(2) << std::hex << pci_bdf.device << "."
				<< std::hex << pci_bdf.function;

	return DOCA_SUCCESS;
}

static doca_error_t
build_devargs(const struct doca_dev *dev, const std::ostringstream &pci_bdf_identifier, const char *usr_devargs, std::ostringstream &out_devargs)
{
	struct ibv_pd *pd{nullptr};
	auto status = doca_dev_get_pd(dev, &pd);
	if (status != DOCA_SUCCESS)
		return status;

	out_devargs << pci_bdf_identifier.str()
		    << ",pd_handle=" << pd->handle << ",cmd_fd=" << priv_doca_dev_to_ibv_ctx(dev)->cmd_fd
		    << "," << usr_devargs;

	return DOCA_SUCCESS;
}

static doca_error_t
observe_dev(struct doca_dev *dev, priv_doca_dev_observer_handle_t &handle)
{
	struct priv_doca_dev_observer_ops observer{0};
	observer.on_dev_close = ports_cleanup_on_dev_close;

	union doca_data cookie{0};
	auto status = priv_doca_dev_observer_register(dev, &observer, cookie, &handle);

	if (status != DOCA_SUCCESS && status != DOCA_ERROR_IN_USE) { // DOCA_ERROR_IN_USE in case the device is already observed.
		DOCA_DLOG_ERR("Failed to register on device: %s", doca_get_error_name(status));
		return status;
	}

	return DOCA_SUCCESS;
}

doca_error_t
update_port_to_dev_map(const char *pci_bdf_identifier, struct doca_dev *dev)
{
	uint16_t pi = RTE_MAX_ETHPORTS;
	struct rte_dev_iterator iter;

	RTE_ETH_FOREACH_MATCHING_DEV(pi, pci_bdf_identifier, &iter) { // iterate over all ethdev ports matching the devargs used for probing
		try {
			// TODO: return error? valid scenario anyway?
			port_to_dev[pi] = dev; // insert if doesn't exist, override dev otherwise
		} catch (...) {
			DOCA_DLOG_ERR("Failed to store DPDK port ID to DOCA device mapping");
			rte_eth_iterator_cleanup(&iter); // breaking before end of loop
			return DOCA_ERROR_NO_MEMORY;
		}
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_dpdk_port_probe(struct doca_dev *dev, const char *devargs)
{
	if (dev == nullptr || devargs == nullptr) {
		DOCA_DLOG_ERR("Invalid input: dev=%p, devargs=%p", dev, devargs);
		return DOCA_ERROR_INVALID_VALUE;
	}

	std::ostringstream pci_bdf_identifier;
	auto status = extract_pci_bdf_identifier(dev, pci_bdf_identifier);
	if (status != DOCA_SUCCESS)
		return status;

	std::ostringstream devargs_prob;
	status = build_devargs(dev, pci_bdf_identifier, devargs, devargs_prob);
	if (status != DOCA_SUCCESS)
		return status;

	char *devargs_prob_str{nullptr};
	try {
		auto s = devargs_prob.str();
		devargs_prob_str = new (std::nothrow) char[s.length()+1]; // devargs_prob_str might be nullptr
		std::strcpy (devargs_prob_str, s.c_str());
	} catch (...) {
		DOCA_DLOG_ERR("Failed to convert string stream devargs to cstr");
		return DOCA_ERROR_NO_MEMORY;
	}

	if (rte_dev_probe(devargs_prob_str) < 0) {
		DOCA_DLOG_ERR("Failed to prob port");
		return DOCA_ERROR_DRIVER;
	}

	priv_doca_dev_observer_handle_t observer_handle{0};
	status = observe_dev(dev, observer_handle);
	if (status != DOCA_SUCCESS) {
		detach_ports(devargs_prob_str);
		return status;
	}

	status = update_port_to_dev_map(devargs_prob_str, dev);
	if (status != DOCA_SUCCESS) {
		detach_ports(devargs_prob_str);
		return status;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_dpdk_port_as_dev(uint16_t port_id, struct doca_dev **dev)
{
	if (dev == NULL) {
		DOCA_DLOG_ERR("Invalid input: dev is null");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto search = port_to_dev.find(port_id);
	if (search != port_to_dev.end()) {
		*dev = search->second;
		return DOCA_SUCCESS;
	}

	DOCA_DLOG_ERR("no DPDK port id %d associated with DOCA device %p", port_id, dev);
	return DOCA_ERROR_NOT_FOUND;
}
