/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdint.h>
#include <stdio.h>

#include <doca_error.h>
#include <private/common_defines.h>

struct doca_error_entry {
	const char *name;
	const char *descr;
};

static const struct doca_error_entry error_arr[] = {
	{
		"DOCA_SUCCESS",
		"Success",
	},
	{
		"DOCA_ERROR_UNKNOWN",
		"Unknown error",
	},
	{
		"DOCA_ERROR_NOT_PERMITTED",
		"Operation not permitted",
	},
	{
		"DOCA_ERROR_IN_USE",
		"Resource already in use",
	},
	{
		"DOCA_ERROR_NOT_SUPPORTED",
		"Operation not supported",
	},
	{
		"DOCA_ERROR_AGAIN",
		"Resource temporarily unavailable, try again",
	},
	{
		"DOCA_ERROR_INVALID_VALUE",
		"Invalid input",
	},
	{
		"DOCA_ERROR_NO_MEMORY",
		"Memory allocation failure",
	},
	{
		"DOCA_ERROR_INITIALIZATION",
		"Resource initialization failure",
	},
	{
		"DOCA_ERROR_TIME_OUT",
		"Timer expired waiting for resource",
	},
	{
		"DOCA_ERROR_SHUTDOWN",
		"Shut down in process or completed",
	},
	{
		"DOCA_ERROR_CONNECTION_RESET",
		"Connection reset by peer",
	},
	{
		"DOCA_ERROR_CONNECTION_ABORTED",
		"Connection aborted",
	},
	{
		"DOCA_ERROR_CONNECTION_INPROGRESS",
		"Connection in progress",
	},
	{
		"DOCA_ERROR_NOT_CONNECTED",
		"Not Connected",
	},
	{
		"DOCA_ERROR_NO_LOCK",
		"Unable to acquire required lock",
	},
	{
		"DOCA_ERROR_NOT_FOUND",
		"Requested Resource Not Found",
	},
	{
		"DOCA_ERROR_IO_FAILED",
		"Input/Output Operation Failed",
	},
	{
		"DOCA_ERROR_BAD_STATE",
		"Bad State",
	},
	{
		"DOCA_ERROR_UNSUPPORTED_VERSION",
		"Unsupported version",
	},
	{
		"DOCA_ERROR_OPERATING_SYSTEM",
		"Operating system call failure",
	},
	{
		"DOCA_ERROR_DRIVER",
		"DOCA Driver call failure",
	},
	{
		"DOCA_ERROR_UNEXPECTED",
		"An unexpected scenario was detected",
	},
};

/**
 *
 * \brief Returns the string representation of an error code name.
 *
 * Returns a string containing the name of an error code in the enum.
 * If the error code is not recognized, "unrecognized error code" is returned.
 *
 * @param[in] error - Error code to convert to string.
 *
 * @return char* pointer to a NULL-terminated string.
 */
const char *
doca_get_error_name(doca_error_t error)
{
	uint32_t array_size = static_cast<uint32_t>(ARRAY_SIZE(error_arr));

	if (error >= 0 && static_cast<uint32_t>(error) < array_size)
		return error_arr[error].name;

	return "Unrecognized error code";
}

/**
 * \brief Returns the description string of an error code.
 *
 *  This function returns the description of an error code.
 *  If the error code is not recognized, "unrecognized error code" is returned.
 *
 * @param[in] error - Error code to convert to description string.
 *
 * @return char* pointer to a NULL-terminated string.
 */
const char *
doca_get_error_string(doca_error_t error)
{
	uint32_t array_size = static_cast<uint32_t>(ARRAY_SIZE(error_arr));

	if (error >= 0 && static_cast<uint32_t>(error) < array_size)
		return error_arr[error].descr;

	return "Unrecognized error code";
}
