/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <unordered_map>
#include <doca_log.h>
#include "doca_hotplug.h"
#include "priv_doca_hotplug.h"
#include "priv_doca_dev.h"
#include "multi_os_prm.h"

DOCA_LOG_REGISTER(CORE::DOCA_HOTPLUG);

static inline enum doca_dev_rep_filter
hotplug_dev_type_to_rep_filter(enum mlx5_hotplug_device_type type)
{
	switch (type) {
	case MLX5_HOTPLUG_DEVICE_TYPE_VIRTIO_FS:
		return DOCA_DEV_REP_FILTER_VIRTIO_FS;
	case MLX5_HOTPLUG_DEVICE_TYPE_VIRTIO_NET:
		return DOCA_DEV_REP_FILTER_VIRTIO_NET;
	case MLX5_HOTPLUG_DEVICE_TYPE_VIRTIO_BLK:
		return DOCA_DEV_REP_FILTER_VIRTIO_BLK;
	case MLX5_HOTPLUG_DEVICE_TYPE_NVME:
		return DOCA_DEV_REP_FILTER_NVME;
	default:
		return DOCA_DEV_REP_FILTER_ALL;
	}
}

static doca_error_t
hotplug_dev_obj_extract_rep_filter(void *hotplug_dev_obj, enum doca_dev_rep_filter *filter)
{
	auto hotplug_dev_type = MLX5_GET(hotplug_device_in, hotplug_dev_obj, hotplug_device_object.device_type);
	enum doca_dev_rep_filter rep_filter = hotplug_dev_type_to_rep_filter(static_cast<enum mlx5_hotplug_device_type>(hotplug_dev_type));

	if (rep_filter == DOCA_DEV_REP_FILTER_ALL) {
		DOCA_DLOG_ERR("Unknown hotplug device type: %d", hotplug_dev_type);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	*filter = rep_filter;
	return DOCA_SUCCESS;
}

static doca_error_t
hotunplug_prm(struct doca_dev *dev, uint16_t vhca_id)
{
	uint8_t in[MLX5_ST_SZ_BYTES(hotunplug_device_in)] = {0};
	uint8_t out[MLX5_ST_SZ_BYTES(hotunplug_device_out)] = {0};
	uint8_t *device_in = MLX5_ADDR_OF(hotunplug_device_in, in, hotplug_device_object);

	MLX5_SET(hotunplug_device_in, in, opcode, MLX5_CMD_OP_HOTUNPLUG_DEVICE);
	MLX5_SET(hotplug_device_obj, device_in, vhca_id, vhca_id);
	auto err = priv_doca_dev_prm_cmd_send(dev, in, sizeof(in), out, sizeof(out));
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("PRM HOTUNPLUG_DEVICE command failed (error=%s)", doca_get_error_name(err));
		return err;
	}

	return DOCA_SUCCESS;
}

static doca_error_t
hotplugged_device_open(struct doca_dev *dev, uint16_t vhca_id, enum doca_dev_rep_filter filter, struct doca_dev_rep **dev_rep)
{
	struct doca_devinfo_rep **devinfo_list{nullptr};
	uint32_t nb_devs_rep = 0;

	auto err = doca_devinfo_rep_list_create(dev, filter, &devinfo_list, &nb_devs_rep);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to create representor device info list (error=%s)", doca_get_error_name(err));
		return err;
	}

	uint16_t vhca_id_tmp = 0;
	uint32_t i = 0;

	for(i = 0; i < nb_devs_rep; i++){
		err = priv_doca_devinfo_rep_vhca_id_get(devinfo_list[i], &vhca_id_tmp);
		if (err != DOCA_SUCCESS){
			DOCA_DLOG_ERR("Failed to query representor device info VHCA ID (error=%s)", doca_get_error_name(err));
			doca_devinfo_rep_list_destroy(devinfo_list);
			return err;
		}
		if (vhca_id_tmp == vhca_id)
			break;
	}

	if (i >= nb_devs_rep){
		DOCA_DLOG_ERR("Failed to query hotplugged device");
		doca_devinfo_rep_list_destroy(devinfo_list);
		return DOCA_ERROR_NOT_FOUND;
	}

	err = doca_dev_rep_open(devinfo_list[i], dev_rep);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to open representor device (error=%s)", doca_get_error_name(err));
		doca_devinfo_rep_list_destroy(devinfo_list);
		return err;
	}

	err = doca_devinfo_rep_list_destroy(devinfo_list);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to destroy representor device info list (error=%s)", doca_get_error_name(err));
		return err;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_dev_rep_hotplug(const struct doca_dev_hotplug_attr *attr, struct doca_dev_rep **dev_rep)
{
	if (attr == NULL || dev_rep == NULL){
		DOCA_DLOG_ERR("Invalid NULL parameter: attr=%p, dev_rep=%p", attr, dev_rep);
		return DOCA_ERROR_INVALID_VALUE;
	}

	uint8_t hotplug_manager_cap = 0;

	auto err = doca_devinfo_get_is_hotplug_manager_supported(doca_dev_as_devinfo(attr->dev), &hotplug_manager_cap);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to query hotplug_manager capability (error=%s)", doca_get_error_name(err));
		return err;
	}

	if (hotplug_manager_cap != 1){ // non-hotplug manager
		DOCA_DLOG_ERR("hotplug on non hotplug manager device is not supported");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	uint8_t hotplug_dev_obj_in[MLX5_ST_SZ_BYTES(hotplug_device_in)] = {0};

	err = attr->hotplug_dev_obj_build_cb(attr, sizeof(hotplug_dev_obj_in), hotplug_dev_obj_in);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Hotplug callback failed (error=%s)", doca_get_error_name(err));
		return err;
	}

	enum doca_dev_rep_filter filter;
	err = hotplug_dev_obj_extract_rep_filter(hotplug_dev_obj_in, &filter);
	if (err != DOCA_SUCCESS)
		return err;

	MLX5_SET(hotplug_device_in, hotplug_dev_obj_in, opcode, MLX5_CMD_OP_HOTPLUG_DEVICE);

	uint8_t out[MLX5_ST_SZ_BYTES(hotplug_device_out)] = {0};

	err = priv_doca_dev_prm_cmd_send(attr->dev, hotplug_dev_obj_in, sizeof(hotplug_dev_obj_in), out, sizeof(out));
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("PRM HOTPLUG_DEVICE command failed (error=%s)", doca_get_error_name(err));
		return err;
	}

	uint16_t vhca_id = MLX5_GET(hotplug_device_out, out, hotplug_device_object.vhca_id);

	err = hotplugged_device_open(attr->dev, vhca_id, filter, dev_rep);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to open hotplugged device (error=%s)", doca_get_error_name(err));
		hotunplug_prm(attr->dev, vhca_id);
		return err;
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_dev_rep_hotunplug(struct doca_dev_rep *rep_dev)
{
	if (rep_dev == NULL){
		DOCA_DLOG_ERR("Invalid NULL rep_dev");
		return DOCA_ERROR_INVALID_VALUE;
	}

	bool is_hotplugged = false;
	auto err = priv_doca_dev_rep_is_hotplug_get(rep_dev, &is_hotplugged);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to query if hotplugged device (error=%s)", doca_get_error_name(err));
		return err;
	}

	if(!is_hotplugged){
		DOCA_DLOG_ERR("Hotunplug of static emulated device is not supported");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	uint16_t vhca_id;

	err = priv_doca_devinfo_rep_vhca_id_get(doca_dev_rep_as_devinfo(rep_dev), &vhca_id);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to query VHCA ID of representor DOCA device (error=%s)", doca_get_error_name(err));
		return err;
	}

	struct doca_dev *dev{nullptr};
	err = priv_doca_dev_rep_parent_get(rep_dev, &dev);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to get local parent device of representor device (error=%s)", doca_get_error_name(err));
		return err;
	}

	err = hotunplug_prm(dev, vhca_id);
	if (err != DOCA_SUCCESS){
		return err;
	}

	err = doca_dev_rep_close(rep_dev);
	if (err != DOCA_SUCCESS){
		DOCA_DLOG_ERR("Failed to close hotplugged representor device (error=%s)", doca_get_error_name(err));
		return err;
	}

	return DOCA_SUCCESS;
}
