/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdbool.h>
#include <string.h>
#include <cstring>
#include <random>
#include <sstream>
#include <iomanip>
#include <json-c/json.h>

#include <doca_log.h>
#include <doca_types.h>
#include "internal_doca_buf.h"
#include "internal_doca_dev.h"
#include <priv_doca_buf.h>
#include <priv_doca_dev.h>
#include <priv_doca_malloc.h>
#include <private/common_defines.h>

#include <doca_mmap.h>
#include <string>
#include <idevice_adapter.hpp>
#include <memory_access_flags.hpp>
#include <priv_doca_os_abstraction_endian.h>
#include <utils.hpp>

#define MAX_MCHUNK_NUM_DEFAULT_VALUE 1
#define MAX_DEVICE_NUM_DEFAULT_VALUE 1
#define MIN_PAGE_SIZE (1 << 12)
#define EXPORT_STR_LEN_DEFAULT_VALUE 256
#define MCHUNK_PAGE_SIZE_DEFULT_VALUE 4096
#define MMAP_EXPORT_VERSION 1

using json_deleter_t = std::function<void(struct json_object *)>;
using doca_malloc_deleter_t = std::function<void(void *)>;
using mmap_deleter_t = std::function<void(struct doca_mmap *)>;

DOCA_LOG_REGISTER(CORE::DOCA_MMAP);

static inline bool
is_power_of_2(uint64_t x)
{
	return x && (x & (x - 1)) == 0;
}

static inline doca_error_t
is_mmap_status_valid_for_operation(const struct doca_mmap *mmap)
{
	if (!mmap->start)
		return DOCA_ERROR_NOT_PERMITTED;

	if (mmap->exported)
		return DOCA_ERROR_NOT_PERMITTED;

	if (mmap->from_export)
		return DOCA_ERROR_NOT_PERMITTED;

	if(mmap->from_representor)
		return DOCA_ERROR_NOT_PERMITTED;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_create(const union doca_data *user_data, struct doca_mmap **mmap)
{
	struct doca_mmap *new_mmap{nullptr};

	if (mmap == NULL) {
		DOCA_DLOG_ERR("Unable to create mmap: Received invalid input - mmap=%p", mmap);
		return DOCA_ERROR_INVALID_VALUE;
	}

	try {

		new_mmap = new doca_mmap();
		if (user_data != NULL)
			new_mmap->m_user_data = *user_data;
		else
			memset(&new_mmap->m_user_data, 0, sizeof(new_mmap->m_user_data));

		new_mmap->max_mchunk_num = MAX_MCHUNK_NUM_DEFAULT_VALUE;
		new_mmap->max_device_num = MAX_DEVICE_NUM_DEFAULT_VALUE;

		*mmap = new_mmap;

		return DOCA_SUCCESS;
	} catch (...) {
		DOCA_DLOG_ERR("Unable to create mmap: Failed to allocate mmap");
		delete new_mmap;
		return DOCA_ERROR_NO_MEMORY;
	}
}

doca_error_t
doca_mmap_destroy(struct doca_mmap *mmap)
{
	uint32_t i, j;

	if (mmap == NULL) {
		DOCA_DLOG_ERR("Unable to destroy mmap: Received invalid input - mmap is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (mmap->doca_buf_num != 0) {
		DOCA_DLOG_ERR("Unable to destroy mmap: There are buffers refering to memory ranges in the mmap - return all buffers previously");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	// Can't just clear the mchunk array because the free callback should be called
	// after the memory region has been released. Therefore a loop must be performed.
	for (i = 0; i < mmap->curr_mchunk_num; i++) {
		struct priv_doca_mchunk *curr_mchunk = &mmap->mchunk_arr[i];
		for (j = 0; j < mmap->curr_device_num; j++) {
			for (auto& mr: curr_mchunk->device_to_mkey_arr[j].memory_regions) {
				if ((mr == nullptr) || (mr->destroy_key() != DOCA_SUCCESS)) {
					DOCA_DLOG_ERR(
						"Unable to destroy mmap: Failed to destroy key for memory range number: %d and device number: %d",
						i, j);
					return DOCA_ERROR_NOT_PERMITTED;
				}
			}
			curr_mchunk->device_to_mkey_arr[j].memory_regions.fill(nullptr);
		}

		/* possible optimization: add a destructor to the mchunk that will call the callback. */
		if (curr_mchunk->free_cb != NULL)
			curr_mchunk->free_cb(curr_mchunk->addr, curr_mchunk->len, curr_mchunk->opaque);
	}

	delete mmap;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_set_max_num_chunks(struct doca_mmap *mmap, uint32_t max_num_chunks)
{
	if (mmap == NULL) {
		DOCA_DLOG_ERR("Unable to set mmap max number of chunks: Received invalid input - mmap=%p", mmap);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (mmap->previously_started) {
		DOCA_DLOG_ERR("Unable to set mmap max number of chunks: Mmap had been started in the past");
		return DOCA_ERROR_NOT_PERMITTED;
	}
	if (max_num_chunks == 0) {
		DOCA_DLOG_ERR("Unable to set mmap max number of chunks: max_num_of_chunks can't be equal to 0");
		return DOCA_ERROR_INVALID_VALUE;
	}

	mmap->max_mchunk_num = max_num_chunks;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_set_max_num_devices(struct doca_mmap *mmap, uint32_t max_num_devices)
{
	if (mmap == NULL) {
		DOCA_DLOG_ERR("Unable to set mmap max number of devices: Received invalid input - mmap=%p", mmap);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (mmap->previously_started) {
		DOCA_DLOG_ERR("Unable to set mmap max number of devices: Mmap had been started in the past");
		return DOCA_ERROR_NOT_PERMITTED;
	}
	if (max_num_devices == 0) {
		DOCA_DLOG_ERR("Unable to set mmap max number of devices: max_num_of_devices can't be equal to 0");
		return DOCA_ERROR_INVALID_VALUE;
	}

	mmap->max_device_num = max_num_devices;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_user_data(struct doca_mmap *mmap, union doca_data *user_data)
{
	if (mmap == NULL || user_data == NULL) {
		DOCA_DLOG_ERR("Unable to get mmap user_data: Received invalid input - mmap=%p, user_data=%p", mmap, user_data);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*user_data = mmap->m_user_data;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_max_num_chunks(struct doca_mmap *mmap, uint32_t *max_num_chunks)
{
	if (mmap == NULL || max_num_chunks == NULL) {
		DOCA_DLOG_ERR("Unable to get mmap max number of chunks: Received invalid input - mmap=%p, max_num_chunks=%p",
			mmap, max_num_chunks);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*max_num_chunks = mmap->max_mchunk_num;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_max_num_devices(struct doca_mmap *mmap, uint32_t *max_num_devices)
{
	if (mmap == NULL || max_num_devices == NULL) {
		DOCA_DLOG_ERR("Unable to get mmap max number devices: Received invalid input - mmap=%p, max_num_devices=%p",
			mmap, max_num_devices);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*max_num_devices = mmap->max_device_num;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_num_bufs(struct doca_mmap *mmap, uint32_t *num_bufs)
{
	if (mmap == NULL || num_bufs == NULL) {
		DOCA_DLOG_ERR("Unable to get mmap number of buffers: Received invalid input - mmap=%p, num_bufs=%p",
			mmap, num_bufs);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*num_bufs = mmap->doca_buf_num;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_exported(struct doca_mmap *mmap, uint8_t *exported)
{
	if (mmap == NULL || exported == NULL) {
		DOCA_DLOG_ERR("Unable to get mmap exported flag value: Received invalid input - mmap=%p, exported=%p",
			mmap, exported);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*exported = !!mmap->exported;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_from_export(struct doca_mmap *mmap, uint8_t *from_export)
{
	if (mmap == NULL || from_export == NULL) {
		DOCA_DLOG_ERR("Unable to get mmap from_export flag value: Received invalid input - mmap=%p, from_export=%p",
			mmap, from_export);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*from_export = !!mmap->from_export;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_start(struct doca_mmap *mmap)
{
	if (mmap == NULL) {
		DOCA_DLOG_ERR("Unable to start mmap: Received invalid input - mmap=%p", mmap);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (mmap->previously_started) {
		mmap->start = true;
		return DOCA_SUCCESS;
	}

	try {
		mmap->mchunk_arr.resize(mmap->max_mchunk_num);
		mmap->device_arr.resize(mmap->max_device_num);
		mmap->start = true;
		mmap->previously_started = true;

		return DOCA_SUCCESS;
	} catch (...) {
		// No need to reset the mchunk and device arrays,
		// they will be deleted when mmap is destroyed (or resized in another start attempt).
		DOCA_DLOG_ERR("Unable to start mmap: Failed to allocate sufficient memory");
		return DOCA_ERROR_NO_MEMORY;
	}
}

doca_error_t
doca_mmap_stop(struct doca_mmap *mmap)
{
	if (mmap == NULL) {
		DOCA_DLOG_ERR("Unable to stop mmap: Received invalid input - mmap is NULL");
		return DOCA_ERROR_INVALID_VALUE;
	}

	mmap->start = false;
	return DOCA_SUCCESS;
}

static doca_error_t
priv_doca_mmap_create_mr(struct priv_doca_device_to_mkey *new_device_to_mkey, void *addr,
			 size_t length, size_t page_size)
{
	auto device_adapter = new_device_to_mkey->device->device_adapter;
	auto pd = new_device_to_mkey->device->pd;

	unsigned int i{0};
	for (auto mkey_access : mkey_accesses) {
		doca_error_t status{DOCA_SUCCESS};
		std::shared_ptr<device_abstraction::imemory_region> mr;

		std::tie(status, mr) = pd->create_mapped_memory_region(addr, length, page_size, mkey_access_to_flags.at(static_cast<uint32_t>(mkey_access)));
		if (status != DOCA_SUCCESS)
			return status;
		new_device_to_mkey->memory_regions[i++] = std::move(mr);
	}

	// Assumption: if status is DOCA_SUCCESS then mr is valid.
	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_dev_add(struct doca_mmap *mmap, struct doca_dev *dev)
{
	struct priv_doca_mchunk *curr_mchunk;
	struct priv_doca_device_to_mkey *new_device_to_mkey;
	uint32_t i = 0, j = 0;
	doca_error_t ret;

	if (mmap == NULL || dev == NULL) {
		DOCA_DLOG_ERR("Unable to add device to mmap: Received invalid input - mmap=%p, dev=%p", mmap, dev);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (is_mmap_status_valid_for_operation(mmap) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to add device to mmap: Mmap stopped or exported or created from export");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	if (mmap->curr_device_num == mmap->max_device_num) {
		DOCA_DLOG_ERR("Unable to add device to mmap: Reached maximum number of devices - %u",
			       mmap->max_device_num);
		return DOCA_ERROR_NO_MEMORY;
	}

	const auto curr_device_num{mmap->curr_device_num};

	/* Check if this device is already exists in doca_mmap */
	for (i = 0; i < curr_device_num; i++) {
		if (dev == mmap->device_arr[i]) {
			DOCA_DLOG_ERR("Unable to add device to mmap: Device already exists in doca_mmap");
			return DOCA_ERROR_IN_USE;
		}
	}

	/* TODO: It is better to convert the MMAP access flags to memory_access_flags. */
	for (i = 0; i < mmap->curr_mchunk_num; i++) {
		curr_mchunk = &mmap->mchunk_arr[i];
		new_device_to_mkey = &(curr_mchunk->device_to_mkey_arr[curr_device_num]);

		new_device_to_mkey->device = dev;
		ret = priv_doca_mmap_create_mr(new_device_to_mkey, curr_mchunk->addr,
					       curr_mchunk->len, curr_mchunk->pg_sz);
		if (ret != DOCA_SUCCESS) {
			DOCA_DLOG_ERR(
				"Unable to add device to mmap: Internal error - Couldn't register MR for memory range number: %d",
				i);
			goto dev_add_error;
		}
	}
	mmap->device_arr[curr_device_num] = dev;
	mmap->curr_device_num++;
	return DOCA_SUCCESS;

dev_add_error:
	for (j = 0; j < i; j++) {
		curr_mchunk = &mmap->mchunk_arr[j];
		new_device_to_mkey = &(curr_mchunk->device_to_mkey_arr[curr_device_num]);
		// There is no reason that mkey destroy will fail when it was just created,
		// therefore there is no need to explicitly call destroy_mkey() and check status.
		new_device_to_mkey->memory_regions.fill(nullptr);
	}
	return DOCA_ERROR_NOT_PERMITTED;
}

doca_error_t
doca_mmap_dev_rm(struct doca_mmap *mmap, struct doca_dev *dev)
{
	uint32_t i = 0, j = 0;
	int dev_to_rm = -1;
	struct priv_doca_mchunk *curr_mchunk;
	struct priv_doca_device_to_mkey *free_device_to_mkey, *last_device_to_mkey;

	if (mmap == NULL || dev == NULL) {
		DOCA_DLOG_ERR("Unable to remove device from mmap: Received invalid input - mmap=%p, dev=%p",
			      mmap, dev);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (is_mmap_status_valid_for_operation(mmap) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to remove device from mmap: Mmap stopped or exported or created from export");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	// The design assumes that the device index is the same as the device_to_mkey_arr index.
	// Therefore finding dev_to_rm is a must and std::find can't be used.
	for (i = 0; i < mmap->curr_device_num; i++) {
		if (dev == mmap->device_arr[i]) {
			dev_to_rm = i;
			break;
		}
	}
	if (dev_to_rm < 0) {
		DOCA_DLOG_ERR(
		      "Unable to remove device from mmap: Received invalid input - Device does not exist in doca_mmap");
		return DOCA_ERROR_INVALID_VALUE;
	}

	for (j = 0; j < mmap->curr_mchunk_num; j++) {
		curr_mchunk = &mmap->mchunk_arr[j];
		free_device_to_mkey = &(curr_mchunk->device_to_mkey_arr[dev_to_rm]);

		for (auto& mr: free_device_to_mkey->memory_regions) {
			if ((mr == nullptr) || (mr->destroy_key() != DOCA_SUCCESS)) {
				DOCA_DLOG_ERR(
					"Unable to remove device from mmap: Internal error - Couldn't deregister MR for memory range number: %d",
					j);
				return DOCA_ERROR_NOT_PERMITTED;
			}
		}
		free_device_to_mkey->memory_regions.fill(nullptr);

		free_device_to_mkey->device = nullptr;

		/*
		 * In order to keep the array contiguous move the last device
		 * to the place of the removed device.
		 * This action should be done only if the device to be removed is not the last -> std::swap takes care of it.
		 */
		last_device_to_mkey = &(curr_mchunk->device_to_mkey_arr[mmap->curr_device_num - 1]);
		std::swap(free_device_to_mkey->device, last_device_to_mkey->device);
		free_device_to_mkey->memory_regions.swap(last_device_to_mkey->memory_regions);
	}

	// A device may be present in the zba map. This section removes all entries that contain the device.
	for (auto &zba_map_entry : mmap->zba_mkey_map) {
		zba_map_entry.second.erase(dev);
	}

	// Same goes here, moving the last device to the place of the removed device.
	mmap->device_arr[dev_to_rm] = mmap->device_arr[mmap->curr_device_num - 1];
	mmap->device_arr[mmap->curr_device_num - 1] = nullptr;
	mmap->curr_device_num--;
	return DOCA_SUCCESS;
}

static doca_error_t
mmap_populate(struct doca_mmap *mmap, void *addr, size_t len, size_t pg_sz,
	      doca_mmap_memrange_free_cb_t *free_cb, void *opaque)
{
	struct priv_doca_mchunk *new_mchunk{nullptr};
	void *curr_mchunk_addr{nullptr};
	size_t curr_mchunk_len{0};
	uint32_t i{0};
	doca_error_t ret;

	if (is_mmap_status_valid_for_operation(mmap) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to add memory range: Mmap stopped or exported or created from export");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	if (mmap->curr_mchunk_num == mmap->max_mchunk_num) {
		DOCA_DLOG_ERR("Unable to add memory range: Reached maximum limit(%u)", mmap->max_mchunk_num);
		return DOCA_ERROR_NO_MEMORY;
	}

	/* Check that the address and len do not intersect with any existing chunk */
	for (i = 0; i < mmap->curr_mchunk_num; i++) {
		curr_mchunk_addr = mmap->mchunk_arr[i].addr;
		curr_mchunk_len = mmap->mchunk_arr[i].len;

		if ((addr > curr_mchunk_addr &&
		    static_cast<size_t>((uint8_t *)addr - (uint8_t *)curr_mchunk_addr) < curr_mchunk_len) ||
		    (addr < curr_mchunk_addr &&
		    static_cast<size_t>((uint8_t *)curr_mchunk_addr - (uint8_t *)addr) < len) ||
		    (addr == curr_mchunk_addr)) {
			DOCA_DLOG_ERR("Unable to add memory range: Address and len intersect with an existing chunk");
			return DOCA_ERROR_NOT_PERMITTED;
		}
	}

	new_mchunk = &(mmap->mchunk_arr[mmap->curr_mchunk_num]);

	try {
		new_mchunk->device_to_mkey_arr.resize(mmap->max_device_num);
	} catch (...) {
		DOCA_DLOG_ERR("Unable to add memory range: Internal error - Failed to allocate device to mkey array");
		return DOCA_ERROR_NO_MEMORY;
	}

	new_mchunk->free_cb = free_cb;
	new_mchunk->opaque = opaque;
	new_mchunk->addr = addr;
	new_mchunk->len = len;
	new_mchunk->pg_sz = pg_sz;
	new_mchunk->mmap = mmap;

	for (i = 0; i < mmap->curr_device_num; i++) {
		auto curr_dev = mmap->device_arr[i];
		auto device_adapter = curr_dev->device_adapter;

		new_mchunk->device_to_mkey_arr[i].device = curr_dev;
		ret = priv_doca_mmap_create_mr(&new_mchunk->device_to_mkey_arr[i], new_mchunk->addr,
						new_mchunk->len, new_mchunk->pg_sz);
		if (ret != DOCA_SUCCESS) {
			DOCA_DLOG_ERR(
			      "Unable to add memory range: Internal error - Couldn't register MR for device number: %d",
			      i);

			// Clearing the device_to_mkey_arr will release all the created memory regions until failure.
			// Since these are memory regions that were just created there is no reason that the destroy_key
			// will fail and therefore it is safe to clear the vector (no need to check error).
			new_mchunk->device_to_mkey_arr.clear();
			return DOCA_ERROR_NOT_PERMITTED;
		}
	}
	mmap->curr_mchunk_num++;
	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_populate(struct doca_mmap *mmap, void *addr, size_t len, size_t pg_sz,
		   doca_mmap_memrange_free_cb_t *free_cb, void *opaque)
{
	if (mmap == NULL || (addr == NULL) || len == 0 || pg_sz < MIN_PAGE_SIZE || !is_power_of_2(pg_sz)) {
		DOCA_DLOG_ERR(
			"Unable to add memory range: Received invalid input - mmap=%p, addr=%p, len=%zu, pg_sz=%zu",
			mmap, addr, len, pg_sz);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return mmap_populate(mmap, addr, len, pg_sz, free_cb, opaque);
}

static inline void
encrypt_decrypt_export_desc(char *dest, const char *src, size_t export_desc_len)
{
	static const char xor_key = 'K';

	for (int i = 0; i < export_desc_len; i++)
		dest[i] = src[i] ^ (xor_key + i);
}

static doca_error_t
mmap_allow_other_vhca(struct doca_mmap *mmap, const struct doca_dev *dev, uint32_t device_index,
		      const device_abstraction::allow_other_vhca_access_key_t &access_key)
{
	for (uint32_t chunk_index = 0; chunk_index < mmap->curr_mchunk_num; chunk_index++) {

		for (auto& mr: mmap->mchunk_arr[chunk_index].device_to_mkey_arr[device_index].memory_regions) {
			auto status = mr->allow_other_vhca_access(access_key);
			if (status != DOCA_SUCCESS) {
				DOCA_DLOG_ERR("Failed to allow other VHCA");
				return status;
			}
		}
	}

	return DOCA_SUCCESS;
}

static void
prepare_mmap_export_message_header(uint8_t *export_buffer, size_t export_buffer_length)
{
	mmap_export_header *header = reinterpret_cast<mmap_export_header *>(export_buffer);
	header->version = MMAP_EXPORT_VERSION;
	header->padding = 0;
	header->length = static_cast<uint16_t>(export_buffer_length);
	// Header is in network order (hton). htobe16 and htobe32 macros implement this for multi OS.
	header->length = htobe16(header->length);
	header->crc = 0; // CRC must be 0 when calculating CRC.
	auto crc = core_utils::calculate_crc32_normal(export_buffer, export_buffer_length);
	header->crc = htobe32(crc);
}

static std::tuple<doca_error_t, uint8_t *, size_t>
prepare_mmap_export_buffer(const char *export_message, size_t export_message_length)
{
	// The export buffer contains the header and the message string.
	auto export_buffer_length = (export_message_length + sizeof(mmap_export_header));

	// The buffer is allocated using malloc because it is passed to the application and is expected to be released
	// using free.
	std::unique_ptr<uint8_t, doca_malloc_deleter_t> export_buffer(
		reinterpret_cast<uint8_t *>(priv_doca_malloc(1, export_buffer_length)),
		[](void *buffer) { priv_doca_free(buffer); });

	if (!export_buffer) {
		DOCA_DLOG_ERR("Failed to alloc memory for export string");
		return {DOCA_ERROR_NO_MEMORY, nullptr, 0};
	}

	// Encrypting the message string directly into the destination buffer.
	uint8_t *message_dest = export_buffer.get() + sizeof(mmap_export_header);
	encrypt_decrypt_export_desc(reinterpret_cast<char *>(message_dest), export_message, export_message_length);

	prepare_mmap_export_message_header(export_buffer.get(), export_buffer_length);

	return {DOCA_SUCCESS, export_buffer.release(), export_buffer_length};
}

static std::pair<doca_error_t, std::unique_ptr<struct json_object, json_deleter_t>>
access_key_to_json(const device_abstraction::allow_other_vhca_access_key_t &access_key)
{
	std::unique_ptr<struct json_object, json_deleter_t> access_key_obj(
		json_object_new_array(), [](struct json_object *jobj) { json_object_put(jobj); });

	for (auto &node : access_key) {
		json_object_array_add(access_key_obj.get(), json_object_new_int(static_cast<int32_t>(node)));
	}

	return {DOCA_SUCCESS, std::move(access_key_obj)};
}

static std::pair<doca_error_t, std::unique_ptr<struct json_object, json_deleter_t>>
mmap_chunks_to_json(struct doca_mmap *mmap, uint32_t device_index)
{
	std::unique_ptr<struct json_object, json_deleter_t> chunks_json_obj(
		json_object_new_array(), [](struct json_object *jobj) { json_object_put(jobj); });
	if (!chunks_json_obj) {
		DOCA_DLOG_ERR("Failed to create json array object");
		return {DOCA_ERROR_NO_MEMORY, std::unique_ptr<struct json_object, json_deleter_t>(nullptr)};
	}

	for (auto chunk = mmap->mchunk_arr.begin(); chunk != mmap->mchunk_arr.begin() + mmap->curr_mchunk_num; ++chunk) {
		struct json_object *mchunk_json_obj = json_object_new_object();

		if (mchunk_json_obj == nullptr) {
			DOCA_DLOG_ERR("Failed to create chunk json object");
			return {DOCA_ERROR_NO_MEMORY, std::unique_ptr<struct json_object, json_deleter_t>(nullptr)};
		}

		struct json_object *mkey_arr_json_obj = json_object_new_array();
		if (mkey_arr_json_obj == nullptr) {
			DOCA_DLOG_ERR("Failed to create mkey array json object");
			json_object_put(mchunk_json_obj);
			return {DOCA_ERROR_NO_MEMORY, std::unique_ptr<struct json_object, json_deleter_t>(nullptr)};
		}

		json_object_object_add(mchunk_json_obj, "addr", json_object_new_int64((uint64_t)chunk->addr));
		json_object_object_add(mchunk_json_obj, "len", json_object_new_int64(chunk->len));

		for (auto mkey_access : mkey_accesses) {
			json_object_array_add(
				mkey_arr_json_obj,
				json_object_new_int(chunk->device_to_mkey_arr[device_index]
							    .memory_regions[static_cast<uint32_t>(mkey_access)]
							    ->get_id()));
		}

		json_object_object_add(mchunk_json_obj, "mkey_arr", mkey_arr_json_obj);
		json_object_object_add(mchunk_json_obj, "page_size", json_object_new_int64(chunk->pg_sz));
		json_object_array_add(chunks_json_obj.get(), mchunk_json_obj);
	}

	return {DOCA_SUCCESS, std::move(chunks_json_obj)};
}

static std::pair<doca_error_t, std::unique_ptr<struct json_object, json_deleter_t>>
mmap_to_json(struct doca_mmap *mmap, uint32_t device_index, uint16_t vhca_id,
	      const device_abstraction::allow_other_vhca_access_key_t &access_key)
{
	std::unique_ptr<struct json_object, json_deleter_t> export_json_obj(
		json_object_new_object(), [](struct json_object *jobj) { json_object_put(jobj); });

	if (!export_json_obj) {
		DOCA_DLOG_ERR("Failed to create json object");
		return {DOCA_ERROR_NO_MEMORY, std::unique_ptr<struct json_object, json_deleter_t> (nullptr)};
	}

	// json_object_object_add has different signatures on different OS versions. Sometimes it returns int and
	// sometimes it returns void. Therefore the method does not check the return value.

	json_object_object_add(export_json_obj.get(), "vhca_id", json_object_new_int(vhca_id));

	doca_error_t status{DOCA_SUCCESS};
	std::unique_ptr<struct json_object, json_deleter_t> access_key_json_obj;
	std::tie(status, access_key_json_obj) = access_key_to_json(access_key);
	if (status!= DOCA_SUCCESS)
		return {status, std::unique_ptr<struct json_object, json_deleter_t>(nullptr)};
	json_object_object_add(export_json_obj.get(), "access_key", access_key_json_obj.release());

	std::unique_ptr<struct json_object, json_deleter_t> chunks_json_obj;
	std::tie(status, chunks_json_obj) = mmap_chunks_to_json(mmap, device_index);
	if (status != DOCA_SUCCESS)
		return {status, std::unique_ptr<struct json_object, json_deleter_t>(nullptr)};

	json_object_object_add(export_json_obj.get(), "mchunks", chunks_json_obj.release());

	return {DOCA_SUCCESS, std::move(export_json_obj)};
}

static std::pair<doca_error_t, int>
find_device_index(struct doca_mmap *mmap, const struct doca_dev *dev)
{
	auto dev_it = find(mmap->device_arr.begin(), mmap->device_arr.end(), dev);
	if (dev_it == mmap->device_arr.end()) {
		DOCA_DLOG_ERR("Unable to export mmap: Device does not exist in doca_mmap");
		return {DOCA_ERROR_INVALID_VALUE, -1};
	}

	return {DOCA_SUCCESS, static_cast<int>(std::distance(mmap->device_arr.begin(), dev_it))};
}

static device_abstraction::allow_other_vhca_access_key_t
generate_access_key(uint16_t vhca_id)
{
	device_abstraction::allow_other_vhca_access_key_t access_key{};

	// Filling the access key with random characters
	std::uniform_int_distribution<> distrib(0, UINT8_MAX);
	std::random_device rd;
	std::default_random_engine gen(rd());
	for (auto &node : access_key) {
		node = static_cast<uint8_t>(distrib(gen));
	}

	// Overwriting the first 2 characters with the VHCA id to ease identifying the source when debugging.
	uint8_t *pvhca_id = reinterpret_cast<uint8_t *>(&vhca_id);
	access_key[0] = pvhca_id[0];
	access_key[1] = pvhca_id[1];

	return access_key;
}

doca_error_t
doca_mmap_export(struct doca_mmap *mmap, const struct doca_dev *dev, void **export_desc, size_t *export_desc_len)
{
	if (mmap == NULL || dev == NULL || export_desc == NULL || export_desc_len == NULL) {
		DOCA_DLOG_ERR(
		  "Unable to export mmap: Received invalid input - mmap=%p, dev=%p, export_desc=%p, export_desc_len=%p",
		  mmap, dev, export_desc, export_desc_len);
		return DOCA_ERROR_INVALID_VALUE;
	}
	if (is_mmap_status_valid_for_operation(mmap) != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to export mmap: Mmap stopped or exported or created from export");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	auto devinfo = &dev->info;
	uint8_t mmap_export{0};
	auto status = doca_devinfo_get_is_mmap_export_supported(devinfo, &mmap_export);
	if (status != DOCA_SUCCESS)
		return status;

	if (mmap_export == 0) {
		DOCA_DLOG_ERR("Unable to export mmap: Device does not support mmap export.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	auto device_adapter = dev->device_adapter;
	device_abstraction::hca_capabilities hca_caps{};
	std::tie(status, hca_caps) = device_adapter->query_hca_caps();
	if (status != DOCA_SUCCESS)
		return status;

	int dev_id_in_mmap{0};
	std::tie(status, dev_id_in_mmap) = find_device_index(mmap, dev);
	if (status != DOCA_SUCCESS)
		return status;

	auto access_key = generate_access_key(hca_caps.vhca_id);

	status = mmap_allow_other_vhca(mmap, dev, dev_id_in_mmap, access_key);
	if (status != DOCA_SUCCESS)
		return status;

	std::unique_ptr<struct json_object, json_deleter_t> export_json;
	std::tie(status, export_json) = mmap_to_json(mmap, dev_id_in_mmap, hca_caps.vhca_id, access_key);
	if (status != DOCA_SUCCESS)
		return status;

	auto json_string = json_object_to_json_string_ext(export_json.get(), JSON_C_TO_STRING_PLAIN);
	if (json_string == nullptr) {
		DOCA_DLOG_ERR("Failed to create json string");
		return DOCA_ERROR_NO_MEMORY;
	}

	// the string length doesn't include the '\0' at the end of the json string, therefore
	// json_string_length is incremented by 1.
	auto json_string_length = (std::strlen(json_string) + 1);

	std::tie(status, *export_desc, *export_desc_len) = prepare_mmap_export_buffer(json_string, json_string_length);
	if (status != DOCA_SUCCESS)
		return status;

	mmap->exported = true;

	return DOCA_SUCCESS;
}

static std::unique_ptr<struct doca_mmap, mmap_deleter_t>
create_mmap(const union doca_data *user_data)
{
	struct doca_mmap *mmap{nullptr};
	auto status = doca_mmap_create(user_data, &mmap);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Failed to create mmap");
		return std::unique_ptr<struct doca_mmap, mmap_deleter_t>(nullptr);
	}

	std::unique_ptr<struct doca_mmap, mmap_deleter_t> unique_mmap(
		mmap, [](struct doca_mmap *mmap_to_delete) { doca_mmap_destroy(mmap_to_delete); });

	return unique_mmap;
}

static doca_error_t
validate_import_buffer(std::vector<uint8_t> &import_buffer)
{
	if (import_buffer.size() < sizeof(mmap_export_header)) {
		DOCA_DLOG_ERR("Invalid size (minimal %zu, received %zu)", sizeof(mmap_export_header),
			      import_buffer.size());
		return DOCA_ERROR_INVALID_VALUE;
	}
	// This method modifies the import buffer
	mmap_export_header *header = reinterpret_cast<mmap_export_header *>(import_buffer.data());
	if (header->version != MMAP_EXPORT_VERSION) {
		DOCA_DLOG_ERR("Invalid version (expected %d, received %d)", MMAP_EXPORT_VERSION, header->version);
		return DOCA_ERROR_INVALID_VALUE;
	}

	uint16_t header_length = be16toh(header->length);
	if (header_length != import_buffer.size()) {
		DOCA_DLOG_ERR("Invalid size (buffer length = %zu, header indicates %u)", import_buffer.size(),
			      (uint32_t)header_length);
		return DOCA_ERROR_INVALID_VALUE;
	}

	uint32_t header_crc = be32toh(header->crc);
	header->crc = 0;
	uint32_t crc = core_utils::calculate_crc32_normal(import_buffer.data(), import_buffer.size());
	if (crc != header_crc) {
		DOCA_DLOG_ERR("Invalid CRC (Header CRC = 0x%X, Calculated CRC = 0x%X)", header_crc, crc);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return DOCA_SUCCESS;
}

static void
decrypt_import_buffer(std::vector<uint8_t> &import_buffer)
{
	auto json_string_length = (import_buffer.size() - sizeof(mmap_export_header));
	auto json_string_buffer = reinterpret_cast<char *>(import_buffer.data() + sizeof(mmap_export_header));

	// The method decrypts the data to the same buffer.
	encrypt_decrypt_export_desc(json_string_buffer, json_string_buffer, json_string_length);
}

std::unique_ptr<struct json_object, json_deleter_t>
import_buffer_to_json(std::vector<uint8_t> &import_buffer)
{
	auto json_string_buffer = reinterpret_cast<char *>(import_buffer.data() + sizeof(mmap_export_header));

	std::unique_ptr<struct json_object, json_deleter_t> import_json(
		json_tokener_parse(json_string_buffer), [](struct json_object *jobj) { json_object_put(jobj); });

	if (!import_json) {
		DOCA_DLOG_ERR("Failed to parse import buffer to json object");
		return std::unique_ptr<struct json_object, json_deleter_t>(nullptr);
	}

	return import_json;
}

static doca_error_t
populate_chunk_from_export(struct doca_mmap *mmap, struct json_object *chunk_obj, uint16_t vhca_id,
			   const device_abstraction::allow_other_vhca_access_key_t &access_key, struct doca_dev *dev)
{
	if (chunk_obj == nullptr) {
		DOCA_DLOG_ERR("Failed to populate chunk, json object is null");
		return DOCA_ERROR_INVALID_VALUE;
	}

	struct json_object *addr_json{nullptr};
	struct json_object *len_json{nullptr};
	struct json_object *mkey_json{nullptr};
	struct json_object *page_size_json{nullptr};

	if (!json_object_object_get_ex(chunk_obj, "addr", &addr_json) ||
	    !json_object_object_get_ex(chunk_obj, "len", &len_json) ||
	    !json_object_object_get_ex(chunk_obj, "mkey_arr", &mkey_json) ||
	    !json_object_object_get_ex(chunk_obj, "page_size", &page_size_json)) {
		DOCA_DLOG_ERR("Unable to create mmap from export: Internal error - Failed to parse json string");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto new_mchunk = &(mmap->mchunk_arr[mmap->curr_mchunk_num]);
	try {
		new_mchunk->device_to_mkey_arr.resize(mmap->max_device_num);
	} catch (...) {
		DOCA_DLOG_ERR("Unable to add memory range: Internal error - Failed to allocate device to mkey array");
		return DOCA_ERROR_NO_MEMORY;
	}

	new_mchunk->free_cb = nullptr;
	new_mchunk->opaque = nullptr;
	new_mchunk->addr = reinterpret_cast<void *>(json_object_get_int64(addr_json));
	new_mchunk->len = static_cast<size_t>(json_object_get_int64(len_json));
	new_mchunk->pg_sz = static_cast<size_t>(json_object_get_int64(page_size_json));
	new_mchunk->mmap = mmap;

	// There is only one device in this MMAP so it is safe to use index 0 (see resize above).
	auto device_to_mkey = &new_mchunk->device_to_mkey_arr[0];
	device_to_mkey->device = dev;

	unsigned int i{0};
	for (auto mkey_access : mkey_accesses) {
		auto mkey_json_obj = json_object_array_get_idx(mkey_json, static_cast<uint32_t>(mkey_access));
		if (!mkey_json_obj) {
			DOCA_DLOG_ERR("Failed to get mkey from export");
			return DOCA_ERROR_INVALID_VALUE;
		}
		auto mkey = static_cast<uint32_t>(json_object_get_int(mkey_json_obj));
		doca_error_t status{DOCA_SUCCESS};
		std::shared_ptr<device_abstraction::imemory_region> mr;
		std::tie(status, mr) = dev->pd->create_alias_memory_region(mkey, vhca_id, access_key);
		device_to_mkey->memory_regions[i++] = std::move(mr);
		if (status != DOCA_SUCCESS) {
			DOCA_DLOG_ERR("Failed to create cross GVMI mkey for chunk %d", mmap->curr_mchunk_num);
			return status;
		}
	}

	mmap->curr_mchunk_num++;

	return DOCA_SUCCESS;
}

static std::pair<doca_error_t, device_abstraction::allow_other_vhca_access_key_t>
json_to_access_key(struct json_object *access_key_jobj)
{
	device_abstraction::allow_other_vhca_access_key_t access_key{};
	auto array_length = json_object_array_length(access_key_jobj);
	if (array_length != access_key.size()) {
		DOCA_DLOG_ERR("Invalid access key size (expected %zu, received %d)", access_key.size(),
			      static_cast<uint32_t>(array_length));
	}

	for (size_t i = 0; i < array_length; i++) {
		auto node_obj = json_object_array_get_idx(access_key_jobj, i);
		if (node_obj)
			access_key[i] = static_cast<uint8_t>(json_object_get_int(node_obj));
		else
			return {DOCA_ERROR_INVALID_VALUE, access_key};
	}

	return {DOCA_SUCCESS, access_key};
}

static doca_error_t
json_to_mmap(struct doca_mmap *mmap, struct json_object *import_jobj, struct doca_dev *dev)
{
	if (import_jobj == nullptr)
		return DOCA_ERROR_INVALID_VALUE;

	struct json_object *vhca_id_jobj{nullptr};
	struct json_object *access_key_jobj{nullptr};
	struct json_object *mchunks_jobj{nullptr};

	/* parse json */
	if (!json_object_object_get_ex(import_jobj, "vhca_id", &vhca_id_jobj) ||
	    !json_object_object_get_ex(import_jobj, "access_key", &access_key_jobj) ||
	    !json_object_object_get_ex(import_jobj, "mchunks", &mchunks_jobj)) {
		DOCA_DLOG_ERR("Unable to create mmap from export: Internal error - Failed to parse json object");
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto vhca_id = (uint16_t)json_object_get_int(vhca_id_jobj);
	doca_error_t status{DOCA_SUCCESS};
	device_abstraction::allow_other_vhca_access_key_t access_key{};
	std::tie(status, access_key) = json_to_access_key(access_key_jobj);
	if (status != DOCA_SUCCESS)
		return status;

	mmap->max_mchunk_num = static_cast<uint32_t>(json_object_array_length(mchunks_jobj));

	// MMAP must be started before population.
	status = doca_mmap_start(mmap);
	if (status != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from export: Internal error - Failed to start mmap");
		return status;
	}

	for (uint32_t chunk_id = 0; chunk_id < mmap->max_mchunk_num; chunk_id++) {
		status = populate_chunk_from_export(mmap, json_object_array_get_idx(mchunks_jobj, chunk_id), vhca_id,
						    access_key, dev);
		if (status != DOCA_SUCCESS)
			return status;
	}

	mmap->from_export = true;
	mmap->device_arr[0] = dev;
	mmap->curr_device_num++;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_create_from_export(const union doca_data *user_data, const void *export_desc, size_t export_desc_len, struct doca_dev *dev,
			     struct doca_mmap **mmap)
{
	if (export_desc == NULL || export_desc_len == 0 || dev == NULL || mmap == NULL) {
		DOCA_DLOG_ERR(
			"Unable to create mmap from export: Received invalid input - export_desc=%p, export_desc_len=%zu, dev=%p, mmap=%p",
			export_desc, export_desc_len, dev, mmap);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto devinfo = &dev->info;
	uint8_t mmap_from_export{0};
	auto status = doca_devinfo_get_is_mmap_from_export_supported(devinfo, &mmap_from_export);
	if (status != DOCA_SUCCESS)
		return status;

	if (mmap_from_export == 0) {
		DOCA_DLOG_ERR("Device does not support creating mmap from export.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	// The import buffer must be copied locally because:
	// 1. It is passed as const and one can not assume that the application will not use it again CRC and decryption
	//	requires changing the content).
	// 2. Decrypting the data over the input buffer exposes the encrypted data.
	std::vector<uint8_t> local_import_buffer;
	try {
		local_import_buffer.resize(export_desc_len);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate local import buffer (size = %zu)", export_desc_len);
		return DOCA_ERROR_NO_MEMORY;
	}

	memcpy(local_import_buffer.data(), export_desc, export_desc_len);

	status = validate_import_buffer(local_import_buffer);
	if (status != DOCA_SUCCESS)
		return status;

	decrypt_import_buffer(local_import_buffer);


	auto import_json = import_buffer_to_json(local_import_buffer);
	if (!import_json)
		return DOCA_ERROR_INVALID_VALUE;

	auto new_mmap = create_mmap(user_data);
	if (!new_mmap) {
		DOCA_DLOG_ERR("Unable to create mmap from export: Internal error - Failed to create mmap");
		return DOCA_ERROR_NO_MEMORY;
	}

	status = json_to_mmap(new_mmap.get(), import_json.get(), dev);
	if (status != DOCA_SUCCESS)
		return status;

	*mmap = new_mmap.release();

	return DOCA_SUCCESS;
}

void
priv_doca_mmap_increase_doca_buf_num(struct doca_mmap *mmap)
{
	mmap->doca_buf_num++;
}

void
priv_doca_mmap_decrease_doca_buf_num(struct doca_mmap *mmap)
{
	mmap->doca_buf_num--;
}

bool
priv_doca_mmap_is_started(struct doca_mmap *mmap)
{
	return mmap->start;
}

struct priv_doca_mchunk *
priv_doca_mmap_is_buf_in_chunk(struct doca_mmap *mmap, void *addr, size_t len)
{
	struct priv_doca_mchunk *curr_mchunk;
	uint32_t i;

	for (i = 0; i < mmap->curr_mchunk_num; i++) {
		curr_mchunk = &mmap->mchunk_arr[i];
		if ((curr_mchunk->addr <= addr) &&
		    ((len + ((uint8_t *)addr - (uint8_t *)curr_mchunk->addr)) <= curr_mchunk->len))
			return curr_mchunk;
	}
	return NULL;
}

doca_error_t
priv_doca_mchunk_extract_mkey_by_device(const struct priv_doca_mchunk *mchunk, const struct doca_dev *dev,
					uint32_t doca_access_flags, uint32_t *lkey)
{
	auto dev_to_mkey = std::find_if(begin(mchunk->device_to_mkey_arr), end(mchunk->device_to_mkey_arr),
					[dev](const priv_doca_device_to_mkey& dev_to_mkey) {
						return dev_to_mkey.device == dev;
					});
	if (dev_to_mkey == std::end(mchunk->device_to_mkey_arr))
		return DOCA_ERROR_INVALID_VALUE;

	uint32_t access{0};
	switch (doca_access_flags)
	{
		case (DOCA_ACCESS_LOCAL_READ):
			{
				access = static_cast<uint32_t>(mkey_access::local_read);
				break;
			}
		case (DOCA_ACCESS_LOCAL_READ | DOCA_ACCESS_LOCAL_WRITE):
			{
				access = static_cast<uint32_t>(mkey_access::local_write);
				break;
			}
		default:
			return DOCA_ERROR_NOT_SUPPORTED;
	}
	(*lkey) = dev_to_mkey->memory_regions[access]->get_id();
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_mmap_create_from_rep(const union doca_data *user_data, struct doca_dev_rep *rep_dev, struct doca_mmap **mmap)
{
	if (rep_dev == NULL || mmap == NULL) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Invalid input rep_dev=%p, mmap=%p",
			      rep_dev, mmap);
		return DOCA_ERROR_INVALID_VALUE;
	}

	struct doca_dev *dev{nullptr};
	auto ret = priv_doca_dev_rep_parent_get(rep_dev, &dev);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to get dev parent");
		return ret;
	}

	device_abstraction::hca_capabilities hca_caps{};
	auto device_adapter = dev->device_adapter;
	std::tie(ret, hca_caps) = device_adapter->query_hca_caps();
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to get hca capabilities");
		return ret;
	}

	if (!hca_caps.crossing_vhca_mkey_supported) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: crossing vhca mkey is not supported");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	struct doca_mmap *new_mmap{nullptr};
	ret = doca_mmap_create(user_data, &new_mmap);
	if(ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to create mmap");
		return ret;
	}

	//mmap from representor holds one device
	ret = doca_mmap_set_max_num_devices(new_mmap, 1);
	if(ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to set max_num devices");
		doca_mmap_destroy(new_mmap);
		return ret;
	}

	//mmap from representor holds one chunk
	ret = doca_mmap_set_max_num_chunks(new_mmap, 1);
	if(ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to set max_num chunks");
		doca_mmap_destroy(new_mmap);
		return ret;
	}

	ret = doca_mmap_start(new_mmap);
	if(ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to start mmap");
		doca_mmap_destroy(new_mmap);
		return ret;
	}

	ret = mmap_populate(new_mmap, 0, SIZE_MAX, MCHUNK_PAGE_SIZE_DEFULT_VALUE, NULL, NULL);
	if(ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to populate mmap");
		doca_mmap_destroy(new_mmap);
		return ret;
	}

	new_mmap->mchunk_arr[0].device_to_mkey_arr[0].device = dev;

	auto vhca_id = priv_doca_dev_vhca_id_get(rep_dev);
	uint32_t introspection_mkey{0};

	ret = priv_doca_dev_rep_introspection_mkey_get(rep_dev, &introspection_mkey);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to get introspection mkey");
		doca_mmap_destroy(new_mmap);
		return ret;
	}

	std::shared_ptr<device_abstraction::imemory_region> mr;
	std::tie(ret, mr) = dev->pd->create_crossing_memory_region(introspection_mkey, vhca_id);
	if (ret != DOCA_SUCCESS) {
		DOCA_DLOG_ERR("Unable to create mmap from representor: Internal error: Failed to create cross mkey");
		doca_mmap_destroy(new_mmap);
		return ret;
	}
	for (int i=0; i < mkey_accesses.size(); i++) {
		new_mmap->mchunk_arr[0].device_to_mkey_arr[0].memory_regions[i] = mr;
	}

	new_mmap->from_representor = true;
	new_mmap->device_arr[0] = dev;
	new_mmap->curr_device_num++;
	*mmap = new_mmap;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_num_chunks(const struct doca_mmap *mmap, uint32_t *num_chunks)
{
	if (mmap == NULL || num_chunks == NULL) {
		DOCA_DLOG_ERR("Unable to get num chunks: Invalid input mmap=%p, num_chunks=%p", mmap, num_chunks);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*num_chunks = mmap->curr_mchunk_num;

	return DOCA_SUCCESS;
}

doca_error_t
doca_mmap_get_chunks(const struct doca_mmap *mmap, struct doca_mmap_chunk_info *chunks, uint32_t *num_chunks)
{
	if (mmap == NULL || chunks == NULL || num_chunks == NULL) {
		DOCA_DLOG_ERR("Unable to get chunks: Invalid input mmap=%p, num_chunks=%p, chunks=%p", mmap, num_chunks,
			      chunks);
		return DOCA_ERROR_INVALID_VALUE;
	}

	*num_chunks = std::min(mmap->curr_mchunk_num, *num_chunks);

	for (uint32_t i = 0; i < (*num_chunks); i++) {
		chunks[i].addr = mmap->mchunk_arr[i].addr;
		chunks[i].len = mmap->mchunk_arr[i].len;
	}

	return DOCA_SUCCESS;
}