/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>

#include <private/common_defines.h>
#include <doca_log.h>
#include <doca_dev.h>
#include "internal_doca_dev.h"
#include <idevice_adapter.hpp>
#include <ireceive_memory_pool.hpp>
#include <priv_doca_rmp.h>

DOCA_LOG_REGISTER(CORE::DOCA_RMP);

using namespace device_abstraction;

doca_error_t
priv_doca_dev_rmp_create(const struct doca_dev *dev, uint32_t wq_size, struct priv_doca_rmp **rmp)
{
	if ((dev == nullptr) || (wq_size == 0) || (rmp == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, wq_size=%u, rmp=%p", dev, wq_size, rmp);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<ireceive_memory_pool> rmp_obj;
	std::tie(status, rmp_obj) = dev->pd->create_receive_memory_pool(wq_size);
	if (status != DOCA_SUCCESS)
		return status;

	*rmp = new (std::nothrow) priv_doca_rmp;
	if (*rmp == nullptr) {
		rmp_obj.reset();
		DOCA_DLOG_ERR("Failed to allocate priv_doca_rmp");
		return DOCA_ERROR_NO_MEMORY;
	}

	(*rmp)->rmp_obj = rmp_obj;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_dev_rmp_destroy(struct priv_doca_rmp *rmp)
{
	if (rmp == nullptr) {
		DOCA_DLOG_ERR("Invalid rmp handle");
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rmp->rmp_obj == nullptr) {
		DOCA_DLOG_ERR("rmp object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	auto status = rmp->rmp_obj->destroy();
	if (status != DOCA_SUCCESS)
		return status;

	delete rmp;
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_rmp_id_get(const struct priv_doca_rmp *rmp, uint32_t *id)
{
	if ((rmp == nullptr) || (id == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input rmp=%p, id=%p", rmp, id);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rmp->rmp_obj == nullptr) {
		DOCA_DLOG_ERR("rmp object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	*id = rmp->rmp_obj->get_id();
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_rmp_wq_size_get(const struct priv_doca_rmp *rmp, uint32_t *wq_size)
{
	if ((rmp == nullptr) || (wq_size == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input rmp=%p, wq_size=%p", rmp, wq_size);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rmp->rmp_obj == nullptr) {
		DOCA_DLOG_ERR("rmp object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	*wq_size = rmp->rmp_obj->get_wq_size();
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_rmp_wq_buffer_get(const struct priv_doca_rmp *rmp, void **wq_buffer)
{
	if ((rmp == nullptr) || (wq_buffer == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input rmp=%p, wq_buffer=%p", rmp, wq_buffer);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rmp->rmp_obj == nullptr) {
		DOCA_DLOG_ERR("rmp object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	*wq_buffer = rmp->rmp_obj->get_wq_buffer();
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_rmp_wq_doorbell_get(const struct priv_doca_rmp *rmp, uint32_t **wq_doorbell)
{
	if ((rmp == nullptr) || (wq_doorbell == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input rmp=%p, wq_doorbell=%p", rmp, wq_doorbell);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (rmp->rmp_obj == nullptr) {
		DOCA_DLOG_ERR("rmp object is null");
		return DOCA_ERROR_UNEXPECTED;
	}

	*wq_doorbell = rmp->rmp_obj->get_wq_doorbell();
	return DOCA_SUCCESS;
}
