/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <memory>

#include <private/common_defines.h>
#include <doca_log.h>
#include <doca_dev.h>
#include "internal_doca_dev.h"
#include <idevice_adapter.hpp>
#include <iuser_access_region.hpp>

DOCA_LOG_REGISTER(CORE::DOCA_UAR);

using namespace device_abstraction;

doca_error_t
priv_doca_dev_uar_create(const struct doca_dev *dev, enum priv_doca_uar_allocation_type allocation_type,
			 struct priv_doca_uar **uar)
{
	if ((dev == nullptr) || (uar == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input dev=%p, uar=%p", dev, uar);
		return DOCA_ERROR_INVALID_VALUE;
	}

	doca_error_t status{DOCA_SUCCESS};
	std::shared_ptr<iuser_access_region> uar_obj{nullptr};
	std::tie(status, uar_obj) = dev->device_adapter->create_user_access_region(
		static_cast<iuser_access_region::allocation_type>(allocation_type));
	if (status != DOCA_SUCCESS)
		return status;

	*uar = new (std::nothrow) struct priv_doca_uar;
	if (*uar == nullptr) {
		DOCA_DLOG_ERR("Failed to allocate doca_uar");
		return DOCA_ERROR_NO_MEMORY;
	}

	(*uar)->uar_obj = uar_obj;
	return DOCA_SUCCESS;
}

void
priv_doca_dev_uar_destroy(struct priv_doca_uar *uar)
{
	// delete works on null as well so there is no need to check.
	// The only thing that the struct holds is a shared pointer to the UAR so deletion
	// will destroy that shared pointer which in turn will lead to the destruction of the UAR.
	delete uar;
}

doca_error_t
priv_doca_uar_id_get(const struct priv_doca_uar *uar, uint32_t *id)
{
	if ((uar == nullptr) || (id == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input uar=%p, id=%p", uar, id);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (uar->uar_obj == nullptr) {
		DOCA_DLOG_ERR("uar object is null");
		return DOCA_ERROR_BAD_STATE;
	}

	*id = uar->uar_obj->get_id();
	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_uar_page_get(const struct priv_doca_uar *uar, void **page)
{
	if ((uar == nullptr) || (page == nullptr)) {
		DOCA_DLOG_ERR("Received invalid input uar=%p, page=%p", uar, page);
		return DOCA_ERROR_INVALID_VALUE;
	}

	if (uar->uar_obj == nullptr) {
		DOCA_DLOG_ERR("uar object is null");
		return DOCA_ERROR_BAD_STATE;
	}

	*page = uar->uar_obj->get_page();
	return DOCA_SUCCESS;
}
