/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <priv_doca_os_abstraction.hpp>
#include <user_memory_registration.hpp>

#include <private/common_defines.h>
#include <priv_doca_umr.h>
#include <priv_doca_buf.h>
#include <doca_buf.h>
#include <doca_log.h>

#include "internal_doca_dev.h"

DOCA_LOG_REGISTER(CORE::DOCA_UMR);

using namespace device_abstraction;

struct priv_doca_umr {
	std::unique_ptr<iuser_memory_registration> umr_obj{nullptr};
	struct doca_dev *dev{nullptr};
};

doca_error_t
priv_doca_dev_umr_create(const struct doca_dev *dev, uint32_t num_elements, struct priv_doca_umr **umr_out)
{
	if (unlikely(dev == nullptr || umr_out == nullptr)) {
		DOCA_DLOG_ERR("Unable to create UMR, received invalid input. dev=%p, umr_out=%p", dev, umr_out);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto umr_ret_obj = dev->pd->create_user_memory_registration(num_elements);
	if (umr_ret_obj.first != DOCA_SUCCESS)
		return umr_ret_obj.first;

	auto umr = new (std::nothrow) struct priv_doca_umr;
	if (unlikely(umr == nullptr))
		return DOCA_ERROR_NO_MEMORY;

	umr->umr_obj = std::move(umr_ret_obj.second);
	umr->dev = const_cast<struct doca_dev *>(dev);

	*umr_out = umr;

	return DOCA_SUCCESS;
}

doca_error_t
priv_doca_umr_buf_to_wqe(struct priv_doca_umr *umr, const struct doca_buf *buf, void *wqe, uint16_t pi,
			 uint32_t qp_num, struct priv_doca_data_segment *indirect_key,
			 bool *need_fence, struct priv_doca_umr_handle **handle_out)
{
	uint32_t max_num_bufs{0};

	auto mkeys = umr->umr_obj->umr_start(&max_num_bufs);
	if (unlikely(mkeys == nullptr)) {
		DOCA_DLOG_ERR("Unable to fill umr wqe from buf. Maximum number of elements exceeded.");
		return DOCA_ERROR_NOT_PERMITTED;
	}

	const struct doca_buf *next_buf = buf;
	uint32_t total_len{0};
	uint32_t num_elements{0};
	doca_error_t ret{DOCA_SUCCESS};

	auto mkey = static_cast<iuser_memory_registration::mkey*>(*mkeys);
	while (next_buf != NULL) {
		++num_elements;
		if (unlikely(num_elements > max_num_bufs))
			break;

		/* Fill KLM array representing the scattered buffers.*/
		auto buf_len = static_cast<uint32_t>(priv_doca_buf_get_data_len(next_buf));
		uint32_t lkey{0};
		ret = doca_buf_get_lkey(next_buf, umr->dev, DOCA_ACCESS_LOCAL_READ, &lkey);
		if (unlikely(ret != DOCA_SUCCESS))
			break;

		mkey->byte_count = htobe32(buf_len);
		mkey->id = htobe32(lkey);
		mkey->address = htobe64(reinterpret_cast<uint64_t>(priv_doca_buf_get_data(next_buf)));

		total_len += buf_len;

		static_cast<void>(priv_doca_buf_list_next(next_buf, &next_buf));
		++mkey;
	}

	if (unlikely(ret != DOCA_SUCCESS || num_elements > max_num_bufs)) {
		DOCA_DLOG_ERR("Unable to fill umr wqe from buf. provided_buffers=%u, permitted_buffers=%u",
			      num_elements, max_num_bufs);
		umr->umr_obj->umr_done(mkeys);
		return DOCA_ERROR_NOT_PERMITTED;
	}

	umr->umr_obj->umr_commit(mkeys, num_elements, total_len, wqe, pi, qp_num,
				 reinterpret_cast<iuser_memory_registration::mkey*>(indirect_key));

	*handle_out = reinterpret_cast<priv_doca_umr_handle*>(mkeys);
	*need_fence = num_elements > 1;

	return DOCA_SUCCESS;
}

void
priv_doca_umr_mkey_free(struct priv_doca_umr *umr, struct priv_doca_umr_handle *handle)
{
	umr->umr_obj->umr_done(reinterpret_cast<iuser_memory_registration::mkey_array*>(handle));
}

doca_error_t
priv_doca_dev_umr_destroy(struct priv_doca_umr *umr)
{
	if (unlikely(umr == nullptr)) {
		DOCA_DLOG_ERR("Unable to destroy umr, received invalid input. umr=%p", umr);
		return DOCA_ERROR_INVALID_VALUE;
	}

	auto ret = umr->umr_obj->destroy();
	if (unlikely(ret != DOCA_SUCCESS))
		return ret;

	delete umr;
	return DOCA_SUCCESS;
}
