/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the buffer, exposing implementation details.
 * This header should not be exposed to applications.
 */

#pragma once

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string>
#include <vector>

#include <priv_doca_compat.h>
#include <doca_error.h>
#include <doca_mmap.h>
#include <doca_buf.h>
#include <idevice_adapter.hpp>
#include <priv_doca_buf.h>
#include "buf_extensions_map.hpp"
#include "mmap.hpp"

/**
 * This struct contains resources that are required by the doca buffer.
 * This indirection is required because the doca buffer memory footprint must be as small as possible.
 * In addition, this file can't include doca_buf_inventory.hpp (creates circular include) but it needs access to the
 * extensions map (buf_extensions_map.hpp) in order to implement the inline extension methods (that are required to be
 * inline for performance).
 */
struct doca_buf_resources {
	struct doca_buf_inventory *inventory{nullptr};	/**< pointer to the inventory that maintains this buffer */
	doca_core::buf_extensions_map const *extensions_map{nullptr}; /**< pointer to the extensions map */
};

/**
 * Doca buf size is super important. Therefore it is packed and the members are ordered by size to minimize the
 * impact of packing.
 */
#pragma pack(push, 1)
struct doca_buf {
	// 64 bit fields
	uint8_t *head{nullptr};					/**< Virtual address of the buffer */
	size_t len{0};						/**< Buffer length */
	uint8_t *data{nullptr};					/**< Data address */
	size_t data_len{0};					/**< Data length */
	struct priv_doca_mchunk *mchunk{nullptr};		/**< Memory chunk that contains the buffer */
	doca_buf_resources const *resources{nullptr};	/**< Buffer resources */

	// 16 bit fields
	uint16_t refcnt{0}; /**< doca_buf reference count */

	// Extensions buffer - MUST BE LAST
	uint8_t extensions[0];
};
#pragma pack(pop)

/************************************************************************/
/*                                                                      */
/*                            Extensions                                */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/* Linked List Extension                                                 */
/************************************************************************/
#pragma pack(push, 1)
struct doca_buf_linked_list_extension {
	struct doca_buf *next{nullptr}; /**< Pointer to next doca_buf */
	bool is_head{true};		/**< Flag to indicate if the buffer is the head of a list.
					 *  An un-chaind buffer is considered the head of a list
					 *  containing only the buffer itself. */
};
#pragma pack(pop)

inline struct doca_buf_linked_list_extension *
get_linked_list_extension(struct doca_buf *buf)
{
	return reinterpret_cast<struct doca_buf_linked_list_extension *>(
		buf->resources->extensions_map->get_extension(buf->extensions,
							      doca_core::buf_extension_type::LINKED_LIST));
}

inline const struct doca_buf_linked_list_extension *
get_linked_list_extension(const struct doca_buf *buf)
{
	return reinterpret_cast<const struct doca_buf_linked_list_extension *>(
		buf->resources->extensions_map->get_extension(buf->extensions,
							      doca_core::buf_extension_type::LINKED_LIST));
}

/**
 * This method can be used if the offset is known
 */
inline struct doca_buf_linked_list_extension *
get_linked_list_extension(struct doca_buf *buf, size_t offset)
{
	return reinterpret_cast<struct doca_buf_linked_list_extension *>(buf->extensions + offset);
}

/**
 * This method can be used if the offset is known
 */
inline const struct doca_buf_linked_list_extension *
get_linked_list_extension(const struct doca_buf *buf, size_t offset)
{
	return reinterpret_cast<const struct doca_buf_linked_list_extension *>(buf->extensions + offset);
}


/** @} */
