/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the device, exposing implementation details.
 * This header should not be exposed to applications.
 */

#pragma once

#include <cstring>
#include <atomic>
#include <stdbool.h>
#include <stdint.h>
#include <unordered_set>
#include <mutex>

#include <idevice_adapter.hpp>

#include <doca_error.h>
#include <doca_dev.h>
#include <priv_doca_dev.h>
#include <priv_doca_compat.h>
#include <priv_doca_dev.h>

struct priv_doca_dev_observer {
	union doca_data cookie;
	struct priv_doca_dev_observer_ops ops;

	priv_doca_dev_observer(const priv_doca_dev_observer_ops &ops, union doca_data cookie):cookie{cookie}
	{
		this->ops.on_dev_close = ops.on_dev_close;
	}

	~priv_doca_dev_observer() = default;
	priv_doca_dev_observer &operator = (priv_doca_dev_observer const &) = delete;

	bool operator == (const priv_doca_dev_observer &observer) const noexcept
	{
		return (!std::memcmp(&cookie, &observer.cookie, sizeof(union doca_data))) && (ops.on_dev_close == observer.ops.on_dev_close);
	}
};

struct priv_doca_dev_observer_hash
{
	std::size_t operator()(priv_doca_dev_observer const& observer) const noexcept
	{
		size_t seed = 0;

		// uses 0x9e3779b9 is the "golden ratio" for "uniformly distribution".
		std::size_t h1 = std::hash<uintptr_t>{}(reinterpret_cast<uintptr_t>(observer.ops.on_dev_close));
		seed ^= h1 + 0x9e3779b9 + (seed << 6) + (seed >> 2);

		std::size_t h2 = std::hash<uintptr_t>{}(reinterpret_cast<uintptr_t>(observer.cookie.ptr));
		seed ^= h2 + 0x9e3779b9 + (seed << 6) + (seed >> 2);

		std::size_t h3 = std::hash<uint64_t>{}(observer.cookie.u64);
		seed ^= h3 + 0x9e3779b9 + (seed << 6) + (seed >> 2);

		return seed;
	}
};

struct doca_devinfo {
	struct doca_dev *dev;
};

struct doca_dev {
	struct doca_devinfo info {};
	std::atomic_long refcount{ 0 };
	std::shared_ptr<device_abstraction::idevice_adapter> device_adapter;
	std::shared_ptr<device_abstraction::iprotection_domain> pd;
	bool zombie {false}; /**< Device no longer exists. */
	std::unordered_set<priv_doca_dev_observer, priv_doca_dev_observer_hash> observers;
};

struct doca_devinfo_rep {
	doca_devinfo_rep() = default;
	doca_devinfo_rep(struct doca_dev *dev_parent, enum doca_pci_func_type func_type,
			    enum priv_doca_rep_type rep_type, uint16_t pf_vhca_id, uint16_t vhca_id,
			    uint16_t pci_bdf, uint16_t vport, uint32_t introspection_mkey,
			    uint16_t nb_total_vfs, uint16_t nb_vfs, bool hotplug)
	: dev_parent(dev_parent)
	, func_type(func_type)
	, rep_type(rep_type)
	, pf_vhca_id(pf_vhca_id)
	, vhca_id(vhca_id)
	, pci_bdf(pci_bdf)
	, vport(vport)
	, introspection_mkey(introspection_mkey)
	, nb_total_vfs(nb_total_vfs)
	, nb_vfs(nb_vfs)
	, hotplug(hotplug) {}

	doca_devinfo_rep(const device_abstraction::emulated_devinfo &emu_fn, struct doca_dev *dev,
			    enum doca_pci_func_type func_type,
			    uint16_t nb_vfs, enum priv_doca_rep_type rep_type, uint16_t vport,
			    uint32_t introspection_mkey)
	: doca_devinfo_rep(dev, func_type, rep_type, emu_fn.vhca_id, emu_fn.vhca_id, emu_fn.pci_bdf,
			      vport, introspection_mkey, emu_fn.nb_total_vfs, nb_vfs, emu_fn.hotplug) {}

	struct doca_dev *dev_parent{nullptr};
	enum doca_pci_func_type func_type { doca_pci_func_type::DOCA_PCI_FUNC_PF};
	enum priv_doca_rep_type rep_type { priv_doca_rep_type::PRIV_DOCA_REP_UNKNOWN};
	uint16_t pf_vhca_id{0};   /**< Holds different value from vhca_id in case func_type is VF.*/
	uint16_t vhca_id{0};
	uint16_t pci_bdf{0};
	uint16_t vport{0};
	uint32_t introspection_mkey{0};
	uint16_t nb_total_vfs{0}; /**< Set for func_type PF only.*/
	uint16_t nb_vfs{0};	  /**< Set for func_type PF only.*/
	bool hotplug{false};	  /**< Set for func_type PF only.*/
};

struct doca_dev_rep {
	struct doca_devinfo_rep info;
};

struct doca_devx_object {
	std::unique_ptr<device_abstraction::idevx_object> devx_object;
};

struct doca_umem_object {
	std::shared_ptr<device_abstraction::imapped_user_memory> umem;
};

struct priv_doca_uar {
	std::shared_ptr<device_abstraction::iuser_access_region> uar_obj;
};

struct priv_doca_cq {
	std::shared_ptr<device_abstraction::icompletion_queue> cq_obj;
};

struct priv_doca_comp_events_channel {
	std::shared_ptr<device_abstraction::icompletion_events_channel> channel_obj;
};

struct priv_doca_channel_service {
	std::shared_ptr<device_abstraction::ichannel_service> channel_service_obj;
};

struct priv_doca_channel_connection {
	std::shared_ptr<device_abstraction::ichannel_connection> channel_connection_obj;
};

struct priv_doca_rmp {
	std::shared_ptr<device_abstraction::ireceive_memory_pool> rmp_obj;
};

struct priv_doca_dct {
	std::shared_ptr<device_abstraction::idynamically_connected_target> dct_obj;
};

/**
 * @brief This method retrieves the device handle from a doca_devinfo
 * @param [in] DOCA devinfo
 * @return idevice or null
 */
device_abstraction::idevice_adapter *priv_doca_devinfo_to_device_handle(const struct doca_devinfo *di);
