/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * This is a private header of the buffer, exposing implementation details.
 * This header should not be exposed to applications.
 */

#pragma once

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string>
#include <vector>
#include <unordered_map>

#include <priv_doca_compat.h>
#include <doca_error.h>
#include <doca_types.h>
#include <doca_mmap.h>
#include <doca_buf.h>
#include <idevice_adapter.hpp>
#include <priv_doca_buf.h>
#include "buf_extensions_map.hpp"

using memory_region_ptr_t = std::shared_ptr<device_abstraction::imemory_region>;

enum class mkey_access : uint32_t {
	local_read = 0,
	local_write = 1,
};

static constexpr std::array<mkey_access, 2> mkey_accesses = {mkey_access::local_read, mkey_access::local_write};

static constexpr std::array<uint32_t, mkey_accesses.size()> mkey_access_to_flags {{ DOCA_ACCESS_LOCAL_READ,
										    DOCA_ACCESS_LOCAL_READ | DOCA_ACCESS_LOCAL_WRITE }};

struct priv_doca_device_to_mkey {
	struct doca_dev *device{nullptr};
	std::array<memory_region_ptr_t, mkey_accesses.size()> memory_regions;
};

struct priv_doca_mchunk {
	void *addr{nullptr};			/**< Virtual address of chunk */
	size_t len{0};				/**< Chunk address length */
	size_t pg_sz{0};			/**< Page size alignment of chunk */
	uint32_t doca_buf_cnt{0};		/**< Counter of total elements pointing to chunk */
	std::vector<priv_doca_device_to_mkey>
		device_to_mkey_arr;		/**< Array for mapping device to its mkey */
	doca_mmap_memrange_free_cb_t
		*free_cb{nullptr};		/**< Free callback */
	void *opaque{nullptr};			/**< Argument passed to the free callback */
	struct doca_mmap *mmap{nullptr};	/**< Memory map to which the chunk belongs */
};

using zba_mkey_map_entry =
	std::unordered_map<const struct doca_dev *, std::shared_ptr<device_abstraction::imemory_region>>;

struct doca_mmap {
	union doca_data m_user_data;			/**< Memory map user_data*/
	std::unordered_map<const struct doca_buf *, zba_mkey_map_entry> zba_mkey_map;
	uint32_t max_mchunk_num{0};			/**< Max Number of memory chunks */
	uint32_t curr_mchunk_num{0};			/**< Current number of memory chunks in memory map */
	std::vector<struct priv_doca_mchunk> mchunk_arr;/**< Array of memory chunks */
	uint32_t max_device_num{0};			/**< Max Number of devices */
	uint32_t curr_device_num{0};			/**< Current number of devices in memory map */
	std::vector<struct doca_dev *>device_arr;	/**< Array of devices pointers */
	uint32_t doca_buf_num{0};			/**< Number of doca_bufs in memory map */

	bool exported{false};				/**< Flag to indicate if mmap is exported */
	bool from_export{false};			/**< Flag to indicate if mmap is from export */
	bool from_representor{false};			/**< Flag to indicate if mmap is from representor */
	bool start{false};				/**< Flag to indicate if mmap is started or unstarted/stop */
	bool previously_started{false};			/**< Flag to indicate if mmap had ever been started */
};

#pragma pack(push, 1)
struct mmap_export_header {
	uint8_t version;
	uint8_t padding;
	uint16_t length;
	uint32_t crc;
};
#pragma pack(pop)


/** @} */
