/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <cmath>

namespace core_utils
{

/**
 * @brief This method calculates CRC32
 * @param buffer [in] buffer to calculate CRC32 for
 * @param length [in] buffer length
 * @return CRC32
 */
template <uint32_t polinomeal>
uint32_t calculate_crc32(const uint8_t *buffer, size_t length)
{
	uint32_t crc{0xFFFFFFFF};

	for (size_t i = 0; i < length; i++) {
		uint8_t node{buffer[i]};
		for (size_t j = 0; j < 8; j++) {
			uint32_t lsb{(node ^ crc) & 1};
			crc >>= 1;
			if (lsb)
				crc = crc ^ polinomeal;
			node >>= 1;
		}
	}

	return ~crc;
}

/**
 * @brief This method calculates normal CRC32 (msb-first)
 * @param buffer [in] buffer to calculate CRC32 for
 * @param length [in] buffer length
 * @return CRC32
 */
static inline uint32_t
calculate_crc32_normal(const uint8_t *buffer, size_t length)
{
	return calculate_crc32<0x04c11db7>(buffer, length);
}

/**
 * @brief This method calculates reversed CRC32 (lsb-first)
 * @param buffer [in] buffer to calculate CRC32 for
 * @param length [in] buffer length
 * @return CRC32
 */
static inline uint32_t
calculate_crc32_reversed(const uint8_t *buffer, size_t length)
{
	return calculate_crc32<0xEDB88320>(buffer, length);
}

static inline uint32_t
align_uint32_t(uint32_t size, uint32_t alignment)
{
	uint32_t remainder = size % alignment;
	if (remainder == 0)
		return size;

	return size + (alignment - remainder);
}

static inline uint32_t
log2_uint32_t(uint32_t x)
{
	if (x == 0)
		/* log(0) is undefined */
		return 0;

	return static_cast<uint32_t>(std::log2(x));
}

static inline bool
is_power_of_two(uint64_t x)
{
    return x && (x & (x - 1)) == 0;
}

} // namespace core_utils

/** @} */
