/*
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#ifdef __linux__
#include <syslog.h>
#endif /* __linux__ */

#include <atomic>
#include <memory>
#include <mutex>
#include <vector>
#include <algorithm>
#include <unordered_map>

#include "priv_doca_os_abstraction.hpp"
#include "doca_log.h"
#include "private/common_defines.h"

using namespace std;

#define DOCA_LOG_PREFIX_FORMAT "[%s:%06u][DOCA][%s][%s:%d]: "
#define DEFAULT_RATE_LIMIT_BUCKET_TIME 1 /* Default value, 1 second */
#define DEFAULT_RATE_LIMIT_QUANTITY 4	 /* Default value, 4 log events */
#define RATE_LIMIT_PRINTED_MASK 0xFFFFFFFF
#define RATE_LIMIT_BUCKET_SHIFT 32
#define MAX_SYSLOG_NAME_LEN (64)

static_assert(sizeof(atomic_ullong) == 8, "atomic_ullong shall be 64 bits");

enum log_backend_type {
	LOG_BACKEND_FILE,
	LOG_BACKEND_FD,
	LOG_BACKEND_BUFFER,
	LOG_BACKEND_SYSLOG,
};

static uint16_t rate_limit_bucket_time = DEFAULT_RATE_LIMIT_BUCKET_TIME;
static uint16_t rate_limit_quantity_per_bucket = DEFAULT_RATE_LIMIT_QUANTITY;

struct rate_bucket {
	std::atomic_flag lock = ATOMIC_FLAG_INIT;	/* Flag to distinguish active thread */
	atomic_ullong meta{0};	/* Left 32 bits - bucket start time, right 32 bits - printed messages */
};

struct logger_sources {
	logger_sources(const char *source_name) : name(source_name) {}
	string name;
	std::vector<std::unique_ptr<rate_bucket>> buckets;	/* Rate buckets array */
};

struct logger_sources_array {
	static logger_sources_array &get_instance()
	{
		static logger_sources_array m_instance;
		return m_instance;
	}
	size_t num_freed{0};			/* Number of freed records in the data structure (teardown) */
	mutex lock;				/* Synchronization object - thread safety */
	std::vector<logger_sources> sources;	/* Source struct array */
};

struct buffer_stream {
	char *buffer{nullptr};				/* User's LogRecord buffer */
	size_t capacity{0};				/* Size (in bytes) of the buffer */
	log_flush_callback flush_handler{nullptr};	/* User's flush callback */
};

struct doca_logger_backend {
	uint32_t level{DOCA_LOG_LEVEL_INFO};
	enum log_backend_type type { LOG_BACKEND_FILE };
	union {
		FILE *fptr;
		int fd;
		struct buffer_stream buffer;
		// Can't use std::string in a union.
		char syslog_name[MAX_SYSLOG_NAME_LEN];
	} stream{0};
};

struct logger_backends_array {
	mutex lock;				/* Synchronization object - thread safety */
	vector<doca_logger_backend> values;	/* The values of the backends themselves */
};

struct logger_backends {
	static logger_backends &get_instance()
	{
		static logger_backends m_instance;
		return m_instance;
	}
	struct doca_logger_backend default_logger;
	struct logger_backends_array backends;
};

#define log_sources logger_sources_array::get_instance()
#define doca_loggers logger_backends::get_instance()
#define global_logger (doca_loggers.default_logger)

doca_error_t
doca_log_stream_redirect(FILE *stream)
{
	if (stream == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	fflush(global_logger.stream.fptr);
	global_logger.stream.fptr = stream;
	return DOCA_SUCCESS;
}

static struct doca_logger_backend *
log_backend_get(uint32_t index)
{
	if (index == 0)
		return &global_logger;

	--index;

	lock_guard<mutex> guard(doca_loggers.backends.lock);
	if (index >= doca_loggers.backends.values.size())
		return NULL;

	return &doca_loggers.backends.values[index];
}

static struct doca_logger_backend *
log_backend_allocate()
{
	lock_guard<mutex> guard(doca_loggers.backends.lock);

	try {
		// No need to worry about capacity, vector will take care of it.
		doca_loggers.backends.values.emplace_back(doca_logger_backend());
		return &doca_loggers.backends.values.back();
	} catch (...) {
		return NULL;
	}
}

uint16_t
doca_log_get_bucket_time(void)
{
	return rate_limit_bucket_time;
}

void
doca_log_set_bucket_time(uint16_t bucket_time)
{
	rate_limit_bucket_time = bucket_time;
}

uint16_t
doca_log_get_quantity(void)
{
	return rate_limit_quantity_per_bucket;
}

void
doca_log_set_quantity(uint16_t quantity)
{
	rate_limit_quantity_per_bucket = quantity;
}

static const char *
get_log_level_name(uint32_t level)
{
	switch (level) {
	case doca_log_level::DOCA_LOG_LEVEL_CRIT:
		return "CRT";
	case doca_log_level::DOCA_LOG_LEVEL_ERROR:
		return "ERR";
	case doca_log_level::DOCA_LOG_LEVEL_WARNING:
		return "WRN";
	case doca_log_level::DOCA_LOG_LEVEL_INFO:
		return "INF";
	case doca_log_level::DOCA_LOG_LEVEL_DEBUG:
		return "DBG";
	default:
		return NULL;
	}
}

doca_error_t
doca_log_backend_level_set(struct doca_logger_backend *logger, uint32_t level)
{
	if (get_log_level_name(level) == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	logger->level = level;
	return DOCA_SUCCESS;
}

/* Exists for Backwards Compatibility reasons (and ease of use) */
doca_error_t
doca_log_global_level_set(uint32_t level)
{
	return doca_log_backend_level_set(&global_logger, level);
}

uint32_t
doca_log_global_level_get(void)
{
	return global_logger.level;
}

static int
source_lookup(const char *source_name)
{
	auto it = find_if(log_sources.sources.begin(), log_sources.sources.end(),
				[&](const logger_sources &source) {
					return source.name == source_name;
				});

	if (it != log_sources.sources.end())
		return static_cast<int>(distance(log_sources.sources.begin(), it));

	return -1;
}

doca_error_t
doca_log_source_register(const char *source_name, int *source)
{
	int id;

	if (unlikely(source == NULL))
		return DOCA_ERROR_INVALID_VALUE;
	*source = -1;

	lock_guard<mutex> guard(doca_loggers.backends.lock);
	/* Don't mix register/destroy (should never happen) */
	if (unlikely(log_sources.num_freed > 0))
		return DOCA_ERROR_BAD_STATE;

	id = source_lookup(source_name);
	if (unlikely(id >= 0)) {
		*source = id;
		return DOCA_SUCCESS;
	}

	try {
		id = static_cast<int>(log_sources.sources.size());
		log_sources.sources.emplace_back(logger_sources(source_name));
	} catch (...) {
		return DOCA_ERROR_NO_MEMORY;
	}

	*source = id;
	return DOCA_SUCCESS;
}

doca_error_t
doca_log_source_destroy(int source)
{
	lock_guard<mutex> guard(log_sources.lock);

	/* Sanity check #1 - Valid source ID */
	if (unlikely(source >= log_sources.sources.size()))
		return DOCA_ERROR_INVALID_VALUE;

	/* Update the overall state */
	log_sources.num_freed++;

	/* There is no need to clear the name or buckets, they will be destroyed once log_sources is destroyed */

	return DOCA_SUCCESS;
}

doca_error_t
doca_log_rate_bucket_register(int source, int *bucket)
{
	if (unlikely(bucket == NULL))
		return DOCA_ERROR_INVALID_VALUE;
	*bucket = -1;

	lock_guard<mutex> guard(log_sources.lock);

	try {
		int id = static_cast<int>(log_sources.sources[source].buckets.size());

		log_sources.sources[source].buckets.emplace_back(std::make_unique<rate_bucket>());
		*bucket = id;

		return DOCA_SUCCESS;
	} catch (...) {
		return DOCA_ERROR_NO_MEMORY;
	}
}

doca_error_t
doca_log_create_file_backend(FILE *fptr, struct doca_logger_backend **backend)
{
	struct doca_logger_backend *logger;

	if (unlikely(backend == NULL))
		return DOCA_ERROR_INVALID_VALUE;
	*backend = NULL;

	logger = log_backend_allocate();
	if (logger == NULL)
		return DOCA_ERROR_NO_MEMORY;

	logger->level = DOCA_LOG_LEVEL_INFO;
	logger->type = LOG_BACKEND_FILE;

	logger->stream.fptr = fptr;

	*backend = logger;
	return DOCA_SUCCESS;
}

doca_error_t
doca_log_create_fd_backend(int fd, struct doca_logger_backend **backend)
{
#ifdef __linux__
	struct doca_logger_backend *logger;

	if (unlikely(backend == NULL))
		return DOCA_ERROR_INVALID_VALUE;
	*backend = NULL;

	logger = log_backend_allocate();
	if (logger == NULL)
		return DOCA_ERROR_NO_MEMORY;

	logger->level = DOCA_LOG_LEVEL_INFO;
	logger->type = LOG_BACKEND_FD;

	logger->stream.fd = fd;

	*backend = logger;
	return DOCA_SUCCESS;
#else /* __linux__ */
	return DOCA_ERROR_NOT_SUPPORTED;
#endif /* __linux__ */
}

doca_error_t
doca_log_create_buffer_backend(char *buffer, size_t capacity, log_flush_callback handler,
				struct doca_logger_backend **backend)
{
	struct doca_logger_backend *logger;

	if (unlikely(handler == NULL || backend == NULL))
		return DOCA_ERROR_INVALID_VALUE;
	*backend = NULL;

	logger = log_backend_allocate();
	if (logger == NULL)
		return DOCA_ERROR_NO_MEMORY;

	logger->level = DOCA_LOG_LEVEL_INFO;
	logger->type = LOG_BACKEND_BUFFER;

	logger->stream.buffer.buffer = buffer;
	logger->stream.buffer.capacity = capacity;
	logger->stream.buffer.flush_handler = handler;

	*backend = logger;
	return DOCA_SUCCESS;
}

doca_error_t
doca_log_create_syslog_backend(const char *name, struct doca_logger_backend **backend)
{
#ifdef __linux__
	struct doca_logger_backend *logger;

	if (unlikely(name == NULL || backend == NULL))
		return DOCA_ERROR_INVALID_VALUE;
	*backend = NULL;

	logger = log_backend_allocate();
	if (logger == NULL)
		return DOCA_ERROR_NO_MEMORY;

	logger->level = DOCA_LOG_LEVEL_INFO;
	logger->type = LOG_BACKEND_SYSLOG;
	os_abstraction::priv_doca_strlcpy(logger->stream.syslog_name, name, MAX_SYSLOG_NAME_LEN);

	*backend = logger;
	return DOCA_SUCCESS;
#else /* __linux__ */
	return DOCA_ERROR_NOT_SUPPORTED;
#endif /* __linux__ */
}

#ifdef __linux__
static inline int
get_syslog_log_level(uint32_t level)
{
	switch (level) {
	case DOCA_LOG_LEVEL_DEBUG:
		return LOG_DEBUG;
	case DOCA_LOG_LEVEL_INFO:
		return LOG_INFO;
	case DOCA_LOG_LEVEL_WARNING:
		return LOG_WARNING;
	case DOCA_LOG_LEVEL_ERROR:
		return LOG_ERR;
	case DOCA_LOG_LEVEL_CRIT:
		return LOG_CRIT;
	default:
		return LOG_ALERT;
	}
}
#endif /* __linux__ */

static doca_error_t
internal_log(uint32_t level, int source, int line, const char *format, va_list ap)
{
	va_list copy_ap;
	char time_string[64]{};
	uint32_t usec{0};
	struct tm time {};
	int i{0}, res{0};
	size_t num_bytes_remaining{0}, num_bytes_used{0};
	FILE *stream{nullptr};
	struct buffer_stream *buffer{nullptr};
	struct doca_logger_backend *logger{nullptr};
	const char *level_name { nullptr };

	if (unlikely(source < 0 || log_sources.sources.size() <= source))
		return DOCA_ERROR_INVALID_VALUE;

	/*
	 * It is valid for a user to use a log level out of bounds and it is easy to deduce the log level in such case
	 * However, there is no obvious deduction from an invalid value within bounds so the method treats such input
	 * as invalid.
	 */
	if (unlikely(level < DOCA_LOG_LEVEL_CRIT))
		level = DOCA_LOG_LEVEL_CRIT;
	else if (unlikely(level > DOCA_LOG_LEVEL_DEBUG))
		level = DOCA_LOG_LEVEL_DEBUG;

	level_name = get_log_level_name(level);
	if (level_name == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (likely(os_abstraction::priv_doca_localtime(time, usec) == DOCA_SUCCESS))
		strftime(time_string, sizeof(time_string), "%H:%M:%S", &time);
	else
		snprintf(time_string, sizeof(time_string), "%s", "Unavailable");

	for (i = 0; (logger = log_backend_get(i)) != NULL; i++) {
		if (logger->level < level)
			continue;

		switch (logger->type) {
		case LOG_BACKEND_FILE:
			stream = logger->stream.fptr != NULL ? logger->stream.fptr : stderr;
			fprintf(stream, DOCA_LOG_PREFIX_FORMAT, time_string, usec, level_name,
				log_sources.sources[source].name.c_str(), line);
			va_copy(copy_ap, ap);
			vfprintf(stream, format, copy_ap);
			va_end(copy_ap);

			fprintf(stream, "\n");
			fflush(stream);
			break;

		case LOG_BACKEND_FD: {
#ifdef __linux__
			int fd = logger->stream.fd;
			dprintf(fd, DOCA_LOG_PREFIX_FORMAT, time_string, usec, level_name,
				log_sources.sources[source].name.c_str(), line);
			va_copy(copy_ap, ap);
			vdprintf(fd, format, copy_ap);
			va_end(copy_ap);

			dprintf(fd, "\n");
			/* file descriptors don't need flushing */
#endif /* __linux__ */
			break;
		}

		case LOG_BACKEND_BUFFER: {
			/* The brackets above are required for the lock guard */
			lock_guard<mutex> guard(doca_loggers.backends.lock);
			buffer = &logger->stream.buffer;
			num_bytes_remaining = buffer->capacity;
			res = snprintf(buffer->buffer, num_bytes_remaining, DOCA_LOG_PREFIX_FORMAT, time_string, usec,
				       level_name, log_sources.sources[source].name.c_str(), line);

			if (unlikely(res < 0 || num_bytes_remaining <= res))
				continue;

			num_bytes_used = res;
			num_bytes_remaining -= res;

			va_copy(copy_ap, ap);
			res = vsnprintf(&buffer->buffer[num_bytes_used], num_bytes_remaining, format, copy_ap);
			va_end(copy_ap);

			if (unlikely(res < 0 || num_bytes_remaining <= res))
				continue;

			num_bytes_used += res;
			num_bytes_remaining -= res;

			res = snprintf(&buffer->buffer[num_bytes_used], num_bytes_remaining, "\n");

			if (unlikely(res < 0 || num_bytes_remaining <= res))
				continue;

			try {
				// Can't tell what does the flush handler do, therefore it has to be under try / catch
				buffer->flush_handler(buffer->buffer);
			}
			catch (...) {
				// There is nothing much that can be done here besides making sure that a flush
				// exception does not crash the log.
			}

			break;
		}
		case LOG_BACKEND_SYSLOG:
#ifdef __linux__
			openlog(logger->stream.syslog_name, LOG_PID | LOG_NDELAY, LOG_USER);
			va_copy(copy_ap, ap);
			vsyslog(get_syslog_log_level(level), format, copy_ap);
			va_end(copy_ap);
			closelog();
#endif /* __linux__ */
			break;
		}
	}

	return DOCA_SUCCESS;
}

doca_error_t
doca_log(uint32_t level, int source, int line, const char *format, ...)
{
	va_list ap;
	doca_error_t res;

	va_start(ap, format);
	res = internal_log(level, source, line, format, ap);
	va_end(ap);

	return res;
}

doca_error_t
doca_log_developer(uint32_t level, int source, int line, const char *format, ...)
{
	static mutex iolock;
	va_list ap;
	doca_error_t res;

	va_start(ap, format);
	lock_guard<mutex> guard(iolock);
	res = internal_log(level, source, line, format, ap);
	va_end(ap);

	return res;
}

static inline void
update_token(atomic_ullong *meta, uint32_t *printed, uint32_t *bucket_start)
{
	uint64_t token = meta->fetch_add(1);

	*printed = token & RATE_LIMIT_PRINTED_MASK;
	*bucket_start = token >> RATE_LIMIT_BUCKET_SHIFT;
}

doca_error_t
doca_log_rate_limit(uint32_t level, int source, int line, int bucket, const char *format, ...)
{
	va_list ap;
	doca_error_t res = DOCA_SUCCESS;

	if (unlikely(source < 0 || log_sources.sources.size() <= source))
		return DOCA_ERROR_INVALID_VALUE;

	if (unlikely(bucket < 0 || log_sources.sources[source].buckets.size() <= bucket))
		return DOCA_ERROR_INVALID_VALUE;

	// Assumption: only the 32 LSB of time is of interest.
	uint32_t now = static_cast<uint32_t>(time(NULL));
	auto rate_limit = log_sources.sources[source].buckets[bucket].get();

	auto meta = &rate_limit->meta;
	auto lock = &rate_limit->lock;
	uint32_t printed, bucket_start;

	update_token(meta, &printed, &bucket_start);

	if (now - bucket_start >= doca_log_get_bucket_time()) {
		for (;;) {
			// atomic flag returns the previous value so if it is false then it means that the test and set
			// locked the lock.
			if (!lock->test_and_set())
				break;
		}

		update_token(meta, &printed, &bucket_start);
		now = MAX(now, bucket_start);
		if (now - bucket_start >= doca_log_get_bucket_time()) {
			meta->store(((uint64_t)now) << RATE_LIMIT_BUCKET_SHIFT);
			printed = 0;
		}
		lock->clear();
	}

	if (printed < doca_log_get_quantity()) {
		va_start(ap, format);
		res = internal_log(level, source, line, format, ap);
		va_end(ap);
	}

	return res;
}
