/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef MLX5_DEVX_CMDS_H_
#define MLX5_DEVX_CMDS_H_

#include <infiniband/verbs.h>
#include <infiniband/mlx5dv.h>

#include <stdbool.h>
#include <stdint.h>

#include <doca_error.h>
#include <priv_doca_compat.h>

#include "mlx5_devx.h"

#ifdef __cplusplus
extern "C" {
#endif

#define MLX5_VUID_BYTES 128

struct mlx5_hca_cap {
	uint16_t vhca_id;
	bool vhca_resource_manager;
	bool hca_cap_2;
	bool hotplug_manager;
	bool emulation_manager_virtio_net;
	bool emulation_manager_virtio_blk;
	bool emulation_manager_nvme;
	bool emulation_manager_virtio_fs;
	bool eswitch_manager;
	uint64_t general_object_types_supported;
	bool ib_port_type;
	bool crossing_vhca_mkey_supported;
};

DOCA_PRIVATE
int mlx5_devx_cmd_query_hca_caps(struct ibv_context *ctx, bool other_function, uint16_t function_id, struct mlx5_hca_cap *caps);

struct mlx5_hca_cap_2 {
	uint32_t cross_vhca_object_to_object_supported;
	bool introspection_mkey_access_allowed;
	uint32_t introspection_mkey;
};

DOCA_PRIVATE
int mlx5_devx_cmd_query_hca_caps_2(struct ibv_context *ctx, bool other_function, uint16_t function_id, struct mlx5_hca_cap_2 *caps);

struct mlx5_vuid {
	uint8_t id[MLX5_VUID_BYTES];
};

DOCA_PRIVATE
doca_error_t mlx5_devx_cmd_query_vuid(struct ibv_context *ctx, uint16_t vhca_id, struct mlx5_vuid *vuids,
				      uint16_t *nb_vuids);


/** Notes:
 *  - Use enum ibv_access_flags for the access field.
 *  - When using IBV_ACCESS_ZERO_BASED, the memory buffer supplied in addr field must be align with memory page
 *    size.
*/
DOCA_PRIVATE
doca_error_t mlx5_devx_cmd_create_mkey(struct ibv_pd *pd, void *addr, size_t length, size_t page_size, uint32_t access,
				       struct mlx5dv_devx_umem **umem_ptr, struct mlx5dv_devx_obj **mkey_obj_ptr,
				       uint32_t *lkey);

DOCA_PRIVATE
doca_error_t mlx5_devx_cmd_destroy_mkey(struct mlx5dv_devx_umem *umem, struct mlx5dv_devx_obj *mkey_obj);


struct desc_data {
	void *addr;
	size_t len;
	uint32_t lkey;
	uint16_t vhca_id;
	char access_key[256];
	size_t access_key_sz;
};

DOCA_PRIVATE
doca_error_t mlx5_devx_cmd_cgmk_export(uint32_t lkey, struct ibv_context *ibv_ctx, void *addr, size_t len,
				       const char *access_key_in, size_t access_key_sz_in, struct desc_data *out_data);

DOCA_PRIVATE
doca_error_t mlx5_devx_cmd_cgmk_from_export_create(struct ibv_pd *pd, uint32_t lkey, uint16_t vhca_id,
						   const char *access_key_in, size_t access_key_sz_in,
						   struct mlx5dv_devx_obj **mkey_obj_out, uint32_t *lkey_out);

DOCA_PRIVATE
doca_error_t mlx5_devx_cmd_cgmk_from_export_destroy(struct mlx5dv_devx_obj *mkey_obj);

struct emulated_fn_info {
	uint16_t vhca_id;
	uint16_t pci_bdf;
	uint16_t nb_total_vfs; /**< Set for PFs only.*/
	bool hotplug;          /**< Set for PFs only.*/
	bool vf_exist;         /**< Set for VFs only.*/
};

/**
 * @brief Returns a list of emulated functions, from PRM command QUERY_EMULATED_FUNCTIONS_INFO.
 *
 * @param [in] pf_vhca_id
 * Ignored in case op_mod is for PFs.
 * @param [out] fns
 * Holds space for the emulated functions. Allocated by the caller.
 * @param [in,out] nb_fns
 * Caller should provide how many elements are allocated in fns.
 * In case op_mod set for PFs then the actual number of PFs is written back by the function.
 * In case op_mod set for VFs then the number of existing VFs is written back instead.
*/
DOCA_PRIVATE
doca_error_t mlx5_devx_cmd_query_emulated_fns_info(struct ibv_context *ctx, enum mlx5_emulated_functions_op_mod op_mod,
					  uint16_t pf_vhca_id, struct emulated_fn_info *fns, uint16_t *nb_fns);

#ifdef __cplusplus
}
#endif

#endif /* MLX5_DEVX_CMDS_H_ */
