/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <infiniband/mlx5dv.h>
#include <stdbool.h>
#include <stdlib.h>

#include <mlx5_prm.h>
#include <doca_log.h>

#include <mlx5_devx_cmds.h>

#define MLX5_GENERAL_OBJ_TYPE_MKEY 0xff01
#define LOCAL_MKEY_TO_REMOTE_MKEY_BIT 0x100
#define ACCESS_KEY_ADDR_MAX_SIZE 32

DOCA_LOG_REGISTER(CORE::MLX5_DEVX_CMDS);

static inline uint64_t
u64log2(uint64_t x)
{
	if (x == 0) {
		/* log(0) is undefined */
		return 0;
	}
	return 63u - __builtin_clzll(x);
}

int
mlx5_devx_cmd_query_hca_caps(struct ibv_context *ctx, bool other_function, uint16_t function_id, struct mlx5_hca_cap *caps)
{
	uint32_t in[MLX5_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	void *out_caps;
	int status, rc;

	/* Set operation code and operation mode of the command. */
	MLX5_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	MLX5_SET(query_hca_cap_in, in, op_mod, MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE | MLX5_HCA_CAP_OPMOD_GET_CUR);
	MLX5_SET(query_hca_cap_in, in, other_function, other_function);
	MLX5_SET(query_hca_cap_in, in, function_id, function_id);

	/* Send the command to FW using DevX. */
	rc = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (rc) {
		DOCA_DLOG_ERR("Failed to query DevX HCA capabilities.");
		return rc;
	}

	/* Inspect the returned status of the command and print the syndrome in case of error. */
	status = MLX5_GET(query_hca_cap_out, out, status);
	if (status) {
		DOCA_DLOG_ERR("Query devx HCA capabilities returned with status 0x%x, syndrome = 0x%x",
			       status, MLX5_GET(query_hca_cap_out, out, syndrome));
		return status;
	}

	/* Extract the capability field from the command output. */
	out_caps = MLX5_ADDR_OF(query_hca_cap_out, out, capability);
	/* Translate the capabilities from PRM format to c structure. */
	*caps = (struct mlx5_hca_cap) {
		.vhca_id = MLX5_GET(cmd_hca_cap, out_caps, vhca_id),
		.hca_cap_2 = MLX5_GET(cmd_hca_cap, out_caps, hca_cap_2),
		.vhca_resource_manager = MLX5_GET(cmd_hca_cap, out_caps, vhca_resource_manager),
		.hotplug_manager = MLX5_GET(cmd_hca_cap, out_caps, hotplug_manager),
		.emulation_manager_virtio_net = MLX5_GET(cmd_hca_cap, out_caps, virtio_net_device_emulation_manager),
		.emulation_manager_virtio_blk = MLX5_GET(cmd_hca_cap, out_caps, virtio_blk_device_emulation_manager),
		.emulation_manager_nvme = MLX5_GET(cmd_hca_cap, out_caps, nvme_device_emulation_manager),
		.emulation_manager_virtio_fs = MLX5_GET(cmd_hca_cap, out_caps, virtio_fs_device_emulation_manager),
		.eswitch_manager = MLX5_GET(cmd_hca_cap, out_caps, eswitch_manager),
		.general_object_types_supported = MLX5_GET64(cmd_hca_cap, out_caps, general_obj_types_63_0),
		.ib_port_type = !(MLX5_GET(cmd_hca_cap, out_caps, port_type)),
		.crossing_vhca_mkey_supported = MLX5_GET(cmd_hca_cap, out_caps, crossing_vhca_mkey),
	};

	return 0;
}

int
mlx5_devx_cmd_query_hca_caps_2(struct ibv_context *ctx, bool other_function, uint16_t function_id, struct mlx5_hca_cap_2 *caps)
{
	uint32_t in[MLX5_ST_SZ_DW(query_hca_cap_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	void *out_caps;
	int status, rc;

	/* Set operation code and operation mode of the command. */
	MLX5_SET(query_hca_cap_in, in, opcode, MLX5_CMD_OP_QUERY_HCA_CAP);
	MLX5_SET(query_hca_cap_in, in, op_mod, MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE_CAP_2 | MLX5_HCA_CAP_OPMOD_GET_CUR);
	MLX5_SET(query_hca_cap_in, in, other_function, other_function);
	MLX5_SET(query_hca_cap_in, in, function_id, function_id);

	/* Send the command to FW using DevX. */
	rc = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, sizeof(out));
	if (rc) {
		DOCA_DLOG_ERR("Failed to query devx HCA capabilities.");
		return rc;
	}
	/* Inspect the returned status of the command and print the syndrome in case of error. */
	status = MLX5_GET(query_hca_cap_out, out, status);
	if (status) {
		DOCA_DLOG_ERR("Query devx HCA capabilities returned with status 0x%x, syndrome = 0x%x",
			       status, MLX5_GET(query_hca_cap_out, out, syndrome));
		return status;
	}
	/* Extract the capability field from the command output. */
	out_caps = MLX5_ADDR_OF(query_hca_cap_out, out, capability);
	/* Translate the capabilities from PRM format to c structure. */
	caps->cross_vhca_object_to_object_supported = MLX5_GET(cmd_hca_cap_2, out_caps, cross_vhca_object_to_object_supported);
	caps->introspection_mkey_access_allowed = MLX5_GET(cmd_hca_cap_2, out_caps, introspection_mkey_access_allowed);
	caps->introspection_mkey = MLX5_GET(cmd_hca_cap_2, out_caps, introspection_mkey);

	return 0;
}

doca_error_t
mlx5_devx_cmd_query_vuid(struct ibv_context *ctx, uint16_t vhca_id, struct mlx5_vuid *vuids, uint16_t *nb_vuids)
{
	uint32_t in[MLX5_ST_SZ_DW(query_vuid_in)] = {0};
	void *out;
	size_t out_len;
	struct mlx5_vuid *out_vuids;
	uint16_t nb_out_vuids, vuid_idx;
	int status, rc;
	doca_error_t err = DOCA_SUCCESS;

	out_len = MLX5_ST_SZ_BYTES(query_vuid_out) + (MLX5_ST_SZ_BYTES(vuid) * (*nb_vuids));
	out = calloc(1, out_len);
	if (out == NULL)
		return DOCA_ERROR_NO_MEMORY;

	MLX5_SET(query_vuid_in, in, opcode, MLX5_CMD_OP_QUERY_VUID);

	const bool query_vfs_vuid = (*nb_vuids > 1);

	MLX5_SET(query_vuid_in, in, query_vfs_vuid, query_vfs_vuid);

	MLX5_SET(query_vuid_in, in, vhca_id, vhca_id);

	/* Send the command to FW using DevX. */
	rc = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, out_len);
	if (rc) {
		DOCA_DLOG_ERR("Failed to query VUID, DevX general command failed. hca_id: %u, vuid_arr_addr: %p, vuid_arr_size: %u",
			      vhca_id, vuids, *nb_vuids);
		err = DOCA_ERROR_DRIVER;
		goto vuid_out_clean;
	}

	/* Inspect the returned status of the command and print the syndrome in case of error. */
	status = MLX5_GET(query_vuid_out, out, status);
	if (status) {
		DOCA_DLOG_ERR("Query DevX VUID returned with status 0x%x, syndrome = 0x%x",
			       status, MLX5_GET(query_hca_cap_out, out, syndrome));
		err = DOCA_ERROR_DRIVER;
		goto vuid_out_clean;
	}

	/* Fail in case user asked for less VUIDs than the number returned.*/
	nb_out_vuids = MLX5_GET(query_vuid_out, out, num_of_entries);
	if (nb_out_vuids > *nb_vuids) {
		DOCA_DLOG_ERR("Query VUID DevX command returned with num_of_entries: %u, while user asked for nb_vuids: %u",
			      nb_out_vuids, *nb_vuids);
		err = DOCA_ERROR_INVALID_VALUE;
		goto vuid_out_clean;
	}
	*nb_vuids = nb_out_vuids;

	/* Extract the vuid array from the command output, then copy them to output.*/
	out_vuids = (struct mlx5_vuid *)MLX5_ADDR_OF(query_vuid_out, out, vuids);
	for (vuid_idx = 0; vuid_idx < nb_out_vuids; vuid_idx++)
		vuids[vuid_idx] = out_vuids[vuid_idx];

vuid_out_clean:
	free(out);

	return err;
}

doca_error_t
mlx5_devx_cmd_create_mkey(struct ibv_pd *pd, void *addr, size_t length, size_t page_size, uint32_t access,
			  struct mlx5dv_devx_umem **umem_ptr, struct mlx5dv_devx_obj **mkey_obj_ptr, uint32_t *lkey)
{
	int ret;
	uint32_t in[MLX5_ST_SZ_DW(create_mkey_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(create_mkey_out)] = {0};
	void *mkc;
	const int variant_part_of_mkey = 0x42;
	const int log_of_4kb_page_size = u64log2(page_size);
	uintptr_t mkey_offset = (access & IBV_ACCESS_ZERO_BASED) ? 0 : (uintptr_t)addr;
	struct mlx5dv_devx_umem_in umem_in = {0};

	umem_in.addr = addr;
	umem_in.size = length;
	umem_in.access = access;
	umem_in.pgsz_bitmap = page_size;

	if (mkey_offset == 0) {
		if (((uintptr_t)addr % page_size) != 0) {
			DOCA_DLOG_ERR(
			     "creation of zero-base mkey allowed only when buffer address align with memory page size");
			return DOCA_ERROR_NOT_SUPPORTED;
		}
		umem_in.access &= ~IBV_ACCESS_ZERO_BASED;
	}
	/* Register umem buf */
	*umem_ptr = mlx5dv_devx_umem_reg_ex(pd->context, &umem_in);
	if (*umem_ptr == NULL) {
		DOCA_DLOG_ERR("Failed to register umem buffer.");
		return DOCA_ERROR_DRIVER;
	}

	/* Cast pd to devx object */
	struct mlx5dv_obj *pd_obj = &(struct mlx5dv_obj){
		.pd.in = pd,
		.pd.out = &(struct mlx5dv_pd){0},
	};
	ret = mlx5dv_init_obj(pd_obj, MLX5DV_OBJ_PD);
	if (ret) {
		DOCA_DLOG_ERR("Unable to expose pdn from pd.");
		goto create_mkey_err;
	}

	MLX5_SET(create_mkey_in, in, opcode, MLX5_CMD_OP_CREATE_MKEY);
	MLX5_SET(create_mkey_in, in, translations_octword_actual_size, 1);
	MLX5_SET(create_mkey_in, in, pg_access, 1);
	MLX5_SET(create_mkey_in, in, mkey_umem_id, (*umem_ptr)->umem_id);
	mkc = MLX5_ADDR_OF(create_mkey_in, in, memory_key_mkey_entry);
	MLX5_SET(mkc, mkc, lr, 0x1);
	if (access & IBV_ACCESS_LOCAL_WRITE)
		MLX5_SET(mkc, mkc, lw, 0x1);
	if (access & IBV_ACCESS_REMOTE_WRITE)
		MLX5_SET(mkc, mkc, rw, 0x1);
	if (access & IBV_ACCESS_REMOTE_READ)
		MLX5_SET(mkc, mkc, rr, 0x1);
	if (access & IBV_ACCESS_REMOTE_ATOMIC)
		MLX5_SET(mkc, mkc, a, 0x1);
	MLX5_SET(mkc, mkc, mkey_7_0, variant_part_of_mkey);
	MLX5_SET(mkc, mkc, qpn, 0xffffff);
	MLX5_SET(mkc, mkc, pd, pd_obj->pd.out->pdn);
	MLX5_SET(mkc, mkc, access_mode_1_0, MLX5_MKC_ACCESS_MODE_MTT);
	MLX5_SET(mkc, mkc, translations_octword_size_crossing_target_mkey, 1);
	MLX5_SET(mkc, mkc, log_entity_size, log_of_4kb_page_size);
	MLX5_SET64(mkc, mkc, start_addr, mkey_offset);
	MLX5_SET64(mkc, mkc, len, length);

	*mkey_obj_ptr = mlx5dv_devx_obj_create(pd->context, in, sizeof(in), out, sizeof(out));
	if (*mkey_obj_ptr == NULL) {
		DOCA_DLOG_ERR("Failed to create devx mkey object. errno = %d", errno);
		goto create_mkey_err;
	}

	/* extend mkey index to 32 bits */
	*lkey = MLX5_GET(create_mkey_out, out, mkey_index) << 8 | variant_part_of_mkey;

	return DOCA_SUCCESS;

create_mkey_err:
	mlx5dv_devx_umem_dereg(*umem_ptr);
	return DOCA_ERROR_DRIVER;
}

doca_error_t
mlx5_devx_cmd_destroy_mkey(struct mlx5dv_devx_umem *umem, struct mlx5dv_devx_obj *mkey_obj)
{
	if (mlx5dv_devx_obj_destroy(mkey_obj)) {
		DOCA_DLOG_ERR("Failed to destroy devx mkey object.");
		return DOCA_ERROR_DRIVER;
	}
	if (mlx5dv_devx_umem_dereg(umem)) {
		DOCA_DLOG_ERR("Failed to deregister umem buffer.");
		return DOCA_ERROR_DRIVER;
	}
	return DOCA_SUCCESS;
}

doca_error_t
mlx5_devx_cmd_cgmk_export(uint32_t lkey, struct ibv_context *ibv_ctx, void *addr, size_t len,
			  const char *access_key_in, size_t access_key_sz_in, struct desc_data *out_data)
{
	int ret;
	struct mlx5_hca_cap caps;
	struct mlx5_hca_cap_2 caps_2;
	uint32_t in[MLX5_ST_SZ_DW(allow_other_vhca_access_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(allow_other_vhca_access_out)] = {0};
	void *access_key;

	out_data->lkey = lkey;
	out_data->access_key_sz = access_key_sz_in;
	memcpy(out_data->access_key, access_key_in, access_key_sz_in);

	/* Check that HCA cross_vhca_object_to_object is supported */
	ret = mlx5_devx_cmd_query_hca_caps(ibv_ctx, 0 /*other_function*/, 0 /*function_id*/, &caps);
	if (ret) {
		DOCA_DLOG_ERR("Failed to query HCA Caps.");
		return DOCA_ERROR_DRIVER;
	}
	if (!caps.hca_cap_2) {
		DOCA_DLOG_ERR("HCA Caps 2 disabled.");
		return DOCA_ERROR_NOT_SUPPORTED;
	}
	ret = mlx5_devx_cmd_query_hca_caps_2(ibv_ctx, 0 /*other_function*/, 0 /*function_id*/, &caps_2);
	if (ret) {
		DOCA_DLOG_ERR("Failed to query HCA Caps 2.");
		return DOCA_ERROR_DRIVER;
	}
	if (!(caps_2.cross_vhca_object_to_object_supported & LOCAL_MKEY_TO_REMOTE_MKEY_BIT)) {
		DOCA_DLOG_ERR("Cross GVMI MKEY not enabled. Cross_vhca_object_to_object_supported: 0x%x",
			      caps_2.cross_vhca_object_to_object_supported);
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	/* Allow crossing mkey */
	MLX5_SET(allow_other_vhca_access_in, in, opcode, MLX5_CMD_OP_ALLOW_OTHER_VHCA_ACCESS);
	MLX5_SET(allow_other_vhca_access_in, in, object_type_to_be_accessed, MLX5_GENERAL_OBJ_TYPE_MKEY);
	MLX5_SET(allow_other_vhca_access_in, in, object_id_to_be_accessed, out_data->lkey >> 8);
	access_key = MLX5_ADDR_OF(allow_other_vhca_access_in, in, access_key);
	memcpy(access_key, out_data->access_key, out_data->access_key_sz);

	ret = mlx5dv_devx_general_cmd(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (ret) {
		DOCA_DLOG_ERR("Failed to allow other vhca access.");
		return DOCA_ERROR_DRIVER;
	}
	out_data->addr = addr;
	out_data->len = len;

	return DOCA_SUCCESS;
}

doca_error_t
mlx5_devx_cmd_cgmk_from_export_create(struct ibv_pd *pd, uint32_t lkey, uint16_t vhca_id,
				      const char *access_key_in, size_t access_key_sz_in,
				      struct mlx5dv_devx_obj **mkey_obj_out, uint32_t *lkey_out)
{
	int ret;
	struct mlx5dv_pd pd_info;
	struct mlx5dv_obj pd_obj;
	struct mlx5dv_devx_obj *alias;
	uint32_t in[MLX5_ST_SZ_DW(create_alias_obj_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(create_alias_obj_out)] = {0};
	void *hdr, *alias_ctx, *access_key_addr, *metadata;
	uint32_t pd_num;

	if (access_key_sz_in > ACCESS_KEY_ADDR_MAX_SIZE) {
		DOCA_DLOG_ERR("Access key size exceeds the limit.");
		return DOCA_ERROR_INVALID_VALUE;
	}

	pd_obj.pd.in = pd;
	pd_obj.pd.out = &pd_info;
	ret = mlx5dv_init_obj(&pd_obj, MLX5DV_OBJ_PD);
	if (ret) {
		DOCA_DLOG_ERR("Unable to expose pdn from pd.");
		return DOCA_ERROR_DRIVER;
	}
	pd_num = htobe32(pd_obj.pd.out->pdn);

	/* Create alias */
	hdr = MLX5_ADDR_OF(create_alias_obj_in, in, hdr);
	alias_ctx = MLX5_ADDR_OF(create_alias_obj_in, in, alias_ctx);

	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJ_TYPE_MKEY);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, alias_object, 1);
	MLX5_SET(alias_context, alias_ctx, vhca_id_to_be_accessed, vhca_id);
	MLX5_SET(alias_context, alias_ctx, object_id_to_be_accessed, lkey >> 8);
	access_key_addr = MLX5_ADDR_OF(alias_context, alias_ctx, access_key);
	memcpy(access_key_addr, access_key_in, access_key_sz_in);
	metadata = MLX5_ADDR_OF(alias_context, alias_ctx, metadata);
	memcpy(metadata, &pd_num, sizeof(pd_num));

	alias = mlx5dv_devx_obj_create(pd->context, in, sizeof(in), out, sizeof(out));
	if (!alias) {
		DOCA_DLOG_ERR("Failed to create MKEY Alias Object.");
		return DOCA_ERROR_DRIVER;
	}
	ret = MLX5_GET(create_alias_obj_out, out, alias_ctx.status);
	if (ret) {
		DOCA_DLOG_ERR("Created mr alias object in bad state.");
		mlx5dv_devx_obj_destroy(alias);
		return DOCA_ERROR_DRIVER;
	}

	*lkey_out = MLX5_GET(create_alias_obj_out, out, hdr.obj_id) << 8;
	*mkey_obj_out = alias;

	return DOCA_SUCCESS;
}

doca_error_t
mlx5_devx_cmd_cgmk_from_export_destroy(struct mlx5dv_devx_obj *mkey_obj)
{
	if (mlx5dv_devx_obj_destroy(mkey_obj)) {
		DOCA_DLOG_ERR("Failed to destroy devx mkey object.");
		return DOCA_ERROR_DRIVER;
	}
	return DOCA_SUCCESS;
}

doca_error_t
mlx5_devx_cmd_query_emulated_fns_info(struct ibv_context *ctx, enum mlx5_emulated_functions_op_mod op_mod,
				      uint16_t pf_vhca_id, struct emulated_fn_info *fns, uint16_t *nb_fns)
{
	uint32_t in[MLX5_ST_SZ_DW(query_emulated_functions_info_in)] = {0};
	uint32_t *out;
	uint8_t *emulated_fn_info_out;
	uint8_t status;
	uint16_t fn_idx, nb_emulated_fns, nb_vf_exist;
	size_t outlen;
	doca_error_t err = DOCA_SUCCESS;
	int ret = 0;

	outlen = MLX5_ST_SZ_BYTES(query_emulated_functions_info_out) +
		 MLX5_ST_SZ_BYTES(emulated_function_info) * (*nb_fns);
	out = calloc(1, outlen);
	if (!out) {
		DOCA_DLOG_ERR("Failed to allocate enough memory for the output. nb_fns: %u, output size: %zu", *nb_fns,
			      outlen);
		return DOCA_ERROR_NO_MEMORY;
	}

	MLX5_SET(query_emulated_functions_info_in, in, opcode,
		 MLX5_CMD_OP_QUERY_EMULATED_FUNCTIONS_INFO);
	MLX5_SET(query_emulated_functions_info_in, in, op_mod, op_mod);
	MLX5_SET(query_emulated_functions_info_in, in, pf_vhca_id, pf_vhca_id);

	ret = mlx5dv_devx_general_cmd(ctx, in, sizeof(in), out, outlen);
	if (ret) {
		DOCA_DLOG_ERR("Failed to query DevX emulated functions information. op_mod: %d, pf_vhca_id:%u, nb_fns:%u",
			      op_mod, pf_vhca_id, *nb_fns);
		err = DOCA_ERROR_NOT_SUPPORTED;
		*nb_fns = 0;
		goto free_out;
	}

	status = MLX5_GET(query_emulated_functions_info_out, out, status);
	if (status) {
		DOCA_DLOG_ERR("Query DevX emulated functions information returned with status %x, syndrome = %x",
			       status, MLX5_GET(query_emulated_functions_info_out, out, syndrome));
		err = DOCA_ERROR_DRIVER;
		goto free_out;
	}

	/*
	 * For PF: this value returns the total number of plugged PCI PFs.
	 * For VFs: this value return the Total VFs (pci spec) number of a PF.
	 */
	nb_emulated_fns = MLX5_GET(query_emulated_functions_info_out, out, num_emulated_functions);
	if (nb_emulated_fns > *nb_fns) {
		DOCA_DLOG_ERR("Query DevX emulated funtions found more devices than requested. found: %u, expected: %u",
			      nb_emulated_fns, *nb_fns);
		err = DOCA_ERROR_NO_MEMORY;
		goto free_out;
	}

	nb_vf_exist = 0;
	for (fn_idx = 0; fn_idx < nb_emulated_fns; fn_idx++) {
		emulated_fn_info_out = MLX5_ADDR_OF(query_emulated_functions_info_out, out,
						    emulated_function_info[fn_idx]);

		fns[fn_idx] = (struct emulated_fn_info) {
			.vhca_id = MLX5_GET(emulated_function_info, emulated_fn_info_out, vhca_id),
			.pci_bdf = MLX5_GET(emulated_function_info, emulated_fn_info_out, pci_bdf),
			.nb_total_vfs = 0,
			.hotplug = MLX5_GET(emulated_function_info, emulated_fn_info_out, hotplug_function),
			.vf_exist = MLX5_GET(emulated_function_info, emulated_fn_info_out, vf_exist),
		};

		if (MLX5_GET(emulated_function_info, emulated_fn_info_out, max_num_vfs_valid))
			fns[fn_idx].nb_total_vfs = MLX5_GET(emulated_function_info, emulated_fn_info_out, max_num_vfs);

		if (fns[fn_idx].vf_exist)
			nb_vf_exist++;
	}

	*nb_fns = (op_mod == MLX5_SET_EMULATED_FUNCTIONS_OP_MOD_VIRTUAL_FUNCTIONS) ? nb_vf_exist : nb_emulated_fns;

free_out:
	free(out);
	return err;
}
