/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * @file typed_generic_pool.hpp
 * @page generic pool
 * Generic Pool library. A generic memory pool of fixed-size elements.
 * The pool allocates memory according to the number of elements and the element size parameters
 * given by the user.
 * The user can ask for an element using the function generic_pool_get
 * and can return an element to the pool using the function generic_pool_put.
 *
 * Note: The pool is not thread safe.
 *
 * @{
 */

#pragma once

#include <stddef.h>
#include <vector>
#include <stdexcept>
#include <functional>

#include <priv_doca_compat.h>
#include <doca_error.h>

/**
 * @brief This class implements parametrized constant size memory pool
 * note: this class doesn't support multi-threading, it's up to caller to synchronize access to the pool.
 */
template <typename T>
class typed_generic_pool
{
public:
	/**
	 * constructor
	 * @param [in] num elements
	 */
	explicit typed_generic_pool(size_t num_elements) throw()
	{
		m_elements.resize(num_elements);
		m_elements_ptrs.resize(num_elements);
		for (size_t i = 0; i < num_elements; i++)
			m_elements_ptrs[i] = &(m_elements[i]);
	}

	/**
	 * destructor
	 */
	~typed_generic_pool() noexcept = default;

	/**
	 * @brief This method returns an element from the pool
	 * @return element or null if pool is empty.
	 */
	inline T*
	acquire() noexcept
	{
		/* Using try to because m_curr is incremented inside the at() access  */
		try {
			// No reason to set m_elements_ptrs.at(m_curr++) to null because it will not be used until
			// release is called for this index (and it will be overwritten).
			return m_elements_ptrs.at(m_curr++);
		}
		catch (...) {
			/* Catching all exceptions to assure that the pool doesn't throw any exception. */
			m_curr = m_elements_ptrs.size();
			return nullptr;
		}
	}

	/**
	 * @brief This method returns an element to the pool
	 * @param [in] element. Do not release the same element twice, the pool will break, this is undefined behavior.
	 */
	inline void
	release(T *element)
	{
		/* Using try to because m_curr is decremented inside the at() access */
		try {
			m_elements_ptrs.at(--m_curr) = element;
		}
		catch (...) {
			/* Catching all exceptions to assure that the pool doesn't throw any exception. */
			m_curr = 0;
		}
	}

	/**
	 * This method returns the number of free elements in the pool
	 * @return uint32_t
	 */
	inline uint32_t
	num_free_elements() const
	{
		return static_cast<uint32_t>(m_elements_ptrs.size() - m_curr);
	}

	/**
	 * This method returns the pool size
	 * @return uint32_t
	 */
	inline uint32_t
	pool_size() const
	{
		return static_cast<uint32_t>(m_elements_ptrs.size());
	}

	/**
	 * This method returns the number of allocated elements in the pool
	 * @return uint32_t
	 */
	inline uint32_t num_allocated_elements() const { return static_cast<uint32_t>(m_curr); }

	/**
	 * @brief This method activates a function on all elements
	 * @param [in] map_fn Method to apply on each element, returns doca_error_t
	 * @return doca_error_t First non-success error
	 */
	inline doca_error_t
	iterate_over_elements(std::function<doca_error_t(T&)> map_fn) noexcept
	{
		for (auto &element : m_elements) {
			auto ret = map_fn(element);
			if (ret != DOCA_SUCCESS)
				return ret;
		}
		return DOCA_SUCCESS;
	}

	typed_generic_pool(typed_generic_pool const &) = delete;
	typed_generic_pool(typed_generic_pool &&) noexcept = delete;
	typed_generic_pool &operator=(typed_generic_pool const &) = delete;
	typed_generic_pool &operator=(typed_generic_pool &&) noexcept = delete;

private:

	/** This vector holds the actual elements. It remains the same throughout the life of the pool */
	std::vector<T> m_elements;
	/** This vector holds references to the elements. It is modified every time that release is called */
	std::vector<T *> m_elements_ptrs;
	size_t m_curr{0};
};

/** @} */
