/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#include <vector>
#include <errno.h>
#include <assert.h>
#include <stdexcept>

#include <doca_log.h>
#include <private/common_defines.h>
#include <generic_pool.h>

DOCA_LOG_REGISTER(CORE::GENERIC_POOL);

/**
 * This class implements the C API pool.
 * Can't use the typed_generic_pool because of the nature of the API.
 */
struct generic_pool
{
public:
	/**
	 * Constructor
	 * @param [in] num elements in the pool
	 * @param [in] element size
	 */
	generic_pool(size_t num_elements, size_t element_size) throw()
		: m_num_elements(num_elements)
	{
		/* new will throw an exception if fails. */
		m_memzone.resize(num_elements * element_size);
		m_elements_ptrs.resize(num_elements);

		/* Initializing the pointers */
		uint8_t *element{m_memzone.data()};
		for (auto i = 0; i < num_elements; i++) {
			m_elements_ptrs[i] = element;
			element += element_size;
		}
	}

	/**
	 * Destructor
	 */
	~generic_pool() = default;

	/**
	 * @brief This method returns an element from the pool
	 * @return element or null if pool is empty.
	 */
	void *get_element()
	{
		/* Using try to because m_curr is incremented inside the at() access */
		try {
			// No reason to set m_elements_ptrs.at(m_curr++) to null because it will not be used until
			// release is called for this index (and it will be overwritten).
			return m_elements_ptrs.at(m_curr++);
		}
		catch (...) {
			/* Catching all exceptions to assure that the pool doesn't throw any exception. */
			m_curr = m_num_elements;
			DOCA_DLOG_ERR("Pool is empty");
			return nullptr;
		}
	}

	/**
	 * @brief This method returns an element to the pool
	 * @param [in] element. Do not return the same element twice, the pool will break, this is undefined behavior.
	 */
	void return_element(void *element)
	{
		/* Using try to because m_curr is decremented inside the at() access */
		try {
			m_elements_ptrs.at(--m_curr) = element;
		}
		catch (...) {
			/* Catching all exceptions to assure that the pool doesn't throw any exception. */
			m_curr = 0;
		}
	}

	generic_pool(generic_pool const &) = delete;
	generic_pool(generic_pool &&) = delete;
	generic_pool &operator=(generic_pool const &) = delete;
	generic_pool &operator=(generic_pool &&) = delete;

private:
	std::vector<void *> m_elements_ptrs;
	std::vector<uint8_t> m_memzone;
	size_t m_curr{0};
	const size_t m_num_elements;
};

struct generic_pool *
generic_pool_create(size_t num_elements, size_t element_size)
{
	if (num_elements == 0 || element_size == 0) {
		DOCA_DLOG_ERR("Received invalid input num_elements %zu or element_size %zu",
			       num_elements, element_size);
		return nullptr;
	}
	try {
		/* new failure or ctor failure will throw, therefore the is no need to check for null. */
		return new generic_pool(num_elements, element_size);
	} catch (...) {
		DOCA_DLOG_ERR("Failed to allocate doca generic pool");
		return nullptr;
	}
}

void
generic_pool_destroy(struct generic_pool *pool)
{
	assert(pool != NULL);

	delete pool;
}

void *
generic_pool_get(struct generic_pool *pool)
{
	assert(pool != NULL);
	return pool->get_element();
}

void
generic_pool_put(struct generic_pool *pool, void *element)
{
	assert(pool != NULL && element != NULL);
	pool->return_element(element);
}
