/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

/**
 * ID allocator used for allocating ids in a range [1, max_id], 0 is considered invalid ID.
 * These IDs can be used as indices to some array later on, providing fast lookup.
 * The get and put operations are considered thread safe.
 */

#ifndef ID_ALLOCATOR_H_
#define ID_ALLOCATOR_H_

#include <stdint.h>

#include <priv_doca_compat.h>

/**
 * Use this macro instead of id_allocator_create and id_allocator_destroy in case you want to allocate a single static
 *  instance for the file. This Macro can only be specified once per file.
 * If using this macro, a static variable with name 'var_name' is created and can be used with put/get methods.
 * Example file src.c:
 * ID_ALLOCATOR_REGISTER(my_id_aloc, 128);
 *
 * int main(){
 *  return id_allocator_get(my_id_aloc);
 * }
 */

#ifdef __cplusplus

struct id_allocator;

#define ID_ALLOCATOR_REGISTER(_var_name_, _max_id_) \
namespace { \
	static struct id_allocator *_var_name_{nullptr}; \
	struct id_allocator_ctor_dtor { \
		id_allocator_ctor_dtor(uint32_t max_id) { _var_name_ = id_allocator_create((max_id)); } \
		~id_allocator_ctor_dtor() \
		{ \
			id_allocator_destroy(_var_name_); \
			_var_name_ = nullptr; \
		} \
	}; \
	static id_allocator_ctor_dtor g_id_allocator_ctor_dtor(_max_id_); \
}

extern "C" {

#else

#define ID_ALLOCATOR_REGISTER(var_name, max_id) \
	static struct id_allocator *var_name; \
	static void __attribute__((constructor, used)) __##id_allocator_constructor(void) \
	{ \
		var_name = id_allocator_create((max_id)); \
	} \
	static void __attribute__((destructor, used)) __##id_allocator_destructor(void) \
	{ \
		id_allocator_destroy(var_name); \
	}

#endif /* __cplusplus */

#define ID_ALLOCATOR_INVALID_ID 0ul

/* Creates an ID allocator that allocates IDs in the range [1, max_id]. */
DOCA_PRIVATE
struct id_allocator *id_allocator_create(uint32_t max_id);

DOCA_PRIVATE
void id_allocator_destroy(struct id_allocator *id_aloc);

/* Returns the smallest free ID in the range [1, max_id]. */
DOCA_PRIVATE
uint32_t id_allocator_get(struct id_allocator *id_aloc);

/* Frees a previously allocated ID, the next may reuse the same ID. */
DOCA_PRIVATE
void id_allocator_put(struct id_allocator *id_aloc, uint32_t id);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* ID_ALLOCATOR_H_ */
