/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string>
#include <vector>
#include <mutex>

#include <private/common_defines.h>
#include <priv_doca_os_abstraction.hpp>
#include <id_allocator.h>
#include <doca_log.h>

DOCA_LOG_REGISTER(CORE::ID_ALLOCATOR);

struct id_allocator {
	std::vector<uint64_t> id_masks;
	uint32_t nb_free_ids{ 0 };
	os_abstraction::doca_spin_lock lock;
};

struct id_allocator *
id_allocator_create(uint32_t max_id)
{
	struct id_allocator *id_aloc{nullptr};
	size_t nb_masks{0};

	try {
		id_aloc = new struct id_allocator();
	} catch (...) {
		/* Catching all exceptions to assure that the pool doesn't throw any exception. */
		DOCA_DLOG_ERR("Failed to allocate id allocator");
		return nullptr;
	}

	nb_masks = max_id / 64;
	if (max_id % 64 != 0)
		++nb_masks;

	try {
		id_aloc->id_masks.resize(nb_masks);
	} catch (...) {
		// Can't guarantee that vector::resize will throw bad alloc so catching all exceptions here.
		DOCA_DLOG_ERR("Failed to allocate ID allocator masks vector with max_id %d", max_id);

		delete id_aloc;
		return nullptr;
	}

	for (auto &id_mask : id_aloc->id_masks) {
		id_mask = 0xFFFFFFFFFFFFFFFF; /* All IDs are available */
	}

	id_aloc->nb_free_ids = max_id;

	return id_aloc;
}

void
id_allocator_destroy(struct id_allocator *id_aloc)
{
	delete id_aloc;
}

uint32_t
id_allocator_get(struct id_allocator *id_aloc)
{
	size_t mask_idx;
	int free_id_idx;

	std::lock_guard<os_abstraction::doca_spin_lock> guard(id_aloc->lock);
	if (id_aloc->nb_free_ids == 0) {
		return ID_ALLOCATOR_INVALID_ID;
	}

	/* Find the first set bit. */
	/* Assumption: The condition above guarantees that one of the id_masks will contain a free id. */
	mask_idx = 0;
	while (id_aloc->id_masks[mask_idx] == 0)
		++mask_idx;

	free_id_idx = os_abstraction::priv_doca_bit_scan_forward(id_aloc->id_masks[mask_idx]);

	/* Unset the found bit. */
	id_aloc->id_masks[mask_idx] ^= (1ull << (free_id_idx - 1));
	id_aloc->nb_free_ids--;

	return static_cast<uint32_t>(mask_idx*64 + free_id_idx);
}

void
id_allocator_put(struct id_allocator *id_aloc, uint32_t id)
{
	const size_t mask_idx = (id - 1) / 64;
	const int id_idx = (id - 1) % 64;

	std::lock_guard<os_abstraction::doca_spin_lock> guard(id_aloc->lock);
	/* Set the bit to mark it as free. */
	id_aloc->id_masks[mask_idx] |= (1ull << id_idx);
	id_aloc->nb_free_ids++;
}
