/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#pragma once

#include <stdint.h>
#include <stdexcept>

#include <doca_error.h>
#include <private/common_defines.h>
#include <priv_doca_compat.h>

#include "priv_doca_os_abstraction_memory.h"

namespace os_abstraction
{

/**
 * @brief This class implements aligned memory
 */
template<typename T = uint8_t>
class aligned_buffer
{
public:
	/**
	 * Constructor
	 * @param [in] alignment
	 * @param [in] size in T
	 */
	aligned_buffer(size_t alignment, size_t size) : m_size(size)
	{
		m_data = reinterpret_cast<T *>(priv_doca_allocate_aligned_buffer(alignment, size * sizeof(T)));
		if (m_data == nullptr)
			throw std::runtime_error("Failed to allocate aligned buffer");

		m_end = begin() + size;
	}

	/**
	 * Destructor
	 */
	virtual ~aligned_buffer()
	{
		priv_doca_free_aligned_buffer(m_data);
	}

	/**
	 * @brief This method returns the buffer size
	 * @return size_t
	 */
	inline size_t size() const noexcept
	{
		return m_size;
	}

	/**
	 * @brief Operator []
	 * The method does not check boundaries
	 * @param [in] index
	 * @return reference to node at the index
	 */
	inline T &operator[](size_t index) noexcept
	{
		return reinterpret_cast<T *>(m_data)[index];
	}

	/**
	 * @brief This method returns the buffer
	 * @return void *
	 */
	inline T *data() const noexcept
	{
		return reinterpret_cast<T *>(m_data);
	}

	/**
	 * @brief This method returns the beginning of the buffer
	 * @return uint8_t * that points to the head of the buffer
	 */
	inline T *begin() const noexcept
	{
		return reinterpret_cast<T *>(m_data);
	}

	/**
	 * @brief This method returns the end of the buffer
	 * @return uint8_t * that points to the end of the buffer
	 */
	inline T *end() const noexcept
	{
		return m_end;
	}

private:

	T *m_data{nullptr};
	T *m_end{nullptr};
	const size_t m_size;

	aligned_buffer(aligned_buffer const &) = delete;
	aligned_buffer &operator=(aligned_buffer const &) = delete;
	aligned_buffer(aligned_buffer &&) = delete;
	aligned_buffer &operator=(aligned_buffer &&) = delete;
};

} // namespace os_abstraction
