/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <doca_log.h>

#include "engine_layer.h"
#include "doca_flow_layer.h"
#include "dpdk_layer.h"

#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "engine_pipe.h"
#include "flow_age.h"
#include "dpdk_pipe_actions.h"

#define DF_MODE_ARG_MAX_LEN (256)
#define ISOLATED_MODE "isolated"
#define MAX_BUFF_SIZE (256)
#define MAX_LCORE (256)

DOCA_LOG_REGISTER(doca_flow);

static enum engine_shared_resource_type shared_resource_type_map[DOCA_FLOW_SHARED_RESOURCE_MAX];
static doca_flow_shared_resource_unbind_cb shared_resources_unbind_cb;
static char tmp_error_buffer[MAX_LCORE][MAX_BUFF_SIZE];

static void
set_error(struct doca_flow_error *error, enum doca_flow_error_type type,
	  const char *format, ...)
{
	va_list args;
	uint16_t lcore_id;

	if (unlikely(error == NULL))
		return;

	error->type = type;

	lcore_id = rte_lcore_id();
	if (unlikely(lcore_id >= MAX_LCORE)) {
		error->message = "invalid lcore buffer";
		return;
	}

	error->message = tmp_error_buffer[lcore_id];
	va_start(args, format);
	vsnprintf(tmp_error_buffer[lcore_id], MAX_BUFF_SIZE, format, args);
	va_end(args);
}

uint8_t*
doca_flow_port_priv_data(struct doca_flow_port *p)
{
	return &p->user_data[0];
}

static int
set_engine_mode(struct engine_model_cfg *model_cfg, enum engine_model_mode_type mode)
{
	if (model_cfg->mode != ENGINE_MODEL_MODE_MAX) {
		model_cfg->mode = ENGINE_MODEL_MODE_MAX;
		return -EINVAL;
	}
	model_cfg->mode = mode;
	return 0;
}

static void
model_parse_extend_param(const char *args, struct engine_model_cfg *model_cfg)
{
	struct mode_param *mode_param = &model_cfg->mode_param;

	if (model_cfg->mode == ENGINE_MODEL_MODE_HYBRID
		&& !strncmp(args, HYBRID_SLAVE_LAG, strlen(HYBRID_SLAVE_LAG)))
		mode_param->hybrid.secondary_lag_id = (uint16_t)atoi(args + strlen(HYBRID_SLAVE_LAG));
}

static int
model_parse_param(const char *mode_args, struct engine_model_cfg *model_cfg)
{
	char mode_str[DF_MODE_ARG_MAX_LEN] = {0};
	int len, ret = 0;
	char *str;

	model_cfg->is_verification_enabled = false;
	model_cfg->mode = ENGINE_MODEL_MODE_MAX;
	model_cfg->isolated = ENGINE_MODEL_NON_ISOLATED;

	if (mode_args == NULL) {
		model_cfg->isolated = ENGINE_MODEL_ISOLATED_UNDEF;
		return -EINVAL;
	}

	len = strlen(mode_args);
	if (len >= DF_MODE_ARG_MAX_LEN - 1) {
		DOCA_LOG_ERR("Unexpected length of mode argments %d, max=%d", len,
			     DF_MODE_ARG_MAX_LEN);
		model_cfg->isolated = ENGINE_MODEL_ISOLATED_UNDEF;
		return -EINVAL;
	}

	memcpy(mode_str, mode_args, len);
	str = strtok(mode_str, ",");
	while ((str != NULL) & !ret) {
		if (!strncmp(str, ENGINE_MODE_STR_VNF, strlen(ENGINE_MODE_STR_VNF))) {
			ret = set_engine_mode(model_cfg, ENGINE_MODEL_MODE_VNF);
		} else if (!strncmp(str, ENGINE_MODE_STR_SWITCH, strlen(ENGINE_MODE_STR_SWITCH))) {
			ret = set_engine_mode(model_cfg, ENGINE_MODEL_MODE_SWITCH);
		} else if (!strncmp(str, ENGINE_MODE_STR_HYBRID, strlen(ENGINE_MODE_STR_HYBRID))) {
			ret = set_engine_mode(model_cfg, ENGINE_MODEL_MODE_HYBRID);
		} else if (!strncmp(str, ENGINE_MODE_STR_REMOTE_VNF,
				    strlen(ENGINE_MODE_STR_REMOTE_VNF))) {
			ret = set_engine_mode(model_cfg, ENGINE_MODEL_MODE_REMOTE_VNF);
		} else if (!strncmp(str, ENGINE_FLAG_STR_VERIFY,
				    strlen(ENGINE_FLAG_STR_VERIFY))) {
			model_cfg->is_verification_enabled = true;
		} else if (!strncmp(str, ENGINE_ISOLATED, strlen(ENGINE_ISOLATED)))
			model_cfg->isolated = ENGINE_MODEL_ISOLATED;
		else
			model_parse_extend_param(str, model_cfg);
		str = strtok(NULL, ",");
	}
	return ret;
}

static int
model_cfg_build(const struct doca_flow_cfg *cfg, struct engine_model_cfg *model_cfg)
{
	model_cfg->nr_pipe_queues = cfg->queues;
	model_cfg->pipe_queue_depth = cfg->queue_depth;
	return model_parse_param(cfg->mode_args, model_cfg);
}

static void
shared_resource_type_map_init()
{
	int res_idx;

	for (res_idx = 0; res_idx < DOCA_FLOW_SHARED_RESOURCE_MAX; res_idx++)
		shared_resource_type_map[res_idx] = ENGINE_SHARED_RESOURCE_MAX;

	shared_resource_type_map[DOCA_FLOW_SHARED_RESOURCE_METER] = ENGINE_SHARED_RESOURCE_METER;
	shared_resource_type_map[DOCA_FLOW_SHARED_RESOURCE_COUNT] = ENGINE_SHARED_RESOURCE_COUNTER;
	shared_resource_type_map[DOCA_FLOW_SHARED_RESOURCE_RSS] = ENGINE_SHARED_RESOURCE_RSS;
	shared_resource_type_map[DOCA_FLOW_SHARED_RESOURCE_CRYPTO] = ENGINE_SHARED_RESOURCE_CRYPTO;
}

static void
shared_rss_cfg_translate(struct doca_flow_shared_resource_cfg *df_res_cfg,
			 struct engine_shared_resource_cfg *eng_shared_cfg)
{
	eng_shared_cfg->rss.nr_queues = df_res_cfg->rss_cfg.nr_queues;
	eng_shared_cfg->rss.queues_array = df_res_cfg->rss_cfg.queues_array;
	eng_shared_cfg->rss.types = dpdk_pipe_actions_get_rss_type(df_res_cfg->rss_cfg.flags);
}

static int
shared_crypto_cfg_translate_nisp(struct doca_flow_resource_crypto_cfg *cfg,
				 struct engine_shared_resource_cfg *engine)
{
	engine->crypto.proto_type = ENGINE_CRYPTO_PROTOCOL_NISP;

	if ((cfg->reformat_type != DOCA_FLOW_CRYPTO_REFORMAT_NONE) &&
	    (cfg->net_type != DOCA_FLOW_CRYPTO_NET_TUNNEL)) {
		DOCA_LOG_RATE_LIMIT_ERR("Invalid network mode for NISP: %d", cfg->net_type);
		return -EINVAL;
	}

	if (cfg->key_sz > sizeof(engine->crypto.key)) {
		DOCA_LOG_RATE_LIMIT_ERR("Key size %u exceeds engine buffer %u",	cfg->key_sz,
					(uint32_t)sizeof(engine->crypto.key));
		return -EINVAL;
	}

	engine->crypto.key_sz = cfg->key_sz;
	if (cfg->key_sz)
		memcpy(engine->crypto.key, cfg->key, cfg->key_sz);

	switch (cfg->reformat_type) {
	case DOCA_FLOW_CRYPTO_REFORMAT_NONE:
		engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_NONE;
		break;
	case DOCA_FLOW_CRYPTO_REFORMAT_ENCAP:
		engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_NISP_TUNNEL_ENCAP;
		break;
	case DOCA_FLOW_CRYPTO_REFORMAT_DECAP:
		engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_NISP_TUNNEL_DECAP;
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Invalid NISP reformat action type: %d",
					cfg->reformat_type);
		return -EINVAL;
	}

	return 0;
}

static int
shared_crypto_cfg_translate_esp(struct doca_flow_resource_crypto_cfg *cfg,
				struct engine_shared_resource_cfg *engine)
{
	engine->crypto.proto_type = ENGINE_CRYPTO_PROTOCOL_ESP;
	engine->crypto.security_ctx = cfg->security_ctx;

	if ((cfg->reformat_type != DOCA_FLOW_CRYPTO_REFORMAT_NONE) &&
	    (cfg->net_type != DOCA_FLOW_CRYPTO_NET_TUNNEL) &&
	    (cfg->net_type != DOCA_FLOW_CRYPTO_NET_TRANSPORT)) {
		DOCA_LOG_RATE_LIMIT_ERR("Invalid network header mode for ESP: %d", cfg->net_type);
		return -EINVAL;
	}

	switch (cfg->reformat_type) {
	case DOCA_FLOW_CRYPTO_REFORMAT_NONE:
		engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_NONE;
		break;
	case DOCA_FLOW_CRYPTO_REFORMAT_ENCAP:
		if (cfg->net_type == DOCA_FLOW_CRYPTO_NET_TUNNEL) {
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TUNNEL_ENCAP;
			break;
		}
		switch (cfg->header_type) {
		case DOCA_FLOW_CRYPTO_HEADER_IPV4:
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV4_ENCAP;
			break;
		case DOCA_FLOW_CRYPTO_HEADER_IPV6:
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV6_ENCAP;
			break;
		case DOCA_FLOW_CRYPTO_HEADER_IPV4_UDP:
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV4_UDP_ENCAP;
			break;
		case DOCA_FLOW_CRYPTO_HEADER_IPV6_UDP:
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV6_UDP_ENCAP;
			break;
		default:
			DOCA_LOG_RATE_LIMIT_ERR("Invalid ESP encap header type: %d", cfg->header_type);
			return -EINVAL;
		}
		break;
	case DOCA_FLOW_CRYPTO_REFORMAT_DECAP:
		if (cfg->net_type == DOCA_FLOW_CRYPTO_NET_TUNNEL) {
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TUNNEL_DECAP;
			break;
		}
		switch (cfg->header_type) {
		case DOCA_FLOW_CRYPTO_HEADER_IPV4:
		case DOCA_FLOW_CRYPTO_HEADER_IPV6:
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_DECAP;
			break;
		case DOCA_FLOW_CRYPTO_HEADER_IPV4_UDP:
		case DOCA_FLOW_CRYPTO_HEADER_IPV6_UDP:
			engine->crypto.reformat_type = ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_UDP_DECAP;
			break;
		default:
			DOCA_LOG_RATE_LIMIT_ERR("Invalid ESP decap header type: %d", cfg->header_type);
			return -EINVAL;
		}
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Invalid ESP reformat action type: %d", cfg->reformat_type);
		return -EINVAL;
	}

	return 0;
}

static int
shared_crypto_cfg_translate(struct doca_flow_resource_crypto_cfg *cfg,
			    struct engine_shared_resource_cfg *engine)
{
	int rc;

	if ((cfg->proto_type != DOCA_FLOW_CRYPTO_PROTOCOL_NISP) &&
	    (cfg->proto_type != DOCA_FLOW_CRYPTO_PROTOCOL_ESP_ENCRYPT) &&
	    (cfg->proto_type != DOCA_FLOW_CRYPTO_PROTOCOL_ESP_DECRYPT)) {
		DOCA_LOG_RATE_LIMIT_ERR("Unsupported crypto action protocol %d", cfg->proto_type);
		return -ENOTSUP;
	}

	if (cfg->reformat_type != DOCA_FLOW_CRYPTO_REFORMAT_NONE) {
		if (cfg->reformat_data_sz > sizeof(engine->crypto.reformat_data)) {
			DOCA_LOG_RATE_LIMIT_ERR("Reformat data size %u exceeds engine buffer %u",
						cfg->reformat_data_sz,
						(uint32_t)sizeof(engine->crypto.reformat_data));
			return -EINVAL;
		}
		engine->crypto.reformat_data_sz = cfg->reformat_data_sz;
		if (cfg->reformat_data_sz)
			memcpy(engine->crypto.reformat_data,
			       cfg->reformat_data, engine->crypto.reformat_data_sz);
	}

	switch (cfg->action_type) {
	case DOCA_FLOW_CRYPTO_ACTION_NONE:
		engine->crypto.crypto_type = ENGINE_CRYPTO_OPERATION_NONE;
		break;
	case DOCA_FLOW_CRYPTO_ACTION_ENCRYPT:
		engine->crypto.crypto_type = ENGINE_CRYPTO_OPERATION_ENCRYPT;
		break;
	case DOCA_FLOW_CRYPTO_ACTION_DECRYPT:
		engine->crypto.crypto_type = ENGINE_CRYPTO_OPERATION_DECRYPT;
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Invalid crypto action type: %d",
					cfg->action_type);
		return -EINVAL;
	}

	switch (cfg->proto_type) {
	case DOCA_FLOW_CRYPTO_PROTOCOL_NISP:
		/* NISP protocol config specific translation */
		rc = shared_crypto_cfg_translate_nisp(cfg, engine);
		break;
	case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_ENCRYPT:
	case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_DECRYPT:
		/* IPsec/ESP protocol config specific translation */
		rc = shared_crypto_cfg_translate_esp(cfg, engine);
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Unsupported crypto action protocol %d", cfg->proto_type);
		rc = -ENOTSUP;
		break;
	}

	if (rc != 0)
		return rc;

	switch (cfg->fwd.type) {
	case DOCA_FLOW_FWD_PORT:
		engine->crypto.fwd.fwd_type = ENGINE_FWD_PORT;
		engine->crypto.fwd.port_id = cfg->fwd.port_id;
		break;
	case DOCA_FLOW_FWD_PIPE:
		engine->crypto.fwd.fwd_type = ENGINE_FWD_PIPE;
		rc = dpdk_engine_next_pipe_group_id_get(cfg->fwd.next_pipe,
							&engine->crypto.fwd.pipe_id);
		if (rc != 0) {
			DOCA_LOG_RATE_LIMIT_ERR("Invalid crypto forward action next pipe - rc=%d", rc);
			return -EINVAL;
		}
		break;
	case DOCA_FLOW_FWD_DROP:
		engine->crypto.fwd.fwd_type = ENGINE_FWD_DROP;
		break;
	case DOCA_FLOW_FWD_NONE:
		if (cfg->proto_type == DOCA_FLOW_CRYPTO_PROTOCOL_ESP_ENCRYPT) {
			engine->crypto.fwd.fwd_type = ENGINE_FWD_NONE;
			engine->crypto.fwd.pipe_id = DOCA_EGRESS_TXMISS_GROUP;
			break;
		}
		DOCA_LOG_RATE_LIMIT_ERR("None forward action type supported for encrypt only");
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Invalid crypto forward action type: %d", cfg->fwd.type);
		return -EINVAL;
	}

	return 0;
}

static int
shared_res_cfg_translate(enum doca_flow_shared_resource_type df_res_type,
			 struct doca_flow_shared_resource_cfg *df_res_cfg,
			 struct engine_shared_resource_cfg *eng_shared_cfg,
			 struct doca_flow_error *error)
{
	switch (df_res_type) {
	case DOCA_FLOW_SHARED_RESOURCE_METER:
		eng_shared_cfg->meter.cir = df_res_cfg->meter_cfg.cir;
		eng_shared_cfg->meter.cbs = df_res_cfg->meter_cfg.cbs;
		return 0;
	case DOCA_FLOW_SHARED_RESOURCE_COUNT:
		return 0;
	case DOCA_FLOW_SHARED_RESOURCE_RSS:
		shared_rss_cfg_translate(df_res_cfg, eng_shared_cfg);
		return 0;
	case DOCA_FLOW_SHARED_RESOURCE_CRYPTO:
		if (shared_crypto_cfg_translate(&df_res_cfg->crypto_cfg, eng_shared_cfg) != 0) {
			set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
				  "Invalid crypto action configuration");
			return -EINVAL;
		}
		return 0;
	default:
		break;
	}

	set_error(error, DOCA_FLOW_ERROR_UNKNOWN,
		  "Unknown shared resource type: %d", df_res_type);
	return -ENOTSUP;
}

void
translate_shared_resource_unbind_cb(enum engine_shared_resource_type type,
				    uint32_t shared_resource_id, struct engine_bindable *bindable)
{
	enum doca_flow_shared_resource_type df_type;

	if (!shared_resources_unbind_cb)
		return;

	switch (type) {
	case ENGINE_SHARED_RESOURCE_METER:
		df_type = DOCA_FLOW_SHARED_RESOURCE_METER;
		break;
	case ENGINE_SHARED_RESOURCE_COUNTER:
		df_type = DOCA_FLOW_SHARED_RESOURCE_COUNT;
		break;
	case ENGINE_SHARED_RESOURCE_CRYPTO:
		df_type = DOCA_FLOW_SHARED_RESOURCE_CRYPTO;
		break;
	case ENGINE_SHARED_RESOURCE_RSS:
		df_type = DOCA_FLOW_SHARED_RESOURCE_RSS;
		break;
	default:
		DOCA_LOG_WARN("Unknown shared resource id %u on bindable %p", shared_resource_id,
			      bindable);
		return;
	}

	shared_resources_unbind_cb(df_type, shared_resource_id, (void *)bindable);
}

static int
shared_resource_cfg_build(const struct doca_flow_cfg *cfg,
			  struct engine_shared_resources_cfg *res_cfg)
{
	int type_idx;

	shared_resource_type_map_init();

	for (type_idx = 0; type_idx < DOCA_FLOW_SHARED_RESOURCE_MAX; type_idx++) {
		int engine_res_idx = shared_resource_type_map[type_idx];

		if (engine_res_idx == ENGINE_SHARED_RESOURCE_MAX) {
			DOCA_LOG_WARN("Unsupported shared resource type %u, continue...", type_idx);
			continue;
		}

		res_cfg->total_resources[engine_res_idx] = cfg->nr_shared_resources[type_idx];
	}

	/* Internal resources: */
	res_cfg->total_resources[ENGINE_SHARED_RESOURCE_INTERNAL_META] = DPDK_META_MAX_SEGMENTS;
	res_cfg->unbind_cb = translate_shared_resource_unbind_cb;
	return 0;
}

static int
engine_cfg_build(const struct doca_flow_cfg *cfg, struct engine_layer_cfg *engine_cfg,
		 struct doca_flow_error *error)
{
	int rc;

	rc = model_cfg_build(cfg, &engine_cfg->model_cfg);
	if (rc < 0) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN, "failed model init - rc=%d", rc);
		return -EINVAL;
	}

	rc = shared_resource_cfg_build(cfg, &engine_cfg->shared_res_cfg);
	if (rc < 0) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN,
			  "failed shared resource init - rc=%d", rc);
		return -EINVAL;
	}

	return 0;
}

int
doca_flow_init(const struct doca_flow_cfg *cfg,
	       struct doca_flow_error *error)
{
	int rc;
	struct engine_layer_cfg engine_cfg = {0};

	if (cfg == NULL) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM, "cfg is null");
		return -EINVAL;
	}

	rc = engine_cfg_build(cfg, &engine_cfg, error);
	if (rc < 0) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN,
			  "failed building engine config with rc=%d", rc);
		return rc;
	}

	rc = engine_layer_init(&engine_cfg);
	if (rc) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN,
			  "failed initializing engine layer with rc=%d", rc);
		return rc;
	}

	rc = dpdk_engine_init(cfg);
	if (rc < 0) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN,
			  "failed initializing dpdk engine layer with rc=%d", rc);
		goto init_failure;
	}

	rc = doca_flow_layer_init();
	if (rc) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN,
			  "failed initializing doca flow layer with rc=%d", rc);
		goto init_failure;
	}

	shared_resources_unbind_cb = cfg->unbind_cb;
	DOCA_LOG_INFO("Doca flow initialized successfully");
	return 0;

init_failure:
	DOCA_LOG_ERR("Doca flow failed initialization - rolling back allocated resources");
	doca_flow_destroy();
	return rc;
}

void
doca_flow_destroy(void)
{
	doca_flow_layer_destroy();
	dpdk_layer_unregister();
	dpdk_engine_destroy();
	engine_layer_destroy();
	DOCA_LOG_INFO("Doca flow destroyed");
}

static int
get_bindable_port_id(struct engine_bindable *obj, uint16_t *port_id)
{
	struct doca_flow_port *port = NULL;

	if (!obj)
		return -EINVAL;

	switch (obj->type) {
	case DF_ENGINE_BINDABLE_TYPE_PORT:
		port = (struct doca_flow_port *)obj;
		break;
	case DF_ENGINE_BINDABLE_TYPE_PIPE:
		port = ((struct doca_flow_pipe *)obj)->port;
		break;
	default:
		return -ENOTSUP;
	}

	if (!port)
		return -EINVAL;
	*port_id = port->port_id;
	return 0;
}

int
doca_flow_shared_resource_cfg(enum doca_flow_shared_resource_type type, uint32_t id,
			      struct doca_flow_shared_resource_cfg *cfg,
			      struct doca_flow_error *error)
{
	enum engine_shared_resource_type res_type;
	struct engine_shared_resource_cfg res_cfg = {0};
	int rc;

	if (unlikely(type >= DOCA_FLOW_SHARED_RESOURCE_MAX)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "unsupported shared resource of type %u", type);
		return -ENOTSUP;
	}

	res_type = shared_resource_type_map[type];
	if (unlikely(!engine_shared_resource_is_initialized(res_type, id))) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "shared resource of type %u and id %u is uninitialized", res_type,
			  id);
		return -EINVAL;
	}
	rc = shared_res_cfg_translate(type, cfg, &res_cfg, error);
	if (rc)
		return rc;

	rc = engine_shared_resource_cfg_set(res_type, id, &res_cfg);
	if (unlikely(rc)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "failed configuring shared resource of type %u and id %u ", res_type,
			  id);
		return rc;
	}

	return 0;
}

static bool
is_override_res_cfg_unneeded(enum engine_shared_resource_type res_type)
{
	return ((res_type != ENGINE_SHARED_RESOURCE_COUNTER) &&
		(res_type != ENGINE_SHARED_RESOURCE_METER) &&
		(res_type != ENGINE_SHARED_RESOURCE_RSS) &&
		(res_type != ENGINE_SHARED_RESOURCE_CRYPTO));
}

static int
override_res_cfg_cb(enum engine_shared_resource_type res_type,
		    uint32_t res_id, void *bindable_obj)
{
	uint16_t port_id;
	int rc;
	struct engine_shared_resource_cfg res_cfg = { 0 };

	if (is_override_res_cfg_unneeded(res_type))
		return 0;

	rc = get_bindable_port_id(bindable_obj, &port_id);
	if (rc) {
		DOCA_LOG_ERR("failed preparing shared resources (%u) id (%u). rc=%d.",
			     res_type, res_id, rc);

		return rc;
	}

	if (res_type != ENGINE_SHARED_RESOURCE_COUNTER) {
		rc = engine_shared_resource_cfg_get(res_type, res_id, &res_cfg);
		if (rc) {
			DOCA_LOG_ERR(
				"failed preparing shared resources - unexpected rc=%d for getting resource type %u, id %u.",
				rc, res_type, res_id);

			return rc;
		}
	}

	switch ((int)res_type) {
	case ENGINE_SHARED_RESOURCE_COUNTER:
		res_cfg.counter.port_id = port_id;
		break;
	case ENGINE_SHARED_RESOURCE_METER:
		res_cfg.meter.port_id = port_id;
		break;
	case ENGINE_SHARED_RESOURCE_RSS:
		res_cfg.rss.port_id = port_id;
		break;
	case ENGINE_SHARED_RESOURCE_CRYPTO:
		res_cfg.crypto.port_id = port_id;
		break;
	}

	rc = engine_shared_resource_cfg_set(res_type, res_id, &res_cfg);
	if (rc) {
		DOCA_LOG_ERR(
			"failed preparing shared resources - unexpected rc=%d for resource type %u, id %u.",
			rc, res_type, res_id);

		return rc;
	}

	return 0;
}

int
doca_flow_shared_resources_bind(enum doca_flow_shared_resource_type type, uint32_t *res_array,
				uint32_t res_array_len, void *bindable_obj,
				struct doca_flow_error *error)
{
	enum engine_shared_resource_type res_type;
	int rc;

	if (unlikely(type >= DOCA_FLOW_SHARED_RESOURCE_MAX)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "unsupported shared resource of type %u", type);
		return -ENOTSUP;
	}

	res_type = shared_resource_type_map[type];

	rc = engine_shared_resource_bind_bulk(res_type, res_array, res_array_len, bindable_obj,
					      override_res_cfg_cb);
	if (unlikely(rc)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "failed binding %u shared resources of type %u",
			  res_array_len, res_type);
		return rc;
	}

	return 0;
}

static void
translate_query_results(enum doca_flow_shared_resource_type type,
		  struct engine_shared_resource_query_result *engine_results_array,
		  struct doca_flow_shared_resource_result *query_results_array,
		  int array_len)
{
	switch (type) {
	case DOCA_FLOW_SHARED_RESOURCE_COUNT:
		for (int i = 0; i < array_len; i++) {
			query_results_array[i].counter.total_bytes =
				engine_results_array[i].counter.total_bytes;
			query_results_array[i].counter.total_pkts =
				engine_results_array[i].counter.total_pkts;
		}
		break;
	default:
		break;
	}
}

int
doca_flow_shared_resources_query(enum doca_flow_shared_resource_type type,
				uint32_t *res_array,
				struct doca_flow_shared_resource_result *query_results_array,
				uint32_t array_len,
				struct doca_flow_error *error)
{
	enum engine_shared_resource_type res_type;
	int rc;

	if (unlikely(type >= DOCA_FLOW_SHARED_RESOURCE_MAX)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "unsupported shared resource of type %u", type);
		return -ENOTSUP;
	}
	if (unlikely(!array_len)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "failed querying resource type %u - array len 0", type);

		return -EINVAL;
	}

	struct engine_shared_resource_query_result engine_results_array[array_len];

	res_type = shared_resource_type_map[type];
	rc = engine_shared_resource_query_bulk(res_type, res_array, array_len,
				engine_results_array, array_len);
	if (unlikely(rc)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "failed querying %u shared resources of type %u",
			  array_len, res_type);
		return rc;
	}
	translate_query_results(type, engine_results_array, query_results_array, array_len);
	return 0;
}

/**
 * @brief
 *
 * @param match
 *
 * @return
 */
struct doca_flow_pipe_entry*
doca_flow_pipe_add_entry(uint16_t pipe_queue,
			 struct doca_flow_pipe *pipe,
			 const struct doca_flow_match *match,
			 const struct doca_flow_actions *actions,
			 const struct doca_flow_monitor *mon,
			 const struct doca_flow_fwd *fwd,
			 uint32_t flags,
			 void *usr_ctx,
			 struct doca_flow_error *error)
{
	struct doca_flow_pipe_entry *entry;
	struct doca_flow_port *port;

	if (pipe == NULL) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "pipe is null");
		return NULL;
	}

	if (unlikely(pipe->type == ENGINE_PIPE_CT)) {
		port = doca_flow_layer_get_pipe_port(pipe);
		if (port->ct_ctx->ct_pipe == NULL) {
			set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
				  "Invalid operation for CT pipe");
			return NULL;
		}
		pipe = port->ct_ctx->ct_pipe;
	}

	entry = dpdk_pipe_entry_add(pipe_queue, 0 /* priority */, pipe, match, NULL /* match_mask */,
				   actions, NULL /* action_descs */, mon, fwd, flags, usr_ctx);
	if (unlikely(entry == NULL)) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN, "dkdp pipe entry add failed");
		return NULL;
	}

	return entry;
}

struct doca_flow_pipe_entry *
doca_flow_pipe_ordered_list_add_entry(uint16_t pipe_queue,
		struct doca_flow_pipe *pipe,
		uint32_t idx,
		const struct doca_flow_ordered_list *ordered_list,
		const struct doca_flow_fwd *fwd,
		enum doca_flow_flags_type flags,
		void *user_ctx,
		struct doca_flow_error *error)
{
	struct doca_flow_pipe_entry *entry;

	if (unlikely(pipe == NULL)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM, "pipe is null");
		return NULL;
	}

	if (unlikely(pipe->type != ENGINE_PIPE_ORDERED_LIST)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM, "pipe is not an ordered list pipe");
		return NULL;
	}

	entry = dpdk_pipe_entry_add_with_index(pipe_queue, pipe, idx, NULL /* actions */,
					       NULL /* action_descs */, NULL /* monitor */,
					       ordered_list, fwd, flags, user_ctx);
	if (unlikely(entry == NULL)) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN, "DPDK pipe entry add failed");
		return NULL;
	}

	return entry;
}

struct doca_flow_pipe_entry*
doca_flow_pipe_lpm_add_entry(uint16_t pipe_queue,
			 struct doca_flow_pipe *pipe,
			 const struct doca_flow_match *match,
			 const struct doca_flow_match *match_mask,
			 const struct doca_flow_actions *actions,
			 const struct doca_flow_monitor *monitor,
			 const struct doca_flow_fwd *fwd,
			 const enum doca_flow_flags_type flags,
			 void *usr_ctx,
			 struct doca_flow_error *error)
{
	struct doca_flow_pipe_entry *entry;

	if (unlikely(pipe == NULL)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "pipe is null");
		return NULL;
	}

	if (unlikely(pipe->type != ENGINE_PIPE_LPM)) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "invalid pipe type");
		return NULL;
	}

	if (match == NULL || match_mask == NULL || fwd == NULL) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "either match %p or match_mask %p or fwd %p is null", match, match_mask, fwd);
		return NULL;
	}

	entry = dpdk_pipe_entry_add(pipe_queue, 0 /* priority */, pipe, match, match_mask, actions,
				   NULL /* action_descs */, NULL /* monitor */, fwd, flags, usr_ctx);
	if (unlikely(entry == NULL)) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN, "dkdp pipe entry add failed");
		return NULL;
	}

	return entry;
}

struct doca_flow_pipe_entry*
doca_flow_pipe_control_add_entry(uint16_t pipe_queue,
			uint32_t priority,
			struct doca_flow_pipe *pipe,
			const struct doca_flow_match *match,
			const struct doca_flow_match *match_mask,
			const struct doca_flow_actions *actions,
			const struct doca_flow_action_descs *action_descs,
			const struct doca_flow_monitor *monitor,
			const struct doca_flow_fwd *fwd,
			struct doca_flow_error *error)
{
	struct doca_flow_pipe_entry *entry;

	if (pipe == NULL) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM,
			  "pipe is null");
		return NULL;
	}

	entry = dpdk_pipe_entry_add(pipe_queue, priority, pipe, match, match_mask,
				   actions, action_descs, monitor, fwd, 0, NULL);
	if (unlikely(entry == NULL)) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN, "dkdp pipe entry add failed");
		return NULL;
	}
	return entry;
}

int
doca_flow_pipe_rm_entry(uint16_t pipe_queue, void *usr_ctx,
			struct doca_flow_pipe_entry *entry)
{
	if (entry == NULL)
		return -EINVAL;
	return dpdk_pipe_entry_remove(pipe_queue, usr_ctx, entry);
}

int
doca_flow_entries_process(struct doca_flow_port *port,
		uint16_t pipe_queue, uint64_t timeout,
		uint32_t max_processed_entries)
{
	return dpdk_flow_entries_process(port, pipe_queue,
				timeout, max_processed_entries);
}

enum doca_flow_entry_status
doca_flow_pipe_entry_get_status(struct doca_flow_pipe_entry *entry)
{
	return dpdk_flow_pipe_entry_status_get(entry);
}

/**
 * @brief
 *
 * @param cfg
 * @param port
 *
 * @return
 */
struct doca_flow_port *
doca_flow_port_start(const struct doca_flow_port_cfg *cfg,
		     struct doca_flow_error *error)

{
	struct doca_flow_port *port = NULL;
	uint16_t port_id;

	if (cfg == NULL)
		return NULL;
	switch (cfg->type) {
	case DOCA_FLOW_PORT_DPDK_BY_ID:
		DOCA_DLOG_INFO("new dpdk port id:%s", cfg->devargs);
		port_id = atoi(cfg->devargs);
		port = dpdk_port_legacy_start(port_id, cfg);
		if (unlikely(port == NULL))
			set_error(error, DOCA_FLOW_ERROR_UNKNOWN, "dpdk port start failed");
		break;
	default:
		set_error(error, DOCA_FLOW_ERROR_UNSUPPORTED,
			  "unsupported port type");
		return NULL;
	}

	return port;
}

/**
 * @brief - close port
 *  release all resources used by port, including all HW rules.
 *
 * @param port
 *
 * @return
 */
int
doca_flow_port_stop(struct doca_flow_port *port)
{
	int rc;

	if (port == NULL)
		return -EINVAL;
	DOCA_DLOG_INFO("port id = %d stopped", port->port_id);

	rc = engine_shared_resource_unbind(&port->bindable_obj);
	if (unlikely(rc))
		DOCA_LOG_CRIT("failed stopping port %u - unbind shared resource for port rc=%d",
			      port->port_id, rc);

	rc = dpdk_port_legacy_stop(port);
	if (unlikely(rc))
		DOCA_LOG_CRIT("failed stopping port %u - dpdk failure rc=%d", port->port_id,
			      rc);
	return rc;
}

int
doca_flow_port_pair(struct doca_flow_port *port, struct doca_flow_port *pair_port)
{
	if (port == NULL || pair_port == NULL)
		return -EINVAL;
	return dpdk_port_legacy_pair(port, pair_port);
}

struct doca_flow_pipe *
doca_flow_pipe_create(const struct doca_flow_pipe_cfg *cfg,
		const struct doca_flow_fwd *fwd,
		const struct doca_flow_fwd *fwd_miss,
		struct doca_flow_error *error)
{
	enum engine_pipe_type pipe_type;
	struct doca_flow_pipe *pipe;

	if (cfg == NULL) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM, "cfg is null");
		return NULL;
	}
	if (cfg->port == NULL) {
		set_error(error, DOCA_FLOW_ERROR_INVALID_PARAM, "port is null");
		return NULL;
	}

	switch (cfg->attr.type) {
	case DOCA_FLOW_PIPE_BASIC:
		pipe_type = ENGINE_PIPE_BASIC;
		break;
	case DOCA_FLOW_PIPE_LPM:
		pipe_type = ENGINE_PIPE_LPM;
		break;
	case DOCA_FLOW_PIPE_CONTROL:
		pipe_type = ENGINE_PIPE_CONTROL;
		break;
	case DOCA_FLOW_PIPE_ORDERED_LIST:
		pipe_type = ENGINE_PIPE_ORDERED_LIST;
		break;
	default:
		set_error(error, DOCA_FLOW_ERROR_UNSUPPORTED,
			  "failed pipe creation - pipe type %u unsupported", cfg->attr.type);
		return NULL;
	}

	pipe = dpdk_pipe_create(pipe_type, cfg, fwd, fwd_miss);
	if (unlikely(pipe == NULL)) {
		set_error(error, DOCA_FLOW_ERROR_UNKNOWN, "dpdk pipe create failed");
		return NULL;
	}

	return pipe;
}

void
doca_flow_pipe_destroy(struct doca_flow_pipe *pipe)
{
	struct doca_flow_port *port;

	if (pipe == NULL)
		return;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed destroying pipe %s - port is null", pipe->name);
		return;
	}

	dpdk_pipe_entries_flush(port, pipe);
	if (dpdk_engine_ct_ops_get() != NULL && pipe->type == ENGINE_PIPE_CT)
		dpdk_engine_ct_ops_get()->destroy(port->ct_ctx);
	engine_shared_resource_unbind(&pipe->bindable_obj);
	dpdk_pipe_destroy(port, pipe);
}

void
doca_flow_port_pipes_flush(struct doca_flow_port *port)
{
	if (port == NULL)
		return;
	dpdk_port_legacy_flush_pipes(port);
}

void
doca_flow_port_destroy(struct doca_flow_port *port)
{
	if (port == NULL)
		return;
	dpdk_port_legacy_destroy(port);
}

void
doca_flow_port_pipes_dump(struct doca_flow_port *port, FILE *f)
{
	if (port == NULL) {
		DOCA_LOG_WARN("Invalid port to dump");
		return;
	}
	dpdk_port_pipes_dump(port, f);
}

void
doca_flow_pipe_dump(struct doca_flow_pipe *pipe, FILE *f)
{
	struct doca_flow_port *port;

	if (pipe == NULL) {
		DOCA_LOG_WARN("Invalid pipe to dump");
		return;
	}

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port))
		DOCA_LOG_RATE_LIMIT_WARN("pipe dump failed for pipe %s - port is null", pipe->name);

	dpdk_pipe_dump(port, pipe, f);
}

int
doca_flow_query(struct doca_flow_pipe_entry *entry,
		struct doca_flow_query *query_stats)
{
	if (entry == NULL || query_stats == NULL)
		return -EINVAL;

	return dpdk_pipe_entry_query(entry, query_stats);
}

int
doca_flow_aging_handle(struct doca_flow_port *port, uint16_t queue,
		uint64_t quota, struct doca_flow_aged_query *entries, int len)
{
	return flow_age_check(port, queue, quota, entries, len);
}

struct doca_flow_port *
doca_flow_port_switch_get(void)
{
	return dpdk_port_legacy_switch_get();
}
