/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stddef.h>
#include <string.h>

#include "private/common_defines.h"
#include "engine_field_mapping.h"
#include "engine_field_opcode.h"
#include "engine_supported_types.h"
#include "engine_to_string.h"
#include "engine_layer.h"

#include "doca_flow_layer.h"
#include "doca_flow_priv.h"

#include <doca_flow.h>
#include <doca_log.h>

DOCA_LOG_REGISTER(doca_flow_layer);

static struct engine_fcp *match_iterator;

#define REG_PROTO(opcode_str, byte_len, offset) \
	do { \
		rc = register_proto(opcode_str, offset, byte_len); \
		if (rc < 0) \
			return rc; \
	} while (0)

static int
register_proto(const char *opcode_str, uint32_t offset, uint16_t len)
{
	struct engine_field_opcode opcode;
	int rc;

	rc = engine_string_to_opcode(opcode_str, strlen(opcode_str), &opcode);
	if (rc < 0) {
		DOCA_LOG_ERR("failed parsing opcode string %s", opcode_str);
		return rc;
	}

	rc = engine_field_mapping_set_length(&opcode, len);
	if (rc < 0) {
		DOCA_LOG_ERR("failed registering field len (opcode=0x%s,len=%u)", opcode_str, len);
		return rc;
	}
	rc = engine_field_mapping_set_offset(&opcode, offset);
	if (rc < 0) {
		DOCA_LOG_ERR("failed registering field offset (opcode=%s,off=%u)", opcode_str,
			     offset);
		return rc;
	}

	DOCA_LOG_DBG("Registered field opcode=%s, offset=%u, len=%u)", opcode_str, offset, len);
	return 0;
}

static int
register_proto_eth()
{
	int rc;

	REG_PROTO("match.packet.outer.eth.dst_mac", 6,
		  offsetof(struct doca_flow_match, out_dst_mac));
	REG_PROTO("match.packet.outer.eth.src_mac", 6,
		  offsetof(struct doca_flow_match, out_src_mac));
	REG_PROTO("match.packet.outer.eth.type", 2,
		  offsetof(struct doca_flow_match, out_eth_type));

	REG_PROTO("match.packet.inner.eth.dst_mac", 6,
		  offsetof(struct doca_flow_match, in_dst_mac));
	REG_PROTO("match.packet.inner.eth.src_mac", 6,
		  offsetof(struct doca_flow_match, in_src_mac));
	REG_PROTO("match.packet.inner.eth.type", 2,
		  offsetof(struct doca_flow_match, in_eth_type));

	return 0;
}

static int
register_proto_eth_vlan()
{
	int rc;

	REG_PROTO("match.packet.outer.eth_vlan.tci", 2,
		  offsetof(struct doca_flow_match, out_vlan_tci));
	REG_PROTO("match.packet.inner.eth_vlan.tci", 2,
		  offsetof(struct doca_flow_match, in_vlan_tci));

	return 0;
}

static int
register_proto_ipv4()
{
	int rc;

	REG_PROTO("match.packet.outer.ipv4.src_ip", 4,
		  offsetof(struct doca_flow_match, out_src_ip.ipv4_addr));
	REG_PROTO("match.packet.outer.ipv4.dst_ip", 4,
		  offsetof(struct doca_flow_match, out_dst_ip.ipv4_addr));
	REG_PROTO("match.packet.outer.ipv4.next_proto", 1,
		  offsetof(struct doca_flow_match, out_l4_type));

	REG_PROTO("match.packet.inner.ipv4.src_ip", 4,
		  offsetof(struct doca_flow_match, in_src_ip.ipv4_addr));
	REG_PROTO("match.packet.inner.ipv4.dst_ip", 4,
		  offsetof(struct doca_flow_match, in_dst_ip.ipv4_addr));
	REG_PROTO("match.packet.inner.ipv4.next_proto", 1,
		  offsetof(struct doca_flow_match, in_l4_type));

	return 0;
}

static int
register_proto_ipv6()
{
	int rc;

	REG_PROTO("match.packet.outer.ipv6.src_ip", 16,
		  offsetof(struct doca_flow_match, out_src_ip.ipv6_addr));
	REG_PROTO("match.packet.outer.ipv6.dst_ip", 16,
		  offsetof(struct doca_flow_match, out_dst_ip.ipv6_addr));
	REG_PROTO("match.packet.outer.ipv6.next_proto", 1,
		  offsetof(struct doca_flow_match, out_l4_type));

	REG_PROTO("match.packet.inner.ipv6.src_ip", 16,
		  offsetof(struct doca_flow_match, in_src_ip.ipv6_addr));
	REG_PROTO("match.packet.inner.ipv6.dst_ip", 16,
		  offsetof(struct doca_flow_match, in_dst_ip.ipv6_addr));
	REG_PROTO("match.packet.inner.ipv6.next_proto", 1,
		  offsetof(struct doca_flow_match, in_l4_type));

	return 0;
}

static int
register_proto_udp()
{
	int rc;

	REG_PROTO("match.packet.outer.udp.src_port", 2,
		  offsetof(struct doca_flow_match, out_src_port));
	REG_PROTO("match.packet.outer.udp.dst_port", 2,
		  offsetof(struct doca_flow_match, out_dst_port));

	REG_PROTO("match.packet.inner.udp.src_port", 2,
		  offsetof(struct doca_flow_match, in_src_port));
	REG_PROTO("match.packet.inner.udp.dst_port", 2,
		  offsetof(struct doca_flow_match, in_dst_port));

	return 0;
}

static int
register_proto_tcp()
{
	int rc;

	REG_PROTO("match.packet.outer.tcp.src_port", 2,
		  offsetof(struct doca_flow_match, out_src_port));
	REG_PROTO("match.packet.outer.tcp.dst_port", 2,
		  offsetof(struct doca_flow_match, out_dst_port));
	REG_PROTO("match.packet.outer.tcp.flags", 1,
		  offsetof(struct doca_flow_match, out_tcp_flags));

	REG_PROTO("match.packet.inner.tcp.src_port", 2,
		  offsetof(struct doca_flow_match, in_src_port));
	REG_PROTO("match.packet.inner.tcp.dst_port", 2,
		  offsetof(struct doca_flow_match, in_dst_port));
	REG_PROTO("match.packet.inner.tcp.flags", 1,
		  offsetof(struct doca_flow_match, in_tcp_flags));

	return 0;
}

static int
register_tun_protocols()
{
	int rc;

	REG_PROTO("match.packet.tunnel.vxlan.vni", 3,
		  offsetof(struct doca_flow_match, tun.vxlan_tun_id));
	REG_PROTO("match.packet.tunnel.gre.key", 4,
		  offsetof(struct doca_flow_match, tun.gre_key));
	REG_PROTO("match.packet.tunnel.gre.protocol", 2,
		  offsetof(struct doca_flow_match, tun.protocol));
	REG_PROTO("match.packet.tunnel.gtp.teid", 4,
		  offsetof(struct doca_flow_match, tun.gtp_teid));
	REG_PROTO("match.packet.tunnel.nisp.hdr", DOCA_FLOW_NISP_HEADER_LEN,
		  offsetof(struct doca_flow_match, tun.nisp_hdr));
	REG_PROTO("match.packet.tunnel.audp.hdr", DOCA_FLOW_AUDP_HEADER_LEN,
		  offsetof(struct doca_flow_match, tun.audp_hdr));
	REG_PROTO("match.packet.tunnel.esp.spi", 4,
		  offsetof(struct doca_flow_match, tun.esp_spi));
	REG_PROTO("match.packet.tunnel.esp.sn", 4,
		  offsetof(struct doca_flow_match, tun.esp_sn));

	return 0;
}

#define REG_ITERATOR_ADD_OPCODE(iter_arg, opcode_str) \
	do { \
		struct engine_field_opcode opcode; \
\
		rc = engine_string_to_opcode(opcode_str, strlen(opcode_str), &opcode);	\
		if (unlikely(rc < 0))	\
			return rc;	\
\
		rc = engine_fcp_add_opcode(iter_arg, &opcode); \
		if (unlikely(rc)) \
			return rc; \
	} while (0)

#define REG_ITERATOR_ADD_SELECTOR_1(fcp, field_offset, field_length, val1, op1_str) \
	do { \
		int rc; \
		struct engine_fcp_selector_cfg selector_cfg = {0}; \
\
		selector_cfg.offset = field_offset; \
		selector_cfg.length = field_length; \
		selector_cfg.selectors = 1; \
\
		selector_cfg.selector[0].value = val1; \
		rc = engine_string_to_opcode(op1_str, strlen(op1_str), \
					     &selector_cfg.selector[0].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		rc = engine_fcp_add_selector(fcp, &selector_cfg); \
		if (unlikely(rc)) \
			return rc; \
	} while (0)

#define REG_ITERATOR_ADD_SELECTOR_2(fcp, field_offset, field_length, val1, op1_str, val2, op2_str) \
	do { \
		int rc; \
		struct engine_fcp_selector_cfg selector_cfg = {0}; \
\
		selector_cfg.offset = field_offset; \
		selector_cfg.length = field_length; \
		selector_cfg.selectors = 2; \
\
		selector_cfg.selector[0].value = val1; \
		rc = engine_string_to_opcode(op1_str, strlen(op1_str), \
					     &selector_cfg.selector[0].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		selector_cfg.selector[1].value = val2; \
		rc = engine_string_to_opcode(op2_str, strlen(op2_str), \
					     &selector_cfg.selector[1].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		rc = engine_fcp_add_selector(fcp, &selector_cfg); \
		if (unlikely(rc)) \
			return rc; \
	} while (0)

#define REG_ITERATOR_ADD_SELECTOR_6(fcp, field_offset, field_length, val1, op1_str, val2, op2_str, \
				    val3, op3_str, val4, op4_str, val5, op5_str, val6, op6_str) \
	do { \
		int rc; \
		struct engine_fcp_selector_cfg selector_cfg = {0}; \
\
		selector_cfg.offset = field_offset; \
		selector_cfg.length = field_length; \
		selector_cfg.selectors = 3; \
\
		selector_cfg.selector[0].value = val1; \
		rc = engine_string_to_opcode(op1_str, strlen(op1_str), \
					     &selector_cfg.selector[0].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		selector_cfg.selector[1].value = val2; \
		rc = engine_string_to_opcode(op2_str, strlen(op2_str), \
					     &selector_cfg.selector[1].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		selector_cfg.selector[2].value = val3; \
		rc = engine_string_to_opcode(op3_str, strlen(op3_str), \
					     &selector_cfg.selector[2].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		selector_cfg.selector[3].value = val4; \
		rc = engine_string_to_opcode(op4_str, strlen(op4_str), \
					     &selector_cfg.selector[3].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		selector_cfg.selector[4].value = val5; \
		rc = engine_string_to_opcode(op5_str, strlen(op5_str), \
					     &selector_cfg.selector[4].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		selector_cfg.selector[5].value = val6; \
		rc = engine_string_to_opcode(op6_str, strlen(op6_str), \
					     &selector_cfg.selector[5].opcode); \
		if (unlikely(rc < 0)) \
			return rc; \
\
		rc = engine_fcp_add_selector(fcp, &selector_cfg); \
		if (unlikely(rc)) \
			return rc; \
	} while (0)


static int
register_match_fcp()
{
	int rc;

	match_iterator = engine_fcp_create(DF_SUP_USER_DEFINED_TYPE_MATCH);
	if (!match_iterator)
		return -ENOMEM;

	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.outer.eth.dst_mac");
	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.outer.eth.src_mac");
	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.outer.eth.type");
	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.outer.eth_vlan.tci");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, out_src_ip.type), 1,
				    DOCA_FLOW_IP4_ADDR, "match.packet.outer.ipv4.src_ip",
				    DOCA_FLOW_IP6_ADDR, "match.packet.outer.ipv6.src_ip");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, out_dst_ip.type), 1,
				    DOCA_FLOW_IP4_ADDR, "match.packet.outer.ipv4.dst_ip",
				    DOCA_FLOW_IP6_ADDR, "match.packet.outer.ipv6.dst_ip");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, out_src_ip.type), 1,
				    DOCA_FLOW_IP4_ADDR, "match.packet.outer.ipv4.next_proto",
				    DOCA_FLOW_IP6_ADDR, "match.packet.outer.ipv6.next_proto");

	REG_ITERATOR_ADD_SELECTOR_1(match_iterator,
				    offsetof(struct doca_flow_match, out_l4_type), 1,
				    DOCA_PROTO_TCP, "match.packet.outer.tcp.flags");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, out_l4_type), 1,
				    DOCA_PROTO_TCP, "match.packet.outer.tcp.src_port",
				    DOCA_PROTO_UDP, "match.packet.outer.udp.src_port");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, out_l4_type), 1,
				    DOCA_PROTO_TCP, "match.packet.outer.tcp.dst_port",
				    DOCA_PROTO_UDP, "match.packet.outer.udp.dst_port");

	REG_ITERATOR_ADD_SELECTOR_6(match_iterator,
				    offsetof(struct doca_flow_match, tun.type), 1,
				    DOCA_FLOW_TUN_VXLAN, "match.packet.tunnel.vxlan.vni",
				    DOCA_FLOW_TUN_GTPU, "match.packet.tunnel.gtp.teid",
				    DOCA_FLOW_TUN_GRE, "match.packet.tunnel.gre.key",
				    DOCA_FLOW_TUN_NISP, "match.packet.tunnel.nisp.hdr",
				    DOCA_FLOW_TUN_AUDP, "match.packet.tunnel.audp.hdr",
				    DOCA_FLOW_TUN_ESP, "match.packet.tunnel.esp.spi");

	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.inner.eth.dst_mac");
	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.inner.eth.src_mac");
	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.inner.eth.type");
	REG_ITERATOR_ADD_OPCODE(match_iterator, "match.packet.inner.eth_vlan.tci");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, in_src_ip.type), 1,
				    DOCA_FLOW_IP4_ADDR, "match.packet.inner.ipv4.src_ip",
				    DOCA_FLOW_IP6_ADDR, "match.packet.inner.ipv6.src_ip");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, in_dst_ip.type), 1,
				    DOCA_FLOW_IP4_ADDR, "match.packet.inner.ipv4.dst_ip",
				    DOCA_FLOW_IP6_ADDR, "match.packet.inner.ipv6.dst_ip");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, in_src_ip.type), 1,
				    DOCA_FLOW_IP4_ADDR, "match.packet.inner.ipv4.next_proto",
				    DOCA_FLOW_IP6_ADDR, "match.packet.inner.ipv6.next_proto");

	REG_ITERATOR_ADD_SELECTOR_1(match_iterator,
				    offsetof(struct doca_flow_match, in_l4_type), 1,
				    DOCA_PROTO_TCP, "match.packet.inner.tcp.flags");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, in_l4_type), 1,
				    DOCA_PROTO_TCP, "match.packet.inner.tcp.src_port",
				    DOCA_PROTO_UDP, "match.packet.inner.udp.src_port");

	REG_ITERATOR_ADD_SELECTOR_2(match_iterator,
				    offsetof(struct doca_flow_match, in_l4_type), 1,
				    DOCA_PROTO_TCP, "match.packet.inner.tcp.dst_port",
				    DOCA_PROTO_UDP, "match.packet.inner.udp.dst_port");

	return 0;
}

int
doca_flow_register_capabilities()
{
	int rc;

	rc = register_proto_eth();
	if (unlikely(rc < 0))
		return rc;

	rc = register_proto_eth_vlan();
	if (unlikely(rc < 0))
		return rc;

	rc = register_proto_ipv4();
	if (unlikely(rc < 0))
		return rc;

	rc = register_proto_ipv6();
	if (unlikely(rc < 0))
		return rc;

	rc = register_proto_udp();
	if (unlikely(rc < 0))
		return rc;

	rc = register_proto_tcp();
	if (unlikely(rc < 0))
		return rc;

	rc = register_tun_protocols();
	if (unlikely(rc < 0))
		return rc;

	rc = register_match_fcp();
	if (unlikely(rc < 0))
		return rc;

	return 0;
}

int
doca_flow_layer_init()
{
	int rc;

	rc = doca_flow_register_capabilities();
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed initializing doca flow registration rc=%d", rc);
		doca_flow_layer_destroy();
		return rc;
	}

	DOCA_LOG_INFO("Doca flow layer initialized");
	return 0;
}

void
doca_flow_layer_destroy()
{
	if (match_iterator) {
		engine_fcp_destroy(match_iterator);
		match_iterator = NULL;
	}

	DOCA_LOG_INFO("Doca flow layer destroyed");
}

struct doca_flow_port *
doca_flow_layer_get_pipe_port(struct doca_flow_pipe *pipe)
{
	return pipe->port;
}

struct engine_fcp *
doca_flow_layer_get_default_match(void)
{
	return match_iterator;
}
