/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DOCA_FLOW_PRIV_H_
#define DOCA_FLOW_PRIV_H_

#include "dpdk_pipe_legacy.h"
#include "dpdk_table.h"
#include "dpdk_flow.h"

#include "engine_bindable.h"
#include "engine_pipe.h"

#include "flow_ct.h"

#ifdef __cplusplus
extern "C" {
#endif

struct doca_flow_port;

extern struct doca_flow_port *flow_switch;

struct flow_age_queue;
struct dpdk_hws_item;
struct dpdk_profiles_ids;
struct dpdk_table_manager;

/**
 * @brief doca flow pipe entry
 */
struct doca_flow_pipe_entry {
	LIST_ENTRY(doca_flow_pipe_entry) next;
	/**< list member on a pipe */
	uint16_t pipe_queue;
	/**< owning queue */
	uint32_t entry_id;
	/**< id of an entry on queue */
	uint32_t table_id;
	/**< table_id */
	enum doca_flow_entry_status status;
	/**< entry handler status */
	uint32_t profile_id;
	/**< profile id */
	uint32_t *meter_id;
	/**< meter id */
	struct flow_age_item *age_item;
	/**< Pointer to aging item */
	struct dpdk_hws_item *hws_item;
	/**< Pointer to hws item */
	struct dpdk_table res;
	/**< Dpdk table */
	struct dpdk_flow_tracker tracker;
	/**< Dpdk flow tracker */
	void *pipe;
	/**< Pointer to pipe */
	int encap_id;
	/**< Encap id */
	void *obj_ctx;
	/**< object context */
	void *user_ctx;
	/**< user context for completion callback */
};

/**
 * @brief doca flow pipe
 */
struct doca_flow_pipe {
	struct engine_bindable bindable_obj;
	/**< bindable object - MUST be the first member */
	enum engine_pipe_type type;
	/**< pipe type */
	struct doca_flow_port *port;
	/**< Pointer to a port */
	LIST_ENTRY(doca_flow_pipe) next;
	/**< list member on a port */
	char name[ENGINE_PIPE_NAME_LEN];
	/**< pipe name */
	uint32_t root_id;
	/**< unique id on root */
	uint32_t group_id;
	/**< dpdk group */
	bool is_root;
	/**< indicate if pipe is on root */
	bool is_connected;
	/**< indicate if pipe attached to port list */
	bool is_aging;
	/**< indicate if pipe has aging */
	void *private_ctx;
	/**< pointer to the pipe-private context */
	uint32_t pipe_size;
	/**< maximum number of control pipe entries */
	struct doca_flow_pipe *meter_pipe;
	/**< pointer to meter pipe */
	uint16_t nb_queues;
	/**< number of queues */
	struct dpdk_table *basic_table;
	/**<pointer to table */
	struct dpdk_table_manager *table_manager;
	/**< pointer to table_manager */
	uint32_t nr_entries;
	/**< nr_entries in pipe */
	struct priv_doca_mempool *flow_entry_pool;
	/**< entries memory pool */
	struct flow_age_queue **age_queues;
	/**< Aging per queues */
	struct dpdk_pipe dpdk_pipe;
	/**< Dpdk pipe */
};

enum port_switch_attr {
	IS_REPRESENTOR = (1 << 0),
	/**< is representor port of SF or PF */
	IS_PF = (1 << 1),
	/**< is PF port */
};
#define IS_PORT_PF(port) ((port)->switch_attr & IS_PF)
#define IS_PORT_REPRESENTOR(port) ((port)->switch_attr & IS_REPRESENTOR)

/**
 * @brief doca flow port
 */
struct doca_flow_port {
	struct engine_bindable bindable_obj;
	/**< bindable object - MUST be the first member */
	uint16_t port_id;
	/**< dpdk port id */
	uint16_t proxy_port_id;
	/**< dpdk proxy port id */
	int idx;
	/**< index in container */
	enum port_switch_attr switch_attr;
	/**< port switch attr */
	struct dpdk_port *dpdk_port;
	/**< Pointer to dpdk port */
	bool rx_encap;
	/**< indicate if encap is done on rx */
	volatile uint16_t state;
	/**< Port state */
	struct doca_flow_port *hairpin_peer;
	/**< Pointer to peer */
	rte_spinlock_t pipe_lock;
	/**< Pipe lock */
	rte_spinlock_t profile_lock;
	/**< Profile lock */
	LIST_HEAD(, doca_flow_pipe) pipe_list;
	/**< Pipe list */
	LIST_HEAD(, doca_flow_pipe_entry) default_flow_list;
	/**< default flow list */
	struct dpdk_table *tx_miss;
	/**< default flow list */
	struct doca_flow_pipe_entry *tx_encrypt_flow;
	/**< empty tx miss table forwarding to wire */
	struct dpdk_profiles_ids *profile_ids;
	/**< profile ids */
	uint32_t profile_id_generator;
	/**< profile id generator */
	struct id_pool *meter_pool;
	/**< meter id pool */
	struct id_pool *meter_policy_pool;
	/**< meter policy id pool */
	struct flow_ct_context *ct_ctx;
	/**< Connection tracking context */
	uint8_t user_data[0];
	/**< User data memory */
};

#ifdef __cplusplus
}
#endif

#endif /* DOCA_FLOW_PRIV_H_ */
