/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "private/common_defines.h"

#include "rte_flow.h"
#include "dpdk_flow.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_flow);

static struct dpdk_flow_ops flow_ops = {0};

static struct dpdk_flow_queue_item *
queue_item_get(struct dpdk_flow_queue *queue)
{
	struct dpdk_flow_queue_item *queue_item = NULL;
	int rc;

	while (!queue_item) {
		queue_item = TAILQ_FIRST(&queue->free_list);
		if (likely(queue_item)) {
			TAILQ_REMOVE(&queue->free_list, queue_item, next);
			queue->in_progress_items++;
			return queue_item;
		}

		rc = flow_ops.queue_poll(queue, 1);
		if (unlikely(rc < 0))
			break;
	}

	DOCA_LOG_CRIT("failed getting queue item - polling failed with rc=%d", rc);
	return NULL;
}

static void
queue_item_put(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item)
{
	TAILQ_INSERT_HEAD(&queue->free_list, queue_item, next);
	queue->in_progress_items--;
}

int
dpdk_flow_init(struct dpdk_flow_cfg *cfg)
{
	int rc;

	if (unlikely(!cfg)) {
		DOCA_LOG_ERR("failed initializing dpdk flow - cfg is null");
		return -EINVAL;
	}

	if (unlikely(!cfg->ops)) {
		DOCA_LOG_ERR("failed initializing dpdk flow - ops is null");
		return -EINVAL;
	}

	flow_ops = *cfg->ops;
	rc = flow_ops.module_init(cfg);
	if (rc) {
		DOCA_LOG_ERR("failed initializing dpdk flow rc=%d", rc);
		return rc;
	}

	DOCA_LOG_INFO("Initializing dpdk flow successfully");
	return 0;
}

static void
item_completion_cb(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item)
{
	if (likely(queue_item->cb))
		queue_item->cb(queue_item->op, queue_item->status, queue_item->callback_ctx);

	queue_item_put(queue, queue_item);
}

struct dpdk_flow_queue *
dpdk_flow_queue_create(struct dpdk_flow_queue_cfg *queue_cfg)
{
	struct dpdk_flow_queue *queue;
	uint32_t item_idx;
	uint32_t queue_depth;
	int rc;

	queue = calloc(1, sizeof(struct dpdk_flow_queue));
	if (!queue) {
		DOCA_LOG_ERR("failed on queue creation - no mem for queue");
		return NULL;
	}

	queue->port_id = queue_cfg->port_id;
	queue->queue_id = queue_cfg->queue_id;
	queue->queue_depth = queue_cfg->queue_depth;
	queue->comp_cb = item_completion_cb;

	if (queue->queue_depth == 0 || queue->queue_depth > MAX_QUEUE_DEPTH) {
		DOCA_DLOG_WARN("Queue depth %u changed to default %u", queue->queue_depth,
			       DEFAULT_QUEUE_DEPTH);
		queue->queue_depth = DEFAULT_QUEUE_DEPTH;
	}

	queue_depth = queue->queue_depth;
	queue->in_progress_array = calloc(queue_depth, sizeof(struct dpdk_flow_queue_item));
	if (!queue->in_progress_array) {
		DOCA_LOG_ERR("failed on queue creation - no mem for queue in progress array");
		free(queue);
		return NULL;
	}

	rc = flow_ops.on_queue_creation(queue);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed on queue creation - got rc=%d", rc);
		free(queue->in_progress_array);
		free(queue);
		return NULL;
	}

	TAILQ_INIT(&queue->free_list);
	queue->in_progress_items = 0;

	for (item_idx = 0; item_idx < queue_depth; item_idx++) {
		struct dpdk_flow_queue_item *item = &queue->in_progress_array[item_idx];

		TAILQ_INSERT_HEAD(&queue->free_list, item, next);
	}

	DOCA_DLOG_DBG("Created dpdk flow queue: port_id=%u, queue_id=%u, queue_depth=%u",
		      queue->port_id, queue->queue_id, queue->queue_depth);
	return queue;
}

static int
queue_wait_for_depletion(struct dpdk_flow_queue *queue)
{
	uint64_t end_tsc;
	int rc = 0;

	end_tsc = rte_rdtsc() +
			(rte_get_timer_hz() / 1000000) * WAIT_FOR_COMPLETION_TIMEOUT_US;

	while (queue->in_progress_items) {
		flow_ops.queue_poll(queue, queue->queue_depth);
		if (rte_rdtsc() > end_tsc)
			break;
	}

	if (queue->in_progress_items) {
		rc = -ENOTRECOVERABLE;
		DOCA_LOG_RATE_LIMIT_ERR("dpdk queue depletion failed on time %u [us]",
					WAIT_FOR_COMPLETION_TIMEOUT_US);
	}

	return rc;
}

int
dpdk_flow_queue_destroy(struct dpdk_flow_queue *queue, bool wait_for_completions)
{
	int rc = 0;

	if (!queue) {
		DOCA_LOG_WARN("failed destroying null queue");
		return -EINVAL;
	}

	if (wait_for_completions) {
		rc = queue_wait_for_depletion(queue);
		if (unlikely(rc))
			DOCA_LOG_ERR("failed on queue deletion - depletion got rc=%d", rc);
	}

	rc = flow_ops.on_queue_deletion(queue);
	if (unlikely(rc))
		DOCA_LOG_ERR("failed on queue deletion op - got rc=%d", rc);

	DOCA_DLOG_DBG("Destroyed dpdk flow queue: port_id=%u, queue_id=%u, wait=%d", queue->port_id,
		      queue->queue_id, wait_for_completions);

	free(queue->in_progress_array);
	free(queue);
	return rc;
}

int
dpdk_flow_cleanup(void)
{
	DOCA_LOG_INFO("Cleanup dpdk flow");
	if (unlikely(!flow_ops.module_cleanup))
		return 0;
	return flow_ops.module_cleanup();
}

static void
tracker_init(struct dpdk_flow_tracker *tracker)
{
	tracker->flow = NULL;
}

int
dpdk_flow_create(struct dpdk_flow_queue *queue, struct dpdk_flow_params *dpdk_flow_params,
		 struct dpdk_flow_request *req)
{
	struct dpdk_flow_queue_item *queue_item = NULL;
	int rc;

	if (unlikely(!queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating flow - queue is null");
		return -EINVAL;
	}

	if (unlikely(!req)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating flow - request is null");
		return -EINVAL;
	}

	if (unlikely(!req->tracker)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating flow - flow tracker is null");
		return -EINVAL;
	}

	if (unlikely(!dpdk_flow_params)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating flow on port %u - dpdk flow is null",
					queue->port_id);
		return -EINVAL;
	}

	queue_item = queue_item_get(queue);
	if (unlikely(!queue_item))
		return -EAGAIN;

	queue_item->callback_ctx = req->callback_ctx;
	queue_item->cb = req->cb;
	queue_item->op = DPDK_FLOW_REQ_OP_CREATE;
	queue_item->status = DPDK_FLOW_REQ_STATUS_FAIL;

	tracker_init(req->tracker);
	rc = flow_ops.create_flow(queue, queue_item, req->tracker, req->wait_for_bulk,
				  dpdk_flow_params);
	if (unlikely(rc)) {
		queue_item_put(queue, queue_item);
		return rc;
	}

	if (req->persistent)
		rc = queue_wait_for_depletion(queue);

	return rc;
}

int
dpdk_flow_destroy(struct dpdk_flow_queue *queue, struct dpdk_flow_request *req)
{
	struct dpdk_flow_queue_item *queue_item;
	int rc;

	if (unlikely(!queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed destroying flow - queue pointer is null");
		return -EINVAL;
	}

	if (unlikely(!req)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed destroying flow - request is null");
		return -EINVAL;
	}

	if (unlikely(!req->tracker)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed destroying flow - flow tracker is null");
		return -EINVAL;
	}

	queue_item = queue_item_get(queue);
	if (unlikely(!queue_item))
		return -EAGAIN;

	queue_item->callback_ctx = req->callback_ctx;
	queue_item->cb = req->cb;
	queue_item->op = DPDK_FLOW_REQ_OP_DESTROY;
	queue_item->status = DPDK_FLOW_REQ_STATUS_FAIL;

	rc = flow_ops.destroy_flow(queue, queue_item, req->tracker, req->wait_for_bulk);
	if (rc) {
		queue_item_put(queue, queue_item);
		return rc;
	}

	if (req->persistent)
		rc = queue_wait_for_depletion(queue);

	return rc;
}

int
dpdk_flow_poll(struct dpdk_flow_queue *queue, uint16_t max_polled_items)
{
	return flow_ops.queue_poll(queue, max_polled_items);
}

struct rte_flow *
dpdk_flow_tracker_get_flow(struct dpdk_flow_tracker *tracker)
{
	return tracker->flow;
}
