/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DPDK_FLOW_H_
#define DPDK_FLOW_H_

#include <stdint.h>
#include <stdbool.h>
#include <sys/queue.h>

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_QUEUE_DEPTH 1024
#define DEFAULT_QUEUE_DEPTH 128
#define WAIT_FOR_COMPLETION_TIMEOUT_US (1000000)
#define FIRMWARE_MAX_TIMEOUT_US (10000000)

struct rte_flow;
struct rte_flow_item;
struct rte_flow_action;
struct rte_flow_template_table;
struct dpdk_flow_queue;

/**
 * @brief Dpdk flow request operation type
 */
enum dpdk_flow_request_op {
	DPDK_FLOW_REQ_OP_CREATE,
	DPDK_FLOW_REQ_OP_DESTROY,
	DPDK_FLOW_REQ_OP_MAX,
};

/**
 * @brief Dpdk flow request status type
 */
enum dpdk_flow_request_status {
	DPDK_FLOW_REQ_STATUS_OK,
	DPDK_FLOW_REQ_STATUS_FAIL,
	DPDK_FLOW_REQ_STATUS_MAX,
};

/**
 * @brief Dpdk flow callback on completion
 */
typedef void (*dpdk_flow_completion_cb)(enum dpdk_flow_request_op op,
					enum dpdk_flow_request_status status, void *ctx);

/**
 * @brief Dpdk flow queue configuration
 */
struct dpdk_flow_queue_cfg {
	uint16_t port_id;
	/**< port id */
	uint16_t queue_id;
	/**< queue id */
	uint32_t queue_depth;
	/**< queue depth */
};

struct dpdk_flow_queue_item {
	TAILQ_ENTRY(dpdk_flow_queue_item) next;

	void *callback_ctx;
	dpdk_flow_completion_cb cb;

	enum dpdk_flow_request_op op;
	enum dpdk_flow_request_status status;
};

typedef void (*queue_item_completion_cb)(struct dpdk_flow_queue *queue,
					 struct dpdk_flow_queue_item *queue_item);

/**
 * @brief Dpdk flow queue
 */
struct dpdk_flow_queue {
	uint16_t port_id;
	/**< port id */
	uint16_t queue_id;
	/**< queue id */
	uint32_t queue_depth;
	/**< queue depth */

	/* section for internal use only */
	uint32_t in_progress_items;
	/**< in progress items */
	queue_item_completion_cb comp_cb;
	/**< item completion callback */
	struct dpdk_flow_queue_item *in_progress_array;
	/**< queue item preallocated array */
	TAILQ_HEAD(, dpdk_flow_queue_item) free_list;
	/**< queue item free list */
	void *priv_ctx;
	/**< private context per submodule */
};

/**
 * @brief Dpdk flow params
 */
struct dpdk_flow_params {
	union {
		uint32_t index;
		/**< index using instead of items in "insert by index" API  */
		struct rte_flow_item *items;
		/**< items array */
	};
	bool use_index;
	/**< whether @c index or @c items member is active */
	struct rte_flow_action *actions;
	/**< actions array */
	struct rte_flow_template_table *table;
	/**< HWS table */

	uint8_t item_template_index;
	/**< item template index on table */
	uint8_t action_template_index;
	/**< action template index on table */

	bool ingress;
	/**< Ingress selector */
	bool egress;
	/**< egress selector */
	bool transfer;
	/**< transfer selector */
	uint32_t group;
	/**< Group id */
	uint32_t priority;
	/**< Priority */
};

/**
 * @brief Dpdk flow tracker
 */
struct dpdk_flow_tracker {
	struct rte_flow *flow;
	/**< Dpdk flow handler */
};

/**
 * @brief Dpdk flow configuration
 */
struct dpdk_flow_cfg {
	struct dpdk_flow_ops *ops;
	/**< Dpdk flow ops */
};

/**
 * @brief Dpdk flow request
 */
struct dpdk_flow_request {
	struct dpdk_flow_tracker *tracker;
	/**< tracker of queue item state */
	void *callback_ctx;
	/**< context for callback */
	dpdk_flow_completion_cb cb;
	/**< callback to call on completion */
	bool persistent;
	/**< block until this request is completed */
	bool wait_for_bulk;
	/**< wait for more requests */
};

/**
 * @brief Dpdk flow ops
 *
 * A struct to define the ops.
 */
struct dpdk_flow_ops {
	/**
	 * @brief called on module initialization
	 *
	 * @param flow_cfg
	 * Pointer to a dpdk flow configuration
	 * @return
	 * 0 on success, otherwise an error code
	 */
	int (*module_init)(struct dpdk_flow_cfg *flow_cfg);

	/**
	 * @brief called on module teardown
	 *
	 * @return
	 * 0 on success, otherwise an error code
	 */
	int (*module_cleanup)();

	/**
	 * @brief called after a new queue was created.
	 *
	 * @param queue
	 * Pointer to a valid queue.
	 * @return
	 * 0 on success, otherwise an error code
	 */
	int (*on_queue_creation)(struct dpdk_flow_queue *queue);

	/**
	 * @brief Polling the queue
	 *
	 * @param queue
	 * Pointer to a valid queue.
	 * @param max_polled_items
	 * Number of items to poll once before bail.
	 * @return
	 * 0 on success, otherwise an error code
	 */
	int (*queue_poll)(struct dpdk_flow_queue *queue, uint16_t max_polled_items);

	/**
	 * @brief called before a queue is deleted.
	 *
	 * @param queue
	 * Pointer to a valid queue.
	 * @return
	 * 0 on success, otherwise an error code:
	 */
	int (*on_queue_deletion)(struct dpdk_flow_queue *queue);

	/**
	 * @brief create a new flow.
	 *
	 * @param queue
	 * Pointer to a valid queue.
	 * @param queue_item
	 * Pointer to an allocated queue item.
	 * @param tracker
	 * Pointer to a stateful tracker.
	 * @param wait_for_bulk
	 * Best effort to wait for more operations to process in bulks.
	 * @param dpdk_flow_params
	 * Pointer to the flow params required by dpdk API.
	 * @return
	 * 0 on success, otherwise the following error codes:
	 * -EINVAL on invalid params
	 * -ENOTRECOVERABLE when the system is stuck.
	 * -EAGAIN try again later.
	 */
	int (*create_flow)(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item,
			   struct dpdk_flow_tracker *tracker, bool wait_for_bulk,
			   struct dpdk_flow_params *dpdk_flow_params);

	/**
	 * @brief destroy existing flow.
	 *
	 * @param queue
	 * Pointer to a valid queue.
	 * @param queue_item
	 * Pointer to an allocated queue item.
	 * @param tracker
	 * Pointer to a stateful tracker.
	 * @param wait_for_bulk
	 * Best effort to wait for more operations to process in bulks.
	 * @return
	 * 0 on success, otherwise the following error codes:
	 * -EINVAL on invalid params
	 * -ENOTRECOVERABLE when the system is stuck.
	 * -EAGAIN try again later.
	 */
	int (*destroy_flow)(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item,
			    struct dpdk_flow_tracker *tracker, bool wait_for_bulk);
};

/**
 * @brief Initialize dpdk flow module.
 *
 * @param cfg
 * Pointer to configuration
 * @return
 * 0 on success, otherwise error code.
 */
int
dpdk_flow_init(struct dpdk_flow_cfg *cfg);

/**
 * @brief Create queue
 *
 * @param queue_cfg
 * Pointer to queue configuration
 * @return
 * handler on success, otherwise NULL.
 */
struct dpdk_flow_queue *
dpdk_flow_queue_create(struct dpdk_flow_queue_cfg *queue_cfg);

/**
 * @brief Destroy queue
 *
 * @param queue
 * Pointer to queue.
 * @param wait_for_completion
 * If true, polls until queue depletion.
 * @return
 * 0 on success, otherwise error code.
 */
int
dpdk_flow_queue_destroy(struct dpdk_flow_queue *queue, bool wait_for_completions);

/**
 * @brief Create dpdk flow
 * create a flow on queue using request.
 *
 * @param queue
 * Pointer to queue.
 * @param dpdk_flow_params
 * Pointer to flow params.
 * @param req
 * Pointer to request.
 * @return
 * 0 on success, otherwise error code.
 */
int
dpdk_flow_create(struct dpdk_flow_queue *queue, struct dpdk_flow_params *dpdk_flow_params,
		 struct dpdk_flow_request *req);

/**
 * @brief Destroy dpdk flow
 * destroys a created flow.
 *
 * @param queue
 * Pointer to queue.
 * @param req
 * Pointer to request.
 * @return
 * 0 on success, otherwise error code.
 */
int
dpdk_flow_destroy(struct dpdk_flow_queue *queue, struct dpdk_flow_request *req);

/**
 * @brief Poll queue
 * polling the queue for a number of completed items.
 *
 * @param queue
 * Pointer to queue.
 * @param max_polled_items
 * Number of maximum polled items.
 * @return
 * >=0 the number of polled items, otherwise error code.
 */
int
dpdk_flow_poll(struct dpdk_flow_queue *queue, uint16_t max_polled_items);

/**
 * @brief Cleanup the module.
 * free all resources
 *
 * @return
 * 0 on success, otherwise error code.
 */
int
dpdk_flow_cleanup(void);

/**
 * @brief Return flow handle
 *
 * @param tracker
 * pointer to stateful tracker
 * @return
 * pointer to a flow handle.
 */
struct rte_flow *
dpdk_flow_tracker_get_flow(struct dpdk_flow_tracker *tracker);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DPDK_FLOW_H_ */
