/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>

#include "dpdk_flow_hws.h"
#include "dpdk_flow.h"

#include "rte_flow.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_flow_hws);

struct hws_queue {
	bool should_push;
	struct rte_flow_op_result dpdk_res[0];
};

static int
dpdk_flow_hws_init(struct dpdk_flow_cfg *cfg)
{
	DOCA_LOG_DBG("Initialized dpdk flow work module to be HW steering");
	return 0;
}

static int
dpdk_flow_hws_on_queue_creation(struct dpdk_flow_queue *queue)
{
	struct hws_queue *hws_queue;
	uint32_t alloc_len;

	alloc_len =
		sizeof(struct hws_queue) + queue->queue_depth * sizeof(struct rte_flow_op_result);

	hws_queue = calloc(1, alloc_len);
	if (!hws_queue) {
		DOCA_LOG_ERR("failed creating queue - no memory for polling results");
		return -ENOMEM;
	}

	queue->priv_ctx = hws_queue;
	return 0;
}

static int
dpdk_flow_hws_on_queue_deletion(struct dpdk_flow_queue *queue)
{
	if (!queue->priv_ctx)
		return -ENOTRECOVERABLE;

	free(queue->priv_ctx);
	return 0;
}

static int
dpdk_flow_hws_cleanup(void)
{
	DOCA_LOG_DBG("Cleanup dpdk flow HW steering module");
	return 0;
}

static int
dpdk_flow_hws_poll(struct dpdk_flow_queue *queue, uint16_t max_polled_items)
{
	struct dpdk_flow_queue_item *queue_item;
	struct rte_flow_error flow_error;
	struct hws_queue *hws_queue;

	int rc;
	int pulled_items;
	int res_idx;

	if (unlikely(!queue))
		return -EINVAL;

	if (unlikely(!queue->priv_ctx))
		return -ENOTRECOVERABLE;

	if (unlikely(max_polled_items > queue->queue_depth))
		max_polled_items = queue->queue_depth;

	hws_queue = queue->priv_ctx;

	if (unlikely(hws_queue->should_push)) {
		rc = rte_flow_push(queue->port_id, queue->queue_id, &flow_error);
		if (unlikely(rc < 0))
			DOCA_LOG_RATE_LIMIT_ERR(
				"failed to poll queue - push, rc=%d, type %d message: %s", rc,
				flow_error.type,
				flow_error.message ? flow_error.message : "(no stated reason)");

		hws_queue->should_push = false;
	}

	rc = rte_flow_pull(queue->port_id, queue->queue_id, hws_queue->dpdk_res, max_polled_items,
			   &flow_error);
	if (rc < 0) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"failed to poll queue - dpdk poll, rc=%d, type %d message: %s", rc, flow_error.type,
			flow_error.message ? flow_error.message : "(no stated reason)");
		return rc;
	}

	pulled_items = rc;
	for (res_idx = 0; res_idx < pulled_items; res_idx++) {
		struct rte_flow_op_result *res = &hws_queue->dpdk_res[res_idx];
		enum dpdk_flow_request_status status = DPDK_FLOW_REQ_STATUS_OK;

		if (unlikely(res->status != RTE_FLOW_OP_SUCCESS))
			status = DPDK_FLOW_REQ_STATUS_FAIL;

		if (!res->user_data)
			continue;

		queue_item = (struct dpdk_flow_queue_item *)res->user_data;
		queue_item->status = status;
		queue->comp_cb(queue, queue_item);
	}
	return pulled_items;
}

static void
queue_set_should_push(struct dpdk_flow_queue *queue)
{
	struct hws_queue *hws_queue;

	if (unlikely(!queue->priv_ctx))
		return;

	hws_queue = queue->priv_ctx;
	hws_queue->should_push = true;
}

static int
dpdk_flow_hws_create(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item,
		     struct dpdk_flow_tracker *tracker, bool wait_for_bulk,
		     struct dpdk_flow_params *params)
{
	struct rte_flow_error flow_error;
	struct rte_flow_op_attr qatr = {
		.postpone = wait_for_bulk ? 1 : 0,
	};

	if (unlikely(params->use_index))
		tracker->flow = rte_flow_async_create_by_index(queue->port_id, queue->queue_id, &qatr,
							       params->table, params->index, params->actions,
							       params->action_template_index, queue_item,
							       &flow_error);
	else
		tracker->flow = rte_flow_async_create(queue->port_id, queue->queue_id, &qatr, params->table,
						      params->items, params->item_template_index,
						      params->actions, params->action_template_index,
						      queue_item, &flow_error);
	if (unlikely(tracker->flow == NULL)) {
		DOCA_LOG_RATE_LIMIT_ERR("Port %u create flow failed, type %d message: %s",
					queue->port_id, flow_error.type,
					flow_error.message ? flow_error.message
							   : "(no stated reason)");
		return -EINVAL;
	}

	if (wait_for_bulk)
		queue_set_should_push(queue);

	queue_item->status = DPDK_FLOW_REQ_STATUS_OK;
	return 0;
}

static int
dpdk_flow_hws_destroy(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item,
		      struct dpdk_flow_tracker *tracker, bool wait_for_bulk)
{
	int rc;
	struct rte_flow_error flow_error;
	struct rte_flow_op_attr qatr = {
		.postpone = !!wait_for_bulk,
	};

	if (!tracker->flow)
		return -EINVAL;

	if (wait_for_bulk)
		queue_set_should_push(queue);

	rc = rte_flow_async_destroy(queue->port_id, queue->queue_id, &qatr, tracker->flow,
				    queue_item, &flow_error);
	if (unlikely(rc)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed destroying flow - port_id=%u queue_id=%u: rc=%d",
					queue->port_id, queue->queue_id, rc);
		return -EINVAL;
	}

	if (wait_for_bulk)
		queue_set_should_push(queue);

	queue_item->status = DPDK_FLOW_REQ_STATUS_OK;
	return 0;
}

static struct dpdk_flow_ops hw_steering_ops = {
	.module_init = dpdk_flow_hws_init,
	.module_cleanup = dpdk_flow_hws_cleanup,

	.on_queue_creation = dpdk_flow_hws_on_queue_creation,
	.queue_poll = dpdk_flow_hws_poll,
	.on_queue_deletion = dpdk_flow_hws_on_queue_deletion,

	.create_flow = dpdk_flow_hws_create,
	.destroy_flow = dpdk_flow_hws_destroy,
};

struct dpdk_flow_ops *
dpdk_flow_hws_get_ops(void)
{
	return &hw_steering_ops;
}
