/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <stdlib.h>
#include <errno.h>


#include "dpdk_port.h"
#include "dpdk_pipe.h"
#include "dpdk_flow_single.h"

#include <rte_flow.h>
#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_flow_single);


static void
rm_flow_single_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			     void *ctx)
{
	struct dpdk_flow_single *single = (struct dpdk_flow_single *)ctx;

	if (unlikely(!single)) {
		DOCA_LOG_CRIT("flow_single deletion completion got null flow single");
		return;
	}

	single->mark_for_delete = true;
}

static int
single_remove_async(struct dpdk_flow_single *single,
		    dpdk_flow_completion_cb cb,
		    bool persistent, bool wait_for_bulk)
{
	struct dpdk_flow_request req = {0};
	int rc;

	req.tracker = &single->tracker;
	req.cb = cb;
	req.callback_ctx = single;
	req.persistent = persistent;
	req.wait_for_bulk = wait_for_bulk;

	rte_spinlock_lock(single->lock);
	rc = dpdk_flow_destroy(single->queue, &req);
	rte_spinlock_unlock(single->lock);
	if (unlikely(rc))
		return rc;
	return 0;
}

static void
add_flow_single_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			      void *ctx)
{
	struct dpdk_flow_single *single = (struct dpdk_flow_single *)ctx;

	if (unlikely(!single)) {
		DOCA_LOG_CRIT("flow_single addition completion got null flow single");
		return;
	}

	DOCA_DLOG_DBG("flow_single %p add completed with status %d", single, status);

	if (status != DPDK_FLOW_REQ_STATUS_OK) {
		single_remove_async(single, rm_flow_single_completion_cb, true, false);
		return;
	}

	single->in_hw = true;
}

static int
insert_rule(struct dpdk_table_port_ctx *table_ctx,
	    struct dpdk_flow_queue *queue,
	    struct dpdk_flow_single *single,
	    struct rte_flow_item *items,
	    struct rte_flow_action *actions,
	    struct rte_flow_attr *attr,
	    rte_spinlock_t *lock)
{
	struct dpdk_flow_request req = {0};
	struct dpdk_flow_params dpdk_flow_params = {0};
	int ret;

	dpdk_flow_params.items = items;
	dpdk_flow_params.actions = actions;
	dpdk_flow_params.table = single->dpdk_table.table;
	dpdk_flow_params.item_template_index = single->dpdk_table.pattern_template_index;
	dpdk_flow_params.action_template_index = single->dpdk_table.action_template_index;

	dpdk_flow_params.ingress = attr->ingress;
	dpdk_flow_params.egress = attr->egress;
	dpdk_flow_params.transfer = attr->transfer;
	dpdk_flow_params.group = attr->group;
	dpdk_flow_params.priority = attr->priority;

	req.tracker = &single->tracker;
	req.cb = add_flow_single_completion_cb;
	req.callback_ctx = single;
	req.persistent = true;
	req.wait_for_bulk = false;

	single->queue = queue;
	single->lock = lock;
	rte_spinlock_lock(lock);
	ret = dpdk_flow_create(queue, &dpdk_flow_params, &req);
	rte_spinlock_unlock(lock);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed insertion fwd_groups rules - ret=%d", ret);
		return ret;
	}

	return 0;
}

static void
fill_dpdk_table_params(struct dpdk_table_params *table_params,
		       uint16_t port_id, struct rte_flow_item *items,
		       struct rte_flow_action **actions, struct rte_flow_action **actions_mask,
		       uint8_t nb_actions, struct rte_flow_attr *attr, uint32_t nb_flows)
{
	table_params->port_id = port_id;
	table_params->items = items;
	table_params->actions = actions;
	table_params->actions_mask = actions_mask;
	table_params->nb_actions = nb_actions;
	table_params->flow_attr = attr;
	table_params->nb_flows = nb_flows;
}

static int
build_table(uint32_t port_id,
	    struct dpdk_table_port_ctx *port_ctx,
	    struct rte_flow_attr *attr,
	    struct rte_flow_item *items,
	    struct rte_flow_action *actions,
	    struct rte_flow_action *actions_masks,
	    struct dpdk_table *dpdk_table,
	    uint8_t nb_actions,
	    uint32_t nb_flows)
{
	struct dpdk_table_params table_params = {0};
	int ret;

	fill_dpdk_table_params(&table_params, port_id, items, &actions,
			       &actions_masks, nb_actions, attr, nb_flows);
	ret = dpdk_table_create(&table_params, port_ctx, dpdk_table);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building dpdk table ret=%d", ret);
		return ret;
	}

	return 0;
}

static void
fill_attr_params(struct rte_flow_attr *attr, struct dpdk_flow_params *params)
{
	attr->ingress = params->ingress;
	attr->egress = params->egress;
	attr->group = params->group;
	attr->priority = params->priority;
	attr->transfer = params->transfer;
}

static bool
flow_single_is_valid(struct dpdk_port *port,
		     struct dpdk_flow_queue *queue,
		     rte_spinlock_t *queue_lock,
		     struct dpdk_flow_single *flow_single)
{
	if (!port) {
		DOCA_LOG_RATE_LIMIT_WARN("Invalid flow_single data - NULL port");
		return false;
	}

	if (!queue) {
		DOCA_LOG_RATE_LIMIT_WARN("Invalid flow_single data - NULL queue");
		return false;
	}

	if (!queue_lock) {
		DOCA_LOG_RATE_LIMIT_WARN("Invalid flow_single data - NULL queue_lock");
		return false;
	}

	if (!flow_single) {
		DOCA_LOG_RATE_LIMIT_WARN("Invalid flow_single data - NULL flow_single");
		return false;
	}

	return true;
}

static bool
flow_single_create_validate(struct dpdk_port *port,
			     struct dpdk_flow_queue *queue,
			     rte_spinlock_t *queue_lock,
			     struct dpdk_flow_single *flow_single)
{
	return flow_single_is_valid(port, queue, queue_lock, flow_single);
}

static bool
flow_single_destroy_validate(struct dpdk_port *port,
			     struct dpdk_flow_queue *queue,
			     rte_spinlock_t *queue_lock,
			     struct dpdk_flow_single *flow_single)
{
	if (!flow_single_is_valid(port, queue, queue_lock, flow_single))
		return false;

	if (!flow_single->in_hw) {
		DOCA_LOG_RATE_LIMIT_WARN("Invalid flow_single data - flow_single not in_hw");
		return false;
	}

	return true;
}

int
dpdk_flow_single_create(struct dpdk_port *port,
			struct dpdk_flow_params *dpdk_flow_params,
			struct dpdk_flow_queue *queue,
			rte_spinlock_t *queue_lock,
			uint8_t nb_actions,
			uint32_t nb_flows,
			struct dpdk_flow_single *flow_single)
{
	struct rte_flow_attr attr = {0};
	struct dpdk_table_port_ctx *table_ctx;
	uint16_t port_id;
	int ret;

	if (!flow_single_create_validate(port, queue, queue_lock, flow_single)) {
		DOCA_LOG_RATE_LIMIT_WARN("Failed creating flow_single data - params not valid");
		return -EINVAL;
	}

	table_ctx = dpdk_port_get_table_ctx(port);
	if (unlikely(!table_ctx)) {
		DOCA_LOG_RATE_LIMIT_WARN("failed freeing flow_single data - NULL table_ctx");
		return -EINVAL;
	}

	port_id = dpdk_port_get_id(port);
	fill_attr_params(&attr, dpdk_flow_params);
	ret = build_table(port_id, table_ctx, &attr, dpdk_flow_params->items, dpdk_flow_params->actions,
			  dpdk_flow_params->actions, &flow_single->dpdk_table, nb_actions, nb_flows);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation flow_single - build dpdk table failure ret %d", ret);
		return ret;
	}

	ret = insert_rule(table_ctx, queue, flow_single, dpdk_flow_params->items, dpdk_flow_params->actions, &attr, queue_lock);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation flow_single - insert rule failure ret=%d", ret);
		goto error;
	}

	return 0;
error:
	ret = dpdk_table_destroy(table_ctx, &flow_single->dpdk_table);
	if (ret < 0)
		DOCA_LOG_CRIT("failed destorying table ret %d", ret);

	return -ENOTRECOVERABLE;
}

void
dpdk_flow_single_destroy(struct dpdk_port *port,
			 struct dpdk_flow_queue *queue,
			 rte_spinlock_t *queue_lock,
			 struct dpdk_flow_single *single)
{
	struct dpdk_table_port_ctx *table_ctx;
	int ret;

	if (!flow_single_destroy_validate(port, queue, queue_lock, single)) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed freeing flow_single data - params not valid");
		return;
	}

	single->lock = queue_lock;
	ret = single_remove_async(single, rm_flow_single_completion_cb, true, false);
	if (ret)
		DOCA_LOG_RATE_LIMIT_ERR("failed freeing flow_single data - flow destroy failure ret=%d", ret);

	table_ctx = dpdk_port_get_table_ctx(port);
	if (unlikely(!table_ctx)) {
		DOCA_LOG_RATE_LIMIT_WARN("failed freeing flow_single data - NULL table_ctx");
		return;
	}

	ret = dpdk_table_destroy(table_ctx, &single->dpdk_table);
	if (ret < 0)
		DOCA_LOG_RATE_LIMIT_ERR("failed freeing flow_single data - table destroy failure ret=%d", ret);
}
