/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>

#include "dpdk_flow_sws.h"
#include "dpdk_flow.h"

#include "rte_flow.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_flow_sws);

struct sws_queue {
	TAILQ_HEAD(, dpdk_flow_queue_item) used_in_queue;
};

static int
dpdk_flow_sws_init(struct dpdk_flow_cfg *cfg)
{
	DOCA_LOG_DBG("Initialized dpdk flow work module to be SW steering");
	return 0;
}

static int
dpdk_flow_sws_on_queue_creation(struct dpdk_flow_queue *queue)
{
	struct sws_queue *sws_queue;

	sws_queue = calloc(1, sizeof(struct sws_queue));
	if (!sws_queue)
		return -ENOMEM;


	TAILQ_INIT(&sws_queue->used_in_queue);

	queue->priv_ctx = sws_queue;
	return 0;
}

static int
queue_poll(struct dpdk_flow_queue *queue, uint16_t max_polled_items)
{
	struct dpdk_flow_queue_item *queue_item;
	struct sws_queue *sws_queue;
	uint16_t polled_items = 0;

	sws_queue = queue->priv_ctx;

	do {
		queue_item = TAILQ_FIRST(&sws_queue->used_in_queue);
		if (!queue_item)
			break;

		TAILQ_REMOVE(&sws_queue->used_in_queue, queue_item, next);
		queue->comp_cb(queue, queue_item);
		polled_items++;
	} while (polled_items < max_polled_items);

	return polled_items;
}

static int
dpdk_flow_sws_on_queue_deletion(struct dpdk_flow_queue *queue)
{
	int rc;

	if (unlikely(!queue->priv_ctx))
		return -ENOTRECOVERABLE;

	rc = queue_poll(queue, queue->queue_depth);

	free(queue->priv_ctx);
	return rc;
}

static int
dpdk_flow_sws_cleanup(void)
{
	DOCA_LOG_DBG("Cleanup dpdk flow SW steering module");
	return 0;
}

static int
dpdk_flow_sws_create(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item,
		     struct dpdk_flow_tracker *tracker, bool wait_for_bulk,
		     struct dpdk_flow_params *dpdk_flow_params)
{
	struct rte_flow_error flow_error;
	struct rte_flow_attr attr = {0};
	struct rte_flow *flow;
	struct sws_queue *sws_queue;

	attr.group = dpdk_flow_params->group;
	attr.ingress = dpdk_flow_params->ingress;
	attr.egress = dpdk_flow_params->egress;
	attr.transfer = !!dpdk_flow_params->transfer;
	attr.priority = dpdk_flow_params->priority;

	flow = rte_flow_create(queue->port_id, &attr, dpdk_flow_params->items,
			       dpdk_flow_params->actions, &flow_error);
	if (unlikely(!flow)) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"failed creating flow on port %u - creation error type %d message: %s",
			queue->port_id, flow_error.type,
			flow_error.message ? flow_error.message : "(no stated reason)");
		queue_item->status = DPDK_FLOW_REQ_STATUS_FAIL;
	} else
		queue_item->status = DPDK_FLOW_REQ_STATUS_OK;

	tracker->flow = flow;
	sws_queue = queue->priv_ctx;
	TAILQ_INSERT_TAIL(&sws_queue->used_in_queue, queue_item, next);
	return 0;
}

static int
dpdk_flow_sws_destroy(struct dpdk_flow_queue *queue, struct dpdk_flow_queue_item *queue_item,
		      struct dpdk_flow_tracker *tracker, bool wait_for_bulk)
{
	struct rte_flow_error flow_error;
	struct sws_queue *sws_queue;
	int ret;

	if (unlikely(!tracker->flow)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed destroying flow on port %u - flow is null",
					queue->port_id);
		queue_item->status = DPDK_FLOW_REQ_STATUS_FAIL;
		return 0;
	}

	ret = rte_flow_destroy(queue->port_id, tracker->flow, &flow_error);
	if (unlikely(ret)) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"failed destroying flow on port %u - destroy error type %d message: %s",
			queue->port_id, flow_error.type,
			flow_error.message ? flow_error.message : "(no stated reason)");

		queue_item->status = DPDK_FLOW_REQ_STATUS_FAIL;
	} else
		queue_item->status = DPDK_FLOW_REQ_STATUS_OK;

	sws_queue = queue->priv_ctx;
	TAILQ_INSERT_TAIL(&sws_queue->used_in_queue, queue_item, next);
	return 0;
}

static int
dpdk_flow_sws_poll(struct dpdk_flow_queue *queue, uint16_t max_polled_items)
{
	if (unlikely(!queue->priv_ctx))
		return -ENOTRECOVERABLE;

	return queue_poll(queue, max_polled_items);

}

static struct dpdk_flow_ops sw_steering_ops = {
	.module_init = dpdk_flow_sws_init,
	.module_cleanup = dpdk_flow_sws_cleanup,

	.on_queue_creation = dpdk_flow_sws_on_queue_creation,
	.queue_poll = dpdk_flow_sws_poll,
	.on_queue_deletion = dpdk_flow_sws_on_queue_deletion,

	.create_flow = dpdk_flow_sws_create,
	.destroy_flow = dpdk_flow_sws_destroy
};

struct dpdk_flow_ops *
dpdk_flow_sws_get_ops(void)
{
	return &sw_steering_ops;
}
