/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <stdlib.h>
#include <errno.h>

#include "dpdk_group_pool.h"
#include "dpdk_port.h"
#include "dpdk_pipe.h"
#include "dpdk_table.h"
#include "dpdk_flow.h"
#include "dpdk_shared_rss.h"
#include "dpdk_fwd_groups.h"
#include "dpdk_flow_single.h"

#include <rte_flow.h>
#include <rte_hash_crc.h>
#include <utils_hash_table.h>

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_fwd_groups);

#define FWD_GROUPS_NAME_BUF_SIZE (32)
#define MAX_FWD_GROUPS (2048)
#define FWD_GROUPS_ACT_TEMPLATE_IDX (0)
#define FWD_GROUPS_ITEM_TEMPLATE_IDX (0)
#define FWD_GROUPS_NB_ACTIONS (1)
#define FWD_GROUPS_NB_FLOWS (1)

struct fwd_single {
	struct dpdk_fwd_groups_cfg fwd_groups_cfg;
	struct dpdk_flow_single flow_single;
};

struct dpdk_fwd_groups {
	struct dpdk_fwd_groups_cfg cfg;
	struct utils_hash_table *fwd_groups_mgr;
};

struct group_item_action {
	struct rte_flow_item_ethdev ethdev;
	union {
		struct rte_flow_action_jump jump;
		struct rte_flow_action_ethdev port;
		struct rte_flow_action_queue queue;
		struct rte_flow_action_rss rss;
	};
};

int
dpdk_fwd_groups_cfg_get_by_id(uint16_t port_id, void **mgmt_queue, void **mgmt_lock)
{
	struct dpdk_fwd_groups *fwd_groups;
	struct dpdk_port *port;

	port = dpdk_port_get_by_id(port_id);
	if (unlikely(!port))
		return -EINVAL;

	fwd_groups = dpdk_port_fwd_groups_get(port);
	if (unlikely(!fwd_groups))
		return -EINVAL;

	*mgmt_queue = fwd_groups->cfg.management_queue;
	*mgmt_lock = fwd_groups->cfg.management_lock;

	return 0;
}

static inline void
format_fwd_groups_name(char *name, uint16_t port_id)
{
	snprintf(name, FWD_GROUPS_NAME_BUF_SIZE, "grp_fwd port %"PRIu16"", port_id);
}

static inline uint32_t
fwd_groups_hash_crc(const void *data, __rte_unused uint32_t data_len, uint32_t init_val)
{
	const struct dpdk_fwd_groups_key *k;

	k = data;
	init_val = rte_hash_crc(&k->port_id, sizeof(k->port_id), init_val);
	init_val = rte_hash_crc(&k->type, sizeof(k->type), init_val);

	switch (k->type) {
	case DPDK_FWD_RSS:
		init_val = rte_hash_crc(&k->rss.shared_rss_id, sizeof(k->rss.shared_rss_id), init_val);
		break;
	case DPDK_FWD_QUEUE:
		init_val = rte_hash_crc(&k->queue.index, sizeof(k->queue.index), init_val);
		break;
	case DPDK_FWD_PORT:
		init_val = rte_hash_crc(&k->port.id, sizeof(k->port.id), init_val);
		break;
	default:
		break;
	}

	return init_val;
}

static inline int
fwd_groups_cmp_key(const void *key1, const void *key2, size_t key_len)
{
	const struct dpdk_fwd_groups_key *src, *dst;

	src = key1;
	dst = key2;
	if ((src->port_id != dst->port_id) || (src->type != dst->type))
		return -EINVAL;

	switch (src->type) {
	case DPDK_FWD_RSS:
		if (src->rss.shared_rss_id != dst->rss.shared_rss_id)
			return -EINVAL;
		break;
	case DPDK_FWD_QUEUE:
		if (src->queue.index != dst->queue.index)
			return -EINVAL;
		break;
	case DPDK_FWD_PORT:
		if (src->port.id != dst->port.id)
			return -EINVAL;
		break;
	case DPDK_FWD_DROP:
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static struct fwd_single *
fwd_single_alloc()
{
	struct fwd_single *single;

	single = calloc(1, sizeof(struct fwd_single));
	return single;
}
static void
fwd_single_free(struct fwd_single *single)
{
	if (single)
		free(single);
}

static struct utils_hash_table *
fwd_groups_mgr_ins_get(uint16_t port_id)
{
	struct dpdk_port *port;
	struct dpdk_fwd_groups *fwd_groups;

	port = dpdk_port_get_by_id(port_id);
	if (!port) {
		DOCA_LOG_RATE_LIMIT_ERR("failure to get fwd_groups_mgr - invalid port_id %u", port_id);
		return NULL;
	}

	fwd_groups = dpdk_port_fwd_groups_get(port);
	if (!fwd_groups) {
		DOCA_LOG_RATE_LIMIT_ERR("failure to get fwd_groups_mgr - null fwd_groups, port_id %u", port_id);
		return NULL;
	}

	return fwd_groups->fwd_groups_mgr;
}

void
fwd_groups_free_key(void *key)
{
	struct fwd_single *data;
	struct dpdk_flow_queue *queue;
	struct utils_hash_table *fwd_groups_mgr;
	rte_spinlock_t *lock;
	uint16_t port_id;
	int ret;

	if (unlikely(!key)) {
		DOCA_LOG_RATE_LIMIT_WARN("failed freeing fwd_group data - null key");
		return;
	}

	port_id = ((struct dpdk_fwd_groups_key *)key)->port_id;
	fwd_groups_mgr = fwd_groups_mgr_ins_get(port_id);
	if (unlikely(!fwd_groups_mgr)) {
		DOCA_LOG_RATE_LIMIT_WARN("failed freeing fwd_group data - null fwd_groups_mgr");
		return;
	}

	ret = utils_hash_table_get_value_by_key(fwd_groups_mgr, key, (void **)&data);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_WARN("failed freeing fwd_group data - data not found");
		return;
	}

	queue = data->fwd_groups_cfg.management_queue;
	lock = data->fwd_groups_cfg.management_lock;
	if (unlikely(!queue) || unlikely(!lock)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed freeing fwd_group data - queue/table_ctx/lock is NULL");
		return;
	}

	dpdk_flow_single_destroy(data->fwd_groups_cfg.port, queue, lock, &data->flow_single);
	fwd_single_free(data);
	data = NULL;
}

static struct utils_hash_table *
fwd_groups_mgr_init(struct dpdk_port *port)
{
	struct utils_hash_table_cfg cfg = {0};
	struct utils_hash_table *fwd_groups_mgr;
	char name[FWD_GROUPS_NAME_BUF_SIZE];
	uint16_t port_id;

	port_id = dpdk_port_get_id(port);
	format_fwd_groups_name(name, port_id);
	cfg.table_size = MAX_FWD_GROUPS;
	cfg.key_len = sizeof(struct dpdk_fwd_groups_key);
	cfg.name = name;
	cfg.hash_func = &fwd_groups_hash_crc;
	cfg.free_key_cb = &fwd_groups_free_key;
	cfg.cmp_eq = &fwd_groups_cmp_key;
	cfg.thread_safe = true;

	fwd_groups_mgr = utils_hash_table_create(&cfg);
	if (fwd_groups_mgr == NULL) {
		DOCA_LOG_ERR("failed to allocate dpdk_fwd_groups_values table for port %"PRIu16"", port_id);
		return NULL;
	}

	return fwd_groups_mgr;
}

static int
fwd_groups_mgr_get_data(struct dpdk_fwd_groups_key *key,
			struct utils_hash_table *fwd_groups_mgr,
			uint32_t *group_id)
{
	int ret;
	struct fwd_single *data;

	ret = utils_hash_table_get_value_by_key(fwd_groups_mgr, key, (void **)&data);
	if (ret)
		return ret;

	*group_id = data->flow_single.group_id;
	return 0;
}

static int
fwd_groups_mgr_put_key_data(struct utils_hash_table *fwd_groups_mgr,
			    struct dpdk_fwd_groups_key *key,
			    struct fwd_single *data)
{
	int ret;

	ret = utils_hash_table_key_set_value(fwd_groups_mgr, key, data);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed putting group_id %u in manager", data->flow_single.group_id);
		return ret;
	}

	return 0;
}

static bool
fwd_groups_create_validate(struct dpdk_fwd_groups_cfg *cfg)
{
	if (unlikely(!cfg)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed fwd_groups_create validation - null fwd_groups_cfg");
		return false;
	}

	if (unlikely(!cfg->port)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed fwd_groups_create validation - null port");
		return false;
	}

	if (unlikely(!cfg->management_queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed fwd_groups_create validation - null management_queue");
		return false;
	}

	if (unlikely(!cfg->management_lock)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed fwd_groups_create validation - null management_lock");
		return false;
	}

	return true;
}

struct dpdk_fwd_groups *
dpdk_fwd_groups_create(struct dpdk_fwd_groups_cfg *cfg)
{
	struct dpdk_fwd_groups *fwd_groups;
	struct utils_hash_table *fwd_groups_mgr;

	if (!fwd_groups_create_validate(cfg))
		return NULL;

	fwd_groups = calloc(1, sizeof(struct dpdk_fwd_groups));
	if (unlikely(!fwd_groups)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation groups_fwd - out of memory");
		return NULL;
	}

	fwd_groups_mgr = fwd_groups_mgr_init(cfg->port);
	if (unlikely(!fwd_groups_mgr)) {
		free(fwd_groups);
		DOCA_LOG_RATE_LIMIT_ERR("failed creation groups_fwd - mgr init failed");
		return NULL;
	}

	fwd_groups->fwd_groups_mgr = fwd_groups_mgr;
	fwd_groups->cfg = *cfg;

	return fwd_groups;
}

static int
fwd_groups_get_validate(struct dpdk_fwd_groups_req *req, struct dpdk_fwd_groups *fwd_groups, uint32_t *group_id)
{
	if (unlikely(!req)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed validating req - null req");
		return -EINVAL;
	}

	if (unlikely(!fwd_groups)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed validating req - null fwd groups");
		return -EINVAL;
	}

	if (unlikely(!fwd_groups->fwd_groups_mgr)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed validating req - null fwd groups manager");
		return -EINVAL;
	}

	if (unlikely(!group_id)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed validating req - null group_id");
		return -EINVAL;
	}

	return 0;
}

static void
build_items(struct rte_flow_item *items)
{
	int nb_count = 0;

	items[nb_count++].type = RTE_FLOW_ITEM_TYPE_ETH;
	items[nb_count].type = RTE_FLOW_ITEM_TYPE_END;
}

static int
build_actions(struct dpdk_fwd_groups_req *cfg, struct rte_flow_action *actions, struct group_item_action *para)
{
	int nb_count = 0;

	switch (cfg->key.type) {
	case DPDK_FWD_JUMP:
		para->jump.group = cfg->key.jump.group;
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_JUMP;
		actions[nb_count++].conf = &para->jump;
		break;
	case DPDK_FWD_QUEUE:
		para->queue.index = cfg->key.queue.index;
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_QUEUE;
		actions[nb_count++].conf = &para->queue;
		break;
	case DPDK_FWD_PORT:
		para->port.port_id = cfg->key.port.id;
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT;
		actions[nb_count++].conf = &para->port;
		break;
	case DPDK_FWD_DROP:
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_DROP;
		actions[nb_count++].conf = NULL;
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("failed building fwd actions - invalid type %d", cfg->key.type);
		return -EINVAL;
	};
	actions[nb_count].type = RTE_FLOW_ACTION_TYPE_END;

	return 0;
}

static void
fill_dpdk_flow_params(struct dpdk_fwd_groups_req *req, struct rte_flow_item *items,
		      struct rte_flow_action *actions, struct dpdk_flow_params *params)
{
	params->items = items;
	params->actions = actions;
	params->table = NULL;

	params->action_template_index = FWD_GROUPS_ACT_TEMPLATE_IDX;
	params->item_template_index = FWD_GROUPS_ITEM_TEMPLATE_IDX;

	params->ingress = req->ingress;
	params->egress = req->egress;
	params->transfer = req->transfer;
	params->priority = req->priority;
	params->group = req->group;
}

int
get_fwd_groups(struct dpdk_fwd_groups_req *req, struct dpdk_fwd_groups *fwd_groups, uint32_t *group_id)
{
	struct fwd_single *single;
	struct dpdk_flow_params dpdk_flow_params = {0};
	struct rte_flow_item items[DPDK_PIPE_MAX_ITEMS] = {0};
	struct rte_flow_action actions[DPDK_PIPE_MAX_ACTIONS] = {0};
	struct dpdk_group_pool *group_pool;
	struct group_item_action group_item_action = {0};
	int ret;

	/* If the req is handled, return group_id. */
	ret = fwd_groups_mgr_get_data(&req->key, fwd_groups->fwd_groups_mgr, group_id);
	if (!ret)
		return 0;

	/* Otherwise alloc group_id and build table/flow */
	DOCA_LOG_RATE_LIMIT_DBG("create dpdk_fwd_group flow, type %d", req->key.type);
	single = fwd_single_alloc();
	if (unlikely(!single)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation dpdk_fwd_group - out of memory");
		return -ENOMEM;
	}

	group_pool = dpdk_port_get_group_pool(fwd_groups->cfg.port);
	if (unlikely(!group_pool)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation dpdk_fwd_group - null group_pool");
		goto error;
	}

	ret = dpdk_group_pool_alloc(group_pool, group_id);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation dpdk_fwd_group - get group_id failed");
		goto error;
	}

	/* create flow */
	single->flow_single.group_id = *group_id;
	req->group = *group_id;
	build_items(items);
	build_actions(req, actions, &group_item_action);
	fill_dpdk_flow_params(req, items, actions, &dpdk_flow_params);
	ret = dpdk_flow_single_create(fwd_groups->cfg.port, &dpdk_flow_params,
				      fwd_groups->cfg.management_queue,
				      fwd_groups->cfg.management_lock, FWD_GROUPS_NB_ACTIONS, FWD_GROUPS_NB_FLOWS,
				      &single->flow_single);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation dpdk_fwd_group - failed to create flow single");
		ret = dpdk_group_pool_free(group_pool, *group_id);
		if (ret)
			DOCA_LOG_RATE_LIMIT_ERR("failed creation dpdk_fwd_group - free group_id %u failure", *group_id);
		goto error;
	}

	/* Store key and data(table/flow_tracker) in hash table*/
	single->fwd_groups_cfg = fwd_groups->cfg;
	fwd_groups_mgr_put_key_data(fwd_groups->fwd_groups_mgr, &req->key, single);

	return 0;
error:
	fwd_single_free(single);
	return -EINVAL;
}

int
dpdk_fwd_groups_get(struct dpdk_fwd_groups *fwd_groups, struct dpdk_fwd_groups_req *req, uint32_t *group_id)
{
	int ret;

	ret = fwd_groups_get_validate(req, fwd_groups, group_id);
	if (ret)
		return ret;

	switch (req->key.type) {
	case DPDK_FWD_RSS:
		/* shared_rss is globally bound and ignore checking the bound state.*/
		return dpdk_shared_rss_get_group(req->key.rss.shared_rss_id, group_id);
	case DPDK_FWD_JUMP:
		*group_id = req->key.jump.group;
		return 0;
	case DPDK_FWD_PORT:
	case DPDK_FWD_QUEUE:
	case DPDK_FWD_DROP:
		return get_fwd_groups(req, fwd_groups, group_id);
	default:
		return -EINVAL;
	}

	return 0;
}

void
dpdk_fwd_groups_destroy(struct dpdk_fwd_groups *fwd_groups)
{
	if (fwd_groups) {
		utils_hash_table_destroy(fwd_groups->fwd_groups_mgr);
		free(fwd_groups);
	}
}
