/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <stdlib.h>
#include <errno.h>
#include <sys/queue.h>

#include "private/common_defines.h"
#include "dpdk_group_pool.h"
#include "engine_spinlock.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_group_pool);

#define MAX_GROUP_ID (UINT32_MAX)

struct dpdk_group_pool_item {
	LIST_ENTRY(dpdk_group_pool_item) next;
	uint32_t group_id;
};

struct dpdk_group_pool {
	LIST_HEAD(, dpdk_group_pool_item) free_list;
	uint32_t min_group_id;
	uint32_t cur_group_id;
	uint32_t in_use;
	struct engine_spinlock lock;
};

static int
min_group_id_get(struct dpdk_group_pool *group_pool)
{
	return group_pool->min_group_id;
}

struct dpdk_group_pool *
dpdk_group_pool_create(struct dpdk_group_pool_cfg *cfg)
{
	struct dpdk_group_pool *group_pool;
	int size;

	if (unlikely(!cfg)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating group pool - cfg null");
		return NULL;
	}

	size = sizeof(struct dpdk_group_pool);
	group_pool = calloc(1, size);
	if (unlikely(!group_pool)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating group pool of size %d - no memory", size);
		return NULL;
	}
	engine_spinlock_init(&group_pool->lock);
	LIST_INIT(&group_pool->free_list);
	group_pool->cur_group_id = cfg->min_group_id;
	group_pool->min_group_id = cfg->min_group_id;

	return group_pool;
}

static void
group_pool_item_release(struct dpdk_group_pool_item *item)
{
	LIST_REMOVE(item, next);
	free(item);
}

void
dpdk_group_pool_destroy(struct dpdk_group_pool *group_pool)
{
	struct dpdk_group_pool_item *item;

	if (unlikely(!group_pool)) {
		DOCA_LOG_RATE_LIMIT_INFO("failed destroying null group pool");
		return;
	}
	if (unlikely(group_pool->in_use))
		DOCA_LOG_WARN("group_pool has %u used groups",
			group_pool->in_use);

	engine_spinlock_lock(&group_pool->lock);
	while ((item = LIST_FIRST(&group_pool->free_list)))
		group_pool_item_release(item);
	engine_spinlock_unlock(&group_pool->lock);
	free(group_pool);
}

int
dpdk_group_pool_alloc(struct dpdk_group_pool *group_pool, uint32_t *group_id)
{
	struct dpdk_group_pool_item *item;

	if (unlikely(!group_pool)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed allocating group - null group_pool");
		return -EINVAL;
	}

	engine_spinlock_lock(&group_pool->lock);
	item = LIST_FIRST(&group_pool->free_list);
	if (item) {
		*group_id = item->group_id;
		group_pool_item_release(item);
	} else {
		if (group_pool->cur_group_id == MAX_GROUP_ID) {
			engine_spinlock_unlock(&group_pool->lock);
			DOCA_LOG_RATE_LIMIT_ERR("failed allocating group - reach max group id");
			return -ERANGE;
		}
		*group_id = group_pool->cur_group_id++;
	}
	group_pool->in_use++;
	engine_spinlock_unlock(&group_pool->lock);

	return 0;
}

int
dpdk_group_pool_free(struct dpdk_group_pool *group_pool, uint32_t group_id)
{
	struct dpdk_group_pool_item *item;
	uint32_t min_group_id;

	if (unlikely(!group_pool)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed freeing group_id %u - null group_pool", group_id);
		return -EINVAL;
	}

	min_group_id = min_group_id_get(group_pool);
	if (group_id < min_group_id) {
		DOCA_LOG_RATE_LIMIT_ERR("failed freeing group_id %u - lower than min group_id %u",
			group_id, min_group_id);
		return -ERANGE;
	}

	item = calloc(1, sizeof(*item));
	if (unlikely(!item)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed freeing group_id %u - add to free_list failed.", group_id);
		return -EINVAL;
	}

	item->group_id = group_id;
	engine_spinlock_lock(&group_pool->lock);
	LIST_INSERT_HEAD(&group_pool->free_list, item, next);
	group_pool->in_use--;
	engine_spinlock_unlock(&group_pool->lock);

	return 0;
}
