/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>

#include <rte_flow.h>
#include <doca_log.h>

#include "private/common_defines.h"
#include "dpdk_resource_manager.h"
#include "dpdk_internal_meta.h"

#define REG_INDEX(offset) ((offset) >> 5)
#define REG_OFFSET(offset) ((offset) & 0x1f)
#define REG_COUNT(offset, width) (REG_INDEX((offset) + (width) - 1) - REG_INDEX((offset)))
#define SEGMENT_SHIFT(data, width, offset) (((data) & ((1ULL << (width)) - 1)) << (offset))

DOCA_LOG_REGISTER(dpdk_interal_meta);

struct meta_segment {
	LIST_ENTRY(meta_segment) next;
	uint32_t bit_width; /** requested bit width */
	uint32_t bit_offset; /* allocated meta offset */
};

LIST_HEAD(meta_segment_list, meta_segment);

struct dpdk_meta {
	bool init_done;
	struct meta_segment segments[DPDK_META_MAX_SEGMENTS];
	struct meta_segment_list tag_list; /* requested complete tag list */
	struct meta_segment_list bit_list; /* requested bit segment list */
	uint32_t user_tag_offset; /* user meta data start tag ID */
} internal_meta;

static int
dpdk_internal_meta_init(uint32_t nr_resources __rte_unused)
{
	if (nr_resources != DPDK_META_MAX_SEGMENTS)
		return -ENOTSUP;

	memset(&internal_meta, 0, sizeof(internal_meta));

	return 0;
}

static int
dpdk_internal_meta_create(uint32_t internal_meta_id, struct engine_shared_resource_cfg *res_cfg)
{
	struct meta_segment *seg;
	struct meta_segment *peer;

	if (internal_meta_id >= DPDK_META_MAX_SEGMENTS)
		return -EINVAL;

	seg = &internal_meta.segments[internal_meta_id];
	if (seg->bit_width != 0)
		return -EINVAL;
	seg->bit_width = res_cfg->meta.bit_width;

	if (REG_OFFSET(seg->bit_width) == 0) {
		/* register a complete tag */
		LIST_INSERT_HEAD(&internal_meta.tag_list, seg, next);
	} else {
		/* register bit segment */
		LIST_FOREACH(peer, &internal_meta.bit_list, next) {
			if (REG_OFFSET(peer->bit_width + seg->bit_width) == 0) {
				/* Tag fully used with peer, move both to tag list */
				LIST_REMOVE(peer, next);
				LIST_INSERT_HEAD(&internal_meta.tag_list, peer, next);
				LIST_INSERT_HEAD(&internal_meta.tag_list, seg, next);
				return 0;
			}
		}
		LIST_INSERT_HEAD(&internal_meta.bit_list, seg, next);
	}
	return 0;
}

static int
dpdk_internal_meta_modify(uint32_t internal_meta_id, struct engine_shared_resource_cfg *res_cfg)
{
	return -ENOTSUP;
}

static int
dpdk_internal_meta_destroy(uint32_t internal_meta_id)
{
	return 0;
}

static int
dpdk_internal_meta_cleanup()
{
	memset(&internal_meta, 0, sizeof(internal_meta));
	return 0;
}

static void
internal_meta_allocate(void)
{
	struct meta_segment *seg;
	uint32_t bit_offset = 0;
	uint32_t tag_offset = 0;

	/* First, iterate complete tag list */
	LIST_FOREACH(seg, &internal_meta.tag_list, next) {
		assert(seg->bit_width > 0);
		seg->bit_offset = bit_offset;
		bit_offset += seg->bit_width;
	}
	bit_offset = RTE_ALIGN_CEIL(bit_offset, 32);

	/* Second, iterate tag segment list, combine where possible */
	LIST_FOREACH(seg, &internal_meta.bit_list, next) {
		assert(seg->bit_width > 0);
		assert(REG_OFFSET(seg->bit_width) > 0);
		if (REG_COUNT(tag_offset, seg->bit_width) > REG_COUNT(0, seg->bit_width)) {
			/* segment over wrapped, end previous segment */
			bit_offset = RTE_ALIGN_CEIL(bit_offset + tag_offset, 32);
			seg->bit_offset = bit_offset;
			tag_offset = seg->bit_width;
		} else if (REG_OFFSET(tag_offset + seg->bit_width) == 0) {
			/* tag complete, end tag. */
			seg->bit_offset = bit_offset + tag_offset;
			bit_offset += tag_offset + seg->bit_width;
			tag_offset = 0;
		} else {
			/* wait next to fill last tag */
			seg->bit_offset = bit_offset + tag_offset;
			tag_offset += seg->bit_width;
		}
	}
	bit_offset += tag_offset;

	/* Last, get user meta tag offset, start from new tag */
	internal_meta.user_tag_offset = RTE_ALIGN_CEIL(bit_offset, 32);

	internal_meta.init_done = true;
}

bool
dpdk_internal_meta_allocate_done(void)
{
	return internal_meta.init_done;
}

uint32_t
dpdk_internal_meta_user_tag_index(void)
{
	if (!internal_meta.init_done)
		internal_meta_allocate();
	return internal_meta.user_tag_offset;
}

int
dpdk_internal_meta_get(uint32_t meta_id, uint32_t *bit_width)
{
	struct meta_segment *seg;

	if (!internal_meta.init_done)
		internal_meta_allocate();
	seg = &internal_meta.segments[meta_id];
	if (bit_width)
		*bit_width = seg->bit_width;
	return (int)seg->bit_offset;
}

int
dpdk_internal_meta_tag_index(uint32_t meta_id)
{
	int ret = dpdk_internal_meta_get(meta_id, NULL);

	if (ret < 0)
		return ret;
	return (int)REG_INDEX(ret);
}

int
dpdk_internal_meta_build(uint32_t meta_id, uint32_t data, uint32_t *out)
{
	uint32_t bit_width;
	int ret = dpdk_internal_meta_get(meta_id, &bit_width);

	if (ret < 0)
		return ret;
	if (out)
		*out = SEGMENT_SHIFT(data, bit_width, ret);
	return 0;
}

static struct dpdk_resource_manager_ops internal_meta_ops = {
	.resources_init = dpdk_internal_meta_init,
	.resources_cleanup = dpdk_internal_meta_cleanup,

	.resource_create = dpdk_internal_meta_create,
	.resource_modify = dpdk_internal_meta_modify,
	.resource_destroy = dpdk_internal_meta_destroy,
};

struct dpdk_resource_manager_ops *
dpdk_internal_meta_get_ops()
{
	return &internal_meta_ops;
}
