/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stddef.h>

#include <rte_ethdev.h>
#include <rte_flow.h>
#include <rte_gre.h>
#include <rte_gtp.h>
#include <rte_vxlan.h>

#include "dpdk_layer.h"
#include "engine_to_string.h"
#include <private/common_defines.h>

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_layer);

#define REG_PROTO(opcode_str, offset) \
	do { \
		rc = register_proto(opcode_str, offset); \
		if (rc < 0) \
			return rc; \
	} while (0)

static int
register_proto(const char *opcode_str, uint32_t offset)
{
	struct engine_field_opcode opcode;
	int rc;

	rc = engine_string_to_opcode(opcode_str, strlen(opcode_str), &opcode);
	if (rc) {
		DOCA_LOG_ERR("failed registering opcode %s - string to opcode rc=%d", opcode_str,
			     rc);
		return rc;
	}

	rc = dpdk_field_mapping_set_offset(&opcode, offset);
	if (rc) {
		DOCA_LOG_ERR("failed registering opcode %s with offset %u rc=%d", opcode_str,
			     offset, rc);
		return rc;
	}

	DOCA_LOG_DBG("Registered dpdk field opcode=%s, offset=%u)", opcode_str, offset);
	return 0;
}

static int
register_proto_eth()
{
	int rc;

	REG_PROTO("match.packet.outer.eth.dst_mac", offsetof(struct rte_flow_item_eth, dst));
	REG_PROTO("match.packet.outer.eth.src_mac", offsetof(struct rte_flow_item_eth, src));
	REG_PROTO("match.packet.outer.eth.type", offsetof(struct rte_flow_item_eth, type));

	REG_PROTO("match.packet.inner.eth.dst_mac", offsetof(struct rte_flow_item_eth, dst));
	REG_PROTO("match.packet.inner.eth.src_mac", offsetof(struct rte_flow_item_eth, src));
	REG_PROTO("match.packet.inner.eth.type", offsetof(struct rte_flow_item_eth, type));

	return 0;
}

static int
register_proto_eth_vlan()
{
	int rc;

	REG_PROTO("match.packet.outer.eth_vlan.tci", offsetof(struct rte_flow_item_vlan, tci));
	REG_PROTO("match.packet.inner.eth_vlan.tci", offsetof(struct rte_flow_item_vlan, tci));

	return 0;
}

static int
register_proto_ipv4()
{
	int rc;

	REG_PROTO("match.packet.outer.ipv4.src_ip",
		  offsetof(struct rte_flow_item_ipv4, hdr.src_addr));
	REG_PROTO("match.packet.outer.ipv4.dst_ip",
		  offsetof(struct rte_flow_item_ipv4, hdr.dst_addr));
	REG_PROTO("match.packet.outer.ipv4.next_proto",
		  offsetof(struct rte_flow_item_ipv4, hdr.next_proto_id));

	REG_PROTO("match.packet.inner.ipv4.src_ip",
		  offsetof(struct rte_flow_item_ipv4, hdr.src_addr));
	REG_PROTO("match.packet.inner.ipv4.dst_ip",
		  offsetof(struct rte_flow_item_ipv4, hdr.dst_addr));
	REG_PROTO("match.packet.inner.ipv4.next_proto",
		  offsetof(struct rte_flow_item_ipv4, hdr.next_proto_id));

	return 0;
}

static int
register_proto_ipv6()
{
	int rc;

	REG_PROTO("match.packet.outer.ipv6.src_ip",
		  offsetof(struct rte_flow_item_ipv6, hdr.src_addr));
	REG_PROTO("match.packet.outer.ipv6.dst_ip",
		  offsetof(struct rte_flow_item_ipv6, hdr.dst_addr));
	REG_PROTO("match.packet.outer.ipv6.next_proto",
		  offsetof(struct rte_flow_item_ipv6, hdr.proto));

	REG_PROTO("match.packet.inner.ipv6.src_ip",
		  offsetof(struct rte_flow_item_ipv6, hdr.src_addr));
	REG_PROTO("match.packet.inner.ipv6.dst_ip",
		  offsetof(struct rte_flow_item_ipv6, hdr.dst_addr));
	REG_PROTO("match.packet.inner.ipv6.next_proto",
		  offsetof(struct rte_flow_item_ipv6, hdr.proto));

	return 0;
}

static int
register_proto_udp()
{
	int rc;

	REG_PROTO("match.packet.outer.udp.src_port",
		  offsetof(struct rte_flow_item_udp, hdr.src_port));
	REG_PROTO("match.packet.outer.udp.dst_port",
		  offsetof(struct rte_flow_item_udp, hdr.dst_port));

	REG_PROTO("match.packet.inner.udp.src_port",
		  offsetof(struct rte_flow_item_udp, hdr.src_port));
	REG_PROTO("match.packet.inner.udp.dst_port",
		  offsetof(struct rte_flow_item_udp, hdr.dst_port));

	return 0;
}

static int
register_proto_tcp()
{
	int rc;

	REG_PROTO("match.packet.outer.tcp.src_port",
		  offsetof(struct rte_flow_item_tcp, hdr.src_port));
	REG_PROTO("match.packet.outer.tcp.dst_port",
		  offsetof(struct rte_flow_item_tcp, hdr.dst_port));
	REG_PROTO("match.packet.outer.tcp.flags",
		  offsetof(struct rte_flow_item_tcp, hdr.tcp_flags));

	REG_PROTO("match.packet.inner.tcp.src_port",
		  offsetof(struct rte_flow_item_tcp, hdr.src_port));
	REG_PROTO("match.packet.inner.tcp.dst_port",
		  offsetof(struct rte_flow_item_tcp, hdr.dst_port));
	REG_PROTO("match.packet.inner.tcp.flags",
		  offsetof(struct rte_flow_item_tcp, hdr.tcp_flags));

	return 0;
}

static int
register_tun_protocols()
{
	int rc;

	REG_PROTO("match.packet.tunnel.vxlan.vni", offsetof(struct rte_flow_item_vxlan, vni));
	REG_PROTO("match.packet.tunnel.gre.key", 0 /* gre key spec is uint32_t */);
	REG_PROTO("match.packet.tunnel.gtp.teid", offsetof(struct rte_flow_item_gtp, teid));

	return 0;
}

static int
dpdk_register_capabilities()
{
	int rc;

	rc = register_proto_eth();
	if (rc < 0)
		goto init_rollback;

	rc = register_proto_eth_vlan();
	if (rc < 0)
		goto init_rollback;

	rc = register_proto_ipv4();
	if (rc < 0)
		goto init_rollback;

	rc = register_proto_ipv6();
	if (rc < 0)
		goto init_rollback;

	rc = register_proto_udp();
	if (rc < 0)
		goto init_rollback;

	rc = register_proto_tcp();
	if (rc < 0)
		goto init_rollback;

	rc = register_tun_protocols();
	if (rc < 0)
		goto init_rollback;

	return 0;

init_rollback:
	dpdk_layer_unregister();
	return rc;
}

int
dpdk_layer_register(struct dpdk_layer_cfg *cfg)
{
	int rc;

	rc = dpdk_table_init(&cfg->table_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed registering dpdk layer - dpdk_table rc=%d", rc);
		return rc;
	}

	rc = dpdk_flow_init(&cfg->flow_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed registering dpdk layer - flow rc=%d", rc);
		return rc;
	}

	rc = dpdk_resource_manager_register();
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed registering dpdk layer - resource manager rc=%d", rc);
		return rc;
	}

	rc = dpdk_field_mapping_init();
	if (rc < 0) {
		DOCA_LOG_ERR("failed registering dpdk field mapping");
		return rc;
	}

	rc = dpdk_register_capabilities();
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed registering dpdk layer capabilities");
		return rc;
	}

	DOCA_LOG_INFO("Dpdk layer register completed");
	return 0;
}

void
dpdk_layer_unregister()
{
	dpdk_field_mapping_destroy();
	dpdk_resource_manager_unregister();
	dpdk_flow_cleanup();
	dpdk_table_cleanup();

	DOCA_LOG_INFO("Dpdk layer unregister completed");
}
