/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <errno.h>

#include <rte_mtr.h>

#include <doca_log.h>

#include "dpdk_meter_controller.h"

DOCA_LOG_REGISTER(dpdk_meter_controller);

#define METER_MARK_STATE_ON 1

struct dpdk_profiles_per_port {
	uint32_t nb_profiles;
	struct rte_flow_meter_profile **pmd_profiles;
};

struct dpdk_profiles_ports {
	uint32_t nb_ports;
	struct dpdk_profiles_per_port *port_profiles;
};

static struct dpdk_profiles_ports dpp = { 0 };

struct dpdk_meter {
	enum dpdk_meter_type type;
};

static struct dpdk_meter dpdk_meter = { 0 };

static int
port_check(uint16_t port_id)
{
	if (port_id >= dpp.nb_ports) {
		DOCA_LOG_ERR("failed to create profile on port - out of range (%u/%u)",
			     port_id, dpp.nb_ports);
		return -EINVAL;
	}

	return 0;
}

static int
port_profile_check(uint16_t port_id, uint32_t profile_id)
{
	int ret;

	ret = port_check(port_id);
	if (ret)
		return ret;

	if (profile_id >= dpp.port_profiles[port_id].nb_profiles) {
		DOCA_LOG_ERR("failed to create profile - out of range (%u/%u)",
			     profile_id, dpp.port_profiles[port_id].nb_profiles);
		return -EINVAL;
	}

	return 0;
}

int
dpdk_meter_profile_ports_set(uint32_t nb_ports)
{
	dpp.port_profiles = calloc(nb_ports,
				   sizeof(struct dpdk_profiles_per_port));
	if (unlikely(!dpp.port_profiles)) {
		DOCA_LOG_ERR("failed to allocate %u profiles ports", nb_ports);
		return -ENOMEM;
	}

	dpp.nb_ports = nb_ports;

	return 0;
}

void
dpdk_meter_profile_ports_unset()
{
	free(dpp.port_profiles);
	dpp.port_profiles = NULL;
	dpp.nb_ports = 0;
}

int
dpdk_meter_profile_alloc(uint16_t port_id, uint32_t nb_profiles)
{
	int ret;
	struct rte_flow_meter_profile **pmd_profiles;
	struct dpdk_profiles_per_port *p;

	ret = port_check(port_id);
	if (ret)
		return ret;

	pmd_profiles = calloc(nb_profiles, sizeof(struct rte_flow_meter_profile *));
	if (unlikely(!pmd_profiles)) {
		DOCA_LOG_ERR("failed to allocate %u pmd profiles", nb_profiles);
		return -ENOMEM;
	}

	p = &dpp.port_profiles[port_id];
	p->pmd_profiles = pmd_profiles;
	p->nb_profiles = nb_profiles;

	return 0;
}

void
dpdk_meter_profile_free(uint16_t port_id)
{
	int ret;

	ret = port_check(port_id);
	if (ret)
		return;

	free(dpp.port_profiles[port_id].pmd_profiles);
	dpp.port_profiles[port_id].nb_profiles = 0;
}

int
dpdk_meter_profile_create(uint16_t port_id,
				 uint32_t profile_id,
				 uint64_t cir,
				 uint64_t cbs)
{
	struct rte_mtr_meter_profile mp;
	struct rte_flow_meter_profile *pmd_profile;
	struct rte_mtr_error mtr_error;
	int ret;

	ret = port_profile_check(port_id, profile_id);
	if (ret)
		return ret;

	memset(&mp, 0, sizeof(struct rte_mtr_meter_profile));
	mp.alg = RTE_MTR_SRTCM_RFC2697;
	mp.srtcm_rfc2697.cir = cir;
	mp.srtcm_rfc2697.cbs = cbs;
	mp.srtcm_rfc2697.ebs = 0;

	/* Shared profile id the shared meter id */
	ret = rte_mtr_meter_profile_add(port_id, profile_id, &mp, &mtr_error);
	if (unlikely(ret)) {
		DOCA_LOG_ERR(
		    "Port %u create Profile id %u error(%d) message: %s\n",
		    port_id, profile_id, mtr_error.type,
		    mtr_error.message ? mtr_error.message : "(no stated reason)");

		return ret;
	}

	/* pmd_profile is a pointer to MLX5 internal representation */
	pmd_profile = rte_mtr_meter_profile_get(port_id, profile_id, &mtr_error);
	if (unlikely(!pmd_profile)) {
		DOCA_LOG_ERR("HW failed on shared meter %u CIR/CBS values. Type %d message %s",
			      profile_id, mtr_error.type, mtr_error.message ?
			      mtr_error.message : "(no stated reason)");

		rte_mtr_meter_profile_delete(port_id, profile_id, &mtr_error);

		return -EIO;
	}

	dpp.port_profiles[port_id].pmd_profiles[profile_id] = pmd_profile;

	return 0;
}

static struct rte_flow_meter_profile *
dpdk_meter_profile_get(uint16_t port_id, uint32_t profile_id)
{
	struct rte_flow_meter_profile *pmd_profile;

	pmd_profile = dpp.port_profiles[port_id].pmd_profiles[profile_id];
	if (unlikely(pmd_profile == NULL))
		DOCA_LOG_ERR("failed finding profile id (%u) driver reference. Was it created?",
			      profile_id);

	return pmd_profile;
}

static void
_dpdk_meter_policy_conf_build(uint32_t action_mtr_id,
				  struct rte_flow_action_meter *meter_policy_conf)
{
	meter_policy_conf->mtr_id = action_mtr_id;
}

void
dpdk_meter_mark_conf_build(uint16_t port_id, uint32_t profile_id,
				  struct rte_flow_action_meter_mark *meter_mark_conf)
{
	struct rte_flow_meter_profile *pmd_profile;

	if (port_profile_check(port_id, profile_id))
		return;

	pmd_profile = dpdk_meter_profile_get(port_id, profile_id);

	meter_mark_conf->profile = pmd_profile;
	meter_mark_conf->state = METER_MARK_STATE_ON;
}

int
dpdk_meter_profile_delete(uint16_t port_id, uint32_t profile_id)
{
	int ret;
	struct rte_mtr_error error;

	ret = port_profile_check(port_id, profile_id);
	if (ret)
		return ret;

	ret = rte_mtr_meter_profile_delete(port_id, profile_id, &error);
	if (ret != 0) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"Port %u delete profile idx(%d) error(%d) message: %s",
			port_id, profile_id, error.type,
			error.message ? error.message : "(no stated reason)");

		return ret;
	}

	dpp.port_profiles[port_id].pmd_profiles[profile_id] = NULL;
	return ret;
}

int
dpdk_meter_set_type(enum dpdk_meter_type type)
{
	switch (type) {
	case DPDK_METER_TYPE_MARK:
	case DPDK_METER_TYPE_POLICY:
		dpdk_meter.type = type;
		return 0;

	default:
		return -EINVAL;
	}
}

bool
dpdk_meter_is_policy_type(void)
{
	return (dpdk_meter.type == DPDK_METER_TYPE_POLICY);
}

bool
dpdk_meter_is_mark_type(void)
{
	return (dpdk_meter.type == DPDK_METER_TYPE_MARK);
}

int
dpdk_meter_data_build(struct rte_flow_action *action, void *conf,
			     uint32_t action_mtr_id, uint32_t profile_id,
			     uint16_t port_id)
{
	int ret = 0;

	if (dpdk_meter.type == DPDK_METER_TYPE_POLICY) {
		_dpdk_meter_policy_conf_build(action_mtr_id, conf);
		action->type = RTE_FLOW_ACTION_TYPE_METER;
		action->conf = conf;
	} else if (dpdk_meter.type == DPDK_METER_TYPE_MARK) {
		if (profile_id == PROFILE_ID_CHANGEABLE)
			conf = NULL;
		else
			dpdk_meter_mark_conf_build(port_id, profile_id, conf);

		action->type = RTE_FLOW_ACTION_TYPE_METER_MARK;
		action->conf = conf;
	} else
		ret = -EINVAL;

	return ret;
}

int
dpdk_meter_contrl_policy_rule_create(int port_id, uint32_t dpdk_policy_id,
				     uint32_t dpdk_mtr_id, uint32_t dpdk_profile_id)
{
	struct rte_mtr_params params;
	struct rte_mtr_error error;
	int ret;

	memset(&params, 0, sizeof(struct rte_mtr_params));
	params.meter_enable = 1;
	params.stats_mask = 0xffff;
	params.use_prev_mtr_color = 0;
	params.dscp_table = NULL;
	params.meter_profile_id = dpdk_profile_id;
	params.meter_policy_id = dpdk_policy_id;
	ret = rte_mtr_create(port_id, dpdk_mtr_id, &params, 1, &error);
	if (ret != 0) {
		DOCA_DLOG_ERR(
		    "Port %u create meter idx(%d) profile(%u) policy(%u) error(%d) message: %s",
		    port_id, dpdk_mtr_id, dpdk_profile_id, dpdk_policy_id,
		    error.type, error.message ? error.message : "(no stated reason)");
		return -ENOENT;
	}
	return 0;
}
