/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DPDK_METER_CONTROLLER_H_
#define DPDK_METER_CONTROLLER_H_

#ifdef __cplusplus
extern "C" {
#endif

enum dpdk_meter_type {
	DPDK_METER_TYPE_UNSPECIFIED,
	DPDK_METER_TYPE_POLICY,
	DPDK_METER_TYPE_MARK,
};

#define PROFILE_ID_CHANGEABLE	(0xffffffff)

/**
 * @brief Init meter profile.
 *
 * @param nb_proifles
 * Number of meter profiles (direct or shared)
 * @return
 * 0 on success, negative on failure.
 */
int
dpdk_meter_profile_init(uint32_t nb_profiles);

/**
 * @brief Create a meter profile.
 *
 * @param port_id
 * DPDK port id
 * @param profile_id.
 * Profile identifier.
 * @param cir
 * Committed information rate.
 * @param cbs
 * Committed burst size
 * @return
 * 0 on success, negative on failure.
 */
int
dpdk_meter_profile_create(uint16_t port_id,
				 uint32_t profile_id,
				 uint64_t cir,
				 uint64_t cbs);

/**
 * @brief Delete a meter profile.
 *
 * @param port_id
 * DPDK port id
 * @param profile_id.
 * Profile identifier.
 * @return
 * 0 on success, negative on failure.
 */
int
dpdk_meter_profile_delete(uint16_t port_id, uint32_t profile_id);

/**
 * @brief Build a meter profile configuration.
 *
 * @param port_id
 * DPDK port id
 * @param profile_id.
 * Profile identifier.
 */
void
dpdk_meter_mark_conf_build(uint16_t port_id, uint32_t profile_id,
				  struct rte_flow_action_meter_mark *meter_mark_conf);

/**
 * @brief Allocate for port id its number of profiles.
 *
 * @param port_id
 * DPDK port id
 * @param nb_profiles
 * Number of profiles
 * @return
 * 0 on success, negative on failure.
 */
int
dpdk_meter_profile_alloc(uint16_t port_id, uint32_t nb_profiles);

/**
 * @brief Free number of profiles belonging to port id.
 *
 * @param port_id
 * DPDK port id
 * @return
 * 0 on success, negative on failure.
 */
void
dpdk_meter_profile_free(uint16_t port_id);

/**
 * @brief Set the max number of ports range expected to have profiles.
 *
 * @param nb_ports
 * Number of ports
 * @return
 * 0 on success, negative on failure.
 */
int
dpdk_meter_profile_ports_set(uint32_t nb_ports);

/**
 * @brief Unset the max number of ports range expected to have profiles.
 *
 * @return
 * 0 on success, negative on failure.
 */
void
dpdk_meter_profile_ports_unset(void);

/**
 * @brief Set meter type.
 *
 * @type
 * Meter type: MARK or POLICY.
 * @return
 * 0 on success, negative on failure.
 */
int
dpdk_meter_set_type(enum dpdk_meter_type type);

/**
 * @brief Confirm if meter is policy based.
 *
 * @return
 * True if meter is policy based, false otherwise.
 */
bool
dpdk_meter_is_policy_type(void);

/**
 * @brief Confirm if meter is mark based.
 *
 * @return
 * True if meter is mark based, false otherwise.
 */
bool
dpdk_meter_is_mark_type(void);

int
dpdk_meter_data_build(struct rte_flow_action *action, void *conf,
			     uint32_t action_mtr_id, uint32_t profile_id,
			     uint16_t port_id);

/**
 * @brief Create a meter based on policy id.
 *
 * @param port id
 * Port identifier on whih the meter objet is created.
 * @param policy id
 * Policy id with which the meter object is created.
 * @param mtr_id
 * meter identifier
 * @param profile_id
 * Profile id with which the meter object is created.
 *
 * @return
 * O on success, negative on failure.
 */
int
dpdk_meter_contrl_policy_rule_create(int port_id, uint32_t policy_id,
				     uint32_t mtr_id, uint32_t profile_id);
#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DPDK_METER_CONTROLLER_H_ */
