/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <string.h>

#include "private/common_defines.h"

#include "dpdk_port.h"
#include "dpdk_flow.h"
#include "dpdk_group_pool.h"
#include "dpdk_fwd_groups.h"
#include "dpdk_table.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_port);

struct dpdk_port {
	struct dpdk_table_port_ctx *table_ctx;
	struct dpdk_flow_queue **queue_array;
	uint16_t port_id;
	uint16_t nr_queues_active;
	uint16_t mgmt_queue_id;
	struct dpdk_group_pool *group_pool;
	struct dpdk_fwd_groups *fwd_groups;
};

static struct dpdk_port *dpdk_ports[DPDK_PORT_MAX_PORT_ID] = {0};

static int
port_register(struct dpdk_port *port)
{
	if (port->port_id > DPDK_PORT_MAX_PORT_ID) {
		DOCA_LOG_RATE_LIMIT_ERR("failure registering port - invalid port_id %u", port->port_id);
		return -EINVAL;
	}

	if (dpdk_ports[port->port_id]) {
		DOCA_LOG_RATE_LIMIT_ERR("failure registering port - port_id %u already taken", port->port_id);
		return -EINVAL;
	}

	dpdk_ports[port->port_id] = port;
	return 0;
}

static void
port_unregister(struct dpdk_port *port)
{
	dpdk_ports[port->port_id] = NULL;
}

struct dpdk_port *
dpdk_port_get_by_id(uint16_t port_id)
{
	return dpdk_ports[port_id];
}

struct dpdk_fwd_groups *
dpdk_port_fwd_groups_get(struct dpdk_port *port)
{
	if (unlikely(!port))
		return NULL;

	return port->fwd_groups;
}

struct dpdk_group_pool *
dpdk_port_get_group_pool(struct dpdk_port *dpdk_port)
{
	if (unlikely(!dpdk_port))
		return NULL;

	return dpdk_port->group_pool;
}

struct dpdk_port *
dpdk_port_create(struct dpdk_port_cfg *cfg)
{
	struct dpdk_port *port;
	struct dpdk_flow_queue_cfg queue_cfg;
	struct dpdk_fwd_groups_cfg fwd_groups_cfg = {0};
	struct dpdk_flow_queue *management_queue;
	int ret;

	if (unlikely(!cfg)) {
		DOCA_LOG_ERR("failed creating dpdk port - configuration is null");
		return NULL;
	}

	if (unlikely(cfg->nr_queues == 0)) {
		DOCA_LOG_ERR("failed creating dpdk port - configured 0 queues");
		return NULL;
	}

	port = calloc(1, sizeof(struct dpdk_port));
	if (unlikely(!port)) {
		DOCA_LOG_ERR("failed creating dpdk port - memory depleted");
		return NULL;
	}

	port->queue_array = calloc(cfg->nr_queues, sizeof(struct dpdk_flow_queue *));
	if (unlikely(!port->queue_array)) {
		DOCA_LOG_ERR("failed creating dpdk port - queue array memory depleted");
		goto rollback;
	}

	port->port_id = cfg->port_id;
	port->mgmt_queue_id = cfg->mgmt_queue_id;
	ret = port_register(port);
	if (ret)
		goto rollback;

	for (uint16_t queue_id = 0; queue_id < cfg->nr_queues; queue_id++) {
		memset(&queue_cfg, 0, sizeof(queue_cfg));
		queue_cfg.port_id = port->port_id;
		queue_cfg.queue_id = queue_id;
		queue_cfg.queue_depth = cfg->queue_depth;

		port->queue_array[queue_id] = dpdk_flow_queue_create(&queue_cfg);
		if (!port->queue_array[queue_id]) {
			DOCA_LOG_ERR("failed creating dpdk port - queue id %u failed creation",
				     queue_id);
			goto rollback;
		}
		port->nr_queues_active++;
	}

	port->group_pool = dpdk_group_pool_create(&cfg->group_pool_cfg);
	if (unlikely(!port->group_pool)) {
		DOCA_LOG_ERR("failed creating dpdk port - group pool failed creation");
		goto rollback;
	}

	port->table_ctx = dpdk_table_port_ctx_create(port->port_id);
	if (port->table_ctx == NULL) {
		DOCA_LOG_ERR("failed creating dpdk port - failed to create table port context");
		goto rollback;
	}

	management_queue = dpdk_port_get_flow_queue(port, port->mgmt_queue_id);
	if (management_queue == NULL) {
		DOCA_LOG_ERR("failed creating dpdk port - failed to get management queue");
		goto rollback;
	}

	fwd_groups_cfg.port = port;
	fwd_groups_cfg.management_lock = cfg->mgmt_lock;
	fwd_groups_cfg.management_queue = management_queue;
	port->fwd_groups = dpdk_fwd_groups_create(&fwd_groups_cfg);
	if (port->table_ctx == NULL) {
		DOCA_LOG_ERR("failed creating dpdk port - failed to create fwd_groups");
		goto rollback;
	}

	DOCA_LOG_INFO("Dpdk port %u initialized successfully with %u queues", cfg->port_id, cfg->nr_queues);
	return port;
rollback:
	dpdk_port_destroy(port);
	return NULL;
}

struct dpdk_flow_queue *
dpdk_port_get_flow_queue(struct dpdk_port *port, uint16_t queue_id)
{
	if (unlikely(!port)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed getting flow queue - port is null");
		return NULL;
	}

	if (unlikely(queue_id >= port->nr_queues_active)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed getting flow queue - queue_id %u invalid",
					queue_id);
		return NULL;
	}

	return port->queue_array[queue_id];
}

struct dpdk_table_port_ctx *
dpdk_port_get_table_ctx(struct dpdk_port *port)
{
	return port->table_ctx;
}

uint16_t
dpdk_port_get_id(struct dpdk_port *port)
{
	return port->port_id;
}

void
dpdk_port_destroy(struct dpdk_port *port)
{

	if (unlikely(!port)) {
		DOCA_LOG_ERR("failed destroying dpdk port - null pointer");
		return;
	}

	if (port->fwd_groups)
		dpdk_fwd_groups_destroy(port->fwd_groups);

	if (port->group_pool)
		dpdk_group_pool_destroy(port->group_pool);

	if (likely(port->queue_array)) {
		for (uint32_t queue_id = 0; queue_id < port->nr_queues_active; queue_id++) {
			dpdk_flow_queue_destroy(port->queue_array[queue_id],
						true /* wait_for_completions */);
		}
		free(port->queue_array);
	}

	if (likely(port->table_ctx))
		dpdk_table_port_ctx_destroy(port->table_ctx);

	port_unregister(port);
	DOCA_LOG_INFO("Dpdk port %u destroyed successfully with %u queues", port->port_id,
		      port->nr_queues_active);
	free(port);
}
