/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DPDK_PORT_H_
#define DPDK_PORT_H_

#include <stdint.h>
#include <stdbool.h>
#include <rte_spinlock.h>

#include "dpdk_group_pool.h"

#ifdef __cplusplus
extern "C" {
#endif

#define DPDK_PORT_MAX_PORT_ID (256)

struct dpdk_port;
struct dpdk_flow_queue;

struct dpdk_port_cfg {
	uint16_t port_id;
	uint16_t nr_queues;
	uint16_t queue_depth;
	uint16_t mgmt_queue_id;
	rte_spinlock_t *mgmt_lock;
	struct dpdk_group_pool_cfg group_pool_cfg;
};

/**
 * @brief Create a dpdk port.
 *
 * @param cfg
 * Pointer to configuration
 * @return
 * Valid port pointer on success, otherwise NULL.
 */
struct dpdk_port *
dpdk_port_create(struct dpdk_port_cfg *cfg);

/**
 * @brief Get a flow queue of port.
 *
 * @param port
 * Pointer to port
 * @param queue_id
 * Queue index
 * @return
 * Valid port pointer on success, otherwise NULL.
 */
struct dpdk_flow_queue *
dpdk_port_get_flow_queue(struct dpdk_port *port, uint16_t queue_id);


/**
 * @brief Get group pool of port.
 *
 * @param port
 * Pointer to port
 * @return
 * Valid port pointer on success, otherwise NULL.
 */
struct dpdk_group_pool *
dpdk_port_get_group_pool(struct dpdk_port *port);

/**
 * @brief Get group fwd
 *
 * @param port
 * Port
 * @return
 * Valid port pointer on success, otherwise NULL.
 */
struct dpdk_fwd_groups *
dpdk_port_fwd_groups_get(struct dpdk_port *port);

/**
 * @brief Get a table ctx of a port.
 *
 * @param port
 * Pointer to port
 * @return
 * Valid table port context pointer on success, otherwise NULL.
 */
struct dpdk_table_port_ctx *
dpdk_port_get_table_ctx(struct dpdk_port *port);

/**
 * @brief Get a port.
 *
 * @param port_id
 * Port_id
 * @return
 * Valid port pointer on success, otherwise NULL.
 */
struct dpdk_port *
dpdk_port_get_by_id(uint16_t port_id);

/**
 * @brief Get port id of a port.
 *
 * @param port
 * Pointer to port
 * @return
 * port id
 */
uint16_t
dpdk_port_get_id(struct dpdk_port *port);

/**
 * @brief Destroys a dpdk port.
 *
 * @param port
 * Pointer to port
 */
void
dpdk_port_destroy(struct dpdk_port *port);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DPDK_PORT_H_ */
