/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>

#include "private/common_defines.h"
#include "dpdk_resource_manager.h"
#include "dpdk_shared_meter.h"
#include "dpdk_shared_counter.h"
#include "dpdk_internal_meta.h"
#include "dpdk_shared_rss.h"
#include "dpdk_shared_crypto.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_resource_manager);

struct shared_resource {
	struct dpdk_resource_manager_ops ops;
	bool is_init;
};

struct shared_resources {
	struct shared_resource resource_per_type[ENGINE_SHARED_RESOURCE_MAX];
};

static struct shared_resources shared_resources = {0};

static struct dpdk_resource_manager_ops *
get_ops_per_type(enum engine_shared_resource_type type)
{
	if (unlikely(type >= ENGINE_SHARED_RESOURCE_MAX))
		return NULL;

	if (unlikely(!shared_resources.resource_per_type[type].is_init))
		return NULL;

	return &shared_resources.resource_per_type[type].ops;
}

static int
dpdk_resources_init(enum engine_shared_resource_type type, uint32_t nr_resources)
{
	struct dpdk_resource_manager_ops *ops = get_ops_per_type(type);

	if (unlikely(!ops))
		return -ENOTSUP;

	return ops->resources_init(nr_resources);
}

static int
dpdk_resource_create(enum engine_shared_resource_type type, uint32_t id,
		     struct engine_shared_resource_cfg *res_cfg)
{
	struct dpdk_resource_manager_ops *ops = get_ops_per_type(type);

	if (unlikely(!ops))
		return -ENOTSUP;

	return ops->resource_create(id, res_cfg);
}

static int
dpdk_resource_modify(enum engine_shared_resource_type type, uint32_t id,
		     struct engine_shared_resource_cfg *res_cfg)
{
	struct dpdk_resource_manager_ops *ops = get_ops_per_type(type);

	if (unlikely(!ops))
		return -ENOTSUP;

	return ops->resource_modify(id, res_cfg);
}

static int
dpdk_resource_query(enum engine_shared_resource_type type, uint32_t id,
		    struct engine_shared_resource_query_result *res_query_rslts)
{
	struct dpdk_resource_manager_ops *ops = get_ops_per_type(type);

	if (unlikely(!ops))
		return -ENOTSUP;

	return ops->resource_query(id, res_query_rslts);
}

static int
dpdk_resource_destroy(enum engine_shared_resource_type type, uint32_t id)
{
	struct dpdk_resource_manager_ops *ops = get_ops_per_type(type);

	if (unlikely(!ops))
		return -ENOTSUP;

	return ops->resource_destroy(id);
}

static int
dpdk_resources_cleanup(enum engine_shared_resource_type type)
{
	struct dpdk_resource_manager_ops *ops = get_ops_per_type(type);

	if (unlikely(!ops))
		return -ENOTSUP;

	return ops->resources_cleanup();
}

static int
register_ops_per_type(enum engine_shared_resource_type type,
		      struct dpdk_resource_manager_ops *ops)
{
	if (type >= ENGINE_SHARED_RESOURCE_MAX)
		return -EINVAL;

	if (!ops)
		return -EINVAL;

	if (!ops->resources_init)
		return -EINVAL;
	if (!ops->resource_create)
		return -EINVAL;
	if (!ops->resource_modify)
		return -EINVAL;
	if (!ops->resource_destroy)
		return -EINVAL;
	if (!ops->resources_cleanup)
		return -EINVAL;

	shared_resources.resource_per_type[type].ops = *ops;
	shared_resources.resource_per_type[type].is_init = true;
	return 0;
}

static struct engine_shared_resources_ops dpdk_resource_manager_ops = {
	.resources_init = dpdk_resources_init,
	.resources_cleanup = dpdk_resources_cleanup,

	.resource_create = dpdk_resource_create,
	.resource_modify = dpdk_resource_modify,
	.resource_query = dpdk_resource_query,
	.resource_destroy = dpdk_resource_destroy,
};

int
dpdk_resource_manager_register()
{
	struct dpdk_resource_manager_ops *ops[ENGINE_SHARED_RESOURCE_MAX] = {0};
	int op_idx;
	int rc;

	ops[ENGINE_SHARED_RESOURCE_METER] = dpdk_shared_meter_get_ops();
	ops[ENGINE_SHARED_RESOURCE_COUNTER] = dpdk_shared_counter_get_ops();
	ops[ENGINE_SHARED_RESOURCE_INTERNAL_META] = dpdk_internal_meta_get_ops();
	ops[ENGINE_SHARED_RESOURCE_RSS] = dpdk_shared_rss_get_ops();
	ops[ENGINE_SHARED_RESOURCE_CRYPTO] = dpdk_shared_crypto_get_ops();

	for (op_idx = 0; op_idx < ENGINE_SHARED_RESOURCE_MAX; op_idx++) {
		rc = register_ops_per_type(op_idx, ops[op_idx]);
		if (unlikely(rc)) {
			DOCA_LOG_ERR("failed registering resource manager - resource %u ops rc=%d",
				     op_idx, rc);
			return rc;
		}
	}

	rc = engine_shared_resource_register_ops(&dpdk_resource_manager_ops);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed registering dpdk resource manager rc=%d", rc);
		return rc;
	}

	DOCA_LOG_INFO("Dpdk resource manager register completed");
	return 0;
}

void
dpdk_resource_manager_unregister()
{
	DOCA_LOG_INFO("Dpdk resource manager unregister completed");
}
