/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <errno.h>

#include <rte_flow.h>

#include "private/common_defines.h"
#include "dpdk_resource_manager.h"

#include <doca_log.h>

#include "engine_model.h"

DOCA_LOG_REGISTER(dpdk_shared_counter);

struct dpdk_shared_ctrs_elems {
	uint32_t port_id;
	struct rte_flow_action_handle *handle;
};

struct dpdk_shared_ctrs {
	uint32_t nb_resources;
	struct dpdk_shared_ctrs_elems *elems;
};

static struct dpdk_shared_ctrs shared_ctrs = {
	.nb_resources = 0,
	.elems = NULL,
};

static struct rte_flow_indir_action_conf conf = {
	.ingress = 1,
	.egress = 0,
	.transfer = 0,
};

static struct rte_flow_action_count act_count = {
	.shared = 0x0,
	.reserved = 0x0,
	.id = 0x0,
};

static struct rte_flow_action action = {
	.type = RTE_FLOW_ACTION_TYPE_COUNT,
	.conf = &act_count,
};

struct dpdk_shared_ctrs_elems *
_action_handle_get(uint32_t counter_id)
{
	return &shared_ctrs.elems[counter_id];
}

static int
dpdk_shared_counter_init(uint32_t nb_resources)
{
	if (unlikely(nb_resources == 0)) {
		DOCA_LOG_ERR("error - number of shared counters must not be 0.");
		return -EINVAL;
	}

	shared_ctrs.elems = (struct dpdk_shared_ctrs_elems *)calloc(nb_resources,
			    sizeof(struct dpdk_shared_ctrs_elems));

	if (!shared_ctrs.elems) {
		DOCA_LOG_ERR("Failed to allocate %u shared counters", nb_resources);
		return -ENOMEM;
	}

	conf.transfer = engine_model_is_mode(ENGINE_MODEL_MODE_SWITCH) ? 1 : 0;
	shared_ctrs.nb_resources = nb_resources;

	return 0;
}

static struct rte_flow_action_handle *
__shared_counter_action_handle_create(uint32_t port_id, uint32_t counter_id)
{
	struct rte_flow_error error;
	struct rte_flow_action_handle *handle;

	handle = rte_flow_action_handle_create(port_id, &conf, &action, &error);
	if (!handle)
		DOCA_LOG_ERR("failed creating count %u on port %u, type %d message %s",
			      counter_id, port_id, error.type, error.message ?
			      error.message : "(no stated reason)");

	return handle;
}

static int
dpdk_shared_counter_create(uint32_t counter_id,
			   struct engine_shared_resource_cfg *res_cfg)
{
	uint16_t port_id;
	struct rte_flow_action_handle *handle;

	if (!shared_ctrs.nb_resources) {
		DOCA_LOG_ERR("failed creating shared counter - no resource initialized");
		return -ENOENT;
	}

	if (unlikely(shared_ctrs.elems[counter_id].handle)) {
		DOCA_LOG_ERR("failed creating shared counter id %u - was already created",
			     counter_id);
		return -EALREADY;
	}

	if (unlikely(!res_cfg)) {
		DOCA_LOG_ERR("failed creating shared counter id %u - missing conf", counter_id);
		return -EINVAL;
	}

	port_id = res_cfg->counter.port_id;
	handle = __shared_counter_action_handle_create(port_id, counter_id);
	if (!handle) {
		DOCA_LOG_ERR("failed creating shared counter id %u port id %u - dpdk error",
			      counter_id, port_id);
		return -EIO;
	}

	shared_ctrs.elems[counter_id].handle = handle;
	shared_ctrs.elems[counter_id].port_id = port_id;
	return 0;
}

static int
dpdk_shared_counter_modify(uint32_t counter_id, struct engine_shared_resource_cfg *res_cfg)
{
	return -ENOTSUP;
}

static int
dpdk_shared_counter_query(uint32_t counter_id,
			  struct engine_shared_resource_query_result *res_query_rslt)
{
	int rc;
	struct rte_flow_error error;
	struct rte_flow_query_count count;
	struct rte_flow_action_handle *handle;
	uint16_t port_id;
	struct dpdk_shared_ctrs_elems *elem;

	if (unlikely(counter_id >= shared_ctrs.nb_resources)) {
		DOCA_LOG_ERR("failed querying shared counter id %u - above max resources.",
			      counter_id);
		return -EINVAL;
	}

	if (unlikely(!res_query_rslt)) {
		DOCA_LOG_ERR("failed querying shared counter id %u - NULL output results pointer.",
			      counter_id);
		return -EINVAL;
	}

	memset(&count, 0, sizeof(count));
	elem = _action_handle_get(counter_id);
	handle = elem->handle;
	if (!handle) {
		DOCA_LOG_ERR("failed querying shared counter id %u - unbound.",
			      counter_id);
		return -ENOENT;
	}

	port_id = elem->port_id;
	rc = rte_flow_action_handle_query(port_id, handle, &count, &error);
	if (rc) {
		DOCA_LOG_ERR("failed querying shared counter id %u type %d message: %s",
			     counter_id, error.type,
			     error.message ? error.message : "(no stated reason)");
		return rc;
	}

	res_query_rslt->counter.total_bytes = count.bytes;
	res_query_rslt->counter.total_pkts = count.hits;

	return 0;
}

static int
dpdk_shared_counter_destroy(uint32_t counter_id)
{
	int rc;
	uint16_t port_id;
	struct rte_flow_action_handle *handle;
	struct rte_flow_error error;
	uint32_t nb_resources = shared_ctrs.nb_resources;
	struct dpdk_shared_ctrs_elems *elem;

	if (unlikely(counter_id >= nb_resources)) {
		DOCA_LOG_ERR("failed destroying shared counter id(%u) - above num of resources(%u)",
			      counter_id, nb_resources);
		return -EINVAL;
	}

	elem = &shared_ctrs.elems[counter_id];
	handle = elem->handle;
	if (!handle) {
		DOCA_LOG_ERR("failed destroying Shared counter id(%u) - not found",
			      counter_id);
		return -ENOENT;
	}

	port_id = elem->port_id;
	rc = rte_flow_action_handle_destroy(port_id, handle, &error);
	if (rc) {
		DOCA_LOG_ERR("Failed destroying Shared count %u on port %u, type %d message %s",
			      counter_id, port_id, error.type, error.message ?
			      error.message : "(no stated reason)");
		return rc;
	}

	memset(&shared_ctrs.elems[counter_id], 0, sizeof(struct dpdk_shared_ctrs_elems));
	return 0;
}

static int
dpdk_shared_counter_cleanup()
{
	if (!shared_ctrs.nb_resources)
		return 0;

	free(shared_ctrs.elems);
	shared_ctrs.elems = NULL;
	shared_ctrs.nb_resources = 0;
	return 0;
}

static struct dpdk_resource_manager_ops counter_ops = {
	.resources_init = dpdk_shared_counter_init,
	.resources_cleanup = dpdk_shared_counter_cleanup,

	.resource_create = dpdk_shared_counter_create,
	.resource_modify = dpdk_shared_counter_modify,
	.resource_query = dpdk_shared_counter_query,
	.resource_destroy = dpdk_shared_counter_destroy,
};

struct rte_flow_action_handle *
dpdk_shared_counter_get_conf(uint32_t counter_id)
{
	uint32_t nb_resources = shared_ctrs.nb_resources;

	if (counter_id >= nb_resources) {
		DOCA_LOG_RATE_LIMIT_ERR("counter_id (%u) is greater than max counters (%u)",
					counter_id, nb_resources);
		return NULL;
	}

	if (!shared_ctrs.elems || !shared_ctrs.elems[counter_id].handle) {
		DOCA_LOG_RATE_LIMIT_ERR("counter_id (%u) was not created.",
					counter_id);
		return NULL;
	}
	return shared_ctrs.elems[counter_id].handle;
}

struct dpdk_resource_manager_ops *
dpdk_shared_counter_get_ops()
{
	return &counter_ops;
}
