/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <rte_flow.h>

#include "private/common_defines.h"
#include "dpdk_resource_manager.h"
#include <engine_model.h>
#include <engine_bindable.h>
#include <engine_shared_resources.h>
#include "dpdk_shared_crypto.h"
#include <doca_ipsec.h>
#include <ipsec_impl.h>
#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_shared_crypto);

struct dpdk_shared_crypto_action {
	struct rte_flow_action_handle *obj;
	enum engine_crypto_protocol_type proto;
	uint16_t port_id;
};

static struct dpdk_shared_crypto_action *obj_array;
static uint32_t obj_num;

#ifndef DPDK_SUPPORTS_NISP
struct rte_flow_action_handle *
rte_pmd_mlx5_nisp_flow_action_create(uint16_t port_id,
				     const struct rte_flow_indir_action_conf *domain,
				     const struct rte_flow_action *action,
				     struct rte_flow_error *error)
{
	rte_flow_error_set(error, ENOTSUP, RTE_FLOW_ERROR_TYPE_ACTION,
			   action, "NISP: HWS not supported");
	return NULL;
}

int
rte_pmd_mlx5_nisp_flow_action_destroy(uint16_t port_id,
				      const struct rte_flow_action_handle *handle,
				      struct rte_flow_error *error)
{
	return rte_flow_error_set(error, ENOTSUP, RTE_FLOW_ERROR_TYPE_ACTION,
				  handle, "NISP: HWS not supported");

}
#endif

static uint32_t __priv_ipsec_sa_to_id(const struct doca_ipsec_sa *sa)
{
	return sa->id;
}

static struct dpdk_shared_crypto_action *
get_action_by_id(uint32_t id)
{
	if (obj_array == NULL || id >= obj_num)
		return NULL;
	return &obj_array[id];
}

static int
dpdk_crypto_init(uint32_t nr_resources)
{
	if (obj_array != NULL || obj_num != 0)
		return -EEXIST;
	if (nr_resources == 0)
		return 0;
	obj_array = calloc(nr_resources, sizeof(struct dpdk_shared_crypto_action));
	if (obj_array == NULL)
		return -ENOMEM;
	obj_num = nr_resources;
	return 0;
}

static int
dpdk_nisp_create(uint32_t nisp_id, struct engine_shared_resource_cfg *cfg)
{
	struct rte_flow_indir_action_conf domain = {
		.transfer = 1,
	};
	struct rte_pmd_mlx5_flow_action_nisp conf = {
		.reformat_type = RTE_PMD_MLX5_NISP_REFORMAT_TYPE_NONE,
		.crypto_type = RTE_PMD_MLX5_NISP_CRYPTO_TYPE_NONE,
	};
	struct rte_flow_action action = {
		.type = RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_NISP,
		.conf = &conf,
	};
	struct rte_flow_action_ethdev port = {
		.port_id = 0,
	};
	struct rte_flow_action_jump jump = {
		.group = 0,
	};
	struct rte_flow_error error;
	struct dpdk_shared_crypto_action *sh = get_action_by_id(nisp_id);
	uint16_t port_id;

	if (sh == NULL)
		return -EINVAL;
	/* create NISP by config */
	if (cfg == NULL)
		return 0; /* just configuration, not actual creation */
	port_id = cfg->crypto.port_id;

	switch (cfg->crypto.reformat_type) {
	case ENGINE_CRYPTO_REFORMAT_NONE:
		conf.reformat_type = RTE_PMD_MLX5_NISP_REFORMAT_TYPE_NONE;
		break;
	case ENGINE_CRYPTO_REFORMAT_NISP_TUNNEL_ENCAP:
		conf.reformat_type = RTE_PMD_MLX5_NISP_REFORMAT_TYPE_ENCAP;
		break;
	case ENGINE_CRYPTO_REFORMAT_NISP_TUNNEL_DECAP:
		conf.reformat_type = RTE_PMD_MLX5_NISP_REFORMAT_TYPE_DECAP;
		break;
	default:
		return -EINVAL;
	}

	switch (cfg->crypto.crypto_type) {
	case ENGINE_CRYPTO_OPERATION_NONE:
		conf.crypto_type = RTE_PMD_MLX5_NISP_CRYPTO_TYPE_NONE;
		break;
	case ENGINE_CRYPTO_OPERATION_ENCRYPT:
		conf.crypto_type = RTE_PMD_MLX5_NISP_CRYPTO_TYPE_ENCRYPT;
		break;
	case ENGINE_CRYPTO_OPERATION_DECRYPT:
		conf.crypto_type = RTE_PMD_MLX5_NISP_CRYPTO_TYPE_DECRYPT;
		break;
	default:
		return -EINVAL;
	}

	if (cfg->crypto.reformat_data_sz) {
		conf.reformat_data_sz = cfg->crypto.reformat_data_sz;
		conf.reformat_data = cfg->crypto.reformat_data;
	}

	if (cfg->crypto.key_sz) {
		conf.key_sz = cfg->crypto.key_sz;
		conf.key = cfg->crypto.key;
	}

	switch (cfg->crypto.fwd.fwd_type) {
	case ENGINE_FWD_PORT:
		conf.fwd.type = RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT;
		conf.fwd.conf = &port;
		port.port_id = cfg->crypto.fwd.port_id;
		break;
	case ENGINE_FWD_PIPE:
		conf.fwd.type = RTE_FLOW_ACTION_TYPE_JUMP;
		conf.fwd.conf = &jump;
		jump.group = cfg->crypto.fwd.pipe_id;
		break;
	case ENGINE_FWD_DROP:
		conf.fwd.type = RTE_FLOW_ACTION_TYPE_DROP;
		conf.fwd.conf = NULL;
		break;
	default:
		return -EINVAL;
	}

	sh->obj = rte_pmd_mlx5_nisp_flow_action_create(port_id, &domain, &action, &error);
	if (sh->obj == NULL) {
		DOCA_LOG_ERR("failed NISP action creation %u (%s)",
				nisp_id, error.message);
		return -rte_errno;
	}
	sh->port_id = port_id;
	sh->proto = ENGINE_CRYPTO_PROTOCOL_NISP;
	return 0;
}

static int
dpdk_esp_create(uint32_t ipsec_id, struct engine_shared_resource_cfg *cfg)
{
	struct rte_flow_indir_action_conf domain = {
		.transfer = !!engine_model_is_mode(ENGINE_MODEL_MODE_SWITCH),
		.ingress = !engine_model_is_mode(ENGINE_MODEL_MODE_SWITCH)
	};
	struct rte_pmd_mlx5_flow_action_ipsec conf = {
		.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_NONE,
		.crypto_type = RTE_PMD_MLX5_IPSEC_CRYPTO_TYPE_NONE,
	};
	struct rte_flow_action action = {
		.type = RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_IPSEC,
		.conf = &conf,
	};
	struct rte_flow_action_ethdev port = {
		.port_id = 0,
	};
	struct rte_flow_action_jump jump = {
		.group = 0,
	};
	struct rte_flow_error error;
	struct dpdk_shared_crypto_action *sh = get_action_by_id(ipsec_id);
	uint16_t port_id;

	if (sh == NULL)
		return -EINVAL;
	/* create IPsec by config */
	if (cfg == NULL)
		return 0; /* just configuration, not actual creation */
	port_id = cfg->crypto.port_id;

	switch (cfg->crypto.reformat_type) {
	case ENGINE_CRYPTO_REFORMAT_NONE:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_NONE;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV4_ENCAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TRANSPORT_IPV4_ENCAP;
		domain.ingress = 0;
		domain.egress = !domain.transfer;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV6_ENCAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TRANSPORT_IPV6_ENCAP;
		domain.ingress = 0;
		domain.egress = !domain.transfer;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV4_UDP_ENCAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TRANSPORT_IPV4_UDP_ENCAP;
		domain.ingress = 0;
		domain.egress = !domain.transfer;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV6_UDP_ENCAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TRANSPORT_IPV6_UDP_ENCAP;
		domain.ingress = 0;
		domain.egress = !domain.transfer;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TUNNEL_ENCAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TUNNEL_ENCAP;
		domain.ingress = 0;
		domain.egress = !domain.transfer;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_DECAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TRANSPORT_DECAP;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_UDP_DECAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TRANSPORT_UDP_DECAP;
		break;
	case ENGINE_CRYPTO_REFORMAT_ESP_TUNNEL_DECAP:
		conf.reformat_type = RTE_PMD_MLX5_IPSEC_REFORMAT_TYPE_TUNNEL_DECAP;
		break;
	default:
		return -EINVAL;
	}

	switch (cfg->crypto.crypto_type) {
	case ENGINE_CRYPTO_OPERATION_NONE:
		conf.crypto_type = RTE_PMD_MLX5_IPSEC_CRYPTO_TYPE_NONE;
		break;
	case ENGINE_CRYPTO_OPERATION_ENCRYPT:
		conf.crypto_type = RTE_PMD_MLX5_IPSEC_CRYPTO_TYPE_ENCRYPT;
		domain.ingress = 0;
		domain.egress = !domain.transfer;
		break;
	case ENGINE_CRYPTO_OPERATION_DECRYPT:
		conf.crypto_type = RTE_PMD_MLX5_IPSEC_CRYPTO_TYPE_DECRYPT;
		domain.ingress = !domain.transfer;
		domain.egress = 0;
		break;
	default:
		return -EINVAL;
	}

	if (cfg->crypto.reformat_data_sz) {
		conf.reformat_data_sz = cfg->crypto.reformat_data_sz;
		conf.reformat_data = cfg->crypto.reformat_data;
	}

	switch (cfg->crypto.fwd.fwd_type) {
	case ENGINE_FWD_PORT:
		conf.fwd.type = RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT;
		conf.fwd.conf = &port;
		port.port_id = cfg->crypto.fwd.port_id;
		break;
	case ENGINE_FWD_NONE:
	/* Force jump to wire for egress encrypt */
	case ENGINE_FWD_PIPE:
		conf.fwd.type = RTE_FLOW_ACTION_TYPE_JUMP;
		conf.fwd.conf = &jump;
		jump.group = cfg->crypto.fwd.pipe_id;
		break;
	case ENGINE_FWD_DROP:
		conf.fwd.type = RTE_FLOW_ACTION_TYPE_DROP;
		conf.fwd.conf = NULL;
		break;
	default:
		return -EINVAL;
	}

	if (cfg->crypto.security_ctx)
		conf.ipsec_id = __priv_ipsec_sa_to_id(cfg->crypto.security_ctx);
	sh->obj = rte_flow_action_handle_create(port_id, &domain, &action, &error);

	if (sh->obj == NULL) {
		DOCA_LOG_ERR("failed IPsec action creation %u (%s)",
			     ipsec_id, error.message);
		return -rte_errno;
	}
	sh->port_id = port_id;
	sh->proto = ENGINE_CRYPTO_PROTOCOL_ESP;
	return 0;
}

static int
dpdk_crypto_create(uint32_t id, struct engine_shared_resource_cfg *cfg)
{
	switch (cfg->crypto.proto_type) {
	case ENGINE_CRYPTO_PROTOCOL_NISP:
		return dpdk_nisp_create(id, cfg);
	case ENGINE_CRYPTO_PROTOCOL_ESP:
		return dpdk_esp_create(id, cfg);
	default:
		DOCA_LOG_ERR("Invalid crypto action (id=%u) protocol (%d)",
			     id, cfg->crypto.proto_type);
		return -EINVAL;
	}
}

static int
__dpdk_crypto_destroy(struct dpdk_shared_crypto_action *sh)
{
	struct rte_flow_error error;

	if (sh == NULL)
		return -EINVAL;
	if (sh->obj) {
		int rc = 0;

		if (sh->proto == ENGINE_CRYPTO_PROTOCOL_NISP)
			rc = rte_pmd_mlx5_nisp_flow_action_destroy(sh->port_id, sh->obj, &error);

		if (sh->proto == ENGINE_CRYPTO_PROTOCOL_ESP)
			rc = rte_flow_action_handle_destroy(sh->port_id, sh->obj, &error);

		if (rc) {
			DOCA_LOG_ERR(
				"failed crypto action destroy %u (%s)",
				(uint32_t)(sh - obj_array), error.message);
			return rc;
		}
	}
	sh->obj = NULL;
	return 0;
}

static int
dpdk_crypto_destroy(uint32_t id)
{
	struct dpdk_shared_crypto_action *sh = get_action_by_id(id);

	return __dpdk_crypto_destroy(sh);
}

static int
dpdk_crypto_modify(uint32_t id, struct engine_shared_resource_cfg *res_cfg)
{
	struct dpdk_shared_crypto_action *sh = get_action_by_id(id);
	struct dpdk_shared_crypto_action save;
	int rc;

	if (sh == NULL)
		return -EINVAL;
	/* Save current action state for restoration on failure case. */
	if (sh->obj)
		memcpy(&save, sh, sizeof(struct dpdk_shared_crypto_action));
	else
		save.obj = NULL;
	rc = dpdk_crypto_create(id, res_cfg);
	/* If creation failed - restore the preceeding action state. */
	if (rc != 0) {
		memcpy(sh, &save, sizeof(struct dpdk_shared_crypto_action));
		return rc;
	}
	__dpdk_crypto_destroy(&save);
	return 0;
}

static int
dpdk_crypto_query(uint32_t id, struct engine_shared_resource_query_result *res_query_rslts)
{
	return -ENOTSUP;
}

static int
dpdk_crypto_cleanup(void)
{
	struct dpdk_shared_crypto_action *sh = obj_array;
	uint32_t n = obj_num;

	obj_array = NULL;
	obj_num = 0;
	if (sh == NULL)
		return 0;
	/* destroy all crypto actions */
	while (n != 0) {
		--n;
		if (sh[n].obj) {
			DOCA_LOG_RATE_LIMIT_ERR("Crypto id %u is still in use\n", n);
			dpdk_crypto_destroy(n);
		}
	}
	free(sh);
	return 0;
}

struct rte_flow_action_handle *
dpdk_shared_crypto_translate(uint32_t id)
{
	struct dpdk_shared_crypto_action *sh = get_action_by_id(id);

	if (sh == NULL)
		return NULL;
	if (sh->obj)
		return sh->obj;
	/* We do not support implicit binding to pipe being created. */
	return NULL;
}

static struct dpdk_resource_manager_ops crypto_ops = {
	.resources_init = dpdk_crypto_init,
	.resources_cleanup = dpdk_crypto_cleanup,

	.resource_create = dpdk_crypto_create,
	.resource_modify = dpdk_crypto_modify,
	.resource_query = dpdk_crypto_query,
	.resource_destroy = dpdk_crypto_destroy,
};

struct dpdk_resource_manager_ops *
dpdk_shared_crypto_get_ops()
{
	return &crypto_ops;
}
