/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <errno.h>

#include <rte_mtr.h>

#include "private/common_defines.h"
#include "dpdk_resource_manager.h"
#include "dpdk_meter_controller.h"

#include <doca_log.h>

#include "engine_model.h"

DOCA_LOG_REGISTER(dpdk_shared_meter);

#define METER_MARK_STATE_ON 1

struct dpdk_shared_meters_elems {
	struct rte_flow_action_handle *handle;
	uint16_t port_id;
	uint64_t cir;
	uint64_t cbs;
	uint16_t port_set : 1;
};

struct dpdk_shared_meters {
	uint32_t nb_resources;
	struct dpdk_shared_meters_elems *elems;
};

struct dpdk_shared_meters shared_meters = {
	.nb_resources = 0,
	.elems = NULL,
};

static int
dpdk_shared_meter_init(uint32_t nb_resources)
{
	/* index range starts from 1 */
	shared_meters.elems = (struct dpdk_shared_meters_elems *)calloc(nb_resources,
			    sizeof(struct dpdk_shared_meters_elems));
	if (unlikely(!shared_meters.elems)) {
		DOCA_LOG_ERR("failed to allocate %u shared meters", nb_resources);
		return -ENOMEM;
	}

	shared_meters.nb_resources = nb_resources;

	return 0;
}

static struct rte_flow_action_handle *
__dpdk_shared_meter_create(uint32_t meter_id, struct engine_shared_resource_cfg *res_cfg,
			   struct dpdk_shared_meters_elems *meter)
{
	struct rte_flow_error flow_error;
	struct rte_flow_action_handle *handle;
	int ret;
	uint16_t port_id;
	struct rte_flow_action_meter_mark meter_mark_action_conf;
	uint32_t profile_id = meter_id;
	uint32_t transfer = engine_model_is_mode(ENGINE_MODEL_MODE_SWITCH) ? 1 : 0;

	meter->port_id = res_cfg->meter.port_id;
	meter->cir = res_cfg->meter.cir;
	meter->cbs = res_cfg->meter.cbs;

	ret = dpdk_meter_profile_create(meter->port_id, profile_id,
					       meter->cir, meter->cbs);
	if (unlikely(ret))
		return NULL;

	/* Create indirect action of modify meter color */
	const struct rte_flow_indir_action_conf indir_conf = {
		.ingress = transfer ^ 1,
		.egress = 0,
		.transfer = transfer,
	};

	port_id = meter->port_id;
	dpdk_meter_mark_conf_build(port_id, profile_id, &meter_mark_action_conf);
	struct rte_flow_action meter_mark_action = {
		.type = RTE_FLOW_ACTION_TYPE_METER_MARK,
		.conf = &meter_mark_action_conf,
	};

	handle = rte_flow_action_handle_create(port_id, &indir_conf, &meter_mark_action,
					       &flow_error);
	if (unlikely(!handle))
		DOCA_LOG_ERR("Shared meter %u create fail on port %u, type %d message %s",
			      meter_id, port_id, flow_error.type,
			      flow_error.message ?  flow_error.message :
						    "(no stated reason)");
	return handle;
}

static int
dpdk_shared_meter_create(uint32_t meter_id, struct engine_shared_resource_cfg *res_cfg)
{
	struct rte_flow_action_handle *handle;

	if (unlikely(!shared_meters.nb_resources)) {
		DOCA_LOG_ERR("Shared meters were not initialized");
		return -EINVAL;
	}

	if (unlikely(!res_cfg)) {
		DOCA_LOG_ERR("Shared meter %u - missing configuration",
			      meter_id);
		return -EINVAL;
	}

	if (unlikely(shared_meters.elems[meter_id].handle)) {
		DOCA_LOG_ERR("Shared meter %u was already created", meter_id);
		return -EALREADY;
	}

	handle = __dpdk_shared_meter_create(meter_id, res_cfg,
					    &shared_meters.elems[meter_id]);
	if (unlikely(!handle)) {
		DOCA_LOG_ERR("failed to create shared meter (%u)", meter_id);
		return -EINVAL;
	}

	shared_meters.elems[meter_id].handle = handle;

	return 0;
}

static int
__dpdk_shared_meter_modify(uint32_t meter_id, struct engine_shared_resource_cfg *res_cfg)
{
	uint16_t port_id;
	struct rte_flow_action_handle *handle;
	struct rte_flow_action_meter_mark meter_mark;
	int ret;
	struct rte_flow_error flow_error;
	uint32_t profile_id = meter_id;

	shared_meters.elems[meter_id].port_id = res_cfg->meter.port_id;
	shared_meters.elems[meter_id].cir = res_cfg->meter.cir;
	shared_meters.elems[meter_id].cbs = res_cfg->meter.cbs;

	handle = shared_meters.elems[meter_id].handle;
	if (unlikely(!handle)) {
		DOCA_LOG_ERR("failed to modify non-created shared meter (%u)", meter_id);
		return -ENOENT;
	}

	port_id = shared_meters.elems[meter_id].port_id;
	ret = dpdk_meter_profile_delete(port_id, profile_id);
	if (unlikely(ret))
		return ret;

	ret = dpdk_meter_profile_create(port_id, profile_id,
					       res_cfg->meter.cir, res_cfg->meter.cbs);
	if (unlikely(ret))
		return ret;

	dpdk_meter_mark_conf_build(port_id, profile_id, &meter_mark);

	struct rte_flow_update_meter_mark updated_meter = {
		.meter_mark = meter_mark,
		.profile_valid = 1,
	};

	ret = rte_flow_action_handle_update(port_id, handle, &updated_meter, &flow_error);
	if (unlikely(ret)) {
		DOCA_LOG_ERR("failed to update shared meter %u message: %s",
			      meter_id,
			      flow_error.message ? flow_error.message :
						   "(no stated reason)");
		return ret;
	}

	return 0;
}

static int
dpdk_shared_meter_modify(uint32_t meter_id, struct engine_shared_resource_cfg *res_cfg)
{
	int ret;
	struct rte_flow_action_handle *handle;

	if (unlikely(!shared_meters.nb_resources)) {
		DOCA_LOG_ERR("failed to modify shared meter id (%u) - resource not initialized",
			     meter_id);
		return -EINVAL;
	}
	if (unlikely(!res_cfg)) {
		DOCA_LOG_ERR("failed to modify shared meter id (%u) - missing configuration",
			      meter_id);
		return -EINVAL;
	}

	shared_meters.elems[meter_id].cir = res_cfg->meter.cir;
	shared_meters.elems[meter_id].cbs = res_cfg->meter.cbs;

	handle = shared_meters.elems[meter_id].handle;
	if (unlikely(!handle)) {
		DOCA_LOG_ERR("failed to modify non-created shared meter (%u)", meter_id);
		return -ENOENT;
	}

	ret = __dpdk_shared_meter_modify(meter_id, res_cfg);
	if (unlikely(ret)) {
		DOCA_LOG_ERR("failed to modify shared meter (%u)", meter_id);
		return ret;
	}

	return 0;
}

static int
dpdk_shared_meter_query(uint32_t meter_id,
		 struct engine_shared_resource_query_result *res_query_rslt)
{
	return -ENOTSUP;
}

static int
dpdk_shared_meter_destroy(uint32_t meter_id)
{
	uint16_t port_id;
	struct rte_flow_action_handle *handle;
	struct rte_flow_error flow_error;
	struct rte_mtr_error mtr_error;
	uint32_t nb_resources = shared_meters.nb_resources;
	uint32_t profile_id = meter_id;

	if (unlikely(meter_id > nb_resources)) {
		DOCA_LOG_ERR("meter id(%u) is above meter resources(%u)",
			      meter_id, nb_resources);
		return -EINVAL;
	}

	handle = shared_meters.elems[meter_id].handle;
	if (unlikely(!handle)) {
		DOCA_LOG_ERR("Shared counter id(%u) was not created",
			      meter_id);
		return -EINVAL;
	}

	port_id = shared_meters.elems[meter_id].port_id;
	if (unlikely(rte_flow_action_handle_destroy(port_id, handle, &flow_error))) {
		DOCA_LOG_ERR("Shared meter %u destroy fail on port %u, type %d message %s",
			      meter_id, port_id, flow_error.type,
			      flow_error.message ?
			      flow_error.message : "(no stated reason)");
		return -EINVAL;
	}

	if (unlikely(rte_mtr_meter_profile_delete(port_id, profile_id, &mtr_error))) {
		DOCA_LOG_ERR("Shared meter profile %u destroy fail on port %u, type %d message %s",
			      meter_id, port_id, mtr_error.type, mtr_error.message ?
			      mtr_error.message : "(no stated reason)");
		return -EINVAL;
	}

	memset(&shared_meters.elems[meter_id], 0, sizeof(shared_meters.elems[0]));

	return 0;
}

static int
dpdk_shared_meter_cleanup()
{
	if (!shared_meters.elems)
		return 0;

	free(shared_meters.elems);
	shared_meters.elems = NULL;
	shared_meters.nb_resources = 0;
	return 0;
}

static struct dpdk_resource_manager_ops meter_ops = {
	.resources_init = dpdk_shared_meter_init,
	.resources_cleanup = dpdk_shared_meter_cleanup,

	.resource_create = dpdk_shared_meter_create,
	.resource_modify = dpdk_shared_meter_modify,
	.resource_query = dpdk_shared_meter_query,
	.resource_destroy = dpdk_shared_meter_destroy,
};

struct rte_flow_action_handle *
dpdk_shared_meter_get_conf(uint32_t meter_id)
{
	uint32_t nb_resources = shared_meters.nb_resources;

	if (unlikely(meter_id >= nb_resources)) {
		DOCA_LOG_ERR("meter_id (%u) is greater than max meters (%u)",
			      meter_id, nb_resources);
		return NULL;
	}

	if (unlikely(!shared_meters.elems || !shared_meters.elems[meter_id].handle)) {
		DOCA_LOG_ERR("meter_id (%u) is unbounded", meter_id);
		return NULL;
	}

	return shared_meters.elems[meter_id].handle;
}

struct dpdk_resource_manager_ops *
dpdk_shared_meter_get_ops()
{
	return &meter_ops;
}
