/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <stdlib.h>
#include <errno.h>

#include "private/common_defines.h"
#include "dpdk_port.h"
#include "dpdk_resource_manager.h"
#include "dpdk_flow.h"
#include "dpdk_flow_single.h"
#include "dpdk_fwd_groups.h"

#include <rte_flow.h>
#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_shared_rss);

#define SHARED_RSS_ITEM_NB (2)
#define SHARED_RSS_ACTION_NB (2)
#define SHARED_RSS_ACT_TEMPLATE_IDX (0)
#define SHARED_RSS_ITEM_TEMPLATE_IDX (0)
#define SHARED_RSS_NB_ACTIONS (1)
#define SHARED_RSS_NB_FLOWS (1)

struct dpdk_shared_rss_ctx {
	struct dpdk_port *port;
	struct dpdk_flow_queue *mgmt_queue;
	rte_spinlock_t *mgmt_lock;
	struct dpdk_flow_single *flow_single;
	bool is_created;
};

struct dpdk_shared_rss {
	struct dpdk_shared_rss_ctx *rss_ctx;
	uint32_t nb_rss;
};

struct dpdk_shared_rss shared_rss = {0};

static int
shared_rss_nb_get(void)
{
	return shared_rss.nb_rss;
}

static struct dpdk_shared_rss_ctx *
shared_rss_ctx_get(uint32_t rss_id)
{
	uint32_t nb_rss;

	nb_rss = shared_rss_nb_get();
	if (rss_id >= nb_rss) {
		DOCA_LOG_RATE_LIMIT_ERR("failed getting rss ctx - invalid rss_id %u, max rss %u",
			rss_id, nb_rss);
		return NULL;
	}

	return &shared_rss.rss_ctx[rss_id];
}

static int
shared_rss_verify(uint32_t rss_id)
{
	uint32_t nb_rss;

	nb_rss = shared_rss_nb_get();
	if (unlikely(rss_id >= nb_rss)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed verifying rss_id %u - larger than nr_resource %u",
				rss_id, nb_rss);
		return -EINVAL;
	}

	if (unlikely(!shared_rss.rss_ctx)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed verifying rss_id %u - rss not initialized", rss_id);
		return -EINVAL;
	}

	return 0;
}

int
dpdk_shared_rss_get_group(uint32_t rss_id, uint32_t *group_id)
{
	struct dpdk_shared_rss_ctx *ctx;
	int ret;

	ret = shared_rss_verify(rss_id);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failure getting rss group - verification failed for rss_id %u", rss_id);
		return ret;
	}

	ctx = shared_rss_ctx_get(rss_id);
	if (unlikely(!ctx)) {
		DOCA_LOG_RATE_LIMIT_ERR("failure getting rss group - null context for rss_id %u", rss_id);
		return -EINVAL;
	}

	if (unlikely(!ctx->flow_single)) {
		DOCA_LOG_RATE_LIMIT_ERR("failure getting rss group - null flow_single for rss_id %u", rss_id);
		return -EINVAL;
	}

	if (!ctx->flow_single->in_hw) {
		DOCA_LOG_RATE_LIMIT_ERR("failure getting rss group - flow_single not in_hw for rss_id %u", rss_id);
		return -EINVAL;
	}

	*group_id = ctx->flow_single->group_id;
	return 0;
}

static int
dpdk_shared_rss_init(uint32_t nr_resources)
{
	if (nr_resources == 0) {
		DOCA_LOG_RATE_LIMIT_ERR("failed initiating rss - invalid nr_resource %u", nr_resources);
		return -EINVAL;
	}

	shared_rss.nb_rss = nr_resources;
	shared_rss.rss_ctx = calloc(nr_resources, sizeof(struct dpdk_shared_rss_ctx));
	if (unlikely(!shared_rss.rss_ctx)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed initiating rss - alloc failure.");
		return -ENOMEM;
	}

	return 0;
}

static void
build_items(struct rte_flow_item *items)
{
	int nb_count = 0;

	items[nb_count++].type = RTE_FLOW_ITEM_TYPE_ETH;
	items[nb_count].type = RTE_FLOW_ITEM_TYPE_END;
}

static void
build_actions(struct engine_shared_resource_cfg *res_cfg, struct rte_flow_action *actions, struct rte_flow_action_rss *rss)
{
	int nb_count = 0;

	rss->queue = res_cfg->rss.queues_array;
	rss->queue_num = res_cfg->rss.nr_queues;
	rss->types = res_cfg->rss.types;
	rss->func = RTE_ETH_HASH_FUNCTION_DEFAULT;

	actions[nb_count].type = RTE_FLOW_ACTION_TYPE_RSS;
	actions[nb_count++].conf = rss;
}

static void
fill_dpdk_flow_params(uint32_t group_id,
		      struct rte_flow_item *items,
		      struct rte_flow_action *actions,
		      struct dpdk_flow_params *params)
{
	params->items = items;
	params->actions = actions;
	params->table = NULL;

	params->action_template_index = SHARED_RSS_ACT_TEMPLATE_IDX;
	params->item_template_index = SHARED_RSS_ITEM_TEMPLATE_IDX;

	params->ingress = 1;
	params->group = group_id;
}

static struct dpdk_flow_single *
flow_single_alloc()
{
	return calloc(1, sizeof(struct dpdk_flow_single));
}

static void
flow_single_free(struct dpdk_flow_single *flow_single)
{
	if (flow_single)
		free(flow_single);
}

static int
dpdk_shared_rss_create(uint32_t rss_id, struct engine_shared_resource_cfg *res_cfg)
{
	struct dpdk_shared_rss_ctx *rss_ctx;
	struct rte_flow_item items[SHARED_RSS_ITEM_NB] = {0};
	struct rte_flow_action actions[SHARED_RSS_ACTION_NB] = {0};
	struct rte_flow_action_rss rss = {0};
	struct dpdk_flow_params dpdk_flow_params = {0};
	struct dpdk_port *port;
	struct dpdk_group_pool *group_pool;
	struct dpdk_flow_queue *mgmt_queue;
	rte_spinlock_t *mgmt_lock;
	struct dpdk_flow_single *flow_single;
	uint16_t port_id;
	uint32_t group_id;
	int ret;

	ret = shared_rss_verify(rss_id);
	if (ret)
		return ret;

	rss_ctx = shared_rss_ctx_get(rss_id);
	if (unlikely(!rss_ctx))
		return -EINVAL;

	/* get group_pool/mgmt_queue/mgmt_lock */
	port_id = res_cfg->rss.port_id;
	port = dpdk_port_get_by_id(port_id);
	if (unlikely(!port)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation shared_rss - null dpdk_port");
		return -EINVAL;
	}

	group_pool = dpdk_port_get_group_pool(port);
	if (unlikely(!group_pool)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation shared_rss - null group_pool");
		return -EINVAL;
	}

	ret = dpdk_fwd_groups_cfg_get_by_id(port_id, (void **)&mgmt_queue, (void **)&mgmt_lock);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation shared_rss - null fwd_groups");
		return -EINVAL;
	}

	/* allocate group_id */
	flow_single = flow_single_alloc();
	if (unlikely(!flow_single)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation shared_rss - alloc flow_single failed");
		return -ENOMEM;
	}

	ret = dpdk_group_pool_alloc(group_pool, &group_id);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creation dpdk_fwd_group - alloc group_id failed");
		flow_single_free(flow_single);
		return ret;
	}

	/* create flows */
	build_items(items);
	build_actions(res_cfg, actions, &rss);
	fill_dpdk_flow_params(group_id, items, actions, &dpdk_flow_params);

	flow_single->group_id = group_id;
	ret = dpdk_flow_single_create(port, &dpdk_flow_params,
				      mgmt_queue, mgmt_lock,
				      SHARED_RSS_NB_ACTIONS, SHARED_RSS_NB_FLOWS, flow_single);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create shared_rss(id=%u) - create flow failure ret=%d", rss_id, ret);
		goto error;
	}

	/* save information */
	rss_ctx->port = port;
	rss_ctx->flow_single = flow_single;
	rss_ctx->mgmt_lock = mgmt_lock;
	rss_ctx->mgmt_queue = mgmt_queue;
	rss_ctx->is_created = true;

	return 0;
error:
	dpdk_group_pool_free(group_pool, group_id);
	flow_single_free(flow_single);
	return ret;
}

static int
dpdk_shared_rss_modify(uint32_t rss_id, struct engine_shared_resource_cfg *res_cfg)
{
	/* delete the existed flow and create new flow based on res_cfg, no change for group_id */

	return 0;
}

static int
dpdk_shared_rss_query(uint32_t rss_id,
		      struct engine_shared_resource_query_result *res_query_rslt)
{
	/* add counter action when creating flow for shared_rss */

	return -ENOTSUP;
}

static int
dpdk_shared_rss_destroy(uint32_t rss_id)
{
	struct dpdk_shared_rss_ctx *rss_ctx;
	struct dpdk_flow_single *flow_single;
	struct dpdk_group_pool *group_pool;
	int ret;

	ret = shared_rss_verify(rss_id);
	if (ret)
		return ret;

	rss_ctx = shared_rss_ctx_get(rss_id);
	if (unlikely(!rss_ctx))
		return -EINVAL;

	if (!rss_ctx->is_created) {
		DOCA_LOG_RATE_LIMIT_WARN("shared_rss %u is not created or destroyed", rss_id);
		return 0;
	}

	group_pool = dpdk_port_get_group_pool(rss_ctx->port);
	if (unlikely(!group_pool)) {
		DOCA_LOG_RATE_LIMIT_WARN("shared_rss %u destory failure - NULL group_pool", rss_id);
		return 0;
	}

	flow_single = rss_ctx->flow_single;
	dpdk_flow_single_destroy(rss_ctx->port, rss_ctx->mgmt_queue, rss_ctx->mgmt_lock, flow_single);
	ret = dpdk_group_pool_free(group_pool, flow_single->group_id);
	if (ret)
		DOCA_LOG_RATE_LIMIT_WARN("shared_rss %u free group_id %u failure.", rss_id, flow_single->group_id);

	flow_single_free(flow_single);
	rss_ctx->is_created = false;
	return ret;
}

static int
dpdk_shared_rss_cleanup()
{
	uint32_t nb_rss;
	int i;

	if (unlikely(!shared_rss.rss_ctx)) {
		DOCA_LOG_RATE_LIMIT_WARN("shared_rss cleanup - rss_ctx is NULL");
		return 0;
	}

	nb_rss = shared_rss_nb_get();
	for (i = 0; i < nb_rss; i++)
		dpdk_shared_rss_destroy(i);

	free(shared_rss.rss_ctx);
	shared_rss.nb_rss = 0;
	shared_rss.rss_ctx = NULL;

	return 0;
}

static struct dpdk_resource_manager_ops rss_ops = {
	.resources_init = dpdk_shared_rss_init,
	.resources_cleanup = dpdk_shared_rss_cleanup,

	.resource_create = dpdk_shared_rss_create,
	.resource_modify = dpdk_shared_rss_modify,
	.resource_query = dpdk_shared_rss_query,
	.resource_destroy = dpdk_shared_rss_destroy,
};

struct dpdk_resource_manager_ops *
dpdk_shared_rss_get_ops()
{
	return &rss_ops;
}
