/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <rte_flow.h>

#include "dpdk_table.h"
#include "private/common_defines.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_table);

static struct dpdk_table_ops table_ops = {0};

static int
validate_ops(struct dpdk_table_ops *table_ops)
{
	if (!table_ops->module_init ||
	    !table_ops->module_cleanup ||
	    !table_ops->port_hook ||
	    !table_ops->port_unhook ||
	    !table_ops->table_create ||
	    !table_ops->table_destroy)
		return -EINVAL;

	return 0;
}

int
dpdk_table_init(struct dpdk_table_cfg *cfg)
{
	int rc;

	if (unlikely(!cfg)) {
		DOCA_LOG_ERR("failed initializing dpdk table - cfg is null");
		return -EINVAL;
	}

	if (unlikely(!cfg->ops)) {
		DOCA_LOG_ERR("failed initializing dpdk table - ops is null");
		return -EINVAL;
	}

	rc = validate_ops(cfg->ops);
	if (rc) {
		DOCA_LOG_ERR("failed initializing dpdk table - some ops have null function pointers");
		return rc;
	}

	table_ops = *cfg->ops;
	rc = table_ops.module_init(cfg);
	if (rc) {
		DOCA_LOG_ERR("failed initializing dpdk table rc=%d", rc);
		return rc;
	}

	DOCA_LOG_INFO("Initializing dpdk table successfully");
	return 0;
}

int
dpdk_table_cleanup(void)
{
	DOCA_LOG_INFO("Cleanup dpdk table");
	if (unlikely(!table_ops.module_cleanup))
		return 0;
	return table_ops.module_cleanup();
}

struct dpdk_table_port_ctx *
dpdk_table_port_ctx_create(uint16_t port_id)
{
	return table_ops.port_hook(port_id);
}

void
dpdk_table_port_ctx_destroy(struct dpdk_table_port_ctx *ctx)
{
	table_ops.port_unhook(ctx);
}

int
dpdk_table_create(struct dpdk_table_params *dpdk_table_params,
		  struct dpdk_table_port_ctx *ctx,
		  struct dpdk_table *res)
{
	if (unlikely(!dpdk_table_params || !res)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating table on port %u - null arguments",
					dpdk_table_params->port_id);
		return -EINVAL;
	}

	if (unlikely(dpdk_table_params->nb_actions > MAX_ACT_TEMPLATE)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating table on port %u - nb_actions %u is larger than %d",
					dpdk_table_params->nb_actions,
					dpdk_table_params->port_id, MAX_ACT_TEMPLATE);
		return -EINVAL;
	}

	return table_ops.table_create(dpdk_table_params, ctx, res);
}

int
dpdk_table_destroy(struct dpdk_table_port_ctx *ctx, struct dpdk_table *res)
{
	if (unlikely(!res)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed destroying flow - res pointer is null");
		return -EINVAL;
	}

	return table_ops.table_destroy(ctx, res);
}
