/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DPDK_TABLE_H_
#define DPDK_TABLE_H_

#include <stdint.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#endif

struct rte_flow_item;
struct rte_flow_attr;
struct rte_flow_action;
struct rte_flow_pattern_template;
struct rte_flow_actions_template;
struct rte_flow_template_table;

struct dpdk_table_port_ctx;

#define MAX_ACT_TEMPLATE 4
#define MAX_PAT_TEMPLATE 1

enum dpdk_table_insertion_type {
	DPDK_TABLE_INSERTION_TYPE_PATTERN,
	DPDK_TABLE_INSERTION_TYPE_INDEX,
	DPDK_TABLE_INSERTION_TYPE_MAX,
};

enum dpdk_table_hash_type {
	DPDK_TABLE_HASH_TYPE_DEFAULT,
	DPDK_TABLE_HASH_TYPE_IDENTITY,
	DPDK_TABLE_HASH_TYPE_CRC32,
	DPDK_TABLE_HASH_TYPE_CRC16,
	DPDK_TABLE_HASH_TYPE_MAX,
};

/**
 * @brief Dpdk table input params
 */
struct dpdk_table_params {
	uint16_t port_id;
	/**< Dpdk port_id */
	struct rte_flow_item *items;
	/**< items array */
	struct rte_flow_action **actions;
	/**< actions arrays */
	struct rte_flow_action **actions_mask;
	/**< actions mask arrays */
	uint8_t nb_actions;
	/**< number of actions and actions_mask arrays */

	struct rte_flow_attr *flow_attr;
	/**< dpdk flow attributes */
	uint32_t nb_flows;
	/**< Table size */

	enum dpdk_table_insertion_type insertion_type;
	/**< DPDK insertion type */
	enum dpdk_table_hash_type hash_type;
	/**< DPDK hash function type */
};

/**
 * @brief Dpdk table configuration
 */
struct dpdk_table_cfg {
	struct dpdk_table_ops *ops;
	/**< Dpdk table ops */
};

/**
 * @brief Dpdk table create result
 */
struct dpdk_table {
	uint16_t port_id;
	/** <Dpdk port_id */
	struct rte_flow_template_table *table;
	/** < pointer to table object */
	uint8_t pattern_template_index;
	/** < pattern template index */
	uint8_t action_template_index;
	/** < actions template index */
};


/**
 * @brief Dpdk table ops
 *
 * A struct to define the ops.
 */
struct dpdk_table_ops {
	/**
	 * @brief called on module initialization
	 *
	 * @param flow_cfg
	 * Pointer to a dpdk flow configuration
	 * @return
	 * 0 on success, otherwise an error code
	 */
	int (*module_init)(struct dpdk_table_cfg *table_cfg);

	/**
	 * @brief called on module teardown
	 *
	 * @return
	 * 0 on success, otherwise an error code
	 */
	int (*module_cleanup)();

	/**
	 * @brief called upon each dpdk port start to initialize module's
	 * resources.
	 *
	 * @param port_id
	 * Dpdk port_id
	 * @return
	 * pointer to struct dpdk_table_port_ctx on success, NULL on failure
	 */
	struct dpdk_table_port_ctx *(*port_hook)(uint16_t port_id);

	void (*port_unhook)(struct dpdk_table_port_ctx *ctx);

	/**
	 * @brief create a new table.
	 *
	 * @param dpdk_table_params
	 * Pointer to the dpdk_table_params
	 * @param ctx
	 * Pointer to struct dpdk_table_port_ctx which is returned from port_hook
	 * @param res
	 * output parameter for the table and indices created
	 * @return
	 * 0 on success, otherwise the following error codes:
	 * -EINVAL on invalid params
	 */
	int (*table_create)(struct dpdk_table_params *dpdk_table_params,
			    struct dpdk_table_port_ctx *ctx,
			    struct dpdk_table *res);

	/**
	 * @brief destroy existing table.
	 *
	 * @param port_id
	 * dpdk port_id.
	 * @param tracker
	 * Pointer to a table.
	 * @return
	 * 0 on success, otherwise the following error codes:
	 * -EINVAL on invalid params
	 */
	int (*table_destroy)(struct dpdk_table_port_ctx *ctx, struct dpdk_table *table);
};

/**
 * @brief Initialize dpdk table module.
 *
 * @param cfg
 * Pointer to configuration
 * @return
 * 0 on success, otherwise error code.
 */
int
dpdk_table_init(struct dpdk_table_cfg *cfg);

int
dpdk_table_cleanup(void);

/**
 * @brief called upon each dpdk port start
 *
 * @param port_id
 * Dpdk port_id
 * @return
 * 0 on success, -1 on failure
 */
struct dpdk_table_port_ctx *
dpdk_table_port_ctx_create(uint16_t port_id);

void
dpdk_table_port_ctx_destroy(struct dpdk_table_port_ctx *ctx);

/**
 * @brief Create dpdk table
 *
 * @param dpdk_table_params
 * Pointer to the dpdk_table_params
 * @param ctx
 * Pointer to struct dpdk_table_port_ctx which is returned from port_hook
 * @param res
 * output parameter for the table and indices created
 * @return
 * 0 on success, otherwise error code.
 */
int
dpdk_table_create(struct dpdk_table_params *dpdk_table_params, struct dpdk_table_port_ctx *ctx,
		  struct dpdk_table *res);

/**
 * @brief Destroy dpdk table
 *
 * @param ctx
 * Pointer to struct dpdk_table_port_ctx which is returned from port_hook
 * @param res
 * Pointer to a valued returned from dpdk_table_create.
 * @return
 * positive value indicating the table's refcount, 0 if the table is no
 * longer used, otherwise error code.
 */
int
dpdk_table_destroy(struct dpdk_table_port_ctx *ctx, struct dpdk_table *table);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DPDK_TABLE_H_ */
