/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <inttypes.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#include "dpdk_table_hws.h"
#include "dpdk_table.h"
#include "dpdk_pipe.h"
#include "dpdk_shared_crypto.h"

#include <rte_flow.h>
#include <rte_hash_crc.h>
#include <utils_hash_table.h>

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_table_hws);

#define TABLE_NAME_BUF_SIZE 32

#define MAX_ACTIONS_TEMPLATES 256
#define MAX_PATTERN_TEMPLATES 256
#define MAX_TABLES 1024

struct dpdk_table_port_ctx {
	struct utils_hash_table *dpdk_template_actions;
	struct utils_hash_table *dpdk_template_items;
	struct utils_hash_table *dpdk_table_mgr;
};

union actions_conf {
	struct rte_flow_action_jump jump;
	struct rte_flow_action_rss rss;
	struct rte_flow_action_mark mark;
	struct rte_flow_action_raw_decap raw_decap;
	struct rte_flow_action_meter meter;
	struct rte_flow_action_queue queue;
	struct rte_flow_action_ethdev ethdev;
	struct rte_flow_action_modify_field modify_field;
	struct rte_flow_action_raw_encap raw_encap;
	struct rte_flow_action_count count;
	struct rte_flow_action_handle *handle;
	struct rte_flow_action_meter_mark meter_mark;
};

struct dpdk_template_actions_item {
	enum rte_flow_action_type type;
	union actions_conf conf;
	bool conf_is_set;
};

union items_mask {
	struct rte_flow_item_eth eth;
	struct rte_flow_item_vlan vlan;
	struct rte_flow_item_ipv4 ipv4;
	struct rte_flow_item_ipv6 ipv6;
	struct rte_flow_item_tcp tcp;
	struct rte_flow_item_udp udp;
	struct rte_flow_item_gtp gtp;
	struct rte_flow_item_vxlan vxlan;
	struct rte_flow_item_gre gre;
	uint32_t gre_key;
	uint32_t meta;
	struct rte_flow_item_ethdev ethdev;
	struct rte_flow_item_tag tag;
	struct mlx5_flow_item_nisp nisp;
	struct mlx5_flow_item_nisp_syndrome nisp_syndrome;
	struct mlx5_flow_item_audp audp;
	struct rte_flow_item_meter_color meter_color;
	struct rte_flow_item_esp esp;
	struct mlx5_flow_item_ipsec_syndrome ipsec_syndrome;
};

struct dpdk_template_items_item {
	enum rte_flow_item_type type;
	union items_mask mask;
	bool mask_is_set;
};

struct dpdk_template_actions_key {
	struct dpdk_template_actions_item actions[DPDK_PIPE_MAX_ACTIONS];
	struct dpdk_template_actions_item mask[DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_actions_template_attr at_attr;
};

struct dpdk_template_items_key {
	struct dpdk_template_items_item items[DPDK_PIPE_MAX_ITEMS];
	struct rte_flow_pattern_template_attr it_attr;
};

struct dpdk_table_mgr_key {
	void *it[MAX_PAT_TEMPLATE];
	uint8_t nb_pattern_templates;
	void *at[MAX_ACT_TEMPLATE];
	uint8_t nb_actions_templates;
	struct rte_flow_template_table_attr tatr;
};

/** Items template manager **/
static inline int
cmp_items(enum rte_flow_item_type type, const union items_mask *dst_mask,
	  const union items_mask *src_mask)
{
	switch ((int)type) {
	case RTE_FLOW_ITEM_TYPE_ETH:
		return memcmp(&dst_mask->eth, &src_mask->eth, sizeof(struct rte_flow_item_eth));
	case RTE_FLOW_ITEM_TYPE_VLAN:
		return memcmp(&dst_mask->vlan, &src_mask->vlan, sizeof(struct rte_flow_item_vlan));
	case RTE_FLOW_ITEM_TYPE_IPV4:
		return memcmp(&dst_mask->ipv4, &src_mask->ipv4, sizeof(struct rte_flow_item_ipv4));
	case RTE_FLOW_ITEM_TYPE_IPV6:
		return memcmp(&dst_mask->ipv6, &src_mask->ipv6, sizeof(struct rte_flow_item_ipv6));
	case RTE_FLOW_ITEM_TYPE_TCP:
		return memcmp(&dst_mask->tcp, &src_mask->tcp, sizeof(struct rte_flow_item_tcp));
	case RTE_FLOW_ITEM_TYPE_UDP:
		return memcmp(&dst_mask->udp, &src_mask->udp, sizeof(struct rte_flow_item_udp));
	case RTE_FLOW_ITEM_TYPE_GTP:
		return memcmp(&dst_mask->gtp, &src_mask->gtp, sizeof(struct rte_flow_item_gtp));
	case RTE_FLOW_ITEM_TYPE_VXLAN:
		return memcmp(&dst_mask->vxlan, &src_mask->vxlan, sizeof(struct rte_flow_item_vxlan));
	case RTE_FLOW_ITEM_TYPE_GRE:
		return memcmp(&dst_mask->gre, &src_mask->gre, sizeof(struct rte_flow_item_gre));
	case RTE_FLOW_ITEM_TYPE_GRE_KEY:
		return dst_mask->gre_key == src_mask->gre_key;
	case RTE_FLOW_ITEM_TYPE_META:
		return dst_mask->meta == src_mask->meta;
	case RTE_FLOW_ITEM_TYPE_REPRESENTED_PORT:
		return memcmp(&dst_mask->ethdev, &src_mask->ethdev, sizeof(struct rte_flow_item_ethdev));
	case RTE_FLOW_ITEM_TYPE_TAG:
		return memcmp(&dst_mask->tag, &src_mask->tag, sizeof(struct rte_flow_item_tag));
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP:
		return memcmp(&dst_mask->nisp, &src_mask->nisp, sizeof(dst_mask->nisp));
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP_SYNDROME:
		return memcmp(&dst_mask->nisp_syndrome, &src_mask->nisp_syndrome, sizeof(dst_mask->nisp_syndrome));
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_AUDP:
		return memcmp(&dst_mask->audp, &src_mask->audp, sizeof(dst_mask->audp));
	case RTE_FLOW_ITEM_TYPE_METER_COLOR:
		return memcmp(&dst_mask->meter_color, &src_mask->meter_color,
			      sizeof(struct rte_flow_item_meter_color));
	case RTE_FLOW_ITEM_TYPE_ESP:
		return memcmp(&dst_mask->esp, &src_mask->esp, sizeof(dst_mask->esp));
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_IPSEC_SYNDROME:
		return memcmp(&dst_mask->ipsec_syndrome, &src_mask->ipsec_syndrome, sizeof(dst_mask->ipsec_syndrome));
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Unsupported item type %d", type);
		return -1;
	}
}

static inline int
copy_items(enum rte_flow_item_type type, union items_mask *dst_mask, const void *src_mask)
{
	switch ((int)type) {
	case RTE_FLOW_ITEM_TYPE_ETH:
		memcpy(&dst_mask->eth, src_mask, sizeof(struct rte_flow_item_eth));
		break;
	case RTE_FLOW_ITEM_TYPE_VLAN:
		memcpy(&dst_mask->vlan, src_mask, sizeof(struct rte_flow_item_vlan));
		break;
	case RTE_FLOW_ITEM_TYPE_IPV4:
		memcpy(&dst_mask->ipv4, src_mask, sizeof(struct rte_flow_item_ipv4));
		break;
	case RTE_FLOW_ITEM_TYPE_IPV6:
		memcpy(&dst_mask->ipv6, src_mask, sizeof(struct rte_flow_item_ipv6));
		break;
	case RTE_FLOW_ITEM_TYPE_TCP:
		memcpy(&dst_mask->tcp, src_mask, sizeof(struct rte_flow_item_tcp));
		break;
	case RTE_FLOW_ITEM_TYPE_UDP:
		memcpy(&dst_mask->udp, src_mask, sizeof(struct rte_flow_item_udp));
		break;
	case RTE_FLOW_ITEM_TYPE_GTP:
		memcpy(&dst_mask->gtp, src_mask, sizeof(struct rte_flow_item_gtp));
		break;
	case RTE_FLOW_ITEM_TYPE_VXLAN:
		memcpy(&dst_mask->vxlan, src_mask, sizeof(struct rte_flow_item_vxlan));
		break;
	case RTE_FLOW_ITEM_TYPE_GRE:
		memcpy(&dst_mask->gre, src_mask, sizeof(struct rte_flow_item_gre));
		break;
	case RTE_FLOW_ITEM_TYPE_GRE_KEY:
		memcpy(&dst_mask->gre_key, src_mask, sizeof(uint32_t));
		break;
	case RTE_FLOW_ITEM_TYPE_META:
		memcpy(&dst_mask->meta, src_mask, sizeof(uint32_t));
		break;
	case RTE_FLOW_ITEM_TYPE_REPRESENTED_PORT:
		memcpy(&dst_mask->ethdev, src_mask, sizeof(struct rte_flow_item_ethdev));
		break;
	case RTE_FLOW_ITEM_TYPE_TAG:
		memcpy(&dst_mask->tag, src_mask, sizeof(struct rte_flow_item_tag));
		break;
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP:
		memcpy(&dst_mask->nisp, src_mask, sizeof(dst_mask->nisp));
		break;
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP_SYNDROME:
		memcpy(&dst_mask->nisp_syndrome, src_mask, sizeof(dst_mask->nisp_syndrome));
		break;
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_AUDP:
		memcpy(&dst_mask->audp, src_mask, sizeof(dst_mask->audp));
		break;
	case RTE_FLOW_ITEM_TYPE_METER_COLOR:
		memcpy(&dst_mask->meter_color, src_mask, sizeof(struct rte_flow_item_meter_color));
		break;
	case RTE_FLOW_ITEM_TYPE_ESP:
		memcpy(&dst_mask->esp, src_mask, sizeof(dst_mask->esp));
		break;
	case RTE_PMD_MLX5_FLOW_ITEM_TYPE_IPSEC_SYNDROME:
		memcpy(&dst_mask->ipsec_syndrome, src_mask, sizeof(dst_mask->ipsec_syndrome));
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Unsupported item type %d", type);
		return -1;
	}
	return 0;
}

static inline uint32_t
pattern_template_hash_crc(const void *data, __rte_unused uint32_t data_len,
			   uint32_t init_val)
{
	const struct dpdk_template_items_key *k;
	int i;

	k = data;
	init_val = rte_hash_crc(&k->it_attr, sizeof(k->it_attr), init_val);
	for (i = 0; k->items[i].type != RTE_FLOW_ITEM_TYPE_END; i++) {
		if (k->items[i].mask_is_set)
			init_val = rte_hash_crc(&k->items[i].mask, sizeof(k->items[i].mask), init_val);
		init_val = rte_hash_crc(&k->items[i].type, sizeof(k->items[i].type), init_val);
	}
	return init_val;
}

static inline int
pattern_template_cmp_item(const struct dpdk_template_items_item *src,
			  const struct dpdk_template_items_item *dst)
{
	if (src->type != dst->type)
		return -EINVAL;

	if (src->mask_is_set) {
		if (dst->mask_is_set) {
			if (cmp_items(src->type, &dst->mask, &src->mask))
				return -EINVAL;
		} else
			return -EINVAL;
	} else if (dst->mask_is_set)
		return -EINVAL;
	return 0;
}

static inline int
pattern_template_cmp_key(const void *key1, const void *key2, size_t key_len)
{
	const struct dpdk_template_items_key *src, *dst;
	int i;

	src = key1;
	dst = key2;
	if (src->it_attr.ingress != dst->it_attr.ingress ||
	    src->it_attr.egress != dst->it_attr.egress ||
	    src->it_attr.transfer != dst->it_attr.transfer)
		return -EINVAL;

	for (i = 0; src->items[i].type != RTE_FLOW_ITEM_TYPE_END; i++) {
		if (pattern_template_cmp_item(&src->items[i], &dst->items[i]))
			return -EINVAL;
	}
	return 0;
}

static inline void
format_pattern_template_name(char *name, uint16_t port_id)
{
	snprintf(name, TABLE_NAME_BUF_SIZE, "p_tmplt_t port %"PRIu16"", port_id);
}

static struct utils_hash_table *
dpdk_template_items_init(uint16_t port_id)
{
	struct utils_hash_table_cfg cfg = {0};
	struct utils_hash_table *ret;
	char name[TABLE_NAME_BUF_SIZE];

	format_pattern_template_name(name, port_id);
	cfg.table_size = MAX_PATTERN_TEMPLATES;
	cfg.key_len = sizeof(struct dpdk_template_items_key);
	cfg.name = name;
	cfg.hash_func = &pattern_template_hash_crc;
	cfg.free_key_cb = NULL;
	cfg.cmp_eq = &pattern_template_cmp_key;
	cfg.thread_safe = true;

	ret = utils_hash_table_create(&cfg);
	if (ret == NULL) {
		DOCA_DLOG_ERR("failed to allocate dpdk_template_items table for port %"PRIu16"", port_id);
		return NULL;
	}
	return ret;
}

static void
dpdk_template_items_destroy(struct utils_hash_table *dpdk_template_items)
{
	utils_hash_table_destroy(dpdk_template_items);
}

static inline int
dpdk_template_items_copy_array(struct dpdk_template_items_item dst[],
			       const struct rte_flow_item src[],
			       int size)
{
	int i, ret;

	for (i = 0; src[i].type != RTE_FLOW_ITEM_TYPE_END && i < size; i++) {
		dst[i].type = src[i].type;
		if (src[i].mask) {
			ret = copy_items(src[i].type, &dst[i].mask, src[i].mask);
			if (ret)
				return -1;
			dst[i].mask_is_set = true;
		} else {
			memset(&dst[i].mask, 0x0, sizeof(dst[i].mask));
			dst[i].mask_is_set = false;
		}
	}
	dst[i].type = RTE_FLOW_ITEM_TYPE_END;
	memset(&dst[i].mask, 0x0, sizeof(dst[i].mask));
	dst[i].mask_is_set = false;
	return 0;
}

static struct rte_flow_pattern_template *
dpdk_template_items_get_template(uint16_t port_id,
				 struct utils_hash_table *dpdk_template_items,
				 const struct rte_flow_pattern_template_attr *it_attr,
				 const struct rte_flow_item items[])
{
	struct rte_flow_pattern_template *data;
	struct dpdk_template_items_key key;
	struct rte_flow_error flow_error;
	int ret;

	memset(&key, 0x0, sizeof(key));
	memcpy(&key.it_attr, it_attr, sizeof(key.it_attr));
	if (dpdk_template_items_copy_array(key.items, items, DPDK_PIPE_MAX_ITEMS))
		return NULL;

	ret = utils_hash_table_get_value_by_key(dpdk_template_items, &key, (void **)&data);
	if (ret < 0) {
		data = rte_flow_pattern_template_create(port_id, it_attr, items, &flow_error);
		if (!data) {
			DOCA_DLOG_ERR("failed pattern template create on port  %"PRIu16" (type %d: %s)",
				      port_id, flow_error.type, flow_error.message);
			return NULL;
		}
		ret = utils_hash_table_key_set_value(dpdk_template_items, &key, data);
		if (ret) {
			ret = rte_flow_pattern_template_destroy(port_id, data, &flow_error);
			if (ret)
				DOCA_DLOG_ERR("failed pattern template destroy on port %"PRIu16" (type %d: %s)",
					      port_id, flow_error.type, flow_error.message);
			return NULL;
		}
	}
	return data;
}

static int
dpdk_template_items_put_template(uint16_t port_id,
				 struct utils_hash_table *dpdk_template_items,
				 struct rte_flow_pattern_template *items_template)
{
	struct rte_flow_error flow_error;
	int ret;

	ret = utils_hash_table_put_value(dpdk_template_items, items_template);
	if (!ret) {
		ret = rte_flow_pattern_template_destroy(port_id, items_template, &flow_error);
		if (ret)
			DOCA_LOG_RATE_LIMIT_ERR("failed to destroy pattern template");
	}
	/* utils_hash_table_put_value might return a positive vale if the template is
	 * not freed indicating it's ref_count, for such cases let the caller know
	 * the call succeeded by returning 0
	 */
	return ret;
}

/** Actions template manager **/
static inline int
cmp_actions(enum rte_flow_action_type type, const union actions_conf *dst_conf,
	    const union actions_conf *src_conf)
{
	switch (type) {
	case RTE_FLOW_ACTION_TYPE_JUMP:
		return memcmp(&dst_conf->jump, &src_conf->jump,
			      sizeof(struct rte_flow_action_jump));
	case RTE_FLOW_ACTION_TYPE_RSS:
		return memcmp(&dst_conf->rss, &src_conf->rss,
			      sizeof(struct rte_flow_action_rss));
	case RTE_FLOW_ACTION_TYPE_MARK:
		return memcmp(&dst_conf->mark, &src_conf->mark,
			      sizeof(struct rte_flow_action_mark));
	case RTE_FLOW_ACTION_TYPE_RAW_DECAP:
		return memcmp(&dst_conf->raw_decap, &src_conf->raw_decap,
			      sizeof(struct rte_flow_action_raw_decap));
	case RTE_FLOW_ACTION_TYPE_METER:
		return memcmp(&dst_conf->meter, &src_conf->meter,
			      sizeof(struct rte_flow_action_meter));
	case RTE_FLOW_ACTION_TYPE_QUEUE:
		return memcmp(&dst_conf->queue, &src_conf->queue,
			      sizeof(struct rte_flow_action_queue));
	case RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT:
		return memcmp(&dst_conf->ethdev, &src_conf->ethdev,
			      sizeof(struct rte_flow_action_ethdev));
	case RTE_FLOW_ACTION_TYPE_MODIFY_FIELD:
		return memcmp(&dst_conf->modify_field, &src_conf->modify_field,
			      sizeof(struct rte_flow_action_modify_field));
	case RTE_FLOW_ACTION_TYPE_RAW_ENCAP:
		return memcmp(&dst_conf->raw_encap, &src_conf->raw_encap,
			      sizeof(struct rte_flow_action_raw_encap));
	case RTE_FLOW_ACTION_TYPE_COUNT:
		return memcmp(&dst_conf->count, &src_conf->count,
			      sizeof(struct rte_flow_action_count));
	case RTE_FLOW_ACTION_TYPE_INDIRECT:
		return memcmp(&dst_conf->handle, &src_conf->handle,
			      sizeof(void *));
	case RTE_FLOW_ACTION_TYPE_METER_MARK:
		return memcmp(&dst_conf->meter_mark, &src_conf->meter_mark,
			      sizeof(struct rte_flow_action_meter_mark));
	default:
		return -1;
	}
}

static inline int
copy_actions(enum rte_flow_action_type type, union actions_conf *dst_conf, const void *src_conf)
{
	switch (type) {
	case RTE_FLOW_ACTION_TYPE_JUMP:
		memcpy(&dst_conf->jump, src_conf, sizeof(struct rte_flow_action_jump));
		break;
	case RTE_FLOW_ACTION_TYPE_RSS:
		memcpy(&dst_conf->rss, src_conf, sizeof(struct rte_flow_action_rss));
		break;
	case RTE_FLOW_ACTION_TYPE_MARK:
		memcpy(&dst_conf->mark, src_conf, sizeof(struct rte_flow_action_mark));
		break;
	case RTE_FLOW_ACTION_TYPE_RAW_DECAP:
		memcpy(&dst_conf->raw_decap, src_conf, sizeof(struct rte_flow_action_raw_decap));
		break;
	case RTE_FLOW_ACTION_TYPE_METER:
		memcpy(&dst_conf->meter, src_conf, sizeof(struct rte_flow_action_meter));
		break;
	case RTE_FLOW_ACTION_TYPE_QUEUE:
		memcpy(&dst_conf->queue, src_conf, sizeof(struct rte_flow_action_queue));
		break;
	case RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT:
		memcpy(&dst_conf->ethdev, src_conf, sizeof(struct rte_flow_action_ethdev));
		break;
	case RTE_FLOW_ACTION_TYPE_MODIFY_FIELD:
		memcpy(&dst_conf->modify_field, src_conf, sizeof(struct rte_flow_action_modify_field));
		break;
	case RTE_FLOW_ACTION_TYPE_RAW_ENCAP:
		memcpy(&dst_conf->raw_encap, src_conf, sizeof(struct rte_flow_action_raw_encap));
		break;
	case RTE_FLOW_ACTION_TYPE_COUNT:
		memcpy(&dst_conf->count, src_conf, sizeof(struct rte_flow_action_count));
		break;
	case RTE_FLOW_ACTION_TYPE_INDIRECT:
		memcpy(&dst_conf->handle, &src_conf, sizeof(void *));
		break;
	case RTE_FLOW_ACTION_TYPE_METER_MARK:
		memcpy(&dst_conf->meter_mark, src_conf, sizeof(struct rte_flow_action_meter_mark));
		break;
	default:
		return -1;
	}
	return 0;
}

static inline uint32_t
actions_template_hash_crc(const void *data, __rte_unused uint32_t data_len,
			  uint32_t init_val)
{
	const struct dpdk_template_actions_key *k;
	int i;

	k = data;
	init_val = rte_hash_crc(&k->at_attr, sizeof(k->at_attr), init_val);
	for (i = 0; k->actions[i].type != RTE_FLOW_ACTION_TYPE_END; i++) {
		if (k->actions[i].conf_is_set) {
			init_val = rte_hash_crc(&k->actions[i].conf, sizeof(k->actions[i].conf), init_val);
			if (k->mask[i].conf_is_set)
				init_val = rte_hash_crc(&k->mask[i].conf, sizeof(k->mask[i].conf), init_val);
		}
		init_val = rte_hash_crc(&k->actions[i].type, sizeof(k->actions[i].type), init_val);
	}
	return init_val;
}

static inline int
dpdk_template_actions_cmp_single_action(const struct dpdk_template_actions_item *action1,
					const struct dpdk_template_actions_item *action2)
{
	if (action1->conf_is_set) {
		if (action2->conf_is_set)
			return cmp_actions(action1->type, &action1->conf, &action2->conf);
		else
			return -1;
	} else if (action2->conf_is_set) {
		return -1;
	}
	return 0;
}

static inline int
actions_template_cmp_key(const void *key1, const void *key2, size_t key_len)
{
	const struct dpdk_template_actions_key *src, *dst;
	int i;

	src = key1;
	dst = key2;
	if (src->at_attr.ingress != dst->at_attr.ingress ||
	    src->at_attr.egress != dst->at_attr.egress ||
	    src->at_attr.transfer != dst->at_attr.transfer)
		return -1;

	for (i = 0; src->actions[i].type != RTE_FLOW_ACTION_TYPE_END; i++) {
		if (src->actions[i].type != dst->actions[i].type)
			return -1;
		if (dpdk_template_actions_cmp_single_action(&src->actions[i], &dst->actions[i]))
			return -1;
		if (dpdk_template_actions_cmp_single_action(&src->mask[i], &dst->mask[i]))
			return -1;
	}
	return 0;
}

static inline void
format_action_template_name(char *name, uint16_t port_id)
{
	snprintf(name, TABLE_NAME_BUF_SIZE, "a_tmplt_t port %"PRIu16"", port_id);
}

static struct utils_hash_table *
dpdk_template_actions_init(uint16_t port_id)
{
	struct utils_hash_table_cfg cfg = {0};
	struct utils_hash_table *ret;
	char name[TABLE_NAME_BUF_SIZE];

	format_action_template_name(name, port_id);
	cfg.table_size = MAX_ACTIONS_TEMPLATES;
	cfg.key_len = sizeof(struct dpdk_template_actions_key);
	cfg.name = name;
	cfg.hash_func = &actions_template_hash_crc;
	cfg.free_key_cb = NULL;
	cfg.cmp_eq = &actions_template_cmp_key;
	cfg.thread_safe = true;

	ret = utils_hash_table_create(&cfg);
	if (ret == NULL) {
		DOCA_DLOG_ERR("failed to allocate dpdk_template_actions table for port %"PRIu16"", port_id);
		return NULL;
	}
	return ret;
}

static void
dpdk_template_actions_destroy(struct utils_hash_table *dpdk_template_actions)
{
	utils_hash_table_destroy(dpdk_template_actions);
}

static inline int
dpdk_template_actions_copy_array(struct dpdk_template_actions_item dst[],
				 const struct rte_flow_action src[],
				 int size)
{
	int i, ret;

	for (i = 0; src[i].type != RTE_FLOW_ACTION_TYPE_END && i < size; i++) {
		dst[i].type = src[i].type;
		if (src[i].conf) {
			ret = copy_actions(src[i].type, &dst[i].conf, src[i].conf);
			if (ret)
				return -EINVAL;
			dst[i].conf_is_set = true;
		} else {
			memset(&dst[i].conf, 0x0, sizeof(dst[i].conf));
			dst[i].conf_is_set = false;
		}
	}
	dst[i].type = RTE_FLOW_ACTION_TYPE_END;
	memset(&dst[i].conf, 0x0, sizeof(dst[i].conf));
	dst[i].conf_is_set = false;
	return 0;
}

static struct rte_flow_actions_template *
dpdk_template_actions_get_template(uint16_t port_id,
				   struct utils_hash_table *dpdk_template_actions,
				   const struct rte_flow_actions_template_attr *at_attr,
				   const struct rte_flow_action actions[],
				   const struct rte_flow_action mask[])
{
	struct rte_flow_actions_template *data;
	struct dpdk_template_actions_key key;
	struct rte_flow_error flow_error;
	int ret;

	memset(&key, 0x0, sizeof(key));
	memcpy(&key.at_attr, at_attr, sizeof(key.at_attr));
	if (dpdk_template_actions_copy_array(key.actions, actions, DPDK_PIPE_MAX_ACTIONS) ||
	    dpdk_template_actions_copy_array(key.mask, mask, DPDK_PIPE_MAX_ACTIONS))
		return NULL;

	ret = utils_hash_table_get_value_by_key(dpdk_template_actions, &key, (void **)&data);
	if (ret < 0) {
		data = rte_flow_actions_template_create(port_id, at_attr, actions, mask, &flow_error);
		if (!data)
			return NULL;
		ret = utils_hash_table_key_set_value(dpdk_template_actions, &key, data);
		if (ret) {
			ret = rte_flow_actions_template_destroy(port_id, data, &flow_error);
			if (ret)
				DOCA_DLOG_ERR("Could not delete actions_template - message (type %d: %s)", flow_error.type, flow_error.message);
			return NULL;
		}
	}
	return data;
}

static int
dpdk_template_actions_put_template(uint16_t port_id,
				   struct utils_hash_table *dpdk_template_actions,
				   struct rte_flow_actions_template *actions_template)
{
	struct rte_flow_error flow_error;
	int ret;

	ret = utils_hash_table_put_value(dpdk_template_actions, actions_template);
	if (!ret) {
		ret = rte_flow_actions_template_destroy(port_id, actions_template, &flow_error);
		if (ret)
			DOCA_LOG_RATE_LIMIT_ERR("failed to destroy actions template");
	}
	/* utils_hash_table_put_value might return a positive vale if the template is
	 * not freed indicating it's ref_count, for such cases let the caller know
	 * the call succeeded by returning 0
	 */
	return ret;
}

/* Table manager */
static inline uint32_t
dpdk_table_mgr_hash_crc(const void *data, __rte_unused uint32_t data_len,
			uint32_t init_val)
{
	const struct dpdk_table_mgr_key *k;
	int i;

	k = data;
	init_val = rte_hash_crc(&k->tatr, sizeof(k->tatr), init_val);
	for (i = 0; i < MAX_PAT_TEMPLATE; i++)
		init_val = rte_hash_crc(&k->it[i], sizeof(struct rte_flow_pattern_template *), init_val);
	for (i = 0; i < MAX_ACT_TEMPLATE; i++)
		init_val = rte_hash_crc(&k->at[i], sizeof(struct rte_flow_actions_template *), init_val);
	return init_val;
}

static inline int
dpdk_table_mgr_cmp_key(const void *key1, const void *key2, size_t key_len)
{
	const struct dpdk_table_mgr_key *src, *dst;
	int i;

	src = key1;
	dst = key2;
	if (src->tatr.flow_attr.ingress != dst->tatr.flow_attr.ingress ||
	    src->tatr.flow_attr.egress != dst->tatr.flow_attr.egress ||
	    src->tatr.flow_attr.transfer != dst->tatr.flow_attr.transfer ||
	    src->tatr.flow_attr.group != dst->tatr.flow_attr.group ||
	    src->tatr.flow_attr.priority != dst->tatr.flow_attr.priority)
		return -1;

	for (i = 0; i < MAX_PAT_TEMPLATE; i++) {
		if (src->it[i] != dst->it[i])
			return -1;
	}
	for (i = 0; i < MAX_ACT_TEMPLATE; i++) {
		if (src->at[i] != dst->at[i])
			return -1;
	}
	return 0;
}

static inline void
format_dpdk_table_mgr_name(char *name, uint16_t port_id)
{
	snprintf(name, TABLE_NAME_BUF_SIZE, "dpdk_tbl_mgr port %"PRIu16"", port_id);
}

static struct utils_hash_table *
dpdk_table_mgr_init(uint16_t port_id)
{
	struct utils_hash_table_cfg cfg = {0};
	struct utils_hash_table *ret;
	char name[TABLE_NAME_BUF_SIZE];

	format_dpdk_table_mgr_name(name, port_id);
	cfg.table_size = MAX_TABLES;
	cfg.key_len = sizeof(struct dpdk_table_mgr_key);
	cfg.name = name;
	cfg.hash_func = &dpdk_table_mgr_hash_crc;
	cfg.free_key_cb = NULL;
	cfg.cmp_eq = &dpdk_table_mgr_cmp_key;
	cfg.thread_safe = true;

	ret = utils_hash_table_create(&cfg);
	if (ret == NULL) {
		DOCA_DLOG_ERR("failed to allocate dpdk_template_items table for port %"PRIu16"", port_id);
		return NULL;
	}
	return ret;
}

static void
dpdk_table_mgr_destroy(struct utils_hash_table *dpdk_table_mgr)
{
	utils_hash_table_destroy(dpdk_table_mgr);
}

static inline void
dpdk_table_mgr_copy_items(void *it_dst[],
			  struct rte_flow_pattern_template *it_src[],
			  uint8_t nb_pattern_templates,
			  void *at_dst[],
			  struct rte_flow_actions_template *at_src[],
			  uint8_t nb_action_templates)
{
	int i;

	for (i = 0; i < nb_pattern_templates; i++)
		it_dst[i] = it_src[i];
	for (i = 0; i < nb_action_templates; i++)
		at_dst[i] = at_src[i];
}

static struct rte_flow_template_table *
dpdk_table_mgr_get_table(uint16_t port_id,
			 struct dpdk_table_port_ctx *ctx,
			 const struct rte_flow_template_table_attr *tatr,
			 struct rte_flow_pattern_template *it[],
			 uint8_t nb_pattern_templates,
			 struct rte_flow_actions_template *at[],
			 uint8_t nb_actions_templates)
{
	struct rte_flow_template_table *data;
	struct rte_flow_error flow_error;
	struct dpdk_table_mgr_key key;
	int ret;
	int i;

	memset(&key, 0x0, sizeof(key));
	memcpy(&key.tatr, tatr, sizeof(key.tatr));
	dpdk_table_mgr_copy_items(key.it, it, nb_pattern_templates, key.at, at, nb_actions_templates);
	key.nb_pattern_templates = nb_pattern_templates;
	key.nb_actions_templates = nb_actions_templates;

	ret = utils_hash_table_get_value_by_key(ctx->dpdk_table_mgr, &key, (void **)&data);
	if (ret < 0) {
		data = rte_flow_template_table_create(port_id, tatr, it, nb_pattern_templates,
						      at, nb_actions_templates, &flow_error);
		if (!data) {
			DOCA_DLOG_ERR("Could not create table template - message (type %d: %s)",
				      flow_error.type, flow_error.message ? flow_error.message : "No stated reason");
			return NULL;
		}
		ret = utils_hash_table_key_set_value(ctx->dpdk_table_mgr, &key, data);
		if (ret) {
			ret = rte_flow_template_table_destroy(port_id, data, &flow_error);
			if (ret)
				DOCA_DLOG_ERR("Could not delete table - message (type %d: %s)",
					      flow_error.type, flow_error.message ? flow_error.message : "No stated reason");
			return NULL;
		}
	} else {
		/* we found a table existing with those templates, in such case a refrence
		 * has been already taken upon creating the original table.
		 * Prior to calling dpdk_table_mgr_get_table another refrence has been taken
		 * but it's not needed since we found a table.
		 */
		for (i = 0; i < key.nb_pattern_templates; i++) {
			if (dpdk_template_items_put_template(port_id, ctx->dpdk_template_items, key.it[i]) < 0)
				return NULL;
		}
		for (i = 0; i < key.nb_actions_templates; i++) {
			if (dpdk_template_actions_put_template(port_id, ctx->dpdk_template_actions, key.at[i]) < 0)
				return NULL;
		}
	}
	return data;
}

static int
dpdk_table_mgr_put_table(struct dpdk_table_port_ctx *ctx,
			 uint16_t port_id,
			 struct rte_flow_template_table *table)
{
	struct dpdk_table_mgr_key *key, key_copy;
	struct rte_flow_error flow_error;
	int ret;
	int i;

	/* get pointer to templates before freeing a table from the hash */
	ret = utils_hash_table_get_key_by_value(ctx->dpdk_table_mgr, table, (void **)&key);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to put table - can't get table key");
		return -1;
	}
	memcpy(&key_copy, key, sizeof(key_copy));
	ret = utils_hash_table_put_value(ctx->dpdk_table_mgr, table);
	if (!ret) {
		int destroy_ret;
		/* table refcount reached 0, delete it. */
		destroy_ret = rte_flow_template_table_destroy(port_id, table, &flow_error);
		if (destroy_ret) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to put table - %p dpdk table destroy failed rc %d - type (%d) message: %s",
						table, ret, flow_error.type,
						flow_error.message ? flow_error.message : "no stated reason");
			return destroy_ret;
		}
		for (i = 0; i < key_copy.nb_pattern_templates; i++) {
			destroy_ret = dpdk_template_items_put_template(port_id, ctx->dpdk_template_items, key_copy.it[i]);
			if (destroy_ret < 0)
				return destroy_ret;
		}
		for (i = 0; i < key_copy.nb_actions_templates; i++) {
			destroy_ret = dpdk_template_actions_put_template(port_id, ctx->dpdk_template_actions, key_copy.at[i]);
			if (destroy_ret < 0)
				return destroy_ret;
		}
	}

	return ret;
}

static int
dpdk_table_hws_init(struct dpdk_table_cfg *cfg)
{
	DOCA_LOG_DBG("Initialized dpdk table work module to be HW steering");
	return 0;
}

static int
dpdk_table_hws_cleanup(void)
{
	DOCA_LOG_DBG("Cleanup dpdk table HW steering module");
	return 0;
}

static struct dpdk_table_port_ctx *
dpdk_table_hws_port_hook(uint16_t port_id)
{
	struct dpdk_table_port_ctx *ret;

	ret = calloc(1, sizeof(*ret));
	if (unlikely(ret == NULL)) {
		DOCA_LOG_ERR("Failed to hook port for port %u - memory depleted", port_id);
		return NULL;
	}
	ret->dpdk_template_actions = dpdk_template_actions_init(port_id);
	if (ret->dpdk_template_actions == NULL) {
		DOCA_LOG_ERR("Failed to init actions template table for port %u", port_id);
		goto err_actions;
	}
	ret->dpdk_template_items = dpdk_template_items_init(port_id);
	if (ret->dpdk_template_items == NULL) {
		DOCA_LOG_ERR("Failed to init items template table for port %u", port_id);
		goto err_items;
	}
	ret->dpdk_table_mgr = dpdk_table_mgr_init(port_id);
	if (ret->dpdk_table_mgr == NULL) {
		DOCA_LOG_ERR("Failed to init dpdk_hws_table_mgr for port %u", port_id);
		goto err_table;
	}
	return ret;

err_table:
	dpdk_template_items_destroy(ret->dpdk_template_items);
err_items:
	dpdk_template_actions_destroy(ret->dpdk_template_actions);
err_actions:
	free(ret);
	return NULL;
}

static void
dpdk_table_hws_port_unhook(struct dpdk_table_port_ctx *ctx)
{
	if (ctx == NULL)
		return;
	dpdk_template_actions_destroy(ctx->dpdk_template_actions);
	dpdk_template_items_destroy(ctx->dpdk_template_items);
	dpdk_table_mgr_destroy(ctx->dpdk_table_mgr);
	free(ctx);
	ctx = NULL;
}

static int
dpdk_table_set_insertion_type_to_attr(enum dpdk_table_insertion_type type, struct rte_flow_template_table_attr *tatr)
{
	static const enum rte_flow_table_insertion_type table[DPDK_TABLE_INSERTION_TYPE_MAX] = {
		[DPDK_TABLE_INSERTION_TYPE_PATTERN] = RTE_FLOW_TABLE_INSERTION_TYPE_PATTERN,
		[DPDK_TABLE_INSERTION_TYPE_INDEX] = RTE_FLOW_TABLE_INSERTION_TYPE_INDEX,
	};

	if (type >= DPDK_TABLE_INSERTION_TYPE_MAX) {
		DOCA_LOG_ERR("unknown DPDK table insertion type %d", type);
		return -EINVAL;
	}
	tatr->insertion_type = table[type];
	return 0;
}

static int
dpdk_table_set_hash_type_to_attr(enum dpdk_table_hash_type type, struct rte_flow_template_table_attr *tatr)
{
	static const enum rte_flow_table_hash_func table[DPDK_TABLE_HASH_TYPE_MAX] = {
		[DPDK_TABLE_HASH_TYPE_DEFAULT] = RTE_FLOW_TABLE_HASH_FUNC_DEFAULT,
		[DPDK_TABLE_HASH_TYPE_IDENTITY] = RTE_FLOW_TABLE_HASH_FUNC_LINEAR,
		[DPDK_TABLE_HASH_TYPE_CRC32] = RTE_FLOW_TABLE_HASH_FUNC_CRC32,
		[DPDK_TABLE_HASH_TYPE_CRC16] = RTE_FLOW_TABLE_HASH_FUNC_CRC16,
	};

	if (type >= DPDK_TABLE_HASH_TYPE_MAX) {
		DOCA_LOG_ERR("unknown DPDK table hash type %d", type);
		return -EINVAL;
	}
	tatr->hash_func = table[type];
	return 0;
}

static int
dpdk_table_hws_create_table(struct dpdk_table_params *dpdk_table_params,
			    struct dpdk_table_port_ctx *ctx,
			    struct dpdk_table *res)
{
	struct rte_flow_pattern_template_attr itr = {0};
	struct rte_flow_actions_template_attr atr = {0};
	struct rte_flow_template_table_attr tatr = {0};
	struct rte_flow_pattern_template *it_array[MAX_PAT_TEMPLATE];
	struct rte_flow_actions_template *at_array[MAX_ACT_TEMPLATE];
	uint8_t nb_pattern_templates, nb_actions_templates;
	struct rte_flow_pattern_template *it;
	struct rte_flow_actions_template *at;
	struct rte_flow_template_table *tbl;
	int i, rc;

	res->pattern_template_index = 0;
	res->action_template_index = 0;
	res->port_id = dpdk_table_params->port_id;

	itr.ingress = dpdk_table_params->flow_attr->ingress;
	itr.egress = dpdk_table_params->flow_attr->egress;
	itr.transfer = dpdk_table_params->flow_attr->transfer;
	it = dpdk_template_items_get_template(res->port_id, ctx->dpdk_template_items,
					      &itr, dpdk_table_params->items);
	if (unlikely(!it)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create items_template for port %u", res->port_id);
		return -ENOENT;
	}
	it_array[0] = it;

	atr.ingress = dpdk_table_params->flow_attr->ingress;
	atr.egress = dpdk_table_params->flow_attr->egress;
	atr.transfer = dpdk_table_params->flow_attr->transfer;
	for (i = 0; i < dpdk_table_params->nb_actions; i++) {
		at = dpdk_template_actions_get_template(res->port_id, ctx->dpdk_template_actions,
							&atr, dpdk_table_params->actions[i],
							dpdk_table_params->actions_mask[i]);
		if (unlikely(!at)) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to create actions_template for port %u", res->port_id);
			dpdk_template_items_put_template(res->port_id, ctx->dpdk_template_items, it);
			return -ENOENT;
		}
		at_array[i] = at;
	}

	tatr.flow_attr.ingress = dpdk_table_params->flow_attr->ingress;
	tatr.flow_attr.egress = dpdk_table_params->flow_attr->egress;
	tatr.flow_attr.transfer = dpdk_table_params->flow_attr->transfer;
	tatr.flow_attr.priority = dpdk_table_params->flow_attr->priority;
	tatr.flow_attr.group = dpdk_table_params->flow_attr->group;
	tatr.nb_flows = dpdk_table_params->nb_flows;
	rc = dpdk_table_set_insertion_type_to_attr(dpdk_table_params->insertion_type, &tatr);
	if (unlikely(rc))
		return rc;
	rc = dpdk_table_set_hash_type_to_attr(dpdk_table_params->hash_type, &tatr);
	if (unlikely(rc))
		return rc;
	/* Current support is only for one pattern template per table */
	nb_pattern_templates = MAX_PAT_TEMPLATE;
	nb_actions_templates = dpdk_table_params->nb_actions;
	tbl = dpdk_table_mgr_get_table(res->port_id, ctx, &tatr, it_array, nb_pattern_templates,
				       at_array, nb_actions_templates);
	if (unlikely(!tbl)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create table for port %u", res->port_id);
		dpdk_template_items_put_template(res->port_id, ctx->dpdk_template_items, it_array[0]);
		for (i = 0; i < dpdk_table_params->nb_actions; i++)
			dpdk_template_actions_put_template(res->port_id, ctx->dpdk_template_actions, at_array[i]);
		return -ENOENT;
	}
	res->table = tbl;
	return 0;
}

static int
dpdk_table_hws_destroy_table(struct dpdk_table_port_ctx *ctx, struct dpdk_table *res)
{
	return dpdk_table_mgr_put_table(ctx, res->port_id, res->table);
}

static struct dpdk_table_ops hw_steering_ops = {
	.module_init = dpdk_table_hws_init,
	.module_cleanup = dpdk_table_hws_cleanup,
	.port_hook = dpdk_table_hws_port_hook,
	.port_unhook = dpdk_table_hws_port_unhook,
	.table_create = dpdk_table_hws_create_table,
	.table_destroy = dpdk_table_hws_destroy_table,
};

struct dpdk_table_ops *
dpdk_table_hws_get_ops(void)
{
	return &hw_steering_ops;
}
