/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdio.h>

#include <rte_vxlan.h>
#include <rte_ethdev.h>
#include <rte_flow.h>
#include <rte_gre.h>

#include <doca_log.h>

#include "doca_flow.h"
#include "dpdk_engine.h"
#include "dpdk_dbg.h"
#include "utils_field.h"
#include "utils_df_translate.h"
#include "dpdk_shared_crypto.h"
#include <private/common_defines.h>

DOCA_LOG_REGISTER(dpdk_dbg);


static void
dump_eth_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_eth *eth_spec, *eth_mask;

	eth_spec = (const struct rte_flow_item_eth *)item->spec;
	eth_mask = (const struct rte_flow_item_eth *)item->mask;
	log_buff("eth");
	if (eth_spec && !rte_is_zero_ether_addr(&eth_spec->src)) {
		log_mac(" src spec ", eth_spec->src.addr_bytes);
		if (eth_mask)
			log_mac(" src mask ", eth_mask->src.addr_bytes);
	}
	if (eth_spec && !rte_is_zero_ether_addr(&eth_spec->dst)) {
		log_mac(" dst spec ", eth_spec->dst.addr_bytes);
		if (eth_mask)
			log_mac(" dst mask ", eth_mask->dst.addr_bytes);
	}
	if (eth_spec && eth_spec->type) {
		log_buff(" type spec 0x%x",
			rte_be_to_cpu_16(eth_spec->type));
		if (eth_mask)
			log_buff(" type mask 0x%x",
				rte_be_to_cpu_16(eth_mask->type));
	}
	if (eth_spec && eth_spec->has_vlan) {
		log_buff(" has_vlan spec %u", eth_spec->has_vlan);
		if (eth_mask)
			log_buff(" has_vlan mask %u", eth_mask->has_vlan);
	}
	log_buff(" / ");
}

static void
dump_ip4_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_ipv4 *ipv4_spec, *ipv4_mask;

	log_buff("ipv4 ");
	ipv4_spec = (const struct rte_flow_item_ipv4 *)item->spec;
	ipv4_mask = (const struct rte_flow_item_ipv4 *)item->mask;
	if (ipv4_spec && ipv4_spec->hdr.src_addr) {
		log_ipv4("src spec ",
			rte_be_to_cpu_32(ipv4_spec->hdr.src_addr));
		if (ipv4_mask)
			log_ipv4("src mask ",
				rte_be_to_cpu_32(ipv4_mask->hdr.src_addr));
	}
	if (ipv4_spec && ipv4_spec->hdr.dst_addr) {
		log_ipv4("dst spec ",
			rte_be_to_cpu_32(ipv4_spec->hdr.dst_addr));
		if (ipv4_mask)
			log_ipv4("dst mask ",
				rte_be_to_cpu_32(ipv4_mask->hdr.dst_addr));
	}
	if (ipv4_spec && ipv4_spec->hdr.next_proto_id) {
		log_buff("proto spec 0x%x ", ipv4_spec->hdr.next_proto_id);
		if (ipv4_mask)
			log_buff("proto mask 0x%x ", ipv4_mask->hdr.next_proto_id);
	}
	log_buff("/ ");
}

static void
dump_upd_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_udp *udp_spec, *udp_mask;

	log_buff("udp ");
	udp_spec = (const struct rte_flow_item_udp *)item->spec;
	udp_mask = (const struct rte_flow_item_udp *)item->mask;
	if (udp_spec && udp_spec->hdr.src_port) {
		log_buff("src spec %u ",
			rte_be_to_cpu_16(udp_spec->hdr.src_port));
		if (udp_mask)
			log_buff("src mask %u ",
				rte_be_to_cpu_16(udp_mask->hdr.src_port));
	}
	if (udp_spec && udp_spec->hdr.dst_port) {
		log_buff("dst spec %u ",
			rte_be_to_cpu_16(udp_spec->hdr.dst_port));
		if (udp_mask)
			log_buff("dst mask 0x%x ",
				rte_be_to_cpu_16(udp_mask->hdr.dst_port));
	}
	log_buff("/ ");
}

static void
dump_tcp_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_tcp *spec, *mask;

	log_buff("tcp ");
	spec = (const struct rte_flow_item_tcp *)item->spec;
	mask = (const struct rte_flow_item_tcp *)item->mask;
	if (spec && spec->hdr.src_port) {
		log_buff("src spec %u ",
			rte_be_to_cpu_16(spec->hdr.src_port));
		if (mask)
			log_buff("src mask %u ",
				rte_be_to_cpu_16(mask->hdr.src_port));
	}
	if (spec && spec->hdr.dst_port) {
		log_buff("dst spec %u ",
			rte_be_to_cpu_16(spec->hdr.dst_port));
		if (mask)
			log_buff("dst mask %u ",
				rte_be_to_cpu_16(mask->hdr.dst_port));
	}
	if (mask != NULL && mask->hdr.tcp_flags != 0) {
		if (spec != NULL)
			log_buff("flags spec 0x%x ",
				rte_be_to_cpu_16(spec->hdr.tcp_flags));
		log_buff("mask 0x%x ",
			rte_be_to_cpu_16(mask->hdr.tcp_flags));
	}
	log_buff("/ ");
}

static void
dump_vxlan_item(const struct rte_flow_item *item, char *buff)
{
	uint32_t vni;
	const struct rte_flow_item_vxlan *spec, *mask;

	log_buff("vxlan ");
	spec = (const struct rte_flow_item_vxlan *)item->spec;
	mask = (const struct rte_flow_item_vxlan *)item->mask;
	if (spec) {
		vni = (spec->vni[0] << 16U | spec->vni[1] << 8U | spec->vni[2]);
		if (vni)
			log_buff("vni spec 0x%x ", vni);
		if (mask) {
			vni = (mask->vni[0] << 16U | mask->vni[1] << 8U | mask->vni[2]);
			if (vni != UINT32_MAX)
				log_buff("vni mask 0x%x ", vni);
		}
	}
	log_buff("/ ");
}

static void
dump_gre_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_gre *spec, *mask;

	log_buff("gre ");
	spec = (const struct rte_flow_item_gre *)item->spec;
	mask = (const struct rte_flow_item_gre *)item->mask;
	if (spec && spec->protocol) {
		log_buff("protocol spec 0x%x ",
			rte_be_to_cpu_16(spec->protocol));
		if (mask)
			log_buff("protocol mask 0x%x ",
				rte_be_to_cpu_16(mask->protocol));
	}
	log_buff("/ ");
}

static void
dump_gre_item_key(const struct rte_flow_item *item, char *buff)
{
	const uint32_t *spec, *mask;

	log_buff("gre_key ");
	spec = (const uint32_t *)item->spec;
	mask = (const uint32_t *)item->mask;
	if (spec) {
		log_buff("value spec 0x%x ", rte_be_to_cpu_32(*spec));
		if (mask)
			log_buff("value mask 0x%x ", rte_be_to_cpu_32(*mask));
	}
	log_buff("/ ");
}

static void
dump_gtp_item_teid(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_gtp *spec, *mask;

	log_buff("gtpu ");
	spec = (const struct rte_flow_item_gtp *)item->spec;
	mask = (const struct rte_flow_item_gtp *)item->mask;
	if (spec && spec->teid) {
		log_buff("teid spec 0x%x ", rte_be_to_cpu_32(spec->teid));
		if (mask)
			log_buff("teid mask 0x%x ",
				 rte_be_to_cpu_32(mask->teid));
	}
	log_buff("/ ");
}

static void
dump_raw_encap(const char *name, char *buff,
		    const struct rte_flow_action_raw_encap *encap)
{
	uint32_t i;

	log_buff("%sset raw_encap 0 raw ", name);
	for (i = 0; i < encap->size; i++)
		log_buff("%02x", encap->data[i]);
	log_buff("(size:%zu)\n", encap->size);
}

static void
dump_raw_decap(const char *name, char *buff,
	       const struct rte_flow_action_raw_decap *decap)
{
	uint32_t i;

	log_buff("%sset raw_decap 0 raw ", name);
	for (i = 0; i < decap->size; i++)
		log_buff("%02x", decap->data[i]);
	log_buff("(size:%zu)\n", decap->size);
}

static void
dump_meta_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_meta *meta =
		(const struct rte_flow_item_meta *)item->spec;
	const struct rte_flow_item_meta *mask =
		(const struct rte_flow_item_meta *)item->mask;

	if (meta)
		log_buff("meta data spec 0x%x ", meta->data);
	if (mask)
		log_buff("data mask 0x%x ", mask->data);
	log_buff("/ ");
}

static void
dump_mark_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_mark *mark =
		(const struct rte_flow_item_mark *)item->spec;

	log_buff("mark id is %u / ", mark->id);
}

static void
dump_tag_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_tag *tag =
		(const struct rte_flow_item_tag *)item->spec;
	const struct rte_flow_item_tag *mask =
		(const struct rte_flow_item_tag *)item->mask;

	log_buff("tag index %hhu value 0x%x ", tag->index, tag->data);
	if (mask)
		log_buff("mask index %hhu value 0x%x ", mask->index, mask->data);
	log_buff("/ ");
}

static void
dump_vlan_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_vlan *spec =
		(const struct rte_flow_item_vlan *)item->spec;
	const struct rte_flow_item_vlan *mask =
		(const struct rte_flow_item_vlan *)item->mask;

	log_buff("vlan tci spec %u ", rte_be_to_cpu_16(spec->tci));
	if (mask)
		log_buff("tci mask 0x%x ", rte_be_to_cpu_16(mask->tci));
	log_buff("/ ");
}

#define MODIFY_FIELD_NAME(item) \
	[RTE_FLOW_FIELD_ ##item] = #item

const char *modify_field_name[] = {
	MODIFY_FIELD_NAME(START),
	MODIFY_FIELD_NAME(MAC_DST),
	MODIFY_FIELD_NAME(MAC_SRC),
	MODIFY_FIELD_NAME(VLAN_TYPE),
	MODIFY_FIELD_NAME(VLAN_ID),
	MODIFY_FIELD_NAME(MAC_TYPE),
	MODIFY_FIELD_NAME(IPV4_DSCP),
	MODIFY_FIELD_NAME(IPV4_TTL),
	MODIFY_FIELD_NAME(IPV4_SRC),
	MODIFY_FIELD_NAME(IPV4_DST),
	MODIFY_FIELD_NAME(IPV6_DSCP),
	MODIFY_FIELD_NAME(IPV6_HOPLIMIT),
	MODIFY_FIELD_NAME(IPV6_SRC),
	MODIFY_FIELD_NAME(IPV6_DST),
	MODIFY_FIELD_NAME(TCP_PORT_SRC),
	MODIFY_FIELD_NAME(TCP_PORT_DST),
	MODIFY_FIELD_NAME(TCP_SEQ_NUM),
	MODIFY_FIELD_NAME(TCP_ACK_NUM),
	MODIFY_FIELD_NAME(TCP_FLAGS),
	MODIFY_FIELD_NAME(UDP_PORT_SRC),
	MODIFY_FIELD_NAME(UDP_PORT_DST),
	MODIFY_FIELD_NAME(VXLAN_VNI),
	MODIFY_FIELD_NAME(GENEVE_VNI),
	MODIFY_FIELD_NAME(GTP_TEID),
	MODIFY_FIELD_NAME(TAG),
	MODIFY_FIELD_NAME(MARK),
	MODIFY_FIELD_NAME(META),
	MODIFY_FIELD_NAME(POINTER),
	MODIFY_FIELD_NAME(VALUE),
	MODIFY_FIELD_NAME(IPV4_ECN),
	MODIFY_FIELD_NAME(IPV6_ECN),
};

#define MODIFY_TYPE_NAME(item) \
	[RTE_FLOW_MODIFY_ ##item] = #item

const char *modify_type_name[] = {
	MODIFY_TYPE_NAME(SET),
	MODIFY_TYPE_NAME(ADD),
	MODIFY_TYPE_NAME(SUB),
};

static void
dump_modify_action_data(const char *name, const struct rte_flow_action_modify_data *data, char *buff)
{
	log_buff("%s", name);
	if (data->field != RTE_FLOW_FIELD_VALUE)
		log_buff("%s offset %u level %u ", modify_field_name[data->field], data->offset, data->level);
	else
		log_buff("value src_value 0x%lx ", *(uint64_t *)data->value);
}

static void
dump_port_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_ethdev *spec =
		(const struct rte_flow_item_ethdev *)item->spec;
	const struct rte_flow_item_ethdev *mask =
		(const struct rte_flow_item_ethdev *)item->mask;

	log_buff("represented_port ethdev_port_id ");
	if (spec)
		log_buff("is %u ", spec->port_id);
	if (mask)
		log_buff("mask 0x%x ", mask->port_id);
	log_buff("/ ");
}

static void
dump_esp_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_esp *spec, *mask;

	log_buff("esp ");
	spec = (const struct rte_flow_item_esp *)item->spec;
	mask = (const struct rte_flow_item_esp *)item->mask;
	if (spec)
		log_buff("protocol spec spi 0x%x ", rte_be_to_cpu_32(spec->hdr.spi));
	if (mask)
		log_buff("protocol mask spi 0x%x ", rte_be_to_cpu_32(mask->hdr.spi));
	if (spec)
		log_buff("protocol spec seq 0x%x ", rte_be_to_cpu_32(spec->hdr.seq));
	if (mask)
		log_buff("protocol mask seq 0x%x ", rte_be_to_cpu_32(mask->hdr.seq));
	log_buff("/ ");
}


static void
dump_nisp_item(const struct rte_flow_item *item, char *buff)
{
	const struct mlx5_flow_item_nisp *spec =
		(const struct mlx5_flow_item_nisp *)item->spec;
	const struct mlx5_flow_item_nisp *mask =
		(const struct mlx5_flow_item_nisp *)item->mask;
	uint32_t i;

	log_buff("nisp header ");
	if (spec) {
		log_buff("is ");
		for (i = 0; i < ARRAY_SIZE(spec->dw); i++)
			log_buff("0x%x ", spec->dw[i]);
	}
	if (mask) {
		log_buff("mask ");
		for (i = 0; i < ARRAY_SIZE(mask->dw); i++)
			log_buff("0x%x ", mask->dw[i]);
	}
	log_buff("\n");
};

static void
dump_audp_item(const struct rte_flow_item *item, char *buff)
{
	const struct mlx5_flow_item_audp *spec =
		(const struct mlx5_flow_item_audp *)item->spec;
	const struct mlx5_flow_item_audp *mask =
		(const struct mlx5_flow_item_audp *)item->mask;
	uint32_t i;

	log_buff("audp header ");
	if (spec) {
		log_buff("is ");
		for (i = 0; i < ARRAY_SIZE(spec->dw); i++)
			log_buff("0x%x ", spec->dw[i]);
	}
	if (mask) {
		log_buff("mask ");
		for (i = 0; i < ARRAY_SIZE(mask->dw); i++)
			log_buff("0x%x ", mask->dw[i]);
	}
	log_buff("\n");
};

static void
dump_nisp_synd_item(const struct rte_flow_item *item, char *buff)
{
	const struct mlx5_flow_item_nisp_syndrome *spec =
		(const struct mlx5_flow_item_nisp_syndrome *)item->spec;
	const struct mlx5_flow_item_nisp_syndrome *mask =
		(const struct mlx5_flow_item_nisp_syndrome *)item->mask;

	log_buff("nisp syndrome ");
	if (spec)
		log_buff("is 0x%x ", spec->syndrome);
	if (mask)
		log_buff("mask 0x%x", mask->syndrome);
	log_buff("\n");
};

static void
dump_ipsec_synd_item(const struct rte_flow_item *item, char *buff)
{
	const struct mlx5_flow_item_ipsec_syndrome *spec =
		(const struct mlx5_flow_item_ipsec_syndrome *)item->spec;
	const struct mlx5_flow_item_ipsec_syndrome *mask =
		(const struct mlx5_flow_item_ipsec_syndrome *)item->mask;

	log_buff("ipsec syndrome ");
	if (spec)
		log_buff("is 0x%x ", spec->syndrome);
	if (mask)
		log_buff("mask 0x%x", mask->syndrome);
	log_buff("\n");
};

static const char
*get_color_str(enum rte_color color)
{
	switch (color) {
	case RTE_COLOR_GREEN:
		return "green";
	case RTE_COLOR_RED:
		return "red";
	default:
		return "unknown";
	}
}

static void
dump_meter_color_item(const struct rte_flow_item *item, char *buff)
{
	const struct rte_flow_item_meter_color *spec =
		(const struct rte_flow_item_meter_color *)item->spec;

	log_buff("meter_color ");
	if (spec)
		log_buff("%s ", get_color_str(spec->color));
	log_buff("/ ");
}

static __rte_always_inline bool
is_rss_changeable(const struct rte_flow_action_rss *rss)
{
	return utils_field_get_type(&rss->queue_num, sizeof(rss->queue_num)) == UTILS_FIELD_TYPE_CHANGEABLE;
}

void
__dpdk_dbg_rte_flow_dump(const char *name, uint16_t port_id,
	      const struct rte_flow_attr *attr,
	      const struct rte_flow_item items[],
	      const struct rte_flow_action actions[],
	      FILE *f)
{
	char buff[MAX_TMP_BUFF];
	char buff2[MAX_TMP_BUFF];

	buff[0] = '\0';
	buff2[0] = '\0';
	log_buff("flow create %u ", port_id);
	if (attr) {
		log_buff("%s%s%sgroup %u priority %u ",
			attr->ingress ? "ingress " : "",
			attr->egress ? "egress " : "",
			attr->transfer ? "transfer " : "", attr->group,
			attr->priority);
	}

	if (items) {
		log_buff("pattern ");
		for (; items->type != RTE_FLOW_ITEM_TYPE_END; items++) {
			int item_type = items->type;

			switch (item_type) {
			case RTE_FLOW_ITEM_TYPE_ETH:
				dump_eth_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_IPV4:
				dump_ip4_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_UDP:
				dump_upd_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_TCP:
				dump_tcp_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_ICMP:
				log_buff("icmp / ");
				break;
			case RTE_FLOW_ITEM_TYPE_VXLAN:
				dump_vxlan_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_GRE:
				dump_gre_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_GRE_KEY:
				dump_gre_item_key(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_GTP:
				dump_gtp_item_teid(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_META:
				dump_meta_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_MARK:
				dump_mark_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_TAG:
				dump_tag_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_VLAN:
				dump_vlan_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_REPRESENTED_PORT:
				dump_port_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_ESP:
				dump_esp_item(items, buff);
				break;
			case RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP:
				dump_nisp_item(items, buff);
				break;
			case RTE_PMD_MLX5_FLOW_ITEM_TYPE_AUDP:
				dump_audp_item(items, buff);
				break;
			case RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP_SYNDROME:
				dump_nisp_synd_item(items, buff);
				break;
			case RTE_PMD_MLX5_FLOW_ITEM_TYPE_IPSEC_SYNDROME:
				dump_ipsec_synd_item(items, buff);
				break;
			case RTE_FLOW_ITEM_TYPE_METER_COLOR:
				dump_meter_color_item(items, buff);
				break;
			default:
				log_buff("unsupported item:%u\n", item_type);
			}
		}
		log_buff("end ");
	}
	if (actions) {
		log_buff("actions ");
		for (; actions->type != RTE_FLOW_ACTION_TYPE_END; actions++) {
			int action_type = actions->type;
			uint32_t queue_num;
			const struct rte_flow_action_mark *mark;
			const struct rte_flow_action_queue *queue;
			const struct rte_flow_action_jump *jump;
			const struct rte_flow_action_port_id *portid;
			const struct rte_flow_action_set_mac *set_mac;
			const struct rte_flow_action_set_ipv4 *set_ipv4;
			const struct rte_flow_action_set_tp *set_tp;
			const struct rte_flow_action_rss *rss;
			const struct rte_flow_action_meter *meter;
			const struct rte_flow_action_raw_encap *encap;
			const struct rte_flow_action_raw_decap *decap;
			const struct rte_flow_action_set_meta *meta;
			const struct rte_flow_action_modify_field *modify;
			const struct rte_flow_action_ethdev *port_ethdev;

			switch (action_type) {
			case RTE_FLOW_ACTION_TYPE_MARK:
				mark = actions->conf;
				if (mark)
					log_buff("mark id %u / ", mark->id);
				break;
			case RTE_FLOW_ACTION_TYPE_QUEUE:
				queue = actions->conf;
				log_buff("queue index %u / ", queue->index);
				break;
			case RTE_FLOW_ACTION_TYPE_JUMP:
				jump = actions->conf;
				log_buff("jump group %u / ", jump ? jump->group : 0);
				break;
			case RTE_FLOW_ACTION_TYPE_RSS:
				rss = actions->conf;
				log_buff("rss queues ");
				if (rss != NULL && !is_rss_changeable(rss)) {
					for (queue_num = 0; queue_num < rss->queue_num; queue_num++)
						log_buff("%u ", rss->queue[queue_num]);
				}
				log_buff("end / ");
				break;
			case RTE_FLOW_ACTION_TYPE_SET_TP_DST:
				set_tp = actions->conf;
				log_buff("set_tp_dst port %u ",
					rte_be_to_cpu_16(set_tp->port));
				log_buff("/ ");
				break;
			case RTE_FLOW_ACTION_TYPE_SET_TP_SRC:
				set_tp = actions->conf;
				log_buff("set_tp_src port %u ",
					rte_be_to_cpu_16(set_tp->port));
				log_buff("/ ");
				break;
			case RTE_FLOW_ACTION_TYPE_SET_IPV4_SRC:
				set_ipv4 = actions->conf;
				log_ipv4("set_ipv4_src ipv4_addr ",
					rte_be_to_cpu_32(set_ipv4->ipv4_addr));
				log_buff("/ ");
				break;
			case RTE_FLOW_ACTION_TYPE_SET_IPV4_DST:
				set_ipv4 = actions->conf;
				log_ipv4("set_ipv4_dst ipv4_addr ",
					rte_be_to_cpu_32(set_ipv4->ipv4_addr));
				log_buff("/ ");
				break;
			case RTE_FLOW_ACTION_TYPE_SET_MAC_DST:
				set_mac = actions->conf;
				log_mac("set_mac_dst mac_addr ", set_mac->mac_addr);
				log_buff("/ ");
				break;
			case RTE_FLOW_ACTION_TYPE_SET_MAC_SRC:
				set_mac = actions->conf;
				log_mac("set_mac_src mac_addr ", set_mac->mac_addr);
				log_buff("/ ");
				break;
			case RTE_FLOW_ACTION_TYPE_PORT_ID:
				portid = actions->conf;
				log_buff("port_id id %u / ", portid->id);
				break;
			case RTE_FLOW_ACTION_TYPE_RAW_DECAP:
				decap = actions->conf;
				dump_raw_decap(name, buff2, decap);
				log_buff("raw_decap index 0 (len:%zu)/ ", decap->size);
				break;
			case RTE_FLOW_ACTION_TYPE_RAW_ENCAP:
				encap = actions->conf;
				if (encap != NULL) {
					dump_raw_encap(name, buff2, encap);
					log_buff("raw_encap index 0 (len:%zu)/ ", encap->size);
				} else
					log_buff("raw_encap NULL / ");
				break;
			case RTE_FLOW_ACTION_TYPE_DROP:
				log_buff("drop / ");
				break;
			case RTE_FLOW_ACTION_TYPE_METER:
				meter = actions->conf;
				log_buff("meter mtr_id %u / ", meter->mtr_id);
				break;
			case RTE_FLOW_ACTION_TYPE_COUNT:
				log_buff("count / ");
				break;
			case RTE_FLOW_ACTION_TYPE_SET_META:
				meta = actions->conf;
				log_buff("set_meta data 0x%x mask 0x%x / ", meta->data, meta->mask);
				break;
			case RTE_FLOW_ACTION_TYPE_MODIFY_FIELD:
				modify = actions->conf;
				log_buff("modify_field op %s ", modify_type_name[modify->operation]);
				dump_modify_action_data("dst_type ", &modify->dst, buff);
				dump_modify_action_data("src_type ", &modify->src, buff);
				log_buff("width %u / ", modify->width);
				break;
			case RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT:
				port_ethdev = actions->conf;
				log_buff("represented_port ethdev_port_id %u / ", port_ethdev->port_id);
				break;
			case RTE_FLOW_ACTION_TYPE_INDIRECT:
				log_buff("indirect conf=0x%lx / ",
					 (uintptr_t)actions->conf);
				break;
			case RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_NISP:
				log_buff("nisp conf=0x%lx / ",
					 (uintptr_t)actions->conf);
				break;
			case RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_IPSEC:
				log_buff("ipsec conf=0x%lx / ",
					 (uintptr_t)actions->conf);
				break;
			case RTE_FLOW_ACTION_TYPE_METER_MARK:
				log_buff("meter_mark / ");
				break;
			default:
				log_buff("unsupported action:%u /", action_type);
			}
		}
		log_buff("end ");
	}
	if (strlen(buff2))
		fprintf(f, "%s", buff2);
	fprintf(f, "%s%s\n", name, buff);
}

static uint8_t
dump_eth_hdr(uint8_t *data, uint32_t *len, char *buff)
{
	uint16_t eth_type;
	struct rte_ether_hdr *eth = (struct rte_ether_hdr *)data;

	log_mac("eth src-mac:", eth->s_addr.addr_bytes);
	log_mac("dst-mac:", eth->d_addr.addr_bytes);
	log_buff(",type:0x%x\n", eth->ether_type);
	eth_type = rte_be_to_cpu_16(eth->ether_type);
	*len = sizeof(struct rte_ether_hdr);
	switch (eth_type) {
	case RTE_ETHER_TYPE_IPV4:
		return DUMP_IPV4;
	default:
		DOCA_DLOG_DBG("eth_type:0x%x not support.\n", eth_type);
		return DUMP_END;
	}
}

static uint8_t
dump_ipv4_hdr(uint8_t *data, uint32_t *len, char *buff)
{
	struct rte_ipv4_hdr *ip_hdr = (struct rte_ipv4_hdr *)data;

	log_ipv4("ip src:", rte_be_to_cpu_32(ip_hdr->src_addr));
	log_ipv4("dst:", rte_be_to_cpu_32(ip_hdr->dst_addr));
	log_buff("proto:%d\n", ip_hdr->next_proto_id);
	*len = sizeof(struct rte_ipv4_hdr);
	switch (ip_hdr->next_proto_id) {
	case DOCA_PROTO_UDP:
		return DUMP_UDP;
	case DOCA_PROTO_GRE:
		return DUMP_GRE;
	case DOCA_PROTO_TCP:
		return DUMP_TCP;
	default:
		DOCA_DLOG_DBG("unsupport layer 4 type:0x%x\n", ip_hdr->next_proto_id);
		return DUMP_END;
	}
}

static uint8_t
dump_gre_hdr(uint8_t *data, uint32_t *len, char *buff)
{
	uint16_t proto;
	uint32_t key;
	struct rte_gre_hdr *gre_hdr = (struct rte_gre_hdr *)data;

	proto = rte_be_to_cpu_16(gre_hdr->proto);
	log_buff("gre proto:0x%x", proto);
	*len = sizeof(struct rte_gre_hdr);
	if (gre_hdr->k) {
		*len += sizeof(uint32_t);
		key = *((uint32_t *)(gre_hdr + 1));
		log_buff(",key:0x%x", rte_be_to_cpu_32(key));
	}
	switch (proto) {
	case RTE_ETHER_TYPE_IPV4:
		log_buff("\n");
		return DUMP_IPV4;
	default:
		DOCA_DLOG_DBG("unsupport gre protocol:0x%x\n", proto);
		return DUMP_END;
	}
}

static uint8_t
dump_vxlan_hdr(uint8_t *data, uint32_t *len, char *buff)
{
	struct rte_vxlan_hdr *vxlan = (struct rte_vxlan_hdr *)data;

	*len = sizeof(struct rte_vxlan_hdr);
	log_buff("vxlan flags:0x%x vni:0x%x\n",
		      rte_be_to_cpu_32(vxlan->vx_flags) >> 24,
		      rte_be_to_cpu_32(vxlan->vx_vni) >> 8);
	return DUMP_ETH;
}

static uint8_t
dump_udp_hdr(uint8_t *data, uint32_t *len, char *buff)
{
	struct rte_udp_hdr *udp = (struct rte_udp_hdr *)data;

	*len = sizeof(struct rte_udp_hdr);
	log_buff("udp src-port:%d dst-port:%d",
		      rte_be_to_cpu_16(udp->src_port),
		      rte_be_to_cpu_16(udp->dst_port));
	if (rte_be_to_cpu_16(udp->dst_port) == DOCA_VXLAN_DEFAULT_PORT) {
		log_buff("\n");
		return DUMP_VXLAN;
	}
	return DUMP_END;
}

static uint8_t
dump_tcp_hdr(uint8_t *data, uint32_t *len, char *buff)
{
	struct rte_tcp_hdr *tcp = (struct rte_tcp_hdr *)data;

	*len = sizeof(struct rte_tcp_hdr);
	log_buff("tcp src-port:%d dst-port:%d\n",
		      rte_be_to_cpu_16(tcp->src_port),
		      rte_be_to_cpu_16(tcp->dst_port));
	return DUMP_END;
}

struct dump_pkt_hdr dump_pkt_hdr[] = {
	{DUMP_ETH,	 dump_eth_hdr},
	{DUMP_IPV4,	 dump_ipv4_hdr},
	{DUMP_UDP,	 dump_udp_hdr},
	{DUMP_TCP,	 dump_tcp_hdr},
	{DUMP_VXLAN, dump_vxlan_hdr},
	{DUMP_GRE,	 dump_gre_hdr},
};

static void
dump_packet_buff(uint8_t *head, char *buff, uint32_t buff_len)
{
	uint32_t length;
	uint8_t next_proto;
	char layer_prefix[MAX_TMP_BUFF];

	layer_prefix[0] = '\0';
	next_proto = DUMP_ETH;
	do {
		log_buff("%s", layer_prefix);
		next_proto = dump_pkt_hdr[next_proto].dump_hdr(head, &length, buff);
		if (buff_len < length)
			return;
		buff_len -= length;
		head += length;
		sprintf(layer_prefix + strlen(layer_prefix), "  ");
	} while (next_proto != DUMP_END);
}

void
__dpdk_dbg_rte_mbuff_dump(const char *name, struct rte_mbuf *mb, FILE *f)
{
	struct rte_ether_hdr _eth_hdr;
	char buff[MAX_TMP_BUFF];
	uint8_t *ethdr;

	buff[0] = '\0';
	log_buff(
		"port:%d mbuff on core:%u,pkt_len:%u,data_len:%u,nb_segs:%u,ol_flags:0x%lx\n",
		mb->port, rte_lcore_id(), mb->pkt_len,
		mb->data_len, mb->nb_segs, mb->ol_flags);
	ethdr = (uint8_t *)((uintptr_t)rte_pktmbuf_read(mb, 0,
		sizeof(_eth_hdr), &_eth_hdr));
	dump_packet_buff(ethdr, buff, MAX_TMP_BUFF - 1);
	fprintf(f, "%s%s\n", name, buff);
}

static const char *
get_l4_type(uint8_t l4_type)
{
	switch (l4_type) {
	case DOCA_PROTO_UDP:
		return "udp";
	case DOCA_PROTO_TCP:
		return "tcp";
	case DOCA_PROTO_GRE:
		return "gre";
	default:
		return "unknown";
	}
}

static void
dump_flow_match(const struct doca_flow_match *match,
		char *buff)
{
	int i;

	if (match == NULL)
		return;
	log_buff("\nmatch items:");
	if (!utils_df_translate_is_mac_ignored(match->out_src_mac))
		log_mac("\n    outer-src-mac: ", match->out_src_mac);
	if (!utils_df_translate_is_mac_ignored(match->out_dst_mac))
		log_mac("\n    outer-dst-mac: ", match->out_dst_mac);
	if (match->out_eth_type)
		log_buff("\n  out-eth-type:0x%x", rte_be_to_cpu_16(match->out_eth_type));
	if (match->out_vlan_tci)
		log_buff("\n  out-vlan-tci:0x%x", rte_be_to_cpu_16(match->out_vlan_tci));
	if (!utils_df_translate_is_ip_ignored(&match->out_src_ip))
		log_ipv4("\n    outer-src-ip: ",
			rte_be_to_cpu_32(match->out_src_ip.ipv4_addr));
	if (!utils_df_translate_is_ip_ignored(&match->out_dst_ip))
		log_ipv4("\n    outer-dst-ip: ",
			rte_be_to_cpu_32(match->out_dst_ip.ipv4_addr));
	if (match->out_l4_type)
		log_buff("\n    outer-l4-type: %u[%s]",
			match->out_l4_type, get_l4_type(match->out_l4_type));
	if (match->out_src_port)
		log_buff("\n    outer-src-port: %u",
			rte_be_to_cpu_16(match->out_src_port));
	if (match->out_dst_port)
		log_buff("\n    outer-dst-port: %u",
			rte_be_to_cpu_16(match->out_dst_port));
	switch (match->tun.type) {
	case DOCA_FLOW_TUN_VXLAN:
		log_buff("\n    tun-type: vxlan,vni: 0x%x",
			rte_be_to_cpu_32(match->tun.vxlan_tun_id) >> 8);
		break;
	case DOCA_FLOW_TUN_GRE:
		log_buff("\n    tun-type: gre,key: 0x%x",
			rte_be_to_cpu_32(match->tun.gre_key));
		break;
	case DOCA_FLOW_TUN_GTPU:
		log_buff("\n    tun-type:gtpu,teid:0x%x",
			rte_be_to_cpu_32(match->tun.gtp_teid));
		break;
	case DOCA_FLOW_TUN_NISP:
		log_buff("\n    tun-type:nisp:");
		log_buff("\n        dw[0]:0x%x, dw[1]:0x%x",
			 rte_be_to_cpu_32(match->tun.nisp_hdr[0]),
			 rte_be_to_cpu_32(match->tun.nisp_hdr[1]));
		log_buff("\n        dw[2]:0x%x, dw[3]:0x%x",
			 rte_be_to_cpu_32(match->tun.nisp_hdr[2]),
			 rte_be_to_cpu_32(match->tun.nisp_hdr[3]));
		log_buff("\n        dw[4]:0x%x, dw[5]:0x%x",
			 rte_be_to_cpu_32(match->tun.nisp_hdr[4]),
			 rte_be_to_cpu_32(match->tun.nisp_hdr[5]));
		log_buff("\n        dw[6]:0x%x, dw[7]:0x%x",
			 rte_be_to_cpu_32(match->tun.nisp_hdr[6]),
			 rte_be_to_cpu_32(match->tun.nisp_hdr[7]));
		log_buff("\n        dw[8]:0x%x, dw[9]:0x%x",
			 rte_be_to_cpu_32(match->tun.nisp_hdr[8]),
			 rte_be_to_cpu_32(match->tun.nisp_hdr[9]));
		break;
	case DOCA_FLOW_TUN_AUDP:
		log_buff("\n    tun-type:nisp:");
		log_buff("\n        dw[0]:0x%x, dw[1]:0x%x",
			 rte_be_to_cpu_32(match->tun.audp_hdr[0]),
			 rte_be_to_cpu_32(match->tun.audp_hdr[1]));
		log_buff("\n        dw[2]:0x%x, dw[3]:0x%x",
			 rte_be_to_cpu_32(match->tun.audp_hdr[2]),
			 rte_be_to_cpu_32(match->tun.audp_hdr[3]));
		log_buff("\n        dw[4]:0x%x, dw[5]:0x%x",
			 rte_be_to_cpu_32(match->tun.audp_hdr[4]),
			 rte_be_to_cpu_32(match->tun.audp_hdr[5]));
		break;
	case DOCA_FLOW_TUN_ESP:
		log_buff("\n    tun-type:ipsec:");
		log_buff("\n        spi:0x%x, sn:0x%x",
			 rte_be_to_cpu_32(match->tun.esp_spi),
			 rte_be_to_cpu_32(match->tun.esp_sn));
		break;
	case DOCA_FLOW_TUN_NONE:
		log_buff("\n    tun-type: none");
		break;
	default:
		log_buff("\n    unknown tun type: %u ", match->tun.type);
	}
	if (match->tun.type != DOCA_FLOW_TUN_NONE) {
		if (!utils_df_translate_is_mac_ignored(match->in_src_mac))
			log_mac("\n    inner-src-mac: ", match->in_src_mac);
		if (!utils_df_translate_is_mac_ignored(match->in_dst_mac))
			log_mac("\n    inner-dst-mac: ", match->in_dst_mac);
		if (match->in_eth_type)
			log_buff("\n  inner-eth-type:0x%x", match->in_eth_type);
		if (match->in_vlan_tci)
			log_buff("\n  inner-vlan-tci:0x%x", match->in_vlan_tci);
		if (!utils_df_translate_is_ip_ignored(&match->in_src_ip))
			log_ipv4("\n    inner-src-ip: ",
				rte_be_to_cpu_32(match->in_src_ip.ipv4_addr));
		if (!utils_df_translate_is_ip_ignored(&match->in_dst_ip))
			log_ipv4("\n    inner-dst-ip: ",
				rte_be_to_cpu_32(match->in_dst_ip.ipv4_addr));
		if (match->in_l4_type)
			log_buff("\n    inner-l4-type: %u[%s]",
				match->in_l4_type, get_l4_type(match->in_l4_type));
		if (match->in_src_port)
			log_buff("\n    inner-src-port: 0x%x",
				rte_be_to_cpu_16(match->in_src_port));
		if (match->in_dst_port)
			log_buff("\n    inner-dst-port: 0x%x",
				rte_be_to_cpu_16(match->in_dst_port));
	}
	if (match->meta.pkt_meta)
		log_buff("\n    pkt_meta: 0x%x", match->meta.pkt_meta);
	for (i = 0; i < RTE_DIM(match->meta.u32); i++) {
		if (match->meta.u32[i])
			log_buff("\n    meta.u32[%d]: 0x%x", i, match->meta.u32[i]);
	}
	if (match->meta.port_meta)
		log_buff("\n    port_meta: 0x%x", match->meta.port_meta);
}

static void
dump_flow_actions(const struct doca_flow_actions *actions,
		  const struct doca_flow_action_descs *desc,
		  char *buff)
{
	if (actions == NULL)
		return;
	log_buff("\nmodify action:");
	if (actions->decap)
		log_buff("\n    decap: %d", actions->decap);
	if (!utils_df_translate_is_mac_ignored(actions->mod_src_mac))
		log_mac("\n    src-mac: ", actions->mod_src_mac);
	if (!utils_df_translate_is_mac_ignored(actions->mod_dst_mac))
		log_mac("\n    dst-mac: ", actions->mod_dst_mac);
	if (!utils_df_translate_is_ip_ignored(&actions->mod_src_ip))
		log_ipv4("\n    src-ipv4: ",
			rte_be_to_cpu_32(actions->mod_src_ip.ipv4_addr));
	if (!utils_df_translate_is_ip_ignored(&actions->mod_dst_ip))
		log_ipv4("\n    dst-ipv4: ",
			rte_be_to_cpu_32(actions->mod_dst_ip.ipv4_addr));
	if (actions->mod_src_port)
		log_buff("\n    src-port: 0x%x",
			rte_be_to_cpu_16(actions->mod_src_port));
	if (actions->mod_dst_port)
		log_buff("\n    dst-port: 0x%x",
			rte_be_to_cpu_16(actions->mod_dst_port));

	if (actions->has_encap) {
		const struct doca_flow_encap_action *encap;

		encap = &actions->encap;
		log_buff("\n    encap info:");
		log_mac("\n        src-mac: ", encap->src_mac);
		log_mac("\n        dst-mac: ", encap->dst_mac);
		log_ipv4("\n        src-ipv4: ",
			rte_be_to_cpu_32(encap->src_ip.ipv4_addr));
		log_ipv4("\n        dst-ipv4: ",
			rte_be_to_cpu_32(encap->dst_ip.ipv4_addr));
		switch (encap->tun.type) {
		case DOCA_FLOW_TUN_VXLAN:
			log_buff("\n        tun-type: vxlan,vni: 0x%x",
				rte_be_to_cpu_32(encap->tun.vxlan_tun_id) >> 8);
			break;
		case DOCA_FLOW_TUN_GRE:
			log_buff("\n        tun-type: gre,key: 0x%x",
				rte_be_to_cpu_32(encap->tun.gre_key));
			break;
		case DOCA_FLOW_TUN_GTPU:
			log_buff("\n        tun-type: gtpu,teid:0x%x",
				rte_be_to_cpu_32(encap->tun.gtp_teid));
			break;
		case DOCA_FLOW_TUN_NISP:
			log_buff("\n        tun-type: nisp (not supported)");
			break;
		case DOCA_FLOW_TUN_AUDP:
			log_buff("\n        tun-type: audp (not supported)");
			break;
		case DOCA_FLOW_TUN_ESP:
			log_buff("\n        tun-type: esp (not supported)");
			break;
		case DOCA_FLOW_TUN_NONE:
			log_buff("\n        tun-type: none");
			break;
		}
	}
	if (actions->meta.pkt_meta)
		log_buff("\n    pkt meta value:0x%x", actions->meta.pkt_meta);
	if (desc && desc->meta.pkt_meta.mask.u32)
		log_buff(" pkt meta mask:0x%x(%d)", desc->meta.pkt_meta.mask.u32, desc->meta.pkt_meta.type);
}

static void
dump_flow_monitor(const struct doca_flow_monitor *monitor,
		  char *buff)
{
	log_buff("\nmonitor:");
	if (monitor == NULL) {
		log_buff(" NULL");
		return;
	}
	if (monitor->flags == DOCA_FLOW_MONITOR_NONE)
		log_buff("\n    no monitor");
	if (monitor->flags & DOCA_FLOW_MONITOR_METER)
		log_buff("\n    meter");
	if (monitor->flags & DOCA_FLOW_MONITOR_COUNT)
		log_buff("\n    count");
	if (monitor->flags & DOCA_FLOW_MONITOR_AGING)
		log_buff("\n    age[timer:%u]", monitor->aging);
}

static void
dump_flow_fwd(const struct doca_flow_fwd *fwd,
	      char *buff)
{
	int i = 0;

	log_buff("\nfwd action:");
	if (fwd == NULL) {
		log_buff(" NULL");
		return;
	}
	log_buff("\n");
	switch (fwd->type) {
	case DOCA_FLOW_FWD_NONE:
		log_buff("    not fwd.");
		break;
	case DOCA_FLOW_FWD_RSS:
		log_buff("    rss: ");
		if (!utils_df_translate_is_fwd_changeable(fwd)) {
			for (i = 0; i < fwd->num_of_queues; i++)
				log_buff("%d ", fwd->rss_queues[i]);
		}
		break;
	case DOCA_FLOW_FWD_PORT:
		log_buff("    port:%u ", fwd->port_id);
		break;
	case DOCA_FLOW_FWD_PIPE:
		log_buff("    pipe:%p ", fwd->next_pipe);
		break;
	case DOCA_FLOW_FWD_DROP:
		log_buff("    drop");
		break;
	case DOCA_FLOW_FWD_ORDERED_LIST_PIPE:
		log_buff("    ordered_list: pipe=%p idx=%u",
				fwd->ordered_list_pipe.pipe,
				fwd->ordered_list_pipe.idx);
		break;
	default:
		log_buff("    unknown fwd type:%u ", fwd->type);
		break;
	}
}

void
__dpdk_dbg_pipe_config_dump(const struct doca_flow_match *match,
		 const struct doca_flow_actions *actions,
		 const struct doca_flow_action_descs *desc,
		 const struct doca_flow_monitor *monitor,
		 const struct doca_flow_fwd *fwd)
{
	char buff[MAX_TMP_BUFF];

	buff[0] = '\0';
	dump_flow_match(match, buff);
	dump_flow_actions(actions, desc, buff);
	dump_flow_monitor(monitor, buff);
	dump_flow_fwd(fwd, buff);
	DOCA_DLOG_DBG("%s\n", buff);
}
