/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdarg.h>
#include <stdio.h>
#include <bsd/string.h>

#include <rte_ethdev.h>
#include <rte_gre.h>
#include <rte_gtp.h>
#include <rte_mtr.h>
#include <rte_vxlan.h>
#include <rte_hash.h>
#include <rte_hash_crc.h>

#include <doca_log.h>

#include "dpdk_engine.h"
#include "dpdk_flow_hws.h"
#include "dpdk_flow_sws.h"
#include "dpdk_table_hws.h"
#include "dpdk_table_sws.h"
#include "dpdk_flow.h"
#include "dpdk_port.h"
#include "dpdk_layer.h"
#include "dpdk_port.h"
#include "dpdk_meter_controller.h"
#include "priv_doca_mempool.h"
#include "dpdk_dbg.h"
#include "doca_flow_priv.h"
#include "encap_table.h"
#include "engine_model.h"
#include "utils_df_translate.h"

#include "doca_flow_layer.h"
#include "flow_age.h"
#include "id_pool.h"
#include "dpdk_flow_hws_legacy.h"
#include "pipe_lpm.h"
#include "dpdk_utilities.h"
#include "dpdk_group_pool.h"
#include "dpdk_fwd_groups.h"

/* #define SUPPORT_METER 0 */
#define HWS_MODE "hws"
#define VERIFY_MODE "verify"
#define VNF_HWS_MODE "vnf,hws"
#define MAX_VNI (0xFFFFFF)
#define DOCA_INGRESS_DEFAULT_GROUP (1)
#define DOCA_INGRESS_DEFAULT_GROUP_ISOLATED (0)
#define DPDK_PORT_DEFAULT_RULE_GROUP (0)
#define MAX_MODE_ARG_LEN (256)
#define CTRL_PIPE_MAX_ACT_TEMPLATE 1

DOCA_LOG_REGISTER(dpdk_engine);

struct dpdk_engine {
	bool has_acl;
	bool layer_registered;
	int mode;
	uint32_t mgmt_id; /* Management queue ID */
	rte_spinlock_t mgmt_lock;
	uint32_t queues_in_use;
	uint32_t nb_meters;
	uint32_t nb_counters;
	uint32_t ingress_default_group;
	struct dpdk_engine_ops ops;
	struct flow_ct_ops *ct_ops;
};

static struct doca_flow_cfg doca_flow_cfg = {0};
struct dpdk_engine dpdk_engine;

static struct doca_flow_port *dpdk_used_ports[DPDK_PORT_MAX_PORT_ID];
struct doca_flow_port *flow_switch;

#define DOCA_ID_POOL_METER_NAME         "meter_id"
#define DOCA_ID_POOL_POLICY_NAME        "policy_id"

enum dpdk_domain {
	DPDK_DOMAIN_NIC = 0,
	DPDK_DOMAIN_FDB,
};

static void
dpdk_flow_age_item_update(void *age_ctx,
			  struct flow_age_item *age_item);
static int
dpdk_flow_age_pkts_query(void *age_ctx, uint64_t *pkts);

static enum dpdk_domain
dpdk_engine_domain_get(struct doca_flow_port *port, const struct doca_flow_fwd *fwd)
{
	enum dpdk_domain domain;

	if ((fwd && (fwd->type == DOCA_FLOW_FWD_RSS))
		|| (!IS_PORT_PF(port) && !IS_PORT_REPRESENTOR(port))
		|| engine_model_is_mode(ENGINE_MODEL_MODE_VNF))
		domain = DPDK_DOMAIN_NIC;
	else
		domain = DPDK_DOMAIN_FDB;

	return domain;
}


static inline uint32_t
dpdk_engine_group_id_get_default(enum dpdk_domain domain)
{
	uint32_t group_id;

	group_id = domain == DPDK_DOMAIN_NIC ? dpdk_engine.ingress_default_group : 0;
	return group_id;
}

static inline uint32_t
dpdk_engine_group_id_get_by_pipe(struct doca_flow_pipe *pipe)
{
	return pipe->group_id;
}

void
dpdk_engine_pipe_q_ctx_group_set(struct doca_flow_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q,
				 const struct doca_flow_fwd *fwd, bool encrypt)
{
	enum dpdk_domain domain;

	domain = dpdk_engine_domain_get(pipe->port, fwd);
	pipe_q->attr.group = dpdk_engine_group_id_get_by_pipe(pipe);
	if (encrypt && (domain != DPDK_DOMAIN_FDB)) {
		pipe_q->attr.egress = 1;
		return;
	}
	pipe_q->attr.transfer = domain == DPDK_DOMAIN_FDB ? 1 : 0;
}

static __rte_always_inline int
is_port_representor(uint16_t portid, bool *is_representor)
{
	struct rte_eth_dev_info dev_info;
	int ret;

	ret = rte_eth_dev_info_get(portid, &dev_info);
	if (ret < 0) {
		DOCA_LOG_ERR("getting device (port %u) info: %s", portid, strerror(-ret));
		return -1;
	}
	*is_representor = !!(*dev_info.dev_flags & RTE_ETH_DEV_REPRESENTOR);
	return 0;
}

bool
mode_is_sws()
{
	return (dpdk_engine.mode & DPDK_MODE_SWS) == DPDK_MODE_SWS;
}

static bool
mode_is_hws()
{
	return (dpdk_engine.mode & DPDK_MODE_HWS) == DPDK_MODE_HWS;
}

void
dpdk_engine_flow_attr_set(struct doca_flow_port *port,
			  struct dpdk_fwd_groups_req *req,
			  const struct doca_flow_fwd *fwd)
{
	enum dpdk_domain domain;

	domain = dpdk_engine_domain_get(port, fwd);
	req->transfer = domain == DPDK_DOMAIN_FDB ? true : false;
	if (req->transfer == 0 || mode_is_sws())
		req->ingress = true;
}

static __rte_always_inline void
management_queue_lock();

static __rte_always_inline void
management_queue_unlock();

const char*
dpdk_engine_mode_name(void)
{
	enum engine_model_mode_type mode;

	mode = engine_model_get_mode();
	if (mode_is_hws() && mode != ENGINE_MODEL_MODE_VNF)
		return "unsupported";

	return engine_model_get_mode_name(mode);
}

static int
dpdk_engine_mode_parse(const char *mode_args)
{
	char mode_str[MAX_MODE_ARG_LEN] = { 0 };
	int mode = 0;
	int len;
	char *str;

	if (mode_args == NULL)
		return DPDK_MODE_UNDEF;

	len = strlen(mode_args);
	if (len >= MAX_MODE_ARG_LEN - 1) {
		DOCA_DLOG_ERR("The mode argments to long: %s , should be less than : %d\n",
			mode_args, MAX_MODE_ARG_LEN);
		return DPDK_MODE_UNDEF;
	}

	memcpy(mode_str, mode_args, len);
	mode = DPDK_MODE_SWS;
	str = strtok(mode_str, ",");
	while (str != NULL) {
		if (!strncmp(str, HWS_MODE, strlen(HWS_MODE)))
			mode = DPDK_MODE_HWS;
		str = strtok(NULL, ",");
	}

	return mode;
}

static int
dpdk_sws_port_configure(struct doca_flow_port *port)
{
	return 0;
}

int
dpdk_flow_sws_create_encap_table(struct dpdk_pipe *dpdk_pipe,
		struct rte_flow_action_raw_decap *decap_conf,
		struct rte_flow_action_raw_encap *encap_conf)
{
	return 0;
}

static uint32_t internal_meta;

int
dpdk_internal_meta_register(uint8_t bit_width)
{
	struct engine_shared_resource_cfg cfg = {.meta.bit_width = bit_width};
	uint32_t meta_id = internal_meta;
	int ret;

	if (dpdk_internal_meta_allocate_done()) {
		DOCA_LOG_ERR("failed to register internal meta - registration finished");
		return -ENOTSUP;
	}

	if (meta_id >= DPDK_META_MAX_SEGMENTS) {
		DOCA_LOG_ERR("failed to register internal meta - too many registered");
		return -ENOTSUP;
	}

	ret = engine_shared_resource_cfg_set(ENGINE_SHARED_RESOURCE_INTERNAL_META, meta_id, &cfg);
	if (ret) {
		DOCA_LOG_ERR("failed to register internal meta  - shared resource configure");
		return ret;
	}

	ret = engine_shared_resource_bind_bulk(ENGINE_SHARED_RESOURCE_INTERNAL_META, &meta_id,
					       1, NULL, NULL);
	if (ret) {
		DOCA_LOG_ERR("failed to register internal meta - bind resource");
		return ret;
	}

	internal_meta++;
	return (int)meta_id;
}

struct flow_ct_ops *
dpdk_engine_ct_ops_get(void)
{
	return dpdk_engine.ct_ops;
}

void
flow_ct_register(struct flow_ct_ops *ops)
{
	assert(dpdk_engine.ct_ops == NULL);
	dpdk_engine.ct_ops = ops;
}

static void
dpdk_engine_register_ops()
{
	struct dpdk_engine_ops *ops;

	ops = &dpdk_engine.ops;
	/* hws ops register */
	if (dpdk_engine.mode & DPDK_MODE_HWS) {
		ops->dpdk_port_configure = dpdk_flow_hws_port_configure;
		ops->dpdk_create_encap_table = dpdk_flow_hws_create_encap_table;
		return;
	}

	/* sws ops register */
	ops->dpdk_port_configure = dpdk_sws_port_configure;
	ops->dpdk_create_encap_table = dpdk_flow_sws_create_encap_table;
}

int
dpdk_engine_encap_table_create(struct dpdk_pipe *dpdk_pipe,
			       struct rte_flow_action_raw_decap *decap_conf,
			       struct rte_flow_action_raw_encap *encap_conf)
{
	return dpdk_engine.ops.dpdk_create_encap_table(dpdk_pipe, decap_conf, encap_conf);
}

static __rte_always_inline void
management_queue_lock()
{
	rte_spinlock_lock(&dpdk_engine.mgmt_lock);
}

static __rte_always_inline void
management_queue_unlock()
{
	rte_spinlock_unlock(&dpdk_engine.mgmt_lock);
}

bool
dpdk_engine_mode_is_hws(void)
{
	return mode_is_hws();
}

bool
dpdk_engine_mode_is_sws(void)
{
	return mode_is_sws();
}

void
dpdk_engine_mgmt_id_lock(void)
{
	management_queue_lock();
}

void
dpdk_engine_mgmt_id_unlock(void)
{
	management_queue_unlock();
}


uint16_t
dpdk_engine_mgmt_id_get(void)
{
	return dpdk_engine.mgmt_id;
}

uint16_t
dpdk_engine_get_queues_in_use()
{
	return dpdk_engine.queues_in_use;
}

uint16_t
dpdk_engine_get_queues()
{
	return doca_flow_cfg.queues;
}

uint16_t
dpdk_engine_get_queue_depth()
{
	return doca_flow_cfg.queue_depth;
}

uint32_t
dpdk_engine_meters_get_number()
{
	return dpdk_engine.nb_meters;
}

uint32_t
dpdk_engine_shared_meters_get_number()
{
	return doca_flow_cfg.nr_shared_resources[DOCA_FLOW_SHARED_RESOURCE_METER];
}

uint32_t
dpdk_engine_counters_get_number()
{
	return dpdk_engine.nb_counters;
}

int
dpdk_engine_init(const struct doca_flow_cfg *cfg)
{
	int mode_type;
	struct dpdk_layer_cfg layer_cfg = {0};
	uint16_t total_queues = cfg->queues + 1; /* Add 1 for mgmt queue */
	struct flow_age_cfg age_cfg = {0};
	int ret = 0;

	if (cfg->queues > rte_lcore_count()) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - queues value can't be %u,larger than total cores (%d)",
			      cfg->queues, rte_lcore_count());
		return -EINVAL;
	}
	if (cfg->queues == 0) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - queues value cannot be zero");
		return -EINVAL;
	}
	doca_flow_cfg = *cfg;
	mode_type = dpdk_engine_mode_parse(cfg->mode_args);
	if (mode_type & DPDK_MODE_UNDEF) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - mode param %s unsupported", cfg->mode_args);
		return -EINVAL;
	}
	memset(&dpdk_engine, 0, sizeof(dpdk_engine));
	memset(dpdk_used_ports, 0, sizeof(dpdk_used_ports));
	dpdk_engine.mode = mode_type;
	if (cfg->queue_depth > MAX_QUEUE_DEPTH) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - invalid queues size %d, must be in (0, %d]",
			     cfg->queue_depth, MAX_QUEUE_DEPTH);
		return -EINVAL;
	}
	if (cfg->queue_depth == 0) {
		doca_flow_cfg.queue_depth = DEFAULT_QUEUE_DEPTH;
		DOCA_LOG_INFO("queue depth is zero, set it to default %d.", DEFAULT_QUEUE_DEPTH);
	}

	if (dpdk_engine_ct_ops_get() != NULL) {
		if (engine_model_is_mode(ENGINE_MODEL_MODE_VNF) || !dpdk_engine_mode_is_hws()) {
			DOCA_LOG_ERR("failed to initialize dpdk engine - connection Tracking is supported on HWS mode and switch mode only");
			return -EINVAL;
		}
	}

	/* index of management queue */
	dpdk_engine.mgmt_id = cfg->queues;
	dpdk_engine.nb_meters = cfg->resource.nb_meters;
	dpdk_engine.nb_counters = cfg->resource.nb_counters;
	dpdk_engine.queues_in_use = total_queues;
	dpdk_engine.ingress_default_group =
			engine_model_is_isolated() ?
			DOCA_INGRESS_DEFAULT_GROUP_ISOLATED :
			DOCA_INGRESS_DEFAULT_GROUP;
	dpdk_engine_register_ops();
	ret = encap_table_init(DPDK_PIPE_ACTIONS_MAX_ENCAPS);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - init encap table");
		goto release;
	}
	age_cfg.update_age_item = dpdk_flow_age_item_update;
	age_cfg.query_pkts = dpdk_flow_age_pkts_query;
	ret = flow_age_init(&age_cfg);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - init flow aging array");
		goto release;
	}
	if (mode_is_hws()) {
		struct dpdk_hws_cfg hws_cfg = {0};

		hws_cfg.queues = doca_flow_cfg.queues + 1;  /* one internal queue */
		hws_cfg.ctrl_queue_id = doca_flow_cfg.queues;
		hws_cfg.queue_depth = doca_flow_cfg.queue_depth;
		hws_cfg.resource = doca_flow_cfg.resource;
		memcpy(hws_cfg.nr_shared_resources, doca_flow_cfg.nr_shared_resources,
		       sizeof(hws_cfg.nr_shared_resources));
		ret = dpdk_flow_hws_init(&hws_cfg);
		if (ret < 0) {
			DOCA_LOG_ERR("failed to initialize dpdk engine - init hws flow, ret=%d", ret);
			goto release;
		}
	}

	if (mode_is_sws()) {
		layer_cfg.flow_cfg.ops = dpdk_flow_sws_get_ops();
		layer_cfg.table_cfg.ops = dpdk_table_sws_get_ops();
		dpdk_meter_set_type(DPDK_METER_TYPE_POLICY);
	} else if (mode_is_hws()) {
		layer_cfg.flow_cfg.ops = dpdk_flow_hws_get_ops();
		layer_cfg.table_cfg.ops = dpdk_table_hws_get_ops();
		dpdk_meter_set_type(DPDK_METER_TYPE_MARK);
	} else {
		DOCA_LOG_ERR("failed to initialize dpdk engine - can't determine dpdk mode");
		goto release;
	}

	ret = dpdk_meter_profile_ports_set(DPDK_PORT_MAX_PORT_ID);
	if (ret) {
		DOCA_LOG_ERR("failed dpdk engine init - meter profile settings failed ret=%d",
			     ret);
		goto release;
	}

	ret = dpdk_pipe_legacy_init_early();
	if (ret < 0) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - dpdk pipe legacy early initialization failed ret=%d",
			     ret);
		goto release;
	}

	ret = dpdk_layer_register(&layer_cfg);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - dpdk layer register failed ret=%d",
			     ret);
		goto release;
	}

	ret = dpdk_pipe_legacy_init_late();
	if (ret < 0) {
		DOCA_LOG_ERR("failed to initialize dpdk engine - dpdk pipe legacy late initialization failed ret=%d",
			     ret);
		goto release;
	}

	dpdk_engine.layer_registered = true;
	rte_spinlock_init(&dpdk_engine.mgmt_lock);
	return 0;
release:
	dpdk_engine_destroy();
	return ret;
}

void
dpdk_engine_destroy(void)
{
	if (mode_is_hws())
		dpdk_flow_hws_destroy();

	if (dpdk_engine.layer_registered)
		dpdk_layer_unregister();

	dpdk_meter_profile_ports_unset();

	internal_meta = 0;
	encap_table_destroy();
}

void
dpdk_acl_enable(void)
{
	dpdk_engine.has_acl = true;
}

int
dpdk_engine_mode_get(void)
{
	return dpdk_engine.mode;
}

void
dpdk_engine_invoke_entry_process_cb(struct doca_flow_pipe_entry *entry, enum doca_flow_entry_op op)
{
	if (!doca_flow_cfg.cb)
		return;

	doca_flow_cfg.cb(entry, entry->status, op, entry->user_ctx);
}

static void
entry_update_status(struct doca_flow_pipe_entry *entry, enum dpdk_flow_request_status status)
{
	switch (status) {
	case DPDK_FLOW_REQ_STATUS_OK:
		entry->status = DOCA_FLOW_ENTRY_STATUS_SUCCESS;
		break;
	default:
		entry->status = DOCA_FLOW_ENTRY_STATUS_ERROR;
		break;
	}
}

/*
 * Should use dpdk_flow, currently not possible due
 * https://redmine.mellanox.com/issues/3168220
 */
static int
dpdk_flow_sws_destroy(uint16_t port_id, struct rte_flow *flow)
{
	struct rte_flow_error error;
	int ret;

	ret = rte_flow_destroy(port_id, flow, &error);
	if (ret) {
		DOCA_DLOG_ERR(
			"Port %u free flow fail, type %d message: %s",
			port_id, error.type, error.message ?
			error.message : "(no stated reason)");
		return ret;
	}
	return 0;
}

static struct doca_flow_port*
get_doca_port_byid(uint8_t port_id)
{
	int i;

	for (i = 0; i < DPDK_PORT_MAX_PORT_ID; i++) {

		if (dpdk_used_ports[i] == NULL)
			continue;

		if (dpdk_used_ports[i]->port_id != port_id)
			continue;

		return dpdk_used_ports[i];
	}
	return NULL;
}

static int
entry_remove_async(struct doca_flow_port *port, struct doca_flow_pipe_entry *entry,
		   dpdk_flow_completion_cb cb, void *usr_ctx, bool persistent, bool wait_for_bulk)
{
	struct dpdk_flow_queue *queue;
	struct dpdk_flow_request req;
	int rc;

	queue = dpdk_port_get_flow_queue(port->dpdk_port, entry->pipe_queue);
	if (!queue)
		return -EINVAL;

	req.tracker = &entry->tracker;
	req.cb = cb;
	req.callback_ctx = entry;
	req.persistent = persistent;
	req.wait_for_bulk = wait_for_bulk;

	entry->user_ctx = usr_ctx;
	rc = dpdk_flow_destroy(queue, &req);
	if (unlikely(rc))
		return rc;
	return 0;
}

int
dpdk_rte_flow_destroy(struct doca_flow_port *port, void *usr_ctx, dpdk_flow_completion_cb cb,
		      struct doca_flow_pipe_entry *entry, bool persistent, bool wait_for_bulk)
{
	int rc;

	rc = entry_remove_async(port, entry, cb, usr_ctx, persistent, wait_for_bulk);
	if (unlikely(rc))
		return rc;
	return 0;
}

static void
add_encap_entry_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			      void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Encap entry addition completion got null entry");
		return;
	}

	DOCA_DLOG_DBG("encap entry %p add completed with status %d", entry, entry->status);

	entry_update_status(entry, status);
}

static int
dpdk_profile_id_destroy(int port_id, uint32_t profile_id)
{
	uint32_t dpdk_profile_id = DPDK_ID(profile_id);

	return dpdk_meter_profile_delete(port_id, dpdk_profile_id);
}

static struct dpdk_hws_item *
ctrl_pipe_table_create(struct doca_flow_port *ctrl_port,
		       struct dpdk_pipe *dpdk_pipe,
		       struct dpdk_pipe_q_ctx *pipe_q,
		       struct rte_flow_item *items,
		       struct rte_flow_action *actions,
		       struct rte_flow_attr *attr,
		       struct dpdk_table *res)
{
	struct rte_flow_action *ctrl_actions_arr[CTRL_PIPE_MAX_ACT_TEMPLATE];
	struct rte_flow_template_table_attr tatr = {0};
	struct dpdk_hws_item *hws_entry;

	tatr.nb_flows = 1;
	tatr.flow_attr = *attr;
	ctrl_actions_arr[0] = actions;

	hws_entry = dpdk_flow_hws_ctrl_table_create(ctrl_port, dpdk_pipe, pipe_q, &tatr, items,
						    ctrl_actions_arr, CTRL_PIPE_MAX_ACT_TEMPLATE);
	if (unlikely(hws_entry == NULL)) {
		DOCA_LOG_ERR("port %u build control flow fail - failed creating ctrl table",
			     ctrl_port->port_id);
		return NULL;
	}
	res->table = hws_entry->table;
	return hws_entry;
}

static void
rm_control_entry_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			       void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Control entry deletion completion got null entry");
		return;
	}
	free(entry);
}

static void
add_control_entry_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			    void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;
	struct doca_flow_port *port;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Control entry addition completion got null entry");
		return;
	}

	DOCA_DLOG_DBG("control entry %p add completed with status %d", entry, entry->status);

	if (status != DPDK_FLOW_REQ_STATUS_OK) {
		port = (struct doca_flow_port *)entry->obj_ctx;
		dpdk_rte_flow_destroy(port, NULL, rm_control_entry_completion_cb, entry,
				      true /* persistent */, false /* wait_for_bulk */);
	}
}

int
dpdk_engine_pipe_get_ctrl_table(struct doca_flow_port *ctrl_port, struct dpdk_pipe *dpdk_pipe,
				struct dpdk_pipe_q_ctx *pipe_q, struct rte_flow_item *items,
				struct rte_flow_action *actions, struct rte_flow_attr *attr,
				struct dpdk_table *res, struct dpdk_hws_item **hws_item)
{
	int rc = 0;

	if (mode_is_sws()) {
		res->table = NULL;
		goto out;
	}

	*hws_item = ctrl_pipe_table_create(ctrl_port, dpdk_pipe, pipe_q, items, actions, attr, res);
	if (unlikely(*hws_item == NULL)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating ctrl table");
		rc = -ENOENT;
	}

out:
	res->pattern_template_index = 0;
	res->action_template_index = 0;
	return rc;
}

static struct doca_flow_pipe_entry *
dpdk_control_flow_create(struct doca_flow_port *port,
			 uint16_t pipe_queue,
			 struct dpdk_control_flow_cfg *cfg)
{
	struct dpdk_flow_queue *queue;
	struct dpdk_flow_request req = {0};
	struct dpdk_flow_params dpdk_flow_params = {0};
	struct dpdk_table res = {0};
	struct rte_flow_action actions[DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_item items[DPDK_PIPE_MAX_ITEMS];
	struct doca_flow_pipe_entry *entry;
	struct dpdk_hws_item *hws_item = NULL;
	uint8_t nb_count;
	int rc;

	if (unlikely(cfg->attr.transfer))
		port = flow_switch;

	queue = dpdk_port_get_flow_queue(port->dpdk_port, pipe_queue);
	if (unlikely(!queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create control flow - invalid queue %u",
					pipe_queue);
		return NULL;
	}

	entry = calloc(1, sizeof(*entry));
	if (!entry) {
		DOCA_LOG_ERR("failed to create control flow - out of mem (port %u)", port->port_id);
		return NULL;
	}
	entry->pipe_queue = pipe_queue;
	entry->obj_ctx = port;

	memset(items, 0x0, sizeof(items));
	memset(actions, 0x0, sizeof(actions));

	/* build items */
	nb_count = 0;
	items[nb_count++].type = RTE_FLOW_ITEM_TYPE_ETH;
	if (cfg->meta.data) {
		items[nb_count].type = RTE_FLOW_ITEM_TYPE_META;
		items[nb_count].spec = &cfg->meta;
		items[nb_count++].mask = &cfg->meta_mask;
	}
	if (cfg->attr.transfer) {
		cfg->ethdev.port_id = port->port_id;
		items[nb_count].type = RTE_FLOW_ITEM_TYPE_REPRESENTED_PORT;
		items[nb_count].mask = &rte_flow_item_ethdev_mask;
		items[nb_count++].spec = &cfg->ethdev;
	}
	items[nb_count].type = RTE_FLOW_ITEM_TYPE_END;

	/* build actions */
	nb_count = 0;
	switch (cfg->fwd_type) {
	case DPDK_FWD_JUMP:
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_JUMP;
		actions[nb_count++].conf = &cfg->jump;
		break;
	case DPDK_FWD_QUEUE:
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_QUEUE;
		actions[nb_count++].conf = &cfg->queue;
		break;
	case DPDK_FWD_PORT:
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_PORT_ID;
		actions[nb_count++].conf = &cfg->port;
		break;
	case DPDK_FWD_RSS:
		actions[nb_count].type = RTE_FLOW_ACTION_TYPE_RSS;
		actions[nb_count++].conf = &cfg->rss.conf;
		if (cfg->mark.conf.id) {
			actions[nb_count].type = RTE_FLOW_ACTION_TYPE_MARK;
			actions[nb_count++].conf = &cfg->mark;
		}
		break;
	default:
		DOCA_LOG_ERR("failed to create entry - unsupported fwd type %u", cfg->fwd_type);
		free(entry);
		return NULL;
	};
	actions[nb_count].type = RTE_FLOW_ACTION_TYPE_END;

	rc = dpdk_engine_pipe_get_ctrl_table(port, NULL, NULL, items, actions, &cfg->attr, &res,
					     &hws_item);
	if (rc) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - table create rc=%d", rc);
		free(entry);
		return NULL;
	}

	dpdk_flow_params.items = items;
	dpdk_flow_params.actions = actions;
	dpdk_flow_params.table = res.table;
	dpdk_flow_params.item_template_index = res.pattern_template_index;
	dpdk_flow_params.action_template_index = res.action_template_index;
	dpdk_flow_params.ingress = cfg->attr.ingress;
	dpdk_flow_params.egress = cfg->attr.egress;
	dpdk_flow_params.transfer = cfg->attr.transfer;
	dpdk_flow_params.group = cfg->attr.group;
	dpdk_flow_params.priority = cfg->attr.priority;

	req.tracker = &entry->tracker;
	req.cb = add_control_entry_completion_cb;
	req.callback_ctx = entry;
	req.persistent = true;
	req.wait_for_bulk = false;

	rc = dpdk_flow_create(queue, &dpdk_flow_params, &req);
	if (rc) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - flow create rc=%d", rc);
		free(entry);
		return NULL;
	}
	if (mode_is_hws())
		hws_item->flow = dpdk_flow_tracker_get_flow(&entry->tracker);

	return entry;
}

struct doca_flow_pipe_entry *
dpdk_engine_encap_entry_alloc(struct doca_flow_pipe *pipe)
{
	struct doca_flow_pipe_entry *encap_entry;

	management_queue_lock();
	encap_entry = dpdk_pipe_entry_alloc(pipe, dpdk_engine.mgmt_id);
	management_queue_unlock();

	if (!encap_entry)
		return NULL;

	encap_entry->pipe_queue = dpdk_engine.mgmt_id;
	return encap_entry;
}

int
dpdk_engine_encap_entry_tx_flow_add(int port_id, int meta_id,
				    struct rte_flow_action_raw_decap *decap_conf,
				    struct rte_flow_action_raw_encap *encap_conf,
				    struct doca_flow_pipe_entry *entry)
{
	struct dpdk_flow_params dpdk_flow_params = {0};
	struct dpdk_flow_request req = {0};
	struct dpdk_flow_queue *queue;
	struct doca_flow_pipe *pipe;
	struct doca_flow_port *port;
	struct rte_flow_action actions[DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_item items[DPDK_PIPE_MAX_ITEMS];
	struct rte_flow_item_meta meta = {0};
	struct rte_flow_attr attr;
	int action_idx = 0;
	int rc;

	port = get_doca_port_byid(port_id);
	if (unlikely(!port)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create encap flow - invalid port %u",
					port_id);
		return -EINVAL;
	}

	queue = dpdk_port_get_flow_queue(port->dpdk_port, entry->pipe_queue /* mgmt_queue */);
	if (unlikely(!queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create encap flow - invalid queue %u",
					entry->pipe_queue);
		return -EINVAL;
	}

	memset(&attr, 0x0, sizeof(struct rte_flow_attr));
	memset(items, 0x0, sizeof(items));
	memset(actions, 0x0, sizeof(actions));

	attr.group = DOCA_EGRESS_DEFAULT_GROUP;
	attr.egress = 1;
	meta.data = meta_id;
	items[0].type = RTE_FLOW_ITEM_TYPE_META;
	items[0].spec = &meta;
	items[1].type = RTE_FLOW_ITEM_TYPE_END;
	if (decap_conf) {
		actions[action_idx].type = RTE_FLOW_ACTION_TYPE_RAW_DECAP;
		actions[action_idx++].conf = decap_conf;
	}
	actions[action_idx].type = RTE_FLOW_ACTION_TYPE_RAW_ENCAP;
	actions[action_idx++].conf = encap_conf;
	actions[action_idx].type = RTE_FLOW_ACTION_TYPE_END;


	if (mode_is_sws()) {
		dpdk_flow_params.table = NULL;
	} else {
		pipe = entry->pipe;

		if (!pipe || !pipe->dpdk_pipe.encap_hws_ctx) {
			DOCA_LOG_RATE_LIMIT_ERR(
				"failed to create encap egress flow - encap ctx is NULL.");
			return -EINVAL;
		}

		dpdk_flow_params.table = pipe->dpdk_pipe.encap_hws_ctx->table;
	}

	dpdk_flow_params.items = items;
	dpdk_flow_params.actions = actions;
	dpdk_flow_params.item_template_index = 0;
	dpdk_flow_params.action_template_index = 0;
	dpdk_flow_params.ingress = false;
	dpdk_flow_params.egress = true;
	dpdk_flow_params.transfer = false;
	dpdk_flow_params.group = DOCA_EGRESS_DEFAULT_GROUP;
	dpdk_flow_params.priority = 0;

	req.tracker = &entry->tracker;
	req.cb = add_encap_entry_completion_cb;
	req.callback_ctx = entry;
	req.persistent = true;
	req.wait_for_bulk = false;

	rc = dpdk_flow_create(queue, &dpdk_flow_params, &req);
	if (rc) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - flow create rc=%d", rc);
		return rc;
	}

	return 0;
}

/*
 * Should use dpdk_flow, currently not possible due
 * https://redmine.mellanox.com/issues/3168220
 */
static int
encap_entry_flow_destroy(uint16_t port_id, struct doca_flow_pipe_entry *entry)
{
	if (mode_is_sws())
		return dpdk_flow_sws_destroy(port_id, dpdk_flow_tracker_get_flow(&entry->tracker));

	struct rte_flow_op_attr qatr = {
		.postpone = 0,
	};
	uint32_t flag = HWS_FLOW_FLAG_PERSISTENT;
	struct doca_flow_port *port;
	int rc;

	port = get_doca_port_byid(port_id);
	/* upon hairpin + tx_encap we might reach a point in which the port is destroyed
	 * when deleting this entry. if we are in this case when we know there are no
	 * in-flight flows, otherwise attempt to deplete the queue before destroying
	 * the encap flow
	 */
	management_queue_lock();
	if (port != NULL) {
		rc = dpdk_flow_entries_process(port, entry->pipe_queue, PULL_TIMEOUT_US, 0);
		if (rc < 0) {
			DOCA_LOG_RATE_LIMIT_ERR("%s failed - could not deplete queue %d rc=%d",
						__func__, entry->pipe_queue, rc);
			goto out;
		}
	}

	rc = dpdk_flow_hws_mgmt_q_flow_destroy_legacy(port_id, entry->pipe_queue, &qatr, entry,
				     NULL, NULL, flag);
	if (rc)
		DOCA_LOG_RATE_LIMIT_ERR("%s failed - destroy flow(%p)",
					__func__, dpdk_flow_tracker_get_flow(&entry->tracker));

out:
	management_queue_unlock();
	return rc;
}

int
dpdk_engine_encap_entry_free(uint16_t port_id, struct doca_flow_pipe_entry *entry)
{
	return encap_entry_flow_destroy(port_id, entry);
}


static int
dpdk_default_flow_create(struct doca_flow_port *port,
			 struct dpdk_control_flow_cfg *cfg)
{
	struct doca_flow_pipe_entry *entry;

	entry = dpdk_control_flow_create(port, dpdk_engine.mgmt_id, cfg);
	if (entry == NULL) {
		DOCA_LOG_ERR("failed creating default flow on port %u", port->port_id);
		return -1;
	}
	LIST_INSERT_HEAD(&port->default_flow_list, entry, next);
	return 0;
}

static int
dpdk_root_rx_jump_create(struct doca_flow_port *port)
{
	struct dpdk_control_flow_cfg cfg = {0};

	cfg.attr.ingress = 1;
	cfg.attr.group = DPDK_PORT_DEFAULT_RULE_GROUP;
	cfg.fwd_type = DPDK_FWD_JUMP;
	cfg.jump.group = dpdk_engine.ingress_default_group;
	return dpdk_default_flow_create(port, &cfg);
}

static int
dpdk_root_tx_jump_create(struct doca_flow_port *port)
{
	struct dpdk_control_flow_cfg cfg = {0};

	cfg.attr.egress = 1;
	cfg.attr.group = DPDK_PORT_DEFAULT_RULE_GROUP;
	cfg.attr.priority = DOCA_EGRESS_DEFAULT_PRIO;
	cfg.fwd_type = DPDK_FWD_JUMP;
	cfg.jump.group = DOCA_EGRESS_DEFAULT_GROUP;
	return dpdk_default_flow_create(port, &cfg);
}

static int
dpdk_port_fdb_flow_create(struct doca_flow_port *port, uint16_t port_id)
{
	struct dpdk_control_flow_cfg cfg = {0};
	int ret;

	cfg.attr.transfer = 1;
	if (mode_is_sws())
		cfg.attr.ingress = 1;
	cfg.attr.priority = MAX_FLOW_PRIO;
	cfg.fwd_type = DPDK_FWD_PORT;
	cfg.port.id = port_id;
	ret = dpdk_default_flow_create(port, &cfg);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to create fdb port flow - pair port %u to %u ret=%d",
			     port->port_id, port_id, ret);
		return ret;
	}
	return 0;
}

static int
dpdk_nic_lag_hairpinq_flow_create(struct doca_flow_port *port,
				  uint16_t meta_index, uint16_t hairpinq)
{
	struct dpdk_control_flow_cfg cfg = {0};
	struct mode_param param;
	uint16_t meta_bit;
	int ret;

	cfg.attr.ingress = 1;
	cfg.attr.group = dpdk_engine.ingress_default_group;
	cfg.attr.priority = MAX_FLOW_PRIO - 1; /* higher than rss rule */
	cfg.fwd_type = DPDK_FWD_QUEUE;

	meta_bit = meta_index == 0 ? LAG_P0_META_BIT : LAG_P1_META_BIT;
	cfg.meta.data = 1 << meta_bit;
	cfg.meta_mask.data = 1 << meta_bit;
	engine_model_get_param(&param);
	cfg.queue.index = (param.hybrid.secondary_lag_id ? meta_index : 0) + hairpinq;
	ret = dpdk_default_flow_create(port, &cfg);
	if (ret < 0) {
		DOCA_LOG_ERR("port %u create lag hairpinq rule on queue %u fail,ret:%d",
			port->port_id, hairpinq, ret);
		return ret;
	}
	return 0;
}

static int
dpdk_nic_hairpinq_flow_create(struct doca_flow_port *port, uint16_t hairpinq_idx)
{
	struct dpdk_control_flow_cfg cfg = {0};
	int ret;

	cfg.attr.ingress = 1;
	cfg.attr.group = dpdk_engine.ingress_default_group;
	cfg.attr.priority = MAX_FLOW_PRIO;
	cfg.fwd_type = DPDK_FWD_QUEUE;
	cfg.queue.index = hairpinq_idx;
	ret = dpdk_default_flow_create(port, &cfg);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to create nic hairpinq flow - rule port %u to hairpin queue %u ret=%d",
			     port->port_id, hairpinq_idx, ret);
		return ret;
	}
	return 0;
}

static int
dpdk_def_rss_flow_create(struct doca_flow_port *port, int queues)
{
	struct dpdk_control_flow_cfg cfg = {0};
	struct dpdk_action_rss_data *rss = &cfg.rss;
	int qidx, ret;

	cfg.attr.ingress = 1;
	cfg.attr.group = dpdk_engine.ingress_default_group;
	cfg.attr.priority = MAX_FLOW_PRIO;
	cfg.fwd_type = DPDK_FWD_RSS;
	memset(rss, 0x0, sizeof(*rss));
	for (qidx = 0; qidx < queues; qidx++)
		rss->queue[qidx] = qidx;
	rss->conf.queue_num = queues;
	rss->conf.func = RTE_ETH_HASH_FUNCTION_DEFAULT;
	rss->conf.types = ETH_RSS_IP | ETH_RSS_UDP | ETH_RSS_TCP;
	rss->conf.queue = rss->queue;
	ret = dpdk_default_flow_create(port, &cfg);
	if (ret < 0) {
		DOCA_LOG_ERR("create port %u def RSS rule fail, ret:%d", port->port_id, ret);
		return ret;
	}
	return 0;
}

static void
dpdk_flow_age_item_update(void *age_ctx,
			  struct flow_age_item *age_item)
{
	struct doca_flow_pipe_entry *entry = age_ctx;

	if (entry == NULL) {
		DOCA_LOG_ERR("update flow age_ctx is NULL.");
		return;
	}
	entry->age_item = age_item;
}

int
dpdk_port_legacy_is_invalid(struct doca_flow_port *port)
{
	if (port == NULL || port->state == PORT_STOPPED)
		return 1;
	return 0;
}

enum doca_flow_entry_status
dpdk_flow_pipe_entry_status_get(struct doca_flow_pipe_entry *entry)
{
	return entry->status;
}

int
dpdk_flow_entries_process(struct doca_flow_port *port,
		uint16_t pipe_queue, uint64_t timeout,
		uint32_t max_processed_entries)
{
	struct dpdk_flow_queue *queue;
	uint64_t end_tsc, us_hz;
	int rc, pulled_items = 0;
	uint32_t max_polls;

	if (dpdk_port_legacy_is_invalid(port)) {
		DOCA_DLOG_ERR("Invalid port");
		return -EINVAL;
	}

	queue = dpdk_port_get_flow_queue(port->dpdk_port, pipe_queue);
	if (unlikely(!queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed processing entries - invalid queue %u", pipe_queue);
		return -EINVAL;
	}

	us_hz = rte_get_timer_hz() / 1000000;  /*micro second*/;
	end_tsc = rte_rdtsc() + us_hz * timeout;
	max_polls = max_processed_entries == 0 ? queue->queue_depth : max_processed_entries;
	do {
		rc = dpdk_flow_poll(queue, max_polls);
		if (rc < 0)
			return rc;

		pulled_items += rc;
		if (pulled_items >= max_polls || !queue->in_progress_items)
			break;
	} while (rte_rdtsc() <= end_tsc);

	return pulled_items; /* polled entries */;
}

static int
dpdk_vnf_mode_init(struct doca_flow_port *port)
{
	if (engine_model_is_isolated()) {
		DOCA_LOG_DBG("Skip VNF default RSS creation due to ISOLATED mode");
		return 0;
	}
	return dpdk_def_rss_flow_create(port, doca_flow_cfg.queues);
}

static int
dpdk_switch_mode_init(struct doca_flow_port *port)
{
	if (engine_model_is_isolated()) {
		DOCA_LOG_DBG("Skip Switch default RSS creation due to ISOLATED mode");
		return 0;
	}
	return dpdk_def_rss_flow_create(port, doca_flow_cfg.queues);
}

static int
dpdk_hyprid_mode_init(struct doca_flow_port *port)
{
	int queues = doca_flow_cfg.queues;
	int ret = 0;

	if (IS_PORT_PF(port)) {
		ret |= dpdk_nic_lag_hairpinq_flow_create(port, 0, queues);
		ret |= dpdk_nic_lag_hairpinq_flow_create(port, 1, queues);
		if (!engine_model_is_isolated())
			ret |= dpdk_def_rss_flow_create(port, queues);
	} else if (IS_PORT_REPRESENTOR(port)) {
		if (!engine_model_is_isolated())
			ret = dpdk_def_rss_flow_create(port, queues);
	} else {
		/* for VF/SF ports, do port pair to self */
		ret = dpdk_port_legacy_pair(port, port);
	}
	return ret;
}

static int
get_fwd_groups(struct doca_flow_port *port, struct doca_flow_fwd *fwd)
{
	struct dpdk_fwd_groups *fwd_groups;
	struct dpdk_fwd_groups_req req = {0};
	uint32_t group_id;
	int ret;

	ret = dpdk_fwd_groups_req_build(port, fwd, &fwd_groups, &req);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building default fwd groups - build drop req failure");
		return ret;
	}

	ret = dpdk_fwd_groups_get(fwd_groups, &req, &group_id);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building default fwd groups - groups get failed for drop");
		return ret;
	}
	return 0;
}

static int
build_port_default_fwd_groups(struct doca_flow_port *port)
{
	struct doca_flow_fwd fwd = {0};
	struct doca_flow_port *req_port;
	int ret;

	if (mode_is_sws())
		return 0;

	switch (engine_model_get_mode()) {
	case ENGINE_MODEL_MODE_VNF:
		fwd.type = DOCA_FLOW_FWD_DROP;
		req_port = port;
		break;
	case ENGINE_MODEL_MODE_REMOTE_VNF:
	case ENGINE_MODEL_MODE_HYBRID:
	case ENGINE_MODEL_MODE_SWITCH:
		fwd.type = DOCA_FLOW_FWD_PORT;
		fwd.port_id = port->port_id;
		req_port = flow_switch;
		break;
	default:
		return 0;
	}

	ret = get_fwd_groups(req_port, &fwd);
	if (ret)
		return ret;

	return 0;
}

int
dpdk_port_legacy_init(struct doca_flow_port *port)
{
	int ret;

	if (dpdk_engine.ops.dpdk_port_configure(port))
		return -1;

	ret = build_port_default_fwd_groups(port);
	if (ret)
		return ret;

	/*
	 * In isolated mode rules should start in group 0 such that miss flows
	 * will go to kernel. In higher groups they are dropped.
	 */
	if (!engine_model_is_isolated()) {
		ret = dpdk_root_rx_jump_create(port);
		if (ret < 0) {
			DOCA_LOG_ERR("failed to initialize dpdk port %u - root rx jump flow", port->port_id);
			return -1;
		}
	}
	/*  NIC egress rules on representors is not supported */
	if (!IS_PORT_REPRESENTOR(port)) {
		ret = dpdk_root_tx_jump_create(port);
		if (ret < 0) {
			DOCA_LOG_WARN(
				"failed to initialize dpdk port %u - root tx jump flow", port->port_id);
		}
	}

	if (engine_model_is_mode(ENGINE_MODEL_MODE_VNF))
		return dpdk_vnf_mode_init(port);
	if (engine_model_is_mode(ENGINE_MODEL_MODE_SWITCH))
		return dpdk_switch_mode_init(port);
	if (engine_model_is_mode(ENGINE_MODEL_MODE_HYBRID))
		return dpdk_hyprid_mode_init(port);
	if (engine_model_is_mode(ENGINE_MODEL_MODE_REMOTE_VNF))
		/* remote-vnf handles the default flows in dpdk_port_legacy_pair API */
		return 0;
	DOCA_LOG_ERR("failed to initialize dpdk port %u - unsupport mode %d", port->port_id, dpdk_engine.mode);
	return -1;
}

int
flow_pipe_id_get(struct doca_flow_pipe *pipe, uint32_t *id)
{
	if (pipe == NULL || id == NULL) {
		DOCA_LOG_ERR("invalid parameter");
		return -EINVAL;
	}
	*id = dpdk_engine_group_id_get_by_pipe(pipe);
	return 0;
}

int
dpdk_engine_group_id_get(struct doca_flow_port *port,
			 struct doca_flow_pipe *pipe,
			 const struct doca_flow_fwd *fwd)
{
	enum dpdk_domain domain;
	struct dpdk_group_pool *group_pool;
	uint32_t group_id;
	int ret;

	domain = dpdk_engine_domain_get(port, fwd);
	if (pipe->is_root)
		group_id = dpdk_engine_group_id_get_default(domain);
	else {
		group_pool = dpdk_port_get_group_pool(port->dpdk_port);
		ret = dpdk_group_pool_alloc(group_pool, &group_id);
		if (ret)
			return ret;
	}

	pipe->group_id = group_id;
	return 0;
}

int
dpdk_engine_next_pipe_group_id_get(struct doca_flow_pipe *next_pipe, uint32_t *group_id)
{
	if (unlikely(!next_pipe)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to get next pipe's group - next_pipe is null");
		return -EINVAL;
	}

	if (next_pipe->is_root && engine_model_is_isolated()) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"failed to get next pipe's group - root pipe jump is forbidden");
		return -EINVAL;
	}

	if (next_pipe->type == ENGINE_PIPE_CT &&
	    doca_flow_layer_get_pipe_port(next_pipe)->ct_ctx->ct_pipe != NULL)
		next_pipe = doca_flow_layer_get_pipe_port(next_pipe)->ct_ctx->ct_pipe;

	*group_id = dpdk_engine_group_id_get_by_pipe(next_pipe);

	return 0;
}

static struct doca_flow_port *
dpdk_doca_port_alloc_byid(uint8_t port_id, const struct doca_flow_port_cfg *cfg)
{
	struct doca_flow_port *port;
	struct engine_bindable *global_bindable = NULL;
	uint32_t size;
	int rc;

	size = sizeof(struct doca_flow_port) + cfg->priv_data_size;
	port = (struct doca_flow_port *)calloc(1, size);
	if (port == NULL) {
		DOCA_LOG_ERR("failed to allocate doca port %u - cannot allocate %u bytes",
			     port_id, size);
		return NULL;
	}

	port->port_id = port_id;
	port->hairpin_peer = NULL;
	port->state = PORT_INIT;
	LIST_INIT(&port->pipe_list);
	LIST_INIT(&port->default_flow_list);
	rte_spinlock_init(&port->pipe_lock);
	rte_spinlock_init(&port->profile_lock);

	rc = engine_bindable_init(&port->bindable_obj, DF_ENGINE_BINDABLE_TYPE_PORT);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed to allocate doca port - stamping bindable content on port, rc=%d", rc);
		goto port_alloc_failed;
	}

	global_bindable = engine_shared_resource_get_global_bindable();
	if (unlikely(!global_bindable)) {
		DOCA_LOG_ERR("failed to allocate doca port - getting global bindable on port");
		goto port_alloc_failed;
	}

	rc = engine_bindable_set_parent(&port->bindable_obj, global_bindable);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed to allocate doca port - setting pipe hierarchy, rc=%d", rc);
		goto port_alloc_failed;
	}

	return port;


port_alloc_failed:
	free(port);
	return NULL;
}

static int
dpdk_port_save(struct doca_flow_port *port)
{
	int i = 0;

	for (i = 0; i < DPDK_PORT_MAX_PORT_ID; i++) {
		if (dpdk_used_ports[i] == NULL) {
			dpdk_used_ports[i] = port;
			port->idx = i;
			return 0;
		}
	}

	DOCA_LOG_ERR("failed to save port %u - out of mem", port->port_id);
	return -1;
}

static int
port_meter_id_pool_initialize(struct doca_flow_port *port)
{
	struct id_pool_cfg pool_cfg = {
		.size = dpdk_engine_meters_get_number(),
		.nb_caches = dpdk_engine.queues_in_use,
		.min = 1};

	snprintf(pool_cfg.name, ID_POOL_NAME_LEN, "%s_p%d",
		 DOCA_ID_POOL_METER_NAME, port->port_id);
	port->meter_pool = id_pool_create(&pool_cfg);

	snprintf(pool_cfg.name, ID_POOL_NAME_LEN, "%s_p%d",
		 DOCA_ID_POOL_POLICY_NAME, port->port_id);
	port->meter_policy_pool = id_pool_create(&pool_cfg);

	if (port->meter_pool == NULL ||
	    port->meter_policy_pool == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to initialize id pool - meter pool or meter policy is NULL");
		return -1;
	}

	return 0;
}

static int
dpdk_port_create_txmiss_table(struct doca_flow_port *port)
{
	struct rte_flow_action actions[DPDK_PIPE_MAX_ACTIONS] = {0};
	struct rte_flow_action action_masks[DPDK_PIPE_MAX_ACTIONS] = {0};
	struct rte_flow_action *actions_arr[1] = {actions};
	struct rte_flow_action *action_masks_arr[1] = {action_masks};
	struct rte_flow_item items[DPDK_PIPE_MAX_ITEMS] = {0};
	struct rte_flow_attr attr = {0};
	int action_idx = 0, action_mask_idx = 0, ret;
	struct rte_flow_action_jump jump = {.group = UINT32_MAX};
	struct dpdk_table_params table_params = {0};
	struct dpdk_table *table;
	uint16_t port_id;

	port_id = dpdk_port_get_id(port->dpdk_port);
	if (port->tx_miss != NULL)
		return 0;

	table = calloc(1, sizeof(*table));
	if (table == NULL) {
		DOCA_LOG_ERR("allocate dpdk_txmiss_table memory failed.");
		return -ENOMEM;
	}
	attr.group = DOCA_EGRESS_TXMISS_GROUP;
	attr.egress = 1;

	items[0].type = RTE_FLOW_ITEM_TYPE_ETH;
	items[1].type = RTE_FLOW_ITEM_TYPE_END;

	/* build actions */
	actions[action_idx].type = RTE_FLOW_ACTION_TYPE_JUMP;
	actions[action_idx++].conf = &jump;
	actions[action_idx++].conf = RTE_FLOW_ACTION_TYPE_END;

	/* build action_masks */
	action_masks[action_mask_idx].type = RTE_FLOW_ACTION_TYPE_JUMP;
	action_masks[action_mask_idx++].conf = NULL;
	action_masks[action_mask_idx].type = RTE_FLOW_ACTION_TYPE_END;

	table_params.port_id = port_id;
	table_params.items = items;
	table_params.actions = actions_arr;
	table_params.actions_mask = action_masks_arr;
	table_params.nb_actions = 1;
	table_params.flow_attr = &attr;
	table_params.nb_flows = DPDK_PIPE_ACTIONS_MAX_ENCAPS;

	ret = dpdk_table_create(&table_params, dpdk_port_get_table_ctx(port->dpdk_port), table);
	if (ret) {
		DOCA_LOG_ERR("%s failed - failed to create dpdk table", __func__);
		free(table);
		return ret;
	}

	port->tx_miss = table;
	return 0;
}

static int
dpdk_flow_hws_destroy_txmiss_table(struct doca_flow_port *port)
{
	struct dpdk_table_port_ctx *ctx;
	int rc;

	if (port == NULL || port->dpdk_port == NULL || port->tx_miss == NULL)
		return 0;

	ctx = dpdk_port_get_table_ctx(port->dpdk_port);
	if (unlikely(ctx == NULL)) {
		DOCA_LOG_RATE_LIMIT_ERR("%s failed - port table ctx is NULL", __func__);
		return -ENOENT;
	}

	rc = dpdk_table_destroy(ctx, port->tx_miss);
	if (rc)
		DOCA_DLOG_ERR("failed to destroy tx miss table - dpdk table destroy, ret=%d", rc);

	free(port->tx_miss);
	port->tx_miss = NULL;
	return rc;
}

struct doca_flow_port *
dpdk_port_legacy_start(uint16_t port_id, const struct doca_flow_port_cfg *cfg)
{
	int ret;
	bool is_representor;
	struct doca_flow_port *port;
	struct dpdk_port_cfg port_cfg;
	char name[PRIV_MEMPOOL_NAME_LEN];
	uint16_t proxy_port_id;
	uint32_t num_profiles;

	if (get_doca_port_byid(port_id) != NULL) {
		DOCA_LOG_ERR("failed to start port %u - it already exists",
			       port_id);
		return NULL;
	}
	ret = is_port_representor(port_id, &is_representor);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to start port %u - check port is representor",
			       port_id);
		return NULL;
	}
	if (is_representor && engine_model_is_mode(ENGINE_MODEL_MODE_VNF)) {
		DOCA_LOG_ERR("failed to start port %u - vnf mode not support representor",
			     port_id);
		return NULL;
	}
	port = dpdk_doca_port_alloc_byid(port_id, cfg);
	if (port == NULL)
		return NULL;

	ret = dpdk_port_save(port);
	if (ret) {
		DOCA_LOG_ERR("failed to start port %u - port save, ret=%d", port_id, ret);
		goto fail_port_start;
	}

	port->state = PORT_STARTED;
	if (is_representor)
		port->switch_attr |= IS_REPRESENTOR;
	/* if not VNF mode, check the eswitch proxy port */
	if (!engine_model_is_mode(ENGINE_MODEL_MODE_VNF)) {
		ret = rte_flow_pick_transfer_proxy(port_id, &proxy_port_id, NULL);
		if (ret) {
			/* in eswitch mode, vfs/sfs can't get the proxy port id */
			DOCA_LOG_INFO("port id %u proxy port is null", port_id);
		} else {
			port->proxy_port_id = proxy_port_id;
			if (port_id == proxy_port_id) {
				port->switch_attr |= IS_PF;
				if (!flow_switch) {
					flow_switch = port;
				} else if (flow_switch->port_id != proxy_port_id) {
					DOCA_LOG_ERR("failed to start port - need to enable LAG configuration for merged switch");
					goto fail_port_start;
				}
			}
		}
	}

	port_cfg.port_id = port_id;
	port_cfg.nr_queues = dpdk_engine.queues_in_use;
	port_cfg.queue_depth = engine_model_get_pipe_queue_depth();
	port_cfg.group_pool_cfg.min_group_id = DPDK_GROUP_POOL_MIN_ID;
	port_cfg.mgmt_queue_id = dpdk_engine.mgmt_id;
	port_cfg.mgmt_lock = &dpdk_engine.mgmt_lock;
	port->dpdk_port = dpdk_port_create(&port_cfg);
	if (unlikely(!port->dpdk_port)) {
		DOCA_LOG_ERR("failed to start port %u - creating dpdk port", port_id);
		goto fail_port_start;
	}

	ret = dpdk_port_legacy_init(port);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to start port %u - init port, ret=%d", port_id, ret);
		goto fail_port_start;
	}

	port->profile_id_generator = dpdk_engine_shared_meters_get_number();
	port->profile_ids = calloc(1, sizeof(*port->profile_ids));
	if (!port->profile_ids) {
		DOCA_DLOG_ERR("failed allocating port (%u) - profile_ids",
			      port_id);
		goto fail_port_start;
	}

	ret = dpdk_port_create_txmiss_table(port);
	if (ret !=  0) {
		DOCA_LOG_ERR("failed to create tx miss table on port(%u), ret=%d.", port->port_id, ret);
		goto fail_port_start;
	}

	snprintf(name, PRIV_MEMPOOL_NAME_LEN, "p%d_prof_hash", port_id);
	ret = dpdk_profile_hash_init(DPDK_MAX_PORT_PROFILES,
				     &port->profile_ids->id_hash, name,
				     RTE_HASH_EXTRA_FLAGS_RW_CONCURRENCY);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to start port %u - init profile hash, ret=%d", port_id, ret);
		goto fail_port_start;
	}

	num_profiles = dpdk_engine_shared_meters_get_number();
	if (dpdk_engine_meters_get_number()) {
		ret = port_meter_id_pool_initialize(port);
		if (ret)
			goto fail_port_start;

		num_profiles += DPDK_MAX_PORT_PROFILES;
	}

	if (num_profiles) {
		ret = dpdk_meter_profile_alloc(port->port_id, num_profiles);
		if (ret)
			goto fail_port_start;
	}

	if (dpdk_engine_ct_ops_get() != NULL) {
		port->ct_ctx = dpdk_engine_ct_ops_get()->init(port->port_id, &doca_flow_cfg, port);
		if (port->ct_ctx == NULL) {
			DOCA_LOG_ERR("failed to start port %hu - init doca-ct module", port_id);
			goto fail_port_start;
		}
	}

	return port;

fail_port_start:
	dpdk_port_legacy_stop(port);
	return NULL;
}

static void
rm_default_flow_entry_completion_cb(enum dpdk_flow_request_op op,
				    enum dpdk_flow_request_status status, void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Entry removal completion got null entry");
		return;
	}

	DOCA_DLOG_DBG("entry %p remove completed with status %d", entry, entry->status);

	free(entry);
}

static void
dpdk_root_tx_encrypt_remove(struct doca_flow_port *port)
{
	if (port->tx_encrypt_flow) {
		int ret;

		ret = dpdk_rte_flow_destroy(port, NULL, rm_default_flow_entry_completion_cb,
					    port->tx_encrypt_flow, true /* persistent */,
					    false /* wait_for_bulk */);
		if (ret)
			DOCA_LOG_WARN("failed destroying tx_encrypt_flow on port %u - ret=%d",
				      port->port_id, ret);
		port->tx_encrypt_flow = NULL;
	}
}

int
dpdk_root_tx_encrypt_create(struct doca_flow_port *port, uint32_t group_id)
{
	struct dpdk_control_flow_cfg cfg = {0};

	cfg.attr.egress = 1;
	cfg.attr.group = DPDK_PORT_DEFAULT_RULE_GROUP;
	cfg.attr.priority = DOCA_EGRESS_ENCRYPT_PRIO;
	cfg.fwd_type = DPDK_FWD_JUMP;
	cfg.jump.group = group_id;
	port->tx_encrypt_flow = dpdk_control_flow_create(port, dpdk_engine_mgmt_id_get(), &cfg);
	if (port->tx_encrypt_flow == NULL) {
		DOCA_LOG_ERR("failed creating tx encrypt flow on port %u", port->port_id);
		return -1;
	}
	return 0;
}

int
dpdk_port_legacy_stop(struct doca_flow_port *port)
{
	int ret, i;
	uint32_t profile_id;
	struct doca_flow_pipe_entry *flow_entry;

	if (port->state == PORT_STOPPED) {
		ret = -EALREADY;
		goto error;
	}

	while ((flow_entry = LIST_FIRST(&port->default_flow_list))) {
		LIST_REMOVE(flow_entry, next);
		ret = dpdk_rte_flow_destroy(port, NULL, rm_default_flow_entry_completion_cb,
					    flow_entry, true /* persistent */,
					    false /* wait_for_bulk */);
		if (ret)
			DOCA_LOG_WARN("failed destroying default flow on port %u - ret=%d",
				      port->port_id, ret);
	}

	if (port->tx_encrypt_flow)
		dpdk_root_tx_encrypt_remove(port);

	if (port->tx_miss)
		dpdk_flow_hws_destroy_txmiss_table(port);

	if (port->profile_ids) {
		dpdk_profile_hash_destroy(port->profile_ids->id_hash);
		for (i = 0; i < DPDK_MAX_PORT_PROFILES; i++) {
			profile_id = port->profile_ids->ids[i];
			if (profile_id)
				dpdk_profile_id_destroy(port->port_id, profile_id);
		}
		free(port->profile_ids);
		port->profile_ids = NULL;
	}

	if (port->meter_policy_pool) {
		id_pool_destroy(port->meter_policy_pool);
		port->meter_policy_pool = NULL;
	}

	if (port->meter_pool) {
		id_pool_destroy(port->meter_pool);
		port->meter_pool = NULL;
	}

	if (port->dpdk_port) {
		dpdk_port_destroy(port->dpdk_port);
		port->dpdk_port = NULL;
	}

	dpdk_meter_profile_free(port->port_id);

	port->state = PORT_STOPPED;
	if (port == flow_switch)
		flow_switch = NULL;
	return 0;
error:
	DOCA_LOG_ERR("failed to stop port %u, ret=%d", port->port_id, ret);
	return ret;
}


static int
dpdk_port_pair_fdb_rules_create(struct doca_flow_port *port, struct doca_flow_port *pair_port)
{
	int ret;

	ret = dpdk_port_fdb_flow_create(port, pair_port->port_id);
	if (ret < 0)
		return ret;
	ret = dpdk_port_fdb_flow_create(pair_port, port->port_id);
	if (ret < 0)
		return ret;
	return 0;
}

static int
dpdk_port_pair_nic_rules_create(struct doca_flow_port *port, struct doca_flow_port *pair_port)
{
	int ret;

	ret = dpdk_nic_hairpinq_flow_create(port, doca_flow_cfg.queues);
	if (ret < 0)
		return ret;
	if (port != pair_port) {
		ret = dpdk_nic_hairpinq_flow_create(pair_port, doca_flow_cfg.queues);
		if (ret < 0)
			return ret;
	}
	return 0;
}

static int
build_pair_port_fwd_groups(struct doca_flow_port *port, struct doca_flow_port *pair_port)
{
	struct doca_flow_fwd fwd = {0};
	int ret;

	if (mode_is_sws())
		return 0;

	fwd.type = DOCA_FLOW_FWD_PORT;
	fwd.port_id = pair_port->port_id;
	ret = get_fwd_groups(port, &fwd);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building pair port - failed getting fwd_groups, port id %u",
			port->port_id);
		return ret;
	}

	if (port->port_id == pair_port->port_id) {
		DOCA_LOG_RATE_LIMIT_DBG("build once for same port and pair_port: id %u", port->port_id);
		return 0;
	}

	fwd.type = DOCA_FLOW_FWD_PORT;
	fwd.port_id = port->port_id;
	ret = get_fwd_groups(pair_port, &fwd);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building pair port - failed getting fwd_groups, port_id %u",
			pair_port->port_id);
		return ret;
	}

	return 0;
}

int
dpdk_port_legacy_pair(struct doca_flow_port *port, struct doca_flow_port *pair_port)
{
	int ret;

	DOCA_LOG_INFO("doca pair ports id %u - %u", port->port_id, pair_port->port_id);
	/* create port pair FDB flows with the lowest priority if one port is representor port */
	if (IS_PORT_REPRESENTOR(port) || IS_PORT_REPRESENTOR(pair_port))
		return dpdk_port_pair_fdb_rules_create(port, pair_port);
	if (port->hairpin_peer || pair_port->hairpin_peer) {
		DOCA_LOG_ERR("port %u can't pair to %u", port->port_id, pair_port->port_id);
		return -1;
	}
	/* save the hairpin peers information */
	port->hairpin_peer = pair_port;
	pair_port->hairpin_peer = port;

	ret = build_pair_port_fwd_groups(port, pair_port);
	if (ret)
		return ret;

	/* create port pair NIC flows with the lowest priority for physical port */
	if (engine_model_is_mode(ENGINE_MODEL_MODE_REMOTE_VNF)
		|| engine_model_is_mode(ENGINE_MODEL_MODE_HYBRID))
		return dpdk_port_pair_nic_rules_create(port, pair_port);
	return 0;
}

void
dpdk_port_legacy_flush_pipes(struct doca_flow_port *port)
{
	struct doca_flow_pipe *pipe;

	if (dpdk_port_legacy_is_invalid(port))
		return;
	rte_spinlock_lock(&port->pipe_lock);
	while ((pipe = LIST_FIRST(&port->pipe_list)))
		dpdk_pipe_destroy(port, pipe);
	rte_spinlock_unlock(&port->pipe_lock);
}

void
dpdk_port_legacy_destroy(struct doca_flow_port *port)
{
	if (port->ct_ctx != NULL) {
		dpdk_engine_ct_ops_get()->close(port->ct_ctx);
		port->ct_ctx = NULL;
	}
	dpdk_port_legacy_flush_pipes(port);
	if (port->state != PORT_STOPPED)
		dpdk_port_legacy_stop(port);

	dpdk_used_ports[port->idx] = NULL;
	free(port);
}

struct doca_flow_port *
dpdk_port_legacy_switch_get(void)
{
	return flow_switch;
}

void
dpdk_port_pipes_dump(struct doca_flow_port *port, FILE *f)
{
	static const char *nic_stats_border = "########################";
	struct doca_flow_pipe *doca_pipe;

	fprintf(f, "\n  %s Pipe info for port %-2d mode %s %s\n", nic_stats_border,
	       port->port_id, dpdk_engine_mode_name(), nic_stats_border);
	rte_spinlock_lock(&port->pipe_lock);
	doca_pipe = LIST_FIRST(&port->pipe_list);
	while (doca_pipe) {
		dpdk_pipe_dump(port, doca_pipe, f);
		doca_pipe = LIST_NEXT(doca_pipe, next);
	}
	rte_spinlock_unlock(&port->pipe_lock);
}

static int
dpdk_flow_age_pkts_query(void *age_ctx, uint64_t *pkts)
{
	struct doca_flow_query stats = {0};
	struct doca_flow_pipe_entry *entry;
	int ret;

	entry = (struct doca_flow_pipe_entry *)age_ctx;
	if (entry == NULL) {
		DOCA_DLOG_ERR("query null flow entry");
		return -1;
	}
	ret = dpdk_pipe_entry_query(entry, &stats);
	if (ret < 0)
		return ret;
	*pkts = stats.total_pkts;
	return 0;
}

struct doca_flow_port *
dpdk_engine_get_hairpin_peer(struct doca_flow_port *port)
{
	if (port)
		return port->hairpin_peer;

	return NULL;
}
