/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DPDK_ENGINE_H_
#define DPDK_ENGINE_H_

#include <stdio.h>

#include <rte_flow.h>
#include <rte_vxlan.h>
#include <rte_ethdev.h>
#include <rte_mtr.h>
#include <rte_gre.h>

#include "dpdk_pipe.h"
#include "dpdk_pipe_legacy.h"
#include "dpdk_pipe_items.h"
#include "dpdk_pipe_actions.h"
#include "doca_flow.h"
#include "dpdk_flow_hws_legacy.h"
#include "dpdk_flow.h"
#include "engine_pipe.h"
#include "dpdk_internal_meta.h"
#include "dpdk_fwd_groups.h"


#ifdef __cplusplus
extern "C" {
#endif

#define MAX_FLOW_PRIO 3
#define DOCA_EGRESS_DEFAULT_PRIO (1)
#define DOCA_EGRESS_ENCRYPT_PRIO (0)
#define DOCA_EGRESS_DEFAULT_GROUP (1)
#define DOCA_EGRESS_TXMISS_GROUP (2)
#define DEFAULT_NB_ACTIONS (1)
#define NB_CTRL_PIPE_ACTIONS_ARR (1)
#define DPDK_ID(doca_id) ((doca_id) - 1)
#define DEFAULT_NB_FLOWS (1 << 13)

enum {
	DPDK_MATCH_METER_COLOR_GREEN =  (1ULL << 29),
	DPDK_MATCH_METER_COLOR_RED =    (1ULL << 30),
	DPDK_META_INTERNAL =            (1ULL << 31),
};

/* group 0-2 are reserved */
#define DPDK_GROUP_POOL_MIN_ID (3)

#define DOCA_PIPE(dpdk_pipe) (container_of(dpdk_pipe, struct doca_flow_pipe, dpdk_pipe))

struct dpdk_table;

enum dpdk_mode {
	DPDK_MODE_HWS = (1 << 0),
	DPDK_MODE_SWS = (1 << 1),
	DPDK_MODE_UNDEF = (1 << 2),
};

enum port_state {
	PORT_INIT = (1 << 0),
	PORT_STARTED = (1 << 1),
	PORT_STOPPED = (1 << 2),
};

struct dpdk_control_flow_cfg {
	struct rte_flow_attr attr;
	/* items */
	struct rte_flow_item_ethdev ethdev;
	struct rte_flow_item_meta meta;
	struct rte_flow_item_meta meta_mask;
	/* actions */
	enum dpdk_fwd_type fwd_type;
	union {
		struct rte_flow_action_jump jump;
		struct rte_flow_action_port_id port;
		struct rte_flow_action_queue queue;
		struct {
			struct dpdk_action_rss_data rss;
			struct dpdk_action_mark_data mark;
		};
	};
};

struct endecap_layer {
	uint16_t layer;
	void (*fill_data)(uint8_t **header,
		const struct doca_flow_encap_action *cfg,
		bool is_encap);
};

enum dpdk_encap_hdr {
	FILL_ETH_HDR = (1 << 1),
	FILL_IPV4_HDR = (1 << 2),
	FILL_UDP_HDR = (1 << 3),
	FILL_VXLAN_HDR = (1 << 4),
	FILL_GRE_HDR = (1 << 5),
	FILL_IPV6_HDR = (1 << 6),
	FILL_GTP_HDR = (1 << 7),
};

struct dpdk_engine_ops {
	int (*dpdk_port_configure)(struct doca_flow_port *port);
	int (*dpdk_create_encap_table)(struct dpdk_pipe *dpdk_pipe,
		struct rte_flow_action_raw_decap *decap_conf,
		struct rte_flow_action_raw_encap *encap_conf);
};

/********* dpdk_meter.c **************/
int dpdk_profile_hash_init(int max_profiles, struct profile_hash **profile_hash,
			   const char *name, uint8_t extra_flag);

void dpdk_profile_hash_destroy(struct profile_hash *profile_hash_ins);

int dpdk_meter_profile_id_get(struct doca_flow_port *port,
			      struct dpdk_pipe_q_ctx *pipe_q,
			      const struct doca_flow_monitor *mon,
			      uint32_t *profile_id);

int dpdk_profile_hash_get_key_idx(struct profile_hash *profile_hash_ins,
				  const struct doca_flow_monitor *mon);

/********* start of dpdk_engine_modify_field.c **************/

void dpdk_meta_flow_item_build(struct dpdk_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q,
			       const struct doca_flow_match *mask);

int dpdk_action_set_meta_build(struct dpdk_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q,
		const struct doca_flow_actions *actions, struct doca_flow_action_desc *desc,
		uint8_t act_arr_idx);

int dpdk_action_descs_build(struct dpdk_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_arr_idx,
		const struct doca_flow_pipe_cfg *cfg);

int dpdk_action_modify_build(struct dpdk_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q,
			     const struct doca_flow_actions *actions, enum rte_flow_field_id field, int level,
			     void *action_val, struct doca_flow_action_desc *desc, uint8_t act_arr_idx);

/* register internal metadata, return meta index. */
int dpdk_internal_meta_register(uint8_t bit_width);

/* return bit offset of user meta, start from tag 0. */
int dpdk_user_meta_tag_resolve(struct doca_flow_match *match, void *addr);

/********* end of dpdk_engine_modify_field.c **************/

int
dpdk_engine_init(const struct doca_flow_cfg *cfg);

void
dpdk_engine_pipe_q_ctx_group_set(struct doca_flow_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q,
				 const struct doca_flow_fwd *fwd, bool encrypt);

uint16_t
dpdk_engine_get_queues_in_use();

uint16_t
dpdk_engine_get_queues();

uint16_t
dpdk_engine_get_queue_depth();

uint32_t
dpdk_engine_meters_get_number();

uint32_t
dpdk_engine_counters_get_number();

const char*
dpdk_engine_mode_name(void);

uint16_t
dpdk_engine_mgmt_id_get(void);

void
dpdk_engine_mgmt_id_lock(void);

void
dpdk_engine_mgmt_id_unlock(void);

bool
dpdk_engine_mode_is_hws(void);

bool
dpdk_engine_mode_is_sws(void);

void
dpdk_engine_invoke_entry_process_cb(struct doca_flow_pipe_entry *entry, enum doca_flow_entry_op op);

int
dpdk_engine_encap_table_create(struct dpdk_pipe *dpdk_pipe,
			       struct rte_flow_action_raw_decap *decap_conf,
			       struct rte_flow_action_raw_encap *encap_conf);

struct doca_flow_pipe_entry *
dpdk_engine_encap_entry_alloc(struct doca_flow_pipe *pipe);

int
dpdk_engine_encap_entry_tx_flow_add(int port_id, int meta_id,
				    struct rte_flow_action_raw_decap *decap_conf,
				    struct rte_flow_action_raw_encap *encap_conf,
				    struct doca_flow_pipe_entry *entry);

int
dpdk_engine_encap_entry_free(uint16_t port_id, struct doca_flow_pipe_entry *entry);

int
dpdk_engine_pipe_get_ctrl_table(struct doca_flow_port *ctrl_port, struct dpdk_pipe *dpdk_pipe,
				struct dpdk_pipe_q_ctx *pipe_q, struct rte_flow_item *items,
				struct rte_flow_action *actions, struct rte_flow_attr *attr,
				struct dpdk_table *res, struct dpdk_hws_item **hws_item);
int
dpdk_engine_group_id_get(struct doca_flow_port *port,
			 struct doca_flow_pipe *pipe,
			 const struct doca_flow_fwd *fwd);

void
dpdk_engine_destroy(void);

int
dpdk_rte_flow_destroy(struct doca_flow_port *port, void *usr_ctx, dpdk_flow_completion_cb cb,
		      struct doca_flow_pipe_entry *entry, bool persistent, bool wait_for_bulk);

int
dpdk_port_legacy_init(struct doca_flow_port *port);

struct doca_flow_port *
dpdk_port_legacy_start(uint16_t port_id, const struct doca_flow_port_cfg *cfg);

int
dpdk_port_legacy_is_invalid(struct doca_flow_port *port);

int
dpdk_port_legacy_stop(struct doca_flow_port *port);

int
dpdk_port_legacy_pair(struct doca_flow_port *port, struct doca_flow_port *pair_port);

void
dpdk_port_legacy_destroy(struct doca_flow_port *port);

struct doca_flow_port *
dpdk_port_legacy_switch_get(void);

void
dpdk_port_legacy_flush_pipes(struct doca_flow_port *port);

void
dpdk_port_pipes_dump(struct doca_flow_port *port, FILE *f);

void
dpdk_acl_enable(void);

int
dpdk_flow_entries_process(struct doca_flow_port *port,
			uint16_t pipe_queue, uint64_t timeout,
			uint32_t max_processed_entries);

enum doca_flow_entry_status
dpdk_flow_pipe_entry_status_get(struct doca_flow_pipe_entry *entry);

int
dpdk_engine_mode_get(void);

int
dpdk_engine_next_pipe_group_id_get(struct doca_flow_pipe *next_pipe,
				   uint32_t *group_id);

int
dpdk_fwd_groups_req_build(struct doca_flow_port *port,
			  const struct doca_flow_fwd *fwd,
			  struct dpdk_fwd_groups **fwd_groups,
			  struct dpdk_fwd_groups_req *req);

void
dpdk_engine_flow_attr_set(struct doca_flow_port *port,
			  struct dpdk_fwd_groups_req *req,
			  const struct doca_flow_fwd *fwd);

int
dpdk_root_tx_encrypt_create(struct doca_flow_port *port, uint32_t group_id);

struct flow_ct_ops *
dpdk_engine_ct_ops_get(void);

struct doca_flow_port*
dpdk_engine_get_hairpin_peer(struct doca_flow_port *port);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DPDK_ENGINE_H_ */
