/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <doca_log.h>

#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "dpdk_pipe_actions.h"
#include "engine_shared_resources.h"
#include "utils_field.h"

DOCA_LOG_REGISTER(dpdk_engine_modify_field);

#define BIT_IS_SET(data, index) (((uint8_t *)data)[(index) >> 3] & (1 << ((index) & 0x7)))
#define BIT_IS_SET_BE(data, index) (((uint8_t *)data)[(index) >> 3] & (1 << (7 - ((index) & 0x7))))
#define BIT_INDEX_INVALID(index) ((index) == -1)
#define BIT_SEGMENT(data, offset, width) (((data) >> (offset)) & ((1ULL << (width)) - 1))
#define SEGMENT_SHIFT(data, width, offset) (((data) & ((1ULL << (width)) - 1)) << (offset))
#define REG_INDEX(offset) ((offset) >> 5)
#define REG_OFFSET(offset) ((offset) & 0x1f)
#define REG_BITS(n_regs) ((n_regs) << 5)
#define BIT2BYTE(n) (((n) >> 3) + (((n) & 0x7) ? 1 : 0))
#define REG_COUNT(offset, width) (REG_INDEX((offset) + (width)) - REG_INDEX((offset)))
#define PKT_META_COUNT (1)
#define PKT_META_BITS REG_BITS(PKT_META_COUNT)
#define DPDK_TAG_INDEX_MASK 0xFF
#define CHK(x)                          \
	do {                            \
		ret = (x);              \
		if (ret)                \
			return ret;     \
	} while (0)

#define member_size(type, member) sizeof(((type *)0)->member)

#define match_member(field, header, level) \
	{ offsetof(struct doca_flow_match, field), member_size(struct doca_flow_match, field), header, level }

uint32_t header_bit_width[] = {
	[RTE_FLOW_FIELD_META] = 4 << 3,
	[RTE_FLOW_FIELD_TAG] = 4 << 3,
	[RTE_FLOW_FIELD_MAC_DST] = 6 << 3,
	[RTE_FLOW_FIELD_MAC_SRC] = 6 << 3,
	[RTE_FLOW_FIELD_MAC_TYPE] = 2 << 3,
	[RTE_FLOW_FIELD_VLAN_TYPE] = 2 << 3,
	[RTE_FLOW_FIELD_VLAN_ID] = 12,
	[RTE_FLOW_FIELD_IPV4_SRC] = 4 << 3,
	[RTE_FLOW_FIELD_IPV4_DST] = 4 << 3,
	[RTE_FLOW_FIELD_IPV6_SRC] = 16 << 3,
	[RTE_FLOW_FIELD_IPV6_DST] = 16 << 3,
	[RTE_FLOW_FIELD_IPV4_TTL] = 1 << 3,
	[RTE_FLOW_FIELD_UDP_PORT_SRC] = 2 << 3,
	[RTE_FLOW_FIELD_UDP_PORT_DST] = 2 << 3,
	[RTE_FLOW_FIELD_TCP_PORT_SRC] = 2 << 3,
	[RTE_FLOW_FIELD_TCP_PORT_DST] = 2 << 3,
	[RTE_FLOW_FIELD_VXLAN_VNI] = 3 << 3,
	[RTE_FLOW_FIELD_GTP_TEID] = 4 << 3,
	[RTE_FLOW_FIELD_GENEVE_VNI] = 4 << 3,
};

uint32_t action_field_offset[] = {
	[RTE_FLOW_FIELD_META] = offsetof(struct doca_flow_actions, meta.pkt_meta),
	[RTE_FLOW_FIELD_TAG] = offsetof(struct doca_flow_actions, meta.u32),
	[RTE_FLOW_FIELD_MAC_DST] = offsetof(struct doca_flow_actions, mod_dst_mac),
	[RTE_FLOW_FIELD_MAC_SRC] = offsetof(struct doca_flow_actions, mod_src_mac),
	[RTE_FLOW_FIELD_IPV4_SRC] = offsetof(struct doca_flow_actions, mod_src_ip.ipv4_addr),
	[RTE_FLOW_FIELD_IPV4_DST] = offsetof(struct doca_flow_actions, mod_dst_ip.ipv4_addr),
	[RTE_FLOW_FIELD_IPV6_SRC] = offsetof(struct doca_flow_actions, mod_src_ip.ipv4_addr),
	[RTE_FLOW_FIELD_IPV6_DST] = offsetof(struct doca_flow_actions, mod_dst_ip.ipv4_addr),
	[RTE_FLOW_FIELD_IPV4_TTL] = offsetof(struct doca_flow_actions, ttl),
	[RTE_FLOW_FIELD_UDP_PORT_SRC] = offsetof(struct doca_flow_actions, mod_src_port),
	[RTE_FLOW_FIELD_UDP_PORT_DST] = offsetof(struct doca_flow_actions, mod_dst_port),
	[RTE_FLOW_FIELD_TCP_PORT_SRC] = offsetof(struct doca_flow_actions, mod_src_port),
	[RTE_FLOW_FIELD_TCP_PORT_DST] = offsetof(struct doca_flow_actions, mod_dst_port),
};

static struct {
	uint32_t start;
	uint32_t width;
	enum rte_flow_field_id header;
	int level;
} match_field_map[] = {
		match_member(meta.pkt_meta, RTE_FLOW_FIELD_META, false),
		match_member(meta.u32[0], RTE_FLOW_FIELD_TAG, 0),
		match_member(meta.u32[1], RTE_FLOW_FIELD_TAG, 1),
		match_member(meta.u32[2], RTE_FLOW_FIELD_TAG, 2),
		match_member(meta.u32[3], RTE_FLOW_FIELD_TAG, 3),
		match_member(out_src_mac, RTE_FLOW_FIELD_MAC_SRC, false),
		match_member(out_dst_mac, RTE_FLOW_FIELD_MAC_DST, false),
		match_member(out_eth_type, RTE_FLOW_FIELD_MAC_TYPE, false),
		match_member(out_vlan_tci, RTE_FLOW_FIELD_VLAN_ID, false),
		match_member(out_src_ip, RTE_FLOW_FIELD_IPV4_SRC, false),
		match_member(out_dst_ip, RTE_FLOW_FIELD_IPV4_DST, false),
		match_member(out_src_port, RTE_FLOW_FIELD_UDP_PORT_SRC, false),
		match_member(out_dst_port, RTE_FLOW_FIELD_UDP_PORT_DST, false),
		match_member(tun, RTE_FLOW_FIELD_VXLAN_VNI, false),
		match_member(in_src_ip, RTE_FLOW_FIELD_IPV4_SRC, true),
		match_member(in_dst_ip, RTE_FLOW_FIELD_IPV4_DST, true),
		match_member(in_vlan_tci, RTE_FLOW_FIELD_VLAN_ID, true),
		match_member(in_src_port, RTE_FLOW_FIELD_UDP_PORT_SRC, true),
		match_member(in_dst_port, RTE_FLOW_FIELD_UDP_PORT_DST, true),
		match_member(tun.vxlan_tun_id, RTE_FLOW_FIELD_VXLAN_VNI, false),
};

static_assert((sizeof(struct doca_flow_meta) % 4) == 0, "Meta data size must align to 4 bytes.");
static_assert(sizeof(struct doca_flow_meta) <= DOCA_FLOW_META_MAX + DOCA_FLOW_META_EXT, "Meta data max size is 20 bytes.");

/**
 * @brief modification field description
 */
struct action_field {
	uint32_t offset; /**< Bit offset from field start. */
	enum rte_flow_field_id type; /**< Field type to copy, auto-detect if address specified. */
	int level; /**< TAG index. */
	bool user_meta; /**< built-in TAG or user requested TAG */
};

/* Get field type and bit offset from matcher field address. */
static int
address_to_field(struct doca_flow_match *match, void *address, struct action_field *field)
{
	uint32_t i;
	uint32_t width;
	uintptr_t offset;

	field->type = RTE_FLOW_FIELD_START;
	field->offset &= ~DPDK_META_INTERNAL;
	offset = (uintptr_t)address - (uintptr_t)(void *)match + (field->offset >> 3);
	field->offset = field->offset % 8;

	for (i = 0; i < RTE_DIM(match_field_map); ++i) {
		width = match_field_map[i].width;
		if (offset < match_field_map[i].start || offset >= match_field_map[i].start + width)
			continue;
		field->type = match_field_map[i].header;
		field->level = match_field_map[i].level;
		field->offset += (offset - match_field_map[i].start) << 3;
		return 0;
	}
	return -EINVAL;
}

static int
dpdk_meta_item_modify(struct dpdk_item_entry *entry, const struct doca_flow_match *match)
{
	struct rte_flow_item_meta *spec = &entry->item_data.meta.spec;

	spec->data = match->meta.pkt_meta;
	return 0;
}

static int
dpdk_tag_item_modify(struct dpdk_item_entry *entry, const struct doca_flow_match *match)
{
	struct rte_flow_item_tag *spec = &entry->item_data.tag.spec;

	if (!(match->flags & DPDK_META_INTERNAL))
		spec->data = match->meta.u32[spec->index - REG_INDEX(dpdk_internal_meta_user_tag_index())];
	else
		spec->data = match->meta.u32[spec->index];
	return 0;
}

void
dpdk_meta_flow_item_build(struct dpdk_pipe *pipe,
			  struct dpdk_pipe_q_ctx *pipe_q,
			  const struct doca_flow_match *mask)
{
	struct dpdk_item_entry *entry;
	struct rte_flow_item *flow_item;
	size_t i;
	uint32_t *meta;
	int user_tag = 0;

	if (!mask)
		return;
	if (!(mask->flags & DPDK_META_INTERNAL))
		user_tag = REG_COUNT(0, dpdk_internal_meta_user_tag_index());
	meta = (void *)&mask->meta;
	for (i = 0; i < MAX_META_REG_COUNT; ++i) {
		if (meta[i] == 0)
			continue;
		entry = &pipe_q->item_entry[pipe->nb_items++];
		flow_item = entry->item;
		if (i < PKT_META_COUNT) {
			flow_item->type = RTE_FLOW_ITEM_TYPE_META;
			entry->item_data.meta.mask.data = meta[i];
			entry->modify_item = dpdk_meta_item_modify;
			flow_item->spec = &entry->item_data.meta.spec;
			flow_item->mask = &entry->item_data.meta.mask;
		} else {
			flow_item->type = RTE_FLOW_ITEM_TYPE_TAG;
			flow_item->spec = &entry->item_data.tag.spec;
			flow_item->mask = &entry->item_data.tag.mask;
			entry->item_data.tag.spec.index = i - PKT_META_COUNT + user_tag;
			entry->item_data.tag.mask.index = DPDK_TAG_INDEX_MASK;
			entry->item_data.tag.mask.data = meta[i];
			entry->modify_item = dpdk_tag_item_modify;
		}
	}
}

static int
dpdk_meta_cb_modify(__rte_unused struct dpdk_pipe *pipe, struct dpdk_action_entry *entry,
		    const struct doca_flow_actions *pkt_action,
		    __rte_unused const struct doca_flow_monitor *mon,
		    __rte_unused const struct doca_flow_fwd *fwd,
		    __rte_unused struct doca_flow_pipe_entry *flow_entry)
{
	struct rte_flow_action_modify_field *modi = &entry->action_data.modify_field.conf;
	uint32_t *data  = (void *)&pkt_action->meta;
	uint32_t val;

	if (modi->dst.field != RTE_FLOW_FIELD_TAG && modi->dst.field != RTE_FLOW_FIELD_META)
		return -EINVAL;
	if (modi->dst.field == RTE_FLOW_FIELD_TAG && modi->dst.level >= MAX_META_REG_COUNT - PKT_META_COUNT)
		return -EINVAL;

	if (modi->dst.field == RTE_FLOW_FIELD_META)
		val = data[0];
	else if (pkt_action->flags & DPDK_META_INTERNAL)
		val = data[modi->dst.level + PKT_META_COUNT];
	else
		val = data[modi->dst.level + PKT_META_COUNT - REG_INDEX(dpdk_internal_meta_user_tag_index())];
	*(uint32_t *)modi->src.value = val;
	return 0;
}

static int
dpdk_field_cb_modify(__rte_unused struct dpdk_pipe *pipe, struct dpdk_action_entry *entry,
		     const struct doca_flow_actions *act,
		     __rte_unused const struct doca_flow_monitor *mon,
		     __rte_unused const struct doca_flow_fwd *fwd,
		     __rte_unused struct doca_flow_pipe_entry *flow_entry)
{
	struct dpdk_action_modify_field_data *modi = &entry->action_data.modify_field;
	void *data = (void *)modi->conf.src.value;
	void *val = (void *)((uintptr_t)act + modi->offset);

	if (!modi->auto_val || utils_field_get_type(val, modi->width) != UTILS_FIELD_TYPE_IGNORED)
		memcpy(data, val, modi->width);
	return 0;
}

static int
dpdk_action_modify_field_build(struct dpdk_pipe *pipe,
		struct dpdk_pipe_q_ctx *pipe_q,
		uint32_t offset,
		uint32_t width,
		struct rte_flow_action_modify_field *conf,
		pipe_entry_action_cb cb,
		const struct doca_flow_actions *actions, uint8_t act_arr_idx,
		bool auto_val)
{
	struct dpdk_action_entry *entry;

	if (pipe->nb_actions_pipe[act_arr_idx] >= DPDK_PIPE_MAX_ACTIONS) {
		DOCA_LOG_ERR("failed to modify action field - too many modify actions %u",
			     pipe->nb_actions_pipe[act_arr_idx]);
		return -ENOTSUP;
	}

	entry = dpdk_pipe_actions_entry_get_next(pipe, pipe_q, act_arr_idx);
	entry->action->type = RTE_FLOW_ACTION_TYPE_MODIFY_FIELD;
	entry->action->conf = &entry->action_data.modify_field.conf;
	entry->action_data.modify_field.offset = offset;
	entry->action_data.modify_field.width = width;
	entry->action_data.modify_field.auto_val = auto_val;
	entry->action_data.modify_field.conf = *conf;
	if (actions && cb) /* Constant modification */
		cb(pipe, entry, actions, NULL, NULL, NULL);
	if (cb && (auto_val || !actions))
		entry->modify_action = cb;
	return 0;
}

static bool
dpdk_field_is_big_endian(enum rte_flow_field_id field)
{
	return field != RTE_FLOW_FIELD_TAG && field != RTE_FLOW_FIELD_META && field != RTE_FLOW_FIELD_MARK;
}

/* Apply mask and width into destination. */
static int
dpdk_action_modify_field_mask_build(struct dpdk_pipe *pipe,
		struct dpdk_pipe_q_ctx *pipe_q,
		uint32_t offset,
		enum rte_flow_field_id field,
		int level,
		void *mask,
		pipe_entry_action_cb cb,
		const struct doca_flow_actions *actions,
		uint8_t act_arr_idx)
{
	uint32_t bit_width = header_bit_width[field];
	uint32_t byte_width = BIT2BYTE(bit_width);
	struct rte_flow_action_modify_field modi = {
		.operation = RTE_FLOW_MODIFY_SET,
		.src = { .field = RTE_FLOW_FIELD_VALUE },
		.dst = { .field = field, .level = level },
	};
	int ret;
	uint32_t bit_idx;
	int mask_start_bit = -1;
	bool big_endian = dpdk_field_is_big_endian(field);

	for (bit_idx = 0; bit_idx < bit_width; ++bit_idx) {
		if (big_endian ? BIT_IS_SET_BE(mask, bit_idx) : BIT_IS_SET(mask, bit_idx)) {
			if (BIT_INDEX_INVALID(mask_start_bit))
				mask_start_bit = bit_idx;
			continue;
		}
		if (BIT_INDEX_INVALID(mask_start_bit))
			continue;
		/* End of set bits. */
		modi.dst.offset = mask_start_bit;
		modi.width = bit_idx - mask_start_bit;
		ret = dpdk_action_modify_field_build(pipe, pipe_q, offset, byte_width, &modi, cb, actions, act_arr_idx,
						     false);
		if (ret)
			return ret;
		mask_start_bit = -1;
	}
	if (!BIT_INDEX_INVALID(mask_start_bit)) {
		modi.dst.offset = mask_start_bit;
		modi.width = bit_idx - mask_start_bit;
		return dpdk_action_modify_field_build(pipe, pipe_q, offset, byte_width, &modi, cb, actions, act_arr_idx,
						      false);
	}
	return 0;
}

int
dpdk_action_modify_build(struct dpdk_pipe *pipe,
		struct dpdk_pipe_q_ctx *pipe_q,
		const struct doca_flow_actions *actions,
		enum rte_flow_field_id field,
		int level,
		void *action_val,
		struct doca_flow_action_desc *desc,
		uint8_t act_arr_idx)
{
	static uint64_t default_field_mask[2] = {UINT64_MAX, UINT64_MAX};
	bool const_modi = desc && desc->type == DOCA_FLOW_ACTION_CONSTANT;
	void *mask = desc && (desc->type == DOCA_FLOW_ACTION_CONSTANT || desc->type == DOCA_FLOW_ACTION_SET) ? desc->mask.u8 : action_val;
	uint32_t width = BIT2BYTE(header_bit_width[field]);
	uintptr_t offset = (uintptr_t)action_val - (uintptr_t)(void *)actions;
	enum utils_field_type field_type;

	if (desc && desc->type == DOCA_FLOW_ACTION_ADD)
		return 0;

	assert(width);
	assert(offset < sizeof(*actions));

	field_type = utils_field_get_type(mask, width);
	switch (field_type) {
	case UTILS_FIELD_TYPE_IGNORED:
		return 0;
	case UTILS_FIELD_TYPE_SPECIFIC:
		if ((!desc || desc->type == DOCA_FLOW_ACTION_AUTO) && field != RTE_FLOW_FIELD_META && field != RTE_FLOW_FIELD_TAG) {
			/* Meta doesn't support AUTO modificaiton due to user defined data struct */
			const_modi = true;
			mask = default_field_mask;
		}
		break;
	case UTILS_FIELD_TYPE_CHANGEABLE:
		break;
	default:
		return -EINVAL;
	}
	if (field == RTE_FLOW_FIELD_TAG && (actions->flags & DPDK_META_INTERNAL) == 0) {
		level += REG_INDEX(dpdk_internal_meta_user_tag_index());
		if (level >= RTE_DIM(actions->meta.u32))
			return -EINVAL;
	}
	return dpdk_action_modify_field_mask_build(pipe, pipe_q, offset, field, level, mask,
			dpdk_field_cb_modify, const_modi ? actions : NULL, act_arr_idx);
}

int
dpdk_action_set_meta_build(struct dpdk_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q,
		const struct doca_flow_actions *actions, struct doca_flow_action_desc *desc,
		uint8_t act_arr_idx)
{
	uint32_t mask;
	bool const_match = desc && desc->type == DOCA_FLOW_ACTION_CONSTANT;
	size_t i;
	int ret;
	int user_tag = 0;

	if (desc && desc->type > DOCA_FLOW_ACTION_SET)
		return 0;

	if (!(actions->flags & DPDK_META_INTERNAL))
		/* skip internal meta area */
		user_tag = REG_COUNT(0, dpdk_internal_meta_user_tag_index());
	if (!const_match)
		actions = NULL;
	for (i = 0; i < MAX_META_REG_COUNT; ++i) {
		if (desc && desc->type != DOCA_FLOW_ACTION_AUTO)
			mask = desc->mask.u32;
		else
			mask = i < PKT_META_COUNT ? actions->meta.pkt_meta : actions->meta.u32[i - 1];
		if (mask == 0)
			continue;
		if (i < PKT_META_COUNT)
			ret = dpdk_action_modify_field_mask_build(pipe, pipe_q, 0, RTE_FLOW_FIELD_META, 0,
					&mask, dpdk_meta_cb_modify, actions, act_arr_idx);
		else
			ret = dpdk_action_modify_field_mask_build(pipe, pipe_q, 0, RTE_FLOW_FIELD_TAG,
					user_tag + i - PKT_META_COUNT, &mask, dpdk_meta_cb_modify,
					actions, act_arr_idx);
		if (ret)
			return ret;
	}
	return 0;
}

static int
dpdk_modify_field_data_convert(struct dpdk_pipe_q_ctx *pipe_q,
		struct action_field *field,
		uint32_t field_bit_width,
		struct rte_flow_action_modify_data *modi)
{
	uint32_t field_bit_len = header_bit_width[field->type];

	if (field_bit_len == 0) {
		DOCA_LOG_ERR("failed to modify data field - unsupported header type %u", field->type);
		return -1;
	}
	if (field->offset + field_bit_width > field_bit_len) {
		DOCA_LOG_ERR("failed to modify data field - header type %u oversize: bit offset %u, field bit width %u",
			     field->type, field->offset, field_bit_width);
		return -1;
	}
	switch (field->type) { /* Handle IPv6 and TCP port. */
	case RTE_FLOW_FIELD_IPV4_SRC:
		modi->field = pipe_q->ip_type == DOCA_FLOW_IP6_ADDR ? RTE_FLOW_FIELD_IPV6_SRC : RTE_FLOW_FIELD_IPV4_SRC;
		break;
	case RTE_FLOW_FIELD_IPV4_DST:
		modi->field = pipe_q->ip_type == DOCA_FLOW_IP6_ADDR ? RTE_FLOW_FIELD_IPV6_DST : RTE_FLOW_FIELD_IPV4_DST;
		break;
	case RTE_FLOW_FIELD_UDP_PORT_SRC:
		modi->field = pipe_q->l4_type == DOCA_PROTO_TCP ? RTE_FLOW_FIELD_TCP_PORT_SRC : RTE_FLOW_FIELD_UDP_PORT_SRC;
		break;
	case RTE_FLOW_FIELD_UDP_PORT_DST:
		modi->field = pipe_q->l4_type == DOCA_PROTO_TCP ? RTE_FLOW_FIELD_TCP_PORT_DST : RTE_FLOW_FIELD_UDP_PORT_DST;
		break;
	default:
		modi->field = field->type;
		break;
	}

	modi->level = field->level;
	if (field->type == RTE_FLOW_FIELD_TAG || field->type == RTE_FLOW_FIELD_META)
		modi->offset = field->offset;
	else
		modi->offset = field_bit_len - (field_bit_width + field->offset); /* Big endian */
	return 0;
}

static int
dpdk_action_copy_build(struct dpdk_pipe *pipe,
		struct dpdk_pipe_q_ctx *pipe_q,
		struct action_field *src_field,
		struct action_field *dst_field,
		uint32_t bit_width,
		uint8_t act_arr_idx)
{
	struct rte_flow_action_modify_field modi = { .operation = RTE_FLOW_MODIFY_SET, .width = bit_width };
	int ret;

	ret = dpdk_modify_field_data_convert(pipe_q, src_field, bit_width, &modi.src);
	if (ret)
		return ret;
	ret = dpdk_modify_field_data_convert(pipe_q, dst_field, bit_width, &modi.dst);
	if (ret)
		return ret;
	return dpdk_action_modify_field_build(pipe, pipe_q, 0, 0, &modi, NULL, NULL, act_arr_idx, false);
}

static int
dpdk_action_desc_add_build(struct dpdk_pipe *pipe,
		struct dpdk_pipe_q_ctx *pipe_q,
		struct action_field *field,
		uint32_t bit_width,
		const struct doca_flow_actions *actions,
		uint8_t act_arr_idx)
{
	struct rte_flow_action_modify_field modi = {.operation = RTE_FLOW_MODIFY_ADD, .src = { .field = RTE_FLOW_FIELD_VALUE }};
	uint32_t bit_len, offset;
	int ret;

	bit_len = bit_width ? bit_width : header_bit_width[field->type];
	offset = action_field_offset[field->type];
	if (bit_len == 0 || offset == 0) {
		DOCA_LOG_ERR("failed to build add action description - field %d of action ADD not supported",
			     field->type);
		return -EINVAL;
	}
	ret = dpdk_modify_field_data_convert(pipe_q, field, bit_len, &modi.dst);
	if (ret)
		return ret;
	modi.width = bit_len;
	if (field->type == RTE_FLOW_FIELD_TAG)
		offset += field->level * 4;
	if (field->user_meta) /* Restore user offset to copy from actions */
		offset -= REG_INDEX(dpdk_internal_meta_user_tag_index()) * 4;

	return dpdk_action_modify_field_build(pipe, pipe_q, offset, BIT2BYTE(bit_len), &modi, dpdk_field_cb_modify,
					      actions, act_arr_idx, true);
}

static int
action_field_init(struct action_field *out,
		struct doca_flow_action_field *field,
		struct doca_flow_match *match,
		enum rte_flow_field_id header,
		int level)
{
	if (field) {
		out->offset = field->offset;
		if (field->address && address_to_field(match, field->address, out))
			goto error;
	}

	if (!out->type) {
		out->type = header;
		out->level = level;
	}

	if (field && out->type != RTE_FLOW_FIELD_TAG && field->offset & DPDK_META_INTERNAL) {
		DOCA_LOG_ERR("unexpected internal meta flag");
		return -EINVAL;
	}

	if (out->type == RTE_FLOW_FIELD_VXLAN_VNI && match->tun.type == DOCA_FLOW_TUN_GTPU)
		out->type = RTE_FLOW_FIELD_GTP_TEID;
	else if (out->type == RTE_FLOW_FIELD_TAG) {
		if (field && field->offset & DPDK_META_INTERNAL)
			out->offset &= ~DPDK_META_INTERNAL;
		else {
			out->user_meta = true;
			out->level += REG_INDEX(dpdk_internal_meta_user_tag_index());
			if (out->level >= RTE_DIM(match->meta.u32))
				return -EINVAL;
		}
	}

	if (out->type != RTE_FLOW_FIELD_START && out->type <= RTE_FLOW_FIELD_VALUE)
		return 0;
error:
	if (field)
		DOCA_LOG_ERR("failed to initialize action field - invalid field info: %p bit offset %u.",
			     field->address, field->offset);
	else
		DOCA_LOG_ERR("failed to initialize action field - invalid field type.");
	return -EINVAL;
}

static int
dpdk_action_desc_build(struct dpdk_pipe *pipe,
		struct dpdk_pipe_q_ctx *pipe_q,
		const struct doca_flow_pipe_cfg *cfg,
		enum rte_flow_field_id header,
		int inner,
		struct doca_flow_action_desc *desc,
		uint8_t act_arr_idx)
{
	struct action_field src = { 0 };
	struct action_field dst = { 0 };

	switch (desc->type) {
	case DOCA_FLOW_ACTION_AUTO:
	case DOCA_FLOW_ACTION_CONSTANT:
	case DOCA_FLOW_ACTION_SET:
		return 0;
	case DOCA_FLOW_ACTION_ADD:
		if (desc->add.dst.offset & ~DPDK_META_INTERNAL) {
			DOCA_LOG_ERR("ADD offset not supported");
			return -EINVAL;
		}
		if (action_field_init(&dst, &desc->add.dst, cfg->match, header, inner))
			return -EINVAL;
		return dpdk_action_desc_add_build(pipe, pipe_q, &dst, desc->add.width, cfg->actions[act_arr_idx], act_arr_idx);
	case DOCA_FLOW_ACTION_COPY:
		if (action_field_init(&src, &desc->copy.src, cfg->match, header, inner) ||
		    action_field_init(&dst, &desc->copy.dst, cfg->match, header, inner))
			return -EINVAL;
		if (src.type == RTE_FLOW_FIELD_START || dst.type == RTE_FLOW_FIELD_START) {
			DOCA_LOG_ERR("failed to build action description - invalid copy field");
			return -EINVAL;
		}
		return dpdk_action_copy_build(pipe, pipe_q, &src, &dst, desc->copy.width, act_arr_idx);
	default:
		break;
	}
	return -EINVAL;
}

int
dpdk_action_descs_build(struct dpdk_pipe *pipe,
		struct dpdk_pipe_q_ctx *pipe_q,
		uint8_t act_arr_idx,
		const struct doca_flow_pipe_cfg *cfg)
{
	struct doca_flow_action_descs *descs;
	int ret;
	int i;

	descs = cfg->action_descs[act_arr_idx];
	if (!descs)
		return 0;

	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_META, false, &descs->meta.pkt_meta, act_arr_idx));
	for (i = 0; i < RTE_DIM(descs->meta.u32); i++)
		CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_TAG, i, &descs->meta.u32[i], act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_MAC_SRC, false, &descs->src_mac, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_MAC_DST, false, &descs->dst_mac, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_MAC_TYPE, false, &descs->eth_type, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_VLAN_ID, false, &descs->vlan_id, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_IPV4_SRC, false, &descs->src_ip, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_IPV4_DST, false, &descs->dst_ip, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_IPV4_TTL, false, &descs->ttl, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_UDP_PORT_SRC, false, &descs->src_port, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_UDP_PORT_DST, false, &descs->dst_port, act_arr_idx));
	CHK(dpdk_action_desc_build(pipe, pipe_q, cfg, RTE_FLOW_FIELD_VXLAN_VNI, false, &descs->tunnel, act_arr_idx));

	return 0;
}

int
dpdk_user_meta_tag_resolve(struct doca_flow_match *match, void *addr)
{
	struct doca_flow_action_field field = {.address = addr};
	struct action_field out = {};
	int ret;

	ret = action_field_init(&out, &field, match, RTE_FLOW_FIELD_START, false);
	if (ret)
		return ret;
	if (out.type != RTE_FLOW_FIELD_TAG)
		return -EINVAL;
	return (out.level * sizeof(match->meta.u32[0])) << 3;
}
