/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <doca_log.h>
#include <dpdk_table.h>
#include <dpdk_port.h>
#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "dpdk_dbg.h"
#include "dpdk_flow_hws_legacy.h"
#include "dpdk_utilities.h"
#include "engine_shared_resources.h"
#include "doca_flow_layer.h"
#include "dpdk_table_manager.h"

DOCA_LOG_REGISTER(dpdk_flow_hws_legacy);

#define DEQUE_SPEC_ENTRY_SUCCESS (0xfffffee)
#define MAX_ENCAP_SIZE (1240)
#define DPDK_NB_POLICIES_MINIMAL 1
#define ACT_TEMPLATE_BUF_SIZE 64
#define ENCAP_MAX_ACT_TEMPALTE 1

struct flow_queue_ctx *flow_queue_ctx_ins;
uint64_t ms_hz;
struct dpdk_pipe default_ctrl_pipe;
struct dpdk_hws_cfg dpdk_hws_cfg = {0};

struct flow_queue_ctx {
	uint16_t nb_queues;
	struct flow_queue_ctx_array *queue_ctx_array[0];
};

void
dpdk_flow_hws_destroy()
{
	uint32_t qidx = 0;
	struct flow_queue_ctx_array *array;

	if (flow_queue_ctx_ins == NULL)
		return;
	for (qidx = 0; qidx < flow_queue_ctx_ins->nb_queues; qidx++) {
		array = flow_queue_ctx_ins->queue_ctx_array[qidx];
		if (array == NULL)
			continue;
		free(array);
		flow_queue_ctx_ins->queue_ctx_array[qidx] = NULL;
	}
	free(flow_queue_ctx_ins);
	flow_queue_ctx_ins = NULL;
}

static uint32_t
get_nr_profiles_per_port()
{
	return dpdk_hws_cfg.resource.nb_meters ? dpdk_hws_cfg.resource.nb_meters :
						 DPDK_NB_POLICIES_MINIMAL;
}

static void
dpdk_flow_port_attr_configure(struct rte_flow_port_attr *port_attr)
{
	uint32_t nb_counters = dpdk_hws_cfg.resource.nb_counters +
		dpdk_hws_cfg.nr_shared_resources[DOCA_FLOW_SHARED_RESOURCE_COUNT];

	uint32_t nb_meters = dpdk_hws_cfg.resource.nb_meters +
		dpdk_hws_cfg.nr_shared_resources[DOCA_FLOW_SHARED_RESOURCE_METER];

	uint32_t nb_profiles = DPDK_MAX_PORT_PROFILES +
		dpdk_hws_cfg.nr_shared_resources[DOCA_FLOW_SHARED_RESOURCE_METER];

	uint32_t nb_policies = get_nr_profiles_per_port();

	*port_attr = (struct rte_flow_port_attr){
		.nb_counters = nb_counters,
		.nb_meters = nb_meters,
		.nb_meter_profiles = nb_profiles,
		.nb_meter_policies = nb_policies,
	};
}

int
dpdk_flow_hws_port_configure(struct doca_flow_port *port)
{
	struct rte_flow_queue_attr queue_attr;
	int total_queues = dpdk_hws_cfg.queues;
	const struct rte_flow_queue_attr *queue_attrs[total_queues];
	struct rte_flow_error err = {0};
	uint16_t rx_peers[RTE_MAX_ETHPORTS], tx_peers[RTE_MAX_ETHPORTS];
	int i, ret, nb_rx_peers, nb_tx_peers;
	struct rte_flow_port_attr port_attr;

	/* unbind current Tx from all peer Rx */
	nb_rx_peers = rte_eth_hairpin_get_peer_ports(port->port_id, rx_peers, RTE_MAX_ETHPORTS, 1);
	for (i = 0; i < nb_rx_peers; i++) {
		ret = rte_eth_hairpin_unbind(port->port_id, rx_peers[i]);
		if (ret < 0) {
			DOCA_LOG_ERR("failed to configure flow hws port - hairpin unbind port tx_port(%u) to rx_port(%u), ret=%d.",
				     port->port_id, rx_peers[i], ret);
			goto failure;
		}
	}
	/* unbind all peer Tx from current Rx */
	nb_tx_peers = rte_eth_hairpin_get_peer_ports(port->port_id, tx_peers, RTE_MAX_ETHPORTS, 0);
	for (i = 0; i < nb_tx_peers; i++) {
		ret = rte_eth_hairpin_unbind(tx_peers[i], port->port_id);
		if (ret < 0) {
			DOCA_LOG_ERR("failed to configure flow hws port - hairpin unbind port tx_port(%u) to rx_port(%u), ret=%d.",
				     tx_peers[i], port->port_id, ret);
			goto failure;
		}
	}

	ret = rte_eth_dev_stop(port->port_id);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to configure flow hws port %u - stop fail, ret=%d.", port->port_id, ret);
		goto failure;
	}

	queue_attr.size = dpdk_hws_cfg.queue_depth;
	for (i = 0; i < total_queues; ++i)
		queue_attrs[i] = &queue_attr;

	dpdk_flow_port_attr_configure(&port_attr);

	DOCA_DLOG_INFO("Port configure: portid %d que_depth %d queues %d meters %d counters %d meter_profiles %d.",
			port->port_id, queue_attr.size, total_queues, dpdk_hws_cfg.resource.nb_meters,
			dpdk_hws_cfg.resource.nb_counters, port_attr.nb_meter_profiles);
	ret = rte_flow_configure(port->port_id, &port_attr, total_queues, queue_attrs, &err);
	if (ret) {
		DOCA_LOG_ERR("failed to configure flow hws port %u - rte flow configure, type %d message: %s",
			     port->port_id, err.type,
			     err.message ? err.message : "(no stated reason)");
		goto failure;
	}

	ret = rte_eth_dev_start(port->port_id);
	if (ret < 0) {
		DOCA_LOG_ERR("Port %u start fail, ret %d.", port->port_id, ret);
		goto failure;
	}
	/* bind current Tx to all peer Rx */
	for (i = 0; i < nb_rx_peers; i++) {
		ret = rte_eth_hairpin_bind(port->port_id, rx_peers[i]);
		if (ret < 0) {
			DOCA_LOG_ERR("failed to configure flow hws port - hairpin bind port tx_port(%u) to rx_port(%u), ret=%d.",
				     port->port_id, rx_peers[i], ret);
			goto failure;
		}
	}
	/* bind all peer Tx to current Rx */
	for (i = 0; i < nb_tx_peers; i++) {
		ret = rte_eth_hairpin_bind(tx_peers[i], port->port_id);
		if (ret < 0) {
			DOCA_LOG_ERR("failed to configure flow hws port - hairpin bind port tx_port(%u) to rx_port(%u), ret=%d.",
				     tx_peers[i], port->port_id, ret);
			goto failure;
		}
	}
	return 0;
failure:
	return ret;
}

int
dpdk_flow_hws_init(const struct dpdk_hws_cfg *cfg)
{
	uint16_t queues = cfg->queues;
	struct flow_queue_ctx_array *array;
	uint32_t size;
	int i, j;

	dpdk_hws_cfg = *cfg;
	size = sizeof(struct flow_queue_ctx) +
		sizeof(struct flow_queue_ctx_array *) * queues;
	flow_queue_ctx_ins = calloc(1, size);
	if (!flow_queue_ctx_ins) {
		DOCA_LOG_ERR("alloc memory for usr_ctx failed.");
		return -1;
	}

	size = sizeof(struct flow_queue_ctx_array) +
			cfg->queue_depth * sizeof(struct flow_queue_ctx_item);
	for (i = 0; i < queues; i++) {
		array = calloc(1, size);
		if (array == NULL) {
			DOCA_LOG_ERR("alloc usr ctx len:%d on queue:%d error.",
					cfg->queue_depth, i);
			goto release;
		}

		array->array_len = cfg->queue_depth;
		LIST_INIT(&array->free_list);
		for (j = 0; j < array->array_len; j++)
			LIST_INSERT_HEAD(&array->free_list, &array->items[j], next);
		flow_queue_ctx_ins->queue_ctx_array[i] = array;
		flow_queue_ctx_ins->nb_queues++;
	}

	memset(&default_ctrl_pipe, 0, sizeof(default_ctrl_pipe));
	LIST_INIT(&(default_ctrl_pipe.hws_ctx.hws_list));
	ms_hz = rte_get_timer_hz() / 1000000;  /*micro second*/;

	return 0;

release:
	dpdk_flow_hws_destroy();
	return -1;
}

static void
flow_usr_ctx_free(uint32_t queue_id, struct flow_queue_ctx_item *item)
{
	struct flow_queue_ctx_array *array;

	array = flow_queue_ctx_ins->queue_ctx_array[queue_id];
	if (item) {
		LIST_INSERT_HEAD(&array->free_list, item, next);
		item->in_use = false;
	}
}

static void *
flow_usr_ctx_add(uint16_t queue,
		struct doca_flow_pipe_entry *entry,
		void *usr_ctx, void *cb,
		enum doca_flow_entry_op op)
{
	struct flow_queue_ctx_array *array;
	struct flow_queue_ctx_item *item;

	array = flow_queue_ctx_ins->queue_ctx_array[queue];
	item = LIST_FIRST(&array->free_list);
	if (!item) {
		DOCA_DLOG_INFO("reach max usr_ctx len:%d", array->array_len);
		return NULL;
	}

	item->ctx = entry;
	item->usr_ctx = usr_ctx;
	item->in_use = true;
	item->cb = cb;
	item->op = op;
	LIST_REMOVE(item, next);

	return item;
}

static int
hws_queue_items_process_cb(uint32_t queue_id,
		int ret, struct rte_flow_op_result *comp,
		struct doca_flow_pipe_entry *entry)
{
	struct flow_queue_ctx_item *item;
	int i, result = 0;

	for (i = 0; i < ret; i++) {
		item = comp[i].user_data;
		if (item == NULL)
			continue;
		if (item->cb)
			item->cb(&comp[i]);
		if (entry && item->ctx == entry)
			result = DEQUE_SPEC_ENTRY_SUCCESS;
		flow_usr_ctx_free(queue_id, item);
	}
	if (entry)
		return result;
	return 0;
}

bool
dpdk_flow_hws_is_ctrl_entry(struct doca_flow_pipe_entry *entry)
{
	struct doca_flow_pipe *entry_pipe;

	if (entry->pipe_queue == dpdk_hws_cfg.ctrl_queue_id)
		return true;
	entry_pipe = entry->pipe;
	if (entry_pipe)
		return (entry_pipe->type == ENGINE_PIPE_CONTROL);
	return false;
}

struct dpdk_hws_item *
dpdk_flow_hws_get_pipe_first_ctx(struct dpdk_pipe *pipe)
{
	return LIST_FIRST(&pipe->hws_ctx.hws_list);
}

static __rte_always_inline int
port_queue_push(uint16_t port_id, uint16_t queue_id, bool force)
{
	int nb_comp;

	nb_comp = dpdk_flow_hws_dequeue(port_id, queue_id, PULL_TIMEOUT_US,
					dpdk_hws_cfg.queue_depth, force, NULL);
	DOCA_DLOG_DBG("Port %u queue %u processed entries nb_comp %d",
		      port_id, queue_id, nb_comp);
	return nb_comp;
}

/* redesign https://redmine.mellanox.com/issues/3165944 */
static inline int
port_queues_flush(uint16_t port_id)
{
	uint64_t end_tsc;
	bool force = true;
	int queue_id;
	int rc = 0;

	end_tsc = rte_rdtsc() + ms_hz * US_PER_S;
	do {
		for (queue_id = 0; queue_id < dpdk_hws_cfg.queues; queue_id++) {
			rc = port_queue_push(port_id, queue_id, force);
			if (rc < 0) {
				DOCA_LOG_ERR("%s failed - port %u queue %d deplete failed rc=%d",
					     __func__, port_id, queue_id, rc);
				return rc;
			}
			force = false;
		}
	} while (rte_rdtsc() <= end_tsc);

	return rc;
}

static int
hws_table_destroy(struct doca_flow_port *port,
		   struct rte_flow_template_table *table)
{
	struct dpdk_table_port_ctx *ctx;
	struct dpdk_table res;
	int rc;

	ctx = dpdk_port_get_table_ctx(port->dpdk_port);
	if (unlikely(ctx == NULL)) {
		DOCA_LOG_RATE_LIMIT_ERR("%s failed - port table ctx is NULL", __func__);
		return -ENOENT;
	}

	res.table = table;
	res.port_id = port->port_id;
	rc = dpdk_table_destroy(ctx, &res);
	if (!rc)
		return 0;

	/* if table destroy failed the assumption is that it's because of having
	 * dangling flows, attempt to deplete the queue and try again
	 */
	rc = port_queues_flush(port->port_id);
	if (rc < 0) {
		DOCA_LOG_ERR("failed to destroy hws table - port %u queues force deplete, rc=%d",
			     port->port_id, rc);
		return -ENOTRECOVERABLE;
	}

	rc = dpdk_table_destroy(ctx, &res);
	if (rc) {
		DOCA_LOG_ERR("failed to destroy hws table - port %u destroy dpdk hws table", port->port_id);
		return -ENOTRECOVERABLE;
	}
	return 0;
}

struct rte_flow *
dpdk_flow_hws_enqueue(uint16_t port_id,
			uint32_t queue_id,
			struct rte_flow_op_attr *qatr,
			struct rte_flow_template_table *table,
			const struct rte_flow_item items[],
			uint8_t item_template_index,
			const struct rte_flow_action actions[],
			uint8_t action_template_index,
			uint32_t flags, void *usr_ctx,
			void *cb, struct doca_flow_pipe_entry *entry)
{
	struct rte_flow_error flow_error;
	uint64_t end_tsc;
	struct rte_flow *flow;
	void *item;
	int ret;
	bool force_push = true;

	end_tsc = rte_rdtsc() + ms_hz * US_PER_S;
	do {
		item = flow_usr_ctx_add(queue_id, entry, usr_ctx, cb, DOCA_FLOW_ENTRY_OP_ADD);
		if (item)
			break;
		dpdk_flow_hws_dequeue(port_id, queue_id, PULL_TIMEOUT_US,
				dpdk_hws_cfg.queue_depth, force_push, NULL);
		force_push = false;
	} while (rte_rdtsc() <= end_tsc);
	if (!item)
		goto error_timeout;

	flow = rte_flow_async_create(port_id, queue_id,
				qatr, table,
				items, item_template_index,
				actions, action_template_index,
				item, &flow_error);
	if (!flow)
		goto error;

	/* handle persitent case */
	if (flags & HWS_FLOW_FLAG_PERSISTENT) {
		force_push = true;
		ret = dpdk_flow_hws_dequeue(port_id, queue_id, PULL_TIMEOUT_US,
				dpdk_hws_cfg.queue_depth, force_push, entry);
		if (entry && (ret != DEQUE_SPEC_ENTRY_SUCCESS))
			goto error_timeout;
	}
	return flow;
error:
	DOCA_LOG_ERR("failed to enqueue hws flow - port %u create rte flow, type %d message: %s", port_id,
		     flow_error.type, flow_error.message ? flow_error.message : "(no stated reason)");
	return NULL;
error_timeout:
	DOCA_LOG_ERR("failed to enqueue hws flow - port %u create rte flow reached timeout.", port_id);
	return NULL;
}

int
dpdk_flow_hws_mgmt_q_flow_destroy_legacy(uint16_t port_id,
					 uint32_t queue_id,
					 struct rte_flow_op_attr *q_ops_attr,
					 struct doca_flow_pipe_entry *entry,
					 void *cb, void *usr_ctx,
					 uint32_t flag)
{
	struct rte_flow_error flow_error;
	struct flow_queue_ctx_item *item;
	uint64_t end_tsc;
	int ret;
	bool force_push = true;

	end_tsc = rte_rdtsc() + ms_hz * US_PER_S;
	do {
		item = flow_usr_ctx_add(queue_id, entry, usr_ctx, cb, DOCA_FLOW_ENTRY_OP_DEL);
		if (item)
			break;
		dpdk_flow_hws_dequeue(port_id, queue_id, PULL_TIMEOUT_US,
				dpdk_hws_cfg.queue_depth, force_push, NULL);
		force_push = false;
	} while (rte_rdtsc() <= end_tsc);
	if (!item)
		goto error_timeout;

	ret = rte_flow_async_destroy(port_id, queue_id,
				     q_ops_attr, dpdk_flow_tracker_get_flow(&entry->tracker),
				     item, &flow_error);
	if (ret)
		goto error_flow;

	ret = dpdk_flow_hws_dequeue(port_id, queue_id, PULL_TIMEOUT_US,
				    dpdk_hws_cfg.queue_depth, force_push, entry);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to destroy hws flow - dequeue hws flow, port %u.", port_id);
		return -ENOTRECOVERABLE;
	}
	return 0;
error_flow:
	DOCA_LOG_ERR("failed to destroy hws flow - destroy rte flow, port %u.", port_id);
	return -1;
error_timeout:
	DOCA_LOG_ERR("failed to destroy hws flow - rte flow not destroyed until timeout, port %u.", port_id);
	return -1;
}

int
dpdk_flow_hws_dequeue(uint16_t port_id,
		       uint32_t queue_id,
		       uint64_t timeout_ms,
		       uint32_t max_processed_entries,
		       bool force_push,
		       struct doca_flow_pipe_entry *entry)
{
	struct rte_flow_error flow_error;
	uint64_t end_tsc;
	uint32_t nb_comp = 0;
	struct rte_flow_op_result comp[dpdk_hws_cfg.queue_depth];
	int ret, result;

	if (force_push) {
		ret = rte_flow_push(port_id, queue_id, &flow_error);
		if (ret)
			DOCA_LOG_RATE_LIMIT_ERR("entries_process: drain failed ret %d port %d",
					ret, port_id);
	}

	end_tsc = rte_rdtsc() + ms_hz * timeout_ms;
	do {
		ret = rte_flow_pull(port_id, queue_id, comp, dpdk_hws_cfg.queue_depth, &flow_error);
		if (ret < 0) {
			DOCA_LOG_RATE_LIMIT_ERR("Deque failed ret %d, type %d message: %s",
						ret, flow_error.type,
						flow_error.message ? flow_error.message : "(no stated reason)");
			break;
		}
		nb_comp += ret;
		if (ret) {
			result = hws_queue_items_process_cb(queue_id, ret, comp, entry);
			if (entry && result == DEQUE_SPEC_ENTRY_SUCCESS)
				return DEQUE_SPEC_ENTRY_SUCCESS;
		}
		if (max_processed_entries && (nb_comp >= max_processed_entries))
			break;
	} while (rte_rdtsc() <= end_tsc);

	return nb_comp;
}

static inline void
fill_dpdk_table_params(struct dpdk_table_params *table_params,
			uint16_t port_id, struct rte_flow_item *items,
			struct rte_flow_action **actions, struct rte_flow_action **actions_mask,
			uint8_t nb_actions, struct rte_flow_attr *attr, uint32_t nb_flows)
{

	table_params->port_id = port_id;
	table_params->items = items;
	table_params->actions = actions;
	table_params->actions_mask = actions_mask;
	table_params->nb_actions = nb_actions;
	table_params->flow_attr = attr;
	table_params->nb_flows = nb_flows;
	table_params->insertion_type = DPDK_TABLE_INSERTION_TYPE_PATTERN;
	table_params->hash_type = DPDK_TABLE_HASH_TYPE_DEFAULT;
}

static inline void
fill_act_template_name(char *buf, int i, bool mask)
{
	if (!mask)
		snprintf(buf, ACT_TEMPLATE_BUF_SIZE, "action template %d\t", i);
	else
		snprintf(buf, ACT_TEMPLATE_BUF_SIZE, "action template mask %d\t", i);
}

static int
get_hws_table(struct doca_flow_port *port,
	       struct rte_flow_item items[],
	       struct rte_flow_action **actions,
	       struct rte_flow_action **actions_masks,
	       uint8_t nb_actions,
	       struct rte_flow_template_table_attr *tatr,
	       struct dpdk_hws_item *hws_entry)
{
	char act_template_name[ACT_TEMPLATE_BUF_SIZE];
	struct dpdk_table_params table_params;
	struct dpdk_table_port_ctx *port_ctx;
	struct dpdk_table res;
	int i, rc;

	if (unlikely(hws_entry == NULL)) {
		DOCA_LOG_ERR("%s failed - hws_entry is NULL", __func__);
		return -EINVAL;
	}

	port_ctx = dpdk_port_get_table_ctx(port->dpdk_port);
	if (unlikely(port_ctx == NULL)) {
		DOCA_LOG_ERR("%s failed - port table ctx is NULL", __func__);
		return -ENOENT;
	}

	dpdk_dbg_rte_flow_dump("pattern template:\t", port->port_id,
			       &tatr->flow_attr, items, NULL, stdout);
	for (i = 0; i < nb_actions; i++) {
		fill_act_template_name(act_template_name, i, false);
		dpdk_dbg_rte_flow_dump(act_template_name, port->port_id,
				       &tatr->flow_attr, NULL, actions[i], stdout);

		fill_act_template_name(act_template_name, i, true);
		dpdk_dbg_rte_flow_dump(act_template_name, port->port_id,
				       &tatr->flow_attr, NULL, actions_masks[i], stdout);
	}

	fill_dpdk_table_params(&table_params, port->port_id, items, actions,
			       actions_masks, nb_actions, &tatr->flow_attr, tatr->nb_flows);

	rc = dpdk_table_create(&table_params, port_ctx, &res);
	if (rc) {
		DOCA_LOG_ERR("%s failed - failed to create dpdk table", __func__);
		return -ENOENT;
	}

	hws_entry->table = res.table;
	hws_entry->pattern_template_index = res.pattern_template_index;
	hws_entry->action_template_index = res.action_template_index;
	return 0;
}

static inline bool
should_build_action_masks(struct dpdk_pipe_q_ctx *pipe_q)
{
	return pipe_q != NULL;
}

struct dpdk_hws_item *
dpdk_flow_hws_ctrl_table_create(struct doca_flow_port *port,
				 struct dpdk_pipe *pipe,
				 struct dpdk_pipe_q_ctx *pipe_q,
				 struct rte_flow_template_table_attr *tatr,
				 const struct rte_flow_item items[],
				 struct rte_flow_action **actions,
				 uint8_t nb_actions)
{
	struct rte_flow_action_modify_field modis[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS] = {};
	struct rte_flow_action action_masks[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_action *actions_mask_arr[MAX_ACT_TEMPLATE];
	struct dpdk_hws_item *hws_entry;
	int rc, i;

	if (!pipe)
		pipe = &default_ctrl_pipe;

	hws_entry = calloc(1, sizeof(struct dpdk_hws_item));
	if (!hws_entry) {
		DOCA_LOG_RATE_LIMIT_ERR("%s failed on port %u - no memory", __func__, port->port_id);
		return NULL;
	}

	/* Since both user defined ctrl pipes and internal ctrl pipes used this function
	 * we might need to pass actions masks to DPDK for actions that need it,
	 * check if it's needed before creating a HW table
	 */
	if (should_build_action_masks(pipe_q)) {
		for (i = 0; i < nb_actions; i++) {
			dpdk_pipe_actions_masks_build(pipe_q, action_masks[i], modis[i], i);
			actions_mask_arr[i] = action_masks[i];
		}
	}
	rc = get_hws_table(port, (struct rte_flow_item *)items, actions, pipe_q ? actions_mask_arr : actions,
			   nb_actions, tatr, hws_entry);
	if (rc) {
		DOCA_LOG_RATE_LIMIT_ERR("%s failed - create hws table failed: port %d, rc=%d",
					__func__, port->port_id, rc);
		free(hws_entry);
		return NULL;
	}

	LIST_INSERT_HEAD(&pipe->hws_ctx.hws_list, hws_entry, next);
	return hws_entry;
}

struct rte_flow *
dpdk_flow_hws_ctrl_entry_create(struct doca_flow_port *port,
				struct dpdk_pipe *pipe,
				const struct rte_flow_item items[],
				const struct rte_flow_action actions[],
				void *usr_ctx, void *cb,
				struct doca_flow_pipe_entry *entry,
				struct dpdk_hws_item *hws_entry)
{
	uint32_t queue_id = dpdk_hws_cfg.ctrl_queue_id;
	struct rte_flow_op_attr qatr = {0};
	struct rte_flow *flow;
	uint16_t port_id;

	if (!pipe)
		pipe = &default_ctrl_pipe;

	port_id = port->port_id;

	flow = dpdk_flow_hws_enqueue(port_id, queue_id, &qatr, hws_entry->table,
				items, 0, actions, 0, HWS_FLOW_FLAG_PERSISTENT,
				usr_ctx, cb, entry);
	if (!flow) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create hws flow control entry - enqueue, port %d.", port_id);
		return NULL;
	}

	hws_entry->flow = flow;
	pipe->hws_ctx.nb_entries++;

	return flow;
}

int
dpdk_flow_hws_ctrl_table_flow_destroy(struct doca_flow_port *port,
				      struct dpdk_hws_item *hws_entry)
{
	int rc;

	rc = hws_table_destroy(port, hws_entry->table);
	if (rc) {
		DOCA_LOG_RATE_LIMIT_ERR("%s failed - destroy hws table, rc=%d", __func__, rc);
		return rc;
	}
	LIST_REMOVE(hws_entry, next);
	free(hws_entry);
	return 0;
}

/* the following code is needed to keep the legacy encap table implementation as is */
static void *
dpdk_hws_alloc_encap_hws_ctx()
{
	return calloc(1, sizeof(struct dpdk_hws_item));
}

static void
dpdk_hws_free_encap_hws_ctx(void *encap_hws_ctx)
{
	free(encap_hws_ctx);
}

static int
dpdk_hws_table_destroy_legacy(uint16_t port_id,
			      struct rte_flow_template_table *table)
{
	struct rte_flow_error flow_error;
	int ret, queue_id, count, nb_comp = 0;
	uint64_t end_tsc;
	bool force_push = true;

	end_tsc = rte_rdtsc() + ms_hz * US_PER_S;
	do {
		ret = rte_flow_template_table_destroy(port_id, table, &flow_error);
		if (!ret)
			break;
		for (queue_id = 0; queue_id < dpdk_hws_cfg.queues; queue_id++) {
			count = dpdk_flow_hws_dequeue(port_id, queue_id, PULL_TIMEOUT_US,
				dpdk_hws_cfg.queue_depth, force_push, NULL);
			nb_comp += count;
		}
		force_push = false;
	} while (rte_rdtsc() <= end_tsc);

	if (ret) {
		DOCA_DLOG_ERR("Port %d table_destory ret %d processed entries nb_comp %d",
			port_id, ret, nb_comp);
	}
	return ret;
}

static struct rte_flow_pattern_template *
dpdk_hws_item_template_create(uint16_t port_id,
			const struct rte_flow_pattern_template_attr *it_attr,
			const struct rte_flow_item items[])
{
	struct rte_flow_pattern_template *it;
	struct rte_flow_error flow_error;
	struct rte_flow_attr attr __rte_unused = {
		.ingress = it_attr->ingress,
		.egress = it_attr->egress,
		.transfer = it_attr->transfer
	};

	dpdk_dbg_rte_flow_dump("pattern template:\t", port_id, &attr, items, NULL, stdout);

	it = rte_flow_pattern_template_create(port_id, it_attr, items, &flow_error);
	if (!it) {
		DOCA_LOG_ERR("failed to create hws item template - port %u create rte hws item template, type %d message: %s",
			     port_id, flow_error.type, flow_error.message ? flow_error.message : "(no stated reason)");
	}
	return it;
}

static int
dpdk_hws_item_template_destroy(uint16_t port_id,
			struct rte_flow_pattern_template *item_template)
{
	struct rte_flow_error flow_error;
	int ret;

	ret = rte_flow_pattern_template_destroy(port_id, item_template, &flow_error);
	if (ret) {
		DOCA_LOG_ERR("failed to destroy hws item template - port %u destroy rte hws item template, type %d message: %s",
			     port_id, flow_error.type, flow_error.message ? flow_error.message : "(no stated reason)");
	}
	return ret;
}

static struct rte_flow_actions_template *
dpdk_hws_action_template_create(uint16_t port_id,
			const struct rte_flow_actions_template_attr *at_attr,
			const struct rte_flow_action actions[],
			const struct rte_flow_action masks[])
{
	struct rte_flow_actions_template *at;
	struct rte_flow_error flow_error;
	struct rte_flow_attr attr __rte_unused = {
		.ingress = at_attr->ingress,
		.egress = at_attr->egress,
		.transfer = at_attr->transfer
	};

	dpdk_dbg_rte_flow_dump("action template:\t", port_id, &attr, NULL, actions, stdout);
	dpdk_dbg_rte_flow_dump("action template mask:\t", port_id, &attr, NULL, masks, stdout);

	at = rte_flow_actions_template_create(port_id, at_attr, actions, masks, &flow_error);
	if (!at) {
		DOCA_LOG_ERR("failed to create hws action template - port %u create rte hws action template, type %d message: %s",
			     port_id, flow_error.type, flow_error.message ? flow_error.message : "(no stated reason)");
	}
	return at;
}

static int
dpdk_hws_action_template_destroy(uint16_t port_id,
			struct rte_flow_actions_template *action_template)
{
	struct rte_flow_error flow_error;
	int ret;

	ret = rte_flow_actions_template_destroy(port_id, action_template, &flow_error);
	if (ret) {
		DOCA_LOG_ERR("failed to destroy hws action template - port %u destroy rte hws action template, type %d message: %s",
			     port_id, flow_error.type, flow_error.message ? flow_error.message : "(no stated reason)");
	}
	return ret;
}

static struct rte_flow_template_table *
dpdk_hws_table_create(uint16_t port_id,
		      const struct rte_flow_template_table_attr *table_attr,
		      struct rte_flow_pattern_template *item_templates[],
		      uint8_t nb_item_templates,
		      struct rte_flow_actions_template *action_templates[],
		      uint8_t nb_action_templates)
{
	struct rte_flow_template_table *tbl;
	struct rte_flow_error flow_error;

	dpdk_dbg_rte_flow_dump("table template:\t\t", port_id,  &table_attr->flow_attr, NULL, NULL, stdout);

	tbl = rte_flow_template_table_create(port_id, table_attr, item_templates,
				nb_item_templates, action_templates,
				nb_action_templates, &flow_error);
	if (!tbl) {
		DOCA_LOG_ERR("failed to create hws table - port %u create rte hws table, type %d message: %s", port_id,
			     flow_error.type,
			     flow_error.message ? flow_error.message
						: "(no stated reason)");
	}
	return tbl;
}

static int
_create_hws_templates_table(uint16_t port_id,
			uint32_t queue_id,
			const struct rte_flow_attr *attr,
			const struct rte_flow_item items[],
			const struct rte_flow_action actions[],
			const struct rte_flow_action action_masks[],
			struct rte_flow_template_table_attr *tatr,
			struct rte_flow_op_attr *qatr,
			struct dpdk_hws_item *hws_entry)
{
	struct rte_flow_pattern_template_attr itr = {0};
	struct rte_flow_actions_template_attr atr = {0};
	struct rte_flow_pattern_template *it_array[DPDK_PIPE_MAX_ITEMS];
	struct rte_flow_actions_template *at_array[DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_pattern_template *it;
	struct rte_flow_actions_template *at;
	struct rte_flow_template_table *tbl;

	itr.ingress = tatr->flow_attr.ingress;
	itr.egress = tatr->flow_attr.egress;
	itr.transfer = tatr->flow_attr.transfer;
	it = dpdk_hws_item_template_create(port_id, &itr, items);
	if (!it) {
		DOCA_DLOG_ERR("failed to create hws table templates - item template");
		return -1;
	}

	atr.ingress = tatr->flow_attr.ingress;
	atr.egress = tatr->flow_attr.egress;
	atr.transfer = tatr->flow_attr.transfer;
	at = dpdk_hws_action_template_create(port_id, &atr, actions, action_masks);
	if (!at) {
		DOCA_DLOG_ERR("failed to create hws table templates - action template");

		dpdk_hws_item_template_destroy(port_id, it);
		return -1;
	}

	/* build table */
	it_array[0] = it;
	at_array[0] = at;
	tbl = dpdk_hws_table_create(port_id, tatr, it_array, 1, at_array, 1);
	if (!tbl) {
		DOCA_DLOG_ERR("failed to create hws table templates - hws table create");
		dpdk_hws_item_template_destroy(port_id, it);
		dpdk_hws_action_template_destroy(port_id, at);
		return -1;
	}

	hws_entry->item_template = it;
	hws_entry->action_template[0] = at;
	hws_entry->table = tbl;
	return 0;
}
/* end of legacy encap table code */

int
dpdk_flow_hws_destroy_encap_table(struct dpdk_pipe *pipe, struct doca_flow_port *port)
{
	uint16_t peer_port_id;
	void *it, *at, *tbl;
	int rc, result = 0;

	peer_port_id = pipe->encap_peer_port_id;
	tbl = pipe->encap_hws_ctx->table;
	rc = dpdk_hws_table_destroy_legacy(peer_port_id, tbl);
	if (rc) {
		DOCA_DLOG_ERR("failed to destroy hws encap table - dpdk table destroy, ret=%d", rc);
		result = rc;
	}

	it = pipe->encap_hws_ctx->item_template;
	rc = dpdk_hws_item_template_destroy(peer_port_id, it);
	if (rc) {
		DOCA_DLOG_ERR("failed to destroy hws encap table - item template destroy, ret=%d", rc);
		result = rc;
	}

	at = pipe->encap_hws_ctx->action_template[0];
	rc = dpdk_hws_action_template_destroy(peer_port_id, at);
	if (rc) {
		DOCA_DLOG_ERR("failed to destroy hws encap table - action template destroy, ret=%d", rc);
		result = rc;
	}
	dpdk_hws_free_encap_hws_ctx(pipe->encap_hws_ctx);
	pipe->encap_hws_ctx = NULL;

	return result;
}

int
dpdk_flow_hws_table_destroy(struct doca_flow_pipe *pipe)
{
	struct dpdk_hws_item *hws_entry;
	struct doca_flow_port *port;
	int rc, result = 0;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_ERR("failed to destroy table - pipe's port is null");
		return -EINVAL;
	}

	if (pipe->dpdk_pipe.encap_hws_ctx) {
		rc = dpdk_flow_hws_destroy_encap_table(&pipe->dpdk_pipe, port);
		if (rc) {
			DOCA_DLOG_ERR("failed to destroy table - destroy encap table, rc=%d", rc);
			result = rc;
		}
	}

	while ((hws_entry = LIST_FIRST(&pipe->dpdk_pipe.hws_ctx.hws_list))) {
		rc = hws_table_destroy(port, hws_entry->table);
		if (rc) {
			DOCA_DLOG_ERR("failed to destroy table, rc=%d", rc);
			result = rc;
		}
		LIST_REMOVE(hws_entry, next);
		free(hws_entry);
		pipe->dpdk_pipe.hws_ctx.nb_entries--;
	}
	return result;
}

int
dpdk_flow_hws_create_encap_table(struct dpdk_pipe *dpdk_pipe,
		struct rte_flow_action_raw_decap *decap_conf,
		struct rte_flow_action_raw_encap *encap_conf)
{
	struct rte_flow_action actions[DPDK_PIPE_MAX_ACTIONS] = {0};
	struct rte_flow_action action_masks[DPDK_PIPE_MAX_ACTIONS] = {0};
	struct rte_flow_item items[DPDK_PIPE_MAX_ITEMS] = {0};
	struct rte_flow_attr attr = {0};
	int action_idx = 0, action_mask_idx = 0, ret;
	struct rte_flow_op_attr qatr = {0};
	struct rte_flow_item_meta meta = {0};
	struct rte_flow_template_table_attr tatr = {0};
	struct rte_flow_action_raw_encap raw_encap = {0};
	uint8_t encap_data[MAX_ENCAP_SIZE] = {0};
	struct doca_flow_port *hairpin_peer;
	struct doca_flow_port *port;
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	uint16_t port_id;

	if (dpdk_pipe->encap_hws_ctx)
		return 0;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_ERR("failed destroy table - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	hairpin_peer = dpdk_engine_get_hairpin_peer(port);
	if (!hairpin_peer) {
		DOCA_LOG_ERR("hairpin_peer is NULL.");
		return -1;
	}
	port_id = hairpin_peer->port_id;

	dpdk_pipe->encap_hws_ctx = dpdk_hws_alloc_encap_hws_ctx();
	if (!dpdk_pipe->encap_hws_ctx) {
		DOCA_LOG_ERR("alloc resource for encap_hws_ctx failed.");
		return -1;
	}
	dpdk_pipe->encap_peer_port_id = port_id;
	attr.group = DOCA_EGRESS_DEFAULT_GROUP;
	attr.egress = 1;

	items[0].type = RTE_FLOW_ITEM_TYPE_META;
	meta.data = UINT32_MAX;
	items[0].mask = &meta;
	items[1].type = RTE_FLOW_ITEM_TYPE_END;

	tatr.flow_attr = attr;
	tatr.nb_flows = DPDK_PIPE_ACTIONS_MAX_ENCAPS;

	/* build actions */
	if (decap_conf) {
		actions[action_idx].type = RTE_FLOW_ACTION_TYPE_RAW_DECAP;
		actions[action_idx++].conf = decap_conf;
	}
	actions[action_idx].type = RTE_FLOW_ACTION_TYPE_RAW_ENCAP;
	raw_encap.size = encap_conf->size;
	raw_encap.data = encap_data;
	actions[action_idx++].conf = &raw_encap;

	/* build action_masks */
	if (decap_conf) {
		action_masks[action_mask_idx].type = RTE_FLOW_ACTION_TYPE_RAW_DECAP;
		action_masks[action_mask_idx++].conf = decap_conf;
	}
	action_masks[action_mask_idx].type = RTE_FLOW_ACTION_TYPE_RAW_ENCAP;
	action_masks[action_mask_idx++].conf = NULL;
	action_masks[action_mask_idx].type = RTE_FLOW_ACTION_TYPE_END;

	ret = _create_hws_templates_table(port_id,
			0, &attr, items, actions, action_masks,
			&tatr, &qatr, dpdk_pipe->encap_hws_ctx);
	if (ret) {
		DOCA_LOG_INFO("create table failed for tx_encap ctx, enable the encap in rx");
		port->rx_encap = true;
		dpdk_hws_free_encap_hws_ctx(dpdk_pipe->encap_hws_ctx);
		dpdk_pipe->encap_hws_ctx = NULL;
	}

	return 0;
}

int
dpdk_flow_hws_templates_table_create(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
				     uint8_t nb_actions, uint32_t nb_flows)
{
	struct rte_flow_action_modify_field modis[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS] = {};
	struct rte_flow_action action_masks[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_action *actions_mask_arr[MAX_ACT_TEMPLATE];
	struct rte_flow_action *actions_arr[MAX_ACT_TEMPLATE];
	struct rte_flow_template_table_attr tatr = {0};
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;
	struct dpdk_hws_item *hws_item;
	struct rte_flow_attr attr = {0};
	int rc, i;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_ERR("failed creating hws table - pipe's port is null");
		return -EINVAL;
	}

	hws_item = calloc(1, sizeof(*hws_item));
	if (!hws_item) {
		DOCA_LOG_ERR("allocate dpdk_hws_item memory failed.");
		return -ENOMEM;
	}

	for (i = 0; i < nb_actions; i++) {
		dpdk_pipe_actions_masks_build(pipe_q, action_masks[i], modis[i], i);
		actions_arr[i] = pipe_q->action_ctx[i].actions;
		actions_mask_arr[i] = action_masks[i];
	}

	attr.ingress = pipe_q->attr.ingress;
	attr.transfer = pipe_q->attr.transfer;
	attr.egress = pipe_q->attr.egress;
	attr.group = pipe_q->attr.group;
	tatr.flow_attr = attr;
	tatr.nb_flows = nb_flows ? nb_flows : DEFAULT_NB_FLOWS;

	rc = get_hws_table(port, pipe_q->items, actions_arr, actions_mask_arr, nb_actions, &tatr,
			   hws_item);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("%s failed - get_hws_table failed with rc=%d", __func__, rc);
		goto err;
	}

	LIST_INSERT_HEAD(&dpdk_pipe->hws_ctx.hws_list, hws_item, next);
	dpdk_pipe->hws_ctx.nb_entries++;
	return 0;

err:
	free(hws_item);
	return -1;
}

/* dpdk_engine register ops */
int
dpdk_hws_flow_entries_process(struct doca_flow_port *port,
		uint16_t pipe_queue, uint64_t timeout,
		uint32_t max_processed_entries)
{
	bool force_push = false;

	return dpdk_flow_hws_dequeue(port->port_id, pipe_queue, timeout,
			max_processed_entries, force_push, NULL);
}
