/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_hash.h>
#include <rte_hash_crc.h>
#include <rte_spinlock.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <doca_log.h>

#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "dpdk_meter_controller.h"

DOCA_LOG_REGISTER(dpdk_meter);

struct profile_hash_key {
	uint64_t cir;
	/* Committed Information Rate (bytes/second). */
	uint64_t cbs;
	/* Committed Burst Size (bytes). */
};

struct profile_hash_entry {
	struct profile_hash_key key;
};

struct profile_hash {
	int max_entries_num;
	struct rte_hash *h;
	struct profile_hash_entry entries[0];
};

static inline uint32_t
profile_hash_crc(const void *data, __rte_unused uint32_t data_len,
		 uint32_t init_val)
{
	const struct profile_hash_key *k;

	k = data;
	init_val = rte_hash_crc_8byte(k->cir, init_val);
	init_val = rte_hash_crc_8byte(k->cbs, init_val);
	return init_val;
}

struct rte_hash_parameters profile_hash_params = {
	.name = "profile_table",
	.entries = DPDK_MAX_PORT_PROFILES,
	.key_len = sizeof(struct profile_hash_key),
	.hash_func = profile_hash_crc,
	.hash_func_init_val = 0,
};

static __rte_always_inline void
build_profile_hash_key(const struct doca_flow_monitor *mon,
		       struct profile_hash_key *key)
{
	*key = (struct profile_hash_key){
		.cir = mon->cir,
		.cbs = mon->cbs,
	};
}

int
dpdk_profile_hash_get_key_idx(struct profile_hash *profile_hash_ins,
			      const struct doca_flow_monitor *mon)
{
	int idx = 0;
	struct profile_hash_key key;

	build_profile_hash_key(mon, &key);
	idx = rte_hash_lookup(profile_hash_ins->h, (void *)&key);
	if (idx >= 0)
		return idx;

	idx = rte_hash_add_key(profile_hash_ins->h, (void *)&key);
	if (idx < 0 || idx > profile_hash_ins->max_entries_num) {
		DOCA_DLOG_WARN("failed adding index");
		return -1;
	}

	profile_hash_ins->entries[idx].key = key;
	return idx;
}


int
dpdk_profile_hash_init(int max_profiles, struct profile_hash **profile_hash,
		       const char *name, uint8_t extra_flag)
{
	int total_size = sizeof(struct profile_hash) +
			 sizeof(struct profile_hash_entry) * max_profiles;

	struct profile_hash *profile_hash_ins;

	*profile_hash = NULL;
	profile_hash_ins = (struct profile_hash *)calloc(1, total_size);
	if (profile_hash_ins == NULL) {
		DOCA_DLOG_ERR("failed to alloc mem");
		return -ENOMEM;
	}

	profile_hash_params.name = name;
	profile_hash_params.entries = max_profiles;
	profile_hash_params.extra_flag = extra_flag;
	profile_hash_ins->max_entries_num = max_profiles;

	profile_hash_ins->h = rte_hash_create(&profile_hash_params);
	if (profile_hash_ins->h == NULL) {
		DOCA_DLOG_ERR("failed to alloc hash table %s with %u entries",
			      profile_hash_params.name,
			      profile_hash_params.entries);
		free(profile_hash_ins);
		return -1;
	}
	DOCA_LOG_DBG("profile table created (%s)", name);
	*profile_hash = profile_hash_ins;
	return 0;
}

void
dpdk_profile_hash_destroy(struct profile_hash *profile_hash_ins)
{
	if (profile_hash_ins == NULL) {
		DOCA_DLOG_WARN("failed to destroy hash table - null pointer");
		return;
	}

	const char *name = (char *)profile_hash_ins->h;

	DOCA_LOG_DBG("profile table destroyed (%s)", name);
	if (profile_hash_ins->h)
		rte_hash_free(profile_hash_ins->h);

	free(profile_hash_ins);
	profile_hash_ins = NULL;
}

static inline int
dpdk_port_meter_profile_id_get(struct doca_flow_port *port,
			       const struct doca_flow_monitor *mon,
			       uint32_t *profile_id)
{
	int idx, ret;

	rte_spinlock_lock(&port->profile_lock);
	idx = dpdk_profile_hash_get_key_idx(port->profile_ids->id_hash, mon);
	if (idx < 0)
		goto unlock;
	*profile_id = port->profile_ids->ids[idx];
	if (*profile_id)
		goto success;

	/* Create a new profile */
	*profile_id = ++port->profile_id_generator;
	port->profile_ids->ids[idx] = *profile_id;
	ret = dpdk_meter_profile_create(port->port_id, DPDK_ID(*profile_id),
					       mon->cir, mon->cbs);
	if (ret < 0)
		goto unlock;
success:
	rte_spinlock_unlock(&port->profile_lock);
	return 0;
unlock:
	rte_spinlock_unlock(&port->profile_lock);
	return -1;
}

int
dpdk_meter_profile_id_get(struct doca_flow_port *port,
			  struct dpdk_pipe_q_ctx *pipe_q,
			  const struct doca_flow_monitor *mon,
			  uint32_t *profile_id)
{
	/*
	 * Search for the profile id with <CIR,CBS> key in local hash
	 * (lockless).
	 * If found - return the profile ID with success.
	 * If not found - get the profile id from the port (port will lock).
	 * The port may either create the profile id for the first time or find
	 * it in its local hash.  The port has a max limit number of 128
	 * profile ids with different keys.
	 * Oncy the profile id is received from the port - update the local
	 * hash.
	 */

	int idx, ret;

	/* Look for the profile ID in the local pipe queue hash */
	idx = dpdk_profile_hash_get_key_idx(pipe_q->profile_ids->id_hash, mon);
	if (idx < 0) {
		DOCA_LOG_ERR("Failed to get profile idx in local pipe queue.");
		return -1;
	}
	/* An index was obtained. Now check if profile id exists */
	*profile_id = pipe_q->profile_ids->ids[idx];
	if (*profile_id)
		return 0;
	/* Not found locally - get it from port (may be created first time) */
	ret = dpdk_port_meter_profile_id_get(port, mon, profile_id);
	if (ret)
		return -1;
	if (*profile_id) {
		/* Update the local pipe queue */
		pipe_q->profile_ids->ids[idx] = *profile_id;
		return 0;
	}
	return -1;
}
