/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_spinlock.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <doca_log.h>

#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "doca_flow_layer.h"

DOCA_LOG_REGISTER(dpdk_meter_pipe);

static struct doca_flow_pipe *
dpdk_pipe_control_meter_create(struct doca_flow_port *port)
{
	struct doca_flow_pipe_cfg pipe_cfg = {0};
	struct doca_flow_pipe *pipe;

	DOCA_DLOG_DBG("port:%u create meter pipe", port->port_id);
	pipe_cfg.port = port;
	pipe_cfg.attr.is_root = false;

	pipe = dpdk_pipe_create(ENGINE_PIPE_CONTROL, &pipe_cfg, NULL, NULL);
	if (!pipe) {
		DOCA_DLOG_ERR("failed to create control meter pipe");
		return NULL;
	}

	return pipe;
}

static int
_dpdk_meter_pipe_create(struct doca_flow_pipe *pipe,
			uint16_t pipe_queue,
			const struct doca_flow_monitor *mon,
			const struct doca_flow_fwd *fwd,
			struct doca_flow_pipe **mtr_pipe)
{
	struct doca_flow_port *port;
	struct doca_flow_pipe *meter_pipe;
	struct doca_flow_match red_match = {0};
	struct doca_flow_match green_match = {0};
	struct doca_flow_fwd fwd_drop = {0};

	red_match.flags |= DPDK_MATCH_METER_COLOR_RED;
	green_match.flags |= DPDK_MATCH_METER_COLOR_GREEN;
	fwd_drop.type = DOCA_FLOW_FWD_DROP;

	if (!mtr_pipe)
		return 0;

	*mtr_pipe = NULL;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed creating internal meter pipe - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	meter_pipe = dpdk_pipe_control_meter_create(port);
	if (!meter_pipe) {
		DOCA_LOG_ERR("failed to create internal meter pipe - creating shared meter (%u) pipe failed",
			     mon->shared_meter_id);
		return -ENOENT;
	}

	struct dpdk_pipe_q_ctx *pipe_q = &meter_pipe->dpdk_pipe.queues[pipe_queue].pipe_q;

	dpdk_engine_pipe_q_ctx_group_set(meter_pipe, pipe_q, fwd, false);

	if (dpdk_pipe_entry_add(pipe_queue, 0 /* priority */,
				meter_pipe, &red_match, NULL /* match_mask */, NULL /* actions */,
				NULL /* action_descs */, NULL /* monitor */, &fwd_drop,
				DOCA_FLOW_NO_WAIT, NULL /* user_ctx */) == NULL) {
		DOCA_LOG_ERR("failed to create internal meter pipe - adding RED policy entry failed, shared meter (%u)",
			     mon->shared_meter_id);

		goto entry_fail;
	}

	if (dpdk_pipe_entry_add(pipe_queue, 0 /* priority */,
				meter_pipe, &green_match, NULL /* match_mask */, NULL /* actions */,
				NULL /* action_descs */, NULL /* monitor */, fwd,
				DOCA_FLOW_NO_WAIT, NULL /* user_ctx */) == NULL) {
		DOCA_LOG_ERR("failed to create internal meter pipe - adding GREEN policy entry failed, shared meter (%u)",
			      mon->shared_meter_id);

		goto entry_fail;

	}

	*mtr_pipe = meter_pipe;

	return 0;

entry_fail:

	if (meter_pipe)
		dpdk_pipe_destroy(port, meter_pipe);

	return -ENOENT;
}

int
dpdk_meter_pipe_create(struct doca_flow_pipe *pipe,
		       uint16_t pipe_queue,
		       const struct doca_flow_monitor *mon,
		       const struct doca_flow_fwd *fwd)
{
	int ret;
	struct doca_flow_pipe *meter_pipe;

	if (!pipe->meter_pipe) {
		ret = _dpdk_meter_pipe_create(pipe,
				pipe_queue,
				mon, fwd,
				&meter_pipe);
		if (ret)
			return ret;

		pipe->meter_pipe = meter_pipe;
	}

	return 0;
}

void
dpdk_meter_pipe_fwd_update(const struct doca_flow_fwd *fwd,
			   struct doca_flow_pipe *pipe,
			   struct doca_flow_fwd *update_fwd,
			   bool *is_new_fwd)
{
	if (pipe->meter_pipe) {
		update_fwd->type = DOCA_FLOW_FWD_PIPE;
		update_fwd->next_pipe = pipe->meter_pipe;
		*is_new_fwd = true;
	} else
		*is_new_fwd = false;
}

void
dpdk_meter_pipe_destroy(struct doca_flow_pipe *pipe)
{
	struct doca_flow_port *port;

	if (pipe->meter_pipe) {
		port = doca_flow_layer_get_pipe_port(pipe);
		if (unlikely(!port)) {
			DOCA_LOG_CRIT(
				"failed destroying internal meter pipe - pipe's port is null");
			return;
		}

		dpdk_pipe_destroy(port, pipe->meter_pipe);
	}
}
