/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_mtr.h>

#include "dpdk_pipe_actions.h"
#include "dpdk_flow_hws_legacy.h"
#include "dpdk_engine.h"
#include "dpdk_pipe_hash.h"
#include "dpdk_utilities.h"
#include "doca_flow_priv.h"

#include "dpdk_shared_counter.h"
#include "dpdk_shared_meter.h"
#include "dpdk_meter_controller.h"
#include "dpdk_shared_crypto.h"
#include "dpdk_meter_pipe.h"
#include "dpdk_port.h"

#include "id_pool.h"
#include "encap_table.h"
#include "utils_df_translate.h"
#include "utils_field.h"

#include "engine_model.h"
#include "doca_flow_layer.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_pipe_actions);

#define EMPTY_GROUP_ID (UINT16_MAX - 1)

/**
 * encap_id is the encap data hash value which start
 * from 0, meta_id will start from encap_id + 1.
 */
#define META_ID(encap_id) ((encap_id) + 1)
#define ENCAP_ID(meta_id) ((meta_id) - 1)

/*
 * Verify that all addresses: action_data, action_data.meter.conf and
 * action_data.meter_mark.conf are the same.
 */
static_assert((void *)&((struct dpdk_action_entry *)0)->action_data ==
	      (void *)&((struct dpdk_action_entry *)0)->action_data.meter.conf,
	      "meter conf and action data addresses must be identical");
static_assert((void *)&((struct dpdk_action_entry *)0)->action_data ==
	      (void *)&((struct dpdk_action_entry *)0)->action_data.meter_mark.conf,
	      "meter mark conf and action data addresses must be identical");

static struct dpdk_pipe_actions_encap_info encap_info_arr[DPDK_PIPE_ACTIONS_MAX_ENCAPS];

struct dpdk_pipe_actions *
dpdk_pipe_actions_ctx_get(struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_ctx_idx)
{
	return &pipe_q->action_ctx[act_ctx_idx];
}

static __rte_always_inline struct dpdk_action_entry *
dpdk_pipe_actions_entry_get(struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_ctx_idx, uint8_t act_idx)
{
	struct dpdk_pipe_actions *action_ctx;

	action_ctx = dpdk_pipe_actions_ctx_get(pipe_q, act_ctx_idx);
	return &action_ctx->action_entry[act_idx];
}

struct rte_flow_action *
dpdk_pipe_actions_get(struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_ctx_idx, uint8_t act_idx)
{
	struct dpdk_pipe_actions *action_ctx;

	action_ctx = dpdk_pipe_actions_ctx_get(pipe_q, act_ctx_idx);
	return &action_ctx->actions[act_idx];
}

static __rte_always_inline struct rte_flow_action *
dpdk_pipe_masks_get(struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_ctx_idx, uint8_t act_idx)
{
	struct dpdk_pipe_actions *action_ctx;

	action_ctx = dpdk_pipe_actions_ctx_get(pipe_q, act_ctx_idx);
	return &action_ctx->masks[act_idx];
}

struct rte_flow_action *
dpdk_pipe_actions_array_get(struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_ctx_idx)
{
	struct dpdk_pipe_actions *action_ctx;

	action_ctx = dpdk_pipe_actions_ctx_get(pipe_q, act_ctx_idx);
	return action_ctx->actions;
}

static __rte_always_inline int
dpdk_pipe_act_idx_get_next(struct dpdk_pipe *dpdk_pipe, uint8_t act_arr_idx)
{
	return dpdk_pipe->nb_actions_pipe[act_arr_idx]++;
}

struct dpdk_action_entry *
dpdk_pipe_actions_entry_get_next(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
				 uint8_t act_arr_idx)
{
	int idx;

	if (act_arr_idx >= dpdk_pipe->nb_actions_array_len) {
		DOCA_LOG_RATE_LIMIT_ERR("invalid act_arr_idx %d, pipe's action array len %d",
					act_arr_idx, dpdk_pipe->nb_actions_array_len);
		return NULL;
	}
	idx = dpdk_pipe_act_idx_get_next(dpdk_pipe, act_arr_idx);
	return dpdk_pipe_actions_entry_get(pipe_q, act_arr_idx, idx);
}

struct dpdk_action_entry *
dpdk_pipe_actions_entry_get_last(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
				 uint8_t act_arr_idx)
{
	int idx;

	if (act_arr_idx >= dpdk_pipe->nb_actions_array_len) {
		DOCA_LOG_RATE_LIMIT_ERR("invalid act_arr_idx %d, pipe's action array len %d",
					act_arr_idx, dpdk_pipe->nb_actions_array_len);
		return NULL;
	}

	idx = dpdk_pipe->nb_actions_pipe[act_arr_idx];
	if (idx == 0)
		return NULL;
	return dpdk_pipe_actions_entry_get(pipe_q, act_arr_idx, idx - 1);
}

static __rte_always_inline bool
is_pipe_has_fwd(struct dpdk_pipe_q_ctx *pipe_q)
{
	return pipe_q->actions_flags & DPDK_PIPE_FWD;
}

static __rte_always_inline bool
is_pipe_fwd_changeable(struct dpdk_pipe_q_ctx *pipe_q)
{
	return pipe_q->fwd_action_index != INVALID_ACTION_INDEX;
}

static __rte_always_inline bool
is_pipe_has_count(struct dpdk_pipe_q_ctx *pipe_q)
{
	return pipe_q->actions_flags & DPDK_PIPE_COUNT;
}

static __rte_always_inline bool
is_pipe_has_meter(struct dpdk_pipe_q_ctx *pipe_q)
{
	return pipe_q->actions_flags & DPDK_PIPE_METER;
}

static __rte_always_inline bool
is_pipe_policy_initialized(struct dpdk_pipe *dpdk_pipe)
{
	return (!!dpdk_pipe->policy_id);
}

static __rte_always_inline bool
is_pipe_action_index_valid(int index)
{
	return ((index >= 0) && (index < DPDK_PIPE_MAX_ACTIONS));
}

enum {
	MON_RESOURCE_NOTHING_BUILD,
	MON_RESOURCE_NON_SHARED_BUILD,
	MON_RESOURCE_SHARED_BUILD,
};

static int
dpdk_pipe_actions_monitor_cb(struct dpdk_pipe *dpdk_pipe,
			     struct dpdk_action_entry *action_entry,
			     __rte_unused const struct doca_flow_actions *pkt_action,
			     const struct doca_flow_monitor *mon,
			     __rte_unused const struct doca_flow_fwd *fwd,
			     struct doca_flow_pipe_entry *entry);


void
dpdk_pipe_actions_init(struct doca_flow_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q)
{
	int act_arr_idx, nb_actions, idx;
	struct dpdk_action_entry *action_entry;

	nb_actions = pipe->dpdk_pipe.nb_actions_array_len;
	for (act_arr_idx = 0; act_arr_idx < nb_actions; act_arr_idx++)
		for (idx = 0; idx < DPDK_PIPE_MAX_ACTIONS; idx++) {
			action_entry = dpdk_pipe_actions_entry_get(pipe_q, act_arr_idx, idx);
			action_entry->action = dpdk_pipe_actions_get(pipe_q, act_arr_idx, idx);
			action_entry->mask = dpdk_pipe_masks_get(pipe_q, act_arr_idx, idx);
		}
}

static int
encap_ip_conv(const struct doca_flow_ip_addr *doca_ip,
	struct encap_table_ip_addr *encap_ip)
{
	switch (doca_ip->type) {
	case DOCA_FLOW_IP4_ADDR:
		encap_ip->type = ENCAP_TABLE_IP4_ADDR;
		encap_ip->ipv4_addr = doca_ip->ipv4_addr;
		break;
	case DOCA_FLOW_IP6_ADDR:
		encap_ip->type = ENCAP_TABLE_IP6_ADDR;
		memcpy(encap_ip->ipv6_addr, doca_ip->ipv6_addr, sizeof(doca_ip->ipv6_addr));
		break;
	default:
		DOCA_LOG_ERR("invalid ip type %d", doca_ip->type);
		return -EINVAL;
	}

	return 0;
}

static int
encap_tun_conv(const struct doca_flow_tun *doca_tun,
	struct encap_table_tun *encap_tun)
{
	switch (doca_tun->type) {
	case DOCA_FLOW_TUN_VXLAN:
		encap_tun->type = ENCAP_TABLE_TUN_VXLAN;
		encap_tun->vxlan_tun_id = doca_tun->vxlan_tun_id;
		break;
	case DOCA_FLOW_TUN_GRE:
		encap_tun->type = ENCAP_TABLE_TUN_GRE;
		encap_tun->gre_key = doca_tun->gre_key;
		break;
	case DOCA_FLOW_TUN_GTPU:
		encap_tun->type = ENCAP_TABLE_TUN_GTPU;
		encap_tun->gtp_teid = doca_tun->gtp_teid;
		break;
	case DOCA_FLOW_TUN_NISP:
	case DOCA_FLOW_TUN_AUDP:
	case DOCA_FLOW_TUN_ESP:
	default:
		DOCA_LOG_ERR("invalid tun type %d", doca_tun->type);
		return -EINVAL;
	}

	return 0;
}

static int
encap_key_conv(uint16_t port_id, const struct doca_flow_encap_action *encap_data,
	struct encap_table_key *key)
{
	int ret;

	if (encap_data->src_ip.type != encap_data->dst_ip.type) {
		DOCA_LOG_ERR("invalid ip type: src %u dst %u",
			encap_data->src_ip.type, encap_data->dst_ip.type);
		return -EINVAL;
	}

	key->port_id = port_id;
	ret = encap_ip_conv(&encap_data->src_ip, &key->src_ip);
	if (ret)
		return ret;
	ret = encap_ip_conv(&encap_data->dst_ip, &key->dst_ip);
	if (ret)
		return ret;
	ret = encap_tun_conv(&encap_data->tun, &key->tun);
	if (ret)
		return ret;
	key->ip_type = key->src_ip.type;

	return 0;
}

static bool
is_transfer_set_in_queue_0(struct dpdk_pipe *dpdk_pipe)
{
	return dpdk_pipe->queues[0].pipe_q.attr.transfer;
}

/*
 * This API is called in pipe create and entry create. For parameter "flow_entry",
 * when it is NULL, it means that it is called during pipe creating. Otherwise,
 * it is called during entry creating.
 */
static int
dpdk_encap_action_modify(struct dpdk_pipe *dpdk_pipe,
			 struct dpdk_action_entry *entry,
			 const struct doca_flow_actions *pkt_action,
			 __rte_unused const struct doca_flow_monitor *mon,
			 __rte_unused const struct doca_flow_fwd *fwd,
			 struct doca_flow_pipe_entry *flow_entry)
{
	struct rte_flow_action_raw_decap *decap_conf = NULL;
	struct dpdk_action_rawencap_data *encap = &entry->action_data.rawencap;
	struct dpdk_action_modify_field_data *meta = &entry->action_data.rawencap.meta;
	const struct doca_flow_encap_action *encap_data = &pkt_action->encap;
	struct rte_flow_action *action = entry->action;
	uint32_t refcnt;
	uint8_t *header;
	int encap_id, meta_id, ret;
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;
	struct doca_flow_pipe_entry *encap_entry;
	struct rte_flow_action_raw_decap conf;
	uint8_t data[128];
	struct doca_flow_port *peer_port;
	bool is_init_encap_action = !flow_entry;
	bool is_add_entry  = !!flow_entry;
	int rc;
	struct encap_table_key key = {0};

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed modifying encap action - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	header = encap->data;
	dpdk_raw_data_build(&header, encap_data, encap->layer, true);

	if (unlikely(flow_entry))
		assert(flow_entry->encap_id < 0);
	encap->conf.data = encap->data;
	encap->conf.size = header - encap->data;

	/* If tx encap doesn't supported or the rules is fdb, will do encap on the original port */
	if (port->rx_encap ||
	    is_transfer_set_in_queue_0(dpdk_pipe)) {
		action->type = RTE_FLOW_ACTION_TYPE_RAW_ENCAP;
		action->conf = &encap->conf;
		return 0;
	}

	ret = encap_key_conv(port->port_id, encap_data, &key);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_WARN(
			"failed modifying encap action - invalid encap_data");
		return ret;
	}
	if (is_add_entry) {
		// refcount update only for add entry usage.
		encap_id = encap_table_add_id(&key, &refcnt);
		if (encap_id < 0) {
			DOCA_LOG_RATE_LIMIT_WARN(
				"failed modifying encap action - cannot add id");
			return -EINVAL;
		}
		meta_id = META_ID(encap_id);
	} else {
		meta_id = UINT32_MAX;
	}

	if (is_add_entry && refcnt == 1) {
		ret = dpdk_utils_build_encap_raw_data(pkt_action, header,
					&decap_conf, &conf, data);
		if (ret) {
			DOCA_LOG_RATE_LIMIT_WARN("unsupported tunnel type (%u) for encap",
				pkt_action->encap.tun.type);
			encap_table_remove_id(encap_id);
			return -ENOTSUP;
		}
		peer_port = dpdk_engine_get_hairpin_peer(port);
		if (unlikely(!peer_port)) {
			DOCA_LOG_RATE_LIMIT_ERR("No hairpin peer port found to attach encap flow.");
			return -EINVAL;
		}
		encap_entry = dpdk_engine_encap_entry_alloc(pipe);
		if (!encap_entry) {
			DOCA_LOG_RATE_LIMIT_ERR("No flow entries in the memory pool.");
			encap_table_remove_id(encap_id);
			return -ENOMEM;
		}
		rc = dpdk_engine_encap_entry_tx_flow_add(peer_port->port_id, meta_id,
							 decap_conf, &encap->conf,
							 encap_entry);
		if (rc) {
			encap_table_remove_id(encap_id);
			dpdk_engine_encap_entry_free(peer_port->port_id, encap_entry);
			return rc;
		}

		encap_entry->encap_id = encap_id;
		encap_info_arr[encap_id].handler = encap_entry;
		encap_info_arr[encap_id].port_id = peer_port->port_id;
	}
	meta->conf.operation = RTE_FLOW_MODIFY_SET;
	meta->conf.width = 32;
	meta->conf.dst.field = RTE_FLOW_FIELD_META;
	meta->conf.dst.level = 0;
	meta->conf.dst.offset = 0;
	meta->conf.src.field = RTE_FLOW_FIELD_VALUE;
	memcpy(&meta->conf.src.value, &meta_id, sizeof(meta_id));
	action->type = RTE_FLOW_ACTION_TYPE_MODIFY_FIELD;
	action->conf = &meta->conf;

	if (is_init_encap_action) {
		ret = dpdk_engine_encap_table_create(dpdk_pipe, decap_conf, &encap->conf);
		if (ret) {
			DOCA_DLOG_ERR("Create tx_encap ctx failed.");
			return ret;
		}
		/* If tx encap is unsupported, update the encap on rx during the first creation */
		if (port->rx_encap) {
			action->type = RTE_FLOW_ACTION_TYPE_RAW_ENCAP;
			action->conf = &encap->conf;
		}
	} else {
		flow_entry->encap_id = encap_id;
	}

	return 0;
}

struct dpdk_pipe_actions_encap_info *
dpdk_pipe_actions_encap_info_put(int encap_id)
{
	struct dpdk_pipe_actions_encap_info *encap_info;

	if (unlikely(encap_id < 0 || encap_id >= DPDK_PIPE_ACTIONS_MAX_ENCAPS))
		return NULL;

	encap_info = &encap_info_arr[encap_id];

	if (!encap_info->handler)
		return NULL;

	if (encap_table_remove_id(encap_id))
		/* returning the pointer means we allow it to be cleaned. */
		return NULL;

	return encap_info;
}

static int
dpdk_modify_pipe_encap(struct dpdk_pipe *dpdk_pipe,
			 struct dpdk_action_entry *entry,
			 const struct doca_flow_actions *pkt_action)
{
	return dpdk_encap_action_modify(dpdk_pipe, entry, pkt_action, NULL, NULL, NULL);
}

static bool
encap_is_changeable(const struct doca_flow_pipe_cfg *cfg, uint8_t act_arr_idx)
{
	struct doca_flow_actions *actions;

	actions = cfg->actions[act_arr_idx];
	if (utils_df_translate_is_mac_changeable(actions->encap.src_mac)
		&& utils_df_translate_is_mac_changeable(actions->encap.dst_mac)
		&& utils_df_translate_is_ip_changeable(&actions->encap.src_ip)
		&& utils_df_translate_is_ip_changeable(&actions->encap.dst_ip))
		return true;

	return false;
}

static void
dpdk_encap_action_build(struct dpdk_action_entry *entry,
			const struct doca_flow_pipe_cfg *cfg,
			uint8_t layer, uint8_t act_arr_idx)
{
	struct rte_flow_action *action = entry->action;
	struct dpdk_action_rawencap_data *encap;
	uint8_t *header;

	encap = &entry->action_data.rawencap;
	header = encap->data;
	encap->layer = layer;

	dpdk_raw_data_build(&header, &cfg->actions[act_arr_idx]->encap,
		layer, true);
	encap->conf.data = encap->data;
	encap->conf.size = header - encap->data;
	action->type = RTE_FLOW_ACTION_TYPE_RAW_ENCAP;
	action->conf = &encap->conf;
	encap->is_changeable = encap_is_changeable(cfg, act_arr_idx);
}

static void
dpdk_decap_action_build(struct dpdk_action_entry *entry,
			const struct doca_flow_pipe_cfg *cfg,
			uint8_t layer)
{
	struct rte_flow_action *action = entry->action;
	struct dpdk_action_rawdecap_data *decap;
	uint8_t *header;

	decap = &entry->action_data.rawdecap;
	header = decap->data;

	dpdk_raw_data_build(&header, NULL, layer, false);
	decap->conf.data = decap->data;
	decap->conf.size = header - decap->data;
	action->type = RTE_FLOW_ACTION_TYPE_RAW_DECAP;
	action->conf = &decap->conf;
}

static void
dpdk_encap_implicit_cfg_set(const struct doca_flow_pipe_cfg *cfg,
			    uint8_t act_arr_idx)
{
	struct doca_flow_encap_action *encap_data = &cfg->actions[act_arr_idx]->encap;
	struct doca_flow_actions *action = cfg->actions[act_arr_idx];

	memcpy(encap_data->src_mac, action->mod_src_mac, sizeof(action->mod_src_mac));
	memcpy(encap_data->dst_mac, action->mod_dst_mac, sizeof(action->mod_dst_mac));
	encap_data->vlan_tci = action->mod_vlan_id;
	encap_data->src_ip.type = action->mod_dst_ip.type;
}

static int
dpdk_mark_action_modify(struct dpdk_pipe *dpdk_pipe,
		      struct dpdk_action_entry *entry,
		      const struct doca_flow_actions *pkt_act,
		      __rte_unused const struct doca_flow_monitor *mon,
		      __rte_unused const struct doca_flow_fwd *fwd,
		      struct doca_flow_pipe_entry *flow_entry)
{
	const struct doca_flow_actions *pkt_action = pkt_act;
	struct dpdk_action_mark_data *mark;

	mark = &entry->action_data.mark;
	if (pkt_action->meta.mark)
		mark->conf.id = pkt_action->meta.mark;
	entry->action->conf = &mark->conf;

	return 0;
}

static int
dpdk_mark_action_build(struct dpdk_pipe *dpdk_pipe,
		       struct dpdk_pipe_q_ctx *pipe_q,
		       const struct doca_flow_pipe_cfg *cfg,
		       uint8_t act_arr_idx,
		       struct doca_flow_actions *doca_flow_actions)
{
	struct dpdk_action_entry *entry;
	struct rte_flow_action *mark_action;
	struct dpdk_action_mark_data *mark;

	entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
	mark = &entry->action_data.mark;
	mark->conf.id = doca_flow_actions->meta.mark;

	mark_action = entry->action;
	mark_action->type = RTE_FLOW_ACTION_TYPE_MARK;
	mark_action->conf = &mark->conf;

	entry->mask->type = RTE_FLOW_ACTION_TYPE_MARK;
	entry->mask->conf = NULL;
	entry->modify_action = dpdk_mark_action_modify;

	return 0;
}

static int
dpdk_decap_build(struct dpdk_pipe *dpdk_pipe,
		 struct dpdk_pipe_q_ctx *pipe_q,
		 const struct doca_flow_pipe_cfg *cfg,
		 uint8_t act_arr_idx)
{
	struct doca_flow_match *match = cfg->match;
	struct dpdk_action_entry *entry;
	uint8_t layer;

	if (!match)
		return 0;

	switch (match->tun.type) {
	case DOCA_FLOW_TUN_VXLAN:
		layer = FILL_ETH_HDR | FILL_IPV4_HDR | FILL_UDP_HDR |
			FILL_VXLAN_HDR;
		entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
		dpdk_decap_action_build(entry, cfg, layer);
		return 0;
	case DOCA_FLOW_TUN_GRE:
		layer = FILL_ETH_HDR | FILL_IPV4_HDR | FILL_GRE_HDR;
		break;
	case DOCA_FLOW_TUN_GTPU:
		layer =
		    FILL_ETH_HDR | FILL_IPV4_HDR | FILL_UDP_HDR | FILL_GTP_HDR;
		break;
	case DOCA_FLOW_TUN_NISP:
	case DOCA_FLOW_TUN_AUDP:
	case DOCA_FLOW_TUN_ESP:
	default:
		DOCA_LOG_RATE_LIMIT_WARN(
			"unsupported decap tunnel type (%u)",
			match->tun.type);
		return -1;
	}

	dpdk_encap_implicit_cfg_set(cfg, act_arr_idx);

	entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
	dpdk_decap_action_build(entry, cfg, layer);
	entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
	dpdk_encap_action_build(entry, cfg, FILL_ETH_HDR, act_arr_idx);
	return 0;
}

static int
get_encap_ip_type(struct doca_flow_actions *actions)
{
	if (actions->encap.src_ip.type != actions->encap.dst_ip.type)
		return -EINVAL;
	return actions->encap.src_ip.type;
}

static void
dpdk_decap_encap_actions_build(struct dpdk_pipe *dpdk_pipe,
			       struct dpdk_pipe_q_ctx *pipe_q,
			       struct dpdk_action_entry *decap_entry,
			       const struct doca_flow_pipe_cfg *cfg,
			       uint8_t decap_layer,
			       uint8_t layer,
			       uint8_t act_arr_idx)
{
	pipe_entry_action_cb dpdk_cb = decap_entry->modify_action;
	struct dpdk_action_entry *entry;

	dpdk_decap_action_build(decap_entry, cfg, decap_layer);
	entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q,
						 act_arr_idx);
	dpdk_encap_action_build(entry, cfg, layer, act_arr_idx);
	entry->modify_action = dpdk_cb;
}

static int
dpdk_encap_build(struct dpdk_pipe *dpdk_pipe,
		 struct dpdk_pipe_q_ctx *pipe_q,
		 const struct doca_flow_pipe_cfg *cfg,
		 uint8_t act_arr_idx)
{
	struct doca_flow_actions *actions = cfg->actions[act_arr_idx];
	struct dpdk_action_entry *entry;
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;
	uint8_t layer, ip_layer = FILL_IPV4_HDR;
	int encap_ip_type;
	uint8_t decap_layer = 0;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed building encap action - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	encap_ip_type = get_encap_ip_type(actions);
	if (encap_ip_type < 0) {
		DOCA_LOG_ERR("failed to build encap action - inconsistent encap ip type, src %d dst %d",
			     actions->encap.src_ip.type, actions->encap.dst_ip.type);
		return -1;
	} else if (encap_ip_type == DOCA_FLOW_IP6_ADDR)
		ip_layer = FILL_IPV6_HDR;

	switch (actions->encap.tun.type) {
	case DOCA_FLOW_TUN_VXLAN:
		layer = FILL_ETH_HDR | ip_layer | FILL_UDP_HDR | FILL_VXLAN_HDR;
		break;
	case DOCA_FLOW_TUN_GRE:
		if (port->rx_encap || pipe_q->attr.transfer)
			decap_layer = FILL_ETH_HDR;

		layer = FILL_ETH_HDR | ip_layer | FILL_GRE_HDR;
		break;
	case DOCA_FLOW_TUN_GTPU:
		if (port->rx_encap || pipe_q->attr.transfer)
			decap_layer = FILL_ETH_HDR;

		layer = FILL_ETH_HDR | ip_layer | FILL_UDP_HDR | FILL_GTP_HDR;
		break;
	case DOCA_FLOW_TUN_NISP:
	case DOCA_FLOW_TUN_AUDP:
	case DOCA_FLOW_TUN_ESP:
	default:
		DOCA_LOG_ERR("failed to build encap action - unsupported encap tunnel type 0x%x",
			     actions->encap.tun.type);
		return -1;
	}

	if (decap_layer &&
	    (pipe_q->attr.transfer || port->rx_encap)) {
		entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
		dpdk_decap_action_build(entry, cfg, decap_layer);
	}
	entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
	dpdk_encap_action_build(entry, cfg, layer, act_arr_idx);
	entry->modify_action = dpdk_encap_action_modify;
	if (!port->rx_encap && !pipe_q->attr.transfer) {
		dpdk_modify_pipe_encap(dpdk_pipe, entry, actions);
		/*
		 * dpdk_modify_pipe_encap() check tx_encap capability and set
		 * rx_encap to true if doesn't support tx_encap.
		 * 1. hairpinq + tx_encap supported, tx_encap.
		 * 2. other cases, rx_encap.
		 */
		if (port->rx_encap && decap_layer)
			dpdk_decap_encap_actions_build(dpdk_pipe, pipe_q, entry, cfg,
						       decap_layer, layer, act_arr_idx);
	}
	return 0;
}

static int
dpdk_shared_resource_check_binding(enum engine_shared_resource_type type,
				   uint32_t id, struct dpdk_pipe *dpdk_pipe)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);

	if (engine_shared_resource_is_bound(type, id, (struct engine_bindable *)pipe))
		return 0;

	return -EPERM;
}

static int
dpdk_shared_nisp_action_modify(struct dpdk_pipe *dpdk_pipe,
			       struct dpdk_action_entry *entry,
			       const struct doca_flow_actions *pkt_action,
			       __rte_unused const struct doca_flow_monitor *mon,
			       __rte_unused const struct doca_flow_fwd *fwd,
			       __rte_unused struct doca_flow_pipe_entry *flow_entry)
{
	struct rte_flow_action_handle *obj;
	uint32_t id;
	int rc;

	if (!pkt_action)
		return 0;

	id = pkt_action->security.crypto_id;
	rc = dpdk_shared_resource_check_binding(ENGINE_SHARED_RESOURCE_CRYPTO, id, dpdk_pipe);
	if (rc != 0) {
		DOCA_LOG_RATE_LIMIT_WARN("failed to modify shared nisp action - invalid NISP action binding id=%u, rc=%d",
					  id, rc);
		return rc;
	}

	obj = dpdk_shared_crypto_translate(id);
	if (obj == NULL) {
		DOCA_LOG_RATE_LIMIT_WARN("NISP action translation failure (id=%u)", id);
		return -EINVAL;
	}

	assert(entry->action->type == RTE_FLOW_ACTION_TYPE_INDIRECT);
	assert(entry->mask->type == RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_NISP);
	entry->action->conf = obj;
	return 0;
}

static int
dpdk_shared_esp_action_modify(struct dpdk_pipe *dpdk_pipe,
			      struct dpdk_action_entry *entry,
			       const struct doca_flow_actions *pkt_action,
			       __rte_unused const struct doca_flow_monitor *mon,
			       __rte_unused const struct doca_flow_fwd *fwd,
			       __rte_unused struct doca_flow_pipe_entry *flow_entry)
{
	uint32_t id = pkt_action->security.crypto_id;
	struct rte_flow_action_handle *obj;
	int rc;

	rc = dpdk_shared_resource_check_binding(ENGINE_SHARED_RESOURCE_CRYPTO, id, dpdk_pipe);
	if (rc != 0) {
		DOCA_LOG_RATE_LIMIT_WARN("Invalid IPsec action binding (id=%u), %d", id, rc);
		return rc;
	}

	obj = dpdk_shared_crypto_translate(id);
	if (obj == NULL) {
		DOCA_LOG_RATE_LIMIT_WARN("IPsec action translation failure (id=%u)", id);
		return -EINVAL;
	}

	assert(entry->action->type == RTE_FLOW_ACTION_TYPE_INDIRECT);
	assert(entry->mask->type == RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_IPSEC);
	entry->action->conf = obj;
	return 0;
}

static int
is_crypto_proto_valid(enum doca_flow_crypto_protocol_type proto)
{
	switch (proto) {
	case DOCA_FLOW_CRYPTO_PROTOCOL_NISP:
	case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_ENCRYPT:
	case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_DECRYPT:
		return 0;
	default:
		return -EINVAL;
	}
}

static int
dpdk_shared_crypto_action_build(struct dpdk_pipe *dpdk_pipe,
				struct dpdk_pipe_q_ctx *pipe_q,
				const struct doca_flow_pipe_cfg *cfg,
				uint8_t act_arr_idx)
{
	struct dpdk_action_entry *entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
	struct dpdk_action_nisp_data *conf = &entry->action_data.nisp;
	uint32_t id = cfg->actions[act_arr_idx]->security.crypto_id;
	enum doca_flow_crypto_protocol_type proto = cfg->actions[act_arr_idx]->security.proto_type;
	int rc;

	rc = is_crypto_proto_valid(proto);
	if (rc != 0) {
		DOCA_LOG_RATE_LIMIT_ERR("unsupported protocol (proto=%d) for crypto action (id=%u), err=%d",
					proto, id, rc);
		return rc;
	}

	switch (proto) {
	case DOCA_FLOW_CRYPTO_PROTOCOL_NISP:
		entry->mask->type = RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_NISP;
		entry->modify_action = dpdk_shared_nisp_action_modify;
		break;
	case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_ENCRYPT:
	case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_DECRYPT:
		entry->mask->type = RTE_RMD_MLX5_RTE_FLOW_ACTION_TYPE_IPSEC;
		entry->modify_action = dpdk_shared_esp_action_modify;
		break;
	default:
		return -EINVAL;
	}

	if (utils_df_translate_is_shared_resource_id_changeable(id) != 0) {
		conf->id = 0;
		conf->obj = NULL;
	} else {
		rc = dpdk_shared_resource_check_binding(ENGINE_SHARED_RESOURCE_CRYPTO, id, dpdk_pipe);
		if (rc != 0) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to build crypto action - id=%u unbound", id);
			return rc;
		}
		conf->obj = dpdk_shared_crypto_translate(id);
		if (conf->obj == NULL) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to build crypto action - object is NULL (id=%u)", id);
			return -EINVAL;
		}
		conf->id = id;
		entry->modify_action = NULL;
	}

	entry->action->type = RTE_FLOW_ACTION_TYPE_INDIRECT;
	entry->action->conf = conf->obj;
	entry->mask->conf = conf->obj;
	return 0;
}

static int
dpdk_action_drop_build(struct dpdk_action_entry *entry)
{
	struct rte_flow_action *action = entry->action;

	action->type = RTE_FLOW_ACTION_TYPE_DROP;

	return 1;
}

static int
dpdk_entry_actions_rss_cb(__rte_unused struct dpdk_pipe *dpdk_pipe,
			  struct dpdk_action_entry *action_entry,
			  __rte_unused const struct doca_flow_actions *pkt_action,
			  __rte_unused const struct doca_flow_monitor *mon,
			  const struct doca_flow_fwd *fwd,
			  __rte_unused struct doca_flow_pipe_entry *entry)
{
	struct dpdk_action_rss_data *rss;
	uint16_t qidx;

	if (unlikely(fwd == NULL))
		return -EINVAL;

	rss = &action_entry->action_data.rss;
	rss->conf.queue_num = fwd->num_of_queues;
	for (qidx = 0; qidx < fwd->num_of_queues; qidx++)
		rss->queue[qidx] = fwd->rss_queues[qidx];
	if (fwd->rss_flags)
		rss->conf.types = dpdk_pipe_actions_get_rss_type(fwd->rss_flags);
	return 0;
}

static int
dpdk_action_rss_build(struct dpdk_pipe *dpdk_pipe,
		      struct dpdk_pipe_q_ctx *pipe_q,
		      struct dpdk_action_entry *entry,
		      const struct doca_flow_fwd *fwd)
{
	struct dpdk_action_rss_data *rss;

	if (dpdk_pipe)
		dpdk_engine_pipe_q_ctx_group_set(DOCA_PIPE(dpdk_pipe), pipe_q, fwd, false);

	entry->action->type = RTE_FLOW_ACTION_TYPE_RSS;
	rss = &entry->action_data.rss;
	entry->action->conf = &rss->conf;
	rss->conf.func = RTE_ETH_HASH_FUNCTION_DEFAULT;
	rss->conf.types = dpdk_pipe_actions_get_rss_type(fwd->rss_flags);
	rss->conf.queue = rss->queue;
	if (utils_df_translate_is_fwd_changeable(fwd))
		entry->modify_action = dpdk_entry_actions_rss_cb;
	else
		dpdk_entry_actions_rss_cb(dpdk_pipe, entry, NULL, NULL, fwd, NULL);
	return 0;
}

static int
dpdk_action_jump_build(struct dpdk_action_entry *entry, int group)
{
	struct rte_flow_action *action = entry->action;
	struct rte_flow_action_jump *jump = &entry->action_data.jump.jump;

	jump->group = group;
	action->type = RTE_FLOW_ACTION_TYPE_JUMP;
	action->conf = jump;

	return 1;
}

static int
dpdk_entry_actions_next_pipe_cb(__rte_unused struct dpdk_pipe *dpdk_pipe,
			  struct dpdk_action_entry *action_entry,
			  __rte_unused const struct doca_flow_actions *pkt_action,
			  __rte_unused const struct doca_flow_monitor *mon,
			  const struct doca_flow_fwd *fwd,
			  __rte_unused struct doca_flow_pipe_entry *entry)
{
	uint32_t group;
	int ret;

	if (unlikely(fwd == NULL || fwd->next_pipe == NULL))
		return -EINVAL;

	ret = dpdk_engine_next_pipe_group_id_get(fwd->next_pipe, &group);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"failed building fwd - get group_id failed for next_pipe");
		return ret;
	}

	action_entry->action_data.jump.jump.group = group;
	return 0;
}

static int
dpdk_action_next_pipe_build(struct dpdk_pipe_q_ctx *pipe_q,
			    struct dpdk_action_entry *entry,
			    const struct doca_flow_fwd *fwd)
{
	int ret;

	ret = dpdk_action_jump_build(entry, UINT32_MAX);

	if (fwd->next_pipe == NULL) /* Changeable next pipe */
		entry->modify_action = dpdk_entry_actions_next_pipe_cb;
	else
		dpdk_entry_actions_next_pipe_cb(NULL, entry, NULL, NULL, fwd, NULL);

	pipe_q->default_pipe_jump = entry->action_data.jump.jump;
	return ret;
}

static int
dpdk_entry_actions_lag_port_cb(__rte_unused struct dpdk_pipe *dpdk_pipe,
			  struct dpdk_action_entry *action_entry,
			  __rte_unused const struct doca_flow_actions *pkt_action,
			  __rte_unused const struct doca_flow_monitor *mon,
			  const struct doca_flow_fwd *fwd,
			  __rte_unused struct doca_flow_pipe_entry *entry)
{
	struct dpdk_action_modify_field_data *modi = &action_entry->action_data.modify_field;
	void *data = (void *)modi->conf.src.value;
	uint32_t meta;

	if (unlikely(fwd == NULL))
		return -EINVAL;

	meta = 1 << (fwd->port_id == 0 ? LAG_P0_META_BIT : LAG_P1_META_BIT);
	memcpy(data, &meta, sizeof(meta));
	return 0;
}

static int
dpdk_action_lag_port_modify_build(struct dpdk_pipe *pipe,
				  struct dpdk_pipe_q_ctx *pipe_q,
				  uint8_t act_arr_idx,
				  const struct doca_flow_fwd *fwd)
{
	struct dpdk_action_entry *action_entry;
	struct doca_flow_actions actions = {0};
	struct doca_flow_action_descs descs = {0};
	int ret;

	/* build meta action, set 2,3 bits mask on meta data */
	descs.meta.pkt_meta.type = DOCA_FLOW_ACTION_SET;
	descs.meta.pkt_meta.mask.u32 = (1 << LAG_P0_META_BIT | 1 << LAG_P1_META_BIT);
	ret = dpdk_action_modify_build(pipe, pipe_q, &actions, RTE_FLOW_FIELD_META, 0,
		&actions.meta.pkt_meta, &descs.meta.pkt_meta, act_arr_idx);
	if (ret < 0) {
		DOCA_LOG_ERR("create LAG port %u action error", fwd->port_id);
		return ret;
	}

	/* amend callback */
	action_entry = dpdk_pipe_actions_entry_get_last(pipe, pipe_q, act_arr_idx);
	action_entry->modify_action = dpdk_entry_actions_lag_port_cb;
	return 1;
}

static int
dpdk_entry_actions_port_cb(__rte_unused struct dpdk_pipe *dpdk_pipe,
			  struct dpdk_action_entry *action_entry,
			  __rte_unused const struct doca_flow_actions *pkt_action,
			  __rte_unused const struct doca_flow_monitor *mon,
			  const struct doca_flow_fwd *fwd,
			  __rte_unused struct doca_flow_pipe_entry *entry)
{
	if (unlikely(fwd == NULL))
		return -EINVAL;

	action_entry->action_data.fwd.port_id_conf.port_id = fwd->port_id;
	return 0;
}

static int
dpdk_action_port_build(struct doca_flow_port *port,
		       struct dpdk_action_entry *entry,
		       const struct doca_flow_fwd *fwd)
{
	struct dpdk_action_fwd_data *fwd_data = &entry->action_data.fwd;
	struct rte_flow_action *action = entry->action;

	/*
	 * For VNF mode, verify_input has checked if it's hairpin peer.
	 * For REMOTE_VNF mode, if port id is hairpin, build jump empty
	 * group action, else build port action.
	 * For SWITCH mode, build the port action.
	 */
	if (engine_model_is_mode(ENGINE_MODEL_MODE_VNF)) {/* NIC Domain */
		fwd_data->queue_conf.index = engine_model_get_pipe_queues();
		action->type = RTE_FLOW_ACTION_TYPE_QUEUE;
		action->conf = &fwd_data->queue_conf;
		return 1;
	}
	if (engine_model_is_mode(ENGINE_MODEL_MODE_REMOTE_VNF) &&
		(port->hairpin_peer && port->hairpin_peer->port_id == fwd->port_id)) {
		return dpdk_action_jump_build(entry, EMPTY_GROUP_ID);
	}
	if (utils_df_translate_is_fwd_changeable(fwd))
		entry->modify_action = dpdk_entry_actions_port_cb;
	dpdk_entry_actions_port_cb(NULL, entry, NULL, NULL, fwd, NULL);
	action->type = RTE_FLOW_ACTION_TYPE_REPRESENTED_PORT;
	action->conf = &fwd_data->port_id_conf;
	return 1;
}

static int
dpdk_entry_ordered_list_pipe_modify_cb(struct dpdk_pipe *dpdk_pipe,
				       struct dpdk_action_entry *action_entry,
				       const struct doca_flow_actions *actions,
				       const struct doca_flow_monitor *monitor,
				       const struct doca_flow_fwd *fwd,
				       struct doca_flow_pipe_entry *entry)
{
	return dpdk_pipe_hash_modify_set_index_action(action_entry,
						      fwd->ordered_list_pipe.idx);
}

static int
dpdk_action_ordered_list_pipe_modify_build(struct dpdk_pipe *dpdk_pipe,
					   struct dpdk_pipe_q_ctx *pipe_q,
					   uint8_t act_arr_idx,
					   const struct doca_flow_fwd *fwd)
{
	struct dpdk_action_entry *action_entry;
	uint32_t index = fwd->ordered_list_pipe.idx;
	int ret;

	ret = dpdk_pipe_hash_build_set_index_action(dpdk_pipe, pipe_q, act_arr_idx, index);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to build ordered list pipe fwd - cannot set entry index");
		return ret;
	}

	/* amend callback */
	action_entry = dpdk_pipe_actions_entry_get_last(dpdk_pipe, pipe_q, act_arr_idx);
	action_entry->modify_action = dpdk_entry_ordered_list_pipe_modify_cb;
	return 1;
}

static int
dpdk_action_ordered_list_pipe_jump_build(struct dpdk_action_entry *action_entry,
					 const struct doca_flow_fwd *fwd)
{
	struct doca_flow_pipe *next_pipe = fwd->ordered_list_pipe.pipe;
	uint32_t group;
	int ret;

	ret = dpdk_engine_next_pipe_group_id_get(next_pipe, &group);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to build ordered list pipe fwd - get group_id failed for pipe");
		return ret;
	}
	return dpdk_action_jump_build(action_entry, group);
}

void
dpdk_pipe_actions_build_complete(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
				 uint8_t act_arr_idx)
{
	struct dpdk_action_entry *action_entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);

	if (action_entry)
		action_entry->action->type = RTE_FLOW_ACTION_TYPE_END;
}

static int
dpdk_fwd_type_build(struct dpdk_pipe *dpdk_pipe,
		    struct dpdk_pipe_q_ctx *pipe_q,
		    struct dpdk_action_entry *action_entry,
		    const struct doca_flow_fwd *fwd)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed to build fwd type %u - pipe's port is null", fwd->type);
		return -ENOTRECOVERABLE;
	}

	switch (fwd->type) {
	case DOCA_FLOW_FWD_RSS:
		return dpdk_action_rss_build(dpdk_pipe, pipe_q, action_entry, fwd);
	case DOCA_FLOW_FWD_PORT:
		if (dpdk_pipe->fwd_port_type == FWD_LAG_PORT)
			return dpdk_action_jump_build(action_entry, EMPTY_GROUP_ID);
		return dpdk_action_port_build(port, action_entry, fwd);
	case DOCA_FLOW_FWD_PIPE:
		return dpdk_action_next_pipe_build(pipe_q, action_entry, fwd);
	case DOCA_FLOW_FWD_DROP:
		return dpdk_action_drop_build(action_entry);
	case DOCA_FLOW_FWD_ORDERED_LIST_PIPE:
		return dpdk_action_ordered_list_pipe_jump_build(action_entry, fwd);
	default:
		DOCA_LOG_RATE_LIMIT_ERR("failed to build fwd type - unsupported forward type %u", fwd->type);
		return -1;
	}
	return 0;
}

static bool
no_need_fwd_build(struct dpdk_pipe_q_ctx *pipe_q)
{
	bool has_fwd = is_pipe_has_fwd(pipe_q);
	bool is_fwd_changeable = dpdk_pipe_actions_is_fwd_changeable(pipe_q);
	bool is_fate_meter = is_pipe_has_meter(pipe_q) &&
			     dpdk_meter_is_policy_type();

	return ((has_fwd && !is_fwd_changeable) || is_fate_meter);
}

void
dpdk_pipe_set_fwd_lag_port_type(struct dpdk_pipe *dpdk_pipe,
				const struct doca_flow_fwd *fwd,
				const struct doca_flow_fwd *fwd_miss)
{
	/*
	 * If fwd type is PORT, fwd miss type is REPRESENTOR_PORT,
	 * set this fwd type with LAG_PORT
	 */
	if (fwd && fwd->type == DOCA_FLOW_FWD_PORT) {
		if (fwd_miss && fwd_miss->type == DOCA_FLOW_FWD_PIPE) {
			if (fwd_miss->next_pipe->dpdk_pipe.fwd_port_type == FWD_REPRESENTOR_PORT) {
				dpdk_pipe->fwd_port_type = FWD_LAG_PORT;
				return;
			}
		}
		dpdk_pipe->fwd_port_type = FWD_REPRESENTOR_PORT;
	}
}

static void
build_queue_for_groups_req(struct dpdk_fwd_groups_req *cfg)
{
	cfg->key.queue.index = engine_model_get_pipe_queues();
	cfg->key.type = DPDK_FWD_QUEUE;
}

static void
build_port_for_groups_req(struct dpdk_fwd_groups_req *cfg, const struct doca_flow_fwd *fwd)
{
	cfg->key.port.id = fwd->port_id;
	cfg->key.type = DPDK_FWD_PORT;
}

static bool
is_hairpin_config(struct doca_flow_port *port,
		  const struct doca_flow_fwd *fwd)
{
	if (port->hairpin_peer && port->hairpin_peer->port_id == fwd->port_id)
		return true;

	return false;
}

static void
build_fwd_port_req(struct doca_flow_port *port,
		   const struct doca_flow_fwd *fwd,
		   struct dpdk_fwd_groups_req *cfg)
{
	/*
	 * For VNF mode, verify_input has checked if it's hairpin peer.
	 * For REMOTE_VNF mode, if port id is hairpin, build jump empty
	 * group action, else build port action.
	 * For SWITCH/HYBRID mode, if hairpin configured, build queue else build the port action.
	 */

	switch (engine_model_get_mode()) {
	case ENGINE_MODEL_MODE_VNF:
		build_queue_for_groups_req(cfg);
		break;
	case ENGINE_MODEL_MODE_REMOTE_VNF:
		if (is_hairpin_config(port, fwd)) {
			cfg->key.type = DPDK_FWD_JUMP;
			cfg->key.jump.group = EMPTY_GROUP_ID;
		} else
			build_port_for_groups_req(cfg, fwd);
		break;
	case ENGINE_MODEL_MODE_SWITCH:
	case ENGINE_MODEL_MODE_HYBRID:
		if (is_hairpin_config(port, fwd))
			build_queue_for_groups_req(cfg);
		else
			build_port_for_groups_req(cfg, fwd);
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("Not supported engine mode %d", engine_model_get_mode());
		break;
	}
}

static int
build_fwd_pipe_req(struct doca_flow_port *port,
		   const struct doca_flow_fwd *fwd,
		   struct dpdk_fwd_groups_req *cfg)
{
	uint32_t group_id;
	int ret;

	ret = dpdk_engine_next_pipe_group_id_get(fwd->next_pipe, &group_id);
	if (ret)
		return ret;

	cfg->key.type = DPDK_FWD_JUMP;
	cfg->key.jump.group = group_id;
	return 0;
}

int
dpdk_fwd_groups_req_build(struct doca_flow_port *port,
			  const struct doca_flow_fwd *fwd,
			  struct dpdk_fwd_groups **fwd_groups,
			  struct dpdk_fwd_groups_req *req)
{
	struct dpdk_fwd_groups *local_groups;

	local_groups = dpdk_port_fwd_groups_get(port->dpdk_port);
	if (unlikely(!local_groups)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building fwd groups req - failed to get fwd groups of port");
		return -EINVAL;
	}

	*fwd_groups = local_groups;
	req->key.port_id = port->port_id;
	dpdk_engine_flow_attr_set(port, req, fwd);

	switch (fwd->type) {
	case DOCA_FLOW_FWD_RSS:
		req->key.type = DPDK_FWD_RSS;
		req->key.rss.shared_rss_id = fwd->shared_rss_id;
		break;
	case DOCA_FLOW_FWD_PORT:
		build_fwd_port_req(port, fwd, req);
		break;
	case DOCA_FLOW_FWD_PIPE:
		build_fwd_pipe_req(port, fwd, req);
		break;
	case DOCA_FLOW_FWD_DROP:
		req->key.type = DPDK_FWD_DROP;
		break;
	default:
		DOCA_LOG_RATE_LIMIT_ERR("failed building fwd groups req - unsupported fwd type %d", fwd->type);
		return -EINVAL;
	}

	return 0;
}

static int
pipe_null_fwd_override_by_entry(struct dpdk_pipe *dpdk_pipe,
		     struct dpdk_action_entry *action_entry,
		     const struct doca_flow_fwd *fwd,
		     struct doca_flow_pipe_entry *flow_entry)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct dpdk_pipe_q_ctx *pipe_q;
	int ret;

	pipe_q = dpdk_pipe_q_ctx_get(pipe, flow_entry->pipe_queue);
	ret = dpdk_fwd_type_build(dpdk_pipe, pipe_q, action_entry, fwd);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to build fwd type %u ", fwd->type);
		return ret;
	}

	return 0;
}

static int
pipe_null_fwd_cb(struct dpdk_pipe *dpdk_pipe,
		 struct dpdk_action_entry *action_entry,
		 __rte_unused const struct doca_flow_actions *pkt_action,
		 __rte_unused const struct doca_flow_monitor *mon,
		 const struct doca_flow_fwd *fwd,
		 struct doca_flow_pipe_entry *flow_entry)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct dpdk_fwd_groups_req req = {0};
	struct dpdk_fwd_groups *fwd_groups;
	struct doca_flow_port *port;
	uint32_t group_id;
	int ret;
	int mode = dpdk_engine_mode_get();

	if (mode == DPDK_MODE_SWS)
		return pipe_null_fwd_override_by_entry(dpdk_pipe, action_entry, fwd, flow_entry);

	if (mode != DPDK_MODE_HWS)
		return 0;

	port = pipe->port;
	ret = dpdk_fwd_groups_req_build(port, fwd, &fwd_groups, &req);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building entry jump action - build fwd groups req failure");
		return ret;
	}

	ret = dpdk_fwd_groups_get(fwd_groups, &req, &group_id);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("failed building entry jump action - get fwd groups failure");
		return ret;
	}

	dpdk_action_jump_build(action_entry, group_id);

	return 0;
}

static int
dpdk_pipe_actions_fwd_build(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
			    uint8_t act_arr_idx,
			    const struct doca_flow_fwd *fwd)
{
	int ret;
	struct dpdk_action_entry *action_entry = NULL;

	if (no_need_fwd_build(pipe_q))
		return 0;

	if (fwd->type == DOCA_FLOW_FWD_PIPE || utils_df_translate_is_fwd_changeable(fwd))
		pipe_q->fwd_action_index = dpdk_pipe->nb_actions_pipe[act_arr_idx];

	if (!(fwd->type == DOCA_FLOW_FWD_PORT && dpdk_pipe->fwd_port_type == FWD_LAG_PORT)) {
		action_entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
		if (action_entry == NULL)
			return -ENOBUFS;
	}

	ret = dpdk_fwd_type_build(dpdk_pipe, pipe_q, action_entry, fwd);
	if (ret < 0) {
		DOCA_LOG_ERR("failed to build fwd type %u ", fwd->type);
		return ret;
	}

	return 0;
}

static int
dpdk_meter_fwd_action_build(struct dpdk_pipe *dpdk_pipe,
			    struct dpdk_action_meter_actions *data,
			    const struct doca_flow_fwd *fwd)
{
	int idx;
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed building meter fwd action with type %u - pipe's port is null",
			      fwd->type);
		return -ENOTRECOVERABLE;
	}

	memset(data, 0x0, sizeof(*data));
	for (idx = 0; idx < MAX_METER_ACTION; idx++)
		data->g_entry[idx].action = &data->g_actions[idx];

	/* G actions */
	if (fwd->type == DOCA_FLOW_FWD_PORT)
		dpdk_action_port_build(doca_flow_layer_get_pipe_port(pipe), &data->g_entry[0], fwd);
	else if (fwd->type == DOCA_FLOW_FWD_RSS)
		dpdk_action_rss_build(NULL, NULL, &data->g_entry[0], fwd);
	else
		return -1;

	/* R actions */
	data->r_actions[0].type = RTE_FLOW_ACTION_TYPE_DROP;
	data->r_actions[0].conf = NULL;
	data->r_actions[1].type = RTE_FLOW_ACTION_TYPE_END;
	data->r_actions[1].conf = NULL;
	return 0;
}

static int
dpdk_meter_policy_create(uint16_t port_id,
			 struct dpdk_action_meter_actions *actions,
			 uint32_t policy_id)
{
	struct rte_mtr_meter_policy_params params;
	struct rte_mtr_error error;
	int ret = 0;
	uint32_t dpdk_policy_id = DPDK_ID(policy_id);

	params.actions[0] = &actions->g_actions[0];
	params.actions[1] = NULL;
	params.actions[2] = &actions->r_actions[0];

	ret = rte_mtr_meter_policy_add(port_id, dpdk_policy_id, &params, &error);
	if (ret) {
		DOCA_DLOG_ERR(
		    "Port %u create policy idx(%d) error(%d) message: %s",
		    port_id, dpdk_policy_id, error.type,
		    error.message ? error.message : "(no stated reason)");
		return -1;
	}
	return 0;
}

static int
__dpdk_meter_policy_build(struct doca_flow_port *port, uint16_t pipe_queue,
			  struct dpdk_pipe *dpdk_pipe, const struct doca_flow_fwd *fwd)
{
	struct dpdk_action_meter_actions actions;
	uint32_t *policy_id;
	int ret;

	if (fwd == NULL || (fwd->type != DOCA_FLOW_FWD_RSS && fwd->type != DOCA_FLOW_FWD_PORT)) {
		DOCA_LOG_ERR("pipe must with rss, CT or hairpin fwd.");
		return -1;
	}
	policy_id = id_pool_alloc(port->meter_policy_pool, pipe_queue);
	if (!policy_id) {
		DOCA_DLOG_ERR("out of meter profile ids");
		return -1;
	}
	ret = dpdk_meter_fwd_action_build(dpdk_pipe, &actions, fwd);
	if (ret < 0)
		goto err;
	ret = dpdk_meter_policy_create(port->port_id, &actions, *policy_id);
	if (ret < 0)
		goto err;
	dpdk_pipe->policy_id = policy_id;
	DOCA_DLOG_DBG("create meter policy id:%d success", *policy_id);
	return 0;
err:
	if (policy_id)
		id_pool_free(port->meter_policy_pool, pipe_queue,
			     policy_id);
	return -1;
}

static int
_dpdk_action_meter_mark_modify(struct dpdk_action_entry *action_entry,
			uint16_t pipe_queue, struct dpdk_pipe *dpdk_pipe,
			struct doca_flow_pipe_entry *pipe_entry,
			const struct doca_flow_monitor *mon)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;
	uint32_t profile_id = 0;
	int ret;
	void *conf;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed modify meter action - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	struct dpdk_pipe_q_ctx *pipe_q = &dpdk_pipe->queues[pipe_queue].pipe_q;

	ret = dpdk_meter_profile_id_get(port, pipe_q, mon, &profile_id);
	if (ret < 0)
		return -ENOMEM; /* Negative error number */

	pipe_entry->profile_id = profile_id;
	conf = &action_entry->action_data;
	dpdk_meter_data_build(action_entry->action, conf,
				     0, DPDK_ID(profile_id), port->port_id);

	return 0;
}

static int
_dpdk_action_meter_policy_modify(struct dpdk_action_entry *action_entry,
				 uint16_t pipe_queue, struct dpdk_pipe *dpdk_pipe,
				 struct doca_flow_pipe_entry *pipe_entry,
				 const struct doca_flow_monitor *mon)
{
	struct dpdk_pipe_q_ctx *pipe_q = &dpdk_pipe->queues[pipe_queue].pipe_q;
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;
	uint32_t *meter_id = NULL;
	uint32_t profile_id = 0;
	uint32_t action_mtr_id = 0;
	int ret;
	void *conf;
	uint32_t dpdk_profile_id;
	uint32_t dpdk_policy_id;
	uint32_t dpdk_mtr_id;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed modify meter action - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	ret = dpdk_meter_profile_id_get(port, pipe_q, mon, &profile_id);
	if (ret < 0)
		goto err;
	meter_id = id_pool_alloc(port->meter_pool, pipe_queue);
	if (!meter_id) {
		DOCA_LOG_RATE_LIMIT_ERR("out of meter ids");
		goto err;
	}

	dpdk_profile_id = DPDK_ID(profile_id);
	dpdk_policy_id = DPDK_ID(*dpdk_pipe->policy_id);
	dpdk_mtr_id = DPDK_ID(*meter_id);
	ret = dpdk_meter_contrl_policy_rule_create(port->port_id, dpdk_policy_id,
						   dpdk_mtr_id, dpdk_profile_id);
	if (ret < 0)
		goto free_meter_id;

	action_mtr_id = DPDK_ID(*meter_id);
	pipe_entry->meter_id = meter_id;
	pipe_entry->profile_id = profile_id;
	conf = &action_entry->action_data;
	dpdk_meter_data_build(action_entry->action, conf,
				     action_mtr_id, 0, port->port_id);

	return 0;

free_meter_id:
	id_pool_free(port->meter_pool, pipe_queue, meter_id);
err:
	return -ENOMEM; /* Negative error number */
}

static int
dpdk_action_meter_modify(struct dpdk_action_entry *action_entry,
			 uint16_t pipe_queue, struct dpdk_pipe *dpdk_pipe,
			 struct doca_flow_pipe_entry *pipe_entry,
			 const struct doca_flow_monitor *mon)
{
	if (dpdk_meter_is_mark_type())
		return _dpdk_action_meter_mark_modify(action_entry, pipe_queue,
							dpdk_pipe, pipe_entry, mon);
	else if (dpdk_meter_is_policy_type())
		return _dpdk_action_meter_policy_modify(action_entry, pipe_queue,
							dpdk_pipe, pipe_entry, mon);
	return -ENOENT;
}

static int
__dpdk_action_meter_build(uint16_t port_id,
			  struct dpdk_pipe *dpdk_pipe,
			  struct dpdk_pipe_q_ctx *pipe_q,
			  uint8_t act_arr_idx,
			  const struct doca_flow_monitor *mon)
{
	struct dpdk_action_entry *action_entry;
	void *conf;

	action_entry = dpdk_pipe_actions_entry_get(pipe_q, act_arr_idx,
			dpdk_pipe->nb_actions_pipe[act_arr_idx]++);
	conf = &action_entry->action_data;
	dpdk_meter_data_build(action_entry->action, conf,
			      0, PROFILE_ID_CHANGEABLE, port_id);
	action_entry->modify_action = dpdk_pipe_actions_monitor_cb;

	return 0;
}

static struct rte_flow_action_count mask_count_conf = {
	.shared = 0x1,
	.reserved = 0x0,
	.id = 0xFF,
};

static struct rte_flow_action_meter_mark mask_meter_mark_conf = {
	.init_color = 1,
};

static void
dpdk_action_shared_counter_modify(struct dpdk_action_entry *action_entry,
				  uint32_t shared_counter_id)
{
	void *indirect_conf = dpdk_shared_counter_get_conf(shared_counter_id);

	action_entry->action->conf = indirect_conf;
}

static int
dpdk_action_shared_resource_build(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
				  uint8_t act_idx,
				  uint32_t shared_resource_id,
				  enum rte_flow_action_type mask_type)
{
	struct dpdk_action_entry *entry;
	void *indirect_conf;
	void *conf;
	pipe_entry_action_cb cb;

	bool is_shared_resource_changeable =
		utils_df_translate_is_shared_resource_id_changeable(shared_resource_id);

	if (is_shared_resource_changeable) {
		indirect_conf = NULL;
		conf = NULL;
		cb = dpdk_pipe_actions_monitor_cb;
	} else {
		switch (mask_type) {
		case RTE_FLOW_ACTION_TYPE_COUNT:
			indirect_conf = dpdk_shared_counter_get_conf(shared_resource_id);
			conf = &mask_count_conf;
			break;

		case RTE_FLOW_ACTION_TYPE_METER_MARK:
			indirect_conf = dpdk_shared_meter_get_conf(shared_resource_id);
			conf = &mask_meter_mark_conf;
			break;

		default:
			return -EINVAL;
		}

		cb = NULL;
	}

	struct dpdk_pipe_actions *act_ctx =
		dpdk_pipe_actions_ctx_get(pipe_q, act_idx);

	entry =  &act_ctx->action_entry[dpdk_pipe->nb_actions_pipe[act_idx]];
	entry->action->type = RTE_FLOW_ACTION_TYPE_INDIRECT;
	entry->action->conf = indirect_conf;
	entry->mask->type = mask_type;
	entry->mask->conf = conf;
	entry->modify_action = cb;
	dpdk_pipe->nb_actions_pipe[act_idx]++;

	return 0;
}

static void
dpdk_action_counter_build(struct dpdk_pipe *dpdk_pipe,
			  struct dpdk_pipe_q_ctx *pipe_q,
			  int act_arr_idx)
{
	struct dpdk_action_entry *entry;

	pipe_q->actions_flags |= DPDK_PIPE_COUNT;
	entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
	entry->modify_action = NULL;
	entry->action->type = RTE_FLOW_ACTION_TYPE_COUNT;
	entry->action->conf = NULL;
}

static int
dpdk_action_meter_build(uint16_t pipe_queue, struct dpdk_pipe *dpdk_pipe,
			struct dpdk_pipe_q_ctx *pipe_q,
			uint8_t act_arr_idx,
			const struct doca_flow_monitor *mon,
			const struct doca_flow_fwd *fwd)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;
	int ret;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed to build meter action - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	if (dpdk_meter_is_policy_type()) {
		if (!is_pipe_policy_initialized(dpdk_pipe)) {
			ret = __dpdk_meter_policy_build(port, pipe_queue, dpdk_pipe, fwd);
			if (ret) {
				DOCA_LOG_ERR("failure to build meter policy ret=%d", ret);
				return ret;
			}
		}
	}

	pipe_q->actions_flags |= DPDK_PIPE_METER;
	ret = __dpdk_action_meter_build(port->port_id, dpdk_pipe, pipe_q, act_arr_idx, mon);
	if (ret)
		DOCA_LOG_ERR("failure to build meter action ret=%d", ret);

	return ret;
}

static void
dpdk_action_shared_meter_modify(struct dpdk_action_entry *action_entry,
				uint32_t shared_meter_id)
{
	void *indirect_conf = dpdk_shared_meter_get_conf(shared_meter_id);

	action_entry->action->conf = indirect_conf;
}

static int
dpdk_monitor_resource_build_type(uint8_t mon_flags,
				 uint8_t requested_flags,
				 uint32_t shared_resource_id)
{
	if (mon_flags & requested_flags)
		return MON_RESOURCE_NON_SHARED_BUILD;

	bool shared_resource_ignored =
		utils_df_translate_is_shared_resource_id_ignored(shared_resource_id);

	if (!shared_resource_ignored)
		return MON_RESOURCE_SHARED_BUILD;

	return MON_RESOURCE_NOTHING_BUILD;
}

bool
dpdk_pipe_actions_is_fwd_changeable(struct dpdk_pipe_q_ctx *pipe_q)
{
	return is_pipe_fwd_changeable(pipe_q);
}

static int
dpdk_pipe_monitor_counter_build(struct dpdk_pipe *dpdk_pipe,
				struct dpdk_pipe_q_ctx *pipe_q,
				uint8_t act_arr_idx,
				uint16_t pipe_queue,
				const struct doca_flow_monitor *mon,
				const struct doca_flow_fwd *fwd)
{
	int rslt;
	uint32_t shared_counter_id = mon->shared_counter_id;

	rslt = dpdk_monitor_resource_build_type(mon->flags,
		(DOCA_FLOW_MONITOR_AGING | DOCA_FLOW_MONITOR_COUNT),
		shared_counter_id);

	switch (rslt) {
	case MON_RESOURCE_NON_SHARED_BUILD:
		dpdk_action_counter_build(dpdk_pipe, pipe_q, act_arr_idx);
		break;

	case MON_RESOURCE_SHARED_BUILD:
		dpdk_action_shared_resource_build(dpdk_pipe, pipe_q, act_arr_idx,
						  mon->shared_counter_id, RTE_FLOW_ACTION_TYPE_COUNT);
		break;
	}

	/* save aging info in pipe. */
	if (mon->flags & DOCA_FLOW_MONITOR_AGING) {
		pipe_q->actions_flags |= DPDK_PIPE_AGE;
		pipe_q->aging = mon->aging;
		pipe_q->user_data = mon->user_data;
	}

	return 0;
}

static int
dpdk_pipe_monitor_meter_build(struct dpdk_pipe *dpdk_pipe,
			      struct dpdk_pipe_q_ctx *pipe_q,
			      uint8_t act_arr_idx,
			      uint16_t pipe_queue,
			      const struct doca_flow_monitor *mon,
			      const struct doca_flow_fwd *fwd,
			      struct doca_flow_error *error)
{
	int ret, rslt;
	uint32_t shared_meter_id = mon->shared_meter_id;

	rslt = dpdk_monitor_resource_build_type(mon->flags, DOCA_FLOW_MONITOR_METER,
					   shared_meter_id);

	switch (rslt) {
	case MON_RESOURCE_NON_SHARED_BUILD:
		ret = dpdk_action_meter_build(pipe_queue, dpdk_pipe, pipe_q, act_arr_idx, mon, fwd);
		if (ret)
			return ret;

		if (dpdk_meter_is_mark_type())
			return dpdk_meter_pipe_create(DOCA_PIPE(dpdk_pipe), pipe_queue, mon, fwd);
		break;
	case MON_RESOURCE_SHARED_BUILD:
		ret = dpdk_action_shared_resource_build(dpdk_pipe, pipe_q, act_arr_idx, shared_meter_id,
							RTE_FLOW_ACTION_TYPE_METER_MARK);
		if (ret)
			return ret;

		if (dpdk_meter_is_mark_type())
			return dpdk_meter_pipe_create(DOCA_PIPE(dpdk_pipe), pipe_queue, mon, fwd);
		break;
	}
	return 0;
}

static __rte_always_inline bool
is_shared_counter(struct dpdk_action_entry *action_entry)
{
	return ((action_entry->action->type == RTE_FLOW_ACTION_TYPE_INDIRECT) &&
	    (action_entry->mask->type == RTE_FLOW_ACTION_TYPE_COUNT));
}

static __rte_always_inline bool
is_shared_meter_mark(struct dpdk_action_entry *action_entry)
{
	return ((action_entry->action->type == RTE_FLOW_ACTION_TYPE_INDIRECT) &&
		   (action_entry->mask->type == RTE_FLOW_ACTION_TYPE_METER_MARK));
}

static int
dpdk_pipe_actions_monitor_cb(struct dpdk_pipe *dpdk_pipe,
			     struct dpdk_action_entry *action_entry,
			     __rte_unused const struct doca_flow_actions *pkt_action,
			     const struct doca_flow_monitor *mon,
			     __rte_unused const struct doca_flow_fwd *fwd,
			     struct doca_flow_pipe_entry *entry)
{
	int rc;

	if (is_shared_counter(action_entry))
		dpdk_action_shared_counter_modify(action_entry, mon->shared_counter_id);
	else if (is_shared_meter_mark(action_entry))
		dpdk_action_shared_meter_modify(action_entry, mon->shared_meter_id);
	else {
		rc = dpdk_action_meter_modify(action_entry, entry->pipe_queue, dpdk_pipe,
					      entry, mon);
		if (rc)
			return rc;
	}

	return 0;
}

static int
dpdk_pipe_modify_build(struct dpdk_pipe *dpdk_pipe,
			struct dpdk_pipe_q_ctx *pipe_q,
			const struct doca_flow_pipe_cfg *cfg,
			uint8_t act_arr_idx)
{
	struct doca_flow_actions *actions = cfg->actions[act_arr_idx];
	static struct doca_flow_action_descs default_descs;
	struct doca_flow_action_descs *descs;
	uint32_t crypto_id;
	int ret = 0;
	int i;

	if (cfg->action_descs && cfg->action_descs[act_arr_idx])
		descs = cfg->action_descs[act_arr_idx];
	else
		descs = &default_descs;
	if (!actions)
		return 0;

	ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_MAC_SRC, 0,
					actions->mod_src_mac, &descs->src_mac, act_arr_idx);
	ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_MAC_DST, 0,
					actions->mod_dst_mac, &descs->dst_mac, act_arr_idx);
	ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_VLAN_ID, 0,
					&actions->mod_vlan_id, &descs->vlan_id, act_arr_idx);
	if (actions->mod_src_ip.type == DOCA_FLOW_IP4_ADDR)
		ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_IPV4_SRC,
						0, &actions->mod_src_ip.ipv4_addr, &descs->src_ip,
						act_arr_idx);
	else if (actions->mod_src_ip.type == DOCA_FLOW_IP6_ADDR)
		ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_IPV6_SRC,
						0, actions->mod_src_ip.ipv6_addr, &descs->src_ip,
						act_arr_idx);
	if (actions->mod_dst_ip.type == DOCA_FLOW_IP4_ADDR)
		ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_IPV4_DST,
						0, &actions->mod_dst_ip.ipv4_addr, &descs->dst_ip,
						act_arr_idx);
	else if (actions->mod_dst_ip.type == DOCA_FLOW_IP6_ADDR)
		ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_IPV6_DST,
						0, actions->mod_dst_ip.ipv6_addr, &descs->dst_ip,
						act_arr_idx);
	if (pipe_q->l4_type == DOCA_PROTO_UDP) {
		ret |= dpdk_action_modify_build(
			dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_UDP_PORT_SRC, 0,
			&actions->mod_src_port, &descs->src_port, act_arr_idx);
		ret |= dpdk_action_modify_build(
			dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_UDP_PORT_DST, 0,
			&actions->mod_dst_port, &descs->dst_port, act_arr_idx);
	} else if (pipe_q->l4_type == DOCA_PROTO_TCP) {
		ret |= dpdk_action_modify_build(
			dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_TCP_PORT_SRC, 0,
			&actions->mod_src_port, &descs->src_port, act_arr_idx);
		ret |= dpdk_action_modify_build(
			dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_TCP_PORT_DST, 0,
			&actions->mod_dst_port, &descs->dst_port, act_arr_idx);
	}
	if (!ret)
		ret |= dpdk_action_modify_build(dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_META, 0,
						&actions->meta.pkt_meta, &descs->meta.pkt_meta,
						act_arr_idx);
	for (i = 0; i < RTE_DIM(actions->meta.u32); i++)
		if (!ret)
			ret |= dpdk_action_modify_build(
				dpdk_pipe, pipe_q, actions, RTE_FLOW_FIELD_TAG, i,
				&actions->meta.u32[i], &descs->meta.u32[i], act_arr_idx);
	crypto_id = cfg->actions[act_arr_idx]->security.crypto_id;
	if ((cfg->actions[act_arr_idx]->security.proto_type != DOCA_FLOW_CRYPTO_PROTOCOL_NONE) &&
	    !utils_df_translate_is_shared_resource_id_ignored(crypto_id)) {
		ret = dpdk_shared_crypto_action_build(dpdk_pipe, pipe_q, cfg, act_arr_idx);
		if (ret != 0) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to build pipe actions - crypto actions, err=%d", ret);
			return ret;
		}
	}

	return ret;
}

static int
dpdk_pipe_null_fwd_build(struct dpdk_pipe *dpdk_pipe,
		       struct dpdk_pipe_q_ctx *pipe_q,
		       uint8_t act_arr_idx)
{
	struct dpdk_action_entry *action_entry;
	struct rte_flow_action *action;

	action_entry = dpdk_pipe_actions_entry_get_next(dpdk_pipe, pipe_q, act_arr_idx);
	action = action_entry->action;
	action->type = RTE_FLOW_ACTION_TYPE_JUMP;
	action->conf = NULL;
	action_entry->modify_action = pipe_null_fwd_cb;

	return 0;
}

#define CHK(item, x) \
	do { \
		ret = (x); \
		if (ret < 0) { \
			DOCA_LOG_ERR("Failed to build " #item); \
			return ret; \
		} \
	} while (0)


/* build an action set */
static int
dpdk_pipe_action_set_build(struct dpdk_pipe *dpdk_pipe,
			   struct dpdk_pipe_q_ctx *pipe_q,
			   int act_arr_idx,
			   struct doca_flow_pipe_cfg *cfg,
			   const struct doca_flow_fwd *fwd)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_actions *actions = cfg->actions ? cfg->actions[act_arr_idx] : NULL;
	struct doca_flow_fwd new_fwd = {0};
	bool is_new_fwd;
	int ret;

	/* reset number of actions */
	dpdk_pipe->nb_actions_pipe[act_arr_idx] = 0;

	/* TAG(MARK) */
	if (actions && actions->meta.mark)
		CHK("aciton mark", dpdk_mark_action_build(dpdk_pipe, pipe_q, cfg, act_arr_idx, actions));

	/* DECAP */
	if (actions && actions->decap)
		CHK("action decap", dpdk_decap_build(dpdk_pipe, pipe_q, cfg, act_arr_idx));

	/* COUNTER */
	if (cfg->monitor)
		CHK("monitor counter", dpdk_pipe_monitor_counter_build(dpdk_pipe, pipe_q, act_arr_idx,
				dpdk_engine_mgmt_id_get(), cfg->monitor, fwd));
	/* METER */
	if (cfg->monitor) {
		CHK("montiro meter", dpdk_pipe_monitor_meter_build(dpdk_pipe, pipe_q, act_arr_idx,
				dpdk_engine_mgmt_id_get(), cfg->monitor, fwd, NULL));
		dpdk_meter_pipe_fwd_update(fwd, DOCA_PIPE(dpdk_pipe), &new_fwd,
					   &is_new_fwd);
		if (is_new_fwd)
			fwd = &new_fwd;
	}

	/* MODIFY */
	if (actions)
		CHK("actions", dpdk_pipe_modify_build(dpdk_pipe, pipe_q, cfg, act_arr_idx));
	if (cfg->action_descs)
		CHK("action descriptions", dpdk_action_descs_build(dpdk_pipe, pipe_q, act_arr_idx, cfg));
	if (fwd && dpdk_pipe->fwd_port_type == FWD_LAG_PORT) {
		CHK("fwd lag port", dpdk_action_lag_port_modify_build(dpdk_pipe, pipe_q, act_arr_idx, fwd));
		return 0;
	}
	if (fwd && fwd->type == DOCA_FLOW_FWD_ORDERED_LIST_PIPE)
		CHK("fwd ordered list pipe set index",
		    dpdk_action_ordered_list_pipe_modify_build(dpdk_pipe, pipe_q, act_arr_idx, fwd));

	/* ENCAP */
	if (actions && actions->has_encap)
		CHK("action encap", dpdk_encap_build(dpdk_pipe, pipe_q, cfg, act_arr_idx));

	/* FT/MISS/TIR/DROP */
	if (fwd) {
		CHK("fwd", dpdk_pipe_actions_fwd_build(dpdk_pipe, pipe_q, act_arr_idx, fwd));
		pipe_q->fwd_type = fwd->type;
	}
	if (!fwd && pipe->type == ENGINE_PIPE_BASIC &&
	    dpdk_pipe_action_crypto_ignored(actions)) {
		CHK("empty fwd", dpdk_pipe_null_fwd_build(dpdk_pipe, pipe_q, act_arr_idx));
	}

	/* END */
	dpdk_pipe_actions_build_complete(dpdk_pipe, pipe_q, act_arr_idx);

	return 0;
}

/* build all action sets of a pipe */
int
dpdk_pipe_actions_build(enum engine_pipe_type type, struct dpdk_pipe *dpdk_pipe,
			struct dpdk_pipe_q_ctx *pipe, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct doca_flow_pipe_cfg cfg = {0};
	int ret, action_set;

	dpdk_pipe_cfg_convert(&cfg, pipe_cfg, type);
	if (pipe_cfg->uds_cfg.fwd && engine_model_is_mode(ENGINE_MODEL_MODE_HYBRID))
		dpdk_pipe_set_fwd_lag_port_type(dpdk_pipe, pipe_cfg->uds_cfg.fwd, pipe_cfg->uds_cfg.fwd_miss);

	for (action_set = 0; action_set < dpdk_pipe->nb_actions_array_len; action_set++) {
		ret = dpdk_pipe_action_set_build(dpdk_pipe, pipe, action_set, &cfg,
						 pipe_cfg->uds_cfg.fwd);
		if (ret < 0)
			return ret;
	}

	if (pipe_cfg->uds_cfg.fwd)
		pipe->actions_flags |= DPDK_PIPE_FWD;

	return 0;
}

int
dpdk_pipe_entry_actions_update(struct dpdk_pipe *dpdk_pipe,
			       struct dpdk_pipe_q_ctx *pipe_q,
			       const struct doca_flow_actions *actions,
			       const struct doca_flow_monitor *mon,
			       const struct doca_flow_fwd *fwd,
			       struct doca_flow_pipe_entry *entry)
{
	struct dpdk_action_entry *action_entry;
	int idex, ret;
	uint8_t act_arr_idx = actions ? actions->action_idx : 0;
	int nb_actions = dpdk_pipe->nb_actions_pipe[act_arr_idx];

	for (idex = 0; idex < nb_actions; idex++) {
		action_entry = dpdk_pipe_actions_entry_get(pipe_q, act_arr_idx, idex);
		if (action_entry->modify_action == NULL)
			continue;
		ret = action_entry->modify_action(dpdk_pipe, action_entry, actions, mon, fwd, entry);
		if (ret) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to modify pipe DPDK actions on index %d and type 0x%x", idex,
						action_entry->action != NULL ? action_entry->action->type : 0);
			return ret;
		}
	}
	return 0;
}

void
dpdk_pipe_actions_masks_build(const struct dpdk_pipe_q_ctx *pipe,
			      struct rte_flow_action *masks,
			      struct rte_flow_action_modify_field *modis,
			      uint8_t act_arr_idx)
{
	const struct rte_flow_action_modify_field *modify;
	int i;

	/* Construct masks of modify field. */
	for (i = 0; ; ++i) {
		if (pipe->action_ctx[act_arr_idx].masks[i].type != 0) {
			masks[i].type = pipe->action_ctx[act_arr_idx].masks[i].type;
			masks[i].conf = pipe->action_ctx[act_arr_idx].masks[i].conf;
			continue;
		}

		masks[i] = pipe->action_ctx[act_arr_idx].actions[i];
		if (pipe->action_ctx[act_arr_idx].actions[i].type == RTE_FLOW_ACTION_TYPE_END)
			break;
		else if (pipe->action_ctx[act_arr_idx].actions[i].type == RTE_FLOW_ACTION_TYPE_RAW_ENCAP
			&& pipe->action_ctx[act_arr_idx].action_entry[i].action_data.rawencap.is_changeable) {
			masks[i].conf = NULL;
			continue;
		} else if (pipe->action_ctx[act_arr_idx].actions[i].type == RTE_FLOW_ACTION_TYPE_JUMP) {
			masks[i].conf = NULL;
			continue;
		} else if (pipe->action_ctx[act_arr_idx].actions[i].type != RTE_FLOW_ACTION_TYPE_MODIFY_FIELD)
			continue;
		masks[i].conf = &modis[i];
		modify = pipe->action_ctx[act_arr_idx].actions[i].conf;
		modis[i].width = UINT32_MAX;
		modis[i].operation = modify->operation;
		modis[i].src.field = modify->src.field;
		modis[i].dst.field = modify->dst.field;
		modis[i].dst.level = UINT32_MAX;
		modis[i].dst.offset = UINT32_MAX;
		if (modify->src.field != RTE_FLOW_FIELD_POINTER &&
		    modify->src.field != RTE_FLOW_FIELD_VALUE) {
			modis[i].src.level = UINT32_MAX;
			modis[i].src.offset = UINT32_MAX;
		} else if (pipe->action_ctx[act_arr_idx].action_entry[i].modify_action == NULL) {
			memset(modis[i].src.value, 0, sizeof(modis[i].src.value));
		}
	}
}
