/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DPDK_PIPE_ACTIONS_H_
#define DPDK_PIPE_ACTIONS_H_

#include <stdio.h>

#include <rte_flow.h>
#include <rte_vxlan.h>
#include <rte_ethdev.h>
#include <rte_mtr.h>
#include <rte_gre.h>

#include "dpdk_pipe.h"
#include "doca_flow.h"

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_METER_ACTION 5
#define INVALID_ACTION_INDEX (-1)
#define DPDK_PIPE_ACTIONS_MAX_ENCAPS (128)
#define DPDK_MAX_RSS_QUEUES (256)

struct dpdk_pipe;
struct dpdk_pipe_q_ctx;
struct dpdk_action_entry;
struct dpdk_pipe_cfg;
enum engine_pipe_type;
struct dpdk_fwd_groups_req;
struct dpdk_fwd_groups;

struct dpdk_pipe_actions_encap_info {
	int port_id;
	void *handler;
};

enum dpdk_pipe_actions_flags {
	DPDK_PIPE_FWD = (1 << 0),
	DPDK_PIPE_METER = (1 << 1),
	DPDK_PIPE_COUNT = (1 << 2),
	DPDK_PIPE_AGE = (1 << 3),
};
struct dpdk_action_mac_data {
	struct rte_flow_action_set_mac set_mac;
};

struct dpdk_action_jump_data {
	struct rte_flow_action_jump jump;
};

struct dpdk_action_ipv4_addr_data {
	struct rte_flow_action_set_ipv4 ipv4;
};

struct dpdk_action_rss_data {
	struct rte_flow_action_rss conf;
	uint8_t key[40];
	uint16_t queue[DPDK_MAX_RSS_QUEUES];
};

struct dpdk_action_mark_data {
	struct rte_flow_action_mark conf;
};

struct dpdk_action_rawdecap_data {
	struct rte_flow_action_raw_decap conf;
	uint8_t data[128];
	uint16_t idx;
};

struct dpdk_action_meter_data {
	struct rte_flow_action_meter conf;
};

struct dpdk_action_meter_mark_data {
	struct rte_flow_action_meter_mark conf;
};

struct dpdk_action_l4_port_data {
	struct rte_flow_action_set_tp l4port;
};

struct dpdk_action_fwd_data {
	struct rte_flow_action_queue queue_conf;
	struct rte_flow_action_ethdev port_id_conf;
};

struct dpdk_action_modify_field_data {
	uint32_t offset; /* Field byte offset to action. */
	uint32_t width; /* Field byte width. */
	bool auto_val; /* Set value if non-zero */
	struct rte_flow_action_modify_field conf;
};

struct dpdk_action_rawencap_data {
	struct rte_flow_action_raw_encap conf;
	struct dpdk_action_modify_field_data meta;
	uint8_t data[128];
	uint8_t preserve[128];
	uint16_t idx;
	uint8_t layer;
	bool is_changeable;
};

struct dpdk_action_nisp_data {
	struct rte_flow_action_handle *obj;
	uint32_t id;
};

struct dpdk_action_data {
	union {
		struct dpdk_action_jump_data jump;
		struct dpdk_action_rss_data rss;
		struct dpdk_action_rawdecap_data rawdecap;
		struct dpdk_action_rawencap_data rawencap;
		struct dpdk_action_meter_data meter;
		struct dpdk_action_meter_mark_data meter_mark;
		struct dpdk_action_fwd_data fwd;
		struct dpdk_action_mark_data mark;
		struct dpdk_action_nisp_data nisp;
		struct dpdk_action_modify_field_data modify_field;
	};
};

typedef int (*pipe_entry_action_cb)(struct dpdk_pipe *dpdk_pipe,
	struct dpdk_action_entry *action_entry,
	const struct doca_flow_actions *act,
	const struct doca_flow_monitor *mon,
	const struct doca_flow_fwd *fwd,
	struct doca_flow_pipe_entry *flow_entry);

struct dpdk_action_entry {
	struct rte_flow_action *action;
	struct rte_flow_action *mask;
	struct dpdk_action_data action_data;
	pipe_entry_action_cb modify_action;
};

struct dpdk_action_meter_actions {
	struct rte_flow_action g_actions[MAX_METER_ACTION];
	struct rte_flow_action r_actions[MAX_METER_ACTION];
	struct dpdk_action_entry g_entry[MAX_METER_ACTION];
};

struct dpdk_pipe_actions {
	struct rte_flow_action actions[DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_action masks[DPDK_PIPE_MAX_ACTIONS];
	struct dpdk_action_entry action_entry[DPDK_PIPE_MAX_ACTIONS];
};

void
dpdk_pipe_actions_init(struct doca_flow_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q);

struct dpdk_action_entry *
dpdk_pipe_actions_entry_get_next(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
				 uint8_t act_arr_idx);

struct dpdk_action_entry *
dpdk_pipe_actions_entry_get_last(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
				 uint8_t act_arr_idx);

struct rte_flow_action *
dpdk_pipe_actions_get(struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_ctx_idx, uint8_t act_idx);

struct rte_flow_action *
dpdk_pipe_actions_array_get(struct dpdk_pipe_q_ctx *pipe_q, uint8_t act_ctx_idx);

struct dpdk_pipe_actions_encap_info *
dpdk_pipe_actions_encap_info_put(int encap_id);

bool
dpdk_pipe_actions_is_fwd_changeable(struct dpdk_pipe_q_ctx *pipe_q);

int
dpdk_pipe_actions_build(enum engine_pipe_type type, struct dpdk_pipe *dpdk_pipe,
			struct dpdk_pipe_q_ctx *pipe, struct dpdk_pipe_cfg *pipe_cfg);

int
dpdk_pipe_entry_actions_update(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
			       const struct doca_flow_actions *actions,
			       const struct doca_flow_monitor *mon,
			       const struct doca_flow_fwd *fwd,
			       struct doca_flow_pipe_entry *entry);

int
dpdk_fwd_groups_req_build(struct doca_flow_port *port,
			  const struct doca_flow_fwd *fwd,
			  struct dpdk_fwd_groups **fwd_groups,
			  struct dpdk_fwd_groups_req *req);

static __rte_always_inline uint64_t
dpdk_pipe_actions_get_rss_type(uint32_t rss_type)
{
	uint64_t rss_flags = 0;

	if (rss_type & DOCA_FLOW_RSS_IP)
		rss_flags |= ETH_RSS_IP;
	if (rss_type & DOCA_FLOW_RSS_UDP)
		rss_flags |= ETH_RSS_UDP;
	if (rss_type & DOCA_FLOW_RSS_TCP)
		rss_flags |= ETH_RSS_TCP;

	return rss_flags;
}

void
dpdk_pipe_actions_masks_build(const struct dpdk_pipe_q_ctx *pipe,
			      struct rte_flow_action *masks,
			      struct rte_flow_action_modify_field *modis,
			      uint8_t act_arr_idx);

bool
dpdk_pipe_action_crypto_ignored(const struct doca_flow_actions *actions);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DPDK_PIPE_ACTIONS_H_ */
