/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dpdk_pipe_basic.h"
#include "dpdk_flow.h"
#include "dpdk_pipe_common.h"

#include "engine_model.h"
#include "utils_df_translate.h"

#include "dpdk_table_manager.h"
#include "dpdk_port.h"
#include "dpdk_flow_hws_legacy.h"

#include "dpdk_engine.h"
#include "dpdk_meter_pipe.h"
#include "doca_flow_layer.h"
#include "doca_flow_priv.h"
#include "dpdk_dbg.h"
#include "priv_doca_mempool.h"

#include <rte_hash.h>

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_pipe_basic);

#define PIPE_BASIC_NR_TABLES 1
#define PIPE_BASIC_DEFAULT_TABLE_IDX 0
#define MAX_ACT_TEMPLATE 4

static int
pipe_basic_verify(struct dpdk_pipe_cfg *pipe_cfg)
{
	return dpdk_pipe_common_verify(pipe_cfg);
}

static struct doca_flow_pipe *
pipe_basic_alloc(uint16_t nr_queues)
{
	return dpdk_pipe_common_alloc(nr_queues);
}

static int
pipe_basic_build(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct dpdk_table_manager_cfg table_manager_cfg;
	int rc;

	table_manager_cfg.nr_tables = PIPE_BASIC_NR_TABLES;
	pipe_legacy->table_manager = dpdk_table_manager_create(&table_manager_cfg);
	if (!pipe_legacy->table_manager) {
		DOCA_LOG_ERR("failed building basic pipe - table manager is null");
		return -ENOMEM;
	}

	rc = dpdk_pipe_basic_build(pipe_legacy, pipe_cfg);
	if (unlikely(rc)) {
		dpdk_table_manager_destroy(pipe_legacy->table_manager);
		pipe_legacy->table_manager = NULL;
	}
	return rc;
}

static int
pipe_basic_submit(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, 0);
	struct rte_flow_action_modify_field modis[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS] = {};
	struct rte_flow_action action_masks[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_action *actions_mask_arr[MAX_ACT_TEMPLATE];
	struct rte_flow_action *actions_arr[MAX_ACT_TEMPLATE];
	struct doca_flow_pipe_cfg cfg = {0};
	uint32_t basic_table_id;
	struct dpdk_table_params table_params;
	int rc;
	int aidx;

	dpdk_pipe_cfg_convert(&cfg, pipe_cfg, ENGINE_PIPE_BASIC);

	if (unlikely(pipe_cfg->uds_cfg.fwd_miss)) {
		rc = dpdk_pipe_fwd_miss_handle(&cfg, pipe_legacy, pipe_cfg->uds_cfg.fwd_miss);
		if (unlikely(rc))
			return rc;
	}

	rc = dpdk_table_manager_alloc_id(pipe_legacy->table_manager, &basic_table_id);
	if (unlikely(rc))
		return rc;

	pipe_legacy->basic_table = dpdk_table_manager_get_by_id(pipe_legacy->table_manager, basic_table_id);
	if (unlikely(!pipe_legacy->basic_table || basic_table_id != PIPE_BASIC_DEFAULT_TABLE_IDX))
		return -ENOTRECOVERABLE;

	dpdk_dbg_rte_flow_dump("pattern template:\t", pipe_cfg->port->port_id,
			       &pipe_q->attr, pipe_q->items, NULL, stdout);
	for (aidx = 0; aidx < pipe_cfg->nb_actions; aidx++) {
		dpdk_pipe_actions_masks_build(pipe_q, action_masks[aidx], modis[aidx], aidx);
		actions_arr[aidx] = dpdk_pipe_actions_array_get(pipe_q, aidx);
		actions_mask_arr[aidx] = action_masks[aidx];

		dpdk_dbg_rte_flow_dump("action template:\t", pipe_cfg->port->port_id,
				       &pipe_q->attr, NULL, actions_arr[aidx], stdout);
		dpdk_dbg_rte_flow_dump("actions template mask:\t", pipe_cfg->port->port_id,
				       &pipe_q->attr, NULL, actions_mask_arr[aidx], stdout);
	}

	dpdk_pipe_common_fill_table_params(&table_params, pipe_cfg->port->port_id,
					   pipe_q->items, actions_arr, actions_mask_arr, pipe_cfg->nb_actions,
					   &pipe_q->attr, pipe_cfg->nb_flows, DPDK_TABLE_INSERTION_TYPE_PATTERN,
					   DPDK_TABLE_HASH_TYPE_DEFAULT);

	rc = dpdk_table_create(&table_params,
			      dpdk_port_get_table_ctx(pipe_cfg->port->dpdk_port), pipe_legacy->basic_table);
	if (unlikely(rc)) {
		dpdk_table_manager_free_id(pipe_legacy->table_manager, PIPE_BASIC_DEFAULT_TABLE_IDX);
		return rc;
	}

	return 0;
}

static void
pipe_basic_free(struct doca_flow_pipe *pipe_legacy)
{
	dpdk_pipe_entries_flush(doca_flow_layer_get_pipe_port(pipe_legacy), pipe_legacy);
	if (pipe_legacy && pipe_legacy->table_manager) {
		if (likely(pipe_legacy->basic_table)) {
			dpdk_table_destroy(dpdk_port_get_table_ctx(pipe_legacy->port->dpdk_port),
								      pipe_legacy->basic_table);
			dpdk_table_manager_free_id(pipe_legacy->table_manager, PIPE_BASIC_DEFAULT_TABLE_IDX);
		}
		dpdk_table_manager_destroy(pipe_legacy->table_manager);
	}

	dpdk_pipe_common_resources_free(pipe_legacy);
	dpdk_pipe_common_legacy_free(pipe_legacy);
}

static int
pipe_basic_queue_verify(struct doca_flow_pipe *pipe_legacy, uint16_t queue_id, uint16_t action_idx,
			uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	return dpdk_pipe_common_entry_verify(pipe_legacy, queue_id, action_idx, uds_cfg);
}

static int
pipe_basic_queue_modify(struct doca_flow_pipe *pipe_legacy, struct doca_flow_pipe_entry *entry,
			uint16_t queue_id, uint16_t action_idx, uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	int rc;

	dpdk_dbg_pipe_config_dump(uds_cfg->match, uds_cfg->actions[action_idx], NULL,
				  uds_cfg->monitor, uds_cfg->fwd);

	rc = dpdk_pipe_common_queue_modify(ENGINE_PIPE_BASIC, queue_id, &pipe_legacy->dpdk_pipe, pipe_q,
				      entry, uds_cfg->match, uds_cfg->actions[action_idx],
				      uds_cfg->monitor, uds_cfg->fwd, 0, NULL);
	if (unlikely(rc))
		return rc;

	rc = dpdk_flow_age_list_add(queue_id, pipe_q, entry, uds_cfg->monitor);
	if (unlikely(rc))
		return rc;

	entry->table_id = PIPE_BASIC_DEFAULT_TABLE_IDX;
	return 0;
}

static int
pipe_basic_queue_submit(struct doca_flow_pipe *pipe_legacy,
			struct doca_flow_pipe_entry *entry, uint16_t queue_id,
			uint16_t action_idx, dpdk_flow_completion_cb entry_completion_cb,
			void *usr_ctx, bool no_wait)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	struct dpdk_flow_params flow_params = {0};

	dpdk_pipe_common_fill_flow_params(&flow_params, pipe_q->items,
					  dpdk_pipe_actions_array_get(pipe_q, action_idx),
					  pipe_legacy->basic_table, &pipe_q->attr, action_idx);

	dpdk_pipe_common_entry_attach(queue_id, pipe_legacy, entry);
	return dpdk_pipe_common_queue_submit(pipe_legacy, entry, queue_id, &flow_params,
					     entry_completion_cb, usr_ctx, no_wait);
}

static int
pipe_basic_queue_remove(struct doca_flow_pipe_entry *entry,
			dpdk_flow_completion_cb cb, void *usr_ctx)
{
	return dpdk_pipe_common_queue_remove(entry, cb, usr_ctx);
}

static int
pipe_basic_queue_cleanup(struct doca_flow_pipe_entry *entry)
{
	dpdk_pipe_entry_release(entry);
	return 0;
}

static struct dpdk_pipe_type_ops pipe_ops = {
	.pipe_legacy_verify = pipe_basic_verify,
	.pipe_legacy_alloc = pipe_basic_alloc,
	.pipe_legacy_build = pipe_basic_build,
	.pipe_legacy_submit = pipe_basic_submit,
	.pipe_legacy_free = pipe_basic_free,

	.pipe_queue_verify = pipe_basic_queue_verify,
	.pipe_queue_modify = pipe_basic_queue_modify,
	.pipe_queue_submit = pipe_basic_queue_submit,
	.pipe_queue_remove = pipe_basic_queue_remove,
	.pipe_queue_cleanup = pipe_basic_queue_cleanup,
};

struct dpdk_pipe_type_ops *
dpdk_pipe_basic_get_ops(void)
{
	return &pipe_ops;
}
