/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dpdk_pipe_common.h"

#include <rte_hash.h>

#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "doca_flow.h"
#include "doca_flow_layer.h"

#include "priv_doca_mempool.h"
#include "id_pool.h"
#include "flow_age.h"

#include "dpdk_meter_pipe.h"

#include "dpdk_pipe_items.h"
#include "dpdk_pipe_actions.h"

#include "dpdk_port.h"
#include "dpdk_table.h"

#include "engine_model.h"
#include "utils_df_translate.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_pipe_common);

#define MAX_BUFF_SIZE (256)
#define MEMPOOL_FLOW_ENTRY_FORMAT ("fe_%"PRIx64"")
#define UNIQUEID_QIDX_NAME_FORMAT ("%"PRIx64"_%x")

struct doca_flow_pipe *
dpdk_pipe_common_alloc(uint16_t nr_queues)
{
	struct doca_flow_pipe *pipe;
	uint32_t pipe_size;

	pipe_size = sizeof(struct doca_flow_pipe) + sizeof(struct dpdk_pipe_queue) * nr_queues;
	pipe = calloc(1, pipe_size);
	if (pipe == NULL) {
		DOCA_LOG_ERR("failed memory allocating a new pipe with size %u bytes", pipe_size);
		return NULL;
	}

	pipe->nb_queues = nr_queues;
	pipe->private_ctx = NULL;
	pipe->meter_pipe = NULL;
	return pipe;
}

static void
destroy_action_ctx_per_q(struct dpdk_pipe_actions *actions_ctx)
{
	if (!actions_ctx)
		return;

	free(actions_ctx);
}


static void
dpdk_nb_actions_free(struct dpdk_pipe *dpdk_pipe)
{
	if (dpdk_pipe->nb_actions_pipe) {
		free(dpdk_pipe->nb_actions_pipe);
		dpdk_pipe->nb_actions_pipe = NULL;
	}
}

static int
dpdk_meter_policy_destroy(int port_id, uint32_t policy_id)
{
	int ret;
	struct rte_mtr_error error;
	uint32_t dpdk_policy_id = DPDK_ID(policy_id);

	ret = rte_mtr_meter_policy_delete(port_id, dpdk_policy_id, &error);
	if (ret != 0) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"Port %u delete policy idx(%d) error(%d) message: %s\n",
			port_id, dpdk_policy_id, error.type,
			error.message ? error.message : "(no stated reason)");
	}
	return ret;
}

static void
dpdk_pipe_flow_destroy(struct doca_flow_port *port,
		       struct doca_flow_pipe *pipe,
		       uint16_t mgmt_qid)
{
	uint32_t *policy_id;

	if (dpdk_port_legacy_is_invalid(port))
		return;
	policy_id = pipe->dpdk_pipe.policy_id;
	if (policy_id) {
		dpdk_meter_policy_destroy(port->port_id, *policy_id);
		dpdk_engine_mgmt_id_lock();
		id_pool_free(port->meter_policy_pool, mgmt_qid, policy_id);
		dpdk_engine_mgmt_id_unlock();
	}
}

void
dpdk_pipe_common_resources_free(struct doca_flow_pipe *pipe_legacy)
{
	struct doca_flow_port *port;
	struct dpdk_pipe_q_ctx *pipe_q;
	int i, rc;

	if (unlikely(!pipe_legacy))
		return;

	port = doca_flow_layer_get_pipe_port(pipe_legacy);
	if (unlikely(!port)) {
		DOCA_LOG_ERR("failed freeing pipe - null port");
		return;
	}

	if (pipe_legacy->dpdk_pipe.encap_hws_ctx) {
		rc = dpdk_flow_hws_destroy_encap_table(&pipe_legacy->dpdk_pipe, port);
		if (rc) {
			DOCA_DLOG_ERR("destroy encap table failed ret %d", rc);
			return;
		}
	}

	for (i = 0; i < pipe_legacy->nb_queues; i++) {
		pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, i);
		if (pipe_q->profile_ids) {
			dpdk_profile_hash_destroy(pipe_q->profile_ids->id_hash);
			free(pipe_q->profile_ids);
			pipe_q->profile_ids = NULL;
		}
		destroy_action_ctx_per_q(pipe_q->action_ctx);
	}

	dpdk_nb_actions_free(&pipe_legacy->dpdk_pipe);
	dpdk_pipe_flow_destroy(port, pipe_legacy, dpdk_engine_mgmt_id_get());
}

void
dpdk_pipe_common_table_destroy(struct doca_flow_pipe *pipe_legacy)
{
	if (dpdk_engine_mode_is_hws()) {
		if (dpdk_flow_hws_table_destroy(pipe_legacy))
			DOCA_LOG_ERR("failed to destroy table, pipe %p", pipe_legacy);
	}
}

void
dpdk_pipe_common_legacy_free(struct doca_flow_pipe *pipe_legacy)
{
	struct doca_flow_port *port;
	struct dpdk_group_pool *group_pool;

	if (pipe_legacy->is_connected)
		LIST_REMOVE(pipe_legacy, next);

	if (pipe_legacy->flow_entry_pool) {
		priv_doca_mempool_destroy(pipe_legacy->flow_entry_pool);
		pipe_legacy->flow_entry_pool = NULL;
	}
	if (!pipe_legacy->is_root) {
		port = doca_flow_layer_get_pipe_port(pipe_legacy);
		group_pool = dpdk_port_get_group_pool(port->dpdk_port);
		if (group_pool)
			dpdk_group_pool_free(group_pool, pipe_legacy->group_id);
		else
			DOCA_LOG_ERR("free group_id %u failed: port %u group pool is null",
				pipe_legacy->group_id, port->port_id);
	}
	dpdk_meter_pipe_destroy(pipe_legacy);
	flow_age_pipe_destroy(pipe_legacy);
	free(pipe_legacy);
}

int
dpdk_pipe_common_get_table(struct doca_flow_pipe *pipe,
	  uint8_t act_arr_idx,
	  struct dpdk_table *res)
{
	struct dpdk_hws_item *hws_item;

	if (dpdk_engine_mode_is_sws()) {
		res->table = NULL;
		res->pattern_template_index = 0;
		res->action_template_index = 0;
		return 0;
	}

	/* this is true because basic pipes only have one hws_item in the list */
	hws_item = dpdk_flow_hws_get_pipe_first_ctx(&pipe->dpdk_pipe);
	if (unlikely(hws_item == NULL))
		return -ENOENT;

	res->table = hws_item->table;
	res->pattern_template_index = hws_item->pattern_template_index;
	res->action_template_index = act_arr_idx;
	return 0;
}


int
dpdk_flow_age_list_add(uint16_t queue,
		       struct dpdk_pipe_q_ctx *pipe_q,
		       struct doca_flow_pipe_entry *entry,
		       const struct doca_flow_monitor *monitor)
{
	uint16_t aging_time_in_sec = 0;
	uint64_t user_data = 0;

	if (monitor && monitor->flags & DOCA_FLOW_MONITOR_AGING) {
		aging_time_in_sec = monitor->aging;
		user_data = monitor->user_data;
	} else if (pipe_q->actions_flags & DPDK_PIPE_AGE) {
		aging_time_in_sec = pipe_q->aging;
		user_data = pipe_q->user_data;
	}
	if (aging_time_in_sec == 0)
		return 0;
	entry->age_item = flow_age_item_alloc(queue, entry,
			aging_time_in_sec, user_data);
	return entry->age_item == NULL ? -1 : 0;
}

int
dpdk_pipe_common_queue_modify(enum engine_pipe_type pipe_type,
		       uint16_t pipe_queue,
		       struct dpdk_pipe *dpdk_pipe,
		       struct dpdk_pipe_q_ctx *pipe_q,
		       struct doca_flow_pipe_entry *entry,
		       const struct doca_flow_match *match,
		       const struct doca_flow_actions *actions,
		       const struct doca_flow_monitor *monitor,
		       const struct doca_flow_fwd *fwd,
		       uint32_t flags,
		       void *usr_ctx)
{
	int ret;

	if (match) {
		ret = dpdk_pipe_matches_modify(dpdk_pipe, pipe_q, match);
		if (ret)
			return ret;
	}
	return dpdk_pipe_entry_actions_update(dpdk_pipe, pipe_q, actions, monitor, fwd, entry);
}

static void
fill_flow_params_internal(struct dpdk_flow_params *flow_params, struct rte_flow_action *actions,
			  struct dpdk_table *table, struct rte_flow_attr *attr, uint16_t action_idx)
{
	flow_params->actions = actions;
	flow_params->table = table->table;
	flow_params->item_template_index = table->pattern_template_index;
	flow_params->action_template_index = action_idx;
	flow_params->ingress = attr->ingress;
	flow_params->egress = attr->egress;
	flow_params->transfer = attr->transfer;
	flow_params->group = attr->group;
	flow_params->priority = attr->priority;
}

void
dpdk_pipe_common_fill_flow_params(struct dpdk_flow_params *flow_params, struct rte_flow_item *items,
				  struct rte_flow_action *actions, struct dpdk_table *table,
				  struct rte_flow_attr *attr, uint16_t action_idx)
{
	fill_flow_params_internal(flow_params, actions, table, attr, action_idx);
	flow_params->items = items;
	flow_params->use_index = false;
}

void
dpdk_pipe_common_fill_indexed_flow_params(struct dpdk_flow_params *flow_params, uint32_t index,
					  struct rte_flow_action *actions, struct dpdk_table *table,
					  struct rte_flow_attr *attr, uint16_t action_idx)
{
	fill_flow_params_internal(flow_params, actions, table, attr, action_idx);
	flow_params->index = index;
	flow_params->use_index = true;
}

static bool
pipe_flow_is_persistent(struct doca_flow_pipe *pipe)
{
	return (pipe->type == ENGINE_PIPE_CONTROL);
}

int
dpdk_pipe_common_queue_submit(struct doca_flow_pipe *pipe_legacy,
			      struct doca_flow_pipe_entry *entry, uint16_t queue_id,
			      struct dpdk_flow_params *flow_params,
			      dpdk_flow_completion_cb entry_completion_cb,
			      void *usr_ctx, bool no_wait)
{
	struct doca_flow_port *port;
	struct dpdk_flow_request req = {0};
	struct dpdk_flow_queue *queue;
	int rc;

	port = doca_flow_layer_get_pipe_port(pipe_legacy);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed submitting entry to queue %u - pipe's port is null", queue_id);
		return -ENOTRECOVERABLE;
	}

	queue = dpdk_port_get_flow_queue(port->dpdk_port, queue_id);
	if (unlikely(!queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed submitting entry to queue - invalid queue %u on pipe %s",
				queue_id, pipe_legacy->name);
		return -EINVAL;
	}

	req.tracker = &entry->tracker;
	req.cb = entry_completion_cb;
	req.callback_ctx = entry;
	req.persistent = pipe_flow_is_persistent(pipe_legacy);
	req.wait_for_bulk = !no_wait;

	entry->status = DOCA_FLOW_ENTRY_STATUS_IN_PROCESS;
	entry->user_ctx = usr_ctx;
	rc = dpdk_flow_create(queue, flow_params, &req);
	if (unlikely(rc)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed submitting entry to queue - flow create rc=%d", rc);
		return rc;
	}
	return 0;
}

void
dpdk_pipe_entry_release(struct doca_flow_pipe_entry *entry)
{
	struct doca_flow_pipe *pipe = entry->pipe, *encap_pipe;
	struct doca_flow_pipe_entry *encap_entry;
	struct dpdk_pipe_actions_encap_info *encap_info;

	encap_info = dpdk_pipe_actions_encap_info_put(entry->encap_id);

	if (encap_info) {
		if (dpdk_engine_encap_entry_free(encap_info->port_id, encap_info->handler)) {
			DOCA_DLOG_ERR("Delete port %d encap flow failed",
				      encap_info->port_id);
		}
		encap_entry = encap_info->handler;
		encap_pipe = encap_entry->pipe;
		assert(encap_entry->pipe_queue == dpdk_engine_mgmt_id_get());
		dpdk_engine_mgmt_id_lock();
		priv_doca_mempool_free(encap_pipe->flow_entry_pool, encap_entry,
				       dpdk_engine_mgmt_id_get());
		dpdk_engine_mgmt_id_unlock();
		encap_info->handler = NULL;
	}
	priv_doca_mempool_free(pipe->flow_entry_pool, entry, entry->pipe_queue);
}

int
dpdk_pipe_common_queue_remove(struct doca_flow_pipe_entry *entry,
			      dpdk_flow_completion_cb cb, void *usr_ctx)
{
	struct doca_flow_pipe *pipe = entry->pipe;

	if (dpdk_flow_tracker_get_flow(&entry->tracker))
		return dpdk_rte_flow_destroy(pipe->port, usr_ctx, cb, entry,
					     true /* persistent */, false /* wait_for_bulk */);
	return 0;
}

struct priv_doca_mempool *
pipe_entry_pool_init(uint32_t nb_flows, const uint64_t unique_id)
{
	char pool_name[MAX_BUFF_SIZE];
	struct priv_doca_mempool_cfg flow_entry_pool_cfg = {
		.size = sizeof(struct doca_flow_pipe_entry),
		.name = pool_name,
		.nb_entries = nb_flows,
		.nb_caches = dpdk_engine_get_queues_in_use(),
		.socket_id = rte_socket_id(),
		.usr_ctx = NULL,
		.usr_init_cb = NULL,
	};

	if (flow_entry_pool_cfg.nb_entries == 0) {
		flow_entry_pool_cfg.nb_entries = DEFAULT_NB_FLOWS;
		DOCA_LOG_INFO("flow_entry_pool set to default %d", DEFAULT_NB_FLOWS);
	}

	snprintf(pool_name, MAX_BUFF_SIZE, MEMPOOL_FLOW_ENTRY_FORMAT, unique_id);

	return priv_doca_mempool_create(&flow_entry_pool_cfg);
}

static int
dpdk_profile_per_q_init(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q, int qidx,
			struct dpdk_pipe_cfg *pipe_cfg)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	int ret;
	char name[PRIV_MEMPOOL_NAME_LEN];
	uint64_t unique_id;

	pipe_q->profile_ids = calloc(1, sizeof(*pipe_q->profile_ids));
	if (!pipe_q->profile_ids) {
		DOCA_LOG_ERR("Failed to allocate profile_ids for pipe %s qidx %u", pipe_cfg->name,
			     qidx);
		return -ENOMEM;
	}

	unique_id = dpdk_pipe_generate_unique_id(pipe_cfg->port, pipe);
	snprintf(name, PRIV_MEMPOOL_NAME_LEN, UNIQUEID_QIDX_NAME_FORMAT, unique_id, qidx);
	ret = dpdk_profile_hash_init(DPDK_MAX_PORT_PROFILES, &pipe_q->profile_ids->id_hash, name,
				     RTE_HASH_EXTRA_FLAGS_RW_CONCURRENCY_LF);
	if (ret < 0) {
		DOCA_LOG_ERR("failed init queue profile hash port %u - ret code %d",
			     pipe_cfg->port->port_id, ret);
		goto free_profiles;
	}
	return 0;

free_profiles:
	free(pipe_q->profile_ids);
	pipe_q->profile_ids = NULL;
	return ret;
}

static int
dpdk_nb_actions_alloc(struct dpdk_pipe *dpdk_pipe)
{
	dpdk_pipe->nb_actions_pipe = calloc(dpdk_pipe->nb_actions_array_len, sizeof(uint8_t));
	if (unlikely(!dpdk_pipe->nb_actions_pipe))
		return -ENOMEM;

	return 0;
}

static struct dpdk_pipe_actions *
init_action_ctx_per_q(int nb_actions)
{
	return calloc(nb_actions, sizeof(struct dpdk_pipe_actions));
}

int
dpdk_pipe_per_q_actions_build(enum engine_pipe_type type, struct dpdk_pipe *dpdk_pipe,
			      struct dpdk_pipe_q_ctx *pipe, struct dpdk_pipe_cfg *pipe_cfg)
{
	int ret;

	ret = dpdk_pipe_actions_build(type, dpdk_pipe, pipe, pipe_cfg);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("build pipe actions fail - ret=%d", ret);
		return ret;
	}

	return 0;
}

static int
__dpdk_pipe_per_q_build(enum engine_pipe_type type, struct dpdk_pipe *dpdk_pipe,
			struct dpdk_pipe_q_ctx *pipe, struct dpdk_pipe_cfg *pipe_cfg)
{
	int ret;

	ret = dpdk_pipe_matches_build(dpdk_pipe, pipe, pipe_cfg->uds_cfg.match,
				      pipe_cfg->uds_cfg.match_mask);
	if (ret) {
		DOCA_LOG_RATE_LIMIT_ERR("build pipe items fail - ret=%d", ret);
		return ret;
	}

	return dpdk_pipe_per_q_actions_build(type, dpdk_pipe, pipe, pipe_cfg);
}

int
dpdk_pipe_basic_build(struct doca_flow_pipe *pipe, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct doca_flow_port *port;
	struct dpdk_pipe_q_ctx *pipe_q;
	struct dpdk_pipe_actions *actions;
	uint64_t unique_id;
	uint32_t nb_actions = pipe_cfg->nb_actions ? pipe_cfg->nb_actions : DEFAULT_NB_ACTIONS;
	uint16_t queue_id;
	int32_t ret;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed building basic pipe - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	pipe->is_root = pipe_cfg->is_root;
	strlcpy(pipe->name, pipe_cfg->name, ENGINE_PIPE_NAME_LEN);

	for (queue_id = 0; queue_id < pipe->nb_queues; queue_id++) {
		pipe_q = dpdk_pipe_q_ctx_get(pipe, queue_id);
		actions = init_action_ctx_per_q(nb_actions);
		if (!actions)
			goto release;

		pipe_q->action_ctx = actions;
	}

	if (pipe->is_root)
		dpdk_pipe_set_root_id(pipe);

	ret = dpdk_engine_group_id_get(port, pipe, pipe_cfg->uds_cfg.fwd);
	if (ret)
		goto release;

	unique_id = dpdk_pipe_generate_unique_id(port, pipe);
	pipe->flow_entry_pool = pipe_entry_pool_init(pipe_cfg->nb_flows, unique_id);
	if (!pipe->flow_entry_pool)
		goto release;

	if (pipe_cfg->is_aging_enabled) {
		pipe->is_aging = true;
		pipe->age_queues = calloc(pipe->nb_queues, sizeof(struct doca_flow_age_queue *));
		if (pipe->age_queues == NULL)
			goto release;

		ret = flow_age_pipe_init(pipe, pipe_cfg->nb_flows);
		if (ret)
			goto release;
	}

	for (queue_id = 0; queue_id < pipe->nb_queues; queue_id++) {
		struct dpdk_pipe_queue *pipe_by_qidx = dpdk_pipe_by_qidx_get(pipe, queue_id);
		struct dpdk_pipe_q_ctx *pipe_q;
		int idx;

		LIST_INIT(&pipe_by_qidx->entry_list);
		pipe_q = dpdk_pipe_q_ctx_get(pipe, queue_id);
		pipe_q->fwd_action_index = INVALID_ACTION_INDEX;
		dpdk_engine_pipe_q_ctx_group_set(pipe, pipe_q, pipe_cfg->uds_cfg.fwd, pipe_cfg->is_encrypt);
		if ((pipe_q->attr.transfer == 0 && pipe_q->attr.egress == 0) ||
		     dpdk_engine_mode_is_sws())
			pipe_q->attr.ingress = 1;
		pipe->dpdk_pipe.nb_actions_array_len =
			(pipe_cfg->nb_actions ? pipe_cfg->nb_actions : 1);

		for (idx = 0; idx < DPDK_PIPE_MAX_ITEMS; idx++)
			pipe_q->item_entry[idx].item = &pipe_q->items[idx];

		dpdk_pipe_actions_init(pipe, pipe_q);
	}

	ret = dpdk_nb_actions_alloc(&pipe->dpdk_pipe);
	if (ret)
		goto release;

	for (queue_id = 0; queue_id < pipe->nb_queues; queue_id++) {
		pipe_q = dpdk_pipe_q_ctx_get(pipe, queue_id);

		ret = dpdk_profile_per_q_init(&pipe->dpdk_pipe, pipe_q, queue_id, pipe_cfg);
		if (ret)
			goto release;

		ret = __dpdk_pipe_per_q_build(pipe->type, &pipe->dpdk_pipe, pipe_q, pipe_cfg);
		if (ret)
			goto release;
	}

	pipe_cfg->nb_actions = pipe->dpdk_pipe.nb_actions_array_len;
	return 0;
release:
	return -ENOMEM;
}



static int
verify_shared_mon_binding(struct doca_flow_pipe *pipe,
		const struct doca_flow_monitor *mon)
{
	enum engine_shared_resource_type res_type;
	uint32_t id;

	if (!pipe) {
		DOCA_LOG_ERR("failed to verify shared monitor binding - pipe is NULL.");
		return -EINVAL;
	}

	for (res_type = 0;
	     res_type < ENGINE_SHARED_RESOURCE_MAX;
	     res_type++) {
		switch (res_type) {
		case ENGINE_SHARED_RESOURCE_COUNTER:
			id = mon->shared_counter_id;
			break;
		case ENGINE_SHARED_RESOURCE_METER:
			id = mon->shared_meter_id;
			break;
		default:
			continue;
		}

		if (!id)
			continue;

		if (!engine_shared_resource_is_bound(res_type, id, &pipe->bindable_obj)) {
			DOCA_LOG_ERR("failed to verify shared monitor binding - resource id (%u) type (%u) is not bounded.",
				     id, res_type);
			return -EINVAL;
		}
	}

	return 0;
}

static int
pipe_verify_fwd(const struct doca_flow_port *port, const struct doca_flow_fwd *fwd,
		const struct doca_flow_actions *actions)
{
	struct doca_flow_port *switch_port;

	if (engine_model_is_mode(ENGINE_MODEL_MODE_VNF) && actions && actions->has_encap &&
	    port->hairpin_peer == NULL) {
		DOCA_LOG_ERR("has_encap must be defined with fwd to port and a hairpin peer");
		return -EINVAL;
	}

	if (fwd->type != DOCA_FLOW_FWD_PORT)
		return 0;

	if (engine_model_is_mode(ENGINE_MODEL_MODE_VNF)) {
		if (port->hairpin_peer == NULL || port->hairpin_peer->port_id != fwd->port_id) {
			DOCA_LOG_ERR("vnf mode portid %u %u are not hairpin peers", port->port_id,
				     fwd->port_id);
			return -EINVAL;
		}
	}

	if (engine_model_is_mode(ENGINE_MODEL_MODE_SWITCH)) {
		switch_port = dpdk_port_legacy_switch_get();
		if (switch_port && port != flow_switch) {
			DOCA_LOG_ERR(
				"transfer flow on port %u is not allowed under switch mode, please use proxy port %u",
				port->port_id, switch_port->port_id);
			return -EINVAL;
		}
	}

	return 0;
}

static int
_dpdk_monitor_resource_check_validity(uint8_t mon_flags,
				     uint8_t requested_flags,
				     uint32_t shared_resource_id,
				     uint32_t nb_resources,
				     const char *name)
{
	bool shared_resource_ignored;

	if (mon_flags & requested_flags) {
		shared_resource_ignored =
			utils_df_translate_is_shared_resource_id_ignored(shared_resource_id);

		if (!shared_resource_ignored) {
			DOCA_LOG_ERR(
				"%s: shared and non-shared resources cannot be supported together",
				name);
			return -EINVAL;
		}

		if (!nb_resources) {
			DOCA_LOG_ERR("%s: configured number of resources is 0.", name);
			return -EINVAL;
		}
	}

	return 0;
}

static int
dpdk_monitor_resource_check_validity(uint8_t flags,
				     uint32_t shared_counter_id,
				     uint32_t shared_meter_id,
				     const struct doca_flow_fwd *fwd)
{
	int ret;

	ret = _dpdk_monitor_resource_check_validity(flags,
			(DOCA_FLOW_MONITOR_AGING | DOCA_FLOW_MONITOR_COUNT),
			shared_counter_id,
			dpdk_engine_counters_get_number(),
			"counter");
	if (ret)
		return ret;

	ret = _dpdk_monitor_resource_check_validity(flags,
			DOCA_FLOW_MONITOR_METER,
			shared_meter_id,
			dpdk_engine_meters_get_number(),
			"meter");
	if (ret)
		return ret;

	uint32_t shared_meter_ignored =
		utils_df_translate_is_shared_resource_id_ignored(shared_meter_id);

	if (!shared_meter_ignored) {
		if (dpdk_engine_mode_is_sws()) {
			DOCA_LOG_ERR(
				"failed building shared meter - no support in SW steering mode.");
			return -ENOTSUP;
		}

		if (fwd && utils_df_translate_is_fwd_changeable(fwd)) {
			DOCA_LOG_ERR("failed building shared meter pipe - fwd is changeable.");
			return -ENOTSUP;
		}
	}

	return 0;
}

bool
dpdk_pipe_action_crypto_ignored(const struct doca_flow_actions *actions)
{
	return (actions == NULL ||
		utils_df_translate_is_shared_resource_id_ignored(actions->security.crypto_id));
}

int
dpdk_pipe_common_verify(struct dpdk_pipe_cfg *pipe_cfg)
{
	const struct doca_flow_port *port;
	const struct doca_flow_actions **actions_arr;
	const struct doca_flow_monitor *monitor;
	const struct doca_flow_fwd *fwd;
	uint8_t act_idx;
	int rc;
	bool is_l3_tunnel;

	port = pipe_cfg->port;
	actions_arr = pipe_cfg->uds_cfg.actions;
	monitor = pipe_cfg->uds_cfg.monitor;
	fwd = pipe_cfg->uds_cfg.fwd;

	is_l3_tunnel = (pipe_cfg->uds_cfg.match &&
		       (pipe_cfg->uds_cfg.match->tun.type == DOCA_FLOW_TUN_GTPU ||
		       (pipe_cfg->uds_cfg.match->tun.type == DOCA_FLOW_TUN_GRE)));

	for (act_idx = 0; actions_arr && act_idx < pipe_cfg->nb_actions; act_idx++) {
		const struct doca_flow_actions *actions = actions_arr[act_idx];

		if (fwd != NULL) {
			if (!dpdk_pipe_action_crypto_ignored(actions)) {
				DOCA_LOG_ERR("fwd action is overidden with crypto action");
				return -EINVAL;
			}

			rc = pipe_verify_fwd(port, fwd, actions);
			if (unlikely(rc))
				return rc;

		} else {
			if (dpdk_engine_mode_is_hws() &&
			    dpdk_pipe_action_crypto_ignored(actions)) {
				DOCA_LOG_ERR("hws must input the FWD action.");
				return -EINVAL;
			}
		}

		if (is_l3_tunnel && actions->decap &&
				!(actions->mod_dst_ip.type == DOCA_FLOW_IP4_ADDR ||
				  actions->mod_dst_ip.type == DOCA_FLOW_IP6_ADDR)) {
			DOCA_LOG_ERR("Decap GTPU or GRE protocols must be configured with mod actions of IPv4 or IPV6 dest IP type.");
			return -EINVAL;
		}
	}

	if (monitor != NULL) {
		rc = dpdk_monitor_resource_check_validity(monitor->flags,
							  monitor->shared_counter_id,
							  monitor->shared_meter_id,
							  fwd);
		if (unlikely(rc))
			return rc;
	}

	return 0;
}


/**
 * @brief: verify user input before adding new entry to the pipe
 *
 * @param match
 *
 * @return
 */
static int
dpdk_pipe_entry_input_verify(uint16_t pipe_queue,
			     struct doca_flow_pipe *pipe,
			     struct doca_flow_port *port,
			     const struct doca_flow_actions *actions,
			     const struct doca_flow_fwd *fwd,
			     const struct doca_flow_monitor *monitor)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe, pipe_queue);
	int ret;

	if (fwd != NULL) {
		ret = pipe_verify_fwd(port, fwd, actions);
		if (ret)
			return ret;

		if (dpdk_pipe_actions_is_fwd_changeable(pipe_q) && pipe_q->fwd_type != fwd->type &&
		    !pipe->meter_pipe) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to verify entry input - PIPE with changeable fwd type %d, input entry fwd type %d is not same",
						pipe_q->fwd_type, fwd->type);
			return -EINVAL;
		}
	}

	if (monitor != NULL) {
		if ((monitor->flags & DOCA_FLOW_MONITOR_AGING) && !pipe->is_aging) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to verify entry input - monitor action is not allowed since the pipe aging is not enable");
			return -EINVAL;
		}

		ret = verify_shared_mon_binding(pipe, monitor);
		if (unlikely(ret))
			return ret;
		ret = dpdk_monitor_resource_check_validity(monitor->flags,
							   monitor->shared_counter_id,
							   monitor->shared_meter_id,
							   NULL);
		if (ret)
			return ret;
	}

	return 0;
}

int
dpdk_pipe_common_entry_verify(struct doca_flow_pipe *pipe_legacy, uint16_t queue_id,
			      uint16_t action_idx, struct dpdk_uds_cfg *uds_cfg)
{
	struct doca_flow_port *port;

	if (likely(!engine_model_is_verification_enabled()))
		return 0;

	port = doca_flow_layer_get_pipe_port(pipe_legacy);
	if (!port)
		return -EINVAL;

	return dpdk_pipe_entry_input_verify(queue_id, pipe_legacy, port,
					    uds_cfg->actions[action_idx], uds_cfg->fwd,
					    uds_cfg->monitor);
}

void
dpdk_pipe_common_fill_table_params(struct dpdk_table_params *table_params,
				   uint16_t port_id,
				   struct rte_flow_item items[],
				   struct rte_flow_action **actions,
				   struct rte_flow_action **actions_masks,
				   uint8_t nb_actions,
				   struct rte_flow_attr *attr,
				   uint32_t nb_flows,
				   enum dpdk_table_insertion_type insertion_type,
				   enum dpdk_table_hash_type hash_type)
{
	table_params->port_id = port_id;
	table_params->items = items;
	table_params->actions = actions;
	table_params->actions_mask = actions_masks;
	table_params->nb_actions = nb_actions;
	table_params->flow_attr = attr;
	table_params->nb_flows = nb_flows;
	table_params->insertion_type = insertion_type;
	table_params->hash_type = hash_type;
}

void
dpdk_pipe_common_entry_attach(uint16_t pipe_queue, struct doca_flow_pipe *pipe,
			      struct doca_flow_pipe_entry *entry)
{
	struct dpdk_pipe_queue *pipe_by_qidx;

	pipe_by_qidx = dpdk_pipe_by_qidx_get(pipe, pipe_queue);

	entry->entry_id = pipe_by_qidx->queue_entry_id++;
	LIST_INSERT_HEAD(&pipe_by_qidx->entry_list, entry, next);
	pipe_by_qidx->nb_entries++;
}

void
dpdk_pipe_common_entry_detach(uint16_t pipe_queue, struct doca_flow_pipe *pipe,
			      struct doca_flow_pipe_entry *entry)
{
	struct dpdk_pipe_queue *pipe_by_qidx;

	pipe_by_qidx = dpdk_pipe_by_qidx_get(pipe, pipe_queue);

	LIST_REMOVE(entry, next);
	pipe_by_qidx->nb_entries--;
}
