/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dpdk_pipe_control.h"
#include "dpdk_pipe_common.h"

#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "dpdk_table_manager.h"
#include "dpdk_table.h"
#include "dpdk_port.h"
#include "dpdk_dbg.h"

#include "doca_flow_layer.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_pipe_control);

#define MAX_NB_CONTROL_ENTRY 64

static bool root_is_set[DPDK_PORT_MAX_PORT_ID];

static int
pipe_control_verify(struct dpdk_pipe_cfg *pipe_cfg)
{
	if (pipe_cfg->uds_cfg.fwd_miss)
		DOCA_LOG_WARN("fwd_miss is ignored for control pipe");

	return dpdk_pipe_common_verify(pipe_cfg);
}

static struct doca_flow_pipe *
pipe_control_alloc(uint16_t nr_queues)
{
	struct doca_flow_pipe *pipe_legacy;

	pipe_legacy = dpdk_pipe_common_alloc(nr_queues);
	if (!pipe_legacy)
		return NULL;

	rte_spinlock_init(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
	return pipe_legacy;
}

static int
pipe_control_build(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct dpdk_table_manager_cfg table_manager_cfg;
	struct doca_flow_port *port;
	int rc;

	if (!pipe_cfg->nb_flows)
		pipe_cfg->nb_flows = MAX_NB_CONTROL_ENTRY;

	rte_spinlock_lock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
	if (pipe_cfg->is_root) {
		port = doca_flow_layer_get_pipe_port(pipe_legacy);
		if (root_is_set[port->port_id]) {
			DOCA_LOG_ERR("Cannot have more than one root pipe for port %u", port->port_id);
			rte_spinlock_unlock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
			return -EINVAL;
		}
		root_is_set[port->port_id] = true;
	}
	rte_spinlock_unlock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);

	table_manager_cfg.nr_tables = pipe_cfg->nb_flows;
	pipe_legacy->table_manager = dpdk_table_manager_create(&table_manager_cfg);
	if (!pipe_legacy->table_manager) {
		DOCA_LOG_ERR("failed building control pipe - table manager is null");
		return -ENOMEM;
	}

	rc = dpdk_pipe_basic_build(pipe_legacy, pipe_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed building control pipe - pipe build failed");
		dpdk_table_manager_destroy(pipe_legacy->table_manager);
		pipe_legacy->table_manager = NULL;
	}
	return rc;
}

static int
pipe_control_submit(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	pipe_legacy->pipe_size = pipe_cfg->nb_flows;
	return 0;
}

static void
pipe_control_free(struct doca_flow_pipe *pipe_legacy)
{
	struct doca_flow_port *port;

	dpdk_pipe_entries_flush(doca_flow_layer_get_pipe_port(pipe_legacy), pipe_legacy);

	if (pipe_legacy && pipe_legacy->table_manager)
		dpdk_table_manager_destroy(pipe_legacy->table_manager);

	rte_spinlock_lock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
	if (pipe_legacy->is_root) {
		port = doca_flow_layer_get_pipe_port(pipe_legacy);
		root_is_set[port->port_id] = false;
	}
	rte_spinlock_unlock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);

	dpdk_pipe_common_resources_free(pipe_legacy);
	dpdk_pipe_common_legacy_free(pipe_legacy);
}

static uint32_t
pipe_get_queues_entries(struct doca_flow_pipe *pipe)
{
	struct dpdk_pipe_queue *pipe_by_qidx;
	int qidx;
	uint32_t sum = 0;

	if (!pipe)
		return 0;

	for (qidx = 0; qidx < pipe->nb_queues; qidx++) {
		pipe_by_qidx = dpdk_pipe_by_qidx_get(pipe, qidx);
		sum += pipe_by_qidx->nb_entries;
	}
	return sum;
}

static int
pipe_control_queue_verify(struct doca_flow_pipe *pipe_legacy, uint16_t queue_id,
			  uint16_t action_idx, uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	uint32_t nb_ctrl_entries;

	if (unlikely(priority > MAX_FLOW_PRIO)) {
		DOCA_LOG_ERR("failed creating control entry - invalid priority %u", priority);
		return -EINVAL;
	}

	rte_spinlock_lock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
	nb_ctrl_entries = pipe_get_queues_entries(pipe_legacy);
	if (nb_ctrl_entries >= pipe_legacy->pipe_size) {
		rte_spinlock_unlock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
		return -EPERM;
	}
	rte_spinlock_unlock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);

	return dpdk_pipe_common_entry_verify(pipe_legacy, queue_id, action_idx, uds_cfg);
}

static void
dpdk_pipe_cfg_build(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg,
		    struct dpdk_uds_cfg *uds_cfg)
{
	pipe_cfg->nb_actions = pipe_legacy->dpdk_pipe.nb_actions_array_len;
	pipe_cfg->nb_flows = pipe_legacy->pipe_size;
	pipe_cfg->is_root = pipe_legacy->is_root;
	pipe_cfg->port = doca_flow_layer_get_pipe_port(pipe_legacy);
	pipe_cfg->uds_cfg = *uds_cfg;
}

static int
pipe_control_queue_modify(struct doca_flow_pipe *pipe_legacy,
		struct doca_flow_pipe_entry *entry, uint16_t queue_id,
		uint16_t action_idx, uint32_t priority,
		struct dpdk_uds_cfg *uds_cfg)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	struct rte_flow_action_modify_field modis[1][DPDK_PIPE_MAX_ACTIONS] = {};
	struct rte_flow_action action_masks[1][DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_action *actions_mask_arr[1];
	struct rte_flow_action *actions_arr[1];
	struct dpdk_table *table;
	struct doca_flow_port *port;
	struct dpdk_table_params table_params;
	struct dpdk_pipe_cfg pipe_cfg;
	int rc;

	port = doca_flow_layer_get_pipe_port(pipe_legacy);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed modifying control queue %u - pipe's port is null", queue_id);
		return -ENOTRECOVERABLE;
	}

	pipe_q->attr.priority = priority;

	/* build control pipe matches */
	memset(pipe_q->item_entry, 0, sizeof(pipe_q->item_entry));
	for (int idx = 0; idx < DPDK_PIPE_MAX_ITEMS; idx++)
		pipe_q->item_entry[idx].item = &pipe_q->items[idx];

	pipe_legacy->dpdk_pipe.nb_items = 0;

	rc = dpdk_pipe_matches_build(&pipe_legacy->dpdk_pipe, pipe_q,
			(struct doca_flow_match *)uds_cfg->match,
			(struct doca_flow_match *)uds_cfg->match_mask);
	if (unlikely(rc))
		return rc;

	dpdk_pipe_cfg_build(pipe_legacy, &pipe_cfg, uds_cfg);

	rc = dpdk_pipe_actions_build(pipe_legacy->type, &pipe_legacy->dpdk_pipe, pipe_q, &pipe_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_CRIT("failed modifying control queue %u - pipe's action build failed", queue_id);
		return rc;
	}

	rc = dpdk_pipe_common_queue_modify(ENGINE_PIPE_CONTROL, queue_id,
			&pipe_legacy->dpdk_pipe, pipe_q, entry, uds_cfg->match,
			uds_cfg->actions[action_idx], uds_cfg->monitor, uds_cfg->fwd, 0,
			NULL);
	if (unlikely(rc)) {
		DOCA_LOG_CRIT("failed modifying control queue %u - pipe's queue modify failed", queue_id);
		return rc;
	}

	rte_spinlock_lock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
	rc = dpdk_table_manager_alloc_id(pipe_legacy->table_manager, &entry->table_id);
	rte_spinlock_unlock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
	if (unlikely(rc)) {
		DOCA_LOG_CRIT("failed modifying control queue %u - pipe's table id alloc failed", queue_id);
		return rc;
	}

	table = dpdk_table_manager_get_by_id(pipe_legacy->table_manager, entry->table_id);
	if (unlikely(!table))
		return -ENOTRECOVERABLE;

	actions_arr[0] = dpdk_pipe_actions_array_get(pipe_q, action_idx);
	actions_mask_arr[0] = action_masks[0];
	dpdk_pipe_actions_masks_build(pipe_q, action_masks[0], modis[0], action_idx);

	dpdk_dbg_rte_flow_dump("pattern template:\t", port->port_id,
			       &pipe_q->attr, pipe_q->items, NULL, stdout);
	dpdk_dbg_rte_flow_dump("action template:\t", port->port_id,
			       &pipe_q->attr, NULL, actions_arr[0], stdout);
	dpdk_dbg_rte_flow_dump("actions template mask:\t", port->port_id,
			       &pipe_q->attr, NULL, actions_arr[0], stdout);

	dpdk_pipe_common_fill_table_params(&table_params, port->port_id,
					   pipe_q->items, actions_arr, actions_mask_arr, 1,
					   &pipe_q->attr, MAX_NB_CONTROL_ENTRY,
					   DPDK_TABLE_INSERTION_TYPE_PATTERN,
					   DPDK_TABLE_HASH_TYPE_DEFAULT);

	rc = dpdk_table_create(&table_params, dpdk_port_get_table_ctx(port->dpdk_port), table);
	if (unlikely(rc)) {
		DOCA_LOG_CRIT("failed modifying control queue %u - pipe's table create failed", queue_id);
		return rc;
	}

	return 0;
}

static int
pipe_control_queue_submit(struct doca_flow_pipe *pipe_legacy,
			  struct doca_flow_pipe_entry *entry, uint16_t queue_id,
			  uint16_t action_idx, dpdk_flow_completion_cb entry_completion_cb,
			  void *usr_ctx, bool no_wait)
{
	struct dpdk_flow_params flow_params = {0};
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	struct dpdk_table *table;
	int rc;

	table = dpdk_table_manager_get_by_id(pipe_legacy->table_manager, entry->table_id);
	if (unlikely(!table))
		return -ENOTRECOVERABLE;

	dpdk_pipe_common_entry_attach(queue_id, pipe_legacy, entry);

	dpdk_pipe_common_fill_flow_params(&flow_params, pipe_q->items,
					  dpdk_pipe_actions_array_get(pipe_q, action_idx),
					  table, &pipe_q->attr, action_idx);

	rc = dpdk_pipe_common_queue_submit(pipe_legacy, entry, queue_id, &flow_params,
					   entry_completion_cb, usr_ctx, no_wait);
	if (unlikely(rc)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed control queue submit - common submit failed");
		return rc;
	}
	pipe_q->actions_flags &= ~DPDK_PIPE_FWD;
	return 0;
}

static int
pipe_control_queue_remove(struct doca_flow_pipe_entry *entry,
			  dpdk_flow_completion_cb cb, void *usr_ctx)
{
	return dpdk_pipe_common_queue_remove(entry, cb, usr_ctx);
}

static int
pipe_control_queue_cleanup(struct doca_flow_pipe_entry *entry)
{
	struct doca_flow_pipe *pipe_legacy = (struct doca_flow_pipe *)entry->pipe;
	struct doca_flow_port *port;
	struct dpdk_table *table;
	int rc;

	port = doca_flow_layer_get_pipe_port(pipe_legacy);
	if (unlikely(!port)) {
		DOCA_LOG_CRIT("failed cleanup queue entry - pipe's port is null");
		return -ENOTRECOVERABLE;
	}

	table = dpdk_table_manager_get_by_id(pipe_legacy->table_manager, entry->table_id);
	if (unlikely(!table))
		return -ENOTRECOVERABLE;

	rc = dpdk_table_destroy(dpdk_port_get_table_ctx(port->dpdk_port), table);
	if (unlikely(rc < 0)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed cleanup queue entry - failed to destroy table");
		return rc;
	}

	/* table refcount reached 0, delete its id. */
	if (!rc) {
		rte_spinlock_lock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
		rc = dpdk_table_manager_free_id(pipe_legacy->table_manager, entry->table_id);
		rte_spinlock_unlock(&pipe_legacy->dpdk_pipe.ctrl_entry_lock);
		if (unlikely(rc)) {
			DOCA_LOG_RATE_LIMIT_ERR("failed cleanup queue entry - failed to free table id");
			return rc;
		}
	}

	dpdk_pipe_entry_release(entry);
	return 0;
}

static struct dpdk_pipe_type_ops pipe_ops = {
	.pipe_legacy_verify = pipe_control_verify,
	.pipe_legacy_alloc = pipe_control_alloc,
	.pipe_legacy_build = pipe_control_build,
	.pipe_legacy_submit = pipe_control_submit,
	.pipe_legacy_free = pipe_control_free,

	.pipe_queue_verify = pipe_control_queue_verify,
	.pipe_queue_modify = pipe_control_queue_modify,
	.pipe_queue_submit = pipe_control_queue_submit,
	.pipe_queue_remove = pipe_control_queue_remove,
	.pipe_queue_cleanup = pipe_control_queue_cleanup,
};

struct dpdk_pipe_type_ops *
dpdk_pipe_control_get_ops(void)
{
	return &pipe_ops;
}
