/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dpdk_pipe_ct.h"

#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "dpdk_pipe_common.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_pipe_ct);


static int
pipe_ct_verify(struct dpdk_pipe_cfg *pipe_cfg)
{
	return dpdk_pipe_common_verify(pipe_cfg);
}

static struct doca_flow_pipe *
pipe_ct_alloc(uint16_t nr_queues)
{
	return dpdk_pipe_common_alloc(nr_queues);
}

static int
pipe_ct_build(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	return dpdk_pipe_basic_build(pipe_legacy, pipe_cfg);
}

static int
pipe_ct_submit(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct doca_flow_pipe_cfg cfg = {0};
	struct flow_ct_context *ctx;
	int rc;

	dpdk_pipe_cfg_convert(&cfg, pipe_cfg, ENGINE_PIPE_CT);

	ctx = cfg.port->ct_ctx;
	rc = dpdk_engine_ct_ops_get()->create(ctx, &cfg, pipe_cfg->uds_cfg.fwd,
					      pipe_cfg->uds_cfg.fwd_miss, pipe_legacy, NULL);
	if (unlikely(rc))
		return rc;

	return 0;
}

static void
pipe_ct_free(struct doca_flow_pipe *pipe_legacy)
{
	dpdk_pipe_common_resources_free(pipe_legacy);
	dpdk_pipe_common_table_destroy(pipe_legacy);
	dpdk_pipe_common_legacy_free(pipe_legacy);
}

static int
pipe_ct_queue_verify(struct doca_flow_pipe *pipe_legacy, uint16_t queue_id,
		     uint16_t action_idx, uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	return dpdk_pipe_common_entry_verify(pipe_legacy, queue_id, action_idx, uds_cfg);
}

static int
pipe_ct_queue_modify(struct doca_flow_pipe *pipe_legacy, struct doca_flow_pipe_entry *entry,
		     uint16_t queue_id, uint16_t action_idx, uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	return 0;
}

static int
pipe_ct_queue_submit(struct doca_flow_pipe *pipe_legacy,
		     struct doca_flow_pipe_entry *entry, uint16_t queue_id,
		     uint16_t action_idx, dpdk_flow_completion_cb entry_completion_cb,
		     void *usr_ctx, bool no_wait)
{
	struct dpdk_flow_params flow_params = {0};
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);

	dpdk_pipe_common_fill_flow_params(&flow_params, pipe_q->items,
					  dpdk_pipe_actions_array_get(pipe_q, action_idx),
					  &entry->res, &pipe_q->attr, action_idx);

	dpdk_pipe_common_entry_attach(queue_id, pipe_legacy, entry);
	return dpdk_pipe_common_queue_submit(pipe_legacy, entry, queue_id, &flow_params,
					     entry_completion_cb, usr_ctx, no_wait);
	return 0;
}

static int
pipe_ct_queue_remove(struct doca_flow_pipe_entry *entry,
		     dpdk_flow_completion_cb cb, void *usr_ctx)
{
	return dpdk_pipe_common_queue_remove(entry, cb, usr_ctx);
}

static int
pipe_ct_queue_cleanup(struct doca_flow_pipe_entry *entry)
{
	dpdk_pipe_entry_release(entry);
	return 0;
}

static struct dpdk_pipe_type_ops pipe_ops = {
	.pipe_legacy_verify = pipe_ct_verify,
	.pipe_legacy_alloc = pipe_ct_alloc,
	.pipe_legacy_build = pipe_ct_build,
	.pipe_legacy_submit = pipe_ct_submit,
	.pipe_legacy_free = pipe_ct_free,

	.pipe_queue_verify = pipe_ct_queue_verify,
	.pipe_queue_modify = pipe_ct_queue_modify,
	.pipe_queue_submit = pipe_ct_queue_submit,
	.pipe_queue_remove = pipe_ct_queue_remove,
	.pipe_queue_cleanup = pipe_ct_queue_cleanup,
};

struct dpdk_pipe_type_ops *
dpdk_pipe_ct_get_ops(void)
{
	return &pipe_ops;
}
