/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dpdk_internal_meta.h"
#include "dpdk_pipe_common.h"
#include "dpdk_table_manager.h"
#include "dpdk_port.h"
#include "dpdk_engine.h"
#include "dpdk_dbg.h"

#include "utils_bitmap.h"

#include "doca_flow_layer.h"
#include "doca_flow_priv.h"
#include "doca_log.h"

#include <rte_flow.h>

DOCA_LOG_REGISTER(dpdk_pipe_hash);

#define DPDK_HASH_TAG_INDEX UINT8_MAX
#define DPDK_HASH_TAG_DATA_BITS 16
#define DPDK_HASH_TAG_DATA_MASK ((1 << DPDK_HASH_TAG_DATA_BITS) - 1)
#define DPDK_HASH_TAG_DATA_SHIFT 16

#define PIPE_HASH_INDEX_BITS 24
#define PIPE_HASH_INDEX_MASK ((1 << PIPE_HASH_INDEX_BITS) - 1)

#define PIPE_HASH_ENTRIES_MAX (1 << PIPE_HASH_INDEX_BITS)

#define PIPE_TABLE_SIZE_SHIFT   DPDK_HASH_TAG_DATA_BITS
#define PIPE_TABLE_SIZE_MAX     (1 << PIPE_TABLE_SIZE_SHIFT)
#define PIPE_IDX_IN_TABLE_MASK  (PIPE_TABLE_SIZE_MAX - 1)

struct table_ctx {
	uint32_t table_id;
	uint32_t group_id;
};

struct pipe_hash_preparation_table {
	struct dpdk_table dpdk_table;
	struct dpdk_flow_tracker tracker[0];
};

struct dpdk_pipe_hash_ctx {
	bool is_linear;
	uint8_t nb_tables;
	uint32_t nb_flows_per_table;
	struct table_ctx *table_ctrl;
	struct pipe_hash_preparation_table *pre_table;
	struct utils_bitmap *entry_bitmap;
};

struct preparation_table_entry {
	enum doca_flow_entry_status status; /**< entry handler status */
};

static struct {
	bool init_early_done;
	bool init_late_done;
	uint32_t meta_id;
	uint32_t tag_index;
	uint32_t tag_offset;
} pipe_hash_meta;

/*
 * Register metadata as early as possible,
 * but do not query its tag and offset to allow other modules to register too.
 */
static int
pipe_hash_meta_init_early(void)
{
	int rc;

	rc = dpdk_internal_meta_register(PIPE_HASH_INDEX_BITS);
	if (rc < 0) {
		DOCA_LOG_ERR("Failed to register internal metadata (rc=%d)", rc);
		return rc;
	}
	pipe_hash_meta.meta_id = rc;
	pipe_hash_meta.init_early_done = true;
	return 0;
}

/* Query tag and offset of the previously registered internal metadata. */
static int
pipe_hash_meta_init_late(void)
{
	uint32_t meta_id, tag_index, tag_offset;
	int rc;

	if (pipe_hash_meta.init_late_done)
		return 0;

	if (!pipe_hash_meta.init_early_done) {
		DOCA_LOG_ERR("Internal metadata early initialization is not done");
		return -EINVAL;
	}
	meta_id = pipe_hash_meta.meta_id;

	rc = dpdk_internal_meta_tag_index(meta_id);
	if (rc < 0) {
		DOCA_LOG_ERR("Failed to get internal metadata %u tag index (rc=%d)", meta_id, rc);
		return rc;
	}
	tag_index = rc;

	rc = dpdk_internal_meta_get(meta_id, NULL);
	if (rc < 0) {
		DOCA_LOG_ERR("Failed to get internal metadata %u tag index (rc=%d)", meta_id, rc);
		return rc;
	}
	tag_offset = rc;

	DOCA_LOG_DBG("Registered internal metadata %u: tag %u offset %u", meta_id, tag_index, tag_offset);
	pipe_hash_meta.init_late_done = true;
	pipe_hash_meta.tag_index = tag_index;
	pipe_hash_meta.tag_offset = tag_offset;
	return 0;
}

static inline uint16_t
pipe_hash_get_table_idx(const struct dpdk_pipe_hash_ctx *ctx, uint32_t index)
{
	if (ctx->is_linear)
		return index >> PIPE_TABLE_SIZE_SHIFT;
	return 0;
}

static inline uint32_t
pipe_hash_get_idx_in_table(const struct dpdk_pipe_hash_ctx *ctx, uint32_t index)
{
	if (ctx->is_linear)
		return index & PIPE_IDX_IN_TABLE_MASK;
	return index;
}

static int
pipe_hash_verify(struct dpdk_pipe_cfg *pipe_cfg)
{
	if (!dpdk_engine_mode_is_hws()) {
		DOCA_LOG_ERR("hash pipes only support HWS mode");
		return -ENOTSUP;
	}

	if (pipe_cfg->hash_type != DPDK_TABLE_HASH_TYPE_IDENTITY) {
		DOCA_LOG_ERR("hash pipes only support identity hash currently");
		return -ENOTSUP;
	}

	if (pipe_cfg->nb_flows > PIPE_HASH_ENTRIES_MAX) {
		DOCA_LOG_ERR("hash pipes are limited to %u entries", PIPE_HASH_ENTRIES_MAX);
		return -EINVAL;
	}

	return dpdk_pipe_common_verify(pipe_cfg);
}

static int
pipe_hash_ctx_init(const struct dpdk_pipe_cfg *cfg, struct dpdk_pipe_hash_ctx **ctx_ptr)
{
	uint32_t nb_flows = cfg->nb_flows ? cfg->nb_flows : DEFAULT_NB_FLOWS;
	uint16_t nb_tables = 1;
	struct table_ctx *tables_ctrl = NULL;
	struct utils_bitmap *bitmap = NULL;
	struct dpdk_pipe_hash_ctx *ctx;
	int rc;

	ctx = calloc(1, sizeof(*ctx));
	if (ctx == NULL) {
		DOCA_LOG_ERR("failed to allocate private context");
		return -ENOMEM;
	}

	switch (cfg->hash_type) {
	case DPDK_TABLE_HASH_TYPE_DEFAULT:
		ctx->is_linear = 0;
		break;
	case DPDK_TABLE_HASH_TYPE_IDENTITY:
		ctx->is_linear = 1;
		nb_tables += pipe_hash_get_table_idx(ctx, nb_flows);
		break;
	default:
		DOCA_LOG_ERR("hash type 0x%x is unsupported", cfg->hash_type);
		rc = -EINVAL;
		goto error;
	}

	bitmap = utils_bitmap_create(nb_flows);
	if (unlikely(bitmap == NULL)) {
		DOCA_LOG_ERR("failed to create bitmap for entries.");
		rc = -ENOMEM;
		goto error;
	}

	tables_ctrl = calloc(nb_tables, sizeof(*tables_ctrl));
	if (unlikely(tables_ctrl  == NULL)) {
		DOCA_LOG_ERR("allocate table control memory failed.");
		rc = -ENOMEM;
		goto error;
	}

	ctx->table_ctrl = tables_ctrl;
	ctx->entry_bitmap = bitmap;
	ctx->nb_tables = nb_tables;
	ctx->nb_flows_per_table = (nb_tables == 1) ? nb_flows : PIPE_TABLE_SIZE_MAX;
	*ctx_ptr = ctx;
	return 0;

error:
	free(tables_ctrl);
	utils_bitmap_destroy(bitmap);
	free(ctx);
	return rc;
}

static void
pipe_hash_ctx_destroy(struct dpdk_pipe_hash_ctx *ctx)
{
	if (ctx == NULL)
		return;
	free(ctx->table_ctrl);
	utils_bitmap_destroy(ctx->entry_bitmap);
	free(ctx);
}

static void
add_preparation_table_entry_completion_cb(enum dpdk_flow_request_op op,
					  enum dpdk_flow_request_status status, void *ctx)
{
	struct preparation_table_entry *entry = (struct preparation_table_entry *)ctx;

	if (op == DPDK_FLOW_REQ_OP_DESTROY)
		return;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Preparation table entry addition completion got null entry");
		return;
	}

	switch (status) {
	case DPDK_FLOW_REQ_STATUS_OK:
		entry->status = DOCA_FLOW_ENTRY_STATUS_SUCCESS;
		break;
	default:
		entry->status = DOCA_FLOW_ENTRY_STATUS_ERROR;
		break;
	}

	DOCA_DLOG_DBG("Preparation table entry %p add completed with status %d", entry,
		      entry->status);
}

static int
pipe_hash_preparation_table_create(struct doca_flow_pipe *pipe,
				   const struct dpdk_uds_cfg *uds_cfg)
{
	struct doca_flow_port *port = doca_flow_layer_get_pipe_port(pipe);
	struct dpdk_flow_queue *queue = dpdk_port_get_flow_queue(port->dpdk_port, 0);
	struct dpdk_table_port_ctx *port_ctx = dpdk_port_get_table_ctx(port->dpdk_port);
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe, 0);

	struct rte_flow_item items[2];
	struct rte_flow_item_tag tag;
	struct rte_flow_action actions[3];
	struct rte_flow_action actions_mask[3];
	struct rte_flow_action_jump jump = {};
	struct rte_flow_action *actions_arr[] = {actions};
	struct rte_flow_action *actions_mask_arr[] = {actions_mask};

	struct dpdk_pipe_hash_ctx *ctx = pipe->private_ctx;
	struct pipe_hash_preparation_table *pre_table;

	struct dpdk_flow_params flow_params;
	struct dpdk_table_params table_params;
	struct dpdk_flow_request flow_request;
	struct preparation_table_entry *entries = NULL;

	uint32_t nb_flows = ctx->nb_tables;
	size_t alloc_size;
	int rc, i = 0, j;

	alloc_size = sizeof(*pre_table) + sizeof(pre_table->tracker[0]) * nb_flows;
	pre_table = calloc(1, alloc_size);
	if (unlikely(pre_table == NULL)) {
		DOCA_LOG_ERR("failed to allocate memory for hash preparation table");
		return -ENOMEM;
	}

	tag = (struct rte_flow_item_tag){
		.index = pipe_hash_meta.tag_index,
	};
	items[0] = (struct rte_flow_item){
		.type = RTE_FLOW_ITEM_TYPE_TAG,
		.spec = &tag,
		.mask = &(struct rte_flow_item_tag){
			.data = (PIPE_HASH_INDEX_MASK & ~DPDK_HASH_TAG_DATA_MASK)
					<< pipe_hash_meta.tag_offset,
			.index = UINT8_MAX,
		},
	};
	items[1] = (struct rte_flow_item){
		.type = RTE_FLOW_ITEM_TYPE_END,
	};

	actions[0] = (struct rte_flow_action){
		.type = RTE_FLOW_ACTION_TYPE_MODIFY_FIELD,
		.conf = &(struct rte_flow_action_modify_field){
			.operation = RTE_FLOW_MODIFY_SET,
			.width = DPDK_HASH_TAG_DATA_BITS,
			.src = (struct rte_flow_action_modify_data){
				.field = RTE_FLOW_FIELD_TAG,
				.level = pipe_hash_meta.tag_index,
				.offset = pipe_hash_meta.tag_offset,
			},
			.dst = (struct rte_flow_action_modify_data){
				.field = RTE_FLOW_FIELD_TAG,
				.level = DPDK_HASH_TAG_INDEX,
				/*
				 * It is irrelevant which tag bits are used as index by HW,
				 * PMD will perform the shift if needed.
				 */
				.offset = 0,
			},
		},
	};
	actions_mask[0] = (struct rte_flow_action){
		.type = RTE_FLOW_ACTION_TYPE_MODIFY_FIELD,
		.conf = &(struct rte_flow_action_modify_field){
			.operation = RTE_FLOW_MODIFY_SET,
			.width = UINT32_MAX,
			.src = (struct rte_flow_action_modify_data){
				.field = RTE_FLOW_FIELD_TAG,
				.level = UINT32_MAX,
				.offset = UINT32_MAX,
			},
			.dst = (struct rte_flow_action_modify_data){
				.field = RTE_FLOW_FIELD_TAG,
				.level = UINT32_MAX,
				.offset = UINT32_MAX,
			},
		},
	};

	actions[1] = (struct rte_flow_action){
		.type = RTE_FLOW_ACTION_TYPE_JUMP,
		.conf = &jump,
	};
	actions_mask[1] = (struct rte_flow_action){
		.type = RTE_FLOW_ACTION_TYPE_JUMP,
	};

	actions[2] = (struct rte_flow_action){
		.type = RTE_FLOW_ACTION_TYPE_END,
	};
	actions_mask[2] = (struct rte_flow_action){
		.type = RTE_FLOW_ACTION_TYPE_END,
	};

	DOCA_LOG_DBG("Creating preparation table for pipe '%s'", pipe->name);
	dpdk_dbg_rte_flow_dump("pattern template:\t", port->port_id, &pipe_q->attr, items, NULL, stdout);
	dpdk_dbg_rte_flow_dump("action template:\t", port->port_id, &pipe_q->attr, NULL, actions, stdout);
	dpdk_dbg_rte_flow_dump("action template mask:\t", port->port_id, &pipe_q->attr, NULL, actions_mask, stdout);

	dpdk_pipe_common_fill_table_params(&table_params, port->port_id, items,
					   actions_arr, actions_mask_arr,
					   1 /* nb_actions */, &pipe_q->attr, nb_flows,
					   DPDK_TABLE_INSERTION_TYPE_PATTERN,
					   DPDK_TABLE_HASH_TYPE_DEFAULT);

	rc = dpdk_table_create(&table_params, port_ctx, &pre_table->dpdk_table);
	if (unlikely(rc < 0)) {
		DOCA_LOG_ERR("failed to create hash preparation table (rc=%d)", rc);
		goto error;
	}

	entries = calloc(nb_flows, sizeof(*entries));
	if (unlikely(entries == NULL)) {
		DOCA_LOG_ERR("failed to allocate memory for hash preparation table entries");
		rc = -ENOMEM;
		goto error;
	}

	dpdk_pipe_common_fill_flow_params(&flow_params, items, actions, &pre_table->dpdk_table,
					  &pipe_q->attr, 0 /* action_idx */);

	flow_request.cb = add_preparation_table_entry_completion_cb;
	flow_request.persistent = false;
	flow_request.wait_for_bulk = true;

	for (i = 0; i < nb_flows; i++) {
		tag.data = i << (DPDK_HASH_TAG_DATA_BITS + pipe_hash_meta.tag_offset);
		jump.group = ctx->table_ctrl[i].group_id;

		flow_request.callback_ctx = entries + i;
		flow_request.tracker = &pre_table->tracker[i];
		if (i == nb_flows - 1) {
			flow_request.persistent = true;
			flow_request.wait_for_bulk = false;
		}

		rc = dpdk_flow_create(queue, &flow_params, &flow_request);
		if (rc < 0) {
			DOCA_LOG_ERR("failed to insert flow rule %u to hash preparation table (rc=%d)", i, rc);
			goto error;
		}
	}

	for (j = 0; j < nb_flows; ++j)
		if (entries[j].status != DOCA_FLOW_ENTRY_STATUS_SUCCESS)
			goto error;

	free(entries);

	ctx->pre_table = pre_table;
	return 0;

error:
	flow_request.callback_ctx = NULL;
	flow_request.persistent = false;
	flow_request.wait_for_bulk = true;

	for (j = 0; j < i; ++j) {
		flow_request.tracker = &pre_table->tracker[j];
		if (j == i - 1) {
			flow_request.persistent = true;
			flow_request.wait_for_bulk = false;
		}

		if (entries[j].status == DOCA_FLOW_ENTRY_STATUS_SUCCESS)
			dpdk_flow_destroy(queue, &flow_request);
	}

	if (entries != NULL)
		free(entries);

	if (pre_table->dpdk_table.table != NULL)
		dpdk_table_destroy(port_ctx, &pre_table->dpdk_table);
	free(pre_table);

	return rc;
}

static void
pipe_hash_preparation_table_destroy(struct doca_flow_pipe *pipe)
{
	struct dpdk_pipe_hash_ctx *ctx = pipe->private_ctx;
	struct dpdk_flow_queue *queue;
	struct doca_flow_port *port;
	struct dpdk_table_port_ctx *port_ctx;
	struct dpdk_flow_request flow_request;
	int i;

	if (ctx->pre_table == NULL)
		return;

	port = doca_flow_layer_get_pipe_port(pipe);
	port_ctx = dpdk_port_get_table_ctx(port->dpdk_port);
	queue = dpdk_port_get_flow_queue(port->dpdk_port, 0);

	flow_request.cb = add_preparation_table_entry_completion_cb;
	flow_request.callback_ctx = NULL;
	flow_request.persistent = false;
	flow_request.wait_for_bulk = true;

	for (i = 0; i < ctx->nb_tables; ++i) {
		flow_request.tracker = &ctx->pre_table->tracker[i];
		if (i == ctx->nb_tables - 1) {
			flow_request.persistent = true;
			flow_request.wait_for_bulk = false;
		}

		dpdk_flow_destroy(queue, &flow_request);
	}

	dpdk_table_destroy(port_ctx, &ctx->pre_table->dpdk_table);
	free(ctx->pre_table);
}

static int
pipe_hash_create_table(struct dpdk_pipe_hash_ctx *pipe_ctx,
		       struct dpdk_table_port_ctx *port_ctx,
		       struct table_ctx *ctrl,
		       struct dpdk_table_params *params,
		       struct dpdk_table_manager *table_manager,
		       struct dpdk_group_pool *group_pool)
{
	struct dpdk_table *table;
	int rc;

	rc = dpdk_table_manager_alloc_id(table_manager, &ctrl->table_id);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed to allocate table ID (rc=%d)", rc);
		return rc;
	}

	table = dpdk_table_manager_get_by_id(table_manager, ctrl->table_id);
	if (unlikely(table == NULL)) {
		DOCA_LOG_ERR("failed to get table by ID %u", ctrl->table_id);
		return -ENOTRECOVERABLE;
	}

	if (pipe_ctx->is_linear) {
		rc = dpdk_group_pool_alloc(group_pool, &ctrl->group_id);
		if (unlikely(rc)) {
			DOCA_LOG_ERR("failed to allocate group ID");
			dpdk_table_manager_free_id(table_manager, ctrl->table_id);
			return rc;
		}

		params->flow_attr->group = ctrl->group_id;
	}

	rc = dpdk_table_create(params, port_ctx, table);
	if (rc) {
		DOCA_LOG_ERR("failed to create DPDK table (rc=%d)", rc);
		dpdk_table_manager_free_id(table_manager, ctrl->table_id);
		if (pipe_ctx->is_linear)
			dpdk_group_pool_free(group_pool, ctrl->group_id);
		return rc;
	}

	return 0;
}

static void
pipe_hash_destroy_table(struct dpdk_pipe_hash_ctx *pipe_ctx,
			struct dpdk_table_port_ctx *port_ctx,
			struct table_ctx *ctrl,
			struct dpdk_table_manager *table_manager,
			struct dpdk_group_pool *group_pool)
{
	struct dpdk_table *table;

	if (pipe_ctx->is_linear)
		dpdk_group_pool_free(group_pool, ctrl->group_id);

	table = dpdk_table_manager_get_by_id(table_manager, ctrl->table_id);
	if (likely(table)) {
		dpdk_table_destroy(port_ctx, table);
		dpdk_table_manager_free_id(table_manager, ctrl->table_id);
	}
}

static int
dpdk_pipe_hash_create_tables(struct doca_flow_pipe *pipe, struct dpdk_pipe_q_ctx *pipe_q,
			     uint8_t nb_actions)
{
	struct rte_flow_action_modify_field modis[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS] = {};
	struct rte_flow_action action_masks[MAX_ACT_TEMPLATE][DPDK_PIPE_MAX_ACTIONS];
	struct rte_flow_action *actions_mask_arr[MAX_ACT_TEMPLATE];
	struct rte_flow_action *actions_arr[MAX_ACT_TEMPLATE];
	struct dpdk_table_port_ctx *port_ctx;
	struct doca_flow_port *port;
	struct dpdk_table_params params;
	struct dpdk_pipe_hash_ctx *pipe_ctx = pipe->private_ctx;
	struct rte_flow_attr flow_attr = pipe_q->attr;
	struct rte_flow_item items[] = {
		{
			.type = RTE_FLOW_ITEM_TYPE_TAG,
			.spec = &(struct rte_flow_item_tag){
				.index = DPDK_HASH_TAG_INDEX,
			},
			.mask = &(struct rte_flow_item_tag){
				.index = UINT8_MAX,
			},
		},
		{
			.type = RTE_FLOW_ITEM_TYPE_END,
		},
	};
	int rc, i, j;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_ERR("pipe port is NULL");
		return -ENOENT;
	}

	port_ctx = dpdk_port_get_table_ctx(port->dpdk_port);
	if (unlikely(port_ctx == NULL)) {
		DOCA_LOG_ERR("port table ctx is NULL");
		return -ENOENT;
	}

	for (i = 0; i < nb_actions; i++) {
		dpdk_pipe_actions_masks_build(pipe_q, action_masks[i], modis[i], i);
		actions_arr[i] = dpdk_pipe_actions_array_get(pipe_q, i);
		actions_mask_arr[i] = action_masks[i];
	}

	dpdk_pipe_common_fill_table_params(&params, port->port_id, items,
					   actions_arr, actions_mask_arr, nb_actions, &flow_attr,
					   pipe_ctx->nb_flows_per_table,
					   DPDK_TABLE_INSERTION_TYPE_INDEX,
					   pipe_ctx->is_linear ? DPDK_TABLE_HASH_TYPE_IDENTITY :
								 DPDK_TABLE_HASH_TYPE_DEFAULT);

	DOCA_LOG_DBG("Defining internal tables for pipe '%s'", pipe->name);
	dpdk_dbg_rte_flow_dump("pattern template:\t", port->port_id, &flow_attr, params.items, NULL, stdout);
	for (i = 0; i < nb_actions; i++) {
		char label[64];

		snprintf(label, RTE_DIM(label), "action template %d\t", i);
		dpdk_dbg_rte_flow_dump(label, port->port_id, &flow_attr, NULL, actions_arr[i], stdout);

		snprintf(label, RTE_DIM(label), "action template mask %d\t", i);
		dpdk_dbg_rte_flow_dump(label, port->port_id, &flow_attr, NULL, actions_mask_arr[i], stdout);
	}

	for (i = 0; i < pipe_ctx->nb_tables; i++) {
		struct table_ctx *ctrl = pipe_ctx->table_ctrl + i;

		rc = pipe_hash_create_table(pipe_ctx, port_ctx, ctrl, &params, pipe->table_manager,
					    dpdk_port_get_group_pool(port->dpdk_port));
		if (unlikely(rc)) {
			DOCA_LOG_ERR("failed to create table %u of %u (rc=%d)",
					i, pipe_ctx->nb_tables, rc);
			goto err;
		}
	}

	return 0;

err:
	for (j = 0; j < i; j++)
		pipe_hash_destroy_table(pipe_ctx, port_ctx, pipe_ctx->table_ctrl + j,
					pipe->table_manager,
					dpdk_port_get_group_pool(port->dpdk_port));

	return rc;
}

static void
dpdk_pipe_hash_destroy_tables(struct doca_flow_pipe *pipe)
{
	struct dpdk_pipe_hash_ctx *ctx = pipe->private_ctx;
	struct dpdk_table_port_ctx *port_ctx;
	struct doca_flow_port *port;
	int i;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_ERR("pipe port is NULL");
		return;
	}

	port_ctx = dpdk_port_get_table_ctx(port->dpdk_port);
	if (unlikely(port_ctx == NULL)) {
		DOCA_LOG_ERR("port table ctx is NULL");
		return;
	}

	for (i = 0; i < ctx->nb_tables; i++) {
		struct table_ctx *ctrl = ctx->table_ctrl + i;

		pipe_hash_destroy_table(ctx, port_ctx, ctrl, pipe->table_manager,
					dpdk_port_get_group_pool(port->dpdk_port));
	}
}

static struct doca_flow_pipe *
pipe_hash_alloc(uint16_t nr_queues)
{
	return dpdk_pipe_common_alloc(nr_queues);
}

static void
pipe_hash_free(struct doca_flow_pipe *pipe_legacy)
{
	dpdk_pipe_entries_flush(doca_flow_layer_get_pipe_port(pipe_legacy), pipe_legacy);
	pipe_hash_preparation_table_destroy(pipe_legacy);
	dpdk_pipe_common_resources_free(pipe_legacy);
	dpdk_pipe_hash_destroy_tables(pipe_legacy);
	dpdk_table_manager_destroy(pipe_legacy->table_manager);
	pipe_hash_ctx_destroy(pipe_legacy->private_ctx);
	dpdk_pipe_common_legacy_free(pipe_legacy);
}

static int
pipe_hash_build(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct dpdk_pipe_hash_ctx *ctx = pipe_legacy->private_ctx;
	struct dpdk_table_manager_cfg table_manager_cfg;
	int rc;

	rc = pipe_hash_meta_init_late();
	if (rc < 0)
		return rc;

	rc = pipe_hash_ctx_init(pipe_cfg, &ctx);
	if (rc < 0)
		return rc;
	pipe_legacy->private_ctx = ctx;

	table_manager_cfg.nr_tables = ctx->nb_tables;
	pipe_legacy->table_manager = dpdk_table_manager_create(&table_manager_cfg);
	if (!pipe_legacy->table_manager) {
		DOCA_LOG_ERR("failed to create table manager");
		pipe_hash_ctx_destroy(ctx);
		return -ENOMEM;
	}

	rc = dpdk_pipe_basic_build(pipe_legacy, pipe_cfg);
	if (rc < 0) {
		dpdk_table_manager_destroy(pipe_legacy->table_manager);
		pipe_legacy->table_manager = NULL;
		pipe_hash_ctx_destroy(ctx);
		return rc;
	}

	return 0;
}

static int
pipe_hash_submit(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct dpdk_pipe_hash_ctx *ctx = pipe_legacy->private_ctx;
	int rc;

	if (unlikely(pipe_cfg->uds_cfg.fwd_miss)) {
		struct doca_flow_pipe_cfg cfg = {0};

		dpdk_pipe_cfg_convert(&cfg, pipe_cfg, ENGINE_PIPE_HASH);

		rc = dpdk_pipe_fwd_miss_handle(&cfg, pipe_legacy, pipe_cfg->uds_cfg.fwd_miss);
		if (unlikely(rc))
			return rc;
	}

	rc = dpdk_pipe_hash_create_tables(pipe_legacy, dpdk_pipe_q_ctx_get(pipe_legacy, 0),
					  pipe_cfg->nb_actions);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed to create tables (rc=%d)", rc);
		return rc;
	}

	if (ctx->is_linear) {
		rc = pipe_hash_preparation_table_create(pipe_legacy, &pipe_cfg->uds_cfg);
		if (unlikely(rc)) {
			DOCA_LOG_ERR("failed to create preparation table (rc=%d)", rc);
			dpdk_pipe_hash_destroy_tables(pipe_legacy);
			return rc;
		}
	}

	return 0;
}

static int
pipe_hash_queue_verify(struct doca_flow_pipe *pipe_legacy, uint16_t queue_id, uint16_t action_idx,
		       uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	struct dpdk_pipe_hash_ctx *ctx = pipe_legacy->private_ctx;
	int rc;

	rc = utils_bitmap_get(ctx->entry_bitmap, uds_cfg->index);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("index (%u) is not free (rc=%d)", uds_cfg->index, rc);
		return -EEXIST;
	}

	return dpdk_pipe_common_entry_verify(pipe_legacy, queue_id, action_idx, uds_cfg);
}

static uint32_t
pipe_hash_entry_index_get(struct doca_flow_pipe_entry *entry)
{
	return (uint32_t)(uintptr_t)entry->obj_ctx;
}

static void
pipe_hash_entry_index_set(struct doca_flow_pipe_entry *entry, uint32_t entry_index)
{
	entry->obj_ctx = (void *)(uintptr_t)entry_index;
}

static int
pipe_hash_entry_fill(struct doca_flow_pipe *pipe, uint32_t index, uint16_t action_idx,
		     struct doca_flow_pipe_entry *entry)
{
	struct dpdk_pipe_hash_ctx *ctx = pipe->private_ctx;
	uint16_t table_idx = pipe_hash_get_table_idx(ctx, index);

	if (unlikely(ctx->table_ctrl == NULL))
		return -ENOENT;

	pipe_hash_entry_index_set(entry, index);
	entry->table_id = ctx->table_ctrl[table_idx].table_id;
	return 0;
}

static int
pipe_hash_queue_modify(struct doca_flow_pipe *pipe_legacy, struct doca_flow_pipe_entry *entry,
		       uint16_t queue_id, uint16_t action_idx, uint32_t priority,
		       struct dpdk_uds_cfg *uds_cfg)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	int rc;

	rc = dpdk_pipe_common_queue_modify(ENGINE_PIPE_HASH, queue_id, &pipe_legacy->dpdk_pipe,
					   pipe_q, entry, NULL, uds_cfg->actions[action_idx],
					   uds_cfg->monitor, uds_cfg->fwd, 0, NULL);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed to modify queue");
		return rc;
	}

	rc = dpdk_flow_age_list_add(queue_id, pipe_q, entry, uds_cfg->monitor);
	if (unlikely(rc))
		return rc;

	rc = pipe_hash_entry_fill(pipe_legacy, uds_cfg->index, action_idx, entry);
	if (unlikely(rc))
		return rc;

	return 0;
}

static int
pipe_hash_queue_submit(struct doca_flow_pipe *pipe_legacy,
		       struct doca_flow_pipe_entry *entry, uint16_t queue_id,
		       uint16_t action_idx, dpdk_flow_completion_cb entry_completion_cb,
		       void *usr_ctx, bool no_wait)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	struct dpdk_pipe_hash_ctx *ctx = pipe_legacy->private_ctx;
	uint32_t entry_index = pipe_hash_entry_index_get(entry);
	uint32_t index = pipe_hash_get_idx_in_table(ctx, entry_index);
	struct rte_flow_action *actions = dpdk_pipe_actions_array_get(pipe_q, action_idx);
	struct dpdk_flow_params flow_params = {0};
	struct dpdk_table *table;
	int rc;

	table = dpdk_table_manager_get_by_id(pipe_legacy->table_manager, entry->table_id);
	if (unlikely(!table))
		return -ENOTRECOVERABLE;

	dpdk_pipe_common_fill_indexed_flow_params(&flow_params, index, actions,
						  table, &pipe_q->attr, action_idx);

	if (ctx->is_linear) {
		uint16_t table_id = pipe_hash_get_table_idx(ctx, entry_index);

		flow_params.group = ctx->table_ctrl[table_id].group_id;
	}

	dpdk_pipe_common_entry_attach(queue_id, pipe_legacy, entry);
	rc = dpdk_pipe_common_queue_submit(pipe_legacy, entry, queue_id, &flow_params,
					   entry_completion_cb, usr_ctx, no_wait);
	if (rc < 0)
		return rc;

	utils_bitmap_set(ctx->entry_bitmap, entry_index, true);
	return 0;
}

static int
pipe_hash_queue_remove(struct doca_flow_pipe_entry *entry,
		       dpdk_flow_completion_cb cb, void *usr_ctx)
{
	return dpdk_pipe_common_queue_remove(entry, cb, usr_ctx);
}

static int
pipe_hash_queue_cleanup(struct doca_flow_pipe_entry *entry)
{
	struct doca_flow_pipe *pipe = entry->pipe;
	struct dpdk_pipe_hash_ctx *ctx = pipe->private_ctx;
	uint32_t entry_index = pipe_hash_entry_index_get(entry);

	utils_bitmap_set(ctx->entry_bitmap, entry_index, false);
	dpdk_pipe_entry_release(entry);
	return 0;
}

static int
pipe_hash_module_init(void)
{
	return pipe_hash_meta_init_early();
}

static struct dpdk_pipe_type_ops pipe_ops = {
	.module_init = pipe_hash_module_init,

	.pipe_legacy_verify = pipe_hash_verify,
	.pipe_legacy_alloc = pipe_hash_alloc,
	.pipe_legacy_build = pipe_hash_build,
	.pipe_legacy_submit = pipe_hash_submit,
	.pipe_legacy_free = pipe_hash_free,

	.pipe_queue_verify = pipe_hash_queue_verify,
	.pipe_queue_modify = pipe_hash_queue_modify,
	.pipe_queue_submit = pipe_hash_queue_submit,
	.pipe_queue_remove = pipe_hash_queue_remove,
	.pipe_queue_cleanup = pipe_hash_queue_cleanup,
};

struct dpdk_pipe_type_ops *
dpdk_pipe_hash_get_ops(void)
{
	return &pipe_ops;
}

int
dpdk_pipe_hash_build_set_index_action(struct dpdk_pipe *dpdk_pipe,
				      struct dpdk_pipe_q_ctx *pipe_q,
				      int act_arr_idx,
				      uint32_t index)
{

	struct doca_flow_actions actions = {};
	struct doca_flow_action_desc action_desc = {};
	uint32_t tag_index = pipe_hash_meta.tag_index;
	uint32_t tag_offset = pipe_hash_meta.tag_offset;

	actions.flags = DPDK_META_INTERNAL;
	actions.meta.u32[tag_index] = (index == UINT32_MAX) ? UINT32_MAX : (index << tag_offset);
	action_desc.type = DOCA_FLOW_ACTION_SET;
	action_desc.mask.u32 = PIPE_HASH_INDEX_MASK << tag_offset;
	return dpdk_action_modify_build(dpdk_pipe, pipe_q, &actions,
					RTE_FLOW_FIELD_TAG, tag_index,
					&actions.meta.u32[tag_index], &action_desc,
					act_arr_idx);
}

int
dpdk_pipe_hash_modify_set_index_action(struct dpdk_action_entry *action_entry,
				       uint32_t index)
{
	struct dpdk_action_modify_field_data *modi = &action_entry->action_data.modify_field;
	void *data = (void *)modi->conf.src.value;
	uint32_t meta = index << pipe_hash_meta.tag_offset;

	memcpy(data, &meta, sizeof(meta));
	return 0;
}

void
dpdk_pipe_hash_build_index_match(struct doca_flow_match *match,
				 struct doca_flow_match *match_mask)
{
	uint32_t tag_index = pipe_hash_meta.tag_index;

	match->meta.u32[tag_index] = UINT32_MAX;
	match->flags |= DPDK_META_INTERNAL;
	match_mask->meta.u32[tag_index] = PIPE_HASH_INDEX_MASK << pipe_hash_meta.tag_offset;
	match_mask->flags |= DPDK_META_INTERNAL;
}

void
dpdk_pipe_hash_modify_index_match(struct doca_flow_match *match, uint32_t index)
{
	match->flags |= DPDK_META_INTERNAL;
	match->meta.u32[pipe_hash_meta.tag_index] = index << pipe_hash_meta.tag_offset;
}
