/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <bsd/string.h>
#include <stdarg.h>
#include <stdio.h>

#include <rte_ethdev.h>
#include <rte_gre.h>
#include <rte_gtp.h>
#include <rte_hash.h>
#include <rte_hash_crc.h>
#include <rte_mtr.h>
#include <rte_vxlan.h>

#include <doca_log.h>

#include "utils_df_translate.h"
#include "utils_field.h"

#include "dpdk_engine.h"
#include "dpdk_pipe_items.h"
#include "doca_flow_priv.h"


DOCA_LOG_REGISTER(dpdk_pipe_items);

#define MAX_VNI (0xFFFFFF)
#define GET_MATCH_FIELD(field) (is_outer ? match->out_##field : match->in_##field)
#define GET_MASK_FIELD(field) (is_outer ? match_mask->out_##field : match_mask->in_##field)

#define GET_ITEM_SPEC(field) (&entry->item_data.field.spec)
#define GET_ITEM_MASK(field) (&entry->item_data.field.mask)

static inline uint8_t
dpdk_match_ip_type_get(const struct doca_flow_match *match, bool is_outer)
{
	struct doca_flow_ip_addr src = GET_MATCH_FIELD(src_ip);
	struct doca_flow_ip_addr dst = GET_MATCH_FIELD(dst_ip);

	if (src.type == DOCA_FLOW_ADDR_NONE && dst.type == DOCA_FLOW_ADDR_NONE)
		return DOCA_FLOW_ADDR_NONE;
	return src.type == DOCA_FLOW_ADDR_NONE ? dst.type : src.type;
}

static inline rte_be16_t
dpdk_ether_type_get(const struct doca_flow_match *match, bool is_outer)
{
	uint16_t protocol, ip_type;

	protocol = GET_MATCH_FIELD(eth_type);
	if (GET_MATCH_FIELD(vlan_tci))
		protocol = RTE_BE16(RTE_ETHER_TYPE_VLAN);
	else {
		ip_type = dpdk_match_ip_type_get(match, is_outer);
		if (ip_type != DOCA_FLOW_ADDR_NONE)
			protocol = RTE_BE16(ip_type == DOCA_FLOW_IP4_ADDR ? RTE_ETHER_TYPE_IPV4
									  : RTE_ETHER_TYPE_IPV6);
	}
	return protocol;
}

static inline uint64_t
dpdk_get_rss_type(uint32_t rss_type)
{
	uint64_t rss_flags = 0;

	if (rss_type & DOCA_FLOW_RSS_IP)
		rss_flags |= ETH_RSS_IP;
	if (rss_type & DOCA_FLOW_RSS_UDP)
		rss_flags |= ETH_RSS_UDP;
	if (rss_type & DOCA_FLOW_RSS_TCP)
		rss_flags |= ETH_RSS_TCP;

	return rss_flags;
}

static int
dpdk_item_eth_modify(struct dpdk_item_entry *entry, const struct doca_flow_match *match)
{
	struct rte_flow_item_eth *spec = GET_ITEM_SPEC(eth);
	struct rte_flow_item_eth *mask = GET_ITEM_MASK(eth);
	bool is_outer = entry->item_data.eth.is_outer;

	if (entry->flags & DPDK_MODIFY_SMAC)
		utils_df_translate_copy_mac_with_mask(GET_MATCH_FIELD(src_mac), (uint8_t *)&mask->src, (uint8_t *)&spec->src);
	if (entry->flags & DPDK_MODIFY_DMAC)
		utils_df_translate_copy_mac_with_mask(GET_MATCH_FIELD(dst_mac), (uint8_t *)&mask->dst, (uint8_t *)&spec->dst);
	if (entry->flags & DPDK_MODIFY_ETH_TYPE)
		spec->type = GET_MATCH_FIELD(eth_type);
	return 0;
}

static void
dpdk_item_eth_build(struct dpdk_item_entry *entry,
		    const struct doca_flow_match *match,
		    const struct doca_flow_match *match_mask,
		    bool is_outer)
{
	struct rte_flow_item_eth *spec = GET_ITEM_SPEC(eth);
	struct rte_flow_item_eth *mask = GET_ITEM_MASK(eth);
	struct rte_flow_item *flow_item = entry->item;

	flow_item->type = RTE_FLOW_ITEM_TYPE_ETH;
	entry->item_data.eth.is_outer = is_outer;
	if (!utils_df_translate_is_mac_ignored(GET_MATCH_FIELD(src_mac))) {
		if (!match_mask)
			utils_df_translate_set_mac_changeable(mask->src.addr_bytes);
		else
			utils_df_translate_copy_mac(GET_MASK_FIELD(src_mac), (uint8_t *)&mask->src);
		utils_df_translate_copy_mac(GET_MATCH_FIELD(src_mac), (uint8_t *)&spec->src);
		if (utils_df_translate_is_mac_changeable(GET_MATCH_FIELD(src_mac)))
			entry->flags |= DPDK_MODIFY_SMAC;
	}
	if (!utils_df_translate_is_mac_ignored(GET_MATCH_FIELD(dst_mac))) {
		if (!match_mask)
			utils_df_translate_set_mac_changeable(mask->dst.addr_bytes);
		else
			utils_df_translate_copy_mac(GET_MASK_FIELD(dst_mac), (uint8_t *)&mask->dst);
		utils_df_translate_copy_mac(GET_MATCH_FIELD(dst_mac), (uint8_t *)&spec->dst);
		if (utils_df_translate_is_mac_changeable(GET_MATCH_FIELD(dst_mac)))
			entry->flags |= DPDK_MODIFY_DMAC;
	}
	if (GET_MATCH_FIELD(vlan_tci)) {
		spec->has_vlan = 1;
		spec->type = RTE_BE16(RTE_ETHER_TYPE_VLAN);
	} else {
		spec->type = GET_MATCH_FIELD(eth_type);
	}
	if (match_mask != NULL)
		mask->type = GET_MASK_FIELD(eth_type);
	else if (spec->type != 0)
		mask->type = UINT16_MAX;
	if (spec->type == UINT16_MAX)
		entry->flags |= DPDK_MODIFY_ETH_TYPE;
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (entry->flags)
		entry->modify_item = dpdk_item_eth_modify;
}

static int
dpdk_item_vlan_modify(struct dpdk_item_entry *entry,
		      const struct doca_flow_match *match)
{
	struct rte_flow_item_vlan *spec = GET_ITEM_SPEC(vlan);
	struct rte_flow_item_vlan *mask = GET_ITEM_MASK(vlan);
	bool is_outer = entry->item_data.vlan.is_outer;

	if ((entry->flags & DPDK_MODIFY_VLAN_TCI) && GET_MATCH_FIELD(vlan_tci))
		spec->tci = GET_MATCH_FIELD(vlan_tci) & mask->tci;
	return 0;
}

static void
dpdk_item_vlan_build(struct dpdk_item_entry *entry,
		     const struct doca_flow_match *match,
		     const struct doca_flow_match *match_mask,
		     bool is_outer)
{
	struct rte_flow_item_vlan *spec = GET_ITEM_SPEC(vlan);
	struct rte_flow_item_vlan *mask = GET_ITEM_MASK(vlan);
	struct rte_flow_item *flow_item = entry->item;

	if (!GET_MATCH_FIELD(vlan_tci))
		return;
	flow_item->type = RTE_FLOW_ITEM_TYPE_VLAN;
	entry->item_data.vlan.is_outer = is_outer;
	spec->tci = GET_MATCH_FIELD(vlan_tci);
	flow_item->spec = spec;
	if (match_mask)
		mask->tci = GET_MASK_FIELD(vlan_tci);
	else
		mask->tci = UINT16_MAX;
	flow_item->mask = mask;
	if (GET_MATCH_FIELD(vlan_tci) == UINT16_MAX) {
		entry->flags |= DPDK_MODIFY_VLAN_TCI;
		entry->modify_item = dpdk_item_vlan_modify;
	}
}

static int
dpdk_item_ipv4_modify(struct dpdk_item_entry *entry,
		      const struct doca_flow_match *match)
{
	struct rte_flow_item_ipv4 *spec = GET_ITEM_SPEC(ipv4);
	bool is_outer = entry->item_data.ipv4.is_outer;
	struct doca_flow_ip_addr src_ip = GET_MATCH_FIELD(src_ip);
	struct doca_flow_ip_addr dst_ip = GET_MATCH_FIELD(dst_ip);
	struct rte_flow_item_ipv4 *mask = GET_ITEM_MASK(ipv4);
	uint8_t l4_type = GET_MATCH_FIELD(l4_type);

	if ((entry->flags & DPDK_MODIFY_SIP) && src_ip.ipv4_addr)
		spec->hdr.src_addr = (src_ip.ipv4_addr & mask->hdr.src_addr);
	if ((entry->flags & DPDK_MODIFY_DIP) && dst_ip.ipv4_addr)
		spec->hdr.dst_addr = (dst_ip.ipv4_addr & mask->hdr.dst_addr);
	if ((entry->flags & DPDK_MODIFY_L4_PROTO) && l4_type)
		spec->hdr.next_proto_id = l4_type;
	return 0;
}

static void
dpdk_item_ipv4_build(struct dpdk_item_entry *entry,
		     const struct doca_flow_match *match,
		     const struct doca_flow_match *match_mask,
		     bool is_outer)
{
	struct rte_flow_item_ipv4 *spec = GET_ITEM_SPEC(ipv4);
	struct rte_flow_item_ipv4 *mask = GET_ITEM_MASK(ipv4);
	struct rte_flow_item *flow_item = entry->item;
	struct doca_flow_ip_addr src_ip = GET_MATCH_FIELD(src_ip);
	struct doca_flow_ip_addr dst_ip = GET_MATCH_FIELD(dst_ip);
	struct doca_flow_ip_addr mask_src_ip;
	struct doca_flow_ip_addr mask_dst_ip;

	flow_item->type = RTE_FLOW_ITEM_TYPE_IPV4;
	entry->item_data.ipv4.is_outer = is_outer;
	if (match_mask) {
		mask_src_ip = GET_MASK_FIELD(src_ip);
		mask_dst_ip = GET_MASK_FIELD(dst_ip);
	} else {
		mask_src_ip.ipv4_addr = UINT32_MAX;
		mask_dst_ip.ipv4_addr = UINT32_MAX;
	}
	if (!utils_df_translate_is_ip_ignored(&src_ip)) {
		spec->hdr.src_addr = src_ip.ipv4_addr;
		mask->hdr.src_addr = mask_src_ip.ipv4_addr;
		if (utils_df_translate_is_ip_changeable(&src_ip))
			entry->flags |= DPDK_MODIFY_SIP;
	}
	if (!utils_df_translate_is_ip_ignored(&dst_ip)) {
		spec->hdr.dst_addr = dst_ip.ipv4_addr;
		mask->hdr.dst_addr = mask_dst_ip.ipv4_addr;
		if (utils_df_translate_is_ip_changeable(&dst_ip))
			entry->flags |= DPDK_MODIFY_DIP;
	}
	if (GET_MATCH_FIELD(l4_type) != 0) {
		spec->hdr.next_proto_id = GET_MATCH_FIELD(l4_type);
		mask->hdr.next_proto_id = UINT8_MAX;
		if (GET_MATCH_FIELD(l4_type) == UINT8_MAX)
			entry->flags |= DPDK_MODIFY_L4_PROTO;
	}
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (entry->flags)
		entry->modify_item = dpdk_item_ipv4_modify;
}

static int
dpdk_item_ipv6_modify(struct dpdk_item_entry *entry,
		      const struct doca_flow_match *match)
{
	struct rte_flow_item_ipv6 *spec = GET_ITEM_SPEC(ipv6);
	struct rte_flow_item_ipv6 *mask = GET_ITEM_MASK(ipv6);
	bool is_outer = entry->item_data.ipv6.is_outer;
	struct doca_flow_ip_addr src_ip = GET_MATCH_FIELD(src_ip);
	struct doca_flow_ip_addr dst_ip = GET_MATCH_FIELD(dst_ip);
	uint8_t l4_type = GET_MATCH_FIELD(l4_type);

	if ((entry->flags & DPDK_MODIFY_SIP) && !utils_df_translate_is_ip_ignored(&src_ip))
		utils_df_translate_copy_ip_with_mask(&src_ip, mask->hdr.src_addr, spec->hdr.src_addr);
	if ((entry->flags & DPDK_MODIFY_DIP) && !utils_df_translate_is_ip_ignored(&dst_ip))
		utils_df_translate_copy_ip_with_mask(&dst_ip, mask->hdr.dst_addr, spec->hdr.dst_addr);
	if ((entry->flags & DPDK_MODIFY_L4_PROTO) && l4_type)
		spec->hdr.proto = l4_type;
	return 0;
}

static void
dpdk_item_ipv6_build(struct dpdk_item_entry *entry,
		     const struct doca_flow_match *match,
		     const struct doca_flow_match *match_mask,
		     bool is_outer)
{
	struct rte_flow_item_ipv6 *spec = GET_ITEM_SPEC(ipv6);
	struct rte_flow_item_ipv6 *mask = GET_ITEM_MASK(ipv6);
	struct doca_flow_ip_addr src_ip = GET_MATCH_FIELD(src_ip);
	struct doca_flow_ip_addr dst_ip = GET_MATCH_FIELD(dst_ip);
	struct rte_flow_item *flow_item = entry->item;
	struct doca_flow_ip_addr mask_src_ip = {0};
	struct doca_flow_ip_addr mask_dst_ip = {0};

	if (match_mask) {
		mask_src_ip = GET_MASK_FIELD(src_ip);
		mask_dst_ip = GET_MASK_FIELD(dst_ip);
	} else {
		mask_src_ip.type = DOCA_FLOW_IP6_ADDR;
		mask_dst_ip.type = DOCA_FLOW_IP6_ADDR;

		utils_df_translate_set_ip_changeable(&mask_src_ip);
		utils_df_translate_set_ip_changeable(&mask_dst_ip);
	}

	flow_item->type = RTE_FLOW_ITEM_TYPE_IPV6;
	entry->item_data.ipv6.is_outer = is_outer;
	if (!utils_df_translate_is_ip_ignored(&src_ip)) {
		utils_df_translate_copy_ip(&src_ip, spec->hdr.src_addr);
		utils_df_translate_copy_ip(&mask_src_ip, mask->hdr.src_addr);
		if (utils_df_translate_is_ip_changeable(&src_ip))
			entry->flags |= DPDK_MODIFY_SIP;
	}
	if (!utils_df_translate_is_ip_ignored(&dst_ip)) {
		utils_df_translate_copy_ip(&dst_ip, spec->hdr.dst_addr);
		utils_df_translate_copy_ip(&mask_dst_ip, mask->hdr.dst_addr);
		if (utils_df_translate_is_ip_changeable(&dst_ip))
			entry->flags |= DPDK_MODIFY_DIP;
	}
	if (GET_MATCH_FIELD(l4_type) != 0) {
		spec->hdr.proto = GET_MATCH_FIELD(l4_type);
		mask->hdr.proto = UINT8_MAX;
		if (GET_MATCH_FIELD(l4_type) == UINT8_MAX)
			entry->flags |= DPDK_MODIFY_L4_PROTO;
	}
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (entry->flags)
		entry->modify_item = dpdk_item_ipv6_modify;
}

/*copy vni first 3 bits*/
#define SET_VNI_BY_TUN(vni, be32_tun_id) \
	memcpy((vni), (uint8_t *)(be32_tun_id), 3)

#define SET_VNI_BY_TUN_WITH_MASK(vni, mask, be32_tun_id) \
	for (int i = 0; i < 3; i++)			\
		vni[i] = ((uint8_t *)(be32_tun_id))[i] & mask[i]

static int
dpdk_item_vxlan_modify(struct dpdk_item_entry *entry,
		       const struct doca_flow_match *match)
{
	struct rte_flow_item_vxlan *spec = GET_ITEM_SPEC(vxlan);
	struct rte_flow_item_vxlan *mask = GET_ITEM_MASK(vxlan);

	if (match->tun.vxlan_tun_id)
		SET_VNI_BY_TUN_WITH_MASK(spec->vni, mask->vni, &match->tun.vxlan_tun_id);
	return 0;
}

static void
dpdk_item_vxlan_build(struct dpdk_item_entry *entry,
		      const struct doca_flow_match *match,
		      const struct doca_flow_match *match_mask)
{
	struct rte_flow_item_vxlan *spec = GET_ITEM_SPEC(vxlan);
	struct rte_flow_item_vxlan *mask = GET_ITEM_MASK(vxlan);
	struct rte_flow_item *flow_item = entry->item;
	uint32_t vni_value = RTE_BE32(match->tun.vxlan_tun_id) >> 8;

	flow_item->type = RTE_FLOW_ITEM_TYPE_VXLAN;
	if (vni_value == 0)
		return;
	if (match_mask)
		SET_VNI_BY_TUN(mask->vni, &match_mask->tun.vxlan_tun_id);
	else
		utils_df_translate_set_vni_changeable(mask->vni);
	SET_VNI_BY_TUN(spec->vni, &match->tun.vxlan_tun_id);
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (vni_value == MAX_VNI) {
		entry->flags |= DPDK_MODIFY_VXLAN_VNI;
		entry->modify_item = dpdk_item_vxlan_modify;
	}
}

static void
dpdk_item_gre_build(struct dpdk_item_entry *entry,
		    const struct doca_flow_match *match,
		    const struct doca_flow_match *match_mask)
{
	struct rte_flow_item_gre *spec = GET_ITEM_SPEC(gre);
	struct rte_flow_item *flow_item = entry->item;
	uint16_t next_proto;
	uint8_t ip_type;

	flow_item->type = RTE_FLOW_ITEM_TYPE_GRE;
	next_proto = match_mask ? match->tun.protocol &
		     match_mask->tun.protocol : match->tun.protocol;

	if (next_proto) {
		spec->protocol = next_proto;
	} else {
		ip_type = dpdk_match_ip_type_get(match, false);
		/* No MPLS for now */
		if (ip_type != DOCA_FLOW_ADDR_NONE)
			spec->protocol = RTE_BE16(ip_type == DOCA_FLOW_IP4_ADDR ?
						  RTE_ETHER_TYPE_IPV4 : RTE_ETHER_TYPE_IPV6);
	}
}

static int
dpdk_item_gre_key_modify(struct dpdk_item_entry *entry,
			 const struct doca_flow_match *match)
{
	doca_be32_t *spec = GET_ITEM_SPEC(gre_key);
	doca_be32_t *mask = GET_ITEM_MASK(gre_key);

	if (match->tun.gre_key)
		*spec = *mask & match->tun.gre_key;
	return 0;
}

static void
dpdk_item_gre_key_build(struct dpdk_item_entry *entry,
			const struct doca_flow_match *match)
{
	doca_be32_t *spec = GET_ITEM_SPEC(gre_key);
	doca_be32_t *mask = GET_ITEM_MASK(gre_key);
	struct rte_flow_item *flow_item = entry->item;

	flow_item->type = RTE_FLOW_ITEM_TYPE_GRE_KEY;
	if (!match->tun.gre_key)
		return;
	*spec = match->tun.gre_key;
	*mask = UINT32_MAX;
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (match->tun.gre_key == UINT32_MAX) {
		entry->flags |= DPDK_MODIFY_GRE_KEY;
		entry->modify_item = dpdk_item_gre_key_modify;
	}
}

static int
dpdk_item_gtp_teid_modify(struct dpdk_item_entry *entry,
			  const struct doca_flow_match *match)
{
	struct rte_flow_item_gtp *spec = GET_ITEM_SPEC(gtp);
	struct rte_flow_item_gtp *mask = GET_ITEM_MASK(gtp);

	if (match->tun.gtp_teid)
		spec->teid = mask->teid & match->tun.gtp_teid;
	return 0;
}

static void
dpdk_item_gtp_build(struct dpdk_item_entry *entry,
		    const struct doca_flow_match *match,
		    const struct doca_flow_match *match_mask)
{
	struct rte_flow_item *flow_item = entry->item;
	struct rte_flow_item_gtp *spec = GET_ITEM_SPEC(gtp);
	struct rte_flow_item_gtp *mask = GET_ITEM_MASK(gtp);

	flow_item->type = RTE_FLOW_ITEM_TYPE_GTP;
	if (!match->tun.gtp_teid)
		return;
	spec->teid = match->tun.gtp_teid;
	mask->teid = match_mask ? match_mask->tun.gtp_teid : UINT32_MAX;
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (match->tun.gtp_teid == UINT32_MAX) {
		entry->flags |= DPDK_MODIFY_GTP_TEID;
		entry->modify_item = dpdk_item_gtp_teid_modify;
	}
}

static int
dpdk_item_nisp_modify(struct dpdk_item_entry *entry,
		      const struct doca_flow_match *match)
{
	struct mlx5_flow_item_nisp *spec = GET_ITEM_SPEC(nisp);
	struct mlx5_flow_item_nisp *mask = GET_ITEM_MASK(nisp);
	int i;

	if (!utils_df_translate_is_nisp_ignored(&match->tun.nisp_hdr))
		for (i = 0; i < DOCA_FLOW_NISP_DWORD; i++)
			spec->dw[i] = mask->dw[i] & match->tun.nisp_hdr[i];
	return 0;
}

static void
dpdk_item_nisp_build(struct dpdk_item_entry *entry,
		     const struct doca_flow_match *match,
		     const struct doca_flow_match *match_mask)
{
	struct rte_flow_item *flow_item = entry->item;
	struct mlx5_flow_item_nisp *spec = GET_ITEM_SPEC(nisp);
	struct mlx5_flow_item_nisp *mask = GET_ITEM_MASK(nisp);

	static_assert(sizeof(match->tun.nisp_hdr) ==
		      sizeof(struct mlx5_flow_item_nisp),
		      "nisp tunnel and header mismatches nisp item");
	flow_item->type = RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP;
	if (utils_df_translate_is_nisp_ignored(&match->tun.nisp_hdr))
		return;
	memcpy(spec, &match->tun.nisp_hdr, sizeof(match->tun.nisp_hdr));
	if (match_mask)
		memcpy(mask, &match_mask->tun.nisp_hdr,
		       sizeof(match_mask->tun.nisp_hdr));
	else
		memset(mask, UINT8_MAX, sizeof(match_mask->tun.nisp_hdr));
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (utils_df_translate_is_nisp_changeable(&match->tun.nisp_hdr)) {
		entry->flags |= DPDK_MODIFY_NISP;
		entry->modify_item = dpdk_item_nisp_modify;
	}
}

static int
dpdk_item_audp_modify(struct dpdk_item_entry *entry,
		      const struct doca_flow_match *match)
{
	struct mlx5_flow_item_audp *spec = GET_ITEM_SPEC(audp);
	struct mlx5_flow_item_audp *mask = GET_ITEM_MASK(audp);
	int i;

	if (!utils_df_translate_is_audp_ignored(&match->tun.audp_hdr))
		for (i = 0; i < DOCA_FLOW_AUDP_DWORD; i++)
			spec->dw[i] = mask->dw[i] & match->tun.audp_hdr[i];
	return 0;
}

static void
dpdk_item_audp_build(struct dpdk_item_entry *entry,
		     const struct doca_flow_match *match,
		     const struct doca_flow_match *match_mask)
{
	struct rte_flow_item *flow_item = entry->item;
	struct mlx5_flow_item_audp *spec = GET_ITEM_SPEC(audp);
	struct mlx5_flow_item_audp *mask = GET_ITEM_MASK(audp);

	assert(sizeof(match->tun.audp_hdr) ==
	       sizeof(struct mlx5_flow_item_audp));
	flow_item->type = RTE_PMD_MLX5_FLOW_ITEM_TYPE_AUDP;
	if (utils_df_translate_is_audp_ignored(&match->tun.audp_hdr))
		return;
	memcpy(spec, &match->tun.audp_hdr, sizeof(match->tun.audp_hdr));
	if (match_mask)
		memcpy(mask, &match_mask->tun.audp_hdr,
		       sizeof(match_mask->tun.audp_hdr));
	else
		memset(mask, UINT8_MAX, sizeof(match_mask->tun.audp_hdr));
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (utils_df_translate_is_audp_changeable(&match->tun.audp_hdr)) {
		entry->flags |= DPDK_MODIFY_AUDP;
		entry->modify_item = dpdk_item_audp_modify;
	}
}

static int
dpdk_item_nisp_syndrome_modify(struct dpdk_item_entry *entry,
			       const struct doca_flow_match *match)
{
	struct mlx5_flow_item_nisp_syndrome *spec = GET_ITEM_SPEC(nisp_synd);
	struct mlx5_flow_item_nisp_syndrome *mask = GET_ITEM_MASK(nisp_synd);

	spec->syndrome = match->meta.nisp_syndrome & mask->syndrome;
	return 0;
}

static void
dpdk_item_nisp_syndrome_build(struct dpdk_pipe *pipe,
			      struct dpdk_pipe_q_ctx *pipe_q,
			      const struct doca_flow_match *match,
			      const struct doca_flow_match *match_mask)
{
	struct mlx5_flow_item_nisp_syndrome *spec;
	struct mlx5_flow_item_nisp_syndrome *mask;
	struct rte_flow_item *flow_item;
	struct dpdk_item_entry *entry;

	if (!match || !match->meta.nisp_syndrome)
		return;
	entry = &pipe_q->item_entry[pipe->nb_items++];
	flow_item = entry->item;
	flow_item->type = RTE_PMD_MLX5_FLOW_ITEM_TYPE_NISP_SYNDROME;
	spec = GET_ITEM_SPEC(nisp_synd);
	mask = GET_ITEM_MASK(nisp_synd);
	spec->syndrome = match->meta.nisp_syndrome;
	mask->syndrome = match_mask ?
			 match_mask->meta.nisp_syndrome : UINT8_MAX;
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (match->meta.nisp_syndrome == UINT8_MAX) {
		entry->flags |= DPDK_MODIFY_NISP_SYNDROME;
		entry->modify_item = dpdk_item_nisp_syndrome_modify;
	}
}

static int
dpdk_item_ipsec_modify(struct dpdk_item_entry *entry,
		       const struct doca_flow_match *match)
{
	struct rte_flow_item_esp *spec = GET_ITEM_SPEC(esp);
	struct rte_flow_item_esp *mask = GET_ITEM_MASK(esp);

	if (entry->flags & DPDK_MODIFY_ESP_SPI)
		spec->hdr.spi = match->tun.esp_spi & mask->hdr.spi;
	if (entry->flags & DPDK_MODIFY_ESP_SN)
		spec->hdr.seq = match->tun.esp_sn & mask->hdr.seq;
	return 0;
}

static void
dpdk_item_ipsec_build(struct dpdk_item_entry *entry,
		      const struct doca_flow_match *match,
		      const struct doca_flow_match *match_mask)
{
	struct rte_flow_item *flow_item = entry->item;
	struct rte_flow_item_esp *spec = GET_ITEM_SPEC(esp);
	struct rte_flow_item_esp *mask = GET_ITEM_MASK(esp);

	flow_item->type = RTE_FLOW_ITEM_TYPE_ESP;
	if (utils_df_translate_is_esp_ignored(&match->tun.esp_spi))
		return;
	spec->hdr.spi = match->tun.esp_spi;
	spec->hdr.seq = match->tun.esp_sn;

	if (match_mask) {
		mask->hdr.spi = match_mask->tun.esp_spi;
		mask->hdr.seq = match_mask->tun.esp_sn;
	} else {
		mask->hdr.spi = UINT32_MAX;
		mask->hdr.seq = 0;
	}
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (utils_field_get_type(&match->tun.esp_spi, sizeof(match->tun.esp_spi)) == UTILS_FIELD_TYPE_CHANGEABLE) {
		entry->flags |= DPDK_MODIFY_ESP_SPI;
		entry->modify_item = dpdk_item_ipsec_modify;
	}
	if (utils_field_get_type(&match->tun.esp_sn, sizeof(match->tun.esp_sn)) == UTILS_FIELD_TYPE_CHANGEABLE) {
		entry->flags |= DPDK_MODIFY_ESP_SN;
		entry->modify_item = dpdk_item_ipsec_modify;
	}
}

static int
dpdk_item_ipsec_syndrome_modify(struct dpdk_item_entry *entry,
				const struct doca_flow_match *match)
{
	struct mlx5_flow_item_ipsec_syndrome *spec = GET_ITEM_SPEC(ipsec_synd);
	struct mlx5_flow_item_ipsec_syndrome *mask = GET_ITEM_MASK(ipsec_synd);

	spec->syndrome = match->meta.ipsec_syndrome & mask->syndrome;
	return 0;
}

static void
dpdk_item_ipsec_syndrome_build(struct dpdk_pipe *pipe,
			       struct dpdk_pipe_q_ctx *pipe_q,
			       const struct doca_flow_match *match,
			       const struct doca_flow_match *match_mask)
{
	struct mlx5_flow_item_ipsec_syndrome *spec;
	struct mlx5_flow_item_ipsec_syndrome *mask;
	struct rte_flow_item *flow_item;
	struct dpdk_item_entry *entry;

	if (match->meta.ipsec_syndrome == 0)
		return;
	entry = &pipe_q->item_entry[pipe->nb_items++];
	flow_item = entry->item;
	flow_item->type = RTE_PMD_MLX5_FLOW_ITEM_TYPE_IPSEC_SYNDROME;
	spec = GET_ITEM_SPEC(ipsec_synd);
	mask = GET_ITEM_MASK(ipsec_synd);
	spec->syndrome = match->meta.ipsec_syndrome;
	mask->syndrome = match_mask ?
			 match_mask->meta.ipsec_syndrome : UINT8_MAX;
	flow_item->spec = spec;
	flow_item->mask = mask;
	if (match->meta.ipsec_syndrome == UINT8_MAX) {
		entry->flags |= DPDK_MODIFY_IPSEC_SYNDROME;
		entry->modify_item = dpdk_item_ipsec_syndrome_modify;
	}
}

static void
dpdk_item_end_build(struct dpdk_item_entry *entry)
{
	struct rte_flow_item *flow_item = entry->item;

	flow_item->type = RTE_FLOW_ITEM_TYPE_END;
}

static int
dpdk_item_tcp_modify(struct dpdk_item_entry *entry,
		     const struct doca_flow_match *match)
{
	struct rte_flow_item_tcp *spec = GET_ITEM_SPEC(tcp);
	struct rte_flow_item_tcp *mask = GET_ITEM_MASK(tcp);
	bool is_outer = entry->item_data.tcp.is_outer;
	rte_be16_t src_port = GET_MATCH_FIELD(src_port);
	rte_be16_t dst_port = GET_MATCH_FIELD(dst_port);
	uint8_t tcp_flags = GET_MATCH_FIELD(tcp_flags);

	if ((entry->flags & DPDK_MODIFY_SPORT) && src_port)
		spec->hdr.src_port = src_port & mask->hdr.src_port;
	if ((entry->flags & DPDK_MODIFY_DPORT) && dst_port)
		spec->hdr.dst_port = dst_port & mask->hdr.dst_port;
	if (entry->flags & DPDK_MODIFY_TCP_FLAGS)
		spec->hdr.tcp_flags = tcp_flags;
	return 0;
}

static void
dpdk_item_tcp_build(struct dpdk_item_entry *entry,
		    const struct doca_flow_match *match,
		    const struct doca_flow_match *match_mask,
		    bool is_outer)
{
	struct rte_flow_item_tcp *spec = &entry->item_data.tcp.spec;
	struct rte_flow_item_tcp *mask = &entry->item_data.tcp.mask;
	struct rte_flow_item *item = entry->item;
	rte_be16_t src_port = GET_MATCH_FIELD(src_port);
	rte_be16_t dst_port = GET_MATCH_FIELD(dst_port);
	uint8_t tcp_flags = GET_MATCH_FIELD(tcp_flags);
	rte_be16_t mask_src_port = match_mask ?
		GET_MASK_FIELD(src_port) : UINT16_MAX;
	rte_be16_t mask_dst_port = match_mask ?
		GET_MASK_FIELD(dst_port) : UINT16_MAX;
	uint8_t mask_tcp_flags = match_mask ?
		GET_MASK_FIELD(tcp_flags) : UINT8_MAX;

	entry->item_data.tcp.is_outer = is_outer;
	item->type = RTE_FLOW_ITEM_TYPE_TCP;
	if (src_port) {
		spec->hdr.src_port = src_port;
		mask->hdr.src_port = mask_src_port;
		if (src_port == UINT16_MAX)
			entry->flags |= DPDK_MODIFY_SPORT;
	}
	if (dst_port) {
		spec->hdr.dst_port = dst_port;
		mask->hdr.dst_port = mask_dst_port;
		if (dst_port == UINT16_MAX)
			entry->flags |= DPDK_MODIFY_DPORT;
	}
	if (tcp_flags) {
		spec->hdr.tcp_flags = tcp_flags;
		mask->hdr.tcp_flags = mask_tcp_flags;
		if (tcp_flags == UINT8_MAX)
			entry->flags |= DPDK_MODIFY_TCP_FLAGS;
	}
	if (src_port == 0 && dst_port == 0 && tcp_flags == 0)
		return;
	item->spec = spec;
	item->mask = mask;
	if (entry->flags)
		entry->modify_item = dpdk_item_tcp_modify;
}

static int
dpdk_item_udp_modify(struct dpdk_item_entry *entry,
		     const struct doca_flow_match *match)
{
	struct rte_flow_item_udp *spec = GET_ITEM_SPEC(udp);
	struct rte_flow_item_udp *mask = GET_ITEM_MASK(udp);
	bool is_outer = entry->item_data.udp.is_outer;
	rte_be16_t src_port = GET_MATCH_FIELD(src_port);
	rte_be16_t dst_port = GET_MATCH_FIELD(dst_port);

	if ((entry->flags & DPDK_MODIFY_SPORT) && src_port)
		spec->hdr.src_port = src_port & mask->hdr.src_port;
	if ((entry->flags & DPDK_MODIFY_DPORT) && dst_port)
		spec->hdr.dst_port = dst_port & mask->hdr.dst_port;
	return 0;
}

static void
dpdk_item_udp_build(struct dpdk_item_entry *entry,
		    const struct doca_flow_match *match,
		    const struct doca_flow_match *match_mask,
		    bool is_outer)
{
	struct rte_flow_item_udp *spec = GET_ITEM_SPEC(udp);
	struct rte_flow_item_udp *mask = GET_ITEM_MASK(udp);
	struct rte_flow_item *item = entry->item;
	rte_be16_t src_port = GET_MATCH_FIELD(src_port);
	rte_be16_t dst_port = GET_MATCH_FIELD(dst_port);
	rte_be16_t mask_src_port = match_mask ?
		GET_MASK_FIELD(src_port) : UINT16_MAX;
	rte_be16_t mask_dst_port = match_mask ?
		GET_MASK_FIELD(dst_port) : UINT16_MAX;

	entry->item_data.udp.is_outer = is_outer;
	item->type = RTE_FLOW_ITEM_TYPE_UDP;
	if (!src_port && !dst_port)
		return;
	if (src_port) {
		spec->hdr.src_port = src_port;
		mask->hdr.src_port = mask_src_port;
		if (src_port == UINT16_MAX)
			entry->flags |= DPDK_MODIFY_SPORT;
	}
	if (dst_port) {
		spec->hdr.dst_port = dst_port;
		mask->hdr.dst_port = mask_dst_port;
		if (dst_port == UINT16_MAX)
			entry->flags |= DPDK_MODIFY_DPORT;
	}
	item->spec = spec;
	item->mask = mask;
	if (entry->flags)
		entry->modify_item = dpdk_item_udp_modify;
}

static int
dpdk_item_port_modify(struct dpdk_item_entry *entry, const struct doca_flow_match *match)
{
	struct rte_flow_item_ethdev *spec = GET_ITEM_SPEC(port);
	struct rte_flow_item_ethdev *mask = GET_ITEM_MASK(port);

	spec->port_id = match->meta.port_meta & mask->port_id;
	return 0;
}

int
dpdk_pipe_items_meta_port_parse(const struct doca_flow_match *match,
				const struct doca_flow_match *mask,
				uint32_t *port_value, uint32_t *port_mask)
{
	*port_value = match->meta.port_meta;
	*port_mask = mask ? mask->meta.port_meta : UINT32_MAX;

	if (*port_mask == 0 || (!mask && (*port_value == 0)))
		return 0;
	return 1;
}

static void
dpdk_item_port_build(struct dpdk_pipe *pipe,
		     struct dpdk_pipe_q_ctx *pipe_q,
		     const struct doca_flow_match *match,
		     const struct doca_flow_match *mask)
{
	uint32_t port_value, port_mask;
	struct rte_flow_item *flow_item;
	struct dpdk_item_entry *entry;

	if (!dpdk_pipe_items_meta_port_parse(match, mask, &port_value, &port_mask))
		return;
	entry = &pipe_q->item_entry[pipe->nb_items++];
	flow_item = entry->item;
	flow_item->type = RTE_FLOW_ITEM_TYPE_REPRESENTED_PORT;
	entry->item_data.port.spec.port_id = port_value;
	entry->item_data.port.mask.port_id = port_mask;
	flow_item->spec = &entry->item_data.port.spec;
	flow_item->mask = &entry->item_data.port.mask;

	if (port_value == UINT32_MAX ||
	   (port_mask == UINT32_MAX && port_value == 0))
		entry->modify_item = dpdk_item_port_modify;
}

static void
dpdk_item_meter_color_build(struct dpdk_item_entry *entry,
			    const struct doca_flow_match *match)
{
	enum rte_color color;

	color = (match->flags & DPDK_MATCH_METER_COLOR_GREEN) ? RTE_COLOR_GREEN :
								RTE_COLOR_RED;

	struct rte_flow_item_meter_color *spec = &entry->item_data.color.spec;
	struct rte_flow_item_meter_color *mask = &entry->item_data.color.mask;
	struct rte_flow_item *item = entry->item;

	spec->color = color;
	mask->color = UINT32_MAX;

	item->type = RTE_FLOW_ITEM_TYPE_METER_COLOR;
	item->spec = spec;
	item->mask = mask;
}

int
dpdk_pipe_matches_build(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
			const struct doca_flow_match *match, const struct doca_flow_match *mask)
{
#define NEXT_ITEM (&pipe_q->item_entry[dpdk_pipe->nb_items++])
	bool is_outer = true;
	uint8_t ip_type, l4_type;
	uint16_t next_proto;

	if (!match)
		goto end_item;

	dpdk_pipe->nb_items = 0;
	pipe_q->l4_type = match->out_l4_type;
	pipe_q->ip_type = dpdk_ether_type_get(match, is_outer) == RTE_BE16(RTE_ETHER_TYPE_IPV6) ?
			DOCA_FLOW_IP6_ADDR : DOCA_FLOW_IP4_ADDR;

	if (mask)
		dpdk_meta_flow_item_build(dpdk_pipe, pipe_q, mask);
	else if (utils_field_get_type(&match->meta, DOCA_FLOW_META_MAX) != UTILS_FIELD_TYPE_IGNORED)
		DOCA_LOG_WARN("Meta data doens't support constant match, ignore");

	if (match->flags & (DPDK_MATCH_METER_COLOR_GREEN | DPDK_MATCH_METER_COLOR_RED))
		dpdk_item_meter_color_build(NEXT_ITEM, match);

	dpdk_item_port_build(dpdk_pipe, pipe_q, match, mask);
	dpdk_item_nisp_syndrome_build(dpdk_pipe, pipe_q, match, mask);
	dpdk_item_ipsec_syndrome_build(dpdk_pipe, pipe_q, match, mask);
	dpdk_item_eth_build(NEXT_ITEM, match, mask, is_outer);
	if (GET_MATCH_FIELD(vlan_tci))
		dpdk_item_vlan_build(NEXT_ITEM, match, mask, is_outer);
	ip_type = dpdk_match_ip_type_get(match, is_outer);
	if (ip_type == DOCA_FLOW_IP4_ADDR)
		dpdk_item_ipv4_build(NEXT_ITEM, match, mask, is_outer);
	else if (ip_type == DOCA_FLOW_IP6_ADDR)
		dpdk_item_ipv6_build(NEXT_ITEM, match, mask, is_outer);
	else
		goto end_item;
	if (match->tun.type != DOCA_FLOW_TUN_NONE) {
		switch (match->tun.type) {
		case DOCA_FLOW_TUN_VXLAN:
			dpdk_item_udp_build(NEXT_ITEM, match, mask, is_outer);
			dpdk_item_vxlan_build(NEXT_ITEM, match, mask);
			goto inner_l2_item;
		case DOCA_FLOW_TUN_GRE:
			dpdk_item_gre_build(NEXT_ITEM, match, mask);
			if (match->tun.gre_key || match->tun.key_present)
				dpdk_item_gre_key_build(NEXT_ITEM, match);
			next_proto = mask ? match->tun.protocol & mask->tun.protocol :
				     match->tun.protocol;
			switch (next_proto) {
			case RTE_BE16(DOCA_ETHER_TYPE_TEB):
				goto inner_l2_item;
			case RTE_BE16(DOCA_ETHER_TYPE_IPV4):
			case RTE_BE16(DOCA_ETHER_TYPE_IPV6):
			case 0:
				goto inner_l3_item;
			/* No MPLS for now */
			default:
				DOCA_LOG_RATE_LIMIT_ERR("failed to build pipe matches - unsupported match GRE protocol %u",
							 match->tun.protocol);
				return -1;
			}
		case DOCA_FLOW_TUN_GTPU:
			dpdk_item_udp_build(NEXT_ITEM, match, mask, is_outer);
			dpdk_item_gtp_build(NEXT_ITEM, match, mask);
			goto inner_l3_item;
		case DOCA_FLOW_TUN_NISP:
			dpdk_item_udp_build(NEXT_ITEM, match, mask, is_outer);
			dpdk_item_nisp_build(NEXT_ITEM, match, mask);
			goto inner_l3_item;
		case DOCA_FLOW_TUN_AUDP:
			dpdk_item_udp_build(NEXT_ITEM, match, mask, is_outer);
			dpdk_item_audp_build(NEXT_ITEM, match, mask);
			goto inner_l3_item;
		case DOCA_FLOW_TUN_ESP:
			l4_type = GET_MATCH_FIELD(l4_type);
			if (l4_type == DOCA_PROTO_UDP)
				dpdk_item_udp_build(NEXT_ITEM, match, mask, is_outer);
			dpdk_item_ipsec_build(NEXT_ITEM, match, mask);
			goto inner_l3_item;
		default:
			DOCA_LOG_RATE_LIMIT_ERR("failed to build pipe matches - unsupported match tunnel type 0x%x",
						match->tun.type);
			return -1;
		}
inner_l2_item:
		is_outer = false;
		if (dpdk_ether_type_get(match, is_outer) ||
		    !utils_df_translate_is_mac_ignored(GET_MATCH_FIELD(src_mac)) ||
		    !utils_df_translate_is_mac_ignored(GET_MATCH_FIELD(dst_mac)))
			dpdk_item_eth_build(NEXT_ITEM, match, mask, is_outer);
		else
			goto end_item;
		if (GET_MATCH_FIELD(vlan_tci))
			dpdk_item_vlan_build(NEXT_ITEM, match, mask, is_outer);
inner_l3_item:
		is_outer = false;
		switch (dpdk_match_ip_type_get(match, is_outer)) {
		case DOCA_FLOW_IP4_ADDR:
			dpdk_item_ipv4_build(NEXT_ITEM, match, mask, is_outer);
			break;
		case DOCA_FLOW_IP6_ADDR:
			dpdk_item_ipv6_build(NEXT_ITEM, match, mask, is_outer);
			break;
		default:
			goto end_item;
		}
	}
	l4_type = GET_MATCH_FIELD(l4_type);
	if (!utils_df_translate_is_l4_type_specific(l4_type))
		goto end_item;
	switch (l4_type) {
	case DOCA_PROTO_TCP:
		dpdk_item_tcp_build(NEXT_ITEM, match, mask, is_outer);
		break;
	case DOCA_PROTO_UDP:
		dpdk_item_udp_build(NEXT_ITEM, match, mask, is_outer);
		break;
	default:
		DOCA_LOG_ERR("failed to build pipe matches - unsupported l4 type %d for %s match",
			     l4_type, is_outer ? "outer" : "inner");
		return -1;
	}
end_item:
	dpdk_item_end_build(NEXT_ITEM);
	return 0;
}

int
dpdk_pipe_matches_modify(struct dpdk_pipe *dpdk_pipe, struct dpdk_pipe_q_ctx *pipe_q,
			 const struct doca_flow_match *match)
{
	struct dpdk_item_entry *item_entry;
	int idex, ret;

	for (idex = 0; idex < dpdk_pipe->nb_items; idex++) {
		item_entry = &pipe_q->item_entry[idex];
		if (item_entry->modify_item == NULL)
			continue;
		ret = item_entry->modify_item(item_entry, match);
		if (ret) {
			DOCA_LOG_RATE_LIMIT_ERR("failed to modify item on index %d and flags 0x%lx", idex,
						item_entry->flags);
			return ret;
		}
	}
	return 0;
}
