/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdarg.h>
#include <stdio.h>

#include <rte_ethdev.h>
#include <rte_gre.h>
#include <rte_gtp.h>
#include <rte_mtr.h>
#include <rte_vxlan.h>
#include <rte_hash_crc.h>

#include <doca_log.h>

#include "engine_model.h"
#include "doca_flow_layer.h"
#include "private/common_defines.h"

#include "utils_df_translate.h"
#include "dpdk_engine.h"
#include "dpdk_layer.h"
#include "dpdk_dbg.h"
#include "doca_flow_priv.h"

#include "dpdk_meter_pipe.h"
#include "pipe_lpm.h"

#include "dpdk_pipe_common.h"
#include "dpdk_pipe_basic.h"
#include "dpdk_pipe_control.h"
#include "dpdk_pipe_ct.h"
#include "dpdk_pipe_hash.h"
#include "dpdk_pipe_lpm.h"
#include "dpdk_pipe_ordered_list.h"
#include "dpdk_table_manager.h"

#include "encap_table.h"
#include "priv_doca_mempool.h"
#include "flow_age.h"
#include "id_pool.h"
#include "dpdk_utilities.h"

#define FWD_MISS_ACTIONS (2)
#define FWD_MISS_ITEMS  (3)

#define DPDK_PIPE_INDEX_NONE UINT32_MAX

DOCA_LOG_REGISTER(dpdk_pipe_legacy);

static struct dpdk_pipe_type_ops *pipe_type_ops[ENGINE_PIPE_MAX];

struct dpdk_pipe_q_ctx *
dpdk_pipe_q_ctx_get(struct doca_flow_pipe *pipe, uint16_t q)
{
	return &pipe->dpdk_pipe.queues[q].pipe_q;
}

struct dpdk_pipe_queue *
dpdk_pipe_by_qidx_get(struct doca_flow_pipe *pipe, uint16_t q)
{
	return &pipe->dpdk_pipe.queues[q];
}

static int
validate_ops(enum engine_pipe_type pipe_type, struct dpdk_pipe_type_ops *ops)
{
	if (unlikely(!ops))
		return -EINVAL;

	RTE_SET_USED(ops->module_init); /* optional */

	if (unlikely(!ops->pipe_legacy_verify))
		return -EINVAL;
	if (unlikely(!ops->pipe_legacy_alloc))
		return -EINVAL;
	if (unlikely(!ops->pipe_legacy_build))
		return -EINVAL;
	if (unlikely(!ops->pipe_legacy_submit))
		return -EINVAL;
	if (unlikely(!ops->pipe_legacy_free))
		return -EINVAL;

	if (unlikely(!ops->pipe_queue_verify))
		return -EINVAL;
	if (unlikely(!ops->pipe_queue_modify))
		return -EINVAL;
	if (unlikely(!ops->pipe_queue_submit))
		return -EINVAL;
	if (unlikely(!ops->pipe_queue_remove))
		return -EINVAL;
	if (unlikely(!ops->pipe_queue_cleanup))
		return -EINVAL;

	return 0;
}

static int
register_ops(enum engine_pipe_type pipe_type, struct dpdk_pipe_type_ops *ops)
{
	int rc;

	if (unlikely(pipe_type >= ENGINE_PIPE_MAX)) {
		DOCA_LOG_CRIT("failed registering pipe ops for type %u", pipe_type);
		return -EINVAL;
	}

	rc = validate_ops(pipe_type, ops);
	if (unlikely(rc)) {
		DOCA_LOG_CRIT("failed registering pipe ops for type %u - invalid ops", pipe_type);
		return rc;
	}

	pipe_type_ops[pipe_type] = ops;
	return 0;
}

static struct dpdk_pipe_type_ops *
get_ops_by_type(enum engine_pipe_type type)
{
	if (unlikely(type >= ENGINE_PIPE_MAX))
		return NULL;

	return pipe_type_ops[type];
}

int
dpdk_pipe_legacy_init_early(void)
{
	int rc;

	memset(pipe_type_ops, 0, sizeof(pipe_type_ops));

	rc = register_ops(ENGINE_PIPE_BASIC, dpdk_pipe_basic_get_ops());
	if (unlikely(rc))
		return rc;

	rc = register_ops(ENGINE_PIPE_CONTROL, dpdk_pipe_control_get_ops());
	if (unlikely(rc))
		return rc;

	rc = register_ops(ENGINE_PIPE_LPM, dpdk_pipe_lpm_get_ops());
	if (unlikely(rc))
		return rc;

	rc = register_ops(ENGINE_PIPE_CT, dpdk_pipe_ct_get_ops());
	if (unlikely(rc))
		return rc;

	rc = register_ops(ENGINE_PIPE_HASH, dpdk_pipe_hash_get_ops());
	if (unlikely(rc))
		return rc;

	rc = register_ops(ENGINE_PIPE_ORDERED_LIST, dpdk_pipe_ordered_list_get_ops());
	if (unlikely(rc))
		return rc;

	dpdk_table_manager_init();
	return 0;
}

int
dpdk_pipe_legacy_init_late(void)
{
	uint32_t i;

	for (i = 0; i < RTE_DIM(pipe_type_ops); i++) {
		struct dpdk_pipe_type_ops *ops = pipe_type_ops[i];
		int rc;

		if (!ops || !ops->module_init)
			continue;
		rc = ops->module_init();
		if (unlikely(rc)) {
			DOCA_LOG_ERR("Failed to initialize module for pipe type %d (rc=%d)", i, rc);
			return rc;
		}
	}
	return 0;
}

uint64_t
dpdk_pipe_generate_unique_id(const struct doca_flow_port *port, struct doca_flow_pipe *pipe)
{
	uint64_t unique_id;
	uint16_t port_id;

	port_id = port->port_id;
	if (pipe->is_root) {
		unique_id = UINT32_MAX;
		unique_id = (unique_id << 32) | (pipe->root_id);
	} else {
		unique_id = port_id;
		unique_id = (unique_id << 32) | (pipe->group_id);
	}

	return unique_id;
}

struct doca_flow_pipe_entry*
dpdk_pipe_entry_alloc(struct doca_flow_pipe *pipe, uint16_t pipe_queue)
{
	struct doca_flow_pipe_entry *entry;

	entry = priv_doca_mempool_alloc(pipe->flow_entry_pool, pipe_queue);
	if (entry == NULL) {
		DOCA_DLOG_ERR("alloc new pipe entry from memory pool fail, pipe:%s, queue:%u",
			      pipe->name, pipe_queue);
		return NULL;
	}
	memset(entry, 0x0, sizeof(*entry));
	entry->encap_id = -1;
	entry->pipe_queue = pipe_queue;
	entry->pipe = pipe;
	return entry;
}

static int
dpdk_meter_id_destroy(int port_id, uint32_t meter_id)
{
	int ret;
	struct rte_mtr_error error;
	uint32_t dpdk_meter_id = DPDK_ID(meter_id);

	ret = rte_mtr_destroy(port_id, dpdk_meter_id, &error);
	if (ret != 0) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"Port %u delete mrt idx(%d) error(%d) message: %s",
			port_id, dpdk_meter_id, error.type,
			error.message ? error.message : "(no stated reason)");
	}
	return ret;
}

void
dpdk_entry_update_status(struct doca_flow_pipe_entry *entry, enum dpdk_flow_request_status status)
{
	switch (status) {
	case DPDK_FLOW_REQ_STATUS_OK:
		entry->status = DOCA_FLOW_ENTRY_STATUS_SUCCESS;
		break;
	default:
		entry->status = DOCA_FLOW_ENTRY_STATUS_ERROR;
		break;
	}
}

static void
rm_entry_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
		       void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;
	struct dpdk_pipe_type_ops *ops;
	struct doca_flow_pipe *pipe;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Entry removal completion got null entry");
		return;
	}

	pipe = (struct doca_flow_pipe *)entry->pipe;
	if (unlikely(!pipe)) {
		DOCA_LOG_CRIT("Entry removal completion got null pipe");
		return;
	}

	DOCA_DLOG_DBG("entry %p remove completed with status %d", entry, entry->status);

	dpdk_entry_update_status(entry, status);

	dpdk_engine_invoke_entry_process_cb(entry, DOCA_FLOW_ENTRY_OP_DEL);

	ops = get_ops_by_type(pipe->type);
	if (likely(ops))
		ops->pipe_queue_cleanup(entry);
}

static void
dpdk_pipe_entry_flow_destroy(struct doca_flow_port *port, void *usr_ctx,
			     struct doca_flow_pipe_entry *entry)
{
	struct doca_flow_pipe *pipe = entry->pipe;
	struct dpdk_pipe_type_ops *ops;
	uint32_t *meter_id = entry->meter_id;
	uint16_t queue_id = entry->pipe_queue;

	if (dpdk_port_legacy_is_invalid(port))
		return;

	entry->meter_id = NULL;
	entry->profile_id = 0;

	ops = get_ops_by_type(pipe->type);
	if (likely(ops))
		ops->pipe_queue_remove(entry, rm_entry_completion_cb, usr_ctx);

	/* Entry may be released at this point and cannot be used. */
	if (meter_id) {
		dpdk_meter_id_destroy(port->port_id, *meter_id);
		id_pool_free(port->meter_pool, queue_id, meter_id);
	}
}

static inline int
_dpdk_pipe_entry_destroy(struct doca_flow_port *port,
			 struct doca_flow_pipe_entry *entry,
			 void *usr_ctx,
			 struct dpdk_pipe_queue *pipe_by_qidx)
{
	if (dpdk_port_legacy_is_invalid(port))
		return -EINVAL;

	if (entry->age_item) {
		flow_age_item_release(entry);
		entry->age_item = NULL;
	}

	dpdk_pipe_entry_flow_destroy(port, usr_ctx, entry);
	return 0;
}

int
dpdk_pipe_entry_remove(uint16_t pipe_queue, void *usr_ctx,
			struct doca_flow_pipe_entry *entry)
{
	struct doca_flow_pipe *pipe = entry->pipe;
	struct dpdk_pipe_queue *pipe_by_qidx;

	/*when create entry, we know the pipe_queue*/
	if (pipe == NULL ||
		pipe->nb_queues < pipe_queue ||
		entry->pipe_queue != pipe_queue)
		return -EINVAL;
	if (pipe->type == ENGINE_PIPE_LPM) {
		DOCA_LOG_WARN("this doca version doesn't support remove lpm pipe entry");
		return -ENOTSUP;
	}

	pipe_by_qidx = dpdk_pipe_by_qidx_get(pipe, pipe_queue);
	if (unlikely(pipe->type == ENGINE_PIPE_CONTROL)) {
		rte_spinlock_lock(&pipe->dpdk_pipe.ctrl_entry_lock);
		pipe_by_qidx->nb_entries--;
		rte_spinlock_unlock(&pipe->dpdk_pipe.ctrl_entry_lock);
		LIST_REMOVE(entry, next);
	} else {
		dpdk_pipe_common_entry_detach(pipe_queue, pipe, entry);
	}
	return _dpdk_pipe_entry_destroy(doca_flow_layer_get_pipe_port(pipe), entry, usr_ctx,
					pipe_by_qidx);
}

static void
add_entry_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Entry addition completion got null entry");
		return;
	}

	DOCA_DLOG_DBG("entry %p add completed with status %d", entry, entry->status);

	dpdk_entry_update_status(entry, status);

	dpdk_engine_invoke_entry_process_cb(entry, DOCA_FLOW_ENTRY_OP_ADD);
}

static struct doca_flow_pipe_entry *
add_entry(uint16_t pipe_queue,
	  uint32_t priority,
	  struct doca_flow_pipe *pipe,
	  uint32_t index,
	  const struct doca_flow_match *match,
	  const struct doca_flow_match *match_mask,
	  const struct doca_flow_actions *actions,
	  const struct doca_flow_action_descs *action_descs,
	  const struct doca_flow_monitor *monitor,
	  const struct doca_flow_ordered_list *ordered_list,
	  const struct doca_flow_fwd *fwd,
	  dpdk_flow_completion_cb entry_completion_cb,
	  uint32_t flags,
	  void *usr_ctx)
{
	struct dpdk_uds_cfg uds_cfg = {0};
	struct dpdk_pipe_type_ops *ops;
	struct doca_flow_pipe_entry *entry;
	uint32_t nb_actions;
	uint8_t act_arr_idx = 0;
	int rc;

	if (unlikely(!pipe)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - invalid pipe");
		return NULL;
	}

	if (unlikely(pipe->type >= ENGINE_PIPE_MAX)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - invalid pipe type %u", pipe->type);
		return NULL;
	}

	if (unlikely(pipe_queue >= pipe->nb_queues)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - invalid queue id %u", pipe_queue);
		return NULL;
	}

	nb_actions = pipe->dpdk_pipe.nb_actions_array_len;
	act_arr_idx = (actions) ? actions->action_idx : 0;
	if (unlikely(act_arr_idx >= nb_actions)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - invalid act_arr_idx %u, nb_actions %u", act_arr_idx, nb_actions);
		return NULL;
	}

	ops = get_ops_by_type(pipe->type);
	if (unlikely(!ops)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - undefined pipe type %u",
					pipe->type);
		return NULL;
	}

	entry = dpdk_pipe_entry_alloc(pipe, pipe_queue);
	if (unlikely(!entry)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - entry pool depleted on pipe %s",
					pipe->name);
		return NULL;
	}

	uds_cfg.index = index;
	uds_cfg.match = match;
	uds_cfg.match_mask = match_mask;
	uds_cfg.monitor = monitor;
	uds_cfg.ordered_lists = &ordered_list;
	uds_cfg.fwd = fwd;

	const struct doca_flow_actions *actions_set[nb_actions];
	const struct doca_flow_action_descs *action_descs_set[nb_actions];

	actions_set[act_arr_idx] = actions;
	action_descs_set[act_arr_idx] = action_descs;
	uds_cfg.actions = actions_set;
	uds_cfg.action_descs = action_descs_set;

	rc = ops->pipe_queue_verify(pipe, pipe_queue, act_arr_idx, priority, &uds_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - verify, rc=%d", rc);
		goto free_pipe_entry;
	}

	rc = ops->pipe_queue_modify(pipe, entry, pipe_queue, act_arr_idx, priority, &uds_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - modify, rc=%d", rc);
		goto free_pipe_entry;
	}

	rc = ops->pipe_queue_submit(pipe, entry, pipe_queue, act_arr_idx,
			entry_completion_cb, usr_ctx,
			flags == DOCA_FLOW_NO_WAIT);
	if (unlikely(rc)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to create entry - submit, rc=%d", rc);
		dpdk_pipe_common_entry_detach(pipe_queue, pipe, entry);
		goto free_pipe_entry;
	}

	return entry;

free_pipe_entry:
	ops->pipe_queue_cleanup(entry);
	return NULL;
}

struct doca_flow_pipe_entry *
dpdk_pipe_entry_add(uint16_t pipe_queue,
		    uint32_t priority,
		    struct doca_flow_pipe *pipe,
		    const struct doca_flow_match *match,
		    const struct doca_flow_match *match_mask,
		    const struct doca_flow_actions *actions,
		    const struct doca_flow_action_descs *action_descs,
		    const struct doca_flow_monitor *monitor,
		    const struct doca_flow_fwd *fwd,
		    uint32_t flags,
		    void *usr_ctx)
{
	return add_entry(pipe_queue, priority, pipe, DPDK_PIPE_INDEX_NONE,
			 match, match_mask, actions, action_descs, monitor,
			 NULL /* ordered_list */, fwd, add_entry_completion_cb,
			 flags, usr_ctx);
}

struct doca_flow_pipe_entry *
dpdk_pipe_entry_add_with_index(uint16_t pipe_queue,
			       struct doca_flow_pipe *pipe,
			       uint32_t index,
			       const struct doca_flow_actions *actions,
			       const struct doca_flow_action_descs *action_descs,
			       const struct doca_flow_monitor *monitor,
			       const struct doca_flow_ordered_list *ordered_list,
			       const struct doca_flow_fwd *fwd,
			       uint32_t flags,
			       void *usr_ctx)
{
	return add_entry(pipe_queue, 0 /* priority */, pipe, index, NULL /* match */,
			 NULL /* match_mask */, actions, action_descs, monitor, ordered_list,
			 fwd, add_entry_completion_cb, flags, usr_ctx);
}

struct doca_flow_pipe_entry *
dpdk_pipe_entry_add_with_cb(uint16_t pipe_queue,
			    uint32_t priority,
			    struct doca_flow_pipe *pipe,
			    const struct doca_flow_match *match,
			    const struct doca_flow_match *match_mask,
			    const struct doca_flow_actions *actions,
			    const struct doca_flow_action_descs *action_descs,
			    const struct doca_flow_monitor *monitor,
			    const struct doca_flow_fwd *fwd,
			    dpdk_flow_completion_cb entry_completion_cb,
			    uint32_t flags,
			    void *usr_ctx)
{
	return add_entry(pipe_queue, priority, pipe, DPDK_PIPE_INDEX_NONE,
			 match, match_mask, actions, action_descs, monitor,
			 NULL /* ordered_list */, fwd, entry_completion_cb, flags, usr_ctx);
}

static void
dpdk_pipe_rte_flow_dump(const char *name, struct dpdk_pipe *dpdk_pipe,
			struct dpdk_pipe_q_ctx *pipe_q, FILE *f)
{
	struct doca_flow_pipe *pipe = DOCA_PIPE(dpdk_pipe);
	struct doca_flow_port *port;
	int i;

	if (unlikely(!pipe))
		return;

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port))
		return;

	for (i = 0; i < dpdk_pipe->nb_actions_array_len; i++)
		dpdk_dbg_rte_flow_dump(name, port->port_id, &pipe_q->attr, pipe_q->items,
				       pipe_q->action_ctx[i].actions, f);
}

void
dpdk_pipe_set_root_id(struct doca_flow_pipe *pipe)
{
	static uint32_t root_id = 1;

	dpdk_engine_mgmt_id_lock();
	pipe->root_id = root_id++;
	dpdk_engine_mgmt_id_unlock();
}

static void
add_miss_entry_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			     void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("Miss entry addition completion got null entry");
		return;
	}

	DOCA_DLOG_DBG("miss entry %p add completed with status %d", entry, entry->status);

	dpdk_entry_update_status(entry, status);
}

static int
dpdk_fwd_miss_entry_add(const struct doca_flow_pipe_cfg *cfg,
			const struct dpdk_pipe_q_ctx *pipe_q,
			const struct doca_flow_fwd *fwd_miss,
			struct doca_flow_pipe_entry *entry)
{
	struct doca_flow_port *port = cfg->port;
	struct dpdk_flow_params dpdk_flow_params = {0};
	struct dpdk_flow_request req = {0};
	struct dpdk_flow_queue *queue;
	struct rte_flow_action actions[FWD_MISS_ACTIONS];
	struct rte_flow_item items[FWD_MISS_ITEMS];
	struct rte_flow_action_jump jump;
	struct rte_flow_attr attr;
	struct dpdk_table res = {0};
	struct dpdk_hws_item *hws_item = NULL;
	uint16_t index = 0;
	int rc;

	queue = dpdk_port_get_flow_queue(port->dpdk_port, entry->pipe_queue);
	if (unlikely(!queue)) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating miss flow - invalid queue %u",
					entry->pipe_queue);
		return -EINVAL;
	}

	memset(actions, 0x0, sizeof(actions));
	memset(items, 0x0, sizeof(items));

	/* build attr */
	memcpy(&attr, &pipe_q->attr, sizeof(attr));

	/* build items */
	items[index++].type = RTE_FLOW_ITEM_TYPE_ETH;
	if (attr.transfer) {
		struct rte_flow_item_ethdev spec, mask;
		uint32_t port_value, port_mask;

		if (cfg->match && dpdk_pipe_items_meta_port_parse(cfg->match, cfg->match_mask,
								  &port_value, &port_mask)) {
			items[index].type = RTE_FLOW_ITEM_TYPE_REPRESENTED_PORT;
			spec.port_id = port_value;
			mask.port_id = port_mask;
			items[index].spec = &spec;
			items[index].mask = &mask;
			index++;
		}
	}
	items[index++].type = RTE_FLOW_ITEM_TYPE_END;

	/* build actions */
	if (fwd_miss->type == DOCA_FLOW_FWD_DROP) {
		attr.priority = MAX_FLOW_PRIO - 1;
		actions[0].type = RTE_FLOW_ACTION_TYPE_DROP;
		actions[0].conf = NULL;
	} else if (fwd_miss->type == DOCA_FLOW_FWD_PIPE) {
		attr.priority = MAX_FLOW_PRIO - 1;
		rc = dpdk_engine_next_pipe_group_id_get(fwd_miss->next_pipe, &jump.group);
		if (rc)
			return -EINVAL;
		actions[0].type = RTE_FLOW_ACTION_TYPE_JUMP;
		actions[0].conf = &jump;
	} else {
		return -ENOTSUP;
	}

	actions[1].type = RTE_FLOW_ACTION_TYPE_END;
	actions[1].conf = NULL;

	rc = dpdk_engine_pipe_get_ctrl_table(port, NULL, NULL, items, actions, &attr, &res,
					     &hws_item);
	if (rc) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating entry - table create rc=%d", rc);
		return -EINVAL;
	}

	dpdk_flow_params.items = items;
	dpdk_flow_params.actions = actions;
	dpdk_flow_params.table = res.table;
	dpdk_flow_params.item_template_index = res.pattern_template_index;
	dpdk_flow_params.action_template_index = res.action_template_index;
	dpdk_flow_params.ingress = attr.ingress;
	dpdk_flow_params.egress = attr.egress;
	dpdk_flow_params.transfer = attr.transfer;
	dpdk_flow_params.group = attr.group;
	dpdk_flow_params.priority = attr.priority;

	req.tracker = &entry->tracker;
	req.cb = add_miss_entry_completion_cb;
	req.callback_ctx = entry;
	req.persistent = true;
	req.wait_for_bulk = false;

	rc = dpdk_flow_create(queue, &dpdk_flow_params, &req);
	if (rc) {
		DOCA_LOG_RATE_LIMIT_ERR("failed creating entry - flow create rc=%d", rc);
		return rc;
	}
	if (dpdk_engine_mode_is_hws())
		hws_item->flow = dpdk_flow_tracker_get_flow(&entry->tracker);

	return 0;
}

int
dpdk_pipe_fwd_miss_handle(const struct doca_flow_pipe_cfg *cfg, struct doca_flow_pipe *pipe,
			  const struct doca_flow_fwd *fwd_miss)
{
	struct doca_flow_pipe_entry *fwd_miss_entry = NULL;
	int rc;

	if (fwd_miss->type != DOCA_FLOW_FWD_DROP &&
	    fwd_miss->type != DOCA_FLOW_FWD_PIPE) {
		DOCA_DLOG_ERR("handle fwd miss failed, invalid fwd_miss type %d.",
			fwd_miss->type);
		return -EINVAL;
	}

	fwd_miss_entry = (struct doca_flow_pipe_entry *)
		dpdk_pipe_entry_alloc(pipe, dpdk_engine_mgmt_id_get());
	if (!fwd_miss_entry) {
		DOCA_DLOG_ERR("allocate entry failed in handle_fwd_miss.");
		return -ENOMEM;
	}

	dpdk_engine_mgmt_id_lock();
	rc = dpdk_fwd_miss_entry_add(cfg, dpdk_pipe_q_ctx_get(pipe, dpdk_engine_mgmt_id_get()),
				     fwd_miss, fwd_miss_entry);
	dpdk_engine_mgmt_id_unlock();
	if (rc) {
		DOCA_DLOG_ERR("Pipe %s: create fwd miss entry fail", pipe->name);
		dpdk_pipe_entry_release(fwd_miss_entry);
		return rc;
	}

	fwd_miss_entry->pipe_queue = dpdk_engine_mgmt_id_get();
	fwd_miss_entry->pipe = pipe;
	dpdk_pipe_common_entry_attach(dpdk_engine_mgmt_id_get(), pipe, fwd_miss_entry);
	return 0;
}

static enum doca_flow_pipe_type
convert_engine_pipe_to_api(enum engine_pipe_type type)
{
	switch (type) {
	case ENGINE_PIPE_BASIC:
		return DOCA_FLOW_PIPE_BASIC;
	case ENGINE_PIPE_CONTROL:
		return DOCA_FLOW_PIPE_CONTROL;
	case ENGINE_PIPE_LPM:
		return DOCA_FLOW_PIPE_LPM;
	case ENGINE_PIPE_HASH:
		DOCA_DLOG_DBG("hash pipe type cannot be converted to API - it is not exposed");
		return DOCA_FLOW_PIPE_BASIC; /* placeholder */
	case ENGINE_PIPE_ORDERED_LIST:
		return DOCA_FLOW_PIPE_ORDERED_LIST;
	default:
		break;
	}

	DOCA_LOG_ERR("unsupported conversion from engine pipe type %u to api", type);
	return DOCA_FLOW_PIPE_BASIC;
};

void
dpdk_pipe_cfg_convert(struct doca_flow_pipe_cfg *cfg, struct dpdk_pipe_cfg *pipe_cfg,
		      enum engine_pipe_type type)
{
	cfg->attr.is_root = pipe_cfg->is_root;
	cfg->attr.type = convert_engine_pipe_to_api(type);
	cfg->attr.nb_actions = pipe_cfg->nb_actions;
	cfg->attr.nb_flows = pipe_cfg->nb_flows;

	cfg->port = (struct doca_flow_port *)pipe_cfg->port;
	cfg->match = (struct doca_flow_match *)pipe_cfg->uds_cfg.match;
	cfg->match_mask = (struct doca_flow_match *)pipe_cfg->uds_cfg.match_mask;
	cfg->actions = (struct doca_flow_actions **)pipe_cfg->uds_cfg.actions;
	cfg->action_descs = (struct doca_flow_action_descs **)pipe_cfg->uds_cfg.action_descs;
	cfg->monitor = (struct doca_flow_monitor *)pipe_cfg->uds_cfg.monitor;
}

static bool
is_age_selected(const struct doca_flow_monitor *monitor)
{
	return monitor && (monitor->flags & DOCA_FLOW_MONITOR_AGING);
}

static int
get_crypto_selection(uint8_t nb_actions,
		     struct doca_flow_actions **actions,
		     bool *res_encrypt)
{
	struct doca_flow_actions *action;
	uint8_t action_set;
	bool encrypt = false, decrypt = false;

	if (actions == NULL)
		return false;
	nb_actions = nb_actions ? nb_actions : DEFAULT_NB_ACTIONS;
	for (action_set = 0; action_set < nb_actions; action_set++) {
		action = actions[action_set];
		if (action) {
			switch (action->security.proto_type) {
			case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_ENCRYPT:
				encrypt = true;
				break;
			case DOCA_FLOW_CRYPTO_PROTOCOL_ESP_DECRYPT:
				decrypt = true;
				break;
			default:
				break;
			}
		}
	}
	if (decrypt && encrypt) {
		DOCA_LOG_ERR("encryption and decription actions require different domains");
		return -1;
	}
	*res_encrypt = !!encrypt;
	return 0;
}

struct doca_flow_pipe *
dpdk_pipe_create(enum engine_pipe_type type,
		 const struct doca_flow_pipe_cfg *cfg,
		 const struct doca_flow_fwd *fwd,
		 const struct doca_flow_fwd *fwd_miss)
{
	struct dpdk_pipe_cfg pipe_cfg = {0};
	struct dpdk_pipe_type_ops *ops;
	uint16_t nr_queues = dpdk_engine_get_queues_in_use();

	struct doca_flow_pipe *pipe;
	struct doca_flow_port *port;
	int rc;

	ops = get_ops_by_type(type);
	if (unlikely(!ops)) {
		DOCA_LOG_ERR("failed creating pipe of type %u - undefined type", type);
		return NULL;
	}

	rc = get_crypto_selection(pipe_cfg.nb_actions, cfg->actions, &pipe_cfg.is_encrypt);
	if (rc < 0)
		return NULL;

	if (pipe_cfg.is_encrypt && cfg->port->tx_encrypt_flow != NULL) {
		DOCA_LOG_ERR("the only one encryption table is supported");
		return NULL;
	}

	pipe_cfg.is_root = cfg->attr.is_root;
	pipe_cfg.is_aging_enabled = is_age_selected(cfg->monitor);
	pipe_cfg.nb_ordered_lists = cfg->attr.nb_ordered_lists;
	pipe_cfg.nb_actions = cfg->attr.nb_actions ? cfg->attr.nb_actions : DEFAULT_NB_ACTIONS;
	pipe_cfg.nb_flows = cfg->attr.nb_flows ? cfg->attr.nb_flows : DEFAULT_NB_FLOWS;
	if (cfg->attr.name)
		strlcpy(pipe_cfg.name, cfg->attr.name, ENGINE_PIPE_NAME_LEN);

	pipe_cfg.port = cfg->port;
	pipe_cfg.dpdk_port = pipe_cfg.port->dpdk_port;
	pipe_cfg.uds_cfg.match = cfg->match;
	pipe_cfg.uds_cfg.match_mask = cfg->match_mask;
	pipe_cfg.uds_cfg.actions = (const struct doca_flow_actions **)cfg->actions;
	pipe_cfg.uds_cfg.action_descs = (const struct doca_flow_action_descs **)cfg->action_descs;
	pipe_cfg.uds_cfg.monitor = cfg->monitor;
	pipe_cfg.uds_cfg.ordered_lists = (const struct doca_flow_ordered_list **)cfg->ordered_lists;
	pipe_cfg.uds_cfg.fwd = fwd;
	pipe_cfg.uds_cfg.fwd_miss = fwd_miss;

	rc = ops->pipe_legacy_verify(&pipe_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed creating pipe of type %u - verification rc=%d", type, rc);
		return NULL;
	}

	pipe = ops->pipe_legacy_alloc(nr_queues);
	if (unlikely(pipe == NULL)) {
		DOCA_LOG_ERR("failed creating pipe of type %u - alloc rc=%d", type, rc);
		return NULL;
	}

	rc = engine_bindable_init(&pipe->bindable_obj, DF_ENGINE_BINDABLE_TYPE_PIPE);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed creating pipe of type %u - bindify failed rc=%d", type, rc);
		ops->pipe_legacy_free(pipe);
		return NULL;
	}

	pipe->type = type;
	pipe->nr_entries = pipe_cfg.nb_flows;
	pipe->port = (struct doca_flow_port *)pipe_cfg.port;

	rc = ops->pipe_legacy_build(pipe, &pipe_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed creating pipe of type %u - build, rc=%d", type, rc);
		ops->pipe_legacy_free(pipe);
		return NULL;
	}

	rc = ops->pipe_legacy_submit(pipe, &pipe_cfg);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed creating pipe of type %u - submit, rc=%d", type, rc);
		ops->pipe_legacy_free(pipe);
		return NULL;
	}

	port = cfg->port;
	rc = engine_bindable_set_parent(&pipe->bindable_obj, &port->bindable_obj);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed to create pipe - setting pipe hierarchy, rc=%d", rc);
		goto free_pipe;
	}

	if (pipe_cfg.is_encrypt) {
		rc = dpdk_root_tx_encrypt_create(port, pipe->group_id);
		if (rc) {
			DOCA_LOG_ERR("failed to create pipe - creating tx encrypt flow, rc=%d", rc);
			goto free_pipe;
		}
	}

	rte_spinlock_lock(&port->pipe_lock);
	LIST_INSERT_HEAD(&port->pipe_list, pipe, next);
	rte_spinlock_unlock(&port->pipe_lock);
	pipe->is_connected = true;

	DOCA_LOG_DBG("pipe of type %u created on port %u", type, port->port_id);
	return pipe;
free_pipe:
	DOCA_DLOG_ERR("failed to create pipe %s, rc=%d", cfg->attr.name, rc);
	dpdk_pipe_destroy(cfg->port, pipe);
	return NULL;
}

void
dpdk_pipe_entries_flush(struct doca_flow_port *port, struct doca_flow_pipe *pipe)
{
	struct doca_flow_pipe_entry *entry;
	struct dpdk_pipe_queue *pipe_by_qidx;
	uint32_t nb_free_entry = 0;
	int i;

	if (dpdk_port_legacy_is_invalid(port))
		return;
	for (i = 0; i < pipe->nb_queues; i++) {
		pipe_by_qidx = dpdk_pipe_by_qidx_get(pipe, i);
		while ((entry = LIST_FIRST(&pipe_by_qidx->entry_list))) {
			dpdk_pipe_common_entry_detach(i, pipe, entry);
			_dpdk_pipe_entry_destroy(port, entry, NULL, pipe_by_qidx);
			nb_free_entry++;
		}
		dpdk_flow_entries_process(port, i, PULL_TIMEOUT_US, dpdk_engine_get_queue_depth());
	}

	DOCA_DLOG_INFO("Pipe %s - all %u entries freed", pipe->name, nb_free_entry);
}

void
dpdk_pipe_destroy(struct doca_flow_port *port,
		  struct doca_flow_pipe *pipe)
{
	struct dpdk_pipe_type_ops *ops;

	ops = get_ops_by_type(pipe->type);
	if (unlikely(!ops)) {
		DOCA_LOG_ERR("failed destroying pipe of type %u - undefined type", pipe->type);
		return;
	}

	DOCA_DLOG_INFO("portid %u destroy pipe %s", port->port_id, pipe->name);
	ops->pipe_legacy_free(pipe);
}

void
dpdk_pipe_dump(struct doca_flow_port *port, struct doca_flow_pipe *pipe, FILE *f)
{
	struct dpdk_pipe_queue *pipe_by_qidx;
	struct dpdk_pipe *dpdk_pipe;
	uint64_t unique_id;
	int i;

	switch (pipe->type) {
	case ENGINE_PIPE_CT:
		dpdk_engine_ct_ops_get()->dump(f);
		break;
	default:
		unique_id = dpdk_pipe_generate_unique_id(port, pipe);
		fprintf(f, "  id:%"PRIx64",name:%s\n", unique_id, pipe->name);
		for (i = 0; i < pipe->nb_queues; i++) {
			dpdk_pipe = &pipe->dpdk_pipe;
			pipe_by_qidx = &dpdk_pipe->queues[i];
			fprintf(f, "Queue #%d: flow entry count:%u\n", i,
				pipe_by_qidx->nb_entries);
			dpdk_pipe_rte_flow_dump("    ", dpdk_pipe,
						&pipe_by_qidx->pipe_q, f);
		}
		break;
	};
}

int
dpdk_pipe_entry_query(struct doca_flow_pipe_entry *entry,
		      struct doca_flow_query *query_stats)
{
	struct rte_flow_error flow_err;
	struct rte_flow_query_count count = {0};
	struct rte_flow_action actions[DPDK_PIPE_MAX_ACTIONS];
	struct doca_flow_pipe *pipe;
	struct doca_flow_port *port;
	uint16_t port_id;
	int ret;

	pipe = (struct doca_flow_pipe *)entry->pipe;
	if (pipe == NULL) {
		DOCA_DLOG_ERR("Pipe of entry %p is NULL", entry);
		return -1;
	}

	port = doca_flow_layer_get_pipe_port(pipe);
	if (unlikely(!port)) {
		DOCA_LOG_RATE_LIMIT_CRIT("failed querying entry - port is null");
		return -ENOTRECOVERABLE;
	}

	port_id = port->port_id;
	actions[0].type = RTE_FLOW_ACTION_TYPE_COUNT;
	actions[1].type = RTE_FLOW_ACTION_TYPE_END;
	ret = rte_flow_query(port_id, dpdk_flow_tracker_get_flow(&entry->tracker),
			     actions, &count, &flow_err);
	if (ret) {
		DOCA_DLOG_ERR("Port %u query flow fail, type %d message: %s",
			     port_id, flow_err.type,
			     flow_err.message ? flow_err.message
					      : "(no stated reason)");
		return -1;
	}
	query_stats->total_bytes = count.bytes;
	query_stats->total_pkts = count.hits;
	return 0;
}
