/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DPDK_PIPE_LEGACY_H_
#define DPDK_PIPE_LEGACY_H_

#include <stdio.h>

#include <rte_flow.h>
#include <rte_spinlock.h>
#include <rte_mtr.h>

#include "doca_flow.h"
#include "dpdk_pipe_items.h"
#include "dpdk_pipe_actions.h"
#include "dpdk_flow.h"

#include "dpdk_flow_hws_legacy.h"
#include "dpdk_table.h"

#include "engine_pipe.h"

#ifdef __cplusplus
extern "C" {
#endif

#define DPDK_MAX_PORT_PROFILES		128

struct doca_flow_port;
struct doca_flow_match;
struct doca_flow_actions;
struct doca_flow_action_descs;
struct doca_flow_monitor;
struct doca_flow_fwd;

struct dpdk_uds_cfg {
	uint32_t index;
	const struct doca_flow_match *match;
	const struct doca_flow_match *match_mask;
	const struct doca_flow_actions **actions;
	const struct doca_flow_action_descs **action_descs;
	const struct doca_flow_monitor *monitor;
	const struct doca_flow_ordered_list **ordered_lists;
	const struct doca_flow_fwd *fwd;
	const struct doca_flow_fwd *fwd_miss;
};

struct dpdk_pipe_cfg {
	struct dpdk_port *dpdk_port;

	uint8_t nb_actions;
	uint8_t nb_ordered_lists;
	uint32_t nb_flows;
	bool is_root;
	bool is_aging_enabled;
	bool is_encrypt;
	char name[ENGINE_PIPE_NAME_LEN];
	enum dpdk_table_hash_type hash_type;

	const struct doca_flow_port *port;

	struct dpdk_uds_cfg uds_cfg;
};

struct dpdk_pipe_type_ops {
	int (*module_init)(void);

	int (*pipe_legacy_verify)(struct dpdk_pipe_cfg *pipe_cfg);

	struct doca_flow_pipe *(*pipe_legacy_alloc)(uint16_t nr_queues);

	int (*pipe_legacy_build)(struct doca_flow_pipe *pipe_legacy,
				 struct dpdk_pipe_cfg *pipe_cfg);

	int (*pipe_legacy_submit)(struct doca_flow_pipe *pipe_legacy,
				  struct dpdk_pipe_cfg *pipe_cfg);

	void (*pipe_legacy_free)(struct doca_flow_pipe *pipe_legacy);

	int (*pipe_queue_verify)(struct doca_flow_pipe *pipe_legacy, uint16_t queue_id,
				 uint16_t action_idx, uint32_t priority, struct dpdk_uds_cfg *uds_cfg);

	int (*pipe_queue_modify)(struct doca_flow_pipe *pipe_legacy,
				 struct doca_flow_pipe_entry *entry, uint16_t queue_id,
				 uint16_t action_idx, uint32_t priority, struct dpdk_uds_cfg *uds_cfg);

	int (*pipe_queue_submit)(struct doca_flow_pipe *pipe_legacy,
				 struct doca_flow_pipe_entry *entry, uint16_t queue_id,
				 uint16_t action_idx, dpdk_flow_completion_cb entry_completion_cb,
				 void *usr_ctx, bool no_wait);

	int (*pipe_queue_remove)(struct doca_flow_pipe_entry *entry,
				  dpdk_flow_completion_cb entry_completion_cb,
				  void *usr_ctx);

	int (*pipe_queue_cleanup)(struct doca_flow_pipe_entry *entry);
};

struct dpdk_profiles_ids {
	struct profile_hash *id_hash;
	/* Profile ids hash */
	uint32_t ids[DPDK_MAX_PORT_PROFILES];
	/* Saved ids based on hash index */
};

struct dpdk_pipe_q_ctx {
	/* recode the meter action index created by pipe */
	int fwd_action_index;
	uint32_t actions_flags;
	/* save aging info */
	uint32_t aging;
	uint64_t user_data;

	enum doca_flow_ip_type ip_type;
	enum doca_flow_fwd_type fwd_type;
	uint8_t l4_type;

	struct rte_flow_attr attr;
	struct rte_flow_item items[DPDK_PIPE_MAX_ITEMS];
	struct dpdk_item_entry item_entry[DPDK_PIPE_MAX_ITEMS];
	struct dpdk_pipe_actions *action_ctx;
	struct dpdk_profiles_ids *profile_ids;
	struct rte_flow_action_jump default_pipe_jump;
} __rte_cache_aligned;

struct dpdk_pipe_queue {
	LIST_HEAD(, doca_flow_pipe_entry) entry_list;
	struct dpdk_pipe_q_ctx pipe_q;
	uint32_t queue_entry_id;
	uint32_t nb_entries;
};

/* only for Hybrid mode */
enum fwd_port_type {
	FWD_REPRESENTOR_PORT = 1,
	FWD_LAG_PORT,
};
#define LAG_P0_META_BIT 0
#define LAG_P1_META_BIT 1

struct dpdk_pipe {
	uint8_t nb_items;
	uint8_t *nb_actions_pipe;
	uint32_t nb_actions_array_len;
	rte_spinlock_t ctrl_entry_lock;
	uint32_t *policy_id;
	enum fwd_port_type fwd_port_type;
	uint16_t encap_peer_port_id;
	struct dpdk_hws_item *encap_hws_ctx;
	struct dpdk_hws_ctx hws_ctx;
	struct dpdk_pipe_queue queues[0];
};

int
dpdk_pipe_legacy_init_early(void);

int
dpdk_pipe_legacy_init_late(void);

struct dpdk_pipe_q_ctx *
dpdk_pipe_q_ctx_get(struct doca_flow_pipe *pipe, uint16_t q);

struct dpdk_pipe_queue *
dpdk_pipe_by_qidx_get(struct doca_flow_pipe *pipe, uint16_t q);

void
dpdk_pipe_cfg_convert(struct doca_flow_pipe_cfg *cfg, struct dpdk_pipe_cfg *pipe_cfg,
		      enum engine_pipe_type type);

void
dpdk_pipe_set_root_id(struct doca_flow_pipe *pipe);

uint64_t
dpdk_pipe_generate_unique_id(const struct doca_flow_port *port, struct doca_flow_pipe *pipe);

int
dpdk_pipe_fwd_miss_handle(const struct doca_flow_pipe_cfg *cfg, struct doca_flow_pipe *pipe,
			  const struct doca_flow_fwd *fwd_miss);

struct doca_flow_pipe_entry *
dpdk_pipe_entry_alloc(struct doca_flow_pipe *pipe, uint16_t pipe_queue);

struct doca_flow_pipe *
dpdk_pipe_create(enum engine_pipe_type type, const struct doca_flow_pipe_cfg *cfg,
		 const struct doca_flow_fwd *fwd, const struct doca_flow_fwd *fwd_miss);

struct doca_flow_pipe_entry *
dpdk_pipe_entry_add(uint16_t pipe_queue,
		    uint32_t priority,
		    struct doca_flow_pipe *pipe,
		    const struct doca_flow_match *match,
		    const struct doca_flow_match *match_mask,
		    const struct doca_flow_actions *actions,
		    const struct doca_flow_action_descs *action_descs,
		    const struct doca_flow_monitor *monitor,
		    const struct doca_flow_fwd *fwd,
		    uint32_t flags,
		    void *usr_ctx);

struct doca_flow_pipe_entry *
dpdk_pipe_entry_add_with_index(uint16_t pipe_queue,
			       struct doca_flow_pipe *pipe,
			       uint32_t index,
			       const struct doca_flow_actions *actions,
			       const struct doca_flow_action_descs *action_descs,
			       const struct doca_flow_monitor *monitor,
			       const struct doca_flow_ordered_list *ordered_list,
			       const struct doca_flow_fwd *fwd,
			       uint32_t flags,
			       void *usr_ctx);

struct doca_flow_pipe_entry *
dpdk_pipe_entry_add_with_cb(uint16_t pipe_queue,
			    uint32_t priority,
			    struct doca_flow_pipe *pipe,
			    const struct doca_flow_match *match,
			    const struct doca_flow_match *match_mask,
			    const struct doca_flow_actions *actions,
			    const struct doca_flow_action_descs *action_descs,
			    const struct doca_flow_monitor *monitor,
			    const struct doca_flow_fwd *fwd,
			    dpdk_flow_completion_cb entry_completion_cb,
			    uint32_t flags,
			    void *usr_ctx);

void
dpdk_entry_update_status(struct doca_flow_pipe_entry *entry, enum dpdk_flow_request_status status);

int
dpdk_pipe_entry_remove(uint16_t pipe_queue, void *usr_ctx, struct doca_flow_pipe_entry *entry);

void
dpdk_pipe_entries_flush(struct doca_flow_port *port, struct doca_flow_pipe *pipe);

void
dpdk_pipe_destroy(struct doca_flow_port *port, struct doca_flow_pipe *pipe);

void
dpdk_pipe_dump(struct doca_flow_port *port, struct doca_flow_pipe *pipe, FILE *f);

int
dpdk_pipe_entry_query(struct doca_flow_pipe_entry *entry,
		      struct doca_flow_query *query_stats);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DPDK_ENGINE_H_ */
