/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "dpdk_pipe_lpm.h"
#include "pipe_lpm.h"
#include "dpdk_table.h"

#include "dpdk_engine.h"
#include "dpdk_table_manager.h"
#include "dpdk_port.h"

#include "dpdk_pipe_common.h"
#include "doca_flow_priv.h"
#include "dpdk_dbg.h"
#include "doca_flow.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(dpdk_pipe_lpm);

#define PIPE_LPM_NR_TABLES 1
#define PIPE_LPM_DEFAULT_TABLE_IDX 0

static int
pipe_lpm_verify(struct dpdk_pipe_cfg *pipe_cfg)
{
	return dpdk_pipe_common_verify(pipe_cfg);
}

static struct doca_flow_pipe *
pipe_lpm_alloc(uint16_t nr_queues)
{
	return dpdk_pipe_common_alloc(nr_queues);
}

static int
pipe_lpm_build(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct dpdk_table_manager_cfg table_manager_cfg;
	struct doca_flow_pipe_cfg cfg = {0};
	struct doca_flow_match lpm_match = {0};
	struct doca_flow_actions lpm_actions = {0};
	struct doca_flow_actions *lpm_actions_array[1];
	struct doca_flow_fwd lpm_fwd = {0};
	struct dpdk_pipe_cfg lpm_pipe_cfg = {0};
	void *lpm = NULL;

	if (pipe_cfg->uds_cfg.match == NULL)
		return -EINVAL;

	dpdk_pipe_cfg_convert(&cfg, pipe_cfg, ENGINE_PIPE_LPM);

	table_manager_cfg.nr_tables = PIPE_LPM_NR_TABLES;
	pipe_legacy->table_manager = dpdk_table_manager_create(&table_manager_cfg);
	if (!pipe_legacy->table_manager) {
		DOCA_LOG_ERR("failed building lpm pipe - table manager is null");
		return -ENOMEM;
	}

	lpm = lpm_create(&cfg, pipe_cfg->uds_cfg.fwd, pipe_cfg->uds_cfg.fwd_miss);
	if (unlikely(lpm == NULL))
		return -ENOMEM;

	pipe_legacy->private_ctx = lpm;

	lpm_extract_match(lpm, &lpm_match);

	lpm_fwd.type = DOCA_FLOW_FWD_PIPE;
	lpm_fwd.next_pipe = lpm_get_main_pipe(lpm);

	lpm_pipe_cfg.is_root = false;
	lpm_pipe_cfg.nb_actions = 1;
	lpm_pipe_cfg.nb_flows = cfg.attr.nb_flows;
	strlcpy(lpm_pipe_cfg.name, pipe_cfg->name, ENGINE_PIPE_NAME_LEN);

	lpm_pipe_cfg.port = cfg.port;
	lpm_pipe_cfg.uds_cfg.match = &lpm_match;
	lpm_actions_array[0] = &lpm_actions;
	lpm_pipe_cfg.uds_cfg.actions = (const struct doca_flow_actions **)lpm_actions_array;
	lpm_pipe_cfg.uds_cfg.fwd = &lpm_fwd;
	lpm_pipe_cfg.uds_cfg.fwd_miss = pipe_cfg->uds_cfg.fwd_miss;

	return dpdk_pipe_basic_build(pipe_legacy, &lpm_pipe_cfg);
}

static int
pipe_lpm_submit(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, 0);
	struct rte_flow_action *actions_arr[1];
	struct doca_flow_pipe_cfg cfg = {0};
	struct doca_flow_match lpm_match = {0};
	struct dpdk_table_params table_params;
	uint32_t lpm_table_id;
	int rc;

	dpdk_pipe_cfg_convert(&cfg, pipe_cfg, ENGINE_PIPE_BASIC);

	if (unlikely(pipe_cfg->uds_cfg.fwd_miss)) {
		rc = dpdk_pipe_fwd_miss_handle(&cfg, pipe_legacy, pipe_cfg->uds_cfg.fwd_miss);
		if (unlikely(rc))
			return rc;
	}

	rc = dpdk_table_manager_alloc_id(pipe_legacy->table_manager, &lpm_table_id);
	if (unlikely(rc))
		return rc;

	pipe_legacy->basic_table = dpdk_table_manager_get_by_id(pipe_legacy->table_manager, lpm_table_id);
	if (unlikely(!pipe_legacy->basic_table || lpm_table_id != PIPE_LPM_DEFAULT_TABLE_IDX))
		return -ENOTRECOVERABLE;

	actions_arr[0] = dpdk_pipe_actions_array_get(pipe_q, 0);
	dpdk_pipe_common_fill_table_params(&table_params, pipe_cfg->port->port_id,
					   pipe_q->items, actions_arr, actions_arr, pipe_cfg->nb_actions,
					   &pipe_q->attr, pipe_cfg->nb_flows, DPDK_TABLE_INSERTION_TYPE_PATTERN,
					   DPDK_TABLE_HASH_TYPE_DEFAULT);

	rc = dpdk_table_create(&table_params,
			       dpdk_port_get_table_ctx(pipe_cfg->port->dpdk_port), pipe_legacy->basic_table);
	if (unlikely(rc)) {
		dpdk_table_manager_free_id(pipe_legacy->table_manager, PIPE_LPM_DEFAULT_TABLE_IDX);
		return rc;
	}

	// add one entry as basic pipe to forward all required traffic to the internal lpm pipe.
	pipe_legacy->type = ENGINE_PIPE_BASIC;
	if (dpdk_pipe_entry_add(0 /* pipe_queue */, 0 /* priority */, pipe_legacy, &lpm_match,
				NULL /* match_mask */, NULL /* actions */, NULL /* action_descs */,
				NULL /* monitor */, NULL /* fwd */, DOCA_FLOW_NO_WAIT,
				NULL /* user_ctx */) == NULL) {
		dpdk_table_manager_free_id(pipe_legacy->table_manager, PIPE_LPM_DEFAULT_TABLE_IDX);
		return -EIO;
	}
	pipe_legacy->type = ENGINE_PIPE_LPM;

	return 0;
}

static void
pipe_lpm_free(struct doca_flow_pipe *pipe_legacy)
{
	struct dpdk_table *table;

	dpdk_pipe_entries_flush(pipe_legacy->port, pipe_legacy);
	if (pipe_legacy && pipe_legacy->table_manager) {
		table = dpdk_table_manager_get_by_id(pipe_legacy->table_manager, PIPE_LPM_DEFAULT_TABLE_IDX);
		if (likely(table)) {
			dpdk_table_destroy(dpdk_port_get_table_ctx(pipe_legacy->port->dpdk_port), table);
			dpdk_table_manager_free_id(pipe_legacy->table_manager, PIPE_LPM_DEFAULT_TABLE_IDX);
		}
		dpdk_table_manager_destroy(pipe_legacy->table_manager);
	}

	if (pipe_legacy->private_ctx)
		lpm_destroy(pipe_legacy->private_ctx);
	dpdk_pipe_common_resources_free(pipe_legacy);
	dpdk_pipe_common_legacy_free(pipe_legacy);
}

static int
pipe_lpm_queue_verify(struct doca_flow_pipe *pipe_legacy, uint16_t queue_id, uint16_t action_idx,
			uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	return dpdk_pipe_common_entry_verify(pipe_legacy, queue_id, action_idx, uds_cfg);
}

static int
pipe_lpm_queue_modify(struct doca_flow_pipe *pipe_legacy, struct doca_flow_pipe_entry *entry,
		      uint16_t queue_id, uint16_t action_idx, uint32_t priority, struct dpdk_uds_cfg *uds_cfg)
{
	int rc;

	if (unlikely(pipe_legacy->private_ctx == NULL))
		return -EINVAL;

	if (unlikely(lpm_entry_lookup(pipe_legacy, uds_cfg->match, uds_cfg->match_mask) == 0))
		return 0;

	dpdk_dbg_pipe_config_dump(uds_cfg->match, uds_cfg->actions[0], NULL /*action_descs*/,
				  NULL /*monitor*/, uds_cfg->fwd);

	rc = lpm_entry_add(entry, queue_id, pipe_legacy, uds_cfg->match, uds_cfg->match_mask,
			    uds_cfg->actions[0], uds_cfg->fwd);
	if (unlikely(rc < 0)) {
		DOCA_DLOG_ERR("failed to add lpm entry, rc=%d", rc);
		return rc;
	}

	return 0;
}

static int
pipe_lpm_queue_submit(struct doca_flow_pipe *pipe_legacy,
		      struct doca_flow_pipe_entry *entry, uint16_t queue_id,
		      uint16_t action_idx, dpdk_flow_completion_cb entry_completion_cb,
		      void *usr_ctx, bool no_wait)
{
	int rc;

	if (no_wait) {
		rc = lpm_flush(queue_id, pipe_legacy, LPM_PUSH);
		if (unlikely(rc < 0)) {
			DOCA_DLOG_ERR("failed to sumbit lpm queue - lpm flush, rc=%d", rc);
			return rc;
		}
	}

	dpdk_pipe_common_entry_attach(queue_id, pipe_legacy, entry);
	entry->table_id = PIPE_LPM_DEFAULT_TABLE_IDX;
	return 0;
}

static int
pipe_lpm_queue_remove(struct doca_flow_pipe_entry *entry,
		      dpdk_flow_completion_cb cb, void *usr_ctx)
{
	return dpdk_pipe_common_queue_remove(entry, cb, usr_ctx);
}

static int
pipe_lpm_queue_cleanup(struct doca_flow_pipe_entry *entry)
{
	dpdk_pipe_entry_release(entry);
	return 0;
}

static struct dpdk_pipe_type_ops pipe_ops = {
	.pipe_legacy_verify = pipe_lpm_verify,
	.pipe_legacy_alloc = pipe_lpm_alloc,
	.pipe_legacy_build = pipe_lpm_build,
	.pipe_legacy_submit = pipe_lpm_submit,
	.pipe_legacy_free = pipe_lpm_free,

	.pipe_queue_verify = pipe_lpm_queue_verify,
	.pipe_queue_modify = pipe_lpm_queue_modify,
	.pipe_queue_submit = pipe_lpm_queue_submit,
	.pipe_queue_remove = pipe_lpm_queue_remove,
	.pipe_queue_cleanup = pipe_lpm_queue_cleanup,
};

struct dpdk_pipe_type_ops *
dpdk_pipe_lpm_get_ops(void)
{
	return &pipe_ops;
}
