/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "doca_flow_layer.h"
#include "doca_flow_priv.h"
#include "dpdk_engine.h"
#include "dpdk_flow.h"
#include "dpdk_pipe_basic.h"
#include "dpdk_pipe_common.h"
#include "dpdk_pipe_hash.h"
#include "dpdk_pipe_legacy.h"
#include "dpdk_pipe_ordered_list.h"
#include "engine_bindable.h"
#include "engine_model.h"
#include "engine_pipe.h"
#include "ordered_list.h"
#include "priv_doca_mempool.h"

#include <doca_log.h>

#include <stdatomic.h>
#include <stdlib.h>

DOCA_LOG_REGISTER(dpdk_pipe_ordered_list);

struct ordered_list_pipe {
	struct priv_doca_mempool *entry_pool;
	struct priv_doca_mempool *entry_ctx_pool;
	struct doca_flow_pipe *hash_pipes[ORDERED_LIST_MAX_TOTAL_ITEMS];
	uint32_t nb_hash_pipes;
	struct ordered_list_cfg sequences[ORDERED_LIST_MAX_NUMBER_OF_SEQUENCES];
	uint32_t nb_sequences;
	struct ordered_list_pipe_cfg cfg;
};

/* Private data attached to each successfully created entry.  */
struct ordered_list_pipe_entry {
	uint32_t nb_hash_pipe_entries;
	struct doca_flow_pipe_entry *hash_pipe_entries[0];
};

/*
 * Data attached to the entry only during insertion and removal.
 * The number of contexts is limited by "queue number" x "queue size",
 * so making it small is not a concern.
 */
struct ordered_list_pipe_entry_ctx {
	struct doca_flow_pipe_entry *entry;
	struct ordered_list_pipe_entry *list_entry;
	dpdk_flow_completion_cb completion_cb;
	void *completion_cb_ctx;
	enum dpdk_flow_request_op completion_op;
	enum dpdk_flow_request_status completion_status;
	uint16_t queue_id;
	uint8_t hash_pipe_entries_left;
	struct {
		struct doca_flow_pipe *next_pipe;
		uint32_t index;
	} frontend_entry;
	struct {
		uint8_t list_idx;
		bool need_rollback;
		bool no_wait;
	} insertion;
};

static struct dpdk_pipe_type_ops *basic_ops;
static struct dpdk_pipe_type_ops *hash_ops;

static int
ordered_list_verify(const struct doca_flow_ordered_list *list, uint32_t idx)
{
	uint32_t i;

	if (list->idx != idx) {
		DOCA_LOG_ERR("ordered_lists[%u]->idx is not equal to %u", idx, idx);
		return -EINVAL;
	}

	if (list->size > ORDERED_LIST_MAX_ELEMENTS_IN_SEQUENCE) {
		DOCA_LOG_ERR("ordered_lists[%u]->size out of range [0; %u]",
			     idx, ORDERED_LIST_MAX_ELEMENTS_IN_SEQUENCE);
		return -EINVAL;
	}

	for (i = 0; i < list->size; i++) {
		if (list->elements[i] == NULL) {
			DOCA_LOG_ERR("ordered_list[%u]->elements[%u] is NULL", idx, i);
			return -EINVAL;
		}

		switch (list->types[i]) {
		case DOCA_FLOW_ORDERED_LIST_ELEMENT_ACTIONS: {
			const struct doca_flow_actions *actions = list->elements[i];

			if (actions->has_encap && i < list->size - 1) {
				DOCA_LOG_ERR("ENCAP must be the last action in the list");
				return -EINVAL;
			}
			break;
		}
		case DOCA_FLOW_ORDERED_LIST_ELEMENT_ACTION_DESCS:
		case DOCA_FLOW_ORDERED_LIST_ELEMENT_MONITOR:
			break;
		default:
			DOCA_LOG_ERR("ordered_lists[%u]->types[%u] == %d is unsupported",
					idx, i, list->types[i]);
			return -EINVAL;
		}
	}
	return 0;
}

static int
ordered_list_pipe_verify(struct dpdk_pipe_cfg *pipe_cfg)
{
	const struct dpdk_uds_cfg *uds_cfg = &pipe_cfg->uds_cfg;
	uint32_t i;
	int rc;

	if (!dpdk_engine_mode_is_hws()) {
		DOCA_LOG_ERR("ordered list pipes only support HWS mode");
		return -ENOTSUP;
	}

	if (pipe_cfg->is_root) {
		DOCA_LOG_ERR("ordered list pipe cannot be a root pipe");
		return -EINVAL;
	}

	if (uds_cfg->ordered_lists == NULL) {
		DOCA_LOG_ERR("ordered_lists is NULL");
		return -EINVAL;
	}
	if (pipe_cfg->nb_ordered_lists < 1 ||
	    pipe_cfg->nb_ordered_lists > ORDERED_LIST_MAX_NUMBER_OF_SEQUENCES) {
		DOCA_LOG_ERR("ordered_lists length out of range [1; %d]",
			     ORDERED_LIST_MAX_NUMBER_OF_SEQUENCES);
		return -EINVAL;
	}
	for (i = 0; i < pipe_cfg->nb_ordered_lists; i++) {
		rc = ordered_list_verify(uds_cfg->ordered_lists[i], i);
		if (rc < 0)
			return rc;
	}

	if (uds_cfg->match || uds_cfg->match_mask || uds_cfg->actions || uds_cfg->action_descs || uds_cfg->monitor)
		DOCA_LOG_WARN("ordered list pipe ignores match, match_mask, actions, actions_desc, and monitor");
	return 0;
}

static struct doca_flow_pipe *
ordered_list_pipe_alloc(uint16_t nr_queues)
{
	struct doca_flow_pipe *pipe_legacy = NULL;
	struct ordered_list_pipe *list_pipe = NULL;

	list_pipe = calloc(1, sizeof(*list_pipe));
	if (list_pipe == NULL)
		return NULL;

	pipe_legacy = dpdk_pipe_common_alloc(nr_queues);
	if (pipe_legacy == NULL) {
		free(list_pipe);
		return NULL;
	}

	pipe_legacy->private_ctx = list_pipe;
	return pipe_legacy;
}

static void
ordered_list_pipe_resources_free(struct ordered_list_pipe *list_pipe)
{
	uint32_t hash_pipe_idx;

	priv_doca_mempool_destroy(list_pipe->entry_ctx_pool);
	priv_doca_mempool_destroy(list_pipe->entry_pool);

	for (hash_pipe_idx = 0; hash_pipe_idx < list_pipe->nb_hash_pipes; hash_pipe_idx++) {
		struct doca_flow_pipe *hash_pipe;
		uint32_t idx = list_pipe->nb_hash_pipes - 1 - hash_pipe_idx;

		hash_pipe = list_pipe->hash_pipes[idx];
		if (hash_pipe == NULL)
			continue;

		hash_ops->pipe_legacy_free(hash_pipe);
		list_pipe->hash_pipes[idx] = NULL;
	}
	list_pipe->nb_hash_pipes = 0;
}

static void
ordered_list_pipe_free(struct doca_flow_pipe *pipe_legacy)
{
	struct ordered_list_pipe *list_pipe = pipe_legacy->private_ctx;

	dpdk_pipe_entries_flush(doca_flow_layer_get_pipe_port(pipe_legacy), pipe_legacy);

	ordered_list_pipe_resources_free(list_pipe);
	free(list_pipe);

	basic_ops->pipe_legacy_free(pipe_legacy);
}

static int
ordered_list_pipe_queue_verify(struct doca_flow_pipe *pipe_legacy,
		uint16_t queue_id,
		uint16_t action_idx,
		uint32_t priority,
		struct dpdk_uds_cfg *uds_cfg)
{
	struct ordered_list_pipe *list_pipe = pipe_legacy->private_ctx;
	const struct doca_flow_ordered_list *ordered_list;
	struct ordered_list_cfg *list_cfg;
	uint32_t i;

	if (priority != 0) {
		DOCA_LOG_RATE_LIMIT_ERR("Ordered list pipe does not support entry priorities");
		return -EINVAL;
	}

	if (uds_cfg->ordered_lists == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("ordered_lists is NULL");
		return -EINVAL;
	}

	ordered_list = uds_cfg->ordered_lists[0];
	if (ordered_list == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("ordered_list is NULL");
		return -EINVAL;
	}

	if (ordered_list->idx >= list_pipe->nb_sequences) {
		DOCA_LOG_RATE_LIMIT_ERR("idx out of range [1; %u]", list_pipe->nb_sequences - 1);
		return -EINVAL;
	}
	list_cfg = &list_pipe->sequences[ordered_list->idx];

	if (ordered_list->size < list_cfg->nb_items) {
		DOCA_LOG_RATE_LIMIT_ERR("submitted list size (%u) exceeds the configured value (%u)",
				ordered_list->size, list_cfg->nb_items);
		return -EINVAL;
	}

	for (i = 0; i < list_cfg->nb_items; i++) {
		int elt_idx;

		elt_idx = list_cfg->items[i].actions_elt_idx;
		if (elt_idx >= 0 && ordered_list->elements[elt_idx] == NULL) {
			DOCA_LOG_RATE_LIMIT_ERR("elements[%d] is NULL while doca_flow_actions is expected", elt_idx);
			return -EINVAL;
		}

		elt_idx = list_cfg->items[i].monitor_elt_idx;
		if (elt_idx >= 0 && ordered_list->elements[elt_idx] == NULL) {
			DOCA_LOG_RATE_LIMIT_ERR("elements[%d] is NULL while doca_flow_monitor is expected", elt_idx);
			return -EINVAL;
		}
	}

	return 0;
}

static int
hash_pipe_queue_modify(struct doca_flow_pipe *hash_pipe,
		struct doca_flow_pipe_entry *hash_pipe_entry,
		uint16_t queue_id,
		uint32_t entry_idx,
		const struct doca_flow_actions *actions,
		const struct doca_flow_monitor *monitor,
		const struct doca_flow_fwd *fwd)
{
	struct dpdk_uds_cfg uds_cfg = {
		.actions = &actions,
		.monitor = monitor,
		.fwd = fwd,
		.index = entry_idx,
	};

	return hash_ops->pipe_queue_modify(hash_pipe, hash_pipe_entry, queue_id,
					   0 /* action_idx */, 0 /* priority */, &uds_cfg);
}

static int
hash_pipe_queue_submit(struct ordered_list_pipe_entry_ctx *list_entry_ctx,
		uint32_t item_idx,
		dpdk_flow_completion_cb completion_cb,
		void *completion_cb_ctx)
{
	struct doca_flow_pipe_entry *hash_pipe_entry = list_entry_ctx->list_entry->hash_pipe_entries[item_idx];
	struct doca_flow_pipe *hash_pipe = hash_pipe_entry->pipe;

	return hash_ops->pipe_queue_submit(hash_pipe, hash_pipe_entry, list_entry_ctx->queue_id,
					   0 /* action_idx */, completion_cb, completion_cb_ctx,
					   list_entry_ctx->insertion.no_wait);
}

static int
ordered_list_pipe_build_lists(struct ordered_list_pipe *list_pipe,
		struct ordered_list_pipe_cfg *pipe_cfg,
		const struct doca_flow_ordered_list **ordered_lists,
		uint8_t nb_ordered_lists)
{
	int rc;

	rc = ordered_list_pipe_cfg_build(pipe_cfg, ordered_lists, nb_ordered_lists, list_pipe->sequences);
	if (rc < 0)
		return rc;

	list_pipe->nb_sequences = nb_ordered_lists;
	return 0;
}

static int
ordered_list_pipe_build_pools(struct ordered_list_pipe *list_pipe,
		uint32_t nb_flows,
		uint32_t unique_id)
{
	struct priv_doca_mempool_cfg entry_pool_cfg = {};
	struct priv_doca_mempool_cfg entry_ctx_pool_cfg = {};
	uint16_t nb_queues = dpdk_engine_get_queues_in_use();
	char name[PRIV_MEMPOOL_NAME_LEN];
	int rc;

	snprintf(name, sizeof(name), "pipe_%u_entry", unique_id);
	entry_pool_cfg.name = name;
	entry_pool_cfg.size = sizeof(struct ordered_list_pipe_entry) +
			list_pipe->nb_hash_pipes * sizeof(struct doca_flow_pipe_entry *);
	entry_pool_cfg.nb_entries = nb_flows;
	entry_pool_cfg.nb_caches = nb_queues;
	entry_pool_cfg.socket_id = rte_socket_id();
	list_pipe->entry_pool = priv_doca_mempool_create(&entry_pool_cfg);
	if (list_pipe->entry_pool == NULL) {
		DOCA_LOG_ERR("Failed to create entry pool");
		rc = -ENOMEM;
		goto error;
	}

	snprintf(name, sizeof(name), "pipe_%u_entry_ctx", unique_id);
	entry_ctx_pool_cfg.name = name;
	entry_ctx_pool_cfg.size = sizeof(struct ordered_list_pipe_entry_ctx);
	entry_ctx_pool_cfg.nb_entries = dpdk_engine_get_queue_depth() * nb_queues;
	entry_ctx_pool_cfg.nb_caches = nb_queues;
	entry_ctx_pool_cfg.socket_id = rte_socket_id();
	list_pipe->entry_ctx_pool = priv_doca_mempool_create(&entry_ctx_pool_cfg);
	if (list_pipe->entry_ctx_pool == NULL) {
		DOCA_LOG_ERR("Failed to create entry context pool");
		rc = -ENOMEM;
		goto error;
	}

	return 0;

error:
	priv_doca_mempool_destroy(list_pipe->entry_ctx_pool);
	priv_doca_mempool_destroy(list_pipe->entry_pool);
	return rc;
}

static int
ordered_list_pipe_build_hash_pipes(struct ordered_list_pipe *list_pipe,
		const struct dpdk_pipe_cfg *pipe_cfg,
		uint32_t unique_id)
{
	struct ordered_list_pipe_cfg *cfg = &list_pipe->cfg;
	int i, list_idx, item_idx, nb_hash_pipes = cfg->nb_hash_pipes;
	uint32_t hash_pipe_idx;
	int rc;

	for (hash_pipe_idx = 0; hash_pipe_idx < nb_hash_pipes; hash_pipe_idx++) {
		struct doca_flow_pipe *hash_pipe;

		hash_pipe = hash_ops->pipe_legacy_alloc(dpdk_engine_get_queues_in_use());
		if (hash_pipe == NULL) {
			DOCA_LOG_ERR("Failed to allocate internal pipe %d", hash_pipe_idx);
			rc = -ENOMEM;
			goto error;
		}

		hash_pipe->type = ENGINE_PIPE_HASH;
		hash_pipe->nr_entries = pipe_cfg->nb_flows;
		hash_pipe->port = (struct doca_flow_port *)pipe_cfg->port;

		rc = engine_bindable_init(&hash_pipe->bindable_obj, DF_ENGINE_BINDABLE_TYPE_PIPE);
		if (unlikely(rc)) {
			DOCA_LOG_ERR("failed creating pipe of type %u - bindify failed rc=%d", ENGINE_PIPE_HASH, rc);
			hash_ops->pipe_legacy_free(hash_pipe);
			goto error;
		}

		list_pipe->hash_pipes[hash_pipe_idx] = hash_pipe;
	}

	for (list_idx = 0; list_idx < list_pipe->nb_sequences; list_idx++) {
		struct ordered_list_cfg *list = &list_pipe->sequences[list_idx];

		for (item_idx = list->nb_items - 1; item_idx >= 0; item_idx--) {
			hash_pipe_idx = list->items[item_idx].hash_pipe_idx;
			struct dpdk_pipe_cfg *hash_pipe_cfg = &cfg->hash_pipe_cfg_arr[hash_pipe_idx];
			struct doca_flow_pipe *hash_pipe = list_pipe->hash_pipes[hash_pipe_idx];
			struct doca_flow_fwd fwd;
			int rc;

			snprintf(hash_pipe_cfg->name, RTE_DIM(hash_pipe_cfg->name),
				 "pipe_%u_subpipe_%u", unique_id, hash_pipe_idx);
			hash_pipe_cfg->hash_type = DPDK_TABLE_HASH_TYPE_IDENTITY;
			hash_pipe_cfg->is_root = false;
			hash_pipe_cfg->dpdk_port = pipe_cfg->dpdk_port;
			hash_pipe_cfg->port = pipe_cfg->port;
			hash_pipe_cfg->nb_flows = pipe_cfg->nb_flows;
			if (item_idx < list->nb_items - 1) {
				uint32_t next_pipe_idx = list->items[item_idx + 1].hash_pipe_idx;

				fwd = (struct doca_flow_fwd){
					.type = DOCA_FLOW_FWD_PIPE,
					.next_pipe = list_pipe->hash_pipes[next_pipe_idx],
				};
				hash_pipe_cfg->uds_cfg.fwd = &fwd;
			} else {
				hash_pipe_cfg->uds_cfg.fwd = pipe_cfg->uds_cfg.fwd;
			}
			hash_pipe_cfg->uds_cfg.fwd_miss = pipe_cfg->uds_cfg.fwd_miss;

			rc = hash_ops->pipe_legacy_build(hash_pipe, hash_pipe_cfg);
			if (rc < 0) {
				DOCA_LOG_ERR("Failed to build internal pipe %d of list %u", item_idx, list_idx);
				goto error;
			}
		}
	}

	list_pipe->nb_hash_pipes = nb_hash_pipes;
	return 0;

error:
	for (i = 0; i < nb_hash_pipes; i++) {
		struct doca_flow_pipe *hash_pipe = list_pipe->hash_pipes[i];

		if (hash_pipe != NULL)
			dpdk_pipe_common_legacy_free(hash_pipe);
	}
	return rc;
}

static int
ordered_list_pipe_build_frontend(struct doca_flow_pipe *pipe_legacy,
				 struct dpdk_pipe_cfg *original_pipe_cfg)
{
	struct ordered_list_pipe *list_pipe = pipe_legacy->private_ctx;
	struct ordered_list_pipe_cfg *cfg = &list_pipe->cfg;
	struct dpdk_pipe_cfg *pipe_cfg = &cfg->frontend.pipe_cfg;
	struct doca_flow_match *match = &cfg->frontend.match;
	struct doca_flow_match *match_mask = &cfg->frontend.match_mask;
	struct doca_flow_fwd *fwd = &cfg->frontend.fwd;

	dpdk_pipe_hash_build_index_match(match, match_mask);
	fwd->type = DOCA_FLOW_FWD_PIPE;
	fwd->next_pipe = NULL; /* Changeable. */

	*pipe_cfg = *original_pipe_cfg;
	pipe_cfg->nb_ordered_lists = 0;
	pipe_cfg->uds_cfg.match = match;
	pipe_cfg->uds_cfg.match_mask = match_mask;
	pipe_cfg->uds_cfg.fwd = fwd;
	return basic_ops->pipe_legacy_build(pipe_legacy, pipe_cfg);
}

static int
ordered_list_pipe_build(struct doca_flow_pipe *pipe_legacy, struct dpdk_pipe_cfg *pipe_cfg)
{
	static uint32_t unique_id_generator;
	struct ordered_list_pipe *list_pipe = pipe_legacy->private_ctx;
	uint32_t nb_flows = pipe_cfg->nb_flows;
	uint32_t unique_id = atomic_fetch_add(&unique_id_generator, 1);
	int rc;

	rc = ordered_list_pipe_build_lists(list_pipe, &list_pipe->cfg,
			pipe_cfg->uds_cfg.ordered_lists, pipe_cfg->nb_ordered_lists);
	if (rc < 0)
		goto error;

	rc = ordered_list_pipe_build_hash_pipes(list_pipe, pipe_cfg, unique_id);
	if (rc < 0)
		goto error;

	rc = ordered_list_pipe_build_pools(list_pipe, nb_flows, unique_id);
	if (rc < 0)
		goto error;

	rc = ordered_list_pipe_build_frontend(pipe_legacy, pipe_cfg);
	if (rc < 0)
		goto error;

	return 0;

error:
	ordered_list_pipe_resources_free(list_pipe);
	return rc;
}

static int
ordered_list_pipe_submit(struct doca_flow_pipe *pipe_legacy,
		struct dpdk_pipe_cfg *original_pipe_cfg)
{
	struct ordered_list_pipe *list_pipe = pipe_legacy->private_ctx;
	struct ordered_list_pipe_cfg *cfg = &list_pipe->cfg;
	uint32_t i, j;
	int rc;

	for (i = 0; i < list_pipe->nb_sequences; i++) {
		const struct ordered_list_cfg *list = &list_pipe->sequences[i];

		for (j = 0; j < list->nb_items; j++) {
			uint32_t hash_pipe_idx = list->items[j].hash_pipe_idx;
			struct dpdk_pipe_cfg *hash_pipe_cfg = &cfg->hash_pipe_cfg_arr[hash_pipe_idx];
			struct doca_flow_pipe *hash_pipe = list_pipe->hash_pipes[hash_pipe_idx];

			rc = hash_ops->pipe_legacy_submit(hash_pipe, hash_pipe_cfg);
			if (rc < 0) {
				DOCA_LOG_ERR("Failed to submit internal pipe %u of list %u", j, i);
				goto error;
			}
		}
	}

	rc = basic_ops->pipe_legacy_submit(pipe_legacy, &cfg->frontend.pipe_cfg);
	if (rc < 0) {
		DOCA_LOG_ERR("Failed to submit frontend pipe (rc=%d)", rc);
		goto error;
	}
	return 0;

error:
	for (i = 0; i < list_pipe->nb_sequences; i++) {
		const struct ordered_list_cfg *list = &list_pipe->sequences[i];

		for (j = 0; j < list->nb_items; j++) {
			uint32_t hash_pipe_idx = list->items[j].hash_pipe_idx;
			struct doca_flow_pipe *hash_pipe = list_pipe->hash_pipes[hash_pipe_idx];

			if (hash_pipe != NULL) {
				hash_ops->pipe_legacy_free(hash_pipe);
				list_pipe->hash_pipes[hash_pipe_idx] = NULL;
			}
		}
	}
	return rc;
}

static void
ordered_list_pipe_entry_free(struct ordered_list_pipe *list_pipe,
		struct ordered_list_pipe_entry *list_entry,
		uint16_t queue_id)
{
	uint32_t i;

	if (unlikely(list_entry == NULL))
		return;
	for (i = 0; i < list_entry->nb_hash_pipe_entries; i++)
		dpdk_pipe_entry_release(list_entry->hash_pipe_entries[i]);
	priv_doca_mempool_free(list_pipe->entry_pool, list_entry, queue_id);
}

static struct ordered_list_pipe_entry *
ordered_list_pipe_entry_alloc(struct ordered_list_pipe *list_pipe,
		uint16_t queue_id, const struct ordered_list_cfg *list)
{
	struct ordered_list_pipe_entry *list_entry;
	uint32_t nb_hash_pipes = list->nb_items;
	uint32_t item_idx;

	list_entry = priv_doca_mempool_alloc(list_pipe->entry_pool, queue_id);
	if (list_entry == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed to allocate pipe entry");
		return NULL;
	}

	list_entry->nb_hash_pipe_entries = nb_hash_pipes;
	memset(list_entry->hash_pipe_entries, 0,
			sizeof(struct doca_flow_pipe_entry *) * nb_hash_pipes);
	for (item_idx = 0; item_idx < nb_hash_pipes; item_idx++) {
		uint32_t hash_pipe_idx = list->items[item_idx].hash_pipe_idx;
		struct doca_flow_pipe *hash_pipe = list_pipe->hash_pipes[hash_pipe_idx];
		struct doca_flow_pipe_entry *hash_pipe_entry;

		hash_pipe_entry = dpdk_pipe_entry_alloc(hash_pipe, queue_id);
		if (hash_pipe_entry == NULL) {
			DOCA_LOG_RATE_LIMIT_ERR("Failed to allocate entry of internal pipe %u", hash_pipe_idx);
			ordered_list_pipe_entry_free(list_pipe, list_entry, queue_id);
			return NULL;
		}
		list_entry->hash_pipe_entries[item_idx] = hash_pipe_entry;
	}

	return list_entry;
}

static int
ordered_list_pipe_queue_modify_backend(struct ordered_list_pipe *list_pipe,
		struct ordered_list_pipe_entry *list_entry,
		uint16_t queue_id,
		const struct ordered_list_cfg *list_cfg,
		const struct dpdk_uds_cfg *uds_cfg,
		uint32_t item_idx)
{
	static const struct doca_flow_actions empty_actions = {0};
	static const struct doca_flow_monitor empty_monitor = {0};
	const struct doca_flow_ordered_list *ordered_list = uds_cfg->ordered_lists[0];
	const struct ordered_list_item *list_item = &list_cfg->items[item_idx];
	const bool is_last = item_idx + 1 == list_cfg->nb_items;
	struct doca_flow_pipe *hash_pipe;
	struct doca_flow_pipe_entry *hash_pipe_entry;
	const struct doca_flow_actions *actions;
	const struct doca_flow_monitor *monitor;
	struct doca_flow_fwd fwd_data;
	const struct doca_flow_fwd *fwd;

	if (!is_last) {
		uint32_t next_idx = list_cfg->items[item_idx + 1].hash_pipe_idx;
		struct doca_flow_pipe *next_pipe = list_pipe->hash_pipes[next_idx];

		fwd_data = (struct doca_flow_fwd){
			.type = DOCA_FLOW_FWD_PIPE,
			.next_pipe = next_pipe,
		};
	}

	hash_pipe = list_pipe->hash_pipes[list_item->hash_pipe_idx];
	hash_pipe_entry = list_entry->hash_pipe_entries[item_idx];
	actions = list_item->actions_elt_idx >= 0 ?
			ordered_list->elements[list_item->actions_elt_idx] :
			&empty_actions;
	monitor = list_item->monitor_elt_idx >= 0 ?
			ordered_list->elements[list_item->monitor_elt_idx] :
			&empty_monitor;
	fwd = is_last ? uds_cfg->fwd : &fwd_data;
	return hash_pipe_queue_modify(hash_pipe, hash_pipe_entry, queue_id, uds_cfg->index,
				      actions, monitor, fwd);
}

static int
ordered_list_pipe_queue_modify(struct doca_flow_pipe *pipe_legacy,
		struct doca_flow_pipe_entry *entry,
		uint16_t queue_id,
		uint16_t action_idx,
		uint32_t priority,
		struct dpdk_uds_cfg *uds_cfg)
{
	struct ordered_list_pipe *list_pipe = pipe_legacy->private_ctx;
	const struct doca_flow_ordered_list *ordered_list = uds_cfg->ordered_lists[0];
	struct ordered_list_cfg *list_cfg = &list_pipe->sequences[ordered_list->idx];
	struct ordered_list_pipe_entry *list_entry = NULL;
	struct ordered_list_pipe_entry_ctx *list_entry_ctx = NULL;
	uint32_t item_idx;
	int rc;

	list_entry = ordered_list_pipe_entry_alloc(list_pipe, queue_id, list_cfg);
	if (list_entry == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed to allocate entry");
		rc = -ENOMEM;
		goto error;
	}

	list_entry_ctx = priv_doca_mempool_alloc(list_pipe->entry_ctx_pool, queue_id);
	if (list_entry_ctx == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed to allocate entry context");
		rc = -ENOMEM;
		goto error;
	}

	for (item_idx = 0; item_idx < list_cfg->nb_items; item_idx++) {
		rc = ordered_list_pipe_queue_modify_backend(list_pipe, list_entry, queue_id,
				list_cfg, uds_cfg, item_idx);
		if (unlikely(rc < 0)) {
			DOCA_LOG_RATE_LIMIT_ERR("Failed to modify queue %hu of backend pipe %u (rc=%d)",
					queue_id, item_idx, rc);
			goto error;
		}
	}

	list_entry_ctx->frontend_entry.next_pipe = list_entry->hash_pipe_entries[0]->pipe;
	list_entry_ctx->frontend_entry.index = uds_cfg->index;

	list_entry_ctx->insertion.list_idx = ordered_list->idx;
	list_entry_ctx->queue_id = queue_id;
	list_entry_ctx->list_entry = list_entry;
	list_entry_ctx->entry = entry;
	entry->obj_ctx = list_entry_ctx;
	return 0;

error:
	if (list_entry_ctx != NULL)
		priv_doca_mempool_free(list_pipe->entry_ctx_pool, list_entry_ctx, queue_id);
	if (list_entry != NULL)
		ordered_list_pipe_entry_free(list_pipe, list_entry, queue_id);
	return rc;
}

static struct ordered_list_pipe *
get_pipe_from_ctx(struct ordered_list_pipe_entry_ctx *list_entry_ctx)
{
	struct doca_flow_pipe *legacy_pipe = list_entry_ctx->entry->pipe;

	return legacy_pipe->private_ctx;
}

static void
ordered_list_pipe_rm_entry_finish(struct ordered_list_pipe_entry_ctx *list_entry_ctx)
{
	struct ordered_list_pipe *list_pipe = get_pipe_from_ctx(list_entry_ctx);
	struct doca_flow_pipe_entry *entry = list_entry_ctx->entry;
	dpdk_flow_completion_cb cb = list_entry_ctx->completion_cb;
	enum dpdk_flow_request_op op = list_entry_ctx->completion_op;
	enum dpdk_flow_request_status status = list_entry_ctx->completion_status;

	entry->obj_ctx = NULL;
	entry->user_ctx = list_entry_ctx->completion_cb_ctx;
	priv_doca_mempool_free(list_pipe->entry_ctx_pool, list_entry_ctx, list_entry_ctx->queue_id);
	cb(op, status, entry);
	dpdk_pipe_entry_release(entry);
}

static void
backend_rm_completion_cb(enum dpdk_flow_request_op op,
		enum dpdk_flow_request_status status, void *ctx)
{
	struct doca_flow_pipe_entry *entry = ctx;
	struct ordered_list_pipe_entry_ctx *list_entry_ctx = entry->user_ctx;

	list_entry_ctx->hash_pipe_entries_left--;
	if (list_entry_ctx->hash_pipe_entries_left == 0)
		ordered_list_pipe_rm_entry_finish(list_entry_ctx);
}

static void
ordered_list_pipe_rm_backend(struct ordered_list_pipe_entry_ctx *list_entry_ctx)
{
	struct ordered_list_pipe_entry *list_entry = list_entry_ctx->list_entry;
	struct ordered_list_pipe *list_pipe = get_pipe_from_ctx(list_entry_ctx);
	uint16_t queue_id = list_entry_ctx->queue_id;
	uint8_t i;

	list_entry_ctx->hash_pipe_entries_left = 0;
	for (i = 0; i < list_entry->nb_hash_pipe_entries; i++) {
		struct doca_flow_pipe_entry *hash_pipe_entry = list_entry->hash_pipe_entries[i];
		struct doca_flow_pipe *hash_pipe = hash_pipe_entry->pipe;
		int rc;

		if (hash_pipe_entry->status == DOCA_FLOW_ENTRY_STATUS_SUCCESS) {
			dpdk_pipe_common_entry_detach(queue_id, hash_pipe, hash_pipe_entry);
			rc = hash_ops->pipe_queue_remove(hash_pipe_entry, backend_rm_completion_cb, list_entry_ctx);
			if (unlikely(rc < 0)) {
				DOCA_LOG_RATE_LIMIT_WARN("Failed to remove internal entry %u", i);
				continue;
			}
			list_entry_ctx->hash_pipe_entries_left++;
			break;
		}
	}

	priv_doca_mempool_free(list_pipe->entry_pool, list_entry, list_entry_ctx->queue_id);
	list_entry_ctx->list_entry = NULL; /* avoid dangling pointer bugs */

	/*
	 * Unable to submit any removals, so need to cleanup and deliver the failure immediately.
	 * Normally it would be done at the last removal completion.
	 */
	if (unlikely(list_entry_ctx->hash_pipe_entries_left == 0))
		ordered_list_pipe_rm_entry_finish(list_entry_ctx);
}

static void
frontend_rm_completion_cb(enum dpdk_flow_request_op op,
		enum dpdk_flow_request_status status, void *ctx)
{
	struct doca_flow_pipe_entry *entry = ctx;

	ordered_list_pipe_rm_backend(entry->user_ctx);
}

static void
ordered_list_pipe_rollback(struct ordered_list_pipe_entry_ctx *list_entry_ctx)
{
	list_entry_ctx->completion_status = DPDK_FLOW_REQ_STATUS_FAIL;
	ordered_list_pipe_rm_backend(list_entry_ctx);
}

static void
frontend_add_completion_cb(enum dpdk_flow_request_op op,
		enum dpdk_flow_request_status status, void *ctx)
{
	struct doca_flow_pipe_entry *entry = ctx;
	struct ordered_list_pipe_entry_ctx *list_entry_ctx = entry->user_ctx;
	struct ordered_list_pipe *list_pipe = get_pipe_from_ctx(list_entry_ctx);
	dpdk_flow_completion_cb completion_cb = list_entry_ctx->completion_cb;

	if (unlikely(status == DPDK_FLOW_REQ_STATUS_FAIL)) {
		ordered_list_pipe_rollback(list_entry_ctx);
		return;
	}

	entry->obj_ctx = list_entry_ctx->list_entry;
	entry->user_ctx = list_entry_ctx->completion_cb_ctx;
	priv_doca_mempool_free(list_pipe->entry_ctx_pool, list_entry_ctx,
			list_entry_ctx->queue_id);
	completion_cb(op, status, entry);
}

static int
ordered_list_pipe_queue_modify_frontend(struct ordered_list_pipe_entry_ctx *list_entry_ctx)
{
	struct doca_flow_pipe_entry *entry = list_entry_ctx->entry;
	struct doca_flow_pipe *pipe_legacy = entry->pipe;
	uint16_t queue_id = list_entry_ctx->queue_id;
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	struct doca_flow_match match = {};
	struct doca_flow_fwd fwd = {
		.type = DOCA_FLOW_FWD_PIPE,
		.next_pipe = list_entry_ctx->frontend_entry.next_pipe,
	};
	int rc;

	dpdk_pipe_hash_modify_index_match(&match, list_entry_ctx->frontend_entry.index);
	rc = dpdk_pipe_common_queue_modify(ENGINE_PIPE_BASIC, queue_id, &pipe_legacy->dpdk_pipe,
					   pipe_q, entry, &match, NULL /* actions */,
					   NULL /* monitor */, &fwd, DOCA_FLOW_NO_WAIT,
					   NULL /* usr_ctx */);
	if (unlikely(rc < 0)) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed to modify queue %hu of frontend pipe (rc=%d)",
					queue_id, rc);
		return rc;
	}

	return 0;
}

static int
ordered_list_pipe_submit_frontend(struct ordered_list_pipe_entry_ctx *list_entry_ctx)
{
	struct doca_flow_pipe_entry *entry = list_entry_ctx->entry;
	struct doca_flow_pipe *pipe_legacy = entry->pipe;
	uint16_t queue_id = list_entry_ctx->queue_id;
	uint8_t action_idx = 0;
	struct dpdk_pipe_q_ctx *pipe_q = dpdk_pipe_q_ctx_get(pipe_legacy, queue_id);
	struct dpdk_flow_params flow_params = {0};

	dpdk_pipe_common_fill_flow_params(&flow_params, pipe_q->items,
					  dpdk_pipe_actions_array_get(pipe_q, action_idx),
					  pipe_legacy->basic_table, &pipe_q->attr, action_idx);
	return dpdk_pipe_common_queue_submit(pipe_legacy, entry, queue_id, &flow_params,
					     frontend_add_completion_cb, list_entry_ctx,
					     list_entry_ctx->insertion.no_wait);
}

static void
backend_add_completion_cb(enum dpdk_flow_request_op op,
		enum dpdk_flow_request_status status, void *ctx)
{
	struct doca_flow_pipe_entry *entry = ctx;
	struct ordered_list_pipe_entry_ctx *list_entry_ctx = entry->user_ctx;
	int rc;

	list_entry_ctx->insertion.need_rollback |= status == DPDK_FLOW_REQ_STATUS_FAIL;

	list_entry_ctx->hash_pipe_entries_left--;
	if (list_entry_ctx->hash_pipe_entries_left > 0)
		return;

	if (list_entry_ctx->insertion.need_rollback) {
		ordered_list_pipe_rollback(list_entry_ctx);
		return;
	}

	rc = ordered_list_pipe_queue_modify_frontend(list_entry_ctx);
	if (unlikely(rc < 0)) {
		ordered_list_pipe_rm_backend(list_entry_ctx);
		return;
	}

	rc = ordered_list_pipe_submit_frontend(list_entry_ctx);
	if (rc < 0) {
		ordered_list_pipe_rm_backend(list_entry_ctx);
		return;
	}
}

static int
ordered_list_pipe_queue_submit(struct doca_flow_pipe *pipe_legacy,
		struct doca_flow_pipe_entry *entry,
		uint16_t queue_id,
		uint16_t action_idx,
		dpdk_flow_completion_cb completion_cb,
		void *completion_cb_ctx,
		bool no_wait)
{
	struct ordered_list_pipe_entry_ctx *list_entry_ctx = entry->obj_ctx;
	struct ordered_list_pipe_entry *list_entry = list_entry_ctx->list_entry;
	uint32_t item_idx;
	int rc;

	dpdk_pipe_common_entry_attach(queue_id, pipe_legacy, entry);

	list_entry_ctx->completion_cb = completion_cb;
	list_entry_ctx->completion_cb_ctx = completion_cb_ctx;
	list_entry_ctx->completion_op = DPDK_FLOW_REQ_OP_CREATE;
	list_entry_ctx->completion_status = DPDK_FLOW_REQ_STATUS_OK;
	list_entry_ctx->insertion.no_wait = no_wait;

	for (item_idx = 0; item_idx < list_entry->nb_hash_pipe_entries; item_idx++) {
		rc = hash_pipe_queue_submit(list_entry_ctx, item_idx,
				backend_add_completion_cb, list_entry_ctx);
		if (unlikely(rc < 0)) {
			list_entry_ctx->insertion.need_rollback = true;
			return rc;
		}
		list_entry_ctx->hash_pipe_entries_left++;
	}
	return 0;
}

static int
ordered_list_pipe_queue_remove(struct doca_flow_pipe_entry *entry,
			       dpdk_flow_completion_cb completion_cb,
			       void *usr_ctx)
{
	struct doca_flow_pipe *legacy_pipe = entry->pipe;
	struct ordered_list_pipe *list_pipe = legacy_pipe->private_ctx;
	struct ordered_list_pipe_entry *list_entry = entry->obj_ctx;
	struct ordered_list_pipe_entry_ctx *list_entry_ctx;
	uint16_t queue_id = entry->pipe_queue;

	list_entry_ctx = priv_doca_mempool_alloc(list_pipe->entry_ctx_pool, queue_id);
	if (unlikely(list_entry_ctx == NULL)) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed to allocate entry context");
		return -ENOMEM;
	}

	list_entry_ctx->entry = entry;
	list_entry_ctx->list_entry = list_entry;
	list_entry_ctx->queue_id = queue_id;
	list_entry_ctx->completion_cb = completion_cb;
	list_entry_ctx->completion_cb_ctx = usr_ctx;
	list_entry_ctx->completion_op = DPDK_FLOW_REQ_OP_DESTROY;
	list_entry_ctx->completion_status = DPDK_FLOW_REQ_STATUS_OK;
	return dpdk_pipe_common_queue_remove(entry, frontend_rm_completion_cb, list_entry_ctx);
}

static int
ordered_list_pipe_queue_cleanup(struct doca_flow_pipe_entry *entry)
{
	struct ordered_list_pipe_entry_ctx *list_entry_ctx;
	struct ordered_list_pipe *list_pipe;

	/*
	 * Normally cleanup will be called after entry removal,
	 * in which case resources are cleaned up in the completion routine.
	 */
	if (likely(entry->status != DOCA_FLOW_ENTRY_STATUS_IN_PROCESS))
		return 0;

	/* Context is not set only in case of .pipe_queue_modify() failure. */
	list_entry_ctx = entry->obj_ctx;
	if (unlikely(list_entry_ctx == NULL))
		return 0;

	list_pipe = ((struct doca_flow_pipe *)entry->pipe)->private_ctx;
	priv_doca_mempool_free(list_pipe->entry_pool, list_entry_ctx->list_entry, entry->pipe_queue);
	priv_doca_mempool_free(list_pipe->entry_ctx_pool, list_entry_ctx, entry->pipe_queue);
	dpdk_pipe_entry_release(entry);
	return 0;
}

static int
ordered_list_module_init(void)
{
	basic_ops = dpdk_pipe_basic_get_ops();
	if (basic_ops == NULL) {
		DOCA_LOG_ERR("Failed to get basic pipe ops");
		return -ENOENT;
	}

	hash_ops = dpdk_pipe_hash_get_ops();
	if (hash_ops == NULL) {
		DOCA_LOG_ERR("Failed to get hash pipe ops");
		return -ENOENT;
	}

	return 0;
}

static struct dpdk_pipe_type_ops pipe_ops = {
	.module_init = ordered_list_module_init,

	.pipe_legacy_verify = ordered_list_pipe_verify,
	.pipe_legacy_alloc = ordered_list_pipe_alloc,
	.pipe_legacy_build = ordered_list_pipe_build,
	.pipe_legacy_submit = ordered_list_pipe_submit,
	.pipe_legacy_free = ordered_list_pipe_free,

	.pipe_queue_verify = ordered_list_pipe_queue_verify,
	.pipe_queue_modify = ordered_list_pipe_queue_modify,
	.pipe_queue_submit = ordered_list_pipe_queue_submit,
	.pipe_queue_remove = ordered_list_pipe_queue_remove,
	.pipe_queue_cleanup = ordered_list_pipe_queue_cleanup,
};

struct dpdk_pipe_type_ops *
dpdk_pipe_ordered_list_get_ops(void)
{
	return &pipe_ops;
}
