/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <rte_flow.h>

#include "dpdk_table.h"
#include "dpdk_table_manager.h"
#include "dpdk_pipe_legacy.h"

#include "id_pool.h"
#include "engine_spinlock.h"

#include "doca_log.h"
#include "private/common_defines.h"

DOCA_LOG_REGISTER(dpdk_table_manager);

#define TABLE_MANAGER_CACHE_SIZE (1)
#define TABLE_MANAGER_CACHE_IDX (0)
#define TABLE_MANAGER_IDPOOL_FORMAT ("tm_%u")

static uint32_t unique_id = 1;

struct dpdk_table_ctx {
	struct dpdk_table table;
	uint32_t *alloced_id;
};

struct dpdk_table_manager {
	struct id_pool *table_id_pool;
	uint32_t nr_tables;
	struct dpdk_table_ctx tables[0];
};

struct engine_spinlock id_lock;

void
dpdk_table_manager_init()
{
	engine_spinlock_init(&id_lock);
}

static void
unique_id_inc()
{
	engine_spinlock_lock(&id_lock);
	unique_id++;
	engine_spinlock_unlock(&id_lock);
}

struct dpdk_table_manager *
dpdk_table_manager_create(struct dpdk_table_manager_cfg *cfg)
{
	uint32_t mem_len;
	struct dpdk_table_manager *table_manager;
	struct id_pool_cfg pool_cfg;

	if (unlikely(!cfg)) {
		DOCA_LOG_ERR("failed creating table manager - cfg is null");
		return NULL;
	}

	if (unlikely(cfg->nr_tables == 0)) {
		DOCA_LOG_ERR("failed creating table manager - no tables defined");
		return NULL;
	}

	mem_len =
		sizeof(struct dpdk_table_manager) + sizeof(struct dpdk_table_ctx) * cfg->nr_tables;
	table_manager = calloc(1, mem_len);
	if (unlikely(!cfg)) {
		DOCA_LOG_ERR("failed creating table manager - memory allocation of %u bytes failed",
			     mem_len);
		return NULL;
	}

	pool_cfg.size = cfg->nr_tables;
	pool_cfg.nb_caches = TABLE_MANAGER_CACHE_SIZE;
	pool_cfg.min = 0;
	snprintf(pool_cfg.name, ID_POOL_NAME_LEN, TABLE_MANAGER_IDPOOL_FORMAT, unique_id);
	unique_id_inc();
	table_manager->table_id_pool = id_pool_create(&pool_cfg);
	if (!table_manager->table_id_pool) {
		DOCA_LOG_ERR("failed creating table manager - id pool creation failed");
		return NULL;
	}

	DOCA_LOG_DBG("table manager %s created", pool_cfg.name);
	table_manager->nr_tables = cfg->nr_tables;
	return table_manager;
}

void
dpdk_table_manager_destroy(struct dpdk_table_manager *table_manager)
{
	if (unlikely(!table_manager))
		return;

	DOCA_LOG_DBG("table manager destroyed");
	id_pool_destroy(table_manager->table_id_pool);
	free(table_manager);
}

static struct dpdk_table_ctx *
get_table_ctx(struct dpdk_table_manager *table_manager, uint32_t table_id)
{
	if (unlikely(!table_manager))
		return NULL;

	if (unlikely(table_id >= table_manager->nr_tables))
		return NULL;

	return &table_manager->tables[table_id];
}

int
dpdk_table_manager_alloc_id(struct dpdk_table_manager *table_manager, uint32_t *table_id)
{
	struct dpdk_table_ctx *table_ctx;
	uint32_t *new_table_id;

	if (unlikely(!table_manager)) {
		DOCA_LOG_ERR("failed allocating id from null table manager");
		return -EINVAL;
	}

	new_table_id = id_pool_alloc(table_manager->table_id_pool, TABLE_MANAGER_CACHE_IDX);
	if (unlikely(!new_table_id)) {
		DOCA_LOG_ERR("failed allocating id from table manager - memory depleted");
		return -ENOMEM;
	}

	table_ctx = get_table_ctx(table_manager, *new_table_id);
	if (unlikely(!table_ctx || table_ctx->alloced_id)) {
		DOCA_LOG_ERR("failed allocating id from table manager - memory depleted");
		return -ENOTSUP;
	}

	table_ctx->alloced_id = new_table_id;
	*table_id = *new_table_id;
	return 0;
}

struct dpdk_table *
dpdk_table_manager_get_by_id(struct dpdk_table_manager *table_manager, uint32_t table_id)
{
	struct dpdk_table_ctx *table_ctx;

	table_ctx = get_table_ctx(table_manager, table_id);
	if (!table_ctx || !table_ctx->alloced_id)
		return NULL;

	return &table_ctx->table;
}

int
dpdk_table_manager_free_id(struct dpdk_table_manager *table_manager, uint32_t table_id)
{
	struct dpdk_table_ctx *table_ctx;

	table_ctx = get_table_ctx(table_manager, table_id);
	if (unlikely(!table_ctx || !table_ctx->alloced_id)) {
		DOCA_LOG_ERR("failed freeing id %u from table manager - does not exists", table_id);
		return -ENOENT;
	}

	if (unlikely(*table_ctx->alloced_id != table_id)) {
		DOCA_LOG_ERR("failed freeing id %u from table manager - api misuse", table_id);
		return -ENOTRECOVERABLE;
	}

	id_pool_free(table_manager->table_id_pool, TABLE_MANAGER_CACHE_IDX, table_ctx->alloced_id);
	table_ctx->alloced_id = NULL;
	return 0;
}
