/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */
#include <rte_gre.h>
#include <rte_gtp.h>

#include <doca_log.h>
#include "dpdk_engine.h"
#include "doca_flow_priv.h"
#include "encap_table.h"
#include "utils_df_translate.h"

#define DEFAULT_VTC_FLOW	(0x60000000)

static void
dpdk_ether_header_build(uint8_t **header,
			const struct doca_flow_encap_action *encap_data,
			bool is_encap)
{
	struct rte_ether_hdr *eth = (struct rte_ether_hdr *)(*header);

	if (is_encap) {
		uint16_t protocol;

		memset(eth, 0, sizeof(struct rte_ether_hdr));
		if (!utils_df_translate_is_mac_ignored(encap_data->src_mac))
			utils_df_translate_copy_mac(encap_data->src_mac, (uint8_t *)&eth->s_addr);
		if (!utils_df_translate_is_mac_ignored(encap_data->dst_mac))
			utils_df_translate_copy_mac(encap_data->dst_mac, (uint8_t *)&eth->d_addr);
		protocol = encap_data->src_ip.type == DOCA_FLOW_IP4_ADDR ?
			RTE_ETHER_TYPE_IPV4 : RTE_ETHER_TYPE_IPV6;
		eth->ether_type = RTE_BE16(protocol);
	}
	*header += sizeof(*eth);

	if (is_encap && encap_data->vlan_tci) {
		struct rte_vlan_hdr *vlan = (struct rte_vlan_hdr *)(*header);

		vlan->eth_proto = eth->ether_type;
		eth->ether_type = RTE_BE16(RTE_ETHER_TYPE_VLAN);
		vlan->vlan_tci = encap_data->vlan_tci;
		*header += sizeof(*vlan);
	}
}

static void
dpdk_ipv4_header_build(uint8_t **header,
		       const struct doca_flow_encap_action *encap_data,
		       bool is_encap)
{
	struct rte_ipv4_hdr *ipv4 = (struct rte_ipv4_hdr *)(*header);

	if (is_encap) {
		memset(ipv4, 0, sizeof(struct rte_ipv4_hdr));
		if (!utils_df_translate_is_ip_ignored(&encap_data->src_ip))
			ipv4->src_addr = encap_data->src_ip.ipv4_addr;
		if (!utils_df_translate_is_ip_ignored(&encap_data->dst_ip))
			ipv4->dst_addr = encap_data->dst_ip.ipv4_addr;
		if (encap_data->tun.type == DOCA_FLOW_TUN_VXLAN)
			ipv4->next_proto_id = DOCA_PROTO_UDP;
		else if (encap_data->tun.type == DOCA_FLOW_TUN_GTPU)
			ipv4->next_proto_id = DOCA_PROTO_UDP;
		else if (encap_data->tun.type == DOCA_FLOW_TUN_GRE)
			ipv4->next_proto_id = DOCA_PROTO_GRE;
		ipv4->version_ihl = 0x45;
	}
	*header += sizeof(*ipv4);
}

static void
dpdk_udp_header_build(uint8_t **header,
		      const struct doca_flow_encap_action *encap_data,
		      bool is_encap)
{
	struct rte_udp_hdr *udp = (struct rte_udp_hdr *)(*header);

	if (is_encap) {
		memset(udp, 0, sizeof(struct rte_flow_item_udp));
		if (encap_data->tun.type == DOCA_FLOW_TUN_VXLAN)
			udp->dst_port = RTE_BE16(DOCA_VXLAN_DEFAULT_PORT);
		else if (encap_data->tun.type == DOCA_FLOW_TUN_GTPU)
			udp->dst_port = RTE_BE16(DOCA_GTPU_PORT);
	}
	*header += sizeof(*udp);
}

static void
dpdk_vxlan_header_build(uint8_t **header,
			const struct doca_flow_encap_action *encap_data,
			bool is_encap)
{
	struct rte_vxlan_hdr *vxlan = (struct rte_vxlan_hdr *)(*header);

	if (is_encap) {
		memset(vxlan, 0, sizeof(struct rte_vxlan_hdr));
		vxlan->vx_flags = RTE_BE32(0x8 << 24); /* Instance flag */
		vxlan->vx_vni = encap_data->tun.vxlan_tun_id;
	}
	*header += sizeof(*vxlan);
}

static void
dpdk_gre_header_build(uint8_t **header,
		      const struct doca_flow_encap_action *encap_data,
		      bool is_encap)
{
	struct rte_gre_hdr *gre = (struct rte_gre_hdr *)(*header);

	if (is_encap) {
		uint32_t *key_data;
		uint16_t protocol;

		memset(gre, 0, sizeof(struct rte_gre_hdr));
		gre->k = 1;
		protocol = encap_data->tun.protocol;
		gre->proto = protocol ? protocol : 0xffff;
		key_data = (uint32_t *)(*header + sizeof(*gre));
		*key_data = encap_data->tun.gre_key;
	}
	*header += sizeof(*gre); /* GRE header */
	*header += sizeof(uint32_t); /* GRE key */
}

static void
dpdk_gtp_header_build(uint8_t **header,
		      const struct doca_flow_encap_action *encap_data,
		      bool is_encap)
{
	struct rte_gtp_hdr *gtp = (struct rte_gtp_hdr *)(*header);

	if (is_encap) {
		memset(gtp, 0, sizeof(struct rte_gtp_hdr));
		gtp->ver = 1;
		gtp->pt = 1;
		gtp->msg_type = 0xff; /* G-PDU */
		gtp->teid = encap_data->tun.gtp_teid;
	}
	*header += sizeof(*gtp);
}

static void
dpdk_ipv6_header_build(uint8_t **header,
		       const struct doca_flow_encap_action *encap_data,
		       bool is_encap)
{
	struct rte_ipv6_hdr *ipv6 = (struct rte_ipv6_hdr *)(*header);

	if (is_encap) {
		memset(ipv6, 0, sizeof(struct rte_ipv6_hdr));
		if (!utils_df_translate_is_ip_ignored(&encap_data->src_ip))
			utils_df_translate_copy_ip(&encap_data->src_ip, ipv6->src_addr);
		if (!utils_df_translate_is_ip_ignored(&encap_data->dst_ip))
			utils_df_translate_copy_ip(&encap_data->dst_ip, ipv6->dst_addr);
		if (encap_data->tun.type == DOCA_FLOW_TUN_VXLAN)
			ipv6->proto = DOCA_PROTO_UDP;
		else if (encap_data->tun.type == DOCA_FLOW_TUN_GTPU)
			ipv6->proto = DOCA_PROTO_UDP;
		else if (encap_data->tun.type == DOCA_FLOW_TUN_GRE)
			ipv6->proto = DOCA_PROTO_GRE;
		ipv6->vtc_flow = RTE_BE32(DEFAULT_VTC_FLOW);
	}
	*header += sizeof(*ipv6);
}

struct endecap_layer doca_endecap_layers[] = {
	{FILL_ETH_HDR, dpdk_ether_header_build},
	{FILL_IPV4_HDR, dpdk_ipv4_header_build},
	{FILL_IPV6_HDR, dpdk_ipv6_header_build},
	{FILL_UDP_HDR, dpdk_udp_header_build},
	{FILL_VXLAN_HDR, dpdk_vxlan_header_build},
	{FILL_GRE_HDR, dpdk_gre_header_build},
	{FILL_GTP_HDR, dpdk_gtp_header_build},
};

void
dpdk_raw_data_build(uint8_t **header,
		    const struct doca_flow_encap_action *encap_data,
		    uint16_t flags, bool is_encap, uint16_t vlan_tci)
{
	struct endecap_layer *layer;
	uint8_t idx;

	for (idx = 0; idx < RTE_DIM(doca_endecap_layers); idx++) {
		layer = &doca_endecap_layers[idx];
		if (flags & layer->layer)
			layer->fill_data(header, encap_data, is_encap);
	}
}

int
dpdk_utils_build_encap_raw_data(const struct doca_flow_actions *pkt_action,
	uint8_t *header, struct rte_flow_action_raw_decap **decap_conf,
	struct rte_flow_action_raw_decap *conf, uint8_t *data)
{
	/* Some tunnel encap action requires a decap action so we
	 * get a decap action (if exists) in the pipe to add both
	 * to the dpdk flow egress table.
	 *
	 * e.g. VXLAN needs only the encap of its headers, GTP need
	 * to first decap l2 (since tunnel is for l3 only).
	 */
	switch (pkt_action->encap.tun.type) {
	case DOCA_FLOW_TUN_GRE:
	case DOCA_FLOW_TUN_GTPU:
			header = data;
			dpdk_raw_data_build(&header, NULL, FILL_ETH_HDR, false, 0);
			conf->data = data;
			conf->size = header - data;
			*decap_conf = conf;
			break;
	case DOCA_FLOW_TUN_VXLAN:
			*decap_conf = NULL;
			break;
	case DOCA_FLOW_TUN_NISP:
	case DOCA_FLOW_TUN_AUDP:
	default:
			return -ENOTSUP;
	}

	return 0;
}

