/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_hash.h>
#include <rte_hash_crc.h>
#include <rte_spinlock.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <doca_log.h>

#include "encap_table.h"

DOCA_LOG_REGISTER(encap_table);

#define ENCAP_HASH_ENTRIES 1024

struct encap_table_entry {
	uint32_t refcnt;
	struct encap_table_key key;
	void *handler;
};

struct encap_table {
	int max;
	rte_spinlock_t lock;
	struct rte_hash *h;
	struct encap_table_entry entries[0];
};

static struct encap_table *encap_table_ins;

static inline uint32_t
encap_hash_crc(const void *data, __rte_unused uint32_t data_len,
	       uint32_t init_val)
{
	const struct encap_table_key *k;
	enum encap_table_ip_type ip_type;
	int i;

	k = data;
	ip_type = k->ip_type;
	switch (ip_type) {
	case ENCAP_TABLE_IP4_ADDR:
		init_val = rte_hash_crc_4byte(k->src_ip.ipv4_addr, init_val);
		init_val = rte_hash_crc_4byte(k->dst_ip.ipv4_addr, init_val);
		break;
	case ENCAP_TABLE_IP6_ADDR:
		for (i = 0; i < ENCAP_TABLE_IPV6_ADDR_4BYTES_LEN; i++) {
			init_val = rte_hash_crc_4byte(k->src_ip.ipv6_addr[i], init_val);
			init_val = rte_hash_crc_4byte(k->dst_ip.ipv6_addr[i], init_val);
		}
		break;
	default:
		DOCA_LOG_ERR("failed calculating encap hash - unsupported ip_type %u", ip_type);
		return init_val;
	}

	switch (k->tun.type) {
	case ENCAP_TABLE_TUN_VXLAN:
		init_val = rte_hash_crc_4byte(k->tun.vxlan_tun_id, init_val);
		break;
	case ENCAP_TABLE_TUN_GRE:
		init_val = rte_hash_crc_4byte(k->tun.gre_key, init_val);
		break;
	case ENCAP_TABLE_TUN_GTPU:
		init_val = rte_hash_crc_4byte(k->tun.gtp_teid, init_val);
		break;
	default:
		break;
	}
	init_val = rte_hash_crc_4byte(k->port_id, init_val);
	init_val = rte_hash_crc_4byte(k->ip_type, init_val);
	return init_val;
}

struct rte_hash_parameters encap_table_hash_params = {
	.name = "encap_table",
	.entries = ENCAP_HASH_ENTRIES,
	.key_len = sizeof(struct encap_table_key),
	.hash_func = encap_hash_crc,
	.hash_func_init_val = 0,
};

int
encap_table_init(int max_encaps)
{
	int total_size = sizeof(struct encap_table) +
			 sizeof(struct encap_table_entry) * max_encaps;

	encap_table_ins = (struct encap_table *)calloc(1, total_size);
	if (encap_table_ins == NULL) {
		DOCA_DLOG_ERR("failed to alloc mem");
		return -ENOMEM;
	}

	encap_table_hash_params.entries = max_encaps;
	encap_table_ins->max = max_encaps;

	encap_table_ins->h = rte_hash_create(&encap_table_hash_params);
	if (encap_table_ins->h == NULL) {
		DOCA_DLOG_ERR("failed to alloc hash table %s with %u entries",
			      encap_table_hash_params.name,
			      encap_table_hash_params.entries);
		free(encap_table_ins);
		return -1;
	}
	rte_spinlock_init(&encap_table_ins->lock);
	DOCA_LOG_INFO("encap table created");
	return 0;
}

void
encap_table_destroy(void)
{
	if (encap_table_ins == NULL) {
		DOCA_DLOG_WARN("failed to destroy hash table - null pointer");
		return;
	}

	if (encap_table_ins->h)
		rte_hash_free(encap_table_ins->h);

	free(encap_table_ins);
	encap_table_ins = NULL;
	DOCA_LOG_INFO("encap table destroyed");
}

static int
encap_table_get_id(struct encap_table_key *key)
{
	int id;

	id = rte_hash_lookup(encap_table_ins->h, (void *)key);
	return id;
}

int
encap_table_add_id(struct encap_table_key *key, uint32_t *refcnt)
{
	int id = 0;

	if (key->ip_type != ENCAP_TABLE_IP4_ADDR && key->ip_type != ENCAP_TABLE_IP6_ADDR) {
		DOCA_DLOG_WARN("invalid ip type %d", key->ip_type);
		return -EINVAL;
	}

	rte_spinlock_lock(&encap_table_ins->lock);
	id = encap_table_get_id(key);
	if (id >= 0) {
		encap_table_ins->entries[id].refcnt++;
		*refcnt = encap_table_ins->entries[id].refcnt;
		rte_spinlock_unlock(&encap_table_ins->lock);
		return id;
	}

	id = rte_hash_add_key(encap_table_ins->h, (void *)key);
	if (id < 0 || id > encap_table_ins->max) {
		DOCA_DLOG_WARN("failed adding id");
		rte_spinlock_unlock(&encap_table_ins->lock);
		return -1;
	}

	encap_table_ins->entries[id].key = *key;
	encap_table_ins->entries[id].refcnt = 1;
	*refcnt = encap_table_ins->entries[id].refcnt;
	rte_spinlock_unlock(&encap_table_ins->lock);
	return id;
}

int
encap_table_remove_id(int id)
{
	if (id < 0 || id > encap_table_ins->max) {
		DOCA_DLOG_WARN("cannot delete invalid id %d", id);
		return -1;
	}

	rte_spinlock_lock(&encap_table_ins->lock);
	if (encap_table_ins->entries[id].refcnt > 0) {
		if (--encap_table_ins->entries[id].refcnt) {
			int refcnt = encap_table_ins->entries[id].refcnt;

			rte_spinlock_unlock(&encap_table_ins->lock);
			return refcnt;
		}

		rte_hash_del_key(encap_table_ins->h,
				 &encap_table_ins->entries[id].key);
		memset(&encap_table_ins->entries[id], 0,
		       sizeof(struct encap_table_entry));
	}
	rte_spinlock_unlock(&encap_table_ins->lock);
	return 0;
}
