/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENCAP_TABLE_H_
#define ENCAP_TABLE_H_


#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief ipv6 4bytes length
 */
#define ENCAP_TABLE_IPV6_ADDR_4BYTES_LEN 4

/**
 * @brief encap table ip type
 */
enum encap_table_ip_type {
	ENCAP_TABLE_ADDR_NONE = 0,
	/**< ip address is not set */
	ENCAP_TABLE_IP4_ADDR = 4,
	/**< ip address is ipv4 */
	ENCAP_TABLE_IP6_ADDR = 6,
	/**< ip address is ipv6 */
};

/**
 * @brief encap table tunnel type
 */
enum encap_table_tun_type {
	ENCAP_TABLE_TUN_NONE = 0,
	/**< tunnel is not set */
	ENCAP_TABLE_TUN_VXLAN,
	/**< tunnel is vxlan type */
	ENCAP_TABLE_TUN_GTPU,
	/**< tunnel is gtpu type */
	ENCAP_TABLE_TUN_GRE,
	/**< tunnel is gre type */
};

/**
 * @brief encap table tunnel type
 */
struct encap_table_ip_addr {
	enum encap_table_ip_type type;
	/**< ip address type */
	union {
		uint32_t ipv4_addr;
		/**< ipv4 address if type is ipv4 */
		uint32_t ipv6_addr[ENCAP_TABLE_IPV6_ADDR_4BYTES_LEN];
		/**< ipv6 address if type is ipv6 */
	};
};

/**
 * @brief encap table tunnel information
 */
struct encap_table_tun {
	enum encap_table_tun_type type;
	/**< tunnel type */
	union {
		struct {
			uint32_t vxlan_tun_id;
			/**< vxlan vni(24) + reserved (8). */
		};
		/**< vxlan information if tunnel is vxlan */
		struct {
			uint32_t gre_key;
			/**< gre key */
		};
		/**< gre information if tunnel is gre */
		struct {
			uint32_t gtp_teid;
			/**< gtp teid */
		};
		/**< gtp information if tunnel is gtp */
	};
};

/**
 * @brief encap table key
 */
struct encap_table_key {
	uint16_t port_id;
	enum encap_table_ip_type ip_type;
	struct encap_table_ip_addr src_ip;
	struct encap_table_ip_addr dst_ip;
	struct encap_table_tun tun;
};

/**
 * @brief
 *
 * @param max_encaps
 *    trying to add more encaps will result in error
 *
 * @return 0 on success
 */
int encap_table_init(int max_encaps);

/**
 * @brief - add encap action
 *      encap is identified by its 4-tuple
 *      port_id, src ip, dst ip and tun id.
 *      in case it exists refcnt++.
 *
 * @param key
 *
 * @param refcnt
 * Output: internal key ref count is written here
 *
 * @return positive value which is the encap idx.
 *  negative value on failure.
 */
int encap_table_add_id(struct encap_table_key *key,
		uint32_t *refcnt);

/**
 * @brief - remove from table.
 *  if id exits with data, data is returned.
 *  del will cause ref--, and on ref == 0,
 *  id will be cleared.
 *
 * @param id
 *
 * @return refcnt value or negative value if doesn't
 *  exists.
 */
int encap_table_remove_id(int id);

void encap_table_destroy(void);

#ifdef __cplusplus
}
#endif

#endif /* ENCAP_TABLE_H_ */
