/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stddef.h>

#include "private/common_defines.h"
#include "engine_bindable.h"

static uint32_t stamp_map[DF_ENGINE_BINDABLE_TYPE_MAX] = {
	0xB67F75BF, /* DF_ENGINE_BINDABLE_TYPE_ALL */
	0x5CD4E03B, /* DF_ENGINE_BINDABLE_TYPE_PORT */
	0x09177EBB  /* DF_ENGINE_BINDABLE_TYPE_PIPE */
};

int
engine_bindable_init(struct engine_bindable *bindable, enum engine_bindable_type type)
{
	if (unlikely(!bindable))
		return -EINVAL;

	switch (type) {
	case DF_ENGINE_BINDABLE_TYPE_ALL:
	case DF_ENGINE_BINDABLE_TYPE_PORT:
	case DF_ENGINE_BINDABLE_TYPE_PIPE:
		bindable->stamp = stamp_map[type];
		bindable->type = type;
		bindable->parent = NULL;
		bindable->priv_ctx = NULL;
		break;
	default:
		return -ENOTSUP;
	}
	return 0;
}

bool
engine_bindable_stamp_verify(const struct engine_bindable *bindable)
{
	if (unlikely(!bindable))
		return false;

	if (unlikely(bindable->type >= DF_ENGINE_BINDABLE_TYPE_MAX))
		return false;

	return bindable->stamp == stamp_map[bindable->type];
}

static inline bool
validate_hierarchy_types(struct engine_bindable *child, struct engine_bindable *parent)
{
	enum engine_bindable_type child_type = child->type;
	enum engine_bindable_type parent_type = parent->type;

	switch (child_type) {
	case DF_ENGINE_BINDABLE_TYPE_PORT:
		return parent_type == DF_ENGINE_BINDABLE_TYPE_ALL;
	case DF_ENGINE_BINDABLE_TYPE_PIPE:
		return parent_type == DF_ENGINE_BINDABLE_TYPE_PORT ||
		       parent_type == DF_ENGINE_BINDABLE_TYPE_ALL;
	case DF_ENGINE_BINDABLE_TYPE_ALL:
	default:
		return false;
	}

}

static bool
find_bindable_in_tree_leaf(const struct engine_bindable *bindable,
			   const struct engine_bindable *tree_leaf)
{
	while (tree_leaf) {
		if (tree_leaf == bindable)
			return true;

		tree_leaf = tree_leaf->parent;
	}

	return false;
}

int
engine_bindable_set_parent(struct engine_bindable *bindable, struct engine_bindable *parent)
{
	if (unlikely(!engine_bindable_stamp_verify(bindable)))
		return -EINVAL;

	if (unlikely(!engine_bindable_stamp_verify(parent)))
		return -EINVAL;

	if (unlikely(!validate_hierarchy_types(bindable, parent)))
		return -EPERM;

	if (unlikely(find_bindable_in_tree_leaf(bindable, parent)))
		return -EEXIST;

	bindable->parent = parent;
	return 0;
}

bool
engine_bindable_hierarchy_verify(const struct engine_bindable *bindable,
				 struct engine_bindable *tree_leaf)
{
	if (unlikely(!engine_bindable_stamp_verify(bindable)))
		return false;

	if (unlikely(!engine_bindable_stamp_verify(tree_leaf)))
		return false;

	return find_bindable_in_tree_leaf(bindable, tree_leaf);
}
