/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENGINE_BINDABLE_H_
#define ENGINE_BINDABLE_H_

#include <stdint.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief engine bindable types
 */
enum engine_bindable_type {
	DF_ENGINE_BINDABLE_TYPE_ALL,
	/**<  Binded to all */
	DF_ENGINE_BINDABLE_TYPE_PORT,
	/**<  Binded to a port */
	DF_ENGINE_BINDABLE_TYPE_PIPE,
	/**<  Binded to a pipe */
	DF_ENGINE_BINDABLE_TYPE_MAX,
	/**<  Maximum bindable types */
};

/**
 * @brief bindable object.
 *
 * NOTE: must be the first member of bindable objects.
 */
struct engine_bindable {
	uint32_t stamp;
	/**<  unique identifier for bindable object type */
	enum engine_bindable_type type;
	/**<  type of bindable object */
	struct engine_bindable *parent;
	/**<  parent of a hierarchy */
	void *priv_ctx;
	/**<  pointer to private context */
};

/**
 * @brief Initialize the bindable object
 *
 * The stamp is unique per bindable type for all calls.
 * The bindable object is used to identify types of objects.
 *
 * @param bindable
 * Pointer to the bindable object, filled accordingly.
 * @param type
 * Type of bindable object
 * @return
 * 0 on success, negative on failure.
 */
int
engine_bindable_init(struct engine_bindable *bindable, enum engine_bindable_type type);

/**
 * @brief Stamp verification
 *
 * Validate the correctness of a stamp by type.
 *
 * @param bindable
 * Pointer to a bindable object.
 * @param bindable_stamped
 * Stamp to validate.
 * @return
 * 0 on success, negative on failure.
 */
bool
engine_bindable_stamp_verify(const struct engine_bindable *bindable);

/**
 * @brief Set the hierarchy parent
 *
 * Sets the parent of a bindable object.
 *
 * @param bindable
 * Pointer to a bindable object
 * @param parent
 * Pointer to a parent candidate
 * @return
 * 0 on success, negative on failure.
 */
int
engine_bindable_set_parent(struct engine_bindable *bindable, struct engine_bindable *parent);

/**
 * @brief Verify bindable is in a leaf hierarchy
 *
 * Checks if a bindable is a part of a hierarchy, will check all parents up to
 * the root.
 *
 * @param bindable
 * Pointer to a bindable object.
 * @param tree_leaf
 * Pointer to a leaf to start the search from.
 * @return
 * true when found
 */
bool
engine_bindable_hierarchy_verify(const struct engine_bindable *bindable,
				 struct engine_bindable *tree_leaf);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* ENGINE_BINDABLE_H_ */
