/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdlib.h>

#include "private/common_defines.h"

#include "engine_field.h"
#include "engine_field_opcode.h"
#include "engine_fcp.h"

#include "doca_log.h"

#define MAX_NODES_IN_ITERATOR (256)
#define MAX_SELECTORS_IN_NODE (6)

DOCA_LOG_REGISTER(engine_fcp);

enum transition_node_type {
	TRANSITION_NODE_TYPE_OPCODE,
	TRANSITION_NODE_TYPE_SELECTOR_FIELD,
	TRANSITION_NODE_TYPE_MAX,
};

struct transition_node_selector {
	struct engine_fcp_selector_cfg cfg;
};

struct transition_node {
	enum transition_node_type type;
	union {
		struct engine_field_opcode opcode;
		struct transition_node_selector selector;
	};
};

struct engine_fcp {
	enum engine_supported_user_defined_type type;
	bool is_valid;
	int nr_nodes;
	struct transition_node nodes[MAX_NODES_IN_ITERATOR];
};

struct field_extract_ctx {
	fcp_iterator_cb iter_cb;
	void *ctx;
};

struct engine_fcp *
engine_fcp_create(enum engine_supported_user_defined_type type)
{
	struct engine_fcp *fcp;

	if (unlikely(type >= DF_SUP_USER_DEFINED_TYPE_MAX)) {
		DOCA_LOG_ERR("failed creating engine fcp - invalid type %u", type);
		return NULL;
	}

	fcp = calloc(1, sizeof(struct engine_fcp));
	if (unlikely(!fcp)) {
		DOCA_LOG_ERR("failed creating engine fcp - memory depleted");
		return NULL;
	}

	fcp->type = type;
	fcp->nr_nodes = 0;
	fcp->is_valid = true;
	return fcp;
}

void
engine_fcp_destroy(struct engine_fcp *fcp)
{
	if (unlikely(!fcp)) {
		DOCA_LOG_ERR("failed destroying engine fcp - null pointer");
		return;
	}

	free(fcp);
}

static bool
is_opcode_type_expected(struct engine_fcp *fcp, const struct engine_field_opcode *opcode)
{
	switch (opcode->pkt.group) {
	case ENGINE_FIELD_OPCODE_GROUP_PACKET:
		return fcp->type == opcode->pkt.type;
	case ENGINE_FIELD_OPCODE_GROUP_FLOW:
		return fcp->type == opcode->flow.type;
	default:
		return false;
	}
}

int
engine_fcp_add_opcode(struct engine_fcp *fcp, const struct engine_field_opcode *opcode)
{
	if (unlikely(!fcp)) {
		DOCA_LOG_ERR("failed adding opcode for fcp - null iterator pointer");
		return -EINVAL;
	}

	if (unlikely(!fcp->is_valid)) {
		DOCA_LOG_ERR("failed adding opcode for fcp - fcp invalid");
		return -EINVAL;
	}

	if (unlikely(!opcode)) {
		DOCA_LOG_ERR("failed adding opcode for fcp - null opcode pointer");
		return -EINVAL;
	}

	if (unlikely(!is_opcode_type_expected(fcp, opcode))) {
		DOCA_LOG_ERR("failed adding opcode for fcp - opcode type %u != iterator type %u",
			     opcode->pkt.type, fcp->type);
		return -EINVAL;
	}

	if (unlikely(fcp->nr_nodes >= MAX_NODES_IN_ITERATOR)) {
		DOCA_LOG_ERR("failed adding opcode for fcp - reached max opcodes %u in iterator",
			     MAX_NODES_IN_ITERATOR);
		return -ENOMEM;
	}

	fcp->nodes[fcp->nr_nodes].type = TRANSITION_NODE_TYPE_OPCODE;
	fcp->nodes[fcp->nr_nodes].opcode = *opcode;
	fcp->nr_nodes++;
	return 0;
}

int
engine_fcp_add_selector(struct engine_fcp *fcp, const struct engine_fcp_selector_cfg *selector_cfg)
{
	const struct engine_field_opcode *opcode;
	int select_idx;

	if (unlikely(!fcp)) {
		DOCA_LOG_ERR("failed adding selector for fcp - null fcp pointer");
		return -EINVAL;
	}

	if (unlikely(!fcp->is_valid)) {
		DOCA_LOG_ERR("failed adding selector for fcp - fcp invalid");
		return -EINVAL;
	}

	if (unlikely(!selector_cfg)) {
		DOCA_LOG_ERR("failed adding selector for fcp - null selector cfg");
		return -EINVAL;
	}

	if (unlikely(selector_cfg->selectors >= ENGINE_FCP_MAX_SELECTORS)) {
		DOCA_LOG_ERR("failed adding selector for fcp - too many selectors %u, max is %u",
			     selector_cfg->selectors, ENGINE_FCP_MAX_SELECTORS);
		return -EINVAL;
	}

	for (select_idx = 0; select_idx < selector_cfg->selectors; select_idx++) {
		opcode = &selector_cfg->selector[select_idx].opcode;

		if (unlikely(!is_opcode_type_expected(fcp, opcode))) {
			DOCA_LOG_ERR(
				"failed adding selector for fcp - opcode type %u != iterator type %u",
				opcode->pkt.type, fcp->type);
			return -EINVAL;
		}
	}

	if (unlikely(fcp->nr_nodes >= MAX_NODES_IN_ITERATOR)) {
		DOCA_LOG_ERR("failed adding selector for fcp - reached max opcodes %u in iterator",
			     MAX_NODES_IN_ITERATOR);
		return -ENOMEM;
	}

	fcp->nodes[fcp->nr_nodes].type = TRANSITION_NODE_TYPE_SELECTOR_FIELD;
	fcp->nodes[fcp->nr_nodes].selector.cfg = *selector_cfg;
	fcp->nr_nodes++;
	return 0;
}

static int
field_extract_iterator_cb(struct engine_field_opcode *opcode, const uint8_t *field_ptr,
			  uint16_t field_len, void *ctx)
{
	struct field_extract_ctx *field_ext_ctx = ctx;

	return field_ext_ctx->iter_cb(opcode, field_ptr, field_len, field_ext_ctx->ctx);
}

static int
process_opcode_transition(struct engine_field_cfg *field_cfg)
{
	return engine_field_extract(field_cfg, field_extract_iterator_cb);
}

static int
process_selector_transition(struct engine_field_cfg *field_cfg,
			    struct engine_fcp_selector_cfg *selector_cfg)
{
	const uint8_t *field_ptr;
	int rc;

	rc = engine_field_get(field_cfg, selector_cfg->offset, selector_cfg->length, &field_ptr);
	if (unlikely(rc))
		return rc;

	for (int select_idx = 0; select_idx < selector_cfg->selectors; select_idx++) {
		struct engine_fcp_selector *selector_node = &selector_cfg->selector[select_idx];

		if (memcmp(field_ptr, &selector_node->value, selector_cfg->length) != 0)
			continue;

		field_cfg->opcode = selector_node->opcode;
		rc = engine_field_extract(field_cfg, field_extract_iterator_cb);
		if (unlikely(rc))
			return rc;
	}

	return 0;
}


int
engine_fcp_iterate(struct engine_fcp *fcp, const void *user_defined_ptr, int user_defined_len,
		   fcp_iterator_cb iter_cb, void *ctx)
{
	struct field_extract_ctx field_ext_ctx = {0};
	struct engine_field_cfg field_cfg = {0};
	int node_idx;
	int rc;

	if (unlikely(!fcp)) {
		DOCA_LOG_ERR("failed iterating on engine fields - null pointer");
		return -EINVAL;
	}

	if (unlikely(!fcp->is_valid)) {
		DOCA_LOG_ERR("failed iterating on engine fields - iterator invalid");
		return -EINVAL;
	}

	if (unlikely(!user_defined_ptr)) {
		DOCA_LOG_ERR("failed iterating on engine fields - null user defined ptr");
		return -EINVAL;
	}

	if (unlikely(user_defined_len <= 0)) {
		DOCA_LOG_ERR("failed iterating on engine fields - invalid user defined length %d",
			     user_defined_len);
		return -EINVAL;
	}

	if (unlikely(!iter_cb)) {
		DOCA_LOG_ERR("failed iterating on engine fields - iterator callback is null");
		return -EINVAL;
	}

	field_ext_ctx.iter_cb = iter_cb;
	field_ext_ctx.ctx = ctx;

	field_cfg.base = (uint8_t *)user_defined_ptr;
	field_cfg.base_len = user_defined_len;
	field_cfg.ctx = &field_ext_ctx;

	for (node_idx = 0; node_idx < fcp->nr_nodes; node_idx++) {
		struct transition_node *node = &fcp->nodes[node_idx];

		switch (node->type) {
		case TRANSITION_NODE_TYPE_OPCODE:
			field_cfg.opcode = node->opcode;

			rc = process_opcode_transition(&field_cfg);
			if (unlikely(rc))
				return rc;
			break;

		case TRANSITION_NODE_TYPE_SELECTOR_FIELD:
			rc = process_selector_transition(&field_cfg, &node->selector.cfg);
			if (unlikely(rc))
				return rc;
			break;
		default:
			return -ENOTSUP;
		}
	}

	return 0;
}
