/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENGINE_FIELD_CHAIN_PROVIDER_H_
#define ENGINE_FIELD_CHAIN_PROVIDER_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include "engine_field_opcode.h"
#include "engine_supported_types.h"

#define ENGINE_FCP_MAX_SELECTORS (6)

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief field chain provider
 *
 * This data structure is responsible to provide a chain of fields.
 * It is configured to go over fields in a user defined struct in some order.
 */
struct engine_fcp;

/**
 * @brief fcp iterator callback
 */
typedef int (*fcp_iterator_cb)(const struct engine_field_opcode *opcode, const uint8_t *field_ptr,
			       uint16_t field_len, void *ctx);

/**
 * @brief fcp selector
 * Choose the opcode if the field value equals this value.
 */
struct engine_fcp_selector {
	uint32_t value;
	struct engine_field_opcode opcode;
};

/**
 * @brief fcp selector configuration
 * Defines a selector field and all opcode <-> value pairing options.
 */
struct engine_fcp_selector_cfg {
	uint32_t offset;
	uint16_t length;
	int selectors;
	struct engine_fcp_selector selector[ENGINE_FCP_MAX_SELECTORS];
};


/**
 * @brief creates a field chain provider
 * Contains the logic of traversing through fields of a user defined struct.
 * Each user defined struct type needs a well-defined chain provider logic.
 *
 * @param type
 * User defined struct type
 * @return
 * Valid fcp pointer on success, NULL on failure.
 */
struct engine_fcp *
engine_fcp_create(enum engine_supported_user_defined_type type);

/**
 * @brief Add opcode on fcp
 * Adds an opcode to the fcp
 *
 * @param fcp
 * Pointer to a field chain provider.
 * @param opcode
 * Pointer to a field opcode
 * @return
 * 0 on success, negative error code on fcp failure.
 */
int
engine_fcp_add_opcode(struct engine_fcp *fcp, const struct engine_field_opcode *opcode);

/**
 * @brief Add selector on fcp
 * The actual value on the user defined field is used to select
 * which field will be iterated next.
 *
 *
 * @param fcp
 * Pointer to a field chain provider.
 * @param selector_cfg
 * Pointer to a selector configuration.
 * @return
 * 0 on success, negative error code on fcp failure.
 */
int
engine_fcp_add_selector(struct engine_fcp *fcp, const struct engine_fcp_selector_cfg *selector_cfg);

/**
 * @brief Iterates over a fcp corresponding to a user defined struct.
 * Iterates on a user defines struct by the traversal logic dictated by the fcp.
 *
 * @param fcp
 * Pointer to a fcp.
 * @param user_defined_ptr
 * Pointer to a user defined struct.
 * @param user_defined_len
 * Length in bytes of the user defined struct.
 * @param iter_cb
 * Callback for each valid field found.
 * @param ctx
 * Context passing to iter_cb.
 * @return
 * 0 on success, negative error code on fcp iterator failure.
 */
int
engine_fcp_iterate(struct engine_fcp *fcp, const void *user_defined_ptr, int user_defined_len,
		   fcp_iterator_cb iter_cb, void *ctx);

/**
 * @brief Destroys a fcp.
 *
 * @param fcp
 * Pointer to a field chain provider.
 */
void
engine_fcp_destroy(struct engine_fcp *fcp);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* ENGINE_FIELD_CHAIN_PROVIDER_H_ */
