/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>

#include "private/common_defines.h"
#include "engine_field.h"
#include "engine_field_mapping.h"
#include "doca_log.h"

#define IGNORED_BYTE_VAL	(0x00)
#define CHANGEABLE_BYTE_VAL	(0xff)

DOCA_LOG_REGISTER(engine_field);

static int
get_field_params(const struct engine_field_cfg *field_cfg, const uint8_t **field_ptr,
		 uint16_t *field_len)
{
	const struct engine_field_map *field_map;
	int rc;

	if (unlikely(!field_cfg))
		return -EINVAL;

	field_map = engine_field_mapping_get(&field_cfg->opcode);
	if (unlikely(!field_map))
		return -ENOTSUP;

	if (unlikely(field_map->offset == ENGINE_INVALID_FIELD_OFFSET)) {
		DOCA_LOG_ERR("failed getting field params - invalid mapping offset");
		return -EINVAL;
	}

	if (unlikely(field_map->length == ENGINE_INVALID_FIELD_LENGTH)) {
		DOCA_LOG_ERR("failed getting field params - invalid mapping length");
		return -EINVAL;
	}

	rc = engine_field_get(field_cfg, field_map->offset, field_map->length, field_ptr);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed getting field params - rc=%d", rc);
		return rc;
	}

	*field_len = field_map->length;
	return 0;
}

int
engine_field_get(const struct engine_field_cfg *field_cfg, uint32_t offset, uint16_t length,
		 const uint8_t **field_ptr)
{
	if (unlikely(!field_cfg))
		return -EINVAL;

	if (unlikely(offset + length > field_cfg->base_len)) {
		DOCA_LOG_ERR("failed getting field - out of bounds - length %u exceeds base len %u",
			     offset + length, field_cfg->base_len);
		return -EINVAL;
	}

	*field_ptr = field_cfg->base + offset;
	return 0;
}

int
engine_field_extract(struct engine_field_cfg *field_cfg, field_extraction_cb extraction_cb)
{
	const uint8_t *field_ptr;
	uint16_t field_len;
	int rc;

	if (unlikely(!field_cfg)) {
		DOCA_LOG_ERR("failed field extraction - field_cfg is null");
		return -EINVAL;
	}

	if (unlikely(!extraction_cb)) {
		DOCA_LOG_ERR("failed field extraction - extraction callback is null");
		return -EINVAL;
	}

	rc = get_field_params(field_cfg, &field_ptr, &field_len);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed field extraction - getting field params got rc=%d", rc);
		return rc;
	}

	rc = extraction_cb(&field_cfg->opcode, field_ptr, field_len, field_cfg->ctx);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed field extraction - callback got rc=%d", rc);
		return rc;
	}

	DOCA_DLOG_DBG("Field extracted: opcode=0x%lx field ptr=%p field length=%u",
		      field_cfg->opcode.opcode, field_ptr, field_len);
	return 0;
}

int
field_get_property(const uint8_t *field_ptr, uint16_t field_len,
		   enum engine_field_property *property)
{
	bool changeable = true;
	bool ignored = true;
	int i;

	if (unlikely(!field_ptr)) {
		DOCA_LOG_ERR("failed getting field property - field_ptr is null");
		return -EINVAL;
	}

	if (unlikely(field_len == 0)) {
		DOCA_LOG_ERR("failed getting field property - field_len is 0");
		return -EINVAL;
	}

	if (unlikely(!property)) {
		DOCA_LOG_ERR("failed getting field property - property is null");
		return -EINVAL;
	}

	*property = DF_ENGINE_FIELD_PROPERTY_SPECIFIC;

	for (i = 0; (changeable || ignored) && i < field_len; i++) {
		if (changeable && field_ptr[i] != CHANGEABLE_BYTE_VAL)
			changeable = false;
		if (ignored && field_ptr[i] != IGNORED_BYTE_VAL)
			ignored = false;
	}

	if (changeable)
		*property = DF_ENGINE_FIELD_PROPERTY_CHANGEABLE;
	else if (ignored)
		*property = DF_ENGINE_FIELD_PROPERTY_IGNORE;

	return 0;
}

int
engine_field_get_property(const uint8_t *field_ptr, uint16_t field_len,
			  enum engine_field_property *property)
{
	return field_get_property(field_ptr, field_len, property);
}

int
engine_field_get_properties(const struct engine_field_cfg *field_cfg,
			    struct engine_field_properties *properties)
{
	const uint8_t *field_ptr;
	uint16_t field_len;
	int rc;

	if (unlikely(!field_cfg)) {
		DOCA_LOG_ERR("failed getting field properties - field_cfg is null");
		return -EINVAL;
	}

	if (unlikely(!properties)) {
		DOCA_LOG_ERR("failed getting field properties - properties is null");
		return -EINVAL;
	}

	rc = get_field_params(field_cfg, &field_ptr, &field_len);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed getting field properties - getting field params got rc=%d",
			     rc);
		return rc;
	}

	rc = field_get_property(field_ptr, field_len, &properties->usability);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed getting field properties - property got rc=%d", rc);
		return rc;
	}

	DOCA_DLOG_DBG("Field properties for opcode=0x%lx: field ptr=%p field length=%u is %u",
		      field_cfg->opcode.opcode, field_ptr, field_len, properties->usability);
	return 0;
}
