/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENGINE_FIELD_H_
#define ENGINE_FIELD_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include "engine_field_opcode.h"
#include "engine_supported_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief field extraction callback
 */
typedef int (*field_extraction_cb)(struct engine_field_opcode *opcode, const uint8_t *field_addr,
				   uint16_t field_len, void *ctx);

/**
 * @brief engine field configuration
 */
struct engine_field_cfg {
	struct engine_field_opcode opcode;
	/**< field opcode */
	const uint8_t *base;
	/**< base address of some packed fields */
	uint16_t base_len;
	/**< address valid length */
	void *ctx;
	/**< context used by registered modules */
};

/**
 * @brief engine field property
 */
enum engine_field_property {
	DF_ENGINE_FIELD_PROPERTY_CHANGEABLE,
	/**< Field is changeable */
	DF_ENGINE_FIELD_PROPERTY_SPECIFIC,
	/**< Field is specific */
	DF_ENGINE_FIELD_PROPERTY_IGNORE,
	/**< Field is ignored */
	DF_ENGINE_FIELD_PROPERTY_MAX,
	/**< delimiter for supported properties */
};

/**
 * @brief engine field properties
 */
struct engine_field_properties {
	enum engine_field_property usability;
};

/**
 * @brief Get field params from a user defined struct.
 *
 * Using the field configuration, this function uses offset and length to decipher
 * if a field in the user defined struct is valid, if it does, field_ptr and field_len
 * are filled.
 *
 * @param field_cfg
 * Pointer to field configuration.
 * @param offset
 * The offset of a field in the user defined struct
 * @param length
 * The length of a field in the user defined struct
 * @param field_ptr
 * Field set pointer
 * @return
 * 0 on success and field_ptr is set, negative on failure.
 */
int
engine_field_get(const struct engine_field_cfg *field_cfg, uint32_t offset, uint16_t length,
		 const uint8_t **field_ptr);

/**
 * @brief Extract a field from a user defined struct.
 *
 * Using the field configuration, this function expects a valid opcode that can be mapped to
 * a field in a user defined struct set on the base pointer.
 * If a field is valid, the extraction callback is being called along with the context pointer
 * given by the user for further processing.
 *
 * @param field_cfg
 * Field configuration.
 * @param extraction_cb
 * Callback for field extraction, will be called when a field is valid.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_field_extract(struct engine_field_cfg *field_cfg, field_extraction_cb extraction_cb);

/**
 * @brief Determine a field property
 *
 * @param field_cfg
 * Field configuration.
 * @param properties
 * Pointer to a field properties struct.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_field_get_properties(const struct engine_field_cfg *field_cfg,
			    struct engine_field_properties *properties);

/**
 * @brief Determine a field property
 *
 * @param field_ptr
 * Field pointer.
 * @param field_len
 * Field length.
 * @param property
 * Pointer to a field property enum.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_field_get_property(const uint8_t *field_ptr, uint16_t field_len,
			  enum engine_field_property *property);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* ENGINE_FIELD_H_ */
