/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <string.h>

#include "private/common_defines.h"
#include "engine_field_mapping.h"
#include "engine_field_opcode.h"

#include "doca_log.h"

DOCA_LOG_REGISTER(engine_field_mapping);

struct protocol_fields {
	struct engine_field_map params_per_field[DF_SUP_PROTO_FIELD_MAX];
};

struct packet_focus {
	struct protocol_fields proto_fields[DF_SUP_PROTO_MAX];
};

static struct packet_focus focus_groups[DF_SUP_FOCUS_MAX];

static struct engine_field_map *
field_map_get(const struct engine_field_opcode *opcode)
{
	if (unlikely(opcode->pkt.type >= DF_SUP_USER_DEFINED_TYPE_MAX))
		return NULL;

	if (unlikely(opcode->pkt.group != ENGINE_FIELD_OPCODE_GROUP_PACKET))
		return NULL;

	if (unlikely(opcode->pkt.focus >= DF_SUP_FOCUS_MAX))
		return NULL;

	if (unlikely(opcode->pkt.proto >= DF_SUP_PROTO_MAX))
		return NULL;

	return &focus_groups[opcode->pkt.focus]
			.proto_fields[opcode->pkt.proto]
			.params_per_field[opcode->pkt.field];
}

static int
field_map_init()
{
	struct engine_field_map *field_map;
	struct engine_field_opcode opcode;
	uint16_t focus;
	uint16_t proto;
	uint16_t field;
	int rc;

	for (focus = 0; focus < DF_SUP_FOCUS_MAX; focus++) {
		for (proto = 0; proto < DF_SUP_PROTO_MAX; proto++) {
			for (field = 0; field < DF_SUP_PROTO_FIELD_MAX; field++) {

				rc = engine_field_opcode_build_pkt(&opcode,
								   DF_SUP_USER_DEFINED_TYPE_MATCH,
								   focus, proto, field);
				if (rc)
					return rc;

				field_map = field_map_get(&opcode);
				field_map->opcode = opcode;
				field_map->offset = ENGINE_INVALID_FIELD_OFFSET;
				field_map->length = ENGINE_INVALID_FIELD_LENGTH;
			}
		}
	}
	return 0;
}

int
engine_field_mapping_init()
{
	int rc;

	rc = field_map_init();
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed initializing field map rc=%d", rc);
		return rc;
	}

	DOCA_LOG_INFO("Engine field mapping initialized with %u focus %u protocols",
		      DF_SUP_FOCUS_MAX, DF_SUP_PROTO_MAX);
	return 0;
}

void
engine_field_mapping_destroy(void)
{
	DOCA_LOG_INFO("Engine field mapping destroyed");
}

int
engine_field_mapping_set_length(const struct engine_field_opcode *opcode, uint16_t field_len)
{
	struct engine_field_map *field_param;

	if (unlikely(field_len == ENGINE_INVALID_FIELD_LENGTH))
		return -EINVAL;

	field_param = field_map_get(opcode);
	if (unlikely(!field_param))
		return -ENOTSUP;

	if (unlikely(field_param->length != ENGINE_INVALID_FIELD_LENGTH)) {
		DOCA_LOG_ERR("failed setting length of opcode 0x%08lx - already set to %u",
			     opcode->opcode, field_param->length);
		return -EPERM;
	}

	field_param->length = field_len;
	field_param->opcode = *opcode;
	DOCA_DLOG_DBG("Setting opcode 0x%lx field length to %u ", opcode->opcode, field_len);
	return 0;
}

int
engine_field_mapping_set_offset(const struct engine_field_opcode *opcode, uint32_t offset)
{
	struct engine_field_map *field_param;

	if (unlikely(offset == ENGINE_INVALID_FIELD_OFFSET))
		return -EINVAL;

	field_param = field_map_get(opcode);
	if (!field_param)
		return -ENOTSUP;

	if (field_param->offset != ENGINE_INVALID_FIELD_OFFSET) {
		DOCA_LOG_ERR("failed setting offset of opcode 0x%08lx - already set to %u",
			     opcode->opcode, field_param->offset);
		return -EPERM;
	}

	field_param->offset = offset;
	field_param->opcode = *opcode;
	DOCA_DLOG_DBG("Setting opcode 0x%lx field offset to %u", opcode->opcode, offset);
	return 0;
}

const struct engine_field_map *
engine_field_mapping_get(const struct engine_field_opcode *opcode)
{
	return field_map_get(opcode);
}

