/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENGINE_FIELD_MAPPING_H_
#define ENGINE_FIELD_MAPPING_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include "engine_field_opcode.h"
#include "engine_field.h"
#include "engine_supported_types.h"

#ifdef __cplusplus
extern "C" {
#endif

#define ENGINE_INVALID_FIELD_OFFSET (UINT32_MAX)
#define ENGINE_INVALID_FIELD_LENGTH (0)

/**
 * @brief Engine field mapping by opcode.
 *
 * Used to identify a field inside a user defined struct by opcode.
 */
struct engine_field_map {
	struct engine_field_opcode opcode;
	/**< engine field opcode */
	uint32_t offset;
	/**< engine field offset */
	uint16_t length;
	/**< engine field length */
};

/**
 * @brief Initializes the engine field mapping module
 *
 * Initializing the memory used in the engine field mapping.
 * This layer maintain the offset and length of a field.
 *
 * @return
 * 0 on success, negative on failure.
 */
int
engine_field_mapping_init(void);

/**
 * @brief Destroys the field mapping module.
 *
 */
void
engine_field_mapping_destroy(void);

/**
 * @brief Sets the field map length by opcode.
 *
 * @param opcode
 * Pointer to an opcode.
 * @param field_len
 * The field length value to map.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_field_mapping_set_length(const struct engine_field_opcode *opcode, uint16_t field_len);

/**
 * @brief Sets the field map offset by opcode.
 *
 * @param opcode
 * Pointer to an opcode.
 * @param offset
 * The field offset value to map.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_field_mapping_set_offset(const struct engine_field_opcode *opcode, uint32_t offset);

/**
 * @brief Gets the field mapping by opcode
 *
 * @param opcode
 * Pointer to an opcode.
 * @return
 * Pointer to a valid field map, NULL if invalid.
 */
const struct engine_field_map *
engine_field_mapping_get(const struct engine_field_opcode *opcode);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* ENGINE_FIELD_MAPPING_H_ */
