/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENGINE_FIELD_OPCODE_H_
#define ENGINE_FIELD_OPCODE_H_

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <assert.h>

#include "engine_supported_types.h"

#ifdef __cplusplus
extern "C" {
#endif

enum engine_field_opcode_group {
	ENGINE_FIELD_OPCODE_GROUP_PACKET,
	ENGINE_FIELD_OPCODE_GROUP_FLOW,
	ENGINE_FIELD_OPCODE_GROUP_MAX,
};

static_assert(ENGINE_FIELD_OPCODE_GROUP_MAX < 4, "there can only be 4 opcode groups");
static_assert(DF_SUP_USER_DEFINED_TYPE_MAX < 16, "there can only be 16 opcode types");
static_assert(DF_SUP_FOCUS_MAX < 65536, "there can only be 65536 opcode focuses");
static_assert(DF_SUP_PROTO_MAX < 65536, "there can only be 65536 opcode protocols");

/**
 * @brief Engine field opcode.
 *
 * Global engine unique opcode that is used to identify user intention.
 * It has a constant size of 64-bit.
 *
 */
struct engine_field_opcode {
	union {
		uint64_t opcode;
		union {
			struct {
				uint64_t group : 2;
				/**< Expected packet group */
				uint64_t type : 4;
				/**<  User defined struct type */
				uint64_t focus : 16;
				/**<  Protocol order (outer/tunnel/inner) */
				uint64_t proto : 16;
				/**<  Protocol type */
				uint64_t field : 8;
				/**<  Protocl field type */
				uint64_t reserved : 18;
			} pkt;

			struct {
				uint64_t group : 2;
				/**< Expected flow group */
				uint64_t type : 4;
				/**<  User defined struct type */
				uint64_t code : 32;
				/**<  Code related to flow */
				uint64_t reserved : 26;
			} flow;
		};
	};
};

static_assert(sizeof(struct engine_field_opcode) == 8, "size of engine field opcode is invalid");

int
engine_field_opcode_build_pkt(struct engine_field_opcode *opcode, uint8_t type, uint16_t focus,
			      uint16_t proto, uint8_t field);

int
engine_field_opcode_build_flow(struct engine_field_opcode *opcode, uint32_t code);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* ENGINE_FIELD_OPCODE_H_ */
