/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <string.h>

#include "private/common_defines.h"
#include "engine_model.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(engine_model);

/**
 * @brief engine model
 */
struct engine_model {
	bool is_init;
	/**< engine model initialized flag */
	bool is_verification_enabled;
	/**< when set enables verification */
	enum engine_model_mode_type mode;
	/**< engine model mode */
	struct mode_param mode_param;
	/**< param for each mode */
	uint16_t nr_pipe_queues;
	/**< engine model number of pipe queues */
	uint32_t pipe_queue_depth;
	/**< engine model pipe queue depth */
	enum engine_model_isolation isolated;
	/**< engine model isolated mode */
};

static struct engine_model engine_model = {0};

int
engine_model_init(struct engine_model_cfg *cfg)
{
	if (unlikely(!cfg)) {
		DOCA_LOG_ERR("failed init engine model - cfg is null");
		return -EINVAL;
	}

	if (engine_model.is_init) {
		DOCA_LOG_WARN("failed init engine model - already initialized");
		return -EALREADY;
	}

	if (cfg->mode >= ENGINE_MODEL_MODE_MAX) {
		DOCA_LOG_ERR("failed init engine model - invalid mode %u", cfg->mode);
		return -EINVAL;
	}

	engine_model.mode = cfg->mode;
	engine_model.nr_pipe_queues = cfg->nr_pipe_queues;
	engine_model.pipe_queue_depth = cfg->pipe_queue_depth;
	engine_model.isolated = (cfg->isolated == ENGINE_MODEL_ISOLATED_UNDEF) ?
				 ENGINE_MODEL_NON_ISOLATED : cfg->isolated;
	engine_model.is_verification_enabled = cfg->is_verification_enabled;
	engine_model.is_init = true;
	engine_model.mode_param = cfg->mode_param;

	DOCA_LOG_INFO("engine model defined with mode=%s",
		      engine_model_get_mode_name(engine_model.mode));
	DOCA_LOG_INFO("engine model defined with nr_pipe_queues=%u", engine_model.nr_pipe_queues);
	DOCA_LOG_INFO("engine model defined with pipe_queue_depth=%u",
		      engine_model.pipe_queue_depth);
	if (engine_model.isolated == ENGINE_MODEL_ISOLATED)
		DOCA_LOG_INFO("engine model defined in isolated mode");
	return 0;
}

bool
engine_model_is_mode(enum engine_model_mode_type expected_mode)
{
	return engine_model.mode == expected_mode;
}

void
engine_model_get_param(struct mode_param *mode_param)
{
	*mode_param = engine_model.mode_param;
}

bool
engine_model_is_verification_enabled()
{
	return engine_model.is_verification_enabled;
}

enum engine_model_mode_type
engine_model_get_mode(void)
{
	return engine_model.mode;
}

const char*
engine_model_get_mode_name(enum engine_model_mode_type mode)
{
	switch (mode) {
	case ENGINE_MODEL_MODE_VNF:
		return ENGINE_MODE_STR_VNF;
	case ENGINE_MODEL_MODE_SWITCH:
		return ENGINE_MODE_STR_SWITCH;
	case ENGINE_MODEL_MODE_REMOTE_VNF:
		return ENGINE_MODE_STR_REMOTE_VNF;
	default:
		return "unsupport";
	}
}


uint16_t
engine_model_get_pipe_queues(void)
{
	return engine_model.nr_pipe_queues;
}

uint32_t
engine_model_get_pipe_queue_depth(void)
{
	return engine_model.pipe_queue_depth;
}

bool
engine_model_is_isolated(void)
{
	return engine_model.isolated == ENGINE_MODEL_ISOLATED;
}

void
engine_model_destroy()
{
	if (!engine_model.is_init) {
		DOCA_LOG_ERR("engine model was never initialized");
		return;
	}

	memset(&engine_model, 0, sizeof(engine_model));
	DOCA_LOG_INFO("engine model destroyed");
}
