/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENGINE_MODEL_H_
#define ENGINE_MODEL_H_

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>

#define ENGINE_MODE_STR_VNF "vnf"
#define ENGINE_MODE_STR_SWITCH "switch"
#define ENGINE_MODE_STR_REMOTE_VNF "remote_vnf"
#define ENGINE_MODE_STR_HYBRID "hybrid"
#define ENGINE_FLAG_STR_VERIFY "verify"
#define ENGINE_ISOLATED "isolated"
#define HYBRID_SLAVE_LAG "slave_lag="

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Engine model mode types
 */
enum engine_model_mode_type {
	ENGINE_MODEL_MODE_VNF,
	/**< VNF mode */
	ENGINE_MODEL_MODE_SWITCH,
	/**< Switch mode */
	ENGINE_MODEL_MODE_REMOTE_VNF,
	/**< Remote-VNF mode */
	ENGINE_MODEL_MODE_HYBRID,
	/**< Hybrid mode */
	ENGINE_MODEL_MODE_MAX,
	/**< Max supported mode type */
};

enum engine_model_isolation {
	ENGINE_MODEL_NON_ISOLATED,
	/**< Non-isolated mode */
	ENGINE_MODEL_ISOLATED,
	/**< Isolated mode */
	ENGINE_MODEL_ISOLATED_UNDEF,
	/**< Undef isolation mode */
};

/**< engine model extend params */
struct mode_param {
	union {
		struct {
			uint16_t secondary_lag_id;
		} hybrid;
	};
};

/**
 * @brief engine model configuration
 */
struct engine_model_cfg {
	bool is_verification_enabled;
	/**< when set enables verification */
	enum engine_model_mode_type mode;
	/**< engine model mode */
	struct mode_param mode_param;
	/**< param for each mode */
	uint16_t nr_pipe_queues;
	/**< engine model number of pipe queues */
	uint32_t pipe_queue_depth;
	/**< engine model queue depth per pipe */
	enum engine_model_isolation isolated;
	/**< engine model isolation */
};

/**
 * @brief Initialize engine model.
 *
 * @return
 * 0 on success, negative on failure.
 */
int
engine_model_init(struct engine_model_cfg *cfg);

/**
 * @brief Engine mode comparison
 *
 * @return
 * true if mode is
 */
bool
engine_model_is_mode(enum engine_model_mode_type expected_mode);

/**
 * @brief Engine pipe param getter
 *
 * @return
 */
void
engine_model_get_param(struct mode_param *mode_param);


/**
 * @brief Engine is_verification flag getter
 *
 * @return
 * true if verification enabled
 */
bool
engine_model_is_verification_enabled();

/**
 * @brief Engine mode getter
 *
 * @return
 * Mode type.
 */
enum engine_model_mode_type
engine_model_get_mode(void);

/**
 * @brief Engine mode name getter
 *
 * @return
 * Mode type string.
 */
const char*
engine_model_get_mode_name(enum engine_model_mode_type mode);

/**
 * @brief Engine number of pipe queues getter
 *
 * @return
 * Number of queues
 */
uint16_t
engine_model_get_pipe_queues(void);

/**
 * @brief Engine pipe queue depth getter
 *
 * @return
 * Queue depth
 */
uint32_t
engine_model_get_pipe_queue_depth(void);

/**
 * @brief Engine isolated status getter.
 *
 * @return
 * true/false if isolated mode was requested.
 */
bool
engine_model_is_isolated(void);

/**
 * @brief Destroy engine model.
 */
void
engine_model_destroy();

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* ENGINE_MODEL_H_ */
