/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "private/common_defines.h"
#include "engine_shared_resources.h"
#include "engine_bindable.h"
#include "engine_spinlock.h"

#include <doca_log.h>

DOCA_LOG_REGISTER(engine_shared_resources);

struct shared_resource {
	struct shared_resource *next;
	struct engine_bindable *bindable;
	uint32_t id;
	enum engine_shared_resource_type type;
	bool is_bound;
	bool pending_cfg;
	struct engine_shared_resource_cfg res_cfg;
};

struct shared_resources {
	struct engine_spinlock big_lock;
	struct engine_shared_resources_cfg cfg;
	struct engine_shared_resources_ops ops;
	bool ops_set;
	bool is_init;
	struct shared_resource *res[ENGINE_SHARED_RESOURCE_MAX];
};

static struct shared_resources resources_engine = {0};
static struct engine_bindable global_bindable_obj = {0};

static int shared_resource_unbind_safe(struct engine_bindable *bindable);
static int invoke_ops_init_on_all_resource_types(void);

int
engine_shared_resources_init(struct engine_shared_resources_cfg *res_cfg)
{
	uint32_t res_type;
	int rc;

	if (resources_engine.is_init)
		return -EALREADY;

	for (res_type = 0; res_type < ENGINE_SHARED_RESOURCE_MAX; res_type++) {
		uint32_t mem_len;

		if (res_cfg->total_resources[res_type] == 0) {
			resources_engine.res[res_type] = NULL;
			continue;
		}

		mem_len = res_cfg->total_resources[res_type] * sizeof(struct shared_resource);
		resources_engine.res[res_type] = calloc(1, mem_len);
		if (!resources_engine.res[res_type]) {
			rc = -ENOMEM;
			goto init_failure;
		}
	}

	engine_spinlock_init(&resources_engine.big_lock);
	resources_engine.cfg = *res_cfg;
	resources_engine.is_init = true;

	if (resources_engine.ops_set) {
		rc = invoke_ops_init_on_all_resource_types();
		if (rc)
			goto init_failure;
	}

	rc = engine_bindable_init(&global_bindable_obj, DF_ENGINE_BINDABLE_TYPE_ALL);
	if (rc) {
		DOCA_LOG_ERR("failed initializing shared resources - global bindable rc=%d", rc);
		goto init_failure;
	}

	DOCA_LOG_INFO("Engine shared resources initialized successfully");
	return 0;

init_failure:
	engine_shared_resources_destroy();
	return rc;
}

struct engine_bindable *
engine_shared_resource_get_global_bindable(void)
{
	if (unlikely(!engine_bindable_stamp_verify(&global_bindable_obj)))
		return NULL;

	return &global_bindable_obj;
}

static int
invoke_ops_init_on_all_resource_types()
{
	uint32_t res_type;
	uint32_t rollback_res_type;
	int rc;
	int rollback_rc;

	for (res_type = 0; res_type < ENGINE_SHARED_RESOURCE_MAX; res_type++) {
		uint32_t nr_resources = resources_engine.cfg.total_resources[res_type];

		if (nr_resources == 0)
			continue;

		rc = resources_engine.ops.resources_init(res_type, nr_resources);
		if (unlikely(rc)) {
			DOCA_LOG_ERR(
				"failed allocating %u resources of type %u - starting rollback",
				nr_resources, res_type);
			goto rollback;
		}

		DOCA_LOG_INFO("Allocated %u shared resources of type %u", nr_resources, res_type);
	}

	return 0;
rollback:

	for (rollback_res_type = 0; rollback_res_type < res_type; rollback_res_type++) {
		if (resources_engine.cfg.total_resources[rollback_res_type] == 0)
			continue;

		rollback_rc = resources_engine.ops.resources_cleanup(rollback_res_type);
		if (rollback_rc)
			DOCA_LOG_CRIT("failed rollback on shared resource type %u rc=%d",
				      rollback_res_type, rollback_rc);
		else
			DOCA_LOG_INFO("Cleanup shared resources type %u - rollback",
				      rollback_res_type);
	}

	return rc;
}

int
engine_shared_resource_register_ops(struct engine_shared_resources_ops *ops)
{
	if (!ops)
		return -EINVAL;

	if (!ops->resources_init)
		return -EINVAL;
	if (!ops->resource_create)
		return -EINVAL;
	if (!ops->resource_modify)
		return -EINVAL;
	if (!ops->resource_query)
		return -EINVAL;
	if (!ops->resource_destroy)
		return -EINVAL;
	if (!ops->resources_cleanup)
		return -EINVAL;

	resources_engine.ops = *ops;
	if (resources_engine.is_init && !resources_engine.ops_set)
		invoke_ops_init_on_all_resource_types();

	resources_engine.ops_set = true;
	return 0;
}

static inline struct shared_resource *
get_shared_resource(enum engine_shared_resource_type type, uint32_t id)
{
	if (unlikely(type >= ENGINE_SHARED_RESOURCE_MAX))
		return NULL;

	if (unlikely(id >= resources_engine.cfg.total_resources[type]))
		return NULL;

	return &resources_engine.res[type][id];
}

static void
shared_resource_pre_destroy(uint32_t type)
{
	struct engine_shared_resource_cfg *cfg;
	struct shared_resource *res;
	int i;
	uint32_t total_res;

	if (type != ENGINE_SHARED_RESOURCE_RSS)
		return;

	/* free queues_array which is allocated in shared_resource_cfg*/
	total_res = resources_engine.cfg.total_resources[type];
	for (i = 0; i < total_res; i++) {
		res = get_shared_resource(type, i);
		if (unlikely(!res)) {
			DOCA_LOG_RATE_LIMIT_CRIT("shared resource of type %u and id %u is invalid at destruction",
				type, i);
			continue;
		}

		cfg = &res->res_cfg;
		if (unlikely(cfg->rss.queues_array)) {
			free(cfg->rss.queues_array);
			cfg->rss.queues_array = NULL;
		}
	}
}

void
engine_shared_resources_destroy(void)
{
	uint32_t nr_resources;
	uint32_t res_type;
	int rc;

	if (engine_bindable_stamp_verify(&global_bindable_obj)) {
		rc = engine_shared_resource_unbind(&global_bindable_obj);
		if (rc)
			DOCA_LOG_CRIT("failed destroying shared resources global bindable rc=%d",
				      rc);
	}

	for (res_type = 0; res_type < ENGINE_SHARED_RESOURCE_MAX; res_type++) {
		if (resources_engine.res[res_type]) {
			nr_resources = resources_engine.cfg.total_resources[res_type];
			if (resources_engine.ops_set)
				resources_engine.ops.resources_cleanup(res_type);

			DOCA_LOG_INFO("Cleanup %u shared resources of type %u completed",
				      nr_resources, res_type);
			shared_resource_pre_destroy(res_type);
			free(resources_engine.res[res_type]);
			resources_engine.res[res_type] = NULL;
		}
	}
	memset(&resources_engine, 0, sizeof(resources_engine));
}

bool
engine_shared_resource_is_initialized(enum engine_shared_resource_type type, uint32_t id)
{
	if (unlikely(type >= ENGINE_SHARED_RESOURCE_MAX))
		return false;

	return id < resources_engine.cfg.total_resources[type];
}

bool
engine_shared_resource_is_bound(enum engine_shared_resource_type type, uint32_t id,
				struct engine_bindable *bindable)
{
	struct shared_resource *res;

	res = get_shared_resource(type, id);
	if (unlikely(!res))
		return false;

	if (unlikely(!res->is_bound))
		return false;

	return engine_bindable_hierarchy_verify(res->bindable, bindable);
}

static int
shared_resource_rss_cfg_copy(struct shared_resource *res,
			     struct engine_shared_resource_cfg *cfg)
{
	uint16_t *queues_array = NULL;
	int nr_queues;

	nr_queues = cfg->rss.nr_queues;
	if (nr_queues) {
		queues_array = calloc(nr_queues, sizeof(uint16_t));
		if (unlikely(!queues_array)) {
			DOCA_LOG_RATE_LIMIT_ERR("failed allocation of internal rss queues array with size %u", nr_queues);
			return -EINVAL;
		}
	}

	/*
	 * queues_array is pointer and can be invalid in binding phase, allocate new pointer.
	 * nr_queues maybe diffrent, free previous pointer if not null.
	 */
	if (nr_queues)
		memcpy(queues_array, cfg->rss.queues_array, nr_queues * sizeof(uint16_t));

	if (unlikely(res->res_cfg.rss.queues_array))
		free(res->res_cfg.rss.queues_array);

	res->res_cfg.rss.types = cfg->rss.types;
	res->res_cfg.rss.mark = cfg->rss.mark;
	res->res_cfg.rss.nr_queues = cfg->rss.nr_queues;
	res->res_cfg.rss.queues_array = queues_array;

	return 0;
}

static int
shared_resource_cfg_copy(enum engine_shared_resource_type type,
			 struct shared_resource *res,
			 struct engine_shared_resource_cfg *cfg)
{
	int rc;

	switch (type) {
	case ENGINE_SHARED_RESOURCE_RSS:
		rc = shared_resource_rss_cfg_copy(res, cfg);
		if (rc)
			return rc;
		break;
	default:
		res->res_cfg = *cfg;
	}

	return 0;
}

int
engine_shared_resource_cfg_set(enum engine_shared_resource_type type, uint32_t id,
			       struct engine_shared_resource_cfg *cfg)
{
	struct shared_resource *res = get_shared_resource(type, id);
	int rc;

	if (unlikely(res == NULL || cfg == NULL))
		return -EINVAL;

	if (unlikely(res->is_bound)) {
		rc = resources_engine.ops.resource_modify(type, id, cfg);
		if (rc)
			return rc;
	} else {
		res->pending_cfg = true;
	}

	rc = shared_resource_cfg_copy(type, res, cfg);
	if (rc)
		return rc;

	return 0;
}

int
engine_shared_resource_cfg_get(enum engine_shared_resource_type type, uint32_t id,
			       struct engine_shared_resource_cfg *cfg)
{
	struct shared_resource *res = get_shared_resource(type, id);

	if (unlikely(!res))
		return -EINVAL;

	if (!res->is_bound && !res->pending_cfg) {
		DOCA_LOG_ERR("failed getting config type (%u) id (%u) - was not defined before.",
			     type, id);
		return -ENOENT;
	}

	*cfg = res->res_cfg;
	return 0;
}

static bool
shared_resources_verify_bound_state(enum engine_shared_resource_type type, uint32_t *ids_arr,
				    uint32_t ids_arr_len, bool expected_bound_state)
{
	struct shared_resource *res;
	uint32_t id_idx;

	for (id_idx = 0; id_idx < ids_arr_len; id_idx++) {
		res = get_shared_resource(type, ids_arr[id_idx]);
		if (unlikely(!res))
			return false;

		if (unlikely(res->is_bound != expected_bound_state))
			return false;
	}

	return true;
}

static void
bindable_resource_push(struct engine_bindable *bindable, struct shared_resource *res)
{
	struct shared_resource *head;

	head = (struct shared_resource *)bindable->priv_ctx;
	res->next = head;
	bindable->priv_ctx = res;
}

static struct shared_resource *
bindable_resource_pop(struct engine_bindable *bindable)
{
	struct shared_resource *res;

	if (unlikely(!bindable || !bindable->priv_ctx))
		return NULL;

	res = (struct shared_resource *)bindable->priv_ctx;
	bindable->priv_ctx = res->next;
	return res;
}

int
engine_shared_resource_bind_bulk(enum engine_shared_resource_type type, uint32_t *ids_arr,
				 uint32_t ids_arr_len, struct engine_bindable *bindable,
				 resource_prebind_cb prebind_cb)
{
	struct shared_resource *res;
	struct engine_shared_resource_cfg *res_cfg;
	uint32_t id_idx;
	uint32_t id;
	int rc;

	if (!bindable)
		bindable = &global_bindable_obj;

	engine_spinlock_lock(&resources_engine.big_lock);

	if (!engine_bindable_stamp_verify(bindable)) {
		engine_spinlock_unlock(&resources_engine.big_lock);
		return -EINVAL;
	}

	if (!shared_resources_verify_bound_state(type, ids_arr, ids_arr_len,
						 false /* bound_state */)) {
		engine_spinlock_unlock(&resources_engine.big_lock);
		return -EPERM;
	}

	for (id_idx = 0; id_idx < ids_arr_len; id_idx++) {

		id = ids_arr[id_idx];
		res = get_shared_resource(type, id);
		if (unlikely(res->next)) {
			rc = -EALREADY;
			goto rollback;
		}

		if (prebind_cb) {
			rc = prebind_cb(type, id, bindable);
			if (unlikely(rc))
				goto rollback;
		}

		res_cfg = (res->pending_cfg ? &res->res_cfg : NULL);
		rc = resources_engine.ops.resource_create(type, id, res_cfg);
		if (unlikely(rc))
			goto rollback;

		res->bindable = bindable;
		res->id = id;
		res->type = type;
		res->is_bound = true;
		res->pending_cfg = false;
		bindable_resource_push(bindable, res);
	}

	engine_spinlock_unlock(&resources_engine.big_lock);
	return 0;

rollback:
	shared_resource_unbind_safe(bindable);
	engine_spinlock_unlock(&resources_engine.big_lock);
	return rc;
}

int
engine_shared_resource_query_bulk(enum engine_shared_resource_type type,
				 uint32_t *ids_arr,
				 uint32_t ids_arr_len,
				 struct engine_shared_resource_query_result *query_result_array,
				 uint32_t query_result_arr_len)
{
	struct engine_shared_resource_query_result *query_result;
	uint32_t id_idx;
	uint32_t id;
	int rc = 0;

	if (query_result_arr_len < ids_arr_len) {
		DOCA_LOG_ERR("failed shared query - result len (%u) is less than ids len (%u)",
			     query_result_arr_len, ids_arr_len);
		return -EINVAL;
	}

	engine_spinlock_lock(&resources_engine.big_lock);

	if (!shared_resources_verify_bound_state(type, ids_arr, ids_arr_len,
						 true /* bound_state */)) {
		engine_spinlock_unlock(&resources_engine.big_lock);
		DOCA_LOG_ERR("failed shared query - some counters are invalid");
		return -EPERM;
	}

	for (id_idx = 0; id_idx < ids_arr_len; id_idx++) {

		id = ids_arr[id_idx];
		query_result = &query_result_array[id_idx];
		rc = resources_engine.ops.resource_query(type, id, query_result);
		if (unlikely(rc)) {
			engine_spinlock_unlock(&resources_engine.big_lock);
			return rc;
		}
	}

	engine_spinlock_unlock(&resources_engine.big_lock);
	return 0;
}

static void
notify_unbind(enum engine_shared_resource_type type, uint32_t id, struct engine_bindable *bindable)
{
	struct engine_shared_resources_cfg *cfg = &resources_engine.cfg;

	if (!cfg->unbind_cb)
		return;

	cfg->unbind_cb(type, id, bindable);
}

static int
shared_resource_unbind_safe(struct engine_bindable *bindable)
{
	struct shared_resource *res;


	if (!engine_bindable_stamp_verify(bindable))
		return -EINVAL;

	do {
		res = bindable_resource_pop(bindable);
		if (!res)
			break;

		resources_engine.ops.resource_destroy(res->type, res->id);

		res->next = NULL;
		res->is_bound = false;
		res->bindable = NULL;
		res->pending_cfg = false;
		notify_unbind(res->type, res->id, bindable);
	} while (res);

	return 0;
}


int
engine_shared_resource_unbind(struct engine_bindable *bindable)
{
	int rc;

	engine_spinlock_lock(&resources_engine.big_lock);
	rc = shared_resource_unbind_safe(bindable);
	engine_spinlock_unlock(&resources_engine.big_lock);
	return rc;
}
