/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef ENGINE_SHARED_RESOURCES_H_
#define ENGINE_SHARED_RESOURCES_H_

#include <stdbool.h>
#include <stdint.h>
#include "engine_pipe.h"

#ifdef __cplusplus
extern "C" {
#endif

struct engine_bindable;

/**
 * @brief Shared resource type
 */
enum engine_shared_resource_type {
	ENGINE_SHARED_RESOURCE_METER,
	/**< shared meter resource */
	ENGINE_SHARED_RESOURCE_COUNTER,
	/**< shared counter resource */
	ENGINE_SHARED_RESOURCE_INTERNAL_META,
	/**< shared meta data resource */
	ENGINE_SHARED_RESOURCE_RSS,
	/**< shared rss resource */
	ENGINE_SHARED_RESOURCE_CRYPTO,
	/**< Shared NISP action type */
	ENGINE_SHARED_RESOURCE_MAX,
	/**< Max supported shared resources */
};

#define ENGINE_CRYPTO_KEY_LEN_MAX 32

#define ENGINE_NISP_DWORD 10
#define ENGINE_CRYPTO_REFORMAT_LEN_MAX	\
	(6 * 2 +			\
	 sizeof(uint16_t) +		\
	 sizeof(uint16_t) * 2 * 2 +	\
	 sizeof(uint32_t) * 15 +	\
	 sizeof(uint32_t) * 2 +		\
	 ENGINE_NISP_DWORD * sizeof(uint32_t))

enum engine_crypto_protocol_type {
	ENGINE_CRYPTO_PROTOCOL_NONE = 0,
	/**< crypto action protocol not defined */
	ENGINE_CRYPTO_PROTOCOL_NISP,
	/**< crypto action for NISP protocol */
	ENGINE_CRYPTO_PROTOCOL_ESP,
	/**< crypto action for IPsec ESP protocol */
};

enum engine_crypto_reformat_type {
	ENGINE_CRYPTO_REFORMAT_NONE = 0,
	/**< no encap and decap operation performed by action */
	ENGINE_CRYPTO_REFORMAT_NISP_TUNNEL_ENCAP,
	/**< remove L2 header and prepend witn NISP tunnel */
	ENGINE_CRYPTO_REFORMAT_NISP_TUNNEL_DECAP,
	/**< remove NISP tunnel header and prepend with L2 */
	ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV4_ENCAP,
	/**< encap with ESP IPv4 transport header */
	ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV6_ENCAP,
	/**< encap with ESP IPv6 transport header */
	ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV4_UDP_ENCAP,
	/**< encap with ESP IPv4 UDP transport header */
	ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_IPV6_UDP_ENCAP,
	/**< encap with ESP IPv6 UDP transport header */
	ENGINE_CRYPTO_REFORMAT_ESP_TUNNEL_ENCAP,
	/**< encap with ESP tunnel header */
	ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_DECAP,
	/**< decap ESP IPv4/6 transport header */
	ENGINE_CRYPTO_REFORMAT_ESP_TRANSPORT_UDP_DECAP,
	/**< decap ESP IPv4/6 UDP transport header */
	ENGINE_CRYPTO_REFORMAT_ESP_TUNNEL_DECAP,
	/**< decap ESP tunnel header */
	ENGINE_CRYPTO_REFORMAT_UNKNOWN,
};

enum engine_crypto_operation_type {
	ENGINE_CRYPTO_OPERATION_NONE = 0,
	/**< no crypto operation performed by action */
	ENGINE_CRYPTO_OPERATION_ENCRYPT,
	/**< do packet encrypt */
	ENGINE_CRYPTO_OPERATION_DECRYPT,
	/**< do packet decrypt */
	ENGINE_CRYPTO_OPERATION_UNKNOWN,
};

/**
 * @brief Shared resource configuration
 */
struct engine_shared_resource_cfg {
	union {
		struct {
			uint32_t port_id;
			/**< Port id */
		} counter;
		struct {
			uint32_t port_id;
			/**< Port id */
			uint64_t cir;
			/**< Committed information rate */
			uint64_t cbs;
			/**< Committed burst size */
		} meter;
		struct  {
			uint32_t port_id;
			/**< Port id */
			enum engine_crypto_protocol_type proto_type;
			/**< action protocol/subtype */
			enum engine_crypto_reformat_type reformat_type;
			/**< packet reformat action */
			enum engine_crypto_operation_type crypto_type;
			/**< crypto action */
			uint16_t reformat_data_sz;
			/**< reformat header length in bytes */
			uint8_t reformat_data[ENGINE_CRYPTO_REFORMAT_LEN_MAX];
			/**< reformat header buffer */
			union {
				struct {
					uint16_t key_sz;
					/**< key size in bytes */
					uint8_t key[ENGINE_CRYPTO_KEY_LEN_MAX];
					/**< key buffer */
				};
				void* security_ctx;
			};
			struct engine_pipe_fwd fwd;
			/**< forward action */
		} crypto;
		struct {
			uint32_t bit_width; /* Requested bit width */
		} meta;
		struct {
			uint32_t port_id;
			/**< Port id */
			uint64_t types;
			/**< rss offload types */
			uint16_t *queues_array;
			/**< rss queues array */
			int nr_queues;
			/**< number of queues */
			uint32_t mark;
			/**< markid of each queues */
		} rss;
	};
};

typedef void (*shared_resource_unbind_cb)(enum engine_shared_resource_type type,
					  uint32_t shared_resource_id,
					  struct engine_bindable *bindable);

/**
 * @brief All shared resources configuration
 */
struct engine_shared_resources_cfg {
	uint32_t total_resources[ENGINE_SHARED_RESOURCE_MAX];
	/**< Total shared resources per type */
	shared_resource_unbind_cb unbind_cb;
	/**< Callback to be called on unbinding of a shared resource */
};

/**
 * @brief Shared resource query
 */
struct engine_shared_resource_query_result {
	union {
		struct {
			uint64_t total_bytes;
			uint64_t total_pkts;
		} counter;
	};
};

/**
 * @brief Shared resource ops
 */
struct engine_shared_resources_ops {
	int (*resources_init)(enum engine_shared_resource_type type, uint32_t nr_resources);
	/**< shared resources initialization functionality */

	int (*resource_create)(enum engine_shared_resource_type type, uint32_t id,
			       struct engine_shared_resource_cfg *res_cfg);
	/**< shared resource creation functionality */

	int (*resource_modify)(enum engine_shared_resource_type type, uint32_t id,
			       struct engine_shared_resource_cfg *res_cfg);
	/**< shared resource modification functionality */

	int (*resource_query)(enum engine_shared_resource_type type, uint32_t id,
		struct engine_shared_resource_query_result *res_query_result);
	/**< shared resource query functionality */

	int (*resource_destroy)(enum engine_shared_resource_type type, uint32_t id);
	/**< destroy shared resource functionality */

	int (*resources_cleanup)(enum engine_shared_resource_type type);
	/**< shared resources cleanup functionality */
};

/**
 * @brief Initialize shared resources
 *
 * Shared resources initialization according to configuration.
 *
 * @param res_cfg
 * Pointer to the configuration struct.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_shared_resources_init(struct engine_shared_resources_cfg *res_cfg);

/**
 * @brief Destroy shared resources
 *
 * Destroy all shared resources.
 *
 */
void
engine_shared_resources_destroy(void);

/**
 * @brief Verify if a shared resource is initialized.
 *
 * @param type
 * Shared resource type.
 * @param id
 * Shared resource id that belongs to that type.
 * @return
 * true when initialized, false otherwise.
 */
bool
engine_shared_resource_is_initialized(enum engine_shared_resource_type type, uint32_t id);

/**
 * @brief Verify if a shared resource was binded.
 *
 * @param type
 * Shared resource type.
 * @param id
 * Shared resource id that belongs to that type.
 * @param bindable
 * Pointer to the bindable object the resources is expected to be bound to.
 * @return
 * true when bound, false otherwise.
 */
bool
engine_shared_resource_is_bound(enum engine_shared_resource_type type, uint32_t id,
				struct engine_bindable *bindable);

/**
 * @brief Configure a shared resource
 *
 * @param type
 * Shared resource type.
 * @param id
 * Shared resource id that belongs to that type.
 * @param cfg
 * Pointer to the shared resource configuration struct.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_shared_resource_cfg_set(enum engine_shared_resource_type type, uint32_t id,
			       struct engine_shared_resource_cfg *cfg);

/**
 * @brief Get a shared resource
 *
 * @param type
 * Shared resource type.
 * @param id
 * Shared resource id that belongs to that type.
 * @param cfg
 * Pointer to the shared resource configuration struct to be filled in.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_shared_resource_cfg_get(enum engine_shared_resource_type type, uint32_t id,
			       struct engine_shared_resource_cfg *cfg);

/**
 * @brief Callback to override system specific parameters before binding.
 *
 * This callback is called just before binding an object to override
 * the resource configuration with system parameters (e.g. port_id).
 *
 * @param type
 * Shared resource type.
 * @param res_arr
 * a resource ID.
 * @param bindable
 * Pointer to the bindable object the resources will be bound to.
 * @return
 * 0 on success, negative on failure.
 */
typedef int
(*resource_prebind_cb)(enum engine_shared_resource_type res_type,
		       uint32_t res_id, void *bindable_obj);

/**
 * @brief Binds a bulk of shared resources to a bindable object.
 *
 * Binds a bulk of shared resources of the same type with arbitrary IDs
 * to a bindable object.
 *
 * In the end of this function, either all requested resources are bound and ready to work or none.
 *
 * @param type
 * Shared resource type.
 * @param ids_arr
 * Pointer to an array of IDs.
 * @param ids_arr_len
 * Length of the array of IDs.
 * @param bindable
 * Pointer to the bindable object the resources will be bound to.
 * @param resource_prebind_cb
 * Callback to update system specific parameters before binding.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_shared_resource_bind_bulk(enum engine_shared_resource_type type, uint32_t *ids_arr,
				 uint32_t ids_arr_len, struct engine_bindable *bindable,
				 resource_prebind_cb cb);

/**
 * @brief Unbinds all shared resources from a bindable object.
 *
 * @param bindable
 * Pointer to the bindable object the resources will be bound to.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_shared_resource_unbind(struct engine_bindable *bindable);

/**
 * @brief Gets the global bindable resource
 *
 * @return
 * 0 on success, negative on failure.
 */
struct engine_bindable *
engine_shared_resource_get_global_bindable(void);

/**
 * @brief Register shared resources operations
 *
 * @param ops
 * Pointer to an ops struct.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_shared_resource_register_ops(struct engine_shared_resources_ops *ops);

/**
 * @brief Query a bulk of shared resources.
 *
 * Query a bulk of shared resources of the same type with arbitrary IDs.
 *
 * @param type
 * Shared resource type.
 * @param ids_arr
 * Pointer to an array of IDs to be querried.
 * @param query_resultst_array
 * Pointer to an array of ids query results.
 * @param ids_arr_len
 * Length of the array of IDs.
 * @return
 * 0 on success, negative on failure.
 */
int
engine_shared_resource_query_bulk(enum engine_shared_resource_type type,
				  uint32_t *ids_arr,
				  uint32_t ids_arr_len,
				  struct engine_shared_resource_query_result *query_result_arr,
				  uint32_t query_result_arr_len);
#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* ENGINE_SHARED_RESOURCES_H_ */
