/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "engine_supported_types.h"
#include "engine_to_string.h"
#include "private/common_defines.h"

#include "doca_log.h"

#define UNKNOWN_STRING "unknown"
#define STRING_SEPARATOR "."
#define MAX_STR_VALUE_SIZE (64)
#define MAX_STR_SEPARATORS (4)
#define MAX_STR_VALUES (MAX_STR_SEPARATORS + 1)
#define MAX_STR_TO_OPCODE (MAX_STR_VALUE_SIZE * MAX_STR_VALUES)

enum {
	PARSE_LEVEL_0 = 0,
	PARSE_LEVEL_1 = 1,
	PARSE_LEVEL_2 = 2,
	PARSE_LEVEL_3 = 3,
	PARSE_LEVEL_4 = 4,
};

DOCA_LOG_REGISTER(engine_to_string);

struct string_value {
	bool used;
	char str[MAX_STR_VALUE_SIZE];
};

struct string_array {
	int len;
	struct string_value str[0];
};

static struct string_array *opcode_group;
static struct string_array *user_defined_type;
static struct string_array *packet_focus;
static struct string_array *packet_proto;

/* protocol fields */
static struct string_array *proto_eth_fields;
static struct string_array *proto_eth_vlan_fields;
static struct string_array *proto_ipv4_fields;
static struct string_array *proto_ipv6_fields;
static struct string_array *proto_udp_fields;
static struct string_array *proto_tcp_fields;
static struct string_array *proto_gre_fields;
static struct string_array *proto_vxlan_fields;
static struct string_array *proto_gtp_fields;
static struct string_array *proto_nisp_fields;
static struct string_array *proto_audp_fields;
static struct string_array *proto_esp_fields;

static struct string_array *
init_string_array(int len)
{
	struct string_array *str_arr;
	uint16_t memlen;

	if (len <= 0)
		return NULL;

	memlen = sizeof(struct string_array) + len * sizeof(struct string_value);

	str_arr = calloc(memlen, 1);
	if (unlikely(!str_arr)) {
		DOCA_LOG_ERR("String array allocation of size %u failed", memlen);
		return NULL;
	}

	str_arr->len = len;
	return str_arr;
}

static int
string_array_set(struct string_array *str_arr, int idx, const char *value)
{
	if (unlikely(!str_arr)) {
		DOCA_LOG_ERR("failed setting value - string array is null");
		return -EINVAL;
	}

	if (unlikely(!value)) {
		DOCA_LOG_ERR("failed setting value - value is null");
		return -EINVAL;
	}

	if (unlikely(idx >= str_arr->len)) {
		DOCA_LOG_ERR("failed setting value - index %d should be lower then %d", idx,
			     str_arr->len);
		return -EINVAL;
	}

	strncpy(str_arr->str[idx].str, value, MAX_STR_VALUE_SIZE - 1);
	str_arr->str[idx].used = true;
	return 0;
}

static const char *
string_array_get(struct string_array *str_arr, int idx)
{
	if (unlikely(!str_arr))
		return UNKNOWN_STRING;

	if (unlikely(idx >= str_arr->len))
		return UNKNOWN_STRING;

	return str_arr->str[idx].str;
}

static int
string_array_find(struct string_array *str_arr, const char *token)
{
	int idx;

	if (unlikely(!str_arr))
		return -EINVAL;

	for (idx = 0; idx < str_arr->len; idx++) {
		if (!str_arr->str[idx].used)
			break;

		if (strcmp(str_arr->str[idx].str, token) == 0)
			return idx;
	}

	return -ENOENT;
}

static bool
string_array_verify_all_used(const struct string_array *str_arr)
{
	bool all_used = true;
	int i;

	if (unlikely(!str_arr)) {
		DOCA_LOG_ERR("failed verifying all used - string array is null");
		return false;
	}

	for (i = 0; all_used && i < str_arr->len; i++)
		all_used &= str_arr->str[i].used;

	return all_used;
}

int
engine_to_string_init()
{
	user_defined_type = init_string_array(DF_SUP_USER_DEFINED_TYPE_MAX);
	string_array_set(user_defined_type, DF_SUP_USER_DEFINED_TYPE_MATCH, "match");
	if (!string_array_verify_all_used(user_defined_type)) {
		DOCA_LOG_ERR("failed init of engine to string - user defined missing type");
		return -EPERM;
	}

	opcode_group = init_string_array(ENGINE_FIELD_OPCODE_GROUP_MAX);
	string_array_set(opcode_group, ENGINE_FIELD_OPCODE_GROUP_PACKET, "packet");
	string_array_set(opcode_group, ENGINE_FIELD_OPCODE_GROUP_FLOW, "flow");
	if (!string_array_verify_all_used(opcode_group)) {
		DOCA_LOG_ERR("failed init of engine to string - opcode group missing type");
		return -EPERM;
	}

	packet_focus = init_string_array(DF_SUP_FOCUS_MAX);
	string_array_set(packet_focus, DF_SUP_FOCUS_OUTER, "outer");
	string_array_set(packet_focus, DF_SUP_FOCUS_TUNNEL, "tunnel");
	string_array_set(packet_focus, DF_SUP_FOCUS_INNER, "inner");
	if (!string_array_verify_all_used(packet_focus)) {
		DOCA_LOG_ERR("failed init of engine to string - packet focus missing type");
		return -EPERM;
	}

	packet_proto = init_string_array(DF_SUP_PROTO_MAX);
	string_array_set(packet_proto, DF_SUP_PROTO_ETH, "eth");
	string_array_set(packet_proto, DF_SUP_PROTO_ETH_VLAN, "eth_vlan");
	string_array_set(packet_proto, DF_SUP_PROTO_IPV4, "ipv4");
	string_array_set(packet_proto, DF_SUP_PROTO_IPV6, "ipv6");
	string_array_set(packet_proto, DF_SUP_PROTO_UDP, "udp");
	string_array_set(packet_proto, DF_SUP_PROTO_TCP, "tcp");
	string_array_set(packet_proto, DF_SUP_PROTO_GRE, "gre");
	string_array_set(packet_proto, DF_SUP_PROTO_VXLAN, "vxlan");
	string_array_set(packet_proto, DF_SUP_PROTO_GTP, "gtp");
	string_array_set(packet_proto, DF_SUP_PROTO_AUDP, "audp");
	string_array_set(packet_proto, DF_SUP_PROTO_NISP, "nisp");
	string_array_set(packet_proto, DF_SUP_PROTO_ESP, "esp");
	if (!string_array_verify_all_used(packet_proto)) {
		DOCA_LOG_ERR("failed init of engine to string - packet proto missing type");
		return -EPERM;
	}

	proto_eth_fields = init_string_array(DF_SUP_PROTO_FIELD_ETH_MAX);
	string_array_set(proto_eth_fields, DF_SUP_PROTO_FIELD_ETH_DST_MAC, "dst_mac");
	string_array_set(proto_eth_fields, DF_SUP_PROTO_FIELD_ETH_SRC_MAC, "src_mac");
	string_array_set(proto_eth_fields, DF_SUP_PROTO_FIELD_ETH_TYPE, "type");
	if (!string_array_verify_all_used(proto_eth_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_eth_fields missing type");
		return -EPERM;
	}

	proto_eth_vlan_fields = init_string_array(DF_SUP_PROTO_FIELD_ETH_VLAN_MAX);
	string_array_set(proto_eth_vlan_fields, DF_SUP_PROTO_FIELD_ETH_VLAN_TCI, "tci");
	if (!string_array_verify_all_used(proto_eth_vlan_fields)) {
		DOCA_LOG_ERR(
			"failed init of engine to string - proto_eth_vlan_fields missing type");
		return -EPERM;
	}

	proto_ipv4_fields = init_string_array(DF_SUP_PROTO_FIELD_IPV4_MAX);
	string_array_set(proto_ipv4_fields, DF_SUP_PROTO_FIELD_IPV4_SRC_IP, "src_ip");
	string_array_set(proto_ipv4_fields, DF_SUP_PROTO_FIELD_IPV4_DST_IP, "dst_ip");
	string_array_set(proto_ipv4_fields, DF_SUP_PROTO_FIELD_IPV4_NEXT_PROTO, "next_proto");
	if (!string_array_verify_all_used(proto_ipv4_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_ipv4_fields missing type");
		return -EPERM;
	}

	proto_ipv6_fields = init_string_array(DF_SUP_PROTO_FIELD_IPV6_MAX);
	string_array_set(proto_ipv6_fields, DF_SUP_PROTO_FIELD_IPV6_SRC_IP, "src_ip");
	string_array_set(proto_ipv6_fields, DF_SUP_PROTO_FIELD_IPV6_DST_IP, "dst_ip");
	string_array_set(proto_ipv6_fields, DF_SUP_PROTO_FIELD_IPV6_NEXT_PROTO, "next_proto");
	if (!string_array_verify_all_used(proto_ipv6_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_ipv6_fields missing type");
		return -EPERM;
	}

	proto_udp_fields = init_string_array(DF_SUP_PROTO_FIELD_UDP_MAX);
	string_array_set(proto_udp_fields, DF_SUP_PROTO_FIELD_UDP_SRC_PORT, "src_port");
	string_array_set(proto_udp_fields, DF_SUP_PROTO_FIELD_UDP_DST_PORT, "dst_port");
	if (!string_array_verify_all_used(proto_udp_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_udp_fields missing type");
		return -EPERM;
	}

	proto_tcp_fields = init_string_array(DF_SUP_PROTO_FIELD_TCP_MAX);
	string_array_set(proto_tcp_fields, DF_SUP_PROTO_FIELD_TCP_SRC_PORT, "src_port");
	string_array_set(proto_tcp_fields, DF_SUP_PROTO_FIELD_TCP_DST_PORT, "dst_port");
	string_array_set(proto_tcp_fields, DF_SUP_PROTO_FIELD_TCP_FLAGS, "flags");
	if (!string_array_verify_all_used(proto_tcp_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_tcp_fields missing type");
		return -EPERM;
	}

	proto_vxlan_fields = init_string_array(DF_SUP_PROTO_FIELD_VXLAN_MAX);
	string_array_set(proto_vxlan_fields, DF_SUP_PROTO_FIELD_VXLAN_TUN_ID, "vni");
	if (!string_array_verify_all_used(proto_vxlan_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_vxlan_fields missing type");
		return -EPERM;
	}

	proto_gre_fields = init_string_array(DF_SUP_PROTO_FIELD_GRE_MAX);
	string_array_set(proto_gre_fields, DF_SUP_PROTO_FIELD_GRE_KEY, "key");
	string_array_set(proto_gre_fields, DF_SUP_PROTO_FIELD_GRE_PROTOCOL, "protocol");
	if (!string_array_verify_all_used(proto_gre_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_gre_fields missing type");
		return -EPERM;
	}

	proto_gtp_fields = init_string_array(DF_SUP_PROTO_FIELD_GTP_MAX);
	string_array_set(proto_gtp_fields, DF_SUP_PROTO_FIELD_GTP_TEID, "teid");
	if (!string_array_verify_all_used(proto_gtp_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_gtp_fields missing type");
		return -EPERM;
	}

	proto_nisp_fields = init_string_array(DF_SUP_PROTO_FIELD_NISP_MAX);
	string_array_set(proto_nisp_fields, DF_SUP_PROTO_FIELD_NISP_HEADER, "hdr");
	if (!string_array_verify_all_used(proto_nisp_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_nisp_fields missing type");
		return -EPERM;
	}

	proto_audp_fields = init_string_array(DF_SUP_PROTO_FIELD_AUDP_MAX);
	string_array_set(proto_audp_fields, DF_SUP_PROTO_FIELD_AUDP_HEADER, "hdr");
	if (!string_array_verify_all_used(proto_audp_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_audp_fields missing type");
		return -EPERM;
	}

	proto_esp_fields = init_string_array(DF_SUP_PROTO_FIELD_ESP_MAX);
	string_array_set(proto_esp_fields, DF_SUP_PROTO_FIELD_ESP_SPI, "spi");
	string_array_set(proto_esp_fields, DF_SUP_PROTO_FIELD_ESP_ESN, "sn");
	if (!string_array_verify_all_used(proto_esp_fields)) {
		DOCA_LOG_ERR("failed init of engine to string - proto_esp_fields missing type");
		return -EPERM;
	}

	return 0;
}

void
engine_to_string_destroy()
{
	free(proto_esp_fields);
	free(proto_audp_fields);
	free(proto_nisp_fields);
	free(proto_gtp_fields);
	free(proto_vxlan_fields);
	free(proto_gre_fields);
	free(proto_tcp_fields);
	free(proto_udp_fields);
	free(proto_ipv6_fields);
	free(proto_ipv4_fields);
	free(proto_eth_vlan_fields);
	free(proto_eth_fields);

	free(opcode_group);
	free(user_defined_type);
	free(packet_focus);
	free(packet_proto);
}

static const char *
to_string_group(enum engine_field_opcode_group group)
{
	return string_array_get(opcode_group, group);
}

static const char *
to_string_user_defined_type(enum engine_supported_user_defined_type type)
{
	return string_array_get(user_defined_type, type);
}

static const char *
to_string_focus(enum engine_supported_focus focus)
{
	return string_array_get(packet_focus, focus);
}

static const char *
to_string_proto(enum engine_supported_proto proto)
{
	return string_array_get(packet_proto, proto);
}

static struct string_array *
get_proto_field_string_array(enum engine_supported_proto proto)
{
	switch (proto) {
	case DF_SUP_PROTO_ETH:
		return proto_eth_fields;
	case DF_SUP_PROTO_ETH_VLAN:
		return proto_eth_vlan_fields;
	case DF_SUP_PROTO_IPV4:
		return proto_ipv4_fields;
	case DF_SUP_PROTO_IPV6:
		return proto_ipv6_fields;
	case DF_SUP_PROTO_UDP:
		return proto_udp_fields;
	case DF_SUP_PROTO_TCP:
		return proto_tcp_fields;
	case DF_SUP_PROTO_GRE:
		return proto_gre_fields;
	case DF_SUP_PROTO_VXLAN:
		return proto_vxlan_fields;
	case DF_SUP_PROTO_GTP:
		return proto_gtp_fields;
	case DF_SUP_PROTO_NISP:
		return proto_nisp_fields;
	case DF_SUP_PROTO_AUDP:
		return proto_audp_fields;
	case DF_SUP_PROTO_ESP:
		return proto_esp_fields;
	default:
		return NULL;
	}
}


static const char *
to_string_proto_field(enum engine_supported_proto proto, uint8_t field)
{
	struct string_array *proto_field;

	proto_field = get_proto_field_string_array(proto);
	if (unlikely(!proto_field))
		return UNKNOWN_STRING;

	return string_array_get(proto_field, field);
}

int
engine_to_string_opcode(struct engine_field_opcode *opcode, char *buff, int len)
{
	int wr = 0;

	if (!opcode)
		return 0;

	switch (opcode->pkt.type) {
	case DF_SUP_USER_DEFINED_TYPE_MATCH:
		wr += snprintf(buff + wr, len - wr, "%s",
			       to_string_user_defined_type(opcode->pkt.type));

		switch (opcode->pkt.group) {
		case ENGINE_FIELD_OPCODE_GROUP_PACKET:
			wr += snprintf(buff + wr, len - wr, STRING_SEPARATOR "%s",
				       to_string_group(opcode->pkt.group));
			wr += snprintf(buff + wr, len - wr, STRING_SEPARATOR "%s",
				       to_string_focus(opcode->pkt.focus));
			wr += snprintf(buff + wr, len - wr, STRING_SEPARATOR "%s",
				       to_string_proto(opcode->pkt.proto));
			wr += snprintf(buff + wr, len - wr, STRING_SEPARATOR "%s",
				       to_string_proto_field(opcode->pkt.proto, opcode->pkt.field));
			break;
		case ENGINE_FIELD_OPCODE_GROUP_FLOW:
			wr += snprintf(buff + wr, len - wr, STRING_SEPARATOR "%s",
				       to_string_group(opcode->flow.group));
			wr += snprintf(buff + wr, len - wr, STRING_SEPARATOR "%u",
				       opcode->flow.code);
			break;
		default:
			wr += snprintf(buff + wr, len - wr, "0x%08lx", opcode->opcode);
		}

		break;
	default:
		wr += snprintf(buff + wr, len - wr, "0x%08lx", opcode->opcode);
	}

	return wr;
}

static int
string_to_user_defined_type(const char *token, enum engine_supported_user_defined_type *type)
{
	int rc;

	rc = string_array_find(user_defined_type, token);
	if (unlikely(rc < 0))
		return rc;

	*type = (enum engine_supported_user_defined_type)rc;
	return 0;
}

static int
string_to_group(const char *token, enum engine_field_opcode_group *group)
{
	int rc;

	rc = string_array_find(opcode_group, token);
	if (unlikely(rc < 0))
		return rc;

	*group = (enum engine_field_opcode_group)rc;
	return 0;
}

static int
string_to_focus(const char *token, enum engine_supported_focus *focus)
{
	int rc;

	rc = string_array_find(packet_focus, token);
	if (unlikely(rc < 0))
		return rc;

	*focus = (enum engine_supported_focus)rc;
	return 0;
}

static int
string_to_proto(const char *token, enum engine_supported_proto *proto)
{
	int rc;

	rc = string_array_find(packet_proto, token);
	if (unlikely(rc < 0))
		return rc;

	*proto = (enum engine_supported_proto)rc;
	return 0;
}

static int
string_to_proto_field(const char *token, enum engine_supported_proto proto, uint8_t *field)
{
	struct string_array *proto_field;
	int rc;

	proto_field = get_proto_field_string_array(proto);
	if (unlikely(!proto_field))
		return -ENOTSUP;

	rc = string_array_find(proto_field, token);
	if (unlikely(rc < 0))
		return rc;

	*field = (uint8_t)rc;
	return 0;
}

int
engine_string_to_opcode(const char *buff, int len, struct engine_field_opcode *opcode)
{
	enum engine_supported_user_defined_type type = DF_SUP_USER_DEFINED_TYPE_MAX;
	enum engine_field_opcode_group group = ENGINE_FIELD_OPCODE_GROUP_MAX;
	enum engine_supported_focus focus = DF_SUP_FOCUS_MAX;
	enum engine_supported_proto proto = DF_SUP_PROTO_MAX;
	uint8_t field = DF_SUP_PROTO_FIELD_MAX;
	char tokenizer[MAX_STR_TO_OPCODE] = {0};
	int parse_level = 0;
	int rc;
	char *token;

	if (unlikely(!buff)) {
		DOCA_LOG_ERR("failed parsing opcode string - buff pointer is null");
		return -EINVAL;
	}

	if (unlikely(!opcode)) {
		DOCA_LOG_ERR("failed parsing opcode string - opcode pointer is null");
		return -EINVAL;
	}

	if (unlikely(len < 0)) {
		DOCA_LOG_ERR("failed parsing opcode string - invalid len %d", len);
		return -EINVAL;
	}

	if (unlikely(len >= MAX_STR_TO_OPCODE)) {
		DOCA_LOG_ERR("failed parsing opcode string - len %d exceeds max %u", len,
			     MAX_STR_TO_OPCODE);
		return -ENOMEM;
	}

	strncpy(tokenizer, buff, len);
	tokenizer[len] = '\0';
	token = strtok(tokenizer, STRING_SEPARATOR);
	while (token != NULL && parse_level < MAX_STR_VALUES) {
		switch (parse_level) {
		case PARSE_LEVEL_0:
			rc = string_to_user_defined_type(token, &type);
			break;
		case PARSE_LEVEL_1:
			rc = string_to_group(token, &group);
			break;
		case PARSE_LEVEL_2:
			rc = string_to_focus(token, &focus);
			break;
		case PARSE_LEVEL_3:
			rc = string_to_proto(token, &proto);
			break;
		case PARSE_LEVEL_4:
			rc = string_to_proto_field(token, proto, &field);
			break;
		default:
			return -ENOTSUP;
		}

		if (unlikely(rc < 0)) {
			DOCA_LOG_ERR(
				"failed parsing opcode string - token %s, parse level %u, rc=%d",
				token, parse_level, rc);
			return -ENOTSUP;
		}

		DOCA_DLOG_DBG("opcode parsed token %s successfully", token);
		token = strtok(NULL, STRING_SEPARATOR);
		parse_level++;
	}

	if (parse_level != MAX_STR_VALUES)
		return -ENOTSUP;

	rc = engine_field_opcode_build_pkt(opcode, type, focus, proto, field);
	if (unlikely(rc < 0)) {
		DOCA_LOG_ERR("failed parsing opcode string - build opcode rc=%d", rc);
		return rc;
	}

	return 0;
}
