/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <doca_log.h>
#include "doca_flow_priv.h"
#include "flow_age.h"

DOCA_LOG_REGISTER(flow_age);

#define CURR_SEC (rte_rdtsc() / rte_get_timer_hz())
static uint64_t ms_hz;
struct flow_age_cfg age_cfg;

#define UPDATE_FLOW_OBJ(age_ctx, obj) \
	age_cfg.update_age_item(age_ctx, obj)
#define QUERY_FLOW_COUNTER(age_ctx, pkts) \
	age_cfg.query_pkts(age_ctx, pkts)

int
flow_age_init(const struct flow_age_cfg *cfg)
{
	age_cfg = *cfg;
	ms_hz = rte_get_timer_hz() / 1000000;

	return 0;
}

void
flow_age_pipe_destroy(struct doca_flow_pipe *pipe)
{
	uint32_t qidx = 0;
	struct flow_age_queue *age_queue;

	if (pipe == NULL || pipe->age_queues == NULL)
		return;
	for (qidx = 0; qidx < pipe->nb_queues; qidx++) {
		age_queue = pipe->age_queues[qidx];
		if (age_queue == NULL)
			continue;
		free(age_queue);
	}
	free(pipe->age_queues);
}

int
flow_age_pipe_init(struct doca_flow_pipe *pipe, uint32_t nb_flows)
{
	uint32_t qidx, len = 0;
	struct flow_age_queue *age_queue;

	len = sizeof(struct flow_age_queue) + nb_flows * sizeof(struct flow_age_item);
	DOCA_DLOG_DBG("Allocating %u age flows for pipe %s", nb_flows, pipe->name);
	for (qidx = 0; qidx < pipe->nb_queues; qidx++) {
		age_queue = calloc(1, len);
		if (age_queue == NULL) {
			DOCA_LOG_RATE_LIMIT_ERR("alloc age list len:%d on queue:%d error.",
				nb_flows, qidx);
			for ( ; qidx < pipe->nb_queues; qidx++)
				pipe->age_queues[qidx] = NULL;
			goto release;
		}
		age_queue->last_idx = -1;
		age_queue->curr_idx = 0;/*read from 0 when array is not null*/
		age_queue->array_len = nb_flows;
		pipe->age_queues[qidx] = age_queue;
	}
	return 0;
release:
	flow_age_pipe_destroy(pipe);
	return -ENOMEM;
}

static void
__flow_age_index_rm(struct flow_age_queue *age_queue, uint32_t index)
{
	struct flow_age_item *item = &age_queue->item_array[index];

	/* remove the flow index from age list*/
	UPDATE_FLOW_OBJ(item->age_ctx, NULL);
	/* if not the last one, put the last item to index */
	if (index != age_queue->last_idx) {
		*item = age_queue->item_array[age_queue->last_idx];
		UPDATE_FLOW_OBJ(item->age_ctx, item);
	}
	age_queue->last_idx--;
}

int
flow_age_item_release(struct doca_flow_pipe_entry *entry)
{
	uint32_t idx;
	uint16_t queue = entry->pipe_queue;
	struct doca_flow_pipe *pipe = entry->pipe;
	struct flow_age_item *item = entry->age_item;
	struct flow_age_queue *age_queue;

	if (queue > pipe->nb_queues || item == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("invalid queue %u or null age item.", queue);
		return -1;
	}
	age_queue = pipe->age_queues[queue];
	idx = item - age_queue->item_array;
	if (idx > age_queue->last_idx) {
		DOCA_LOG_RATE_LIMIT_ERR("release age index %u error", idx);
		return -1;
	}
	__flow_age_index_rm(age_queue, idx);
	return 0;
}

struct flow_age_item *
flow_age_item_alloc(uint16_t queue, struct doca_flow_pipe_entry *entry,
		uint16_t time_out, uint64_t user_data)
{
	struct flow_age_queue *age_queue;
	struct flow_age_item *item;
	struct doca_flow_pipe *pipe = (struct doca_flow_pipe *) entry->pipe;

	age_queue = pipe->age_queues[queue];
	if (age_queue->last_idx + 1 >= age_queue->array_len) {
		DOCA_DLOG_ERR("Error, reach max aged len:%d", age_queue->array_len);
		return NULL;
	}
	/*new item always add in the last index*/
	item = &age_queue->item_array[++age_queue->last_idx];
	item->age_ctx = entry;
	item->user_data = user_data;
	item->aging_time_in_sec = time_out;
	item->expr_time_in_sec = CURR_SEC + item->aging_time_in_sec;
	return item;
}

int
flow_age_check(struct doca_flow_port *port, uint16_t queue,
		uint64_t quota, struct doca_flow_aged_query *entries, int len)
{
	uint64_t curr_tsc = rte_rdtsc();
	uint64_t end_tsc = curr_tsc + quota * ms_hz;
	uint64_t curr_pkts = 0;
	struct flow_age_queue *age_queue;
	struct flow_age_item *item;
	uint32_t curr_sec, nb_aged_pipe, nb_aged = 0;
	int checked_pipes = 0, nb_pipes = 0;
	int ret;
	struct doca_flow_pipe *pipe;

	LIST_FOREACH(pipe, &port->pipe_list, next) {
		if (!pipe->is_aging)
			continue;
		age_queue = pipe->age_queues[queue];
		nb_aged_pipe = 0;
		nb_pipes++;
		while (curr_tsc < end_tsc) {
			curr_tsc = rte_rdtsc();
			if (curr_tsc >= end_tsc) {
				checked_pipes--;
				goto out;
			}
			/* age array is empty*/
			if (age_queue->last_idx == -1) {
				checked_pipes++;
				break;
			}
			if (age_queue->curr_idx > age_queue->last_idx) {
				checked_pipes++;
				age_queue->curr_idx = 0;
				break;
			}
			/* curr_idx is at the end*/
			item = &age_queue->item_array[age_queue->curr_idx];
			/* if query error, we aged it */
			ret = QUERY_FLOW_COUNTER(item->age_ctx, &curr_pkts);
			if (ret < 0)
				goto check_aged;
			/* flow hit update */
			curr_sec = CURR_SEC;
			if (curr_pkts != item->pkts) {
				item->pkts = curr_pkts;
				item->expr_time_in_sec = curr_sec + item->aging_time_in_sec;
				age_queue->curr_idx++;
				continue;
			}
			/* not expired */
			if (item->expr_time_in_sec > curr_sec) {
				age_queue->curr_idx++;
				continue;
			}
check_aged:
			entries[nb_aged].user_data = item->user_data ?
			item->user_data : (uint64_t)item->age_ctx;
			nb_aged++;
			nb_aged_pipe++;
			__flow_age_index_rm(age_queue, age_queue->curr_idx);
			if (nb_aged >= len)
				goto out;
		}
	}
out:
	if (nb_aged)
		return nb_aged;
	else
		return checked_pipes < nb_pipes ? 0 : -1;
}
