/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef FLOW_CT_H_
#define FLOW_CT_H_

#include <stdint.h>
#include <stdio.h>
#include <errno.h>

#include "doca_flow.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Number of max worker threads
 */
#define FLOW_CT_MAX_WORKER 256

/**
 * @brief CT context
 */
struct flow_ct_context {
	struct doca_flow_pipe *ct_pipe;	  /**< CT 6-tupple pipe */
	uint16_t n_rq_ids; /**< Number of worker RQ IDs */
	uint16_t rq_ids[FLOW_CT_MAX_WORKER]; /**< Worker RQ IDs */
};

/**
 * @brief module initilization callback
 *
 * @param port_id
 *   Port ID.
 * @param flow_cfg
 *   Doca Flow configuration
 * @param port
 *   Doca Flow port
 * @return
 *   CT context on success, NULL on error.
 */
typedef struct flow_ct_context *(*ct_init_cb)(uint16_t port_id, const struct doca_flow_cfg *flow_cfg,
			struct doca_flow_port *port);

/**
 * @brief module close callback
 *
 * @param ctx
 *   CT context
 */
typedef void (*ct_close_cb)(struct flow_ct_context *ctx);

/**
 * @brief CT pipe creation callback
 *
 * @param ctx
 *   CT context
 * @param cfg
 *   Pipe configuration
 * @param fwd
 *   PIpe forwarding actions
 * @param fwd_miss
 *   Pipe forwarding miss actions
 * @param pipe
 *   Pipe
 * @param error
 *   Error information
 * @return
 *   0 on success, negative error number otherwise
 */
typedef int (*ct_pipe_create_cb)(struct flow_ct_context *ctx, const struct doca_flow_pipe_cfg *cfg,
			const struct doca_flow_fwd *fwd, const struct doca_flow_fwd *fwd_miss,
			struct doca_flow_pipe *pipe, struct doca_flow_error *error);

/**
 * @brief CT pipe destroy callback
 *
 * @param ctx
 *   CT context
 */
typedef void (*ct_pipe_destroy_cb)(struct flow_ct_context *ctx);


/**
 * @brief Dump CT internal pipes statistics
 *
 * @param f
 *   The output file of the pipe information.
 */
typedef void (*ct_pipe_dump_cb)(FILE *f);

/**
 * @brief CT operation callbacks
 */
struct flow_ct_ops {
	ct_init_cb init; /**< CT module init */
	ct_close_cb close; /**< CT module close */
	ct_pipe_create_cb create; /**< CT pipe create */
	ct_pipe_destroy_cb destroy; /**< CT pipe destroy */
	ct_pipe_dump_cb dump; /**< CT pipe statistics dump */
};

/**
 * @brief Regisrer Doca CT operations.
 *
 * Doca CT module register its operations to Doca flow.
 *
 * @param ops
 * Doca CT implementation callbacks.
 */
__DOCA_EXPERIMENTAL
void flow_ct_register(struct flow_ct_ops *ops);

/**
 * @brief Get DPDK group ID from pipe.
 *
 * @param pipe
 * Pointer to doca flow pipe.
 * @param id
 * Pointer to save pipe ID.
 * @return
 * 0 on success, negative on error.
 */
__DOCA_EXPERIMENTAL
int flow_pipe_id_get(struct doca_flow_pipe *pipe, uint32_t *id);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* FLOW_CT_H_ */