/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <doca_log.h>
#include "id_pool.h"
#include "priv_doca_mempool.h"

DOCA_LOG_REGISTER(id_pool);

struct id_pool {
	struct priv_doca_mempool *mempool;
	struct id_pool_cfg cfg;
};

static void
id_pool_obj_cb(void *ctx, void *addr, uint32_t idx)
{
	struct id_pool *pool = (struct id_pool *)ctx;
	uint32_t *id = (uint32_t *)addr;

	*id = idx + pool->cfg.min;
}

struct id_pool *
id_pool_create(struct id_pool_cfg *cfg)
{
	struct id_pool *pool;
	struct priv_doca_mempool_cfg mempool_cfg;

	if (cfg == NULL || cfg->size <= 0 || cfg->min < 0)
		return NULL;

	pool = (struct id_pool *)calloc(1, sizeof(struct id_pool));
	if (!pool)
		return NULL;

	pool->cfg = *cfg;
	memset(&mempool_cfg, 0, sizeof(mempool_cfg));
	mempool_cfg.name = (char *)cfg->name;
	mempool_cfg.nb_entries = cfg->size;
	mempool_cfg.size = sizeof(uint32_t);
	mempool_cfg.nb_caches = cfg->nb_caches;
	mempool_cfg.usr_ctx = pool;
	mempool_cfg.usr_init_cb = id_pool_obj_cb;

	pool->mempool = priv_doca_mempool_create(&mempool_cfg);
	if (!pool->mempool) {
		DOCA_LOG_ERR(
		    "Failed creation of id pool with size %u - mempool is null",
		    cfg->size);
		goto err;
	}

	DOCA_LOG_INFO("Initialized ID Pool %s with address %p of size %u, min index %u",
		      cfg->name, pool, cfg->size, cfg->min);

	return pool;
err:
	if (pool)
		free(pool);
	return NULL;
}

uint32_t *
id_pool_alloc(struct id_pool *pool, int qidx)
{
	void *id;

	if (pool == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed allocating id - pool is null");
		return NULL;
	}

	if (qidx >= pool->cfg.nb_caches) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"%s: Failed allocating id - invalid pool cache %u index",
			pool->cfg.name, qidx);
		return NULL;
	}
	id = priv_doca_mempool_alloc(pool->mempool, qidx);

	if (!id)
		return NULL;

	return (uint32_t *)id;
}

void
id_pool_free(struct id_pool *pool, int qidx, uint32_t *id)
{
	if (pool == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("Failed freeing id - pool is null");
		return;
	}

	if (qidx >= pool->cfg.nb_caches) {
		DOCA_LOG_RATE_LIMIT_ERR(
			"Failed freeing id - invalid pool cache %u index",
			qidx);
		return;
	}
	priv_doca_mempool_free(pool->mempool, (void *)id, qidx);
}

void
id_pool_destroy(struct id_pool *pool)
{
	if (!pool) {
		DOCA_LOG_ERR(
		    "Failed to destroy id pool with a NULL pool");
		return;
	}
	priv_doca_mempool_destroy(pool->mempool);
	pool->mempool = NULL;
	DOCA_LOG_INFO("Destroyed ID Pool %s with address %p of size %u",
		      pool->cfg.name, pool, pool->cfg.size);
	free(pool);
}
