/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "ordered_list.h"

#include <doca_log.h>

#include <stdbool.h>

struct ordered_list_build_ctx {
	const struct doca_flow_ordered_list *ordered_list;
	struct ordered_list_pipe_cfg *pipe_cfg;
	struct ordered_list_cfg *list_cfg;
	struct doca_flow_action_descs *action_descs;
	int actions_elt_idx;
	int monitor_elt_idx;
};

DOCA_LOG_REGISTER(ordered_list);

static void
ordered_list_build_ctx_reset(struct ordered_list_build_ctx *list_ctx)
{
	list_ctx->action_descs = NULL;
	list_ctx->actions_elt_idx = -1;
	list_ctx->monitor_elt_idx = -1;
}

static void
ordered_list_build_ctx_init(struct ordered_list_build_ctx *list_ctx,
		const struct doca_flow_ordered_list *ordered_list,
		struct ordered_list_pipe_cfg *pipe_cfg,
		struct ordered_list_cfg *list_cfg)
{
	list_ctx->pipe_cfg = pipe_cfg;
	list_ctx->ordered_list = ordered_list;
	list_ctx->list_cfg = list_cfg;
	ordered_list_build_ctx_reset(list_ctx);
}

static void
ordered_list_build_ctx_advance(struct ordered_list_build_ctx *list_ctx)
{
	const struct doca_flow_ordered_list *ordered_list = list_ctx->ordered_list;
	struct ordered_list_pipe_cfg *pipe_cfg = list_ctx->pipe_cfg;
	struct doca_flow_actions *actions;
	struct doca_flow_monitor *monitor;
	struct dpdk_pipe_cfg *hash_pipe_cfg;
	struct ordered_list_item *list_item;
	uint32_t hash_pipe_idx;

	actions = list_ctx->actions_elt_idx >= 0 ?
			(struct doca_flow_actions *)ordered_list->elements[list_ctx->actions_elt_idx] :
			NULL;
	monitor = list_ctx->monitor_elt_idx >= 0 ?
			(struct doca_flow_monitor *)ordered_list->elements[list_ctx->monitor_elt_idx] :
			NULL;
	list_item = &list_ctx->list_cfg->items[list_ctx->list_cfg->nb_items];

	/*
	 * Possible optimization: determine which hash pipe to use.
	 * Previously processed lists may define monitor template for each pipe.
	 * If the monitor is the same, the pipe can be reused to add new actions.
	 * It the monitor differs, a new pipe must be selected.
	 */
	hash_pipe_idx = pipe_cfg->nb_hash_pipes++;
	hash_pipe_cfg = &pipe_cfg->hash_pipe_cfg_arr[hash_pipe_idx];

	pipe_cfg->actions_arr[hash_pipe_idx] = actions;
	pipe_cfg->action_descs_arr[hash_pipe_idx] = list_ctx->action_descs;
	hash_pipe_cfg->nb_actions = 1;
	hash_pipe_cfg->uds_cfg.actions = (const struct doca_flow_actions **)
			&pipe_cfg->actions_arr[hash_pipe_idx];
	hash_pipe_cfg->uds_cfg.action_descs = (const struct doca_flow_action_descs **)
			&pipe_cfg->action_descs_arr[hash_pipe_idx];
	hash_pipe_cfg->uds_cfg.monitor = monitor;

	list_item->hash_pipe_idx = hash_pipe_idx;
	list_item->actions_elt_idx = list_ctx->actions_elt_idx;
	list_item->monitor_elt_idx = list_ctx->monitor_elt_idx;
	list_ctx->list_cfg->nb_items++;

	ordered_list_build_ctx_reset(list_ctx);
}

static bool
next_is_action_descs(uint32_t idx, const struct doca_flow_ordered_list *list)
{
	return !!(idx < list->size - 1 &&
		  list->types[idx + 1] == DOCA_FLOW_ORDERED_LIST_ELEMENT_ACTION_DESCS);
}

static int
build(struct ordered_list_pipe_cfg *pipe_cfg,
		const struct doca_flow_ordered_list *ordered_list,
		struct ordered_list_cfg *list)
{
	struct ordered_list_build_ctx list_ctx;
	uint32_t elt_idx;

	ordered_list_build_ctx_init(&list_ctx, ordered_list, pipe_cfg, list);
	for (elt_idx = 0; elt_idx < ordered_list->size; elt_idx++) {
		switch (ordered_list->types[elt_idx]) {
		case DOCA_FLOW_ORDERED_LIST_ELEMENT_ACTIONS:
			list_ctx.actions_elt_idx = elt_idx;
			if (!next_is_action_descs(elt_idx, ordered_list))
				ordered_list_build_ctx_advance(&list_ctx);
			break;
		case DOCA_FLOW_ORDERED_LIST_ELEMENT_ACTION_DESCS:
			list_ctx.action_descs = (struct doca_flow_action_descs *)
					ordered_list->elements[elt_idx];
			ordered_list_build_ctx_advance(&list_ctx);
			break;
		case DOCA_FLOW_ORDERED_LIST_ELEMENT_MONITOR:
			list_ctx.monitor_elt_idx = elt_idx;
			ordered_list_build_ctx_advance(&list_ctx);
			break;
		}
	}
	return 0;
}

int
ordered_list_pipe_cfg_build(struct ordered_list_pipe_cfg *pipe_cfg,
		const struct doca_flow_ordered_list **ordered_lists,
		uint8_t nb_ordered_lists,
		struct ordered_list_cfg *lists)
{
	uint32_t list_idx;
	int rc;

	for (list_idx = 0; list_idx < nb_ordered_lists; list_idx++) {
		rc = build(pipe_cfg, ordered_lists[list_idx], &lists[list_idx]);
		if (rc < 0) {
			DOCA_LOG_ERR("Failed to build list idx=%u (rc=%d)", list_idx, rc);
			return rc;
		}
	}
	return 0;
}
