/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <stdint.h>

#include "private/common_defines.h"
#include <priv_doca_malloc.h>
#include <doca_log.h>


#include "doca_flow_priv.h"
#include <dpdk_engine.h>

#include "utils_df_translate.h"
#include "lpm_utils.h"
#include "pipe_lpm.h"

DOCA_LOG_REGISTER(pipe_lpm);

#define LPM_DEFAULT_META_TAG 0

/** Maximum cidr_bits value possible for IPv4 LPM. */
#define LPM_MAX_IPV4_DEPTH               32  // 4 bytes * 8 bits
#define LPM_MAX_IPV6_DEPTH               128 // 16 bytes * 8 bits
#define LPM_MAX_MAC_DEPTH                48 //  6 bytes * 8 bits

#define LPM_MAX_NUM_TABLES 128
#define LPM_UNKNOWN_TABLE -1

#define LPM_MAX_ADDR_LEN 16 // max of ipv4 len, MAC len and ipv6 len

#define LPM_ACTION_TABLE 1234

#define LPM_MAX_SUPPORTED_TABLE_TREE_DEPTH 8 // log(128) + 1
#define LPM_MAX_SUPPORTED_ENTRY_TREE_DEPTH 64

#define LPM_NB_ACTIONS_ARR	1
#define LPM_DEFAULT_TIMEOUT_US (10000)		/* Timeout for processing pipe entries */

enum lpm_addr_type {
	LPM_ADDR_TYPE_ETHER_SRC_OUT,
	LPM_ADDR_TYPE_ETHER_DST_OUT,
	LPM_ADDR_TYPE_ETHER_SRC_IN,
	LPM_ADDR_TYPE_ETHER_DST_IN,
	LPM_ADDR_TYPE_IPV4_SRC_OUT,
	LPM_ADDR_TYPE_IPV4_DST_OUT,
	LPM_ADDR_TYPE_IPV4_SRC_IN,
	LPM_ADDR_TYPE_IPV4_DST_IN,
	LPM_ADDR_TYPE_IPV6_SRC_OUT,
	LPM_ADDR_TYPE_IPV6_DST_OUT,
	LPM_ADDR_TYPE_IPV6_SRC_IN,
	LPM_ADDR_TYPE_IPV6_DST_IN,
};

enum lpm_state {
	LPM_STATE_INIT,
	LPM_STATE_COLLECT,
	LPM_STATE_FLUSH_START,
	LPM_STATE_FLUSH_DONE,
};

enum lpm_tree_type {
	LPM_TREE_MAIN,
	LPM_TREE_SECONDARY
};

enum lpm_entry_type {
	LPM_ENTRY_MARKED_TYPE,
	LPM_ENTRY_BMP_ORIG_TYPE,
	LPM_ENTRY_BMP_UPDATED_TYPE,
	LPM_ENTRY_TO_REMOVE
};

struct lpm_tree_table_node;

struct lpm_entry_node_data {
	int node_tree_depth;
	uint8_t bytes_in_depth_value;  // in bytes. [cidr_bits/8 + (cidr_bits%8)?1:0]  for cidr_bits 32 bytes_in_depth_value = 4 ; for cidr_bits 30 bytes_in_depth_value = 4
	enum lpm_entry_type entry_type;
	uint32_t bmp_tag;
	uint8_t addr_masked[LPM_MAX_ADDR_LEN];
	struct lpm_tree_table_node *tree_table_node;
	struct lpm_tree_table_node *jump_table_node;
	struct doca_flow_fwd fwd;
	struct doca_flow_actions actions;
	struct doca_flow_pipe_entry *flow_pipe_entry;
	struct doca_flow_pipe_entry *flow_action_entry;
	struct doca_flow_pipe_entry *doca_entry;
};

struct lpm_tree_entry_node {
	struct lpm_entry_node_data node_data;
	struct lpm_tree_entry_node *left;
	struct lpm_tree_entry_node *right;
};

struct lpm_table_node_data {
	int node_tree_depth;
	uint8_t cidr_bits;
	uint8_t cidr_mask[LPM_MAX_ADDR_LEN];
	struct doca_flow_match match_mask;
	struct doca_flow_match match;
	struct lpm_tree_entry_node *root_entry;
	struct doca_flow_pipe *pipe;
};

struct lpm_tree_table_node {
	struct lpm_table_node_data node_data;
	struct lpm_tree_table_node *left;
	struct lpm_tree_table_node *right;
	struct lpm_tree_table_node *parent;
};

struct lpm_tree {
	struct lpm_tree_table_node *tables_array[LPM_MAX_NUM_TABLES];
	struct lpm_tree_table_node *lpm_tree_root;
	struct doca_flow_pipe *action_pipe;
};

struct lpm_priv_s {
	enum lpm_addr_type addr_type;
	uint8_t len_addr;
	uint32_t rule_counter;
	struct lpm_tree main_tree;
	struct lpm_tree secondary_tree;
	struct lpm_tree *curr_tree;
	struct lpm_tree *tree_to_release;
	enum lpm_state state;
	enum lpm_tree_type tree_type;
	struct doca_flow_pipe *dispatcher_pipe;
	struct doca_flow_port port;
	uint32_t nb_flows;
	int meta_tag;
	rte_spinlock_t lock;
	struct lpm_tree_entry_node *last_created_node;
};

static int
lpm_get_address_type(const struct doca_flow_match *match, enum lpm_addr_type *addr_type)
{
	uint8_t set_cnt = 0;

	if (!utils_df_translate_is_mac_ignored(match->out_src_mac)) {
		*addr_type = LPM_ADDR_TYPE_ETHER_SRC_OUT;
		set_cnt++;
	}
	if (!utils_df_translate_is_mac_ignored(match->out_dst_mac)) {
		*addr_type = LPM_ADDR_TYPE_ETHER_DST_OUT;
		set_cnt++;
	}
	if (!utils_df_translate_is_mac_ignored(match->in_src_mac)) {
		*addr_type = LPM_ADDR_TYPE_ETHER_SRC_IN;
		set_cnt++;
	}
	if (!utils_df_translate_is_mac_ignored(match->in_dst_mac)) {
		*addr_type = LPM_ADDR_TYPE_ETHER_DST_IN;
		set_cnt++;
	}
	if (!utils_df_translate_is_ip_ignored(&match->out_src_ip)) {
		if (match->out_src_ip.type == DOCA_FLOW_IP6_ADDR)
			*addr_type = LPM_ADDR_TYPE_IPV6_SRC_OUT;
		else
			*addr_type = LPM_ADDR_TYPE_IPV4_SRC_OUT;
		set_cnt++;
	}
	if (!utils_df_translate_is_ip_ignored(&match->out_dst_ip)) {
		if (match->out_dst_ip.type == DOCA_FLOW_IP6_ADDR)
			*addr_type = LPM_ADDR_TYPE_IPV6_DST_OUT;
		else
			*addr_type = LPM_ADDR_TYPE_IPV4_DST_OUT;
		set_cnt++;
	}
	if (!utils_df_translate_is_ip_ignored(&match->in_src_ip)) {
		if (match->in_src_ip.type == DOCA_FLOW_IP6_ADDR)
			*addr_type = LPM_ADDR_TYPE_IPV6_SRC_IN;
		else
			*addr_type = LPM_ADDR_TYPE_IPV4_SRC_IN;
		set_cnt++;
	}
	if (!utils_df_translate_is_ip_ignored(&match->in_dst_ip)) {
		if (match->in_dst_ip.type == DOCA_FLOW_IP6_ADDR)
			*addr_type = LPM_ADDR_TYPE_IPV6_DST_IN;
		else
			*addr_type = LPM_ADDR_TYPE_IPV4_DST_IN;
		set_cnt++;
	}
	if (set_cnt == 1)
		return 0;
	else
		return -1;
}

static void
lpm_get_addr_len(enum lpm_addr_type addr_type, uint8_t *len_addr)
{
	switch (addr_type) {
	case LPM_ADDR_TYPE_ETHER_SRC_OUT: /* fall through */
	case LPM_ADDR_TYPE_ETHER_DST_OUT: /* fall through */
	case LPM_ADDR_TYPE_ETHER_SRC_IN:  /* fall through */
	case LPM_ADDR_TYPE_ETHER_DST_IN:
		*len_addr = 6;
		break;
	case LPM_ADDR_TYPE_IPV4_SRC_OUT: /* fall through */
	case LPM_ADDR_TYPE_IPV4_DST_OUT: /* fall through */
	case LPM_ADDR_TYPE_IPV4_SRC_IN:  /* fall through */
	case LPM_ADDR_TYPE_IPV4_DST_IN:
		*len_addr = 4;
		break;
	case LPM_ADDR_TYPE_IPV6_SRC_OUT: /* fall through */
	case LPM_ADDR_TYPE_IPV6_DST_OUT: /* fall through */
	case LPM_ADDR_TYPE_IPV6_SRC_IN:  /* fall through */
	case LPM_ADDR_TYPE_IPV6_DST_IN:
		*len_addr = 16;
		break;
	}
}

static void
lpm_create_match_for_main_pipe(enum lpm_addr_type addr_type, struct doca_flow_match *match)
{
	switch (addr_type) {
	case LPM_ADDR_TYPE_ETHER_SRC_OUT:
	case LPM_ADDR_TYPE_ETHER_DST_OUT:
	case LPM_ADDR_TYPE_ETHER_SRC_IN:
	case LPM_ADDR_TYPE_ETHER_DST_IN:
		break;
	case LPM_ADDR_TYPE_IPV4_SRC_OUT:
		match->out_src_ip.type = DOCA_FLOW_IP4_ADDR;
		break;
	case LPM_ADDR_TYPE_IPV4_DST_OUT:
		match->out_dst_ip.type = DOCA_FLOW_IP4_ADDR;
		break;
	case LPM_ADDR_TYPE_IPV4_SRC_IN:
		match->in_src_ip.type = DOCA_FLOW_IP4_ADDR;
		break;
	case LPM_ADDR_TYPE_IPV4_DST_IN:
		match->in_dst_ip.type = DOCA_FLOW_IP4_ADDR;
		break;
	case LPM_ADDR_TYPE_IPV6_SRC_OUT:
		match->out_src_ip.type = DOCA_FLOW_IP6_ADDR;
		break;
	case LPM_ADDR_TYPE_IPV6_DST_OUT:
		match->out_dst_ip.type = DOCA_FLOW_IP6_ADDR;
		break;
	case LPM_ADDR_TYPE_IPV6_SRC_IN:
		match->in_src_ip.type = DOCA_FLOW_IP6_ADDR;
		break;
	case LPM_ADDR_TYPE_IPV6_DST_IN:
		match->in_dst_ip.type = DOCA_FLOW_IP6_ADDR;
		break;
	default:
		break;
	}
}

static int
lpm_create_action_pipe(struct lpm_priv_s *lpm_priv,
		struct doca_flow_actions **actions,
		const struct doca_flow_fwd *fwd,
		const struct doca_flow_fwd *fwd_miss)
{
	struct doca_flow_pipe_cfg pipe_cfg = {0};
	struct doca_flow_match match = {0};
	struct doca_flow_match match_mask = {0};

	pipe_cfg.port = &lpm_priv->port;
	pipe_cfg.attr.is_root = false;
	pipe_cfg.attr.name = "ACTION_PIPE";
	pipe_cfg.attr.type = DOCA_FLOW_PIPE_BASIC;
	pipe_cfg.match = &match;
	pipe_cfg.match_mask = &match_mask;
	pipe_cfg.actions = actions;
	pipe_cfg.attr.nb_actions = LPM_NB_ACTIONS_ARR;

	match_mask.meta.u32[lpm_priv->meta_tag] = UINT32_MAX;

	lpm_priv->main_tree.action_pipe = dpdk_pipe_create(ENGINE_PIPE_BASIC, &pipe_cfg,
			fwd, fwd_miss);
	if (lpm_priv->main_tree.action_pipe == NULL) {
		DOCA_DLOG_ERR("create main lpm action pipe error");
		return -1;
	}

	lpm_priv->secondary_tree.action_pipe = dpdk_pipe_create(ENGINE_PIPE_BASIC, &pipe_cfg,
			fwd, fwd_miss);
	if (lpm_priv->secondary_tree.action_pipe == NULL) {
		DOCA_DLOG_ERR("create secondary lpm action pipe error");
		return -1;
	}
	return 0;
}

static int
lpm_create_dispatcher_pipe(struct lpm_priv_s *lpm_priv)
{
	struct doca_flow_pipe_cfg pipe_cfg = {0};
	struct doca_flow_match match = {0};
	struct doca_flow_actions actions = {0};
	struct doca_flow_actions *actions_arr[LPM_NB_ACTIONS_ARR];

	pipe_cfg.port = &lpm_priv->port;
	pipe_cfg.attr.is_root = false;
	pipe_cfg.attr.name = "LPM_DISPATCHER_PIPE";
	pipe_cfg.attr.type = DOCA_FLOW_PIPE_BASIC;
	pipe_cfg.match = &match;
	actions_arr[0] = &actions;
	pipe_cfg.actions = actions_arr;
	pipe_cfg.attr.nb_actions = LPM_NB_ACTIONS_ARR;

	actions.meta.u32[lpm_priv->meta_tag] = UINT32_MAX;

	lpm_priv->dispatcher_pipe = dpdk_pipe_create(ENGINE_PIPE_BASIC,
							&pipe_cfg,
							NULL /* fwd */,
							NULL /* fwd_miss */);
	if (lpm_priv->dispatcher_pipe == NULL) {
		DOCA_DLOG_ERR("create lpm dispatcher pipe error");
		return -1;
	}
	return 0;
}

static int
_lpm_init(struct lpm_priv_s *lpm_priv,
		const struct doca_flow_pipe_cfg *cfg,
		const struct doca_flow_fwd *fwd,
		const struct doca_flow_fwd *fwd_miss)
{
	lpm_priv->port = *cfg->port;
	lpm_priv->nb_flows = cfg->attr.nb_flows;
	lpm_priv->rule_counter = 0;

	memset(lpm_priv->main_tree.tables_array, 0,
	       sizeof(struct lpm_tree_table_node *) * LPM_MAX_NUM_TABLES);
	memset(lpm_priv->secondary_tree.tables_array, 0,
	       sizeof(struct lpm_tree_table_node *) * LPM_MAX_NUM_TABLES);
	lpm_priv->curr_tree = &lpm_priv->main_tree;
	lpm_priv->tree_to_release = NULL;
	lpm_priv->state = LPM_STATE_INIT;
	lpm_priv->tree_type = LPM_TREE_MAIN;

	rte_spinlock_init(&lpm_priv->lock);

	if (lpm_create_action_pipe(lpm_priv, cfg->actions, fwd, fwd_miss) < 0)
		goto init_error;

	if (lpm_create_dispatcher_pipe(lpm_priv) < 0)
		goto init_error;

	return 0;

init_error:
	if (lpm_priv->dispatcher_pipe) {
		dpdk_pipe_destroy(cfg->port, lpm_priv->dispatcher_pipe);
		lpm_priv->dispatcher_pipe = NULL;
	}
	if (lpm_priv->main_tree.action_pipe) {
		dpdk_pipe_destroy(cfg->port, lpm_priv->main_tree.action_pipe);
		lpm_priv->main_tree.action_pipe = NULL;
	}
	if (lpm_priv->secondary_tree.action_pipe) {
		dpdk_pipe_destroy(cfg->port, lpm_priv->secondary_tree.action_pipe);
		lpm_priv->secondary_tree.action_pipe = NULL;
	}

	return -1;
}

void *
lpm_create(const struct doca_flow_pipe_cfg *cfg,
		const struct doca_flow_fwd *fwd,
		const struct doca_flow_fwd *fwd_miss)
{
	struct lpm_priv_s *lpm_priv;

	lpm_priv = (struct lpm_priv_s *)priv_doca_malloc(1, sizeof(struct lpm_priv_s));
	if (lpm_priv == NULL) {
		DOCA_LOG_ERR("failed to create lpm - out of memory");
		return NULL;
	}
	memset(lpm_priv, 0, sizeof(struct lpm_priv_s));

	if (lpm_get_address_type(cfg->match, &lpm_priv->addr_type) < 0) {
		DOCA_LOG_ERR("failed to create lpm - selected field undetected");
		goto create_error;
	}

	lpm_get_addr_len(lpm_priv->addr_type, &lpm_priv->len_addr);

	lpm_priv->meta_tag = LPM_DEFAULT_META_TAG;

	if (_lpm_init(lpm_priv, cfg, fwd, fwd_miss) < 0)
		goto create_error;

	return (void *)lpm_priv;

create_error:
	lpm_destroy(lpm_priv);
	return NULL;
}

void
lpm_extract_match(void *lpm, struct doca_flow_match *match)
{
	struct lpm_priv_s *lpm_priv = lpm;

	lpm_create_match_for_main_pipe(lpm_priv->addr_type, match);
}

static void
lpm_destroy_entry_node(struct lpm_tree_entry_node *node)
{
	if (node == NULL)
		return;
	lpm_destroy_entry_node(node->left);
	lpm_destroy_entry_node(node->right);
	if (node->node_data.flow_action_entry)
		dpdk_pipe_entry_remove(node->node_data.flow_action_entry->pipe_queue, NULL,
				       node->node_data.flow_action_entry);
	if (node->node_data.fwd.rss_queues)
		priv_doca_free(node->node_data.fwd.rss_queues);
	priv_doca_free(node);
}

static void
lpm_destroy_table_node(struct lpm_priv_s *lpm_priv, struct lpm_tree_table_node *node)
{
	if (node == NULL)
		return;
	lpm_destroy_table_node(lpm_priv, node->left);
	lpm_destroy_table_node(lpm_priv, node->right);
	if (node->node_data.pipe)
		dpdk_pipe_destroy(&lpm_priv->port, node->node_data.pipe);
	lpm_destroy_entry_node(node->node_data.root_entry);
	priv_doca_free(node);
}


static void
lpm_destoy_tree(struct lpm_priv_s *lpm_priv, struct lpm_tree *tree)
{
	if (tree == NULL)
		return;
	memset(tree->tables_array, 0, sizeof(struct lpm_tree_table_node *)*LPM_MAX_NUM_TABLES);
	lpm_destroy_table_node(lpm_priv, tree->lpm_tree_root);
	tree->lpm_tree_root = NULL;
}

void
lpm_destroy(void *lpm)
{
	struct lpm_priv_s *lpm_priv = (struct lpm_priv_s *)lpm;

	if (lpm_priv == NULL)
		return;

	lpm_destoy_tree(lpm_priv, lpm_priv->curr_tree);
	lpm_destoy_tree(lpm_priv, lpm_priv->tree_to_release);

	if (lpm_priv->main_tree.action_pipe)
		dpdk_pipe_destroy(&lpm_priv->port, lpm_priv->main_tree.action_pipe);

	if (lpm_priv->secondary_tree.action_pipe)
		dpdk_pipe_destroy(&lpm_priv->port, lpm_priv->secondary_tree.action_pipe);

	if (lpm_priv->dispatcher_pipe)
		dpdk_pipe_destroy(&lpm_priv->port, lpm_priv->dispatcher_pipe);

	priv_doca_free(lpm_priv);
}

struct doca_flow_pipe *
lpm_get_main_pipe(void *lpm)
{
	if (lpm == NULL)
		return NULL;
	return ((struct lpm_priv_s *)lpm)->dispatcher_pipe;
}

static inline void *
lpm_get_addr_masked(struct lpm_tree_entry_node *node)
{
	return &node->node_data.addr_masked;
}

static inline struct lpm_tree_entry_node *
lpm_tree_table_get_root_entry(struct lpm_tree_table_node *node) {
	if (node == NULL)
		return NULL;
	return node->node_data.root_entry;
}

static inline uint8_t
lpm_tree_entry_get_bytes_in_depth(struct lpm_tree_entry_node *node)
{
	if (node == NULL)
		return 0;
	return node->node_data.bytes_in_depth_value;
}

static int
lpm_entry_tree_get_height(struct lpm_tree_entry_node *node)
{
	if (node == NULL)
		return 0;
	return node->node_data.node_tree_depth;
}
static int
lpm_entry_tree_get_balance(struct lpm_tree_entry_node *node)
{
	if (node == NULL)
		return 0;
	return (lpm_entry_tree_get_height(node->left) - lpm_entry_tree_get_height(node->right));
}

static struct lpm_tree_entry_node *
lpm_entry_tree_right_rotate(struct lpm_tree_entry_node *y)
{
	struct lpm_tree_entry_node *x = y->left;
	struct lpm_tree_entry_node *T2 = x->right;

	x->right = y;
	y->left = T2;

	y->node_data.node_tree_depth = MAX(lpm_entry_tree_get_height(y->left), lpm_entry_tree_get_height(y->right))+1;
	x->node_data.node_tree_depth = MAX(lpm_entry_tree_get_height(x->left), lpm_entry_tree_get_height(x->right))+1;

	return x;
}

static struct lpm_tree_entry_node *
lpm_entry_tree_left_rotate(struct lpm_tree_entry_node *x)
{
	struct lpm_tree_entry_node *y = x->right;
	struct lpm_tree_entry_node *T2 = y->left;

	y->left = x;
	x->right = T2;

	x->node_data.node_tree_depth = MAX(lpm_entry_tree_get_height(x->left), lpm_entry_tree_get_height(x->right))+1;
	y->node_data.node_tree_depth = MAX(lpm_entry_tree_get_height(y->left), lpm_entry_tree_get_height(y->right))+1;

	return y;
}

static struct  lpm_tree_entry_node *
lpm_balance_entry_tree(struct lpm_tree_entry_node *node, uint8_t *addr_masked, uint8_t bytes_in_depth_value)
{
	int res;

	node->node_data.node_tree_depth = 1 + MAX(lpm_entry_tree_get_height(node->left), lpm_entry_tree_get_height(node->right));

	int balance = lpm_entry_tree_get_balance(node);

	if (balance > 1) {
		res = utl_compare_array(addr_masked, lpm_get_addr_masked(node->left), bytes_in_depth_value);
		if (res == -1)
			return lpm_entry_tree_right_rotate(node);
		if (res == 1) {
			node->left =  lpm_entry_tree_left_rotate(node->left);
			return lpm_entry_tree_right_rotate(node);
		}
	}

	if (balance < -1) {
		res = utl_compare_array(addr_masked, lpm_get_addr_masked(node->right), bytes_in_depth_value);

		if (res == 1)
			return lpm_entry_tree_left_rotate(node);

		if (res == -1) {
			node->right = lpm_entry_tree_right_rotate(node->right);
			return lpm_entry_tree_left_rotate(node);
		}
	}

	return node;
}

static int
lpm_tree_get_height(const struct lpm_tree_table_node *node)
{
	if (node == NULL)
		return 0;
	return node->node_data.node_tree_depth;
}

static int
lpm_tree_get_balance(const struct lpm_tree_table_node *node)
{
	if (node == NULL)
		return 0;
	return (lpm_tree_get_height(node->left) - lpm_tree_get_height(node->right));
}

static struct lpm_tree_table_node *
lpm_tree_right_rotate(struct lpm_tree_table_node *y)
{
	struct lpm_tree_table_node *x = y->left;
	struct lpm_tree_table_node *T2 = x->right;

	x->right = y;
	y->left = T2;

	y->node_data.node_tree_depth = MAX(lpm_tree_get_height(y->left), lpm_tree_get_height(y->right))+1;
	x->node_data.node_tree_depth = MAX(lpm_tree_get_height(x->left), lpm_tree_get_height(x->right))+1;

	return x;
}


static struct lpm_tree_table_node *
lpm_tree_left_rotate(struct lpm_tree_table_node *x)
{
	struct lpm_tree_table_node *y = x->right;
	struct lpm_tree_table_node *T2 = y->left;

	y->left = x;
	x->right = T2;

	x->node_data.node_tree_depth = MAX(lpm_tree_get_height(x->left), lpm_tree_get_height(x->right))+1;
	y->node_data.node_tree_depth = MAX(lpm_tree_get_height(y->left), lpm_tree_get_height(y->right))+1;

	return y;
}

static void
lpm_fill_match_type(enum lpm_addr_type addr_type, struct doca_flow_match *match)
{
	switch (addr_type) {
	case LPM_ADDR_TYPE_ETHER_SRC_OUT: /* fall through */
	case LPM_ADDR_TYPE_ETHER_DST_OUT: /* fall through */
	case LPM_ADDR_TYPE_ETHER_SRC_IN:  /* fall through */
	case LPM_ADDR_TYPE_ETHER_DST_IN:  /* fall through */
		return;
	case LPM_ADDR_TYPE_IPV4_SRC_OUT:
		match->out_src_ip.type = DOCA_FLOW_IP4_ADDR;
		return;
	case LPM_ADDR_TYPE_IPV4_DST_OUT:
		match->out_dst_ip.type = DOCA_FLOW_IP4_ADDR;
		return;
	case LPM_ADDR_TYPE_IPV4_SRC_IN:
		match->in_src_ip.type = DOCA_FLOW_IP4_ADDR;
		return;
	case LPM_ADDR_TYPE_IPV4_DST_IN:
		match->in_dst_ip.type = DOCA_FLOW_IP4_ADDR;
		return;
	case LPM_ADDR_TYPE_IPV6_SRC_OUT:
		match->out_src_ip.type = DOCA_FLOW_IP6_ADDR;
		return;
	case LPM_ADDR_TYPE_IPV6_DST_OUT:
		match->out_dst_ip.type = DOCA_FLOW_IP6_ADDR;
		return;
	case LPM_ADDR_TYPE_IPV6_SRC_IN:
		match->in_src_ip.type = DOCA_FLOW_IP6_ADDR;
		return;
	case LPM_ADDR_TYPE_IPV6_DST_IN:
		match->in_dst_ip.type = DOCA_FLOW_IP6_ADDR;
		return;
	default:
		DOCA_DLOG_ERR("LPM pipe field type %d is invalid.", addr_type);
		return;
	}
}

static uint8_t *
lpm_get_match_addr(enum lpm_addr_type addr_type, const struct doca_flow_match *match)
{
	switch (addr_type) {
	case LPM_ADDR_TYPE_ETHER_SRC_OUT:
		return (uint8_t *)match->out_src_mac;
	case LPM_ADDR_TYPE_ETHER_DST_OUT:
		return (uint8_t *)match->out_dst_mac;
	case LPM_ADDR_TYPE_ETHER_SRC_IN:
		return (uint8_t *)match->in_src_mac;
	case LPM_ADDR_TYPE_ETHER_DST_IN:
		return (uint8_t *)match->in_dst_mac;
	case LPM_ADDR_TYPE_IPV4_SRC_OUT:
		return (uint8_t *)&match->out_src_ip.ipv4_addr;
	case LPM_ADDR_TYPE_IPV4_DST_OUT:
		return (uint8_t *)&match->out_dst_ip.ipv4_addr;
	case LPM_ADDR_TYPE_IPV4_SRC_IN:
		return (uint8_t *)&match->in_src_ip.ipv4_addr;
	case LPM_ADDR_TYPE_IPV4_DST_IN:
		return (uint8_t *)&match->in_dst_ip.ipv4_addr;
	case LPM_ADDR_TYPE_IPV6_SRC_OUT:
		return (uint8_t *)match->out_src_ip.ipv6_addr;
	case LPM_ADDR_TYPE_IPV6_DST_OUT:
		return (uint8_t *)match->out_dst_ip.ipv6_addr;
	case LPM_ADDR_TYPE_IPV6_SRC_IN:
		return (uint8_t *)match->in_src_ip.ipv6_addr;
	case LPM_ADDR_TYPE_IPV6_DST_IN:
		return (uint8_t *)match->in_dst_ip.ipv6_addr;
	default:
		DOCA_DLOG_ERR("LPM pipe field type %d is invalid.", addr_type);
		return NULL;
	}
}

static int
lpm_fill_match(struct doca_flow_match *match, uint8_t *entry_addr_mask, struct lpm_priv_s *lpm_priv)
{

	uint8_t *p_addr = lpm_get_match_addr(lpm_priv->addr_type, match);

	if (p_addr == NULL)
		return -1;

	memcpy(p_addr, entry_addr_mask, lpm_priv->len_addr);

	lpm_fill_match_type(lpm_priv->addr_type, match);
	return 0;
}

static struct lpm_tree_table_node *
lpm_add_tree_node(struct lpm_priv_s *lpm_priv, uint8_t cidr_bits)
{
	struct lpm_tree_table_node *node = (struct lpm_tree_table_node *)priv_doca_malloc(1, sizeof(struct lpm_tree_table_node));

	if (node == NULL)
		return NULL;

	node->node_data.cidr_bits = cidr_bits;
	node->node_data.root_entry = NULL;
	memset(node->node_data.cidr_mask, 0xff, lpm_priv->len_addr);
	utl_field_mask_invoke(node->node_data.cidr_mask, lpm_priv->len_addr, cidr_bits);
	lpm_fill_match(&node->node_data.match_mask, node->node_data.cidr_mask, lpm_priv);

	node->left = NULL;
	node->right = NULL;
	node->parent = NULL;
	node->node_data.pipe = NULL;
	lpm_priv->curr_tree->tables_array[cidr_bits-1] = node;

	return node;
}

static struct  lpm_tree_table_node *
lpm_balance_table_tree(struct lpm_tree_table_node *node, int table_num)
{
	node->node_data.node_tree_depth = 1 + MAX(lpm_tree_get_height(node->left), lpm_tree_get_height(node->right));

	int balance = lpm_tree_get_balance(node);

	if (balance > 1 && table_num < node->left->node_data.cidr_bits)
		return lpm_tree_right_rotate(node);

	if (balance < -1 && table_num > node->right->node_data.cidr_bits)
		return lpm_tree_left_rotate(node);

	if (balance > 1 && table_num > node->left->node_data.cidr_bits) {
		node->left = lpm_tree_left_rotate(node->left);
		return lpm_tree_right_rotate(node);
	}

	if (balance < -1 && table_num < node->right->node_data.cidr_bits) {
		node->right = lpm_tree_right_rotate(node->right);
		return lpm_tree_left_rotate(node);
	}

	return node;
}

static struct  lpm_tree_table_node *
lpm_insert_new_table(struct lpm_priv_s *lpm_priv, struct lpm_tree_table_node *node, int table_num)
{
	if (node == NULL)
		return(lpm_add_tree_node(lpm_priv, table_num));

	if (table_num < node->node_data.cidr_bits)
		node->left  = lpm_insert_new_table(lpm_priv, node->left, table_num);
	else if (table_num > node->node_data.cidr_bits)
		node->right = lpm_insert_new_table(lpm_priv, node->right, table_num);
	else // Equal keys are not allowed in BST
		return node;

	return lpm_balance_table_tree(node, table_num);
}

static uint32_t
lpm_get_rule_number(struct lpm_priv_s *lpm_priv)
{
	lpm_priv->rule_counter++;
	if (lpm_priv->rule_counter == UINT_MAX)
		lpm_priv->rule_counter = 0;
	return lpm_priv->rule_counter;
}

static int lpm_cpy_fwd(const struct doca_flow_fwd *src_fwd, struct doca_flow_fwd *dst_fwd)
{
	if (unlikely(dst_fwd == NULL))
		return -EINVAL;

	if (src_fwd == NULL)
		return 0;

	memcpy(dst_fwd, src_fwd, sizeof(struct doca_flow_fwd));
	if (src_fwd->type == DOCA_FLOW_FWD_RSS && src_fwd->num_of_queues) {
		dst_fwd->rss_queues = (uint16_t *)priv_doca_malloc(src_fwd->num_of_queues, sizeof(uint16_t));
		if (dst_fwd->rss_queues == NULL)
			return -ENOMEM;
		memcpy(dst_fwd->rss_queues, src_fwd->rss_queues, src_fwd->num_of_queues*sizeof(uint16_t));
	}

	return 0;
}

static struct lpm_tree_entry_node *
lpm_add_tree_entry_node(struct lpm_priv_s *lpm_priv, uint8_t *addr_masked,
				uint8_t bytes_in_depth_value, const struct doca_flow_fwd *fwd,
				const struct doca_flow_actions *actions,
				bool is_marked, struct lpm_tree_table_node *jump_node)
{
	struct lpm_tree_entry_node *node = (struct lpm_tree_entry_node *)priv_doca_malloc(1, sizeof(struct lpm_tree_entry_node));

	if (node == NULL)
		return NULL;

	node->node_data.jump_table_node = jump_node;

	if (is_marked) {
		node->node_data.bmp_tag = 0;
		node->node_data.entry_type = LPM_ENTRY_MARKED_TYPE;
	} else {
		node->node_data.bmp_tag = lpm_get_rule_number(lpm_priv);
		node->node_data.entry_type = LPM_ENTRY_BMP_ORIG_TYPE;
	}
	node->node_data.bytes_in_depth_value = bytes_in_depth_value;

	memcpy(lpm_get_addr_masked(node), addr_masked, LPM_MAX_ADDR_LEN);

	if (lpm_cpy_fwd(fwd, &node->node_data.fwd) < 0) {
		priv_doca_free(node);
		return NULL;
	}

	if (actions)
		node->node_data.actions = *actions;

	node->left = NULL;
	node->right = NULL;
	node->node_data.node_tree_depth = 1;
	node->node_data.flow_pipe_entry = NULL;
	node->node_data.flow_action_entry = NULL;

	lpm_priv->last_created_node = node;
	return node;
}

static struct lpm_tree_entry_node *
lpm_insert_new_entry(struct lpm_priv_s *lpm_priv, struct lpm_tree_entry_node *root,
			uint8_t *addr_masked,  uint8_t bytes_in_depth_value, const struct doca_flow_fwd *fwd,
			const struct doca_flow_actions *actions,
			bool is_marked, struct lpm_tree_table_node *jump_node)
{
	struct lpm_tree_entry_node *node;
	struct lpm_tree_entry_node *curr, *prev;
	uint16_t max_loop = LPM_MAX_SUPPORTED_ENTRY_TREE_DEPTH;
	struct lpm_tree_entry_node *path[LPM_MAX_SUPPORTED_ENTRY_TREE_DEPTH];
	int n_path = 0;
	int res;


	if (root == NULL)
		return lpm_add_tree_entry_node(lpm_priv, addr_masked, bytes_in_depth_value, fwd,  actions, is_marked, jump_node);

	curr = root;
	while (curr != NULL && --max_loop) {
		prev = curr;
		path[n_path++] = curr;
		res = utl_compare_array(addr_masked, lpm_get_addr_masked(curr), bytes_in_depth_value);
		if (res > 0)
			curr = curr->right;
		else if (res < 0)
			curr = curr->left;
		else
			return root;
	}
	node = lpm_add_tree_entry_node(lpm_priv, addr_masked, bytes_in_depth_value, fwd,  actions, is_marked, jump_node);
	if (res > 0)
		prev->right = node;
	else if (res < 0)
		prev->left = node;

	for (int i = n_path-1; i > 0; i--) {
		curr = lpm_balance_entry_tree(path[i], addr_masked, bytes_in_depth_value);
		if (curr != path[i]) {
			if (path[i-1]->left == path[i])
				path[i-1]->left = curr;
			else if (path[i-1]->right == path[i])
				path[i-1]->right = curr;
		}
	}
	return lpm_balance_entry_tree(path[0], addr_masked, bytes_in_depth_value);
}

static int
lpm_get_mask_depth(enum lpm_addr_type addr_type, const struct doca_flow_match *match_mask,
			uint8_t len_addr, uint8_t *cidr_bits)
{
	int ret;

	switch (addr_type) {
	case LPM_ADDR_TYPE_ETHER_SRC_OUT:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->out_src_mac, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_ETHER_DST_OUT:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->out_dst_mac, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_ETHER_SRC_IN:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->in_src_mac, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_ETHER_DST_IN:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->in_dst_mac, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV4_SRC_OUT:
		ret = utl_get_mask_sequential_bits((uint8_t *)&match_mask->out_src_ip.ipv4_addr, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV4_DST_OUT:
		ret = utl_get_mask_sequential_bits((uint8_t *)&match_mask->out_dst_ip.ipv4_addr, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV4_SRC_IN:
		ret = utl_get_mask_sequential_bits((uint8_t *)&match_mask->in_src_ip.ipv4_addr, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV4_DST_IN:
		ret = utl_get_mask_sequential_bits((uint8_t *)&match_mask->in_dst_ip.ipv4_addr, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV6_SRC_OUT:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->out_src_ip.ipv6_addr, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV6_DST_OUT:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->out_dst_ip.ipv6_addr, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV6_SRC_IN:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->in_src_ip.ipv6_addr, len_addr, cidr_bits);
		break;
	case LPM_ADDR_TYPE_IPV6_DST_IN:
		ret = utl_get_mask_sequential_bits((uint8_t *)match_mask->in_dst_ip.ipv6_addr, len_addr, cidr_bits);
		break;
	default:
		DOCA_DLOG_ERR("LPM pipe field type %d is invalid", addr_type);
		ret = -1;
		break;
	}
	return ret;
}

static int
lpm_get_address(enum lpm_addr_type addr_type, const struct doca_flow_match *match,
			uint8_t len_addr, uint8_t *addr_value)
{
	uint8_t *p_addr = lpm_get_match_addr(addr_type, match);

	if (p_addr == NULL)
		return -1;

	memcpy(addr_value, p_addr, len_addr);
	return 0;
}

static int
lpm_add_new_entry(struct lpm_priv_s *lpm_priv,
			struct doca_flow_pipe_entry *doca_entry,
			uint8_t cidr_bits, int bytes_in_depth_value, uint8_t *addr_masked,
			const struct doca_flow_actions *actions,
			const struct doca_flow_fwd *fwd)
{
	if (lpm_priv->curr_tree->tables_array[cidr_bits-1] == NULL) { // node doesn't exist yet, insert new node to the tree
		lpm_priv->curr_tree->lpm_tree_root = lpm_insert_new_table(lpm_priv, lpm_priv->curr_tree->lpm_tree_root, cidr_bits);
		if (lpm_priv->curr_tree->lpm_tree_root == NULL)
			return -1;
	}

	struct lpm_tree_table_node *node = lpm_priv->curr_tree->tables_array[cidr_bits-1];

	node->node_data.root_entry = lpm_insert_new_entry(lpm_priv, lpm_tree_table_get_root_entry(node), addr_masked,
			bytes_in_depth_value, fwd, actions, false /* is_marked */, NULL);
	if (node->node_data.root_entry == NULL)
		return -1;
	if (lpm_priv->last_created_node) {
		lpm_priv->last_created_node->node_data.tree_table_node = node;
		lpm_priv->last_created_node->node_data.doca_entry = doca_entry;
		doca_entry->obj_ctx = lpm_priv->last_created_node;
		lpm_priv->last_created_node = NULL;
	}
	return 0;
}

static void
lpm_copy_one_node(struct lpm_priv_s *lpm_priv, struct lpm_tree_entry_node *node, uint8_t cidr_bits)
{
	if (node == NULL)
		return;

	if (node->node_data.entry_type == LPM_ENTRY_BMP_ORIG_TYPE)
		lpm_add_new_entry(lpm_priv, node->node_data.doca_entry,
				cidr_bits, lpm_tree_entry_get_bytes_in_depth(node),
				lpm_get_addr_masked(node), &node->node_data.actions,
				&node->node_data.fwd);
	lpm_copy_one_node(lpm_priv, node->left, cidr_bits);
	lpm_copy_one_node(lpm_priv, node->right, cidr_bits);
}

static void
lpm_create_alternate_tree(struct lpm_priv_s *lpm_priv, struct lpm_tree_table_node *node)
{
	if (node == NULL)
		return;
	lpm_copy_one_node(lpm_priv, lpm_tree_table_get_root_entry(node), node->node_data.cidr_bits);
	lpm_create_alternate_tree(lpm_priv, node->left);
	lpm_create_alternate_tree(lpm_priv, node->right);
}

struct lpm_tree_entry_node *
lpm_get_entry_node_by_value(struct lpm_tree_entry_node *node, uint8_t *value, uint8_t nr_to_compare)
{
	if (node == NULL)
		return NULL;

	int res = utl_compare_array(value, lpm_get_addr_masked(node), nr_to_compare);

	if (res < 0)
		return lpm_get_entry_node_by_value(node->left, value, nr_to_compare);
	else if (res > 0)
		return lpm_get_entry_node_by_value(node->right, value, nr_to_compare);
	else
		return node;
}

/**
 * * @return
 *   -EINVAL for incorrect arguments, -ENOENT on lookup miss, 0 on lookup hit
 *
 */
int
lpm_entry_lookup(const struct doca_flow_pipe *pipe,
		const struct doca_flow_match *match,
		const struct doca_flow_match *match_mask)
{
	uint8_t addr_masked[LPM_MAX_ADDR_LEN] = {0};
	struct lpm_tree_entry_node *found_entry;
	uint8_t cidr_bits;
	int ret;

	struct lpm_priv_s *lpm_priv = pipe->private_ctx;

	ret = lpm_get_mask_depth(lpm_priv->addr_type, match_mask, lpm_priv->len_addr, &cidr_bits);
	if (ret < 0) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to look up lmp entry - match mask invalid");
		return -EINVAL;
	}

	lpm_get_address(lpm_priv->addr_type, match, lpm_priv->len_addr, addr_masked);
	utl_field_mask_invoke(addr_masked, lpm_priv->len_addr, cidr_bits);

	for (int i = 0; i < LPM_MAX_NUM_TABLES; i++) {
		if (lpm_priv->curr_tree->tables_array[i]) {
			found_entry = lpm_get_entry_node_by_value(lpm_tree_table_get_root_entry(lpm_priv->curr_tree->tables_array[i]), addr_masked, lpm_priv->len_addr);
			if (found_entry)
				return 0;
		}
	}
	return -ENOENT;
}

int
lpm_entry_add(struct doca_flow_pipe_entry *doca_entry,
		uint16_t pipe_queue,
		const struct doca_flow_pipe *pipe,
		const struct doca_flow_match *match,
		const struct doca_flow_match *match_mask,
		const struct doca_flow_actions *actions,
		const struct doca_flow_fwd *fwd)
{
	uint8_t cidr_bits;
	uint8_t addr_masked[LPM_MAX_ADDR_LEN] = {0};
	int ret;

	struct lpm_priv_s *lpm_priv = pipe->private_ctx;

	ret = lpm_get_mask_depth(lpm_priv->addr_type, match_mask, lpm_priv->len_addr, &cidr_bits);
	if (ret < 0) {
		DOCA_LOG_RATE_LIMIT_ERR("failed to add lpm entry - match mask invalid");
		return -1;
	}
	lpm_get_address(lpm_priv->addr_type, match, lpm_priv->len_addr, addr_masked);

	int bytes_in_depth_value = utl_get_nb_bytes_from_bits(cidr_bits);

	utl_field_mask_invoke(addr_masked, lpm_priv->len_addr, cidr_bits);

	rte_spinlock_lock(&lpm_priv->lock);

	if (lpm_priv->state == LPM_STATE_FLUSH_DONE) {
		struct lpm_tree *prev_tree = lpm_priv->curr_tree;

		if (lpm_priv->tree_type == LPM_TREE_MAIN) {
			lpm_priv->curr_tree = &lpm_priv->secondary_tree;
			lpm_priv->tree_type = LPM_TREE_SECONDARY;
		} else {
			lpm_priv->curr_tree = &lpm_priv->main_tree;
			lpm_priv->tree_type = LPM_TREE_MAIN;
		}

		lpm_create_alternate_tree(lpm_priv, prev_tree->lpm_tree_root);
		lpm_priv->tree_to_release = prev_tree;
	}

	lpm_priv->state = LPM_STATE_COLLECT;

	ret = lpm_add_new_entry(lpm_priv, doca_entry, cidr_bits,  bytes_in_depth_value, addr_masked, actions, fwd);

	rte_spinlock_unlock(&lpm_priv->lock);

	return ret;

}

int
lpm_entry_remove(uint16_t pipe_queue, void *usr_ctx,
		struct doca_flow_pipe_entry *entry)
{
	struct lpm_tree_entry_node *entry_node = (struct lpm_tree_entry_node *)entry->obj_ctx;

	if (entry_node == NULL) {
		DOCA_DLOG_DBG("lpm  entry is null");
		return -1;
	}
	if (entry_node->node_data.entry_type != LPM_ENTRY_BMP_ORIG_TYPE) {
		DOCA_DLOG_DBG("lpm  entry type - '%d' is wrong", entry_node->node_data.entry_type);
		return -1;
	}
	/* the removed entry will not be copied to the secondary tree after the flush*/
	entry_node->node_data.entry_type = LPM_ENTRY_TO_REMOVE;

	/* full version will be support next version only */
	//	dpdk_pipe_entry_remove(pipe_queue, usr_ctx, entry_node->flow_pipe_entry);
	//	dpdk_pipe_entry_remove(pipe_queue, usr_ctx, entry_node->flow_action_entry);
	// !!! need to remove bmp_updated entry as well and add marked entry !!!!

	return 0;
}

// set parents in node's tree
static void
lpm_set_parents(struct lpm_tree_table_node *node)
{
	if (node == NULL)
		return;
	if (node->left)	{
		node->left->parent = node;
		lpm_set_parents(node->left);
	}
	if (node->right) {
		node->right->parent = node;
		lpm_set_parents(node->right);
	}
}

static void
lpm_add_marks_to_parent(struct lpm_tree_entry_node *entry_node,
				struct lpm_tree_table_node *table_node,
				struct lpm_priv_s *lpm_priv)
{
	if (entry_node == NULL)
		return;
	// checking only bmp entries
	if (entry_node->node_data.entry_type == LPM_ENTRY_BMP_ORIG_TYPE) {
		struct lpm_tree_table_node *parent = table_node->parent;
		struct lpm_tree_table_node *curr_node = table_node;
		uint8_t max_loop = LPM_MAX_SUPPORTED_TABLE_TREE_DEPTH;

		while (parent && --max_loop) {
			// curr_node is on the right side of the parent
			if (parent->node_data.cidr_bits < curr_node->node_data.cidr_bits) {
				struct lpm_tree_entry_node *parent_root_entry = lpm_tree_table_get_root_entry(parent);

				if (parent_root_entry == NULL)
					return;
				struct lpm_tree_entry_node *parent_entry = lpm_get_entry_node_by_value(parent_root_entry,
						lpm_get_addr_masked(entry_node),
						lpm_tree_entry_get_bytes_in_depth(parent_root_entry));
				// not found, add marked entry to parent
				if (parent_entry == NULL) {
					uint8_t addr_masked[LPM_MAX_ADDR_LEN] = {0};

					memcpy(addr_masked, lpm_get_addr_masked(entry_node), lpm_priv->len_addr);
					utl_field_mask_invoke(addr_masked, lpm_priv->len_addr, parent->node_data.cidr_bits);
					parent->node_data.root_entry = lpm_insert_new_entry(lpm_priv, parent_root_entry,
							addr_masked,  lpm_tree_entry_get_bytes_in_depth(parent_root_entry),
							NULL, NULL, true, table_node);
					if (lpm_priv->last_created_node) {
						lpm_priv->last_created_node->node_data.tree_table_node = parent;
						lpm_priv->last_created_node = NULL;
					}
				}
				curr_node = parent;
				parent = parent->parent;
			} else
				break; // break from while. we don't need to continue
		}
		if (max_loop == 0) {
			DOCA_DLOG_CRIT("overloop in add marks to parents. ");
			return;
		}
	}
	lpm_add_marks_to_parent(entry_node->left, table_node, lpm_priv);
	lpm_add_marks_to_parent(entry_node->right, table_node, lpm_priv);
}

static void
lpm_add_node_mark(struct lpm_tree_table_node *node, struct lpm_priv_s *lpm_priv)
{
	// our node is on left side , we don't need to add marked node to the parent
	if (node->parent == NULL || node->parent->node_data.cidr_bits > node->node_data.cidr_bits)
		return;

	lpm_add_marks_to_parent(lpm_tree_table_get_root_entry(node), node, lpm_priv);
}

static void
lpm_add_marks(struct lpm_tree_table_node *node, struct lpm_priv_s *lpm_priv)
{
	if (node == NULL)
		return;
	lpm_add_marks(node->left, lpm_priv);
	lpm_add_marks(node->right, lpm_priv);
	lpm_add_node_mark(node, lpm_priv);
}

static void lpm_update_marked_entry(struct lpm_priv_s *lpm_priv, struct lpm_tree_entry_node *entry_node)
{
	struct lpm_tree_table_node *node;
	struct lpm_tree_entry_node *best_found_entry = NULL;
	uint8_t addr_masked[LPM_MAX_ADDR_LEN] = {0};
	struct lpm_tree_entry_node *found_entry;
	struct lpm_tree_entry_node *root_entry;
	uint8_t entry_node_cidr_bits;
	uint8_t best_found_cidr_bits;

	if (entry_node->node_data.entry_type != LPM_ENTRY_MARKED_TYPE)
		return;

	entry_node_cidr_bits  = entry_node->node_data.tree_table_node->node_data.cidr_bits;
	best_found_cidr_bits = 0;

	for (int i = 0; i < LPM_MAX_NUM_TABLES; i++) {
		node = lpm_priv->curr_tree->tables_array[i];
		if (node == NULL || node->node_data.cidr_bits >= entry_node_cidr_bits)
			continue;

		memcpy(addr_masked, lpm_get_addr_masked(entry_node), lpm_priv->len_addr);
		utl_field_mask_invoke(addr_masked, lpm_priv->len_addr, node->node_data.cidr_bits);
		root_entry = lpm_tree_table_get_root_entry(node);
		if (root_entry == NULL)
			return;
		found_entry = lpm_get_entry_node_by_value(root_entry, addr_masked, lpm_tree_entry_get_bytes_in_depth(root_entry));
		if (found_entry && found_entry->node_data.entry_type == LPM_ENTRY_BMP_ORIG_TYPE) {
			if (node->node_data.cidr_bits > best_found_cidr_bits) {
				best_found_cidr_bits = node->node_data.cidr_bits;
				best_found_entry = found_entry;
			}
		}
	}

	if (best_found_entry) {
		entry_node->node_data.entry_type = LPM_ENTRY_BMP_UPDATED_TYPE;
		entry_node->node_data.bmp_tag = best_found_entry->node_data.bmp_tag;
		entry_node->node_data.actions = best_found_entry->node_data.actions;
		if (lpm_cpy_fwd(&best_found_entry->node_data.fwd, &entry_node->node_data.fwd) < 0)
			return;
	}

}

static void
lpm_update_node_bmp(struct lpm_priv_s *lpm_priv, struct lpm_tree_entry_node *entry_node)
{
	if (entry_node == NULL)
		return;

	lpm_update_marked_entry(lpm_priv, entry_node);
	lpm_update_node_bmp(lpm_priv, entry_node->left);
	lpm_update_node_bmp(lpm_priv, entry_node->right);
}

static void
lpm_update_bmp(struct lpm_priv_s *lpm_priv, struct lpm_tree_table_node *node)
{
	if (node == NULL)
		return;
	lpm_update_node_bmp(lpm_priv, lpm_tree_table_get_root_entry(node));
	lpm_update_bmp(lpm_priv, node->left);
	lpm_update_bmp(lpm_priv, node->right);
}

static int
lpm_create_tree_pipe(struct lpm_tree_table_node *node, struct lpm_priv_s *lpm_priv)
{
	struct doca_flow_pipe_cfg pipe_cfg = {0};
	struct doca_flow_actions actions = {0};
	struct doca_flow_actions *actions_arr[LPM_NB_ACTIONS_ARR];
	struct doca_flow_match match = {0};
	struct doca_flow_fwd fwd_miss = {0};
	uint8_t match_field[LPM_MAX_ADDR_LEN];

	fwd_miss.type = DOCA_FLOW_FWD_PIPE;
	if (node->left)
		fwd_miss.next_pipe = node->left->node_data.pipe;
	else
		fwd_miss.next_pipe = lpm_priv->curr_tree->action_pipe;

	pipe_cfg.port = &lpm_priv->port;
	pipe_cfg.attr.is_root = false;
	pipe_cfg.attr.name = "NODE_PIPE";
	pipe_cfg.attr.type = DOCA_FLOW_PIPE_BASIC;
	pipe_cfg.match = &match;
	pipe_cfg.match_mask = &node->node_data.match_mask;
	actions_arr[0] = &actions;
	pipe_cfg.actions = actions_arr;
	pipe_cfg.attr.nb_actions = LPM_NB_ACTIONS_ARR;

	memset(match_field, 0xff, lpm_priv->len_addr);
	lpm_fill_match(&match, match_field, lpm_priv);

	actions.meta.u32[lpm_priv->meta_tag] = UINT32_MAX;


	node->node_data.pipe = dpdk_pipe_create(ENGINE_PIPE_BASIC, &pipe_cfg,
						NULL /* fwd */, &fwd_miss);

	if (node->node_data.pipe == NULL)
		return -1;
	return 0;
}

static int
lpm_create_tree_pipes(struct lpm_tree_table_node *node, struct lpm_priv_s *lpm_priv)
{
	int rc;

	if (node == NULL)
		return 0;
	rc = lpm_create_tree_pipes(node->left, lpm_priv);
	if (rc < 0)
		return rc;
	rc = lpm_create_tree_pipes(node->right, lpm_priv);
	if (rc < 0)
		return rc;

	return lpm_create_tree_pipe(node, lpm_priv);
}

static void
lpm_add_entry_completion_cb(enum dpdk_flow_request_op op, enum dpdk_flow_request_status status,
			void *ctx)
{
	struct doca_flow_pipe_entry *entry = (struct doca_flow_pipe_entry *)ctx;

	if (unlikely(!entry)) {
		DOCA_LOG_CRIT("lpm Entry addition completion got null entry");
		return;
	}

	DOCA_DLOG_DBG("lpm entry %p add completed with status %d", entry, entry->status);

	dpdk_entry_update_status(entry, status);
}

static int
lpm_add_current_entry_to_pipe(uint16_t pipe_queue,
					struct lpm_tree_entry_node *entry_node,
					struct lpm_tree_table_node *node,
					struct lpm_priv_s *lpm_priv)
{
	struct doca_flow_match match = {0};
	struct doca_flow_actions actions = {0};
	struct doca_flow_actions *pactions = NULL;
	struct doca_flow_fwd fwd = {0};

	if (entry_node->node_data.entry_type == LPM_ENTRY_BMP_ORIG_TYPE ||
			entry_node->node_data.entry_type == LPM_ENTRY_BMP_UPDATED_TYPE) {
		/*fill action to node pipe*/
		actions.meta.u32[lpm_priv->meta_tag] = entry_node->node_data.bmp_tag;
		actions.action_idx = 0;
		pactions = &actions;

		/* add entry to the action pipe */
		match.meta.u32[lpm_priv->meta_tag] = entry_node->node_data.bmp_tag;

		entry_node->node_data.flow_action_entry =
			dpdk_pipe_entry_add_with_cb(pipe_queue, 0 /* priority */,
					    lpm_priv->curr_tree->action_pipe,
					    &match, NULL/*match_mask*/,
					    &entry_node->node_data.actions, NULL /* action_descs */,
					    NULL /* mon */, &entry_node->node_data.fwd, lpm_add_entry_completion_cb,
					    DOCA_FLOW_WAIT_FOR_BATCH, NULL /* usr_ctx */);

		if (entry_node->node_data.flow_action_entry == NULL)
			return -1;
	}
	fwd.type = DOCA_FLOW_FWD_PIPE;
	if (node->right)
		fwd.next_pipe = (entry_node->node_data.jump_table_node == NULL ? node->right->node_data.pipe:entry_node->node_data.jump_table_node->node_data.pipe);
	else
		fwd.next_pipe = lpm_priv->curr_tree->action_pipe;

	memset(&match, 0, sizeof(struct doca_flow_match));
	lpm_fill_match(&match, lpm_get_addr_masked(entry_node), lpm_priv);

	entry_node->node_data.flow_pipe_entry =
		dpdk_pipe_entry_add_with_cb(pipe_queue, 0 /* priority */,
					    node->node_data.pipe,
					    &match, NULL, /* match_mask */
					    pactions, NULL, /* action_descs */
					    NULL /* mon */, &fwd, lpm_add_entry_completion_cb,
					    DOCA_FLOW_WAIT_FOR_BATCH, NULL /* usr_ctx */);

	if (entry_node->node_data.flow_pipe_entry == NULL)
		return -1;

	return 0;
}

static int
lpm_add_entry_to_pipe(uint16_t pipe_queue, struct lpm_tree_entry_node *entry_node,
				struct lpm_tree_table_node *node,
				struct lpm_priv_s *lpm_priv)
{
	if (entry_node == NULL)
		return 0;

	if (lpm_add_current_entry_to_pipe(pipe_queue, entry_node, node, lpm_priv) == -1)
		return -1;

	if (lpm_add_entry_to_pipe(pipe_queue, entry_node->left, node, lpm_priv) == -1)
		return -1;
	if (lpm_add_entry_to_pipe(pipe_queue, entry_node->right, node, lpm_priv) == -1)
		return -1;

	return 0;
}

static int
lpm_add_entries_to_pipe(uint16_t pipe_queue, struct lpm_tree_table_node *node, struct lpm_priv_s *lpm_priv)
{
	if (node == NULL)
		return 0;

	if (lpm_add_entry_to_pipe(pipe_queue, lpm_tree_table_get_root_entry(node), node, lpm_priv) == -1)
		return -1;

	if (lpm_add_entries_to_pipe(pipe_queue, node->left, lpm_priv) == -1)
		return -1;
	if (lpm_add_entries_to_pipe(pipe_queue, node->right, lpm_priv) == -1)
		return -1;

	return 0;
}

static int
lpm_fwd_traffic_to_root_pipe(uint16_t pipe_queue, struct lpm_priv_s *lpm_priv)
{
	struct doca_flow_match match = {0};
	struct doca_flow_actions actions = {0};
	struct doca_flow_fwd fwd = {0};
	struct doca_flow_pipe_entry *entry;

	fwd.type = DOCA_FLOW_FWD_PIPE;
	fwd.next_pipe = lpm_priv->curr_tree->lpm_tree_root->node_data.pipe;

	actions.meta.u32[lpm_priv->meta_tag] = 0;

	entry = dpdk_pipe_entry_add_with_cb(pipe_queue,
			    0 /*priority*/,
			    lpm_priv->dispatcher_pipe,
			    &match, NULL /*match_mask*/, &actions, NULL /* action_descs */,
			    NULL /* mon */, &fwd, lpm_add_entry_completion_cb,
			    DOCA_FLOW_WAIT_FOR_BATCH, NULL /* usr_ctx */);
	if (entry == NULL)
		return -1;

	return 0;
}

/*create pipe for each node  add entries for each type with actions and fwd*/
static int
lpm_offload_pipe_entries(uint16_t pipe_queue, struct lpm_priv_s *lpm_priv)
{
	/* create pipe for each node with miss fwd rule */
	if (lpm_create_tree_pipes(lpm_priv->curr_tree->lpm_tree_root, lpm_priv) < 0)
		return -1;

	/* add entries for each node pipe +  fwd rule + action for each bmp entry  */
	if (lpm_add_entries_to_pipe(pipe_queue, lpm_priv->curr_tree->lpm_tree_root, lpm_priv) < 0)
		return -1;

	// add entry to fwd to the main lpm pipe to lpm_priv.curr_tree pipe
	if (lpm_fwd_traffic_to_root_pipe(pipe_queue, lpm_priv) < 0)
		return -1;

	return 0;
}

static int
lpm_build_tree(struct lpm_priv_s *lpm_priv)
{
	lpm_set_parents(lpm_priv->curr_tree->lpm_tree_root);

	lpm_add_marks(lpm_priv->curr_tree->lpm_tree_root, lpm_priv);

	lpm_update_bmp(lpm_priv, lpm_priv->curr_tree->lpm_tree_root);

	return true;
}

int
lpm_flush(uint16_t pipe_queue,
		struct doca_flow_pipe *pipe,
		enum lpm_oper_type oper_type)
{
	struct lpm_priv_s *lpm_priv = pipe->private_ctx;
	int rc;

	rte_spinlock_lock(&lpm_priv->lock);

	lpm_priv->state = LPM_STATE_FLUSH_START;

	if (oper_type == LPM_BUILD_ONLY || oper_type == LPM_PUSH)
		lpm_build_tree(lpm_priv);

	// add HW rules
	if (oper_type == LPM_OFFLOAD_ONLY || oper_type == LPM_PUSH) {

		rc = dpdk_flow_entries_process(&lpm_priv->port, pipe_queue, LPM_DEFAULT_TIMEOUT_US, 0);
		if (rc < 0) {
			DOCA_LOG_RATE_LIMIT_ERR("%s failed - could not deplete queue %d rc=%d",
						__func__, pipe_queue, rc);
			return rc;
		}

		lpm_offload_pipe_entries(pipe_queue, lpm_priv);

		rc = dpdk_flow_entries_process(&lpm_priv->port, pipe_queue, LPM_DEFAULT_TIMEOUT_US, 0);
		if (rc < 0) {
			DOCA_LOG_RATE_LIMIT_ERR("%s failed - could not deplete queue %d rc=%d",
						__func__, pipe_queue, rc);
			return rc;
		}

		if (lpm_priv->tree_to_release) {
			lpm_destoy_tree(lpm_priv, lpm_priv->tree_to_release);
			lpm_priv->tree_to_release = NULL;
		}
		lpm_priv->state = LPM_STATE_FLUSH_DONE;
	}

	rte_spinlock_unlock(&lpm_priv->lock);

	return 0;
}

void
get_number_of_compiled_entry_nodes(struct lpm_tree_entry_node *node, uint32_t *n_entry_rules)
{
	if (node == NULL)
		return;
	get_number_of_compiled_entry_nodes(node->left, n_entry_rules);
	get_number_of_compiled_entry_nodes(node->right, n_entry_rules);

	if (node->node_data.entry_type == LPM_ENTRY_BMP_ORIG_TYPE ||
		node->node_data.entry_type == LPM_ENTRY_BMP_UPDATED_TYPE) {
		// one rule for the node pipe, one rule for the action pipe
		*n_entry_rules += 2;
	}
	if (node->node_data.entry_type == LPM_ENTRY_MARKED_TYPE)
		*n_entry_rules += 1;  // one rule for the node pipe
}


void
get_number_of_compiled_tree_nodes(struct lpm_tree_table_node *node, uint8_t *n_tree_rules, uint32_t *n_entry_rules)
{
	if (node == NULL)
		return;
	get_number_of_compiled_tree_nodes(node->left, n_tree_rules, n_entry_rules);
	get_number_of_compiled_tree_nodes(node->right, n_tree_rules, n_entry_rules);

	get_number_of_compiled_entry_nodes(lpm_tree_table_get_root_entry(node), n_entry_rules);

	(*n_tree_rules)++;
}

uint32_t
lpm_get_number_of_built_rule(struct doca_flow_pipe *pipe)
{
	struct lpm_priv_s *lpm_priv = (struct lpm_priv_s *)pipe->private_ctx;
	uint8_t n_tree_rules = 0;
	uint32_t n_entry_rules = 0;
	uint32_t n_compiled_rules;

	get_number_of_compiled_tree_nodes(lpm_priv->curr_tree->lpm_tree_root, &n_tree_rules, &n_entry_rules);

	n_compiled_rules = n_tree_rules + // one rule for each created pipe ( fwd_miss )
			n_entry_rules + // two rule for BMP entry and one rule for Marked entry
			1; // one rule for forwarding traffic to the root tree table
	return n_compiled_rules;
}

static void
lpm_tree_lookup(struct lpm_priv_s *lpm_priv,
		struct lpm_tree_table_node *node,
		uint8_t *value, uint8_t *table_num, int *nr_steps, uint32_t *bmp)
{
	if (node == NULL)
		return;
	uint8_t addr_mask[LPM_MAX_ADDR_LEN] = {0};
	struct lpm_tree_entry_node *entry;
	struct lpm_tree_entry_node *root_entry;

	root_entry = lpm_tree_table_get_root_entry(node);
	if (root_entry == NULL)
		return;

	(*nr_steps)++;

	memcpy(addr_mask, value, lpm_priv->len_addr);
	utl_field_mask_invoke(addr_mask, lpm_priv->len_addr, node->node_data.cidr_bits);
	entry = lpm_get_entry_node_by_value(root_entry, addr_mask, lpm_tree_entry_get_bytes_in_depth(root_entry));
	if (entry == NULL)
		lpm_tree_lookup(lpm_priv, node->left, value, table_num, nr_steps, bmp);
	else {
		if (entry->node_data.entry_type == LPM_ENTRY_BMP_ORIG_TYPE ||
			entry->node_data.entry_type == LPM_ENTRY_BMP_UPDATED_TYPE) {
			*bmp = entry->node_data.bmp_tag;
			*table_num = node->node_data.cidr_bits;
		}
		lpm_tree_lookup(lpm_priv, node->right, value, table_num, nr_steps, bmp);
	}
}

int
lpm_lookup(struct doca_flow_pipe *pipe,	uint8_t *address)
{
	struct lpm_priv_s *lpm_priv = pipe->private_ctx;
	uint8_t table_num = 0;
	int nr_steps = 0;
	uint32_t bmp = 0;

	lpm_tree_lookup(lpm_priv, lpm_priv->curr_tree->lpm_tree_root, address, &table_num, &nr_steps, &bmp);
	nr_steps--;

	if (bmp > 0) {
		DOCA_LOG_INFO("rule number with BMP %u found in table %u for %d steps", bmp, table_num, nr_steps);
		return 0;
	}
	DOCA_LOG_INFO("rule number is not found in the tree");
	return -1;
}
