/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdint.h>
#include <stdlib.h>

#include <rte_malloc.h>
#include <rte_lcore.h>
#include <rte_mempool.h>

#include <doca_log.h>

#include "priv_doca_mempool.h"

#define MAX_BUFF_SIZE (256)
DOCA_LOG_REGISTER(priv_doca_mempool);

struct priv_doca_mempool {
	struct priv_doca_mempool_cfg cfg;
	void *data;
	void *cache[0];
};

static void
entry_init(__rte_unused struct rte_mempool *pool,
		void *arg, void *addr, unsigned int n)
{
	struct priv_doca_mempool_cfg *cfg = arg;

	cfg->usr_init_cb(cfg->usr_ctx, addr, n);
}

struct priv_doca_mempool *
priv_doca_mempool_create(struct priv_doca_mempool_cfg *cfg)
{
	struct priv_doca_mempool *pool;
	char name[MAX_BUFF_SIZE];
	int i;

	if (unlikely(cfg == NULL || cfg->name == NULL ||
			cfg->socket_id >= rte_socket_count())) {
		if (cfg == NULL)
			DOCA_LOG_ERR("mempool cfg is NULL");
		else
			DOCA_LOG_ERR("mempool cfg name %s, socket_id %u, nb_caches %u",
					cfg->name ? cfg->name : "NULL",
					cfg->socket_id, cfg->nb_caches);
		return NULL;
	}
	pool = (struct priv_doca_mempool *)rte_zmalloc(cfg->name,
		sizeof(struct priv_doca_mempool) + sizeof(void *) *
		cfg->nb_caches, RTE_CACHE_LINE_SIZE);
	if (pool == NULL) {
		DOCA_LOG_ERR("Failed to allocate mempool, name %s", cfg->name);
		return NULL;
	}
	pool->cfg = *cfg;
	pool->cfg.name = rte_zmalloc(NULL, strlen(cfg->name) + 1,
			RTE_CACHE_LINE_SIZE);
	if (pool->cfg.name == NULL) {
		DOCA_LOG_ERR("Failed to allocate mempool cfg, name %s", cfg->name);
		goto err;
	}
	snprintf(pool->cfg.name, strlen(cfg->name) + 1, "%s", cfg->name);
	snprintf(name, sizeof(name), "%s_%d", cfg->name, cfg->socket_id);
	pool->data = rte_mempool_create(name, cfg->nb_entries,
			cfg->size, 0, 0, NULL, NULL, cfg->usr_init_cb ? entry_init :
			NULL, &pool->cfg, cfg->socket_id, 0);
	if (pool->data == NULL) {
		DOCA_LOG_ERR("Failed to allocate mempool memory, name %s", cfg->name);
		goto err;
	}
	for (i = 0; i < cfg->nb_caches; i++) {
		pool->cache[i] = rte_mempool_cache_create(
				RTE_MEMPOOL_CACHE_MAX_SIZE, cfg->socket_id);
		if (pool->cache[i] == NULL) {
			DOCA_LOG_ERR("Failed to allocate mempool cache, name %s index %d",
					cfg->name, i);
			goto err;
		}
	}
	return pool;
err:
	priv_doca_mempool_destroy(pool);
	return NULL;
}

void *
priv_doca_mempool_alloc(struct priv_doca_mempool *pool, uint16_t cache_idx)
{
	void *obj;

	if (unlikely(pool == NULL || pool->data == NULL ||
	    cache_idx >= pool->cfg.nb_caches || pool->cache[cache_idx] == NULL)) {
		DOCA_LOG_ERR("Failed to allocate memory, pool/data/cache is NULL");
		return NULL;
	}
	if (!rte_mempool_generic_get(pool->data, &obj, 1,
				pool->cache[cache_idx]))
		return obj;
	return NULL;
}

void
priv_doca_mempool_free(struct priv_doca_mempool *pool, void *addr, uint16_t cache_idx)
{
	if (unlikely(pool == NULL || addr == NULL || pool->data == NULL ||
	    cache_idx >= pool->cfg.nb_caches || pool->cache[cache_idx] == NULL)) {
		DOCA_LOG_ERR("Failed to free memory, pool/addr/data/cache is NULL");
		return;
	}
	rte_mempool_generic_put(pool->data, &addr, 1, pool->cache[cache_idx]);
}

void
priv_doca_mempool_destroy(struct priv_doca_mempool *pool)
{
	int i = 0;

	if (pool == NULL)
		return;
	for (i = 0; i < pool->cfg.nb_caches; i++)
		if (pool->cache[i] != NULL) {
			if (pool->data != NULL)
				rte_mempool_cache_flush(pool->cache[i], pool->data);
			rte_mempool_cache_free(pool->cache[i]);
		}
	if (pool->data != NULL)
		rte_mempool_free(pool->data);
	if (pool->cfg.name != NULL)
		rte_free(pool->cfg.name);
	rte_free(pool);
}
