/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef DOCA_MEMPOOL_H_
#define DOCA_MEMPOOL_H_

#include <stdint.h>
#include <stdbool.h>

#ifdef __cplusplus
extern "C" {
#endif

#define PRIV_MEMPOOL_NAME_LEN (32)

typedef void (*obj_init_t)(void *usr_ctx, void *obj,
		unsigned int obj_idx);

/**
 *  DOCA memory pool configuration structure.
 */
struct priv_doca_mempool_cfg {
	uint32_t size; /**< Entry size. */
	uint32_t nb_entries; /**< Entry number. */
	uint16_t nb_caches; /**< Cache number. */
	uint8_t socket_id; /**< The numa node. */
	obj_init_t usr_init_cb; /**< Predefined callback to init each entry. */
	void *usr_ctx; /**< Upper layer handle. */
	char *name; /**< Pool name. */
};

/**
 * @brief create a memory pool
 *
 * Create memory pool using specific configuration.
 *
 * @param cfg
 * Pointer to configuration.
 * @return
 * Memory pool handler, NULL if fail.
 * @note
 * Will create a pool for the specific NUMA node.
 */
struct priv_doca_mempool *priv_doca_mempool_create(struct priv_doca_mempool_cfg *cfg);

/**
 * @brief Destroy a pool.
 *
 * @param pool
 * Memory pool to be destroyed.
 */
void priv_doca_mempool_destroy(struct priv_doca_mempool *pool);

/**
 * @brief allocate an entry
 *
 * Allocate an entry from pool.
 *
 * @param pool
 * Memory pool handler where to allocate.
 * @param cache_idx
 * Index of local cache which is lock free.
 * @return
 * Entry address, NULL if fail
 * @note
 * Not thread safe
 */
void *priv_doca_mempool_alloc(struct priv_doca_mempool *pool, uint16_t cache_idx);

/**
 * @brief Free an entry.
 *
 * Free an entry identified by address.
 *
 * @param pool
 * Pointer to memory pool.
 * @param addr
 * The address to be free.
 * @param cache_idx
 * Index of local cache which is lock free.
 * @note
 * Not thread safe
 */
void priv_doca_mempool_free(struct priv_doca_mempool *pool, void *addr, uint16_t cache_idx);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* DOCA_MEMPOOL_H_ */
