/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <stdlib.h>
#include <errno.h>

#include <private/common_defines.h>
#include <doca_log.h>

#include "utils_bitmap.h"

#define BITS_IN_BYTE_SHIFT	(3)
#define BITS_IN_BYTE		(1 << BITS_IN_BYTE_SHIFT)
#define BITS_IN_BYTE_MASK	(BITS_IN_BYTE - 1)

DOCA_LOG_REGISTER(utils_bitmap);


struct utils_bitmap {
	uint32_t nr_bits;
	uint8_t bitmap_array[0];
};

static inline uint8_t
get_bit_offset_in_byte(uint32_t bit)
{
	return bit & BITS_IN_BYTE_MASK;
}

static inline void
get_bit_offset(uint32_t bit, uint32_t *byte_offset, uint8_t *bit_offset_in_byte)
{
	*byte_offset = bit >> BITS_IN_BYTE_SHIFT;
	*bit_offset_in_byte = 1 << get_bit_offset_in_byte(bit);
}

static uint32_t
calc_needed_bytes(uint32_t nr_bits)
{
	uint32_t nb_bytes = nr_bits >> BITS_IN_BYTE_SHIFT;

	if (get_bit_offset_in_byte(nr_bits) > 0)
		nb_bytes++;

	return nb_bytes;
}

struct utils_bitmap *
utils_bitmap_create(uint32_t nr_bits)
{
	struct utils_bitmap *bitmap;
	uint32_t nb_bytes = calc_needed_bytes(nr_bits);
	size_t size_alloc = sizeof(struct utils_bitmap) + nb_bytes * sizeof(uint8_t);

	bitmap = calloc(1, size_alloc);
	if (bitmap == NULL) {
		DOCA_LOG_ERR("failed to allocate memory for bitmap of %u bits", nr_bits);
		return NULL;
	}

	bitmap->nr_bits = nr_bits;

	return bitmap;
}

void
utils_bitmap_destroy(struct utils_bitmap *bitmap)
{
	if (unlikely(bitmap == NULL))
		return;

	free(bitmap);
}

int
utils_bitmap_get(struct utils_bitmap *bitmap, uint32_t bit)
{
	uint32_t idx;
	uint8_t offset;

	if (unlikely(bitmap == NULL || bit >= bitmap->nr_bits))
		return -EINVAL;

	get_bit_offset(bit, &idx, &offset);

	return (bitmap->bitmap_array[idx] & offset) ? 1 : 0;
}

int
utils_bitmap_set(struct utils_bitmap *bitmap, uint32_t bit, bool on)
{
	uint32_t idx;
	uint8_t offset;

	if (unlikely(bitmap == NULL || bit >= bitmap->nr_bits))
		return -EINVAL;

	get_bit_offset(bit, &idx, &offset);

	if (on)
		bitmap->bitmap_array[idx] |= offset;
	else
		bitmap->bitmap_array[idx] &= ~offset;

	return 0;
}
