/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <rte_hash.h>
#include <rte_hash_crc.h>
#include <rte_spinlock.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <doca_log.h>
#include <inttypes.h>
#include <private/common_defines.h>
#include "utils_hash_table.h"

DOCA_LOG_REGISTER(utils_hash_table);

struct key_value_context {
	void **key;
	void *value;
	bool found;
};

struct hash_table_entry {
	uint32_t refcnt;
	void *key;
};

struct utils_hash_table {
	int size;
	rte_spinlock_t lock;
	struct rte_hash *h;
	size_t key_len;
	void (*free_key_cb)(void *key);
	void (*table_lock)(rte_spinlock_t *lock);
	void (*table_unlock)(rte_spinlock_t *lock);
	struct hash_table_entry entries[0];
};

static void
table_lock_dummy(rte_spinlock_t *lock)
{
}

static void
table_unlock_dummy(rte_spinlock_t *lock)
{
}

static void
table_lock(rte_spinlock_t *lock)
{
	rte_spinlock_lock(lock);
}

static void
table_unlock(rte_spinlock_t *lock)
{
	rte_spinlock_unlock(lock);
}

struct utils_hash_table *
utils_hash_table_create(struct utils_hash_table_cfg *cfg)
{
	struct rte_hash_parameters hash_table_params;
	struct utils_hash_table *inst;
	int total_size;

	if (unlikely(cfg == NULL)) {
		DOCA_LOG_ERR("failed creating hash table - cfg is NULL");
		return NULL;
	}
	memset(&hash_table_params, 0x0, sizeof(hash_table_params));
	if (unlikely(cfg->name == NULL)) {
		DOCA_DLOG_DBG("hash table name is NULL, fallback to \"utils_hash\"");
		hash_table_params.name = "utils_hash";
	} else {
		hash_table_params.name = cfg->name;
	}
	total_size = sizeof(struct utils_hash_table) + sizeof(struct hash_table_entry) * cfg->table_size;
	inst = (struct utils_hash_table *)calloc(1, total_size);
	if (inst == NULL) {
		DOCA_LOG_ERR("failed creating hash table \"%s\" - insufficient memory", hash_table_params.name);
		return NULL;
	}

	hash_table_params.entries = cfg->table_size;
	hash_table_params.key_len = cfg->key_len;
	hash_table_params.hash_func_init_val = 0;
	hash_table_params.hash_func = cfg->hash_func ? cfg->hash_func : rte_hash_crc;
	inst->size = cfg->table_size;
	inst->key_len = cfg->key_len;
	inst->free_key_cb = cfg->free_key_cb;

	inst->h = rte_hash_create(&hash_table_params);
	if (inst->h == NULL) {
		DOCA_LOG_ERR("failed to alloc dpdk hash table %s with %u entries",
		hash_table_params.name,
		hash_table_params.entries);
		free(inst);
		return NULL;
	}

	if (cfg->cmp_eq != NULL)
		rte_hash_set_cmp_func(inst->h, cfg->cmp_eq);

	inst->table_lock = &table_lock_dummy;
	inst->table_unlock = &table_unlock_dummy;
	if (cfg->thread_safe) {
		inst->table_lock = &table_lock;
		inst->table_unlock = &table_unlock;
		rte_spinlock_init(&inst->lock);
	}
	DOCA_LOG_INFO("hash table %s created", hash_table_params.name);
	return inst;
}

void
utils_hash_table_destroy(struct utils_hash_table *t)
{
	const void *next_key;
	uint32_t iter = 0;
	void *next_data;
	int ret;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to destroy hash table - null pointer");
		return;
	}

	if (t->free_key_cb) {
		ret = rte_hash_iterate(t->h, &next_key, &next_data, &iter);
		while (ret >= 0) {
			if (next_key != NULL)
				t->free_key_cb((void *)next_key);
			ret = rte_hash_iterate(t->h, &next_key, &next_data, &iter);
		}
	}
	if (t->h)
		rte_hash_free(t->h);

	DOCA_LOG_INFO("hash table destroyed");
	free(t);
}

int
utils_hash_table_get_count(struct utils_hash_table *t)
{
	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to get count of hash table - null pointer");
		return -1;
	}
	return rte_hash_count(t->h);
}

int
utils_hash_table_iterate(struct utils_hash_table *t, utils_hash_table_iter_fn iter_fn,
			 void *usr_ctx)
{
	uint32_t next = 0;
	const void *key;
	void *data;
	int ret;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to iterate hash table - null pointer");
		return -1;
	}

	if (unlikely(iter_fn == NULL))
		return -EINVAL;

	t->table_lock(&t->lock);
	ret = rte_hash_iterate(t->h, &key, &data, &next);
	while (ret >= 0) {
		if (key != NULL) {
			t->table_unlock(&t->lock);
			ret = iter_fn(key, data, ret, usr_ctx);
			switch (ret) {
			case -ITERATE_ERR:
				return -ENOENT;
			case -ITERATE_STOP:
				return 0;
			default:
				break;
			}
			t->table_lock(&t->lock);
		}
		ret = rte_hash_iterate(t->h, &key, &data, &next);
	}
	t->table_unlock(&t->lock);
	return 0;
}

static void *
key_alloc(size_t key_len)
{
	return malloc(key_len);
}

static void
key_free(void *key)
{
	free(key);
}

static inline int
add_key(struct utils_hash_table *t, void *key)
{
	int id;

	id = rte_hash_add_key(t->h, key);
	if (id < 0 || id > t->size) {
		DOCA_LOG_RATE_LIMIT_ERR("failed getting id for key %p", key);
		return -ENOENT;
	}

	t->entries[id].refcnt = 1;
	t->entries[id].key = key_alloc(t->key_len);
	if (t->entries[id].key == NULL) {
		DOCA_LOG_RATE_LIMIT_ERR("memory depleted for id %d", id);
		return -ENOMEM;
	}
	memcpy(t->entries[id].key, key, t->key_len);
	return id;
}

int
utils_hash_table_lookup_key(struct utils_hash_table *t, void *key)
{
	int id;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to lookup hash table key - null pointer");
		return -1;
	}
	t->table_lock(&t->lock);
	id = rte_hash_lookup(t->h, key);
	t->table_unlock(&t->lock);
	return id;
}

int
utils_hash_table_get_id(struct utils_hash_table *t, void *key, uint32_t *refcnt)
{
	int id = 0;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to get id - null pointer");
		return -1;
	}
	id = utils_hash_table_lookup_key(t, key);
	t->table_lock(&t->lock);
	if (id >= 0) {
		t->entries[id].refcnt++;
		*refcnt = t->entries[id].refcnt;
		DOCA_DLOG_DBG("found id %d refcnt is %"PRIu32"", id, *refcnt);
		t->table_unlock(&t->lock);
		return id;
	}
	id = add_key(t, key);
	if (id < 0)
		goto out;
	*refcnt = t->entries[id].refcnt;
out:
	t->table_unlock(&t->lock);
	return id;
}

int
utils_hash_table_put_id(struct utils_hash_table *t, int id)
{
	if (id < 0 || id > t->size) {
		DOCA_LOG_RATE_LIMIT_ERR("cannot delete invalid id %d", id);
		return -EINVAL;
	}
	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to put id - null pointer");
		return -1;
	}

	t->table_lock(&t->lock);
	DOCA_DLOG_DBG("put id %d refcnt is %"PRIu32"", id, t->entries[id].refcnt);
	if (t->entries[id].refcnt > 0) {
		if (--t->entries[id].refcnt) {
			t->table_unlock(&t->lock);
			return t->entries[id].refcnt;
		}
		rte_hash_del_key(t->h, t->entries[id].key);
		if (t->free_key_cb)
			t->free_key_cb(t->entries[id].key);
		key_free(t->entries[id].key);
		t->entries[id].key = NULL;
		DOCA_DLOG_DBG("freed id %d refcnt is %"PRIu32"", id, t->entries[id].refcnt);
	}

	t->table_unlock(&t->lock);
	return 0;
}

int
utils_hash_table_get_value_by_key(struct utils_hash_table *t, void *key, void **value)
{
	int id = 0;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to get value by key - null pointer");
		return -1;
	}
	t->table_lock(&t->lock);
	id = rte_hash_lookup_data(t->h, key, value);
	if (id < 0 || id > t->size) {
		DOCA_DLOG_DBG("key %p was not found, lookup returned %d", key, id);
		t->table_unlock(&t->lock);
		return -ENOENT;
	}
	t->entries[id].refcnt++;
	t->table_unlock(&t->lock);
	return 0;
}

static int
get_key_by_value_iter(const void *key, const void *value, int id, void *usr_ctx)
{
	struct key_value_context *ctx = usr_ctx;

	if (ctx->value == value) {
		*ctx->key = (void *)key;
		ctx->found = true;
		return -ITERATE_STOP;
	}
	return -ITERATE_CONT;
}

int
utils_hash_table_get_key_by_value(struct utils_hash_table *t, void *value, void **key)
{
	struct key_value_context ctx;
	int ret;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to get key by value - null pointer");
		return -1;
	}
	ctx.value = value;
	ctx.key = key;
	ctx.found = false;
	ret = utils_hash_table_iterate(t, get_key_by_value_iter, &ctx);
	if (ret || !ctx.found)
		return -ENOENT;
	return 0;
}

int
utils_hash_table_key_set_value(struct utils_hash_table *t, void *key, void *value)
{
	int ret;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to set key value - null pointer");
		return -1;
	}
	t->table_lock(&t->lock);
	ret = add_key(t, key);
	if (ret < 0)
		goto out;
	ret = rte_hash_add_key_data(t->h, key, value);
out:
	t->table_unlock(&t->lock);
	return ret;
}

static int
put_value_iter(const void *key, const void *value, int id, void *usr_ctx)
{
	struct ctx {
		struct utils_hash_table *t;
		void *value;
		uint32_t refcnt;
	} ctx;
	int ret = 0;

	ctx.value = ((struct ctx *)usr_ctx)->value;
	ctx.t = ((struct ctx *)usr_ctx)->t;
	if (value == ctx.value) {
		ret = utils_hash_table_put_id(ctx.t, id);
		if (ret < 0)
			return -ITERATE_ERR;
		((struct ctx *)usr_ctx)->refcnt = ret;
		return -ITERATE_STOP;
	}
	return -ITERATE_CONT;
}

int
utils_hash_table_put_value(struct utils_hash_table *t, void *value)
{
	struct ctx {
		struct utils_hash_table *t;
		void *value;
		uint32_t refcnt;
	} ctx;
	int ret;

	if (unlikely(t == NULL)) {
		DOCA_LOG_WARN("failed to put value - null pointer");
		return -1;
	}
	ctx.value = value;
	ctx.t = t;
	ctx.refcnt = 0;
	ret = utils_hash_table_iterate(t, put_value_iter, &ctx);
	if (ret)
		return -ENOENT;
	return ctx.refcnt;
}
