/*
 * Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#ifndef UTILS_HASH_TABLE_H_
#define UTILS_HASH_TABLE_H_

#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

struct utils_hash_table_cfg {
	uint32_t table_size;
	/** < Max number of entries in table */
	size_t key_len;
	/** < Length of key in hash table */
	const char *name;
	/** < Hash table name */
	bool thread_safe;
	/** < if set, then calls to functions will be protected by a lock */
	uint32_t (*hash_func)(const void *data, uint32_t data_len, uint32_t init_val);
	/* < Function pointer to a user-defined function for hash calculation, if
	 * set to NULL then default rte_crc_hash function will be used.
	 */
	void (*free_key_cb)(void *key);
	/** < Function pointer to be used upon freeing a key */
	int (*cmp_eq)(const void *key1, const void *key2, size_t key_len);
	/** < Function pointer to be used upon comparing keys, is set not NULL memcmp will be used */
};

/** < Possible return values for utils_hash_table_iter_fn */
enum utils_iter_return_types {
	ITERATE_DONE = 0,
	/** < Returned when itrator finished iterating over all entries */
	ITERATE_STOP,
	/** < indication for the iterator to stop iterating*/
	ITERATE_CONT,
	/** < indication for the iterator to continue iterating */
	ITERATE_ERR,
	/** < indication for the iterator to stop iterating due to error */
};

/** @brierf
 * Type of function used to iterate the hash table.
 *
 * @param key
 * a key from the hash table
 * @param value
 * a value corresponsing to a key
 * @param id
 * unique id for the specific @param key
 * @param usr_ctx
 * user defined context which will be passed with each iteration to the function
 *
 * @return
 * -ITERATE_STOP to indicate to the iterator to stop.
 * -ITERATE_ERR to indicate to the iterator to stop due to error.
 * -ITERATE_CONT to indicate to the iterator to continue.
 */
typedef int (*utils_hash_table_iter_fn)(const void *key, const void *value, int id, void *usr_ctx);

/**
 * @brief
 * Function to create a hash table
 *
 * @param cfg
 * utils hash table config
 *
 * @return
 * Pointer to instance of hash table on success, NULL on failure
 */
struct utils_hash_table *
utils_hash_table_create(struct utils_hash_table_cfg *cfg);

/**
 * @brief
 * Function to destroy a hash table
 *
 * @param t
 * Pointer to a hash table created using hash_table_init
 */
void
utils_hash_table_destroy(struct utils_hash_table *t);

/**
 * @brief
 * Function to return number of keys in the hash table
 *
 * @param t
 * Pointer to a hash table created using hash_table_init
 *
 * @return
 * number of keys in the hash table.
 */
int
utils_hash_table_get_count(struct utils_hash_table *t);

/**
 * @brief
 * Iterate through the hash table, and run @param iter function for each
 * key-value pair.
 *
 * @param t
 * Hash table to iterate
 * @param iter
 * pointer to struct utils_hash_table_iterator.
 * @param usr_ctx
 * user context to be passed to the iterator function.
 *
 * @return
 * 0 on success.
 * -ENOENT if iter_fn fails.
 * -EINVAL if parameters are invalid.
 */
int
utils_hash_table_iterate(struct utils_hash_table *t, utils_hash_table_iter_fn iter_fn,
			 void *usr_ctx);

/**
 * @brief
 * Lookup for a key in hash table
 *
 * @param t
 * Pointer to a hash table created using hash_table_init
 * @param key
 * Pointer to a key to lookup
 *
 * @return
 * positive value indicating a unique id for a given key if found.
 * -ENOENT otherwise.
 */
int
utils_hash_table_lookup_key(struct utils_hash_table *t, void *key);

/**
 * @brief
 * Allocates an id for a given @param key, if it already exists
 * increments @param refcnt.
 *
 * @param t
 * Pointer to a hash table created using hash_table_init
 * @param key
 * Pointer to a key in the hash table
 * @param refcnt
 * pointer to a user-managed refrence count which indicates how many times
 * function hash_table_get_id has been called with the same key.
 *
 * @return
 * positive value indicating a unique id for a given key.
 * -ENOENT if table is full.
 * -ENOMEM if can't allocate memory for key.
 */
int
utils_hash_table_get_id(struct utils_hash_table *t, void *key, uint32_t *refcnt);

/**
 * @brief
 * Frees an id for a given @param key, if it's refcount reaches 0 then removes
 * the key from the hash table
 *
 * @param t
 * Pointer to a hash table created using hash_table_init
 * @param id
 * Unique id which is returned by calling hash_table_get_id
 *
 * @return
 * refcount value, if returned 0 it can be assumed that the key has
 * been deleted from the hash table.
 * -EINVAL for invalid id.
 */
int
utils_hash_table_put_id(struct utils_hash_table *t, int id);

/**
 * @brief
 * Returns value for a given @param key.
 *
 * @param t
 * Pointer to a hash table created using hash_table_init
 * @param key
 * Pointer to a key in the hash table
 * @param value
 * Output parameter to store the value returned by the key
 *
 * @return
 * 0 on success
 * -ENOENT if no value is found.
 */
int
utils_hash_table_get_value_by_key(struct utils_hash_table *t, void *key, void **value);

/**
 * @brief
 * Returns key for a given @param value.
 *
 * @param t
 * Pointer to a hash table created using hash_table_init
 * @param value
 * Pointer to a value in the hash table
 * @param key
 * Output parameter to store the key returned by the value
 *
 * @return
 * 0 on success.
 * -ENOENT no key is found.
 */
int
utils_hash_table_get_key_by_value(struct utils_hash_table *t, void *value, void **key);

/**
 * @brief
 * sets a given key's in a hash table.
 * This function assumes the @param key was not inserted prior its call
 * This function inserts @param key to the hash table.
 *
 * @param t
 * Pointer to a hash table created using hash_table_init.
 * @param key
 * Pointer to a key to insert.
 * @param value
 * Pointer to value to insert.
 *
 * @return
 * 0 on success.
 * -ENOENT if end of the hash table.
 * -ENOMEM if can't allocate memory for key.
 */
int
utils_hash_table_key_set_value(struct utils_hash_table *t, void *key, void *value);

/**
 * @brief
 * Frees value in a hash table, if the value is no longer refrenced frees
 * its key.
 *
 * @param t
 * Pointer to a hash table created using hash_table_init.
 * @param value
 * Pointer to value to free.
 *
 * @return
 * positive number indicating how many times the value is refrenced on success.
 * -ENOENT if no value was found.
 */
int
utils_hash_table_put_value(struct utils_hash_table *t, void *value);

#ifdef __cplusplus
}
#endif

#endif /* UTILS_HASH_TABLE_H_ */
