/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include "private/common_defines.h"

#include "utils_df_translate.h"
#include "utils_field.h"

#include "doca_log.h"

#define DF_FIELD_IPV4_ADDR_LEN (4) /**< length of ipv4 address length. */
#define DF_FIELD_IPV6_ADDR_LEN (16) /**< length of ipv6 address length. */
#define DF_FIELD_VXLAN_VNI_ADDR_LEN (3) /**< length of vxlan vni address length. */
#define DF_FIELD_PORT_ID_LEN (2) /**< length of port id. */
#define DF_FIELD_RSS_NUMBER_LEN (4) /**< length of number of rss length. */
#define DF_FIELD_L4_TYPE_LEN (1) /**< length of L4 protocol type. */
#define DF_FIELD_SHARED_RESOURCE_ID_LEN (4) /**<  length of share coutner */
#define DF_FIELD_ORDERED_LIST_IDX_LEN (4) /**< length of an ordered list pipe index */
#define DF_HEADER_NISP_LEN DOCA_FLOW_NISP_HEADER_LEN /**< length of NISP protocol header. */
#define DF_HEADER_AUDP_LEN DOCA_FLOW_AUDP_HEADER_LEN /**< length of AUDP protocol header. */
#define DF_HEADER_ESP_LEN (8) /**<  length of esp header */

#define CHANGEABLE_BYTE (0xff)

DOCA_LOG_REGISTER(utils_df_translate);

bool
utils_df_translate_is_mac_ignored(const uint8_t *mac_addr)
{
	if (unlikely(!mac_addr)) {
		DOCA_LOG_ERR("failed determine if mac is ignored - address is null");
		return false;
	}

	return utils_field_get_type(mac_addr, DOCA_ETHER_ADDR_LEN) == UTILS_FIELD_TYPE_IGNORED;
}

bool
utils_df_translate_is_mac_changeable(const uint8_t *mac_addr)
{
	if (unlikely(!mac_addr)) {
		DOCA_LOG_ERR("failed determine if mac is ignored - address is null");
		return false;
	}

	return utils_field_get_type(mac_addr, DOCA_ETHER_ADDR_LEN) == UTILS_FIELD_TYPE_CHANGEABLE;
}

void
utils_df_translate_set_mac_changeable(uint8_t *mac_addr)
{
	if (unlikely(!mac_addr)) {
		DOCA_LOG_ERR("failed setting mac as changeable - address is null");
		return;
	}

	utils_field_set(mac_addr, DOCA_ETHER_ADDR_LEN, CHANGEABLE_BYTE);
}

void
utils_df_translate_copy_mac(const uint8_t *src_mac_addr, uint8_t *dst_mac_addr)
{
	if (unlikely(!src_mac_addr)) {
		DOCA_LOG_ERR("failed copying mac - src address is null");
		return;
	}

	if (unlikely(!dst_mac_addr)) {
		DOCA_LOG_ERR("failed copying mac - dst address is null");
		return;
	}

	utils_field_copy(dst_mac_addr, src_mac_addr, DOCA_ETHER_ADDR_LEN);
}

void
utils_df_translate_copy_mac_with_mask(const uint8_t *src_mac_addr, const uint8_t *mac_mask_addr, uint8_t *dst_mac_addr)
{
	int i;

	if (unlikely(!src_mac_addr)) {
		DOCA_LOG_ERR("failed copying mac with mask - src address is null");
		return;
	}

	if (unlikely(!mac_mask_addr)) {
		DOCA_LOG_ERR("failed copying mac with mask - mask address is null");
		return;
	}

	if (unlikely(!dst_mac_addr)) {
		DOCA_LOG_ERR("failed copying mac with mask - dst address is null");
		return;
	}

	for (i = 0; i < DOCA_ETHER_ADDR_LEN; i++)
		dst_mac_addr[i] = src_mac_addr[i] & mac_mask_addr[i];
}

static int
get_ip_params(const struct doca_flow_ip_addr *ip_addr, uint8_t **field_ptr, int *field_len)
{
	if (unlikely(!ip_addr)) {
		DOCA_LOG_ERR("failed getting ip params - ip address is null");
		return -EINVAL;
	}

	if (unlikely(!field_ptr || !field_len)) {
		DOCA_LOG_ERR("failed getting ip params - field ptr %p or field len %p null",
			     field_ptr, field_len);
		return -EINVAL;
	}

	switch (ip_addr->type) {
	case DOCA_FLOW_ADDR_NONE:
		*field_ptr = NULL;
		*field_len = 0;
		break;
	case DOCA_FLOW_IP4_ADDR:
		*field_ptr = (uint8_t *)&ip_addr->ipv4_addr;
		*field_len = DF_FIELD_IPV4_ADDR_LEN;
		break;
	case DOCA_FLOW_IP6_ADDR:
		*field_ptr = (uint8_t *)ip_addr->ipv6_addr;
		*field_len = DF_FIELD_IPV6_ADDR_LEN;
		break;
	default:
		DOCA_LOG_ERR("failed getting ip params - ip type %u is unsupported", ip_addr->type);
		return -EINVAL;
	}

	return 0;
}

bool
utils_df_translate_is_ip_ignored(const struct doca_flow_ip_addr *ip_addr)
{
	uint8_t *field_ptr;
	int field_len;
	int rc;

	rc = get_ip_params(ip_addr, &field_ptr, &field_len);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed determine if ip is ignored - params returned rc=%d", rc);
		return false;
	}

	if (!field_ptr)
		return true;

	return utils_field_get_type(field_ptr, field_len) == UTILS_FIELD_TYPE_IGNORED;
}

bool
utils_df_translate_is_ip_changeable(const struct doca_flow_ip_addr *ip_addr)
{
	uint8_t *field_ptr;
	int field_len;
	int rc;

	rc = get_ip_params(ip_addr, &field_ptr, &field_len);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed determine if ip is changeable - params returned rc=%d", rc);
		return false;
	}

	if (!field_ptr)
		return false;

	return utils_field_get_type(field_ptr, field_len) == UTILS_FIELD_TYPE_CHANGEABLE;
}

void
utils_df_translate_set_ip_changeable(struct doca_flow_ip_addr *ip_addr)
{
	uint8_t *field_ptr;
	int field_len;
	int rc;

	rc = get_ip_params(ip_addr, &field_ptr, &field_len);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed setting ip as changeable - params returned rc=%d", rc);
		return;
	}

	if (!field_ptr) {
		DOCA_LOG_ERR("failed setting ip as changeable - ip type must be set");
		return;
	}

	utils_field_set(field_ptr, field_len, CHANGEABLE_BYTE);
}

void
utils_df_translate_copy_ip(const struct doca_flow_ip_addr *src_ip_addr, uint8_t *dst_ptr)
{
	uint8_t *field_ptr;
	int field_len;
	int rc;

	if (unlikely(!dst_ptr)) {
		DOCA_LOG_ERR("failed copying ip - dst pointer is null");
		return;
	}

	rc = get_ip_params(src_ip_addr, &field_ptr, &field_len);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed copying ip - params returned rc=%d", rc);
		return;
	}

	utils_field_copy(dst_ptr, field_ptr, field_len);
}

void
utils_df_translate_copy_ip_with_mask(const struct doca_flow_ip_addr *src_ip_addr, const uint8_t *mask_ptr, uint8_t *dst_ptr)
{
	uint8_t *field_ptr;
	int field_len;
	int i;
	int rc;

	if (unlikely(!dst_ptr)) {
		DOCA_LOG_ERR("failed copying ip - dst pointer is null");
		return;
	}

	rc = get_ip_params(src_ip_addr, &field_ptr, &field_len);
	if (unlikely(rc)) {
		DOCA_LOG_ERR("failed copying ip - params returned rc=%d", rc);
		return;
	}

	for (i = 0; i < field_len; i++)
		dst_ptr[i] = field_ptr[i] & mask_ptr[i];
}

void
utils_df_translate_set_vni_changeable(void *vni_addr)
{
	if (unlikely(!vni_addr)) {
		DOCA_LOG_ERR("failed setting vni as changeable - address is null");
		return;
	}

	utils_field_set(vni_addr, DF_FIELD_VXLAN_VNI_ADDR_LEN, CHANGEABLE_BYTE);
}

bool
utils_df_translate_is_fwd_changeable(const struct doca_flow_fwd *fwd)
{
	if (unlikely(!fwd)) {
		DOCA_LOG_ERR("failed determine if fwd is changeable - address is null");
		return false;
	}
	if (fwd->type == DOCA_FLOW_FWD_PORT)
		return utils_field_get_type(&fwd->port_id, DF_FIELD_PORT_ID_LEN) == UTILS_FIELD_TYPE_CHANGEABLE;
	else if (fwd->type == DOCA_FLOW_FWD_RSS)
		return utils_field_get_type(&fwd->num_of_queues, DF_FIELD_RSS_NUMBER_LEN) == UTILS_FIELD_TYPE_CHANGEABLE;
	else if (fwd->type == DOCA_FLOW_FWD_ORDERED_LIST_PIPE)
		return utils_field_get_type(&fwd->ordered_list_pipe.idx, DF_FIELD_ORDERED_LIST_IDX_LEN) ==
				UTILS_FIELD_TYPE_CHANGEABLE;
	DOCA_LOG_RATE_LIMIT_WARN("set unsupport fwd type %u changeable is false", fwd->type);
	return false;
}

bool
utils_df_translate_is_l4_type_specific(uint8_t l4_type)
{
	return utils_field_get_type(&l4_type, DF_FIELD_L4_TYPE_LEN) == UTILS_FIELD_TYPE_SPECIFIC;
}

bool
utils_df_translate_is_nisp_ignored(const void *nisp)
{
	if (unlikely(!nisp)) {
		DOCA_LOG_ERR("failed determine if nisp header is ignored - address is null");
		return false;
	}

	return utils_field_get_type(nisp, DF_HEADER_NISP_LEN) == UTILS_FIELD_TYPE_IGNORED;
}

bool
utils_df_translate_is_nisp_changeable(const void *nisp)
{
	if (unlikely(!nisp)) {
		DOCA_LOG_ERR("failed determine if nisp header is changable - address is null");
		return false;
	}

	return utils_field_get_type(nisp, DF_HEADER_NISP_LEN) == UTILS_FIELD_TYPE_CHANGEABLE;
}

bool
utils_df_translate_is_audp_ignored(const void *audp)
{
	if (unlikely(!audp)) {
		DOCA_LOG_ERR("failed determine if audp header is ignored - address is null");
		return false;
	}

	return utils_field_get_type(audp, DF_HEADER_AUDP_LEN) == UTILS_FIELD_TYPE_IGNORED;
}

bool
utils_df_translate_is_audp_changeable(const void *audp)
{
	if (unlikely(!audp)) {
		DOCA_LOG_ERR("failed determine if audp header is changable - address is null");
		return false;
	}

	return utils_field_get_type(audp, DF_HEADER_AUDP_LEN) == UTILS_FIELD_TYPE_CHANGEABLE;
}

bool
utils_df_translate_is_esp_ignored(const void *esp)
{
	if (unlikely(!esp)) {
		DOCA_LOG_ERR("failed determine if esp header is ignored - address is null");
		return false;
	}

	return utils_field_get_type(esp, DF_HEADER_ESP_LEN) == UTILS_FIELD_TYPE_IGNORED;
}

bool
utils_df_translate_is_esp_changeable(const void *esp)
{
	if (unlikely(!esp)) {
		DOCA_LOG_ERR("failed determine if esp header is changable - address is null");
		return false;
	}

	return utils_field_get_type(esp, DF_HEADER_ESP_LEN) == UTILS_FIELD_TYPE_CHANGEABLE;
}


bool
utils_df_translate_is_shared_resource_id_specific(uint32_t shared_resource_id)
{
	return utils_field_get_type(&shared_resource_id, DF_FIELD_SHARED_RESOURCE_ID_LEN)
		== UTILS_FIELD_TYPE_SPECIFIC;
}

bool
utils_df_translate_is_shared_resource_id_ignored(uint32_t shared_resource_id)
{
	return utils_field_get_type(&shared_resource_id, DF_FIELD_SHARED_RESOURCE_ID_LEN)
		== UTILS_FIELD_TYPE_IGNORED;
}

bool
utils_df_translate_is_shared_resource_id_changeable(uint32_t shared_resource_id)
{
	return utils_field_get_type(&shared_resource_id, DF_FIELD_SHARED_RESOURCE_ID_LEN)
		== UTILS_FIELD_TYPE_CHANGEABLE;
}
