/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <doca_log.h>
#include <priv_doca_ctx.h>
#include <priv_doca_dev.h>
#include <priv_doca_malloc.h>
#include <multi_os_prm.h>
#include <private/common_defines.h>

#include <infiniband/mlx5dv.h>
#include <infiniband/verbs.h>

#include "doca_ipsec.h"
#include "ipsec_workq_ctx.h"

DOCA_LOG_REGISTER(DOCA_IPSEC);

#define IPSEC_MAX_WORKQ_NUM 16

struct doca_ipsec {
	struct doca_ctx base;
	struct doca_dev *dev;
	uint8_t nb_workq_attached;
};

static doca_error_t
ipsec_start(struct doca_ctx *ctx)
{
	struct doca_ipsec *ipsec_ctx = NULL;

	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ipsec_ctx = container_of(ctx, struct doca_ipsec, base);
	if (ipsec_ctx->dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	return DOCA_SUCCESS;
}

static doca_error_t
ipsec_stop(struct doca_ctx *ctx)
{
	struct doca_ipsec *ipsec_ctx = NULL;

	if (ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ipsec_ctx = container_of(ctx, struct doca_ipsec, base);
	if (ipsec_ctx->dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	if (ipsec_ctx->nb_workq_attached > 0)
		return DOCA_ERROR_IN_USE;
	return DOCA_SUCCESS;
}

static doca_error_t
ipsec_dev_add(struct doca_ctx *ctx, struct doca_dev *dev)
{
	doca_error_t ret = DOCA_SUCCESS;
	struct doca_ipsec *ipsec_ctx = NULL;

	ipsec_ctx = container_of(ctx, struct doca_ipsec, base);

	if (ipsec_ctx->dev != NULL) {
		if (ipsec_ctx->dev == dev)
			return DOCA_SUCCESS;
		else
			return DOCA_ERROR_IN_USE;
	}
	ipsec_ctx->dev = dev;
	return ret;
}

static doca_error_t
ipsec_dev_rm(struct doca_ctx *ctx, struct doca_dev *dev)
{
	struct doca_ipsec *ipsec_ctx = NULL;

	ipsec_ctx = container_of(ctx, struct doca_ipsec, base);

	if (ipsec_ctx->dev != NULL) {
		if (ipsec_ctx->dev == dev) {
			ipsec_ctx->dev = NULL;
			return DOCA_SUCCESS;
		}
	}
	return DOCA_ERROR_INVALID_VALUE;
}

static doca_error_t
ipsec_progress(union doca_data data_priv, struct doca_event *ev)
{
	struct ipsec_workq_ctx *workq_ctx = NULL;
	doca_error_t res;

	workq_ctx = (struct ipsec_workq_ctx *)(data_priv.ptr);
	res = ipsec_workq_progress(workq_ctx, ev);
	if (res == DOCA_SUCCESS)
		return res;
	if (res == DOCA_ERROR_INITIALIZATION)
		return DOCA_ERROR_AGAIN;
	ev->result.u64 = res;
	return DOCA_ERROR_IO_FAILED;
}

static doca_error_t
ipsec_workq_add(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data *data_priv)
{
	struct doca_ipsec *ipsec_ctx = NULL;
	struct ipsec_workq_ctx *workq_ctx = NULL;
	struct priv_doca_workq_progress_entry *poller = NULL;
	doca_error_t ret = DOCA_SUCCESS;

	ipsec_ctx = container_of(ctx, struct doca_ipsec, base);
	if (ipsec_ctx->dev == NULL)
		return DOCA_ERROR_NOT_PERMITTED;
	if ((ipsec_ctx->nb_workq_attached + 1) > IPSEC_MAX_WORKQ_NUM)
		return DOCA_ERROR_NOT_PERMITTED;

	ret = ipsec_workq_create(&workq_ctx, ipsec_ctx->dev);
	if (ret != DOCA_SUCCESS)
		return ret;

	poller = priv_doca_malloc(1, sizeof(*poller));
	if (poller == NULL) {
		ret = DOCA_ERROR_NO_MEMORY;
		goto err_poller_malloc;
	}

	poller->progress_cb = ipsec_progress;
	poller->data_priv.ptr = (void *)(workq_ctx);

	ret = priv_doca_workq_progress_entry_add(workq, poller);
	if (ret != DOCA_SUCCESS)
		goto err_clean_poller;

	data_priv->ptr = (void *)poller;
	ipsec_ctx->nb_workq_attached += 1;

	return DOCA_SUCCESS;
err_clean_poller:
	priv_doca_free(poller);
err_poller_malloc:
	ipsec_workq_ctx_destroy(workq_ctx);
	return ret;
}

static doca_error_t
ipsec_workq_rm(struct doca_ctx *ctx, struct doca_workq *workq, union doca_data data_priv)
{
	struct doca_ipsec *ipsec_ctx = NULL;
	struct ipsec_workq_ctx *workq_ctx = NULL;
	struct priv_doca_workq_progress_entry *poller = NULL;
	doca_error_t ret = DOCA_SUCCESS;

	ipsec_ctx = container_of(ctx, struct doca_ipsec, base);
	if (ipsec_ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	poller = (struct priv_doca_workq_progress_entry *)(data_priv.ptr);
	ret = priv_doca_workq_progress_entry_rm(workq, poller);
	if (ret != DOCA_SUCCESS)
		return ret;

	workq_ctx = (struct ipsec_workq_ctx *)(poller->data_priv.ptr);
	ipsec_workq_ctx_destroy(workq_ctx);
	ipsec_ctx->nb_workq_attached--;

	priv_doca_free(poller);
	poller = NULL;
	return DOCA_SUCCESS;
}

static doca_error_t
ipsec_job_submit(union doca_data data_priv, struct doca_job const *job)
{
	struct ipsec_workq_ctx *workq_ctx = NULL;
	struct priv_doca_workq_progress_entry *poller = NULL;

	poller = (struct priv_doca_workq_progress_entry *)(data_priv.ptr);
	workq_ctx = poller->data_priv.ptr;

	switch (job->type) {
	case DOCA_IPSEC_JOB_SA_CREATE:
	case DOCA_IPSEC_JOB_SA_DESTROY:
		return ipsec_workq_job_submit(workq_ctx, job);
	default:
		DOCA_DLOG_ERR("Unsupported ipsec job type [%d] is found", job->type);
		return DOCA_ERROR_INVALID_VALUE;
	}
}

static const struct priv_doca_ctx_ops ipsec_ops = {.start = ipsec_start,
						   .stop = ipsec_stop,
						   .dev_add = ipsec_dev_add,
						   .dev_rm = ipsec_dev_rm,
						   .workq_add = ipsec_workq_add,
						   .workq_rm = ipsec_workq_rm,
						   .job_submit = ipsec_job_submit};

doca_error_t
doca_ipsec_create(struct doca_ipsec **ipsec_ctx)
{
	struct doca_ipsec *ctx = NULL;
	doca_error_t ret = DOCA_SUCCESS;

	if (ipsec_ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	ctx = priv_doca_malloc(1, sizeof(struct doca_ipsec));
	if (ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	memset(ctx, 0, sizeof(struct doca_ipsec));

	ret = priv_doca_ctx_init(&(ctx->base), true);
	if (ret != DOCA_SUCCESS)
		goto err_priv_doca_ctx_init;

	ctx->base.ops = ipsec_ops;

	*ipsec_ctx = ctx;

	return DOCA_SUCCESS;

err_priv_doca_ctx_init:
	priv_doca_free(ctx);
	return ret;
}

doca_error_t
doca_ipsec_destroy(struct doca_ipsec *ipsec_ctx)
{
	if (ipsec_ctx == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	if (ipsec_ctx->nb_workq_attached > 0)
		return DOCA_ERROR_IN_USE;

	ipsec_ctx->nb_workq_attached = 0;

	ipsec_ctx->dev = NULL;

	priv_doca_ctx_destroy(&(ipsec_ctx->base));

	priv_doca_free(ipsec_ctx);
	return DOCA_SUCCESS;
}

struct doca_ctx *
doca_ipsec_as_ctx(struct doca_ipsec *ipsec_ctx)
{
	if (ipsec_ctx == NULL)
		return NULL;
	return &(ipsec_ctx->base);
}

doca_error_t
doca_ipsec_job_get_supported(struct doca_devinfo *devinfo,
	enum doca_ipsec_job_types job_type)
{
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	void *atts = MLX5_ADDR_OF(query_hca_cap_out, out, capability);
	doca_error_t ret;

	ret = priv_doca_devinfo_hca_caps_by_opmod_query(devinfo,
							MLX5_HCA_CAP_OP_MOD_GENERAL_DEVICE | MLX5_HCA_CAP_OPMOD_GET_CUR,
							out, sizeof(out));
	if (ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to query device capabilities. Maybe old FW?");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	uint32_t ipsec_offload = MLX5_GET(cmd_hca_cap, atts, ipsec_offload);

	switch (job_type) {
	case DOCA_IPSEC_JOB_SA_CREATE:
	case DOCA_IPSEC_JOB_SA_DESTROY:
		return ipsec_offload ? DOCA_SUCCESS : DOCA_ERROR_NOT_SUPPORTED;
	default:
		DOCA_DLOG_ERR("Unsupported ipsec enum [%d] is found", job_type);
		return DOCA_ERROR_INVALID_VALUE;
	}
}

doca_error_t
doca_ipsec_is_mode_supported(const struct doca_devinfo *devinfo,
	enum doca_ipsec_sa_mode mode)
{
	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	switch (mode) {
	case DOCA_IPSEC_SA_MODE_TUNNEL:
		return DOCA_SUCCESS;
	case DOCA_IPSEC_SA_MODE_TRANSPORT:
		return DOCA_ERROR_NOT_SUPPORTED;
	default:
		DOCA_DLOG_ERR("Unsupported ipsec enum [%d] is found", mode);
		return DOCA_ERROR_INVALID_VALUE;
	}
}

doca_error_t
doca_ipsec_is_offload_supported(const struct doca_devinfo *devinfo,
	enum doca_ipsec_sa_offload offload)
{
	uint32_t out[MLX5_ST_SZ_DW(query_hca_cap_out)] = {0};
	void *atts = MLX5_ADDR_OF(query_hca_cap_out, out, capability);
	uint32_t ret = 0;

	doca_error_t cap_ret = priv_doca_devinfo_hca_caps_by_opmod_query(devinfo,
							MLX5_HCA_CAP_OP_MOD_IPSEC | MLX5_HCA_CAP_OPMOD_GET_CUR,
							out, sizeof(out));

	if (cap_ret != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to query device capabilities. Maybe old FW?");
		return DOCA_ERROR_NOT_SUPPORTED;
	}

	switch (offload) {
	case DOCA_IPSEC_SA_OFFLOAD_FULL:
		ret = MLX5_GET(ipsec_cap, atts, ipsec_full_offload);
		break;
	case DOCA_IPSEC_SA_OFFLOAD_CRYPTO:
		ret = MLX5_GET(ipsec_cap, atts, ipsec_crypto_offload);
		break;
	default:
		DOCA_DLOG_ERR("Unsupported ipsec enum [%d] is found", offload);
		return DOCA_ERROR_INVALID_VALUE;
	}

	return ret ? DOCA_SUCCESS : DOCA_ERROR_NOT_SUPPORTED;
}

doca_error_t
doca_ipsec_is_protocol_supported(const struct doca_devinfo *devinfo,
	enum doca_ipsec_sa_protocol protocol)
{
	if (devinfo == NULL)
		return DOCA_ERROR_INVALID_VALUE;

	switch (protocol) {
	case DOCA_IPSEC_SA_PROTO_ESP:
		return DOCA_SUCCESS;
	case DOCA_IPSEC_SA_PROTO_AH:
		return DOCA_ERROR_NOT_SUPPORTED;
	default:
		DOCA_DLOG_ERR("Unsupported ipsec enum [%d] is found", protocol);
		return DOCA_ERROR_INVALID_VALUE;
	}
}

