/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <infiniband/mlx5dv.h>
#include <infiniband/verbs.h>
#include <multi_os_prm.h>
#include <stdint.h>
#include <stdio.h>

#include "doca_dev.h"
#include "doca_ipsec.h"
#include "doca_log.h"
#include "ipsec_impl.h"
#include "priv_doca_dev.h"

DOCA_LOG_REGISTER(IPSEC);

#define MLX5_GENERAL_OBJ_TYPE_IPSEC_OFFLOAD 0x13
#define MLX5_GENERAL_OBJECT_TYPES_ENCRYPTION_KEY 0xc
#define MLX5_GENERAL_OBJECT_TYPE_ENCRYPTION_KEY_TYPE_IPSEC 0x2

struct mlx5_ifc_encryption_key_obj_bits {
	u8 modify_field_select[0x40];

	u8 state[0x8];
	u8 reserved_at_48[0xC];
	u8 key_size[0x4];
	u8 has_keytag[0x1];
	u8 reserved_at_59[0x3];
	u8 key_purpose[0x4];

	u8 reserved_at_60[0x8];
	u8 pd[0x18];

	u8 reserved_at_80[0x180];
	u8 key[0x100];

	u8 reserved_at_300[0x500];
};

struct mlx5_ifc_create_encryption_key_in_bits {
	struct mlx5_ifc_general_obj_in_cmd_hdr_bits general_obj_in_cmd_hdr;
	struct mlx5_ifc_encryption_key_obj_bits encryption_key_object;
};

enum win_size {
	WIN_32BIT = 0x0,
	WIN_64BIT = 0x1,
	WIN_128BIT = 0x2,
	WIN_256BIT = 0x3,
};

struct mlx5dv_devx_obj *
op_encryption_key(struct doca_ipsec_sa *sa)
{
	uint32_t in[MLX5_ST_SZ_DW(create_encryption_key_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	struct ibv_context *ibv_ctx = priv_doca_dev_to_ibv_ctx(sa->dev);
	void *hdr = MLX5_ADDR_OF(create_encryption_key_in, in, general_obj_in_cmd_hdr);
	void *key_obj = MLX5_ADDR_OF(create_encryption_key_in, in, encryption_key_object);
	void *key_field = MLX5_ADDR_OF(encryption_key_obj, key_obj, key);
	void *raw_key = sa->attr.key.aes_gcm.raw_key;
	uint32_t pd_id = 0;
	uint32_t bytes_to_copy = 0, key_size = 0;

	priv_doca_dev_pd_id_get(sa->dev, &pd_id);

	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJECT_TYPES_ENCRYPTION_KEY);

	MLX5_SET(encryption_key_obj, key_obj, key_purpose, MLX5_GENERAL_OBJECT_TYPE_ENCRYPTION_KEY_TYPE_IPSEC);
	MLX5_SET(encryption_key_obj, key_obj, pd, pd_id);

	switch (sa->attr.key.type) {
	case DOCA_ENCRYPTION_KEY_AESGCM_128:
		key_size = 0;
		bytes_to_copy = 128 / 8;
		key_field += bytes_to_copy;
		break;
	case DOCA_ENCRYPTION_KEY_AESGCM_256:
		key_size = 1;
		bytes_to_copy = 256 / 8;
		break;
	default:
		DOCA_DLOG_ERR("Wrong type ipsec key type");
		return NULL;
	}

	memcpy(key_field, raw_key, bytes_to_copy);
	memcpy(sa->raw_key, raw_key, bytes_to_copy);

	MLX5_SET(encryption_key_obj, key_obj, key_size, key_size);

	sa->devx_key_obj = mlx5dv_devx_obj_create(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (sa->devx_key_obj == NULL) {
		DOCA_DLOG_ERR("Failed to create ipsec key");
		return NULL;
	}
	sa->key_id = MLX5_GET(general_obj_out_cmd_hdr, out, obj_id);

	return sa->devx_key_obj;
}

struct mlx5dv_devx_obj *
create_encryption_key(struct doca_ipsec_sa *sa)
{
	return op_encryption_key(sa);
}

doca_error_t
sa_destroy(const struct doca_ipsec_sa *sa)
{
	return mlx5dv_devx_obj_destroy(sa->devx_ipsec_obj) | mlx5dv_devx_obj_destroy(sa->devx_key_obj);
}

struct mlx5dv_devx_obj *
sa_op(struct doca_ipsec_sa *sa)
{
	uint32_t in[MLX5_ST_SZ_DW(create_ipsec_obj_in)] = {0};
	uint32_t out[MLX5_ST_SZ_DW(general_obj_out_cmd_hdr)] = {0};
	struct ibv_context *ibv_ctx = priv_doca_dev_to_ibv_ctx(sa->dev);
	void *hdr = MLX5_ADDR_OF(create_ipsec_obj_in, in, general_obj_in_cmd_hdr);
	void *ipsec_obj = MLX5_ADDR_OF(create_ipsec_obj_in, in, ipsec_object);
	void *out_hdr = MLX5_ADDR_OF(query_ipsec_obj_out, out, general_obj_out_cmd_hdr);
	uint32_t icv_length;
	uint32_t pd_id = 0;
	struct doca_ipsec_sa_attrs *attr = &sa->attr;

	priv_doca_dev_pd_id_get(sa->dev, &pd_id);

	sa->devx_key_obj = op_encryption_key(sa);
	if (sa->devx_key_obj == NULL) {
		DOCA_DLOG_ERR("Failed to create ipsec key");
		return NULL;
	}

	MLX5_SET(general_obj_in_cmd_hdr, hdr, opcode, MLX5_CMD_OP_CREATE_GENERAL_OBJECT);
	MLX5_SET(general_obj_in_cmd_hdr, hdr, obj_type, MLX5_GENERAL_OBJ_TYPE_IPSEC_OFFLOAD);

	switch (attr->icv_length) {
	case DOCA_IPSEC_ICV_LENGTH_8:
		icv_length = MLX5_IPSEC_OBJECT_ICV_LEN_8B;
		break;
	case DOCA_IPSEC_ICV_LENGTH_12:
		icv_length = MLX5_IPSEC_OBJECT_ICV_LEN_12B;
		break;
	case DOCA_IPSEC_ICV_LENGTH_16:
		icv_length = MLX5_IPSEC_OBJECT_ICV_LEN_16B;
		break;
	default:
		DOCA_DLOG_ERR("Failed to create ipsec object unknown length = %d", attr->icv_length);
		goto error;
	}

	MLX5_SET(ipsec_obj, ipsec_obj, full_offload, attr->offload == DOCA_IPSEC_SA_OFFLOAD_FULL);
	MLX5_SET(ipsec_obj, ipsec_obj, esn_en, attr->esn_enabled);
	MLX5_SET(ipsec_obj, ipsec_obj, esn_overlap, attr->esn_overlap);
	MLX5_SET(ipsec_obj, ipsec_obj, icv_length, icv_length);
	MLX5_SET(ipsec_obj, ipsec_obj, aso_return_reg, 0);
	MLX5_SET(ipsec_obj, ipsec_obj, dekn, sa->key_id);
	MLX5_SET(ipsec_obj, ipsec_obj, salt, attr->key.aes_gcm.salt);
	MLX5_SET64(ipsec_obj, ipsec_obj, implicit_iv, attr->key.aes_gcm.implicit_iv);

	void *ipsec_aso_obj = MLX5_ADDR_OF(ipsec_obj, ipsec_obj, ipsec_aso);
	uint32_t win_size = 0;

	if (attr->antireplay.antireplay_enable) {
		switch (attr->antireplay.replay_win_sz) {
		case DOCA_IPSEC_REPLAY_WIN_SIZE_32:
			win_size = WIN_32BIT;
			break;
		case DOCA_IPSEC_REPLAY_WIN_SIZE_64:
			win_size = WIN_64BIT;
			break;
		case DOCA_IPSEC_REPLAY_WIN_SIZE_128:
			win_size = WIN_128BIT;
			break;
		case DOCA_IPSEC_REPLAY_WIN_SIZE_256:
			win_size = WIN_256BIT;
			break;
		default:
			DOCA_DLOG_ERR("Failed to create ipsec object unknown length = %d",
				      attr->antireplay.replay_win_sz);
			goto error;
		}
	}

	uint32_t mode = attr->antireplay.antireplay_enable;

	if (attr->antireplay.antireplay_enable)
		mode = attr->direction ? 1 : 2;

	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, valid, 1);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, mode, mode);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, window_size, win_size);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, soft_lifetime_arm, attr->antireplay.soft_lifetime_arm);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, hard_lifetime_arm, attr->antireplay.hard_lifetime_arm);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, remove_flow_enable, attr->antireplay.remove_flow_enable);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, ens_event_arm, attr->antireplay.esn_overlap_event_arm);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, remove_flow_packet_count, attr->antireplay.remove_flow_packet_count);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, remove_flow_soft_lifetime, attr->antireplay.remove_flow_soft_lifetime);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, mode_parameter, attr->antireplay.sn);
	MLX5_SET(ipsec_aso_obj, ipsec_aso_obj, replay_protection_window,
		 (uint32_t)(uintptr_t)attr->antireplay.replay_win_state);

	sa->devx_ipsec_obj = mlx5dv_devx_obj_create(ibv_ctx, in, sizeof(in), out, sizeof(out));
	if (sa->devx_ipsec_obj == NULL) {
		DOCA_DLOG_ERR("Failed to create ipsec object");
		goto error;
	}
	sa->id = MLX5_GET(general_obj_out_cmd_hdr, out_hdr, obj_id);

	return sa->devx_ipsec_obj;
error:
	mlx5dv_devx_obj_destroy(sa->devx_key_obj);
	return NULL;
}

struct mlx5dv_devx_obj *
sa_create(struct doca_ipsec_sa *sa)
{
	return sa_op(sa);
}

