/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include "ipsec_workq_ctx.h"
#include "doca_ipsec.h"
#include "ipsec_impl.h"
#include <doca_log.h>
#include <errno.h>
#include <priv_doca_malloc.h>
#include <private/common_defines.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

DOCA_LOG_REGISTER(DOCA_IPSEC_WORKQ);

struct doca_ipsec_sa *
doca_ipsec_sa_from_result(struct doca_event *ev)
{
	return (struct doca_ipsec_sa *)(ev->result.u64);
}

doca_error_t
ipsec_workq_job_submit(struct ipsec_workq_ctx *ctx, struct doca_job const *job)
{
	struct doca_ipsec_job *ipsec_job = priv_doca_malloc(1, sizeof(struct doca_ipsec_job));

	ipsec_job->type = job->type;
	ipsec_job->user_data = job->user_data;
	struct doca_ipsec_sa_create_job *job_create;
	struct doca_ipsec_sa_destroy_job *job_destroy;

	switch (job->type) {
	case DOCA_IPSEC_JOB_SA_CREATE:
		job_create =
			(struct doca_ipsec_sa_create_job *)container_of(job, struct doca_ipsec_sa_create_job, base);
		ipsec_job->attr = job_create->sa_attrs;
		break;
	case DOCA_IPSEC_JOB_SA_DESTROY:
		job_destroy =
			(struct doca_ipsec_sa_destroy_job *)container_of(job, struct doca_ipsec_sa_destroy_job, base);
		ipsec_job->sa = job_destroy->sa;
		break;
	}
	ipsec_job->next_job = NULL;

	if (ctx->jobs_head == NULL) {
		ctx->jobs_head = ipsec_job;
		ctx->jobs_tail = ipsec_job;
	} else {
		ctx->jobs_tail->next_job = ipsec_job;
		ctx->jobs_tail = ipsec_job;
	}
	return DOCA_SUCCESS;
}

doca_error_t
ipsec_workq_progress(struct ipsec_workq_ctx *ctx, struct doca_event *ev)
{
	struct doca_ipsec_job *job = ctx->jobs_head;
	struct mlx5dv_devx_obj *devx_obj;
	struct doca_ipsec_sa *sa;
	doca_error_t res = DOCA_SUCCESS;

	if (job == NULL)
		return DOCA_ERROR_INITIALIZATION;

	ev->type = job->type;
	ev->user_data = job->user_data;

	switch (job->type) {
	case DOCA_IPSEC_JOB_SA_CREATE:
		sa = priv_doca_malloc(1, sizeof(struct doca_ipsec_sa));
		if (sa == NULL) {
			res = DOCA_ERROR_NO_MEMORY;
		} else {
			sa->attr = job->attr;
			sa->dev = ctx->dev;
			ev->result.u64 = (uint64_t)sa;
			devx_obj = sa_create(sa);
			if (devx_obj == NULL) {
				free(sa);
				res = DOCA_ERROR_INVALID_VALUE;
			}
		}
		break;
	case DOCA_IPSEC_JOB_SA_DESTROY:
		if (sa_destroy(job->sa) != DOCA_SUCCESS)
			res = DOCA_ERROR_IO_FAILED;
		else {
			ev->result.u64 = DOCA_SUCCESS;
			free(job->sa);
		}
	}

	ctx->jobs_head = ctx->jobs_head->next_job;
	free(job);
	return res;
}

doca_error_t
ipsec_workq_create(struct ipsec_workq_ctx **ipsec_ctx, struct doca_dev *dev)
{
	struct ipsec_workq_ctx *ctx;

	if (ipsec_ctx == NULL || dev == NULL)
		return DOCA_ERROR_INVALID_VALUE;
	ctx = priv_doca_malloc(1, sizeof(struct ipsec_workq_ctx));
	if (ctx == NULL)
		return DOCA_ERROR_NO_MEMORY;

	ctx->dev = dev;
	ctx->jobs_head = NULL;
	ctx->jobs_tail = NULL;

	*ipsec_ctx = ctx;

	return DOCA_SUCCESS;
}

void
ipsec_workq_ctx_destroy(struct ipsec_workq_ctx *ctx)
{
	if (ctx == NULL)
		return;
	while (ctx->jobs_head != NULL) {
		struct doca_ipsec_job *tmp = ctx->jobs_head->next_job;

		priv_doca_free(ctx->jobs_head);
		ctx->jobs_head = tmp;
	}
	free(ctx);
	ctx = NULL;
}
